/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * vector3d header
 *
 */

#ifndef __SP_DIMENSION_VECTOR3D_H__
#define __SP_DIMENSION_VECTOR3D_H__


#include "spStandard.hpp"

#include <math.h>


namespace sp
{
namespace dim
{


template <typename T> class vector3d
{
    
    public:
        
        vector3d() : X(0), Y(0), Z(0)
        {
        }
        vector3d(T Size) : X(Size), Y(Size), Z(Size)
        {
        }
        vector3d(T NewX, T NewY, T NewZ) : X(NewX), Y(NewY), Z(NewZ)
        {
        }
        vector3d(const vector3d<T> &other)
        {
            X = other.X; Y = other.Y; Z = other.Z;
        }
        ~vector3d()
        {
        }
        
        /* === Operators - comparisions === */
        
        inline bool operator == (const vector3d<T> &other) const
        {
            return X == other.X && Y == other.Y && Z == other.Z;
        }
        inline bool operator != (const vector3d<T> &other) const
        {
            return X != other.X || Y != other.Y || Z != other.Z;
        }
        
        inline bool operator < (const vector3d<T> &other) const
        {
            return X < other.X && Y < other.Y && Z < other.Z;
        }
        inline bool operator > (const vector3d<T> &other) const
        {
            return X > other.X && Y > other.Y && Z > other.Z;
        }
        
        inline bool operator <= (const vector3d<T> &other) const
        {
            return X <= other.X && Y <= other.Y && Z <= other.Z;
        }
        inline bool operator >= (const vector3d<T> &other) const
        {
            return X >= other.X && Y >= other.Y && Z >= other.Z;
        }
        
        /* === Operators - addition, subtraction, division, multiplication === */
        
        inline vector3d<T>& operator ++ ()
        {
            ++X; ++Y, ++Z; return *this;
        }
        inline vector3d<T>& operator -- ()
        {
            --X; --Y, --Z; return *this;
        }
        
        inline vector3d<T> operator + (const vector3d<T> &other) const
        {
            return vector3d<T>(X + other.X, Y + other.Y, Z + other.Z);
        }
        inline vector3d<T>& operator += (const vector3d<T> &other)
        {
            X += other.X; Y += other.Y; Z += other.Z; return *this;
        }
        
        inline vector3d<T> operator - (const vector3d<T> &other) const
        {
            return vector3d<T>(X - other.X, Y - other.Y, Z - other.Z);
        }
        inline vector3d<T>& operator -= (const vector3d<T> &other)
        {
            X -= other.X; Y -= other.Y; Z -= other.Z; return *this;
        }
        
        inline vector3d<T> operator / (const vector3d<T> &other) const
        {
            return vector3d<T>(X / other.X, Y / other.Y, Z / other.Z);
        }
        inline vector3d<T>& operator /= (const vector3d<T> &other)
        {
            X /= other.X; Y /= other.Y; Z /= other.Z; return *this;
        }
        
        inline vector3d<T> operator / (T Size) const
        {
            return vector3d<T>(X / Size, Y / Size, Z / Size);
        }
        inline vector3d<T>& operator /= (T Size)
        {
            X /= Size; Y /= Size; Z /= Size; return *this;
        }
        
        inline vector3d<T> operator ^ (T Exp) const
        {
            return vector3d<T>( static_cast<T>(pow(X, Exp)), static_cast<T>(pow(Y, Exp)), static_cast<T>(pow(Z, Exp)) );
        }
        inline vector3d<T>& operator ^= (T Exp)
        {
            X = static_cast<T>(pow(X, Exp)); Y = static_cast<T>(pow(Y, Exp)); Z = static_cast<T>(pow(Z, Exp)); return *this;
        }
        
        inline vector3d<T> operator - () const
        {
            return vector3d<T>(-X, -Y, -Z);
        }
        
        // (Standard product: *)
        inline vector3d<T> operator * (const vector3d<T> &other) const
        {
            return vector3d<T>(X * other.X, Y * other.Y, Z * other.Z);
        }
        inline vector3d<T>& operator *= (const vector3d<T> &other)
        {
            X *= other.X; Y *= other.Y; Z *= other.Z; return *this;
        }
        
        inline vector3d<T> operator * (T Size) const
        {
            return vector3d<T>(X * Size, Y * Size, Z * Size);
        }
        inline vector3d<T>& operator *= (T Size)
        {
            X *= Size; Y *= Size; Z *= Size; return *this;
        }
        
        // (Cross/ Vector product: %)
        inline vector3d<T> operator % (const vector3d<T> &other) const
        {
            return vector3d<T>(
                Y*other.Z - other.Y*Z,
                Z*other.X - other.Z*X,
                X*other.Y - other.X*Y
            );
        }
        inline vector3d<T>& operator %= (const vector3d<T> &other)
        {
            f32 NewX = X, NewY = Y, NewZ = Z;
            
            NewX = Y*other.Z - other.Y*Z;
            NewY = Z*other.X - other.Z*X;
            NewZ = X*other.Y - other.X*Y;
            
            X = NewX; Y = NewY; Z = NewZ;
            
            return *this;
        }
        
        /* === Additional operators === */
        
        inline const T operator [] (s32 i) const
        {
            switch (i)
            {
                case 0: return X;
                case 1: return Y;
                case 2: return Z;
            }
            return 0;
        }
        
        inline T& operator [] (s32 i)
        {
            return *(&X + i);
        }
        
        /* === Extra functions === */
        
        inline T getLength() const
        {
            return static_cast<T>( sqrt(X*X + Y*Y + Z*Z) );
        }
        inline T getLengthSq() const
        {
            return X*X + Y*Y + Z*Z;
        }
        
        inline T getScalarProduct(const vector3d<T> &other) const // Scalar/ Dot product
        {
            return X*other.X + Y*other.Y + Z*other.Z;
        }
        inline vector3d<T> getCrossProduct(const vector3d<T> &other) const // Cross/ Vector product
        {
            return vector3d<T>(
                Y*other.Z - other.Y*Z,
                Z*other.X - other.Z*X,
                X*other.Y - other.X*Y
            );
        }
        
        inline T getAngle(const vector3d<T> &other) const
        {
            return static_cast<T>(acos( getScalarProduct(other) / (getLength()*other.getLength()) )*180.0/M_PI);
        }
        
        inline vector3d<T>& setInvert()
        {
            X = -X; Y = -Y; Z = -Z; return *this;
        }
        inline vector3d<T> getInvert() const
        {
            return vector3d<T>(-X, -Y, -Z);
        }
        
        inline vector3d<T>& setRound(s32 Precision)
        {
            s32 presicion = static_cast<s32>(pow(10, Precision));
            X = (T)((s32)(X*presicion)) / presicion;
            Y = (T)((s32)(Y*presicion)) / presicion;
            Z = (T)((s32)(Z*presicion)) / presicion;
            return *this;
        }
        inline vector3d<T> getRound(s32 Precision) const
        {
            s32 presicion = static_cast<s32>(pow(10, Precision));
            return vector3d<T>(
                (T)((s32)(X*presicion)) / presicion,
                (T)((s32)(Y*presicion)) / presicion,
                (T)((s32)(Z*presicion)) / presicion
            );
        }
        
        inline bool isEqual(const vector3d<T> &other, f32 Precision = 1.0e-6) const
        {
            return
                (X + Precision > other.X) && (X - Precision < other.X) &&
                (Y + Precision > other.Y) && (Y - Precision < other.Y) &&
                (Z + Precision > other.Z) && (Z - Precision < other.Z);
        }
        inline bool isEmpty() const
        {
            return isEqual(0);
        }
        
        inline void make2DProjection()
        {
            X =   X * f32(__Screen_Width /2) + __Screen_Width /2;
            Y = - Y * f32(__Screen_Height/2) + __Screen_Height/2;
            Z = 0;
        }
        inline void make2DProjection(f32 FOV)
        {
            X =   X / Z * FOV + __Screen_Width /2;
            Y = - Y / Z * FOV + __Screen_Height/2;
        }
        
        inline vector3d<T>& setAbs()
        {
            X = X > 0 ? X : -X;
            Y = Y > 0 ? Y : -Y;
            Z = Z > 0 ? Z : -Z;
            return *this;
        }
        inline vector3d<T> getAbs() const
        {
            return vector3d<T>(
                X > 0 ? X : -X,
                Y > 0 ? Y : -Y,
                Z > 0 ? Z : -Z
            );
        }
        
        inline vector3d<T>& normalize()
        {
            T n = X*X + Y*Y + Z*Z;
            
            if (n == 0 || n == 1)
                return *this;
            
            n = static_cast<T>(1.0 / sqrt(n));
            
            X *= n;
            Y *= n;
            Z *= n;
            
            return *this;
        }
        inline vector3d<T>& sign()
        {
            if (X > 0) X = 1; else if (X < 0) X = -1;
            if (Y > 0) Y = 1; else if (Y < 0) Y = -1;
            if (Z > 0) Z = 1; else if (Z < 0) Z = -1;
            return *this;
        }
        
        inline void set(T NewX, T NewY, T NewZ)
        {
            X = NewX; Y = NewY; Z = NewZ;
        }
        inline void get(T &NewX, T &NewY, T &NewZ) const
        {
            NewX = X; NewY = Y; NewZ = Z;
        }
        
        inline vector3d<T>& make(T NewX, T NewY, T NewZ)
        {
            X = NewX; Y = NewY; Z = NewZ;
            return *this;
        }
        
        inline void setLength(T Length)
        {
            normalize();
            *this *= Length;
        }
        
        inline T getDistanceFromSq(const vector3d<T> &other) const
        {
            return vector3d<T>(X - other.X, Y - other.Y, Z - other.Z).getLengthSq();
        }
        
        inline bool isBetweenPoints(const vector3d<T> &Begin, const vector3d<T> &End) const
        {
            T tmp = (End - Begin).getLengthSq();
            return getDistanceFromSq(Begin) < tmp && getDistanceFromSq(End) < tmp;
        }
        
        inline vector3d<T> getInterpolatedQuadratic(const vector3d<T> &v2, const vector3d<T> &v3, const T d) const
        {
            const T inv = (T)1.0 - d;
            const T mul0 = inv * inv;
            const T mul1 = (T)2.0 * d * inv;
            const T mul2 = d * d;
            
            return vector3d<T>(
                X * mul0 + v2.X * mul1 + v3.X * mul2,
                Y * mul0 + v2.Y * mul1 + v3.Y * mul2,
                Z * mul0 + v2.Z * mul1 + v3.Z * mul2
            );
        }
        
        inline void getArray3(T* Array) const
        {
            Array[0] = X; Array[1] = Y; Array[2] = Z;
        }
        inline void getArray4(T* Array) const
        {
            Array[0] = X; Array[1] = Y; Array[2] = Z; Array[3] = 1;
        }
        
        template <typename B> inline vector3d<B> cast() const
        {
            return vector3d<B>(static_cast<B>(X), static_cast<B>(Y), static_cast<B>(Z));
        }
        
        /* Members */
        
        T X, Y, Z;
        
};

typedef vector3d<s32> vector3di;
typedef vector3d<f32> vector3df;


} // /namespace dim

} // /namespace sp


#endif







// ================================================================================
