/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * main device header
 *
 */

#ifndef __SP_MAINDEVICE_H__
#define __SP_MAINDEVICE_H__


#include "spStandard.hpp"
#include "spDimension.hpp"
#include "spInputOutput.hpp"
#include "spVideoDriver.hpp"
#include "spSceneManager.hpp"
#include "spUserControl.hpp"
#include "spScriptExtension.hpp"
#include "spInputOutputOSInformator.hpp"
#include "spInputOutputFileSystem.hpp"

#include <time.h>
#include <string>

#if defined(LINUX)
#   include <gl/glx.h>
#   include <X11/Xlib.h>
#   include <X11/Xutil.h>
#endif


namespace sp
{


/*
 * Macros
 */

#if defined(WIN32)
#   ifndef WM_MOUSEWHEEL
#       define WM_MOUSEWHEEL    0x020A
#   endif
#   ifndef WHEEL_DELTA
#       define WHEEL_DELTA      120
#   endif
#endif


static const s32 DEF_SCRIPT_OBJECTCOUNT = 100;
static const s32 DEF_MULTISAMPLE        = 2;
static const s32 DEF_COLORDEPTH         = 32;
static const f32 DEF_FRAMERATE          = 70.0f;


/* 
 * Enumerations
 */

enum ETimeTypes
{
    TIME_SECOND = 0,
    TIME_MINUTE,
    TIME_HOURE,
    TIME_DAY,
    TIME_MONTH,
    TIME_YEAR
};


/*
 * Structures
 */

struct SDeviceFlags
{
    SDeviceFlags(
        bool ResizAble = false, bool Vsync = true, bool AntiAlias = false, s32 AntiAliasSamples = DEF_MULTISAMPLE)
        : isResizAble(ResizAble), isVsync(Vsync), isAntiAlias(AntiAlias), MultiSample(AntiAliasSamples)
    {
    }
    ~SDeviceFlags()
    {
    }
    
    /* Members */
    
    bool isResizAble;
    bool isVsync;
    bool isAntiAlias;
    s32 MultiSample;
};

static const SDeviceFlags DEVICEFLAG_HQ = SDeviceFlags(false, true, true, 16);


/*
 * SoftPixelDevice class
 */

class SP_EXPORT SoftPixelDevice
{
    
    public:
        
        SoftPixelDevice(
            video::EDriverTypes VideoDriverType,
            HWND hWndParent, dim::size2di Dimension, s32 ColorDepth, io::stringc Title,
            bool isFullScreen, bool isResizeAble, bool isVsync, bool isAntiAlias, s32 MultiSample
        );
        ~SoftPixelDevice();
        
        /* Video driver for the video buffer */
        video::VideoDriver* getVideoDriver();
        
        /* Scene manager for 3d rendering operations */
        scene::SceneManager* getSceneManager();
        
        /* Collision detector for collision detection and intersection arithmetic */
        scene::CollisionDetector* getCollisionDetector();
        
        /* User control for user input-/ output communication */
        UserControl* getUserControl();
        
        /* Filesystem for read-/ write files */
        io::FileLoader* getFileSystem();
        
        /* OS informator for informations about the operating-system */
        io::OSInformator* getOSInformator();
        
        /* Script loader (script extension) for simple file loading (textures, sounds, 3d models) */
        script::ScriptLoader* getScriptLoader(
            u32 TexMax = DEF_SCRIPT_OBJECTCOUNT, u32 SoundMax = DEF_SCRIPT_OBJECTCOUNT, u32 ModelMax = DEF_SCRIPT_OBJECTCOUNT
        );
        
        void deleteDevice(); // closes the screen, window and releases the device context
        bool updateEvent(); // calls the WindowCallback function
        
        #if defined(WIN32)
        void changeGFXMode(
            HWND hWndParent, dim::size2di Dimension, s32 ColorDepth, io::stringc Title,
            bool isFullScreen = false, bool isResizeAble = false
        );
        #elif defined(LINUX)
        
        #endif
        
        void resizeScreen(dim::size2di Dimension);
        inline dim::size2di getScreenDimension() // (Screen resolution)
        {
            return ScreenDimension_;
        }
        
        void clearDevice(); // (not usful for users)
        
        /* ===== Settings ===== */
        
        void setResizeMode(bool isResizeUpdate);
        
        void setAntiAlias(bool isAntiAlias, s32 MultiSample = DEF_MULTISAMPLE);
        
        void manipulateScreenSize(dim::size2di ScreenSize);
        
        void setFrameRate(f32 FrameRate = DEF_FRAMERATE); // FPS (Frames per seconds) 0 (Zero) is max. speed
        
        f32 getFPS();
        u64 getMilliseconds(); // returns the latest time from the booting-start in milliceconds
        u16 getTime(u8 Type);
        
        /* ===== Window management ===== */
        
        void setWindowTitle(io::stringc Title);
        io::stringc getWindowTitle() const;
        
        inline dim::size2di getWindowSize() const
        {
            return WindowDimension_;
        }
        bool hasWindowResized() const;
        
        bool isWindowActive() const;
        
        io::stringc getUserCharList() const;
        
        #if defined(WIN32)
        
        inline HWND getWindowHandle()
        {
            return hMainWnd_;
        }
        
        #elif defined(LINUX)
        
        #endif
        
        io::stringc getVersion();
        
        void sleep(u32 Milliseconds);
        void beep(u32 Milliseconds = 100, u32 Frequency = 1000);
        
    private:
        
        /* ===== Private functions ==== */
        
        void initEngine();
        
        #if defined(WIN32)
        bool openScreen(
            HWND hWndParent, dim::size2di Dimension, s32 ColorDepth, io::stringc Title,
            bool isFullScreen, bool isResizeAble, bool isAntiAlias
        );
        #elif defined(LINUX)
        
        #endif
        
        void closeScreen();
        
        void autoDetectVideoDriver();
        
        /* ===== Members ===== */
        
        video::EDriverTypes DriverType_;
        
        #if defined(WIN32)
        
        HWND hMainWnd_;
        MSG Msg_;
        
        c8 ClassName_[22]; // "SoftPixelEngine_Class"
        
        #elif defined(LINUX)
        
        Display* hDisplay_;
        XVisualInfo* Visual_;
        Window hMainWnd_;
        
        GLXWindow glxWin_;
        GLXContext Contect_;
        
        #endif
        
        bool isWindowOpened_, isResizAble_;
        bool isResizeUpdate_, hasWindowResized_;
        bool isAntiAlias_;
        
        dim::size2di ScreenDimension_, WindowDimension_;
        s32 ColorDepth_;
        io::stringc Title_;
        bool isFullScreen_;
        
        f32 FrameRate_;
        
        s32 MultiSampleFormat_, MultiSamplePixelFormat_;
        
};


/*
 * Static functions
 */

/* Creates the main device (only one device - screen - is possible) */
SP_EXPORT SoftPixelDevice* createDevice(
    video::EDriverTypes VideoDriverType,
    dim::size2di Dimension, s32 ColorDepth = DEF_COLORDEPTH, io::stringc Title = "",
    bool isFullScreen = false, SDeviceFlags Flags = SDeviceFlags()
);

/* Creates the main device (for window-based application) */
#if defined(WIN32)
SP_EXPORT SoftPixelDevice* createDeviceWnd(
    video::EDriverTypes VideoDriverType,
    HWND hWndParent, dim::size2di Dimension, s32 ColorDepth = DEF_COLORDEPTH, io::stringc Title = "",
    SDeviceFlags Flags = SDeviceFlags()
);
#elif defined(LINUX)

#endif

SP_EXPORT void deleteDevice();

#if defined(WIN32)
SP_EXPORT LRESULT CALLBACK spWindowCallback(HWND hWnd, UINT Message, WPARAM wParam, LPARAM lParam);
#elif defined(LINUX)

#endif


} // /namespace sp


#endif







// ================================================================================
