/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * Header for model loading 3DS-fileformat
 *
 */

#ifndef __SP_MODELLOADER_3DS_H__
#define __SP_MODELLOADER_3DS_H__


#include "spStandard.hpp"
#include "spInputOutput.hpp"
#include "spDimension.hpp"
#include "spModelLoaderRAW.hpp"

#include <io.h>
#include <vector>


namespace sp
{
namespace scene
{


class SP_EXPORT ModelLoader3DS : public ModelLoaderRAW
{
    
    public:
        
        ModelLoader3DS();
        ~ModelLoader3DS();
        
        Entity* loadModel(io::stringc &Filename, io::stringc &TexturePath);
        
    private:
        
        /* === Enumerations === */
        
        enum EModel3DSChunkIDTypes
        {
            MODEL_3DS_PRIMARY           = 0x4D4D,
            
            MODEL_3DS_VERSION           = 0x0002,
            MODEL_3DS_OBJECTINFO        = 0x3D3D,
            MODEL_3DS_OBJECT            = 0x4000,
            MODEL_3DS_MATERIAL          = 0xAFFF,
            MODEL_3DS_EDITKEYFRAME      = 0xB000,
            
            MODEL_3DS_OBJECT_MESH       = 0x4100,
            MODEL_3DS_OBJECT_VERTICES   = 0x4110,
            MODEL_3DS_OBJECT_FACES      = 0x4120,
            MODEL_3DS_OBJECT_MATERIAL   = 0x4130,
            MODEL_3DS_OBJECT_TEXCOORDS  = 0x4140,
            MODEL_3DS_OBJECT_MATRIX     = 0x4160,
            
            MODEL_3DS_MATNAME           = 0xA000,
            MODEL_3DS_MATDIFFUSE        = 0xA020,
            MODEL_3DS_MATMAP            = 0xA200,
            MODEL_3DS_MATMAPFILE        = 0xA300,
            
            MODEL_3DS_KEYF_TRACKINFO    = 0xB002,
            MODEL_3DS_KEYF_CURTIME      = 0xB009,
            MODEL_3DS_KEYF_BONENAME     = 0xB010,
            MODEL_3DS_KEYF_PIVOTPOINT   = 0xB013,
            MODEL_3DS_KEYF_BOUNDBOX     = 0xB014,
            MODEL_3DS_KEYF_BONEPOSITION = 0xB020,
            MODEL_3DS_KEYF_BONEROTATION = 0xB021,
            MODEL_3DS_KEYF_BONESCALE    = 0xB022,
            MODEL_3DS_KEYF_NODE_ID      = 0xB030,
        };
        
        /* === Structures === */
        
        struct SChunk3DS
        {
            u16 ID;
            u32 Length, ReadedBytes;
        };
        
        struct SMaterial3DS
        {
            c8 strName[255];
            c8 strFile[255];
            s8 color[3];
            s32 TextureID;
        };
        
        struct SFace3DS
        {
            u16 VertexIndex[3];
            u16 Flags;
        };
        
        struct SSurfaceObject3DS
        {
            c8 strName[255];
            
            s32 VerticesCounter;
            s32 FacesCounter;
            s32 MapCoordCounter;
            s32 MaterialID;
            bool hasTexture;
            
            SFace3DS* pFaces;
            
            dim::vector3df* VerticesList;
            dim::point2df* MapCoordList;
            
            dim::matrix4f TransformationMatrix;
        };
        
        struct SSkeletonBone3DS
        {
            /* General data */
            u32 ID, ParentBoneID;
            io::stringc Name;
            
            dim::vector3df Translation, Scale;
            dim::quaternion Quaternion;
            
            /* Animation container */
            std::vector<dim::vector3df> PositionList, ScaleList;
            std::vector<dim::quaternion> RotationList;
        };
        
        /* === Private functions === */
        
        void init();
        void clear();
        
        Entity* buildModel();
        void buildAnimation();
        
        s32 getString(FILE* CurFile, char* pBuffer);
        
        void readChunk(SChunk3DS* pChunk);
        
        void processNextChunk           (SChunk3DS* pPreviousChunk);
        void processNextObjectChunk     (SChunk3DS* pPreviousChunk, SSurfaceObject3DS* Object);
        void processNextMaterialChunk   (SChunk3DS* pPreviousChunk);
        void processNextKeyframeChunk   (SChunk3DS* pPreviousChunk);
        
        void readColorChunk     (SChunk3DS* pPreviousChunk, SMaterial3DS* Material);
        void readVerticesIndices(SChunk3DS* pPreviousChunk, SSurfaceObject3DS* Object);
        void readTexCoords      (SChunk3DS* pPreviousChunk, SSurfaceObject3DS* Object);
        void readVertices       (SChunk3DS* pPreviousChunk, SSurfaceObject3DS* Object);
        void readObjectMaterial (SChunk3DS* pPreviousChunk, SSurfaceObject3DS* Object);
        void readObjectMatrix   (SChunk3DS* pPreviousChunk, SSurfaceObject3DS* Object);
        
        void animCreateNewBone(u32 ID);
        void animSetBoneName(io::stringc Name);
        void animAddBonePositionList();
        void animAddBoneRotationList();
        void animAddBoneScaleList();
        
        /* === Members === */
        
        FILE* hFile_;
        
        SChunk3DS* CurrentChunk_;
        SChunk3DS* TempChunk_;
        
        u32 MaterialCounter_;
        u32 ObjectCounter_;
        bool hasAnimation_;
        
        std::vector<SMaterial3DS> MaterialList_;
        std::vector<SSurfaceObject3DS> ObjectList_;
        
        /* Animation building information */
        std::vector<dim::vector3df> BonePositionList_, BoneScaleList_;
        std::vector<dim::quaternion> BoneRotationList_;
        
        std::vector<SSkeletonBone3DS*> AnimBoneList_;
        SSkeletonBone3DS* CurBone_;
        
};


} // /namespace scene

} // /namespace sp


#endif





// ================================================================================
