/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * Header for entities
 *
 */

#ifndef __SP_SCENEMANAGER_ENTITY_H__
#define __SP_SCENEMANAGER_ENTITY_H__


#include "spStandard.hpp"
#include "spSceneManagerNode.hpp"
#include "spVideoDriverTexture.hpp"
#include "spInputOutputString.hpp"
#include "spVideoDriverMaterial.hpp"

#include <vector>
#include <list>


namespace sp
{
namespace scene
{


/*
 * Macros
 */

static const u32 SURFACE_IGNORE     = UINT_MAX;

static const u32 DEF_OCTTREE_LENGTH = 2;


/*
 * Pre-declarations
 */

class OctTree;
class Entity;
class Camera;
class SceneManager;


/*
 * Enumerations
 */

enum ESkyBoxTypes
{
    SKYBOX_NORMAL = 0,
    SKYBOX_SWAPED
};

enum ENormalsTypes
{
    NORMALS_SIMPLE = 0,
    NORMALS_SMOOTH
};


/*
 * Structures
 */

struct SNormalCoord
{
    u32 Index; // Vertex index
    dim::vector3df Normal;
    dim::vector3df Position;
};

struct SCollisionData
{
    u32 ID;
    dim::triangle3df Triangle;
    dim::vector3df Point;
    dim::vector3df Normal;
    Entity* Object;
};


/*
 * STexCoord structure
 */

struct STexCoord
{
    
    STexCoord()
        : U(0.0f), V(0.0f), W(0.0f)
    {
    }
    STexCoord(const STexCoord &TexCoord3D)
        : U(TexCoord3D.U), V(TexCoord3D.V), W(TexCoord3D.W)
    {
    }
    template <typename T> STexCoord(const T TexU)
        : U((f32)TexU), V(0.0f), W(0.0f)
    {
    }
    template <typename T> STexCoord(const T TexU, const T TexV)
        : U((f32)TexU), V((f32)TexV), W(0.0f)
    {
    }
    template <typename T> STexCoord(const T TexU, const T TexV, const T TexW)
        : U((f32)TexU), V((f32)TexV), W((f32)TexW)
    {
    }
    template <typename T> STexCoord(const T TexU, const T TexV, const T TexW, const T Factor)
        : U((f32)(TexU*Factor)), V((f32)(TexV*Factor)), W((f32)(TexW*Factor))
    {
    }
    template <typename T> STexCoord(const dim::point2d<T> TexCoord2D)
        : U((f32)TexCoord2D.X), V((f32)TexCoord2D.Y), W(0.0f)
    {
    }
    template <typename T> STexCoord(const dim::vector3d<T> TexCoord3D)
        : U((f32)TexCoord3D.X), V((f32)TexCoord3D.Y), W((f32)TexCoord3D.Z)
    {
    }
    STexCoord(const dim::quaternion TexCoord4D)
        : U(TexCoord4D.X*TexCoord4D.W), V(TexCoord4D.Y*TexCoord4D.W), W(TexCoord4D.Z*TexCoord4D.W)
    {
    }
    ~STexCoord()
    {
    }
    
    /* Functions */
    
    template <typename T> inline dim::point2d<T> vec2() const
    {
        return dim::point2d<T>((T)U, (T)V);
    }
    
    template <typename T> inline dim::vector3d<T> vec3() const
    {
        return dim::vector3d<T>((T)U, (T)V, (T)W);
    }
    
    /* Members */
    
    union
    {
        f32 U, X;
    };
    union
    {
        f32 V, Y;
    };
    union
    {
        f32 W, Z;
    };
    
};


/*
 * Entity class
 * entities represent the 3d objects/ meshes
 */

class SP_EXPORT Entity : public Node
{
    
    public:
        
        /* === Constructors === */
        
        Entity();
        Entity(video::Texture* TextureList[6], f32 Radius, u8 Type);
        Entity(video::Texture* hTextureHighMap, s32 Detail);
        ~Entity();
        
        /* === Textureing === */
        
        void addTexture(video::Texture* hTexture, u8 Layer = video::TEXTURE_IGNORE);
        void removeTexture(u8 Layer = 0);
        
        void doTextureTranslate(u8 Layer, dim::point2df Direction);
        void doTextureTransform(u8 Layer, dim::size2df Size);
        void doTextureTurn(u8 Layer, f32 Rotation);
        
        void setTexture(u8 Layer, video::Texture* hTexture);
        video::Texture* getTexture(u8 Layer = 0) const;
        
        void setTextureMatrix(u8 Layer, const dim::matrix4f &Matrix);
        dim::matrix4f getTextureMatrix(u8 Layer) const;
        
        void setTextureEnv(u8 Layer, video::ETextureEnvTypes Type);
        video::ETextureEnvTypes getTextureEnv(u8 Layer) const;
        
        void setMappingGen(u8 Layer, video::EMappingGenTypes Type);
        video::EMappingGenTypes getMappingGen(u8 Layer) const;
        
        void setMappingGenCoords(u8 Layer, s32 Coords);
        s32 getMappingGenCoords(u8 Layer) const;
        
        std::list<video::Texture*> getTextureList(u32 Surface = SURFACE_IGNORE) const;
        
        /* === Mesh control === */
        
        // !!! u32 Surface = SURFACE_IGNORE <- can be wasted and exchaged by "selectSurface" -> CurSurface_ !!!
        
        void updateNormals(u32 Surface = SURFACE_IGNORE);
        
        void doMeshTranslate(dim::vector3df Direction, u32 Surface = SURFACE_IGNORE);
        void doMeshTransform(dim::vector3df Size, u32 Surface = SURFACE_IGNORE);
        void doMeshTransform(dim::matrix4f Matrix, u32 Surface = SURFACE_IGNORE);
        void doMeshTurn(dim::vector3df Rotation, u32 Surface = SURFACE_IGNORE);
        
        void flipMesh(u32 Surface = SURFACE_IGNORE);
        void flipMesh(bool isXAxis, bool isYAxis, bool isZAxis, u32 Surface = SURFACE_IGNORE);
        
        void fitMesh(dim::vector3df Position, dim::vector3df Size);
        
        void flipTriangles(u32 Surface = SURFACE_IGNORE);
        
        void getBoundingBox(dim::vector3df &LeftBottomFrontCorner, dim::vector3df &RightTopBackCorner) const;
        dim::line3df getBoundingBox() const;
        
        /* === Surfaces === */
        
        u32 createSurface();
        
        void removeSurface(u32 Surface);
        
        void optimizeSurfaces();
        
        inline void selectSurface(u32 Surface)
        {
            CurSurface_ = Surface;
        }
        inline u32 getMeshSurfaceCount() const
        {
            return SurfaceList_.size();
        }
        
        inline void setSurfaceName(u32 Surface, io::stringc Name)
        {
            SurfaceList_[Surface].Name = Name;
        }
        inline io::stringc getSurfaceName(u32 Surface) const
        {
            return SurfaceList_[Surface].Name;
        }
        
        void clearTriangleList(u32 Surface);
        
        /* === Vertices (the most for the following functions are indebted with the 'selectSurface' function) === */
        
        u32 getMeshVerticesCount(u32 Surface = SURFACE_IGNORE) const;
        u32 getMeshTriangleCount(u32 Surface = SURFACE_IGNORE) const;
        
        void getMeshSurfaceByTriangle(u32 &TriangleIndex, u32 &SurfaceIndex) const;
        
        /* Coordinates */
        
        inline dim::vector3df getVertexCoord(u32 VertexIndex) const
        {
            return SurfaceList_[CurSurface_].VerticesList[VertexIndex + VertexCounter_].Position;
        }
        inline void setVertexCoord(u32 VertexIndex, dim::vector3df Position)
        {
            SurfaceList_[CurSurface_].VerticesList[VertexIndex + VertexCounter_].Position = Position;
        }
        
        /* Texture coordinates */
        
        inline STexCoord getVertexTexCoord(u32 VertexIndex, u8 Layer = 0) const
        {
            return SurfaceList_[CurSurface_].VerticesList[VertexIndex + VertexCounter_].TexCoord[Layer];
        }
        void setVertexTexCoord(u32 VertexIndex, STexCoord UVMapping, u8 Layer = video::TEXTURE_IGNORE);
        
        /* Color */
        
        inline video::color getVertexColor(u32 VertexIndex) const
        {
            return SurfaceList_[CurSurface_].VerticesList[VertexIndex + VertexCounter_].Color;
        }
        inline void setVertexColor(u32 VertexIndex, video::color Color)
        {
            SurfaceList_[CurSurface_].VerticesList[VertexIndex + VertexCounter_].Color = Color;
        }
        
        /* Normals */
        
        inline dim::vector3df getVertexNormal(u32 VertexIndex) const
        {
            return SurfaceList_[CurSurface_].VerticesList[VertexIndex + VertexCounter_].Normal;
        }
        inline void setVertexNormal(u32 VertexIndex, dim::vector3df Normal)
        {
            SurfaceList_[CurSurface_].VerticesList[VertexIndex + VertexCounter_].Normal = Normal;
        }
        
        /* Fog */
        
        inline f32 getVertexFog(u32 VertexIndex) const
        {
            return SurfaceList_[CurSurface_].VerticesList[VertexIndex + VertexCounter_].Fog;
        }
        inline void setVertexFog(u32 VertexIndex, f32 Fog)
        {
            SurfaceList_[CurSurface_].VerticesList[VertexIndex + VertexCounter_].Fog = Fog;
        }
        
        /* Edit vertices */
        
        u32 addVertex(
            dim::vector3df Position,
            STexCoord TexCoord      = STexCoord(),
            video::color Color      = video::color(255)
        );
        u32 addVertex(
            dim::vector3df Position,
            dim::vector3df Normal,
            STexCoord TexCoord      = STexCoord(),
            video::color Color      = video::color(255),
            f32 Fog                 = 0.0f
        );
        
        void removeVertex(u32 VertexIndex);
        
        /* Edit triangles */
        
        u32 addTriangle(u32 Vertex1, u32 Vertex2, u32 Vertex3);
        
        void removeTriangle(u32 TriangleIndex);
        
        /* Setting functions */
        
        void getTriangleVertices(u32 TriangleIndex, u32 Vertices[3]) const;
        
        inline void setVertexCounter(u32 VertexCounter)
        {
            VertexCounter_ = VertexCounter;
        }
        inline void addVertexCounter(u32 VertexCounter)
        {
            VertexCounter_ += VertexCounter;
        }
        
        void setVertexCoordArray(dim::vector3df* pVector);
        void getVertexCoordArray(dim::vector3df* pVector);
        
        void getTriangleArray(u32* pTriangle);
        
        void getTriangleList(dim::triangle3df* pTriangle, u32 &TriangelCount);
        void getTriangleList(dim::ptriangle3df* pTriangle, u32 &TriangelCount);
        
        void updateModeling(bool isNormalsUpdate = true); // must called after fucntions like 'addVertex(), setVertex...()' etc.
        void updateBuffersVBO(u32 Surface);
        
        /* === Collision detection === */
        
        void resetCollision(); // (Not usful for users)
        void simulateCollision(SCollisionData CollisionInformation); // (Not usful for users)
        
        bool detectCollision(u32 CollisionID);
        
        //bool detectIntersection(Entity* other);
        
        inline u32 getCollisionCount() const
        {
            return CollisionList_.size();
        }
        inline std::list<SCollisionData> &getCollisionList()
        {
            return CollisionList_;
        }
        
        inline dim::vector3df getCollisionPoint() const
        {
            return itCurColl_->Point;
        }
        inline dim::triangle3df getCollisionTriangle() const
        {
            return itCurColl_->Triangle;
        }
        inline Entity* getCollisionObject() const
        {
            return itCurColl_->Object;
        }
        dim::vector3df getCollisionNormal();
        
        inline void setCollisionUpdate(bool isCollUpdate)
        {
            isCollUpdate_ = isCollUpdate;
        }
        inline bool getCollisionUpdate() const
        {
            return isCollUpdate_;
        }
        
        /* === Oct-tree optimization === */
        
        void createOctTree(u32 Length = DEF_OCTTREE_LENGTH);
        void deleteOctTree();
        
        inline OctTree* getOctTreeRoot() const
        {
            return OctTreeRoot_;
        }
        
        /* === Material === */
        
        inline void setMaterial(video::SMeshMaterial Material)
        {
            Material_ = Material;
        }
        inline video::SMeshMaterial getMaterial() const
        {
            return Material_;
        }
        inline video::SMeshMaterial &getThisMaterial()
        {
            return Material_;
        }
        
        /* Render modes */
        inline void setWireframe(s32 Type)
        {
            setWireframe(Type, Type);
        }
        void setWireframe(s32 TypeFront, s32 TypeBack);
        
        inline void setRenderFace(s32 Face)
        {
            Material_.RenderFace = Face;
        }
        
        /* Shininess */
        void setShininess(f32 Intensity)
        {
            Material_.Shininess = 128.0f - Intensity*128;
        }
        f32 getShininess() const
        {
            return 1.0f - Material_.Shininess/128;
        }
        
        /* Lighting */
        inline void setLighting(bool isLighting)
        {
            Material_.isLighting = isLighting;
        }
        inline bool getLighting() const
        {
            return Material_.isLighting;
        }
        
        inline void setLightingColor(video::color Diffuse, video::color Ambient, video::color Specular, video::color Emission)
        {
            Material_.ColorDiffuse  = Diffuse;
            Material_.ColorAmbient  = Ambient;
            Material_.ColorSpecular = Specular;
            Material_.ColorEmission = Emission;
        }
        
        /* Fog effect */
        inline void setFog(bool isFog)
        {
            Material_.isFog = isFog;
        }
        inline bool getFog() const
        {
            return Material_.isFog;
        }
        
        /* Color material */
        inline void setColorMaterial(bool isColorMaterial)
        {
            Material_.isColorMaterial = isColorMaterial;
        }
        inline bool getColorMaterial() const
        {
            return Material_.isColorMaterial;
        }
        
        /* Order (ZBuffer) */
        inline void setOrder(s8 Order)
        {
            Order_ = Order;
            Material_.isZBuffer = (Order == 0);
        }
        inline s8 getOrder() const
        {
            return Order_;
        }
        
        /* Normals type */
        inline void setNormalsType(u8 NormalsType)
        {
            NormalsType_ = NormalsType;
        }
        inline u8 getNormalsType() const
        {
            return NormalsType_;
        }
        
        /* Visibility */
        inline void setVisible(bool isVisible)
        {
            isCollUpdate_ = isVisible_ = isVisible;
        }
        
        /* Blending mode */
        inline void setBlendingMode(u8 BlendingMode)
        {
            setBlending(BlendingMode, Material_.BlendSource, Material_.BlendTarget);
        }
        void setBlendingMode(s32 BlendSource, s32 BlendTarget);
        
        inline void setShader(video::Shader* hShader)
        {
            Material_.hShader = hShader;
        }
        inline video::Shader* getShader() const
        {
            return Material_.hShader;
        }
        
        /* Depth- and alpha modes */
        void setDepthMode(video::ESizeComparisionTypes Mode);
        void setAlphaMode(video::ESizeComparisionTypes Mode, f32 AlphaReference = -1.0f);
        
        /* === Other settings === */
        
        void setColor(video::color Color, u32 Surface = SURFACE_IGNORE, bool isForce = false);
        
        /* Matrix transformation function */
        void updateModelviewMatrix();
        
        /* === Rendering & copying === */
        
        Entity* copy();
        
        void render();
        
        void renderCelShading(dim::vector3df LightDirection = dim::vector3df(0, 0, 1));
        
        void drawNormals(const video::color Color = video::color(64, 64, 255), const f32 NormalLength = 1.0f, Camera* hCamera = 0);
        
    private:
        
        /* === Friends === */
        
        friend class SceneManager;
        friend bool sortEntitiesDefault(Entity* &obj1, Entity* &obj2);
        friend bool sortEntitiesFast(Entity* &obj1, Entity* &obj2);
        
        /* === Structures === */
        
        struct SMeshSurface3D
        {
            SMeshSurface3D()
                : ID(0), IndicesCount(0), TriangleCount(0)
            {
            }
            ~SMeshSurface3D()
            {
            }
            
            // Identity (VBO names & surface name)
            u32 ID;
            io::stringc Name;
            
            // Counters
            u32 IndicesCount;
            u32 TriangleCount;
            
            // Lists
            std::vector<video::SSurfaceTexture> TextureList;
            std::vector<video::SMeshVertex3D> VerticesList;
            std::vector<u32> TriangleIndicesList;
        };
        
        /* === Functions === */
        
        void updateCopyBuffersVBO();
        
        void createSkyBox(video::Texture* TextureList[6], f32 Radius, u8 Type);
        void createHighMap(video::Texture* hTextureHighMap, s32 Detail);
        
        void init();
        void clear();
        
        void calcSurfaceNoramls(u32 Surface);
        
        /* === Members === */
        
        /* Mesh information lists */
        
        std::vector<SMeshSurface3D> SurfaceList_;
        std::list<SCollisionData> CollisionList_;
        std::list<SCollisionData>::iterator itCurColl_;
        
        /* General information */
        
        f32 PerspectiveDistance_;
        u32 VertexCounter_, CurSurface_;
        u8 RenderMode_;
        u8 NormalsType_;
        s8 Order_;
        bool isCollUpdate_;
        
        /* Material information */
        video::SMeshMaterial Material_;
        
        /* Oct-tree-root-node */
        OctTree* OctTreeRoot_;
        
};


} // /namespace scene

} // /namespace sp


#endif







// ================================================================================
