/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * Header for nodes
 *
 */

#ifndef __SP_SCENEMANAGER_NODE_H__
#define __SP_SCENEMANAGER_NODE_H__


#include "spStandard.hpp"
#include "spDimension.hpp"
#include "spInputOutputStandard.hpp"
#include "spVideoDriverShader.hpp"
#include "spVideoDriverMaterial.hpp"
#include "spMath.hpp"

#include <gl/gl.h>
#include <gl/glext.h>
#include <list>


namespace sp
{
namespace scene
{


/*
 * Global members
 */

extern dim::matrix4f spProjectionMatrix;
extern dim::matrix4f spViewMatrix;
extern dim::matrix4f spWorldMatrix;
extern dim::matrix4f spTextureMatrix[video::MAX_COUNT_OF_TEXTURES];
extern dim::matrix4f spColorMatrix;


/* 
 * Enumerations
 */

enum ENodeTypes
{
    NODE_ENTITY = 0,    // 3D objects/ meshes
    NODE_CAMERA,        // Camera
    NODE_LIGHT,         // Lights
    NODE_SPRITE,        // Sprites/ particle
    NODE_TERRAIN,       // Terrains
    NODE_JOINT,         // simple node
};

enum EBlendingModeTypes
{
    BLENDING_NORMAL = 0,
    BLENDING_BRIGHTNESS,
    BLENDING_DARKNESS
};

enum ENodeOrderTypes
{
    ORDER_FOREGROUND = -1,
    ORDER_NORMAL = 0,
    ORDER_BACKGROUND = 1,
    
    ORDER_UNCUT = 1
};


/*
 * Node class, the parent class of all 3d scene obects
 */

class SP_EXPORT Node
{
    
    public:
        
        /* Public functions */
        
        Node(ENodeTypes Type);
        ~Node();
        
        /* Fast localisation */
        
        virtual inline void setPosition(dim::vector3df Position)
        {
            Position_ = Position;
        }
        virtual inline dim::vector3df getPosition() const
        {
            return Position_;
        }
        
        virtual inline void setRotation(dim::vector3df Rotation)
        {
            Rotation_ = Rotation;
        }
        virtual inline dim::vector3df getRotation() const
        {
            return Rotation_;
        }
        
        virtual inline void setScale(dim::vector3df Scale)
        {
            Scale_ = Scale;
        }
        virtual inline dim::vector3df getScale() const
        {
            return Scale_;
        }
        
        /* Detailed localisation */
        
        virtual void setPosition(dim::vector3df Position, bool isGlobal);
        
        virtual dim::vector3df getPosition(bool isGlobal) const;
        
        virtual void setRotation(dim::vector3df Rotation, bool isGlobal);
        
        virtual dim::vector3df getRotation(bool isGlobal) const;
        
        virtual void setScale(dim::vector3df Scale, bool isGlobal);
        
        virtual dim::vector3df getScale(bool isGlobal) const;
        
        /* Movement */
        
        virtual void doMove(dim::vector3df Direction);
        
        virtual inline void doTranslate(dim::vector3df Direction)
        {
            Position_ += Direction;
        }
        virtual inline void doTransform(dim::vector3df Size)
        {
            Scale_ += Size;
        }
        virtual inline void doTurn(dim::vector3df Rotation)
        {
            Rotation_ += Rotation;
        }
        
        /* Matrix localisation */
        
        virtual inline void setAutoMatrix(bool isAutoMatrix)
        {
            isAutoMatrix_ = isAutoMatrix;
        }
        virtual inline bool getAutoMatrix() const
        {
            return isAutoMatrix_;
        }
        
        virtual inline void setMatrix(const dim::matrix4f &Matrix)
        {
            Matrix_ = Matrix;
        }
        virtual inline dim::matrix4f &getMatrix()
        {
            return Matrix_;
        }
        
        virtual void updateMatrix() const;
        
        virtual dim::vector3df getDirection(dim::vector3df upVector = dim::vector3df(0, 0, 1)) const;
        
        virtual void lookAt(dim::vector3df Position, bool isGlobal = false);
        
        /* Visibility */
        
        virtual inline void setVisible(bool isVisible)
        {
            isVisible_ = isVisible;
        }
        virtual inline bool getVisible() const
        {
            return isVisible_;
        }
        
        /* Colors */
        
        virtual inline void setColor(video::color Color)
        {
            Color_ = Color;
        }
        virtual inline video::color getColor() const
        {
            return Color_;
        }
        
        /* Properties */
        
        virtual inline ENodeTypes getType() const
        {
            return Type_;
        }
        
        /* Parents */
        
        virtual inline void setParent(Node* ParentNode)
        {
            ParentNode_ = ParentNode;
        }
        virtual inline Node* getParent() const
        {
            return ParentNode_;
        }
        
        /* Name identification */
        
        virtual inline void setName(io::stringc Name)
        {
            Name_ = Name;
        }
        virtual inline io::stringc getName() const
        {
            return Name_;
        }
        
        /* User data */
        
        virtual inline void setUserData(void* Data)
        {
            UserData_ = Data;
        }
        virtual inline void* getUserData() const
        {
            return UserData_;
        }
        
        /* Parent system */
        
        virtual void updateParent() const;
        
        virtual dim::matrix4f getGlobalLocation() const;
        virtual dim::matrix4f getGlobalLocationInverse() const;
        
        virtual void updateGlobalLocation() const;
        virtual void updateGlobalLocationInverse() const;
        
        /* Copying attributes */
        
        virtual void copyAttributes(Node* OtherNode);
        
        /* Matrix transformation */
        
        virtual void updateModelviewMatrix();
        
    protected:
        
        /* Protected functions */
        
        virtual void setBlending(s32 Mode, s32 &Blend1, s32 &Blend2);
        
        /* Members */
        
        ENodeTypes Type_;
        io::stringc Name_;
        
        dim::matrix4f Matrix_;
        
        dim::vector3df Position_;
        dim::vector3df Rotation_;
        dim::vector3df Scale_;
        
        video::color Color_;
        bool isVisible_, isAutoMatrix_;
        
    private:
        
        /* Members */
        
        Node* ParentNode_;
        
        void* UserData_;
        
};


} // /namespace scene

} // /namespace sp


#endif







// ================================================================================
