/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * Header for the "Ray" Software video driver
 *
 */

#ifndef __SP_SOFTWAREDRIVER_H__
#define __SP_SOFTWAREDRIVER_H__


#include "spStandard.hpp"
#include "spVideoDriver.hpp"
#include "spSoftwareTexture.hpp"

#include "spRayTexturedGeometry.h"


namespace sp
{
namespace video
{


/*
 * ========== The "Ray" Software video driver class ==========
 */


class SP_EXPORT SoftwareDriver : public VideoDriver
{
    
    public:
        
        SoftwareDriver();
        ~SoftwareDriver();
        
        /*
         * === Video driver identity ===
         */
        
        io::stringc getRenderer()           const;
        io::stringc getVersion()            const;
        io::stringc getVendor()             const;
        
        bool supportAntiAlias()             const;
        bool supportMultitex()              const;
        bool supportVBO()                   const;
        bool supportAnisotropicFiltering()  const;
        
        s32 getMultitexCount()              const;
        s32 getMaxAnisotropicFilter()       const;
        
        /*
         * === User controll functions ===
         */
        
        /* Video buffer manipulation */
        void clearBuffers();
        void flipBuffers();
        
        
        /*
         * === Setting-/ getting functions ===
         */
        
        void setShadeMode(EShadeModeTypes ShadeMode);
        
        /* Color manipulation */
        void setClearColor(color Color);
        
        /*
         * === Context functions (for initializing the engine) ===
         */
        
        void initRenderer();
        
        void changeDisplaySettings(bool isFullScreen, dim::size2di Dimension, s32 ColorDepth);
        
        bool initScreen(HWND hWnd, bool isAntiAlias, s32 MultiSamplePixelFormat, s32 ColorDepth);
        void clearScreen(HWND &hWnd);
        
        
        /*
         * ========== Rendering 3D scenes ==========
         */
        
        /* Material states */
        void setMaterialStatesTriangleList(SMeshMaterial &Material, dim::vector3df &Rotation);
        
        void setMaterialStatesSprite(SMeshMaterial &Material);
        
        /* Render 3d objects */
        void drawTriangleList(
            u32 &BufferID, u32* IndicesList, u32 &IndicesCount, std::vector<SSurfaceTexture> &TextureList
        );
        
        void drawTriangleList(u32 &BufferID, u32* IndicesList, u32 &IndicesCount);
        
        void drawSprite(
            Texture* &hTexture, f32 &RotationZ, video::color &Color, dim::vector3df &Scale, dim::point2df &BasePosition
        );
        
        /* Update light */
        void updateLight(
            u32 &LightID, scene::ELightTypes &LightingType, bool &isVolumetric,
            dim::vector3df Direction, f32 &NearSpot, f32 &FarSpot,
            f32 &AttenuationConstant, f32 &AttenuationLinear, f32 &AttenuationQuadratic
        );
        
        /* Vertex-buffer-objects (VBO) */
        void createVertexBuffer(u32 &BufferID);
        void deleteVertexBuffer(u32 &BufferID);
        void updateVertexBuffer(u32 &BufferID, SMeshVertex3D* pVerticesList, s32 VerticesCount);
        
        /* Disable render states */
        void disableTriangleListStates();
        void disable3DRenderStates();
        void disableTexturing();
        
        void setDefaultAlphaBlending();
        
        void enableBlending();
        void disableBlending();
        
        /* Special effects actor */
        scene::effect::SpecialEffectsActor* createSpecialEffectsActor();
        
        
        /*
         * === Stack operation functions ===
         */
        
        void pushViewport();
        void popViewport();
        
        
        /*
         * === Texturing ===
         */
        
        void bindTexture(u32 ID);
        void unbindTexture();
        
        void setTextureSubImage(u32 ID, s32 X, s32 Y, s32 Width, s32 Height, const u8* ImageData);
        
        
        /*
         * ========== Drawing 2D objects ==========
         */
        
        /* === General setting function === */
        
        /* Begin & end 2d drawing */
        void beginDrawing2D(bool isSettingViewport = false);
        void endDrawing2D();
        
        /* Begin & end 3d drawing */
        void beginDrawing3D(bool isLighting = false, bool isFog = false);
        void endDrawing3D();
        
        /* begin & end alpha blending */
        void beginAlphaBlending(u8 Type = BLENDING_BRIGHTNESS);
        void endAlphaBlending();
        
        /* Begin & end masking */
        void beginMasking(u8 Type);
        void endMasking();
        
        /* Begin & end texture drawing */
        void beginTextureDrawing(Texture* hTexture, s32 RenderMode = TEXTURE_SMOOTH, u32 TextureDataType = TEXTURE_RGB);
        void endTextureDrawing(bool isUpdateImageData = true);
        
        /* Drawing rect (scissoring) */
        void beginDrawingRect(dim::rect2di Dimension);
        void beginDrawingRect(dim::point2di Position, dim::size2di Size);
        void endDrawingRect();
        
        /* Drawing area (viewport) */
        void setDrawingArea(dim::rect2di Dimension);
        void setDrawingArea(dim::point2di Position, dim::size2di Size);
        
        /* Size of lines & points */
        void setLineSize(s32 Size = 1);
        void setPointSize(s32 Size = 1);
        
        void setSolidMode(bool isSolidMode);
        
        /* === Texture drawing === */
        
        void draw2DImage(
            Texture* hTexture, dim::point2di Position,
            color Color = color(255), bool isTransFilterUsing = true
        );
        
        void draw2DImage(
            Texture* hTexture, dim::rect2di Position, dim::rect2df Clipping = dim::rect2df(0.0f, 0.0f, 1.0f, 1.0f),
            color Color = color(255), bool isTransFilterUsing = true
        );
        
        void draw2DImage(
            Texture* hTexture, dim::point2di Position, f32 Rotation, f32 Radius,
            color Color = color(255), bool isTransFilterUsing = true
        );
        
        void draw2DImage(
            Texture* hTexture,
            dim::rect2di Position,
            dim::rect2df Clipping,
            bool isTransFilterUsing,
            f32 Rotation,
            dim::point2df RotationPoint = dim::point2df(0.5f, 0.5f),
            color lefttopColor = color(255),
            color righttopColor = color(255),
            color rightbottomColor = color(255),
            color leftbottomColor = color(255)
        );
        
        void draw2DImage(
            Texture* hTexture,
            dim::point2di lefttopPosition,
            dim::point2di righttopPosition,
            dim::point2di rightbottomPosition,
            dim::point2di leftbottomPosition,
            dim::point2df lefttopClipping = dim::point2df(0.0f, 0.0f),
            dim::point2df righttopClipping = dim::point2df(1.0f, 0.0f),
            dim::point2df rightbottomClipping = dim::point2df(1.0f, 1.0f),
            dim::point2df leftbottomClipping = dim::point2df(0.0f, 1.0f),
            bool isTransFilterUsing = true,
            color lefttopColor = color(255),
            color righttopColor = color(255),
            color rightbottomColor = color(255),
            color leftbottomColor = color(255)
        );
        
        /* === Primitive drawing === */
        
        color getPixel(dim::point2di Position);
        
        void draw2DPixel(dim::point2di Position, color Color = color(255));
        
        void draw2DLine(dim::point2di Position1, dim::point2di Position2, color Color = color(255));
        void draw2DLine(dim::point2di Position1, dim::point2di Position2, color Color1, color Color2);
        void draw2DLine(dim::point2di Position1, dim::point2di Position2, color Color, s32 DotLength);
        
        void draw2DRectangle(dim::rect2di Position, color Color = color(255), EDrawingModeTypes Type = DRAW_SOLID);
        void draw2DRectangle(
            dim::rect2di Position, color lefttopColor, color righttopColor,
            color rightbottomColor, color leftbottomColor, EDrawingModeTypes Type = DRAW_SOLID
        );
        
        void draw2DRoundRectangle(dim::rect2di Position, dim::size2di CornerRadius, color Color = video::color(255));
        
        void draw2DCircle(dim::point2di Position, s32 Radius, color Color = color(255));
        
        void draw2DEllipse(dim::point2di Position, dim::size2di Radius, color Color = color(255));
        void draw2DEllipse(
            dim::point2di Position, dim::size2di Radius, f32 RotationDegree,
            color Color = color(255), bool isSolid = false
        );
        
        void draw2DPolygon(SPrimitiveVertex* VerticesList, u32 AryCount, EDrawingModeTypes Type = DRAW_SOLID);
        void draw2DPolygonImage(
            Texture* hTexture, SPrimitiveVertex* VerticesList, u32 AryCount, EDrawingModeTypes Type = DRAW_SOLID
        );
        
        /* === Extra drawing functions === */
        
        void draw2DArc(
            dim::point2di Position, dim::size2di Radius, s32 RadialDegree1, s32 RadialDegree2,
            f32 RotationDegree = 0.0f, color Color = color(255)
        );
        
        void draw2DBezier(
            dim::point2di Position1, dim::point2di Position2, dim::point2di Radial1, dim::point2di Radial2,
            color Color = color(255), f32 Detail = DEF_BEZIER_DETAIL
        );
        
        void draw2DSpline(
            math::Spline2D &Spline, s32 Detail = DEF_SPLINE_DETAIL,
            dim::point2di Pos = dim::point2di(), color Color = color(255)
        );
        
        void draw2DFloodFill(dim::point2di Position, color FillColor, color BorderColor = color()); // alpha(0) is someone border
        
        /* === 3D drawing functions === */
        
        void draw3DPixel(dim::vector3df Position, color Color = color(255), scene::Camera* hCamera = 0);
        void draw3DLine(
            dim::vector3df Position1, dim::vector3df Position2,
            color Color = color(255), scene::Camera* hCamera = 0
        );
        void draw3DLine(
            dim::vector3df Position1, dim::vector3df Position2,
            color Color1, color Color2, scene::Camera* hCamera = 0
        );
        void draw3DEllipse(
            dim::vector3df Position, dim::vector3df Rotation = dim::vector3df(),
            dim::size2df Radius = dim::size2df(0.5f), color Color = color(255), scene::Camera* hCamera = 0
        );
        void draw3DTriangle(
            Texture* hTexture, dim::triangle3df Triangle,
            dim::vector3df Position = dim::vector3df(), dim::vector3df Rotation = dim::vector3df(),
            dim::vector3df Scale = dim::vector3df(1.0f), color Color = color(255), scene::Camera* hCamera = 0
        );
        void draw3DBox(
            dim::vector3df Corner1, dim::vector3df Corner2,
            dim::vector3df Position = dim::vector3df(), dim::vector3df Rotation = dim::vector3df(),
            dim::vector3df Scale = dim::vector3df(1.0f), color Color = color(255), scene::Camera* hCamera = 0
        );
        
        void draw3DBone(scene::Entity* Obj, dim::matrix4f &BoneTransformation, scene::Camera* hCamera);
        
        /*
         * ========== Texture loading & creating ==========
         */
        
        /* Most important function (loads the format which are integrated) */
        Texture* loadTexture(io::stringc Filename);
        
        /* Check the correctness of the texture */
        bool isTexture(Texture* hTexture);
        
        /* Copy texture */
        Texture* copyTexture(Texture* hTexture);
        
        /* Create new texture */
        Texture* createTexture(
            dim::size2di Dimension, u32 TextureDataType = TEXTURE_RGB, io::stringc Filename = ""
        );
        
        /* Create new texture (with own image data) */
        Texture* createTexture(
            dim::size2di Dimension, u8* ImageData, u32 TextureDataType = TEXTURE_RGB, io::stringc Filename = ""
        );
        
        /* Create screenshot */
        Texture* createScreenShot(dim::point2di Position = 0, dim::size2di Size = 0);
        void createScreenShot(Texture* hTexture, dim::point2di Position = 0);
        
        
        /*
         * ========== Font loading & drawing ==========
         */
        
        Font* loadFont(
            io::stringc FontName = "",
            dim::size2di FontSize = dim::size2di(__SP_IGNORE, __SP_IGNORE),
            s32 TypeFlag = 0,
            s32 RotationAngle = 0,
            bool isSymbolUsing = false,
            bool is3DSpaceUsing = false,
            f32 FontDepth = 0.1f,
            f32 DetailDeviation = 0.1f,
            EDrawingModeTypes Wireframe = DRAW_SOLID
        );
        
        Font* loadFont(
            io::stringc FontName,
            s32 FontSize,
            s32 TypeFlag = 0,
            s32 RotationAngle = 0,
            bool isSymbolUsing = false,
            bool is3DSpaceUsing = false,
            f32 FontDepth = 0.1f,
            f32 DetailDeviation = 0.1f,
            EDrawingModeTypes Wireframe = DRAW_SOLID
        );
        
        Font* loadFont(Texture* hTextureFont, dim::size2di FontSize, dim::size2di ElementCount = dim::size2di(16, 16));
        
        void deleteFont(Font* hFont);
        
        /* Drawing text */
        void draw2DText(Font* hFont, dim::point2di Position, io::stringc Text, color Color = color(255));
        
        
        /*
         * === Matrix controll ===
         */
        
        void updateModelviewMatrix();
        
        void setProjectionMatrix(const dim::matrix4f &Matrix);
        void setTextureMatrix   (const dim::matrix4f &Matrix, u8 TextureLayer = 0);
        
        
        /*
         * === Other renderer option functions ===
         */
        
        void setViewport(s32 X, s32 Y, s32 Width, s32 Height);
        
        void setClippingRange(f32 Near, f32 Far);
        
    private:
        
        friend class SoftwareTexture;
        
        /* === Typedefinitions === */
        
        typedef void (*PFNWIREFRAMEPROC)(void);
        
        /* === Functions === */
        
        /* Building mipmapps */
        void build2DMipmaps(s32 InternalFormat, s32 Format, s32 Width, s32 Height, const u8* ImageData);
        
        /* === Inline functions === */
        
        inline void setRayRenderState(RAYenum Mode, bool Enable)
        {
            if (Enable)
                rayEnable(Mode);
            else
                rayDisable(Mode);
        }
        
        /* === Members === */
        
        HGLRC hRC_;
        
        /* === Render informations === */
        
        s32 Wireframe_;
        
        /* Vertex buffer objects (pseudo VBO) */
        std::vector<SMeshVertex3D*> VertexBufferList_;
        std::vector<u32*> VertexBufferIDList_;
        
};


/* 
 * Static functions for software rendering
 */

SP_EXPORT void raySetTextureFilter(ETextureRenderModeTypes Filter);

SP_EXPORT void raySetMultiTexturing(u32 MaxTextureLayer);


} // /namespace video

} // /namespace sp


#endif






// ================================================================================
