/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * Header for all stencil operations (stencil shadow, stencil mirror)
 *
 */

#ifndef __SP_STENCILMANAGER_H__
#define __SP_STENCILMANAGER_H__


#include "spStandard.hpp"
#include "spMemoryManagement.hpp"
#include "spSceneManagerEntity.hpp"
#include "spSceneManagerCamera.hpp"
#include "spSceneManagerLight.hpp"
#include "spInputOutputStandard.hpp"
#include "spMath.hpp"

#include <list>


namespace sp
{
namespace scene
{


/*
 * Macros
 */

#define __SP_STENCIL_POINTS_COUNT   10000
#define __SP_STENCIL_PLANES_COUNT   10000


/*
 * Enumerations
 */

enum EMirrorPlaneTypes
{
    MIRROR_X = 0,
    MIRROR_Y,
    MIRROR_Z
};


/*
 * Structures
 */

struct SShadowLightSource
{
    SShadowLightSource(
        scene::Node* LightSource, video::color Color = video::color(0, 0, 0, 128),
        f32 InitShadowLength = 100.0f, f32 InitRadius = -1.0f, bool InitVolumetric = false)
        : Object(LightSource), ShadowColor(Color), ShadowLength(InitShadowLength),
        Radius(InitRadius), Volumetric(InitVolumetric), Visible(true)
    {
    }
    ~SShadowLightSource()
    {
    }
    
    /* Members */
    
    scene::Node* Object;
    video::color ShadowColor;
    f32 ShadowLength;
    f32 Radius;
    bool Volumetric;
    bool Visible;
    
};


/*
 * CastCloudObject class
 */

// Pre-decleration
class StencilManager;
class CastCloudObject;

typedef void (*PFNSHADOWINTENSITYCALLBACK)(f32 &Intensity, const CastCloudObject* Mesh, const SShadowLightSource* LightSource);


class SP_EXPORT CastCloudObject
{
    
    public:
        
        CastCloudObject(scene::Entity* Object);
        ~CastCloudObject();
        
        /* Functions */
        
        void addLightSource(SShadowLightSource* LightSource);
        void removeLightSource(SShadowLightSource* LightSource = 0);
        
        /* Inline functions */
        
        inline scene::Entity* getMesh() const
        {
            return Object_;
        }
        
        inline void setShadowIntensity(f32 Intensity)
        {
            ShadowIntensity_ = Intensity;
        }
        inline f32 getShadowIntensity() const
        {
            return ShadowIntensity_;
        }
        
        inline void setShadowIntensityCallback(PFNSHADOWINTENSITYCALLBACK pFunc)
        {
            ShadowIntensityCallback_ = pFunc;
        }
        
    private:
        
        friend class StencilManager;
        
        /* Strutcures */
        
        struct SPlaneEq
        {
            f32 a, b, c, d;
        };
        
        struct SPlane
        {
            u32 p[3], Neigh[3];
            SPlaneEq PlaneEq;
            bool isVisible;
        };
        
        /* Members */
        
        scene::Entity* Object_;
        
        f32 ShadowIntensity_;
        
        SPlane pPlanes_[10000];
        //SPlane* pPlanes_;
        s32 CountOfPlanes_;
        
        dim::vector3df pMeshVertices_[10000];
        //dim::vector3df* pMeshVertices_;
        s32 CountOfMeshVertices_;
        
        dim::vector3df* pShadowVertices_;
        s32 CountOfShadowVertices_;
        
        std::vector<SShadowLightSource*> LightSourcesList_;
        
        PFNSHADOWINTENSITYCALLBACK ShadowIntensityCallback_;
        
};


/*
 * StencilManager class
 */

class SP_EXPORT StencilManager
{
    
    public:
        
        StencilManager();
        ~StencilManager();
        
        /* === Stencil shadows === */
        
        CastCloudObject* addEntityCastCloud(Entity* hEntity);
        void removeEntityCastCloud(Entity* hEntity);
        
        void addLightSource(SShadowLightSource* LightSource);
        void removeLightSource(SShadowLightSource* LightSource);
        
        void updateStencilShadow(Entity* hEntity);
        
        void renderStencilShadows(Camera* hCamera);
        
        /* === Stencil clipping mirrors === */
        
        void addMirrorPlane(Entity* hEntity, u8 Type);
        void removeMirrorPlane(Entity* hEntity);
        
        void addMirrorObject(Entity* hEntity);
        void removeMirrorObject(Entity* hEntity);
        
        void updateMirrorPlane(Entity* hEntity);
        
        void renderStencilMirrors(Camera* hCamera);
        
        /* Inline functions */
        
        inline void setMultiShadows(bool MultiShadows)
        {
            MultiShadows_ = MultiShadows;
        }
        inline bool getMultiShadows() const
        {
            return MultiShadows_;
        }
        
        inline void setSingleShadowColor(video::color ShadowColor)
        {
            SingleShadowColor_ = ShadowColor;
        }
        inline video::color getSingleShadowColor() const
        {
            return SingleShadowColor_;
        }
        
    private:
        
        /* Functions */
        
        void updateConnectivity(CastCloudObject* Object);
        void updateCalculationPlanes(CastCloudObject* Object);
        void updateShadowVolume(CastCloudObject* Object, const dim::vector3df LightPos, f32 ShadowLength);
        
        /* Members */
        
        std::vector<CastCloudObject*> ObjectList_;
        
        bool MultiShadows_;
        video::color SingleShadowColor_;
        
};


} // /namespace scene

} // /namespace sp


#endif





// ================================================================================
