/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * Header for texture managing
 *
 */

#ifndef __SP_TEXTUREMANAGER_H__
#define __SP_TEXTUREMANAGER_H__


#include "spStandard.hpp"
#include "spInputOutput.hpp"
#include "spVideoDriverColor.hpp"
#include "spMath.hpp"

#include <cstdio>


namespace sp
{
namespace video
{


enum ETextureImageDataTypes
{
    /* 1 component */
    TEXTURE_INDEX,
    TEXTURE_STENCIL,
    TEXTURE_DEPTH,
    TEXTURE_RED,
    TEXTURE_GREEN,
    TEXTURE_BLUE,
    TEXTURE_ALPHA,
    TEXTURE_GRAY,
    
    /* 2 components */
    TEXTURE_GRAYALPHA,
    
    /* 3 components */
    TEXTURE_RGB,
    //TEXTURE_BGR,
    
    /* 4 components */
    TEXTURE_RGBA,
    //TEXTURE_BGRA,
};

typedef struct STextureRAWInfo
{
    STextureRAWInfo();
    
    s32 Width;          // Dimension: width
    s32 Height;         // Dimension: height
    
    s32 bpp;            // bits per pixel
    s32 Format;         // TEXTURE_RGB, TEXTURE_RGBA ...
    s32 InternalFormat; // 1/ 2/ 3/ 4
    
    u32 ID;             // Identification number
    
    u8* ImageData;      // Image RAW data
}
textureRAWInfo;


/*
 * ======= Texture manager (for pixel & buffer manipulations) =======
 */

class SP_EXPORT TextureManager
{
    
    public:
        
        TextureManager();
        ~TextureManager();
        
        /* Flips the color values (BGR -> RGB) */
        static void flipImageColors(s32 Width, s32 Height, u8* ImageData, s32 InternalFormat = 3);
        
        /* Flips the image data on the x-axis */
        static void flipImageDataX(s32 Width, s32 Height, u8* ImageData, s32 InternalFormat = 3);
        
        /* Flips the image data on the y-axis */
        static void flipImageDataY(s32 Width, s32 Height, u8* ImageData, s32 InternalFormat = 3);
        
        /* Scales the image to a new size */
        static void scaleImageData(u8* &ImageData, s32 Width, s32 Height, s32 NewWidth, s32 NewHeight, s32 InternalFormat = 3);
        
        /* Scales the image to a half size (smooth) */
        static void scaleHalfImageData(u8* &ImageData, s32 Width, s32 Height, s32 InternalFormat = 3);
        
        /* Converts the image data formats (e.g. RGB -> RGBA) */
        static void convertImageData(u8* &ImageData, s32 Width, s32 Height, s32 Type, s32 NewType);
        
        /* Blurs the image data (for mipmap levels) */
        static void blurImageData(u8* &ImageData, s32 Width, s32 Height, s32 InternalFormat = 3);
        
        /* Universal function (calls the other needed functions) */
        static void changeImageData(u8* &ImageData, s32 Width, s32 Height, s32 Type, s32 NewWidth, s32 NewHeight, s32 NewType);
        
        /* Get the internal format to the specified type */
        static s32 getInternalFormat(s32 Type);
        
        /* Get the correct image size (2^x) */
        static bool checkImageSize(dim::size2di &InputSize);
        
        /* Get the count of mipmap levels */
        static s32 getMipmapLevelsCount(s32 Width, s32 Height);
        
};


/*
 * ======= Texture load RAW (parent class for all image file format loader) =======
 */

class SP_EXPORT TextureLoaderRAW
{
    
    public:
        
        virtual textureRAWInfo* loadImageData() = 0;
        
        virtual ~TextureLoaderRAW();
        
    protected:
        
        TextureLoaderRAW(io::stringc &Filename);
        
        inline bool isFileError() const
        {
            return isFileError_;
        }
        
        /* Members */
        
        FILE* hFile_;
        bool isFileError_;
        io::stringc Filename_;
        
};


/*
 * ======= Texture save RAW (parent class for all image file format saver) =======
 */

class SP_EXPORT TextureSaverRAW
{
    
    public:
        
        virtual bool saveImageData(textureRAWInfo* texture) = 0;
        
        virtual ~TextureSaverRAW();
        
    protected:
        
        TextureSaverRAW(io::stringc &Filename);
        
        inline bool isFileError() const
        {
            return isFileError_;
        }
        
        /* Members */
        
        FILE* hFile_;
        bool isFileError_;
        io::stringc Filename_;
        
};



} // /namespace video

} // /namespace sp


#endif






// ================================================================================
