//
// =============== SoftPixel Demo Firework (12.4.2008) ===============
//

#include <SoftPixelEngine.hpp>

using namespace sp;

#include "../../help/tutorials/common.hpp"

SoftPixelDevice* spDevice               = 0;
UserControl* spControl                  = 0;
video::VideoDriver* spDriver            = 0;
sound::SensitiveSoundDevice* spListener = 0;

sound::Sound* snd[3];

const s32 ScrWidth = 640, ScrHeight = 480;

s32 FireworkCount = 4;
io::stringc FireworkFilename[] = {
    "Firework1.txt", "Firework2.txt", "Firework3.txt", "Firework4.txt"
};

void Init();
void Drawing();

int main(void)
{
    
    Init();
    
    while (spDevice->updateEvent() && !spControl->isKeyPressed(KEY_ESCAPE))
    {
        spDriver->clearBuffers();
        
        spDriver->beginDrawing2D();
        
        Drawing();
        
        spDriver->endDrawing2D();
        
        spDriver->flipBuffers();
    }
    
    spDevice->deleteDevice();
    
    return 0;
}


/*
 * Class - Firework
 */

#define GRAVITY     0.02f
#define SPARKDIST   4

enum FireWorkTypes
{
    ROCKET,
    SPARK
};

class Firework
{
    
    public:
        
        Firework();
        Firework(dim::point2df Position, dim::point2df Direction, f32 Gravity, video::color Color);
        ~Firework();
        
        /* Functions */
        
        void Update();
        void Render();
        
        /* Inline functions */
        
        inline bool isDelete() const
        {
            return isDelete_;
        }
        
    private:
        
        /* Functions */
        
        void CreateSparks();
        
        /* Members */
        
        s32 Type_;
        dim::point2df Position_, Direction_;
        dim::point2di LastPosition_;
        video::color Color_;
        f32 Gravity_;
        bool isDelete_;
        
};

std::list<Firework> FireworkList;

Firework::Firework()
{
    Type_           = ROCKET;
    Position_       = dim::point2df(ScrWidth/2, ScrHeight);
    Direction_      = dim::point2df((f32)math::Rnd(10)/20 - 5.0f/20, -(f32)(math::Rnd(70, 90))/20);
    LastPosition_   = Position_.cast<s32>();
    Color_          = video::color(math::Rnd(255), math::Rnd(255), math::Rnd(255), 255);
    Gravity_        = GRAVITY;
    isDelete_       = false;
}
Firework::Firework(dim::point2df Position, dim::point2df Direction, f32 Gravity, video::color Color)
{
    Type_           = SPARK;
    Position_       = Position;
    Direction_      = Direction;
    LastPosition_   = Position_.cast<s32>();
    Color_          = video::color(
        math::MinMax(math::Random(Color.Red  -25, Color.Red  +25), 0, 255),
        math::MinMax(math::Random(Color.Green-25, Color.Green+25), 0, 255),
        math::MinMax(math::Random(Color.Blue -25, Color.Blue +25), 0, 255)
    );
    Gravity_        = Gravity;
    isDelete_       = false;
}
Firework::~Firework()
{
}

void Firework::Update()
{
    
    Direction_.Y += Gravity_;
    Position_ += Direction_;
    
    if (Type_ == ROCKET)
    {
        if (Direction_.Y > 0.5f)
        {
            spListener->playSound(snd[math::Random(2)]);
            CreateSparks();
            isDelete_ = true;
        }
    }
    else if (Type_ == SPARK)
    {
        if (--Color_.Alpha <= 0)
            isDelete_ = true;
    }
    
}

void Firework::Render()
{
    LastPosition_ = Position_.cast<s32>();
    spDriver->draw2DPixel(LastPosition_, Color_);
}

void Firework::CreateSparks()
{
    
    dim::point2df Pos, Direct;
    dim::size2di Size;
    std::ifstream hFile(FireworkFilename[math::Random(FireworkCount-1)].c_str());
    
    if (hFile)
    {
        
        hFile >> Size.Width;
        hFile >> Size.Height;
        ++Size.Width;
        ++Size.Height;
        
        c8* Line = new c8[Size.Width];
        bool* isSpark = new bool[Size.Width*Size.Height];
        
        s32 x, y = 0, i, j;
        while (!hFile.eof() && hFile.getline(Line, Size.Width))
        {
            for (x = 0; x < Size.Width; ++x)
                isSpark[y * Size.Width + x] = (Line[x] == '*');
            ++y;
        }
        
        // Add the picture sparks
        for (y = 0; y < Size.Height; ++y)
        {
            for (x = 0; x < Size.Width; ++x)
            {
                if (isSpark[y * Size.Width + x])
                {
                    Direct.set( (f32)(x-Size.Width/2)/25, (f32)(y-Size.Height/2)/15 );
                    FireworkList.push_back(Firework(Position_, Direct * 2, GRAVITY, Color_));
                }
            }
        }
        
        // Add some other sparks
        f32 Angle, Radius;
        for (i = 0, j = math::Random(75, 150); i < j; ++i)
        {
            // Randomize the direction
            Angle   = (f32)math::Random(3600)/10;
            Radius  = (f32)math::Random(100)/100;
            Direct.set(SIN(Angle)*Radius, COS(Angle)*Radius);
            // Add the spark
            FireworkList.push_back(Firework(Position_, Direct * 2, (f32)math::Random(10, 30)/1000, Color_));
        }
        
        hFile.close();
        
        spDeleteArray(Line);
        spDeleteArray(isSpark);
        
    }
    
}

/*
 * All the other functions
 */


void Init()
{
    spDevice = createDevice(
        ChooseRenderer(), dim::size2di(ScrWidth, ScrHeight), 32, "SoftPixel Demo Firework", false, true
    );
    spListener = sound::createListener();
    
    spDriver = spDevice->getVideoDriver();
    
    math::SeedRandom();
    spDriver->setPointSize(3);
    spDevice->setResizeMode(false);
    
    for (s32 i = 0; i < 3; ++i)
    {
        snd[i] = spListener->loadSound(io::stringc("explosion")+io::stringc(i+1)+io::stringc(".wav"), sound::SOUNDQUALITY_IGNORE, 1);
        spListener->setSoundVolume(snd[i], 0.15f);
    }
    
    FireworkList.push_back(Firework());
}

void Drawing()
{
    
    for (std::list<Firework>::iterator it = FireworkList.begin(); it != FireworkList.end();)
    {
        it->Update();
        it->Render();
        if (it->isDelete())
            it = FireworkList.erase(it);
        else
            ++it;
    }
    
    if (math::Random(100) == 1)
        FireworkList.push_back(Firework());
    
}










// ===================================================================
