/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * Direct3D9 video driver
 *
 */

#include "spDirect3D9Driver.hpp"
#include "spInternalDeclerations.hpp"
#include "spSceneManagerCamera.hpp"
#include "spSceneManagerSpecialEffectsActor.hpp" // !!!


namespace sp
{
namespace video
{


#if defined(WIN32)

/*
 * ========== Internal members ==========
 */

const io::stringc d3dDllFileName = io::stringc("d3dx9_") + io::stringc((s32)D3DX_SDK_VERSION) + io::stringc(".dll");

const s32 D3DCompareList[] = {
    D3DCMP_NEVER, D3DCMP_EQUAL, D3DCMP_NOTEQUAL, D3DCMP_LESS, D3DCMP_LESSEQUAL,
    D3DCMP_GREATER, D3DCMP_GREATEREQUAL, D3DCMP_ALWAYS,
};

const s32 D3DMappingGenList[] = {
    D3DTSS_TCI_PASSTHRU, D3DTSS_TCI_PASSTHRU, D3DTSS_TCI_CAMERASPACEPOSITION, D3DTSS_TCI_SPHEREMAP,
    D3DTSS_TCI_CAMERASPACENORMAL, D3DTSS_TCI_CAMERASPACEREFLECTIONVECTOR,
};

const s32 D3DTextureEnvList[] = {
    D3DTOP_MODULATE, D3DTOP_SELECTARG1, D3DTOP_ADD, D3DTOP_ADDSIGNED,
    D3DTOP_SUBTRACT, D3DTOP_LERP, D3DTOP_DOTPRODUCT3,
};

const s32 D3DBlendingList[] = {
    D3DBLEND_ZERO, D3DBLEND_ONE, D3DBLEND_SRCCOLOR, D3DBLEND_INVSRCCOLOR, D3DBLEND_SRCALPHA, D3DBLEND_INVSRCALPHA,
    D3DBLEND_DESTCOLOR, D3DBLEND_INVDESTCOLOR, D3DBLEND_DESTALPHA, D3DBLEND_INVDESTALPHA,
};


/*
 * ========== Constructors & destructor ==========
 */


Direct3D9Driver::Direct3D9Driver()
    : VideoDriver(DRIVER_DIRECT3D9), pDirect3D_(0), pDirect3DDevice_(0), LastRenderTarget_(0), CurD3DTexture_(0)
{
    /* Initialization */
    init();
}
Direct3D9Driver::~Direct3D9Driver()
{
    /* Release all Direct3D9 device contexts */
    clear();
}


/*
 * ========== Video driver identity ==========
 */

io::stringc Direct3D9Driver::getRenderer() const
{
    D3DADAPTER_IDENTIFIER9 adapter;
    
    pDirect3D_->GetAdapterIdentifier(D3DADAPTER_DEFAULT, 0, &adapter);
    
    return io::stringc(adapter.Description);
}
io::stringc Direct3D9Driver::getVersion() const
{
    return "Direct3D9";
}
io::stringc Direct3D9Driver::getVendor() const
{
    D3DADAPTER_IDENTIFIER9 adapter;
    
    pDirect3D_->GetAdapterIdentifier(D3DADAPTER_DEFAULT, 0, &adapter);
    
    io::stringc VendorName;
    
    switch (adapter.VendorId)
    {
        case 0x1002:
            VendorName = "ATI Technologies Inc."; break;
        case 0x10DE:
            VendorName = "NVIDIA Corporation"; break;
        case 0x102B:
            VendorName = "Matrox Electronic Systems Ltd."; break;
        case 0x121A:
            VendorName = "3dfx Interactive Inc"; break;
        case 0x5333:
            VendorName = "S3 Graphics Co., Ltd."; break;
        case 0x8086:
            VendorName = "Intel Corporation"; break;
        default:
            VendorName = io::stringc("Unknown VendorId: ") + io::stringc((u32)adapter.VendorId); break;
    }
    
    return VendorName;
}

bool Direct3D9Driver::supportAntiAlias() const // !!!
{
    return true;
}

bool Direct3D9Driver::supportMultitex() const
{
    return getMultitexCount() > 1;
}

bool Direct3D9Driver::supportVBO() const // !!!
{
    return true;
}

bool Direct3D9Driver::supportAnisotropicFiltering() const
{
    return getMaxAnisotropicFilter() > 1;
}

s32 Direct3D9Driver::getMultitexCount() const // !!!
{
    return 8;
}

s32 Direct3D9Driver::getMaxAnisotropicFilter() const // !!!
{
    return 16;
}


/*
 * ========== User controll functions (clear buffers, flip buffers, 2d drawing etc.) ==========
 */

void Direct3D9Driver::clearBuffers()
{
    setViewport(0, 0, __Screen_Width, __Screen_Height);
    
    pDirect3DDevice_->Clear(0, 0, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, ClearColor_.getSingle(), 1.0f, 0);
    
    pDirect3DDevice_->BeginScene();
}

void Direct3D9Driver::flipBuffers()
{
    pDirect3DDevice_->EndScene();
    
    pDirect3DDevice_->Present(0, 0, 0, 0);
}


/*
 * ========== Setting-/ getting functions ==========
 */

void Direct3D9Driver::setShadeMode(EShadeModeTypes ShadeMode)
{
    switch (ShadeMode)
    {
        case SHADEMODE_SMOOTH:
            pDirect3DDevice_->SetRenderState(D3DRS_SHADEMODE, D3DSHADE_GOURAUD);
            break;
        case SHADEMODE_FLAT:
            pDirect3DDevice_->SetRenderState(D3DRS_SHADEMODE, D3DSHADE_FLAT);
            break;
    }
}

void Direct3D9Driver::setRenderMode(EDriverRenderModes RenderMode)
{
    switch (RenderMode)
    {
        case RENDERMODE_LINE_LINEAR:
            pDirect3DDevice_->SetRenderState(D3DRS_ANTIALIASEDLINEENABLE, false);
            break;
        case RENDERMODE_LINE_SMOOTH:
            pDirect3DDevice_->SetRenderState(D3DRS_ANTIALIASEDLINEENABLE, true);
            break;
            
        case RENDERMODE_POINT_LINEAR:
            break;
        case RENDERMODE_POINT_SMOOTH:
            break;
    }
}

void Direct3D9Driver::setPolygonMode(EDriverRenderModes RenderFace, EDriverRenderModes RenderMode)
{
    
}

void Direct3D9Driver::setClearColor(color Color)
{
    ClearColor_ = Color;
}

void Direct3D9Driver::setVsync(bool isVsync)
{
    if (isVsync)
        Presenter_.PresentationInterval = D3DPRESENT_INTERVAL_ONE;
    else
        Presenter_.PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;
}

void Direct3D9Driver::setAntiAlias(bool isAntiAlias)
{
    pDirect3DDevice_->SetRenderState(D3DRS_MULTISAMPLEANTIALIAS, isAntiAlias);
}


/*
 * ======= Context functions (for initializing the engine) =======
 */

void Direct3D9Driver::initRenderer()
{
    /* Default settings */
    pDirect3DDevice_->SetRenderState(D3DRS_ZFUNC, D3DCMP_LESSEQUAL);
    pDirect3DDevice_->SetRenderState(D3DRS_ALPHATESTENABLE, true);
    pDirect3DDevice_->SetRenderState(D3DRS_SPECULARENABLE, true);
    pDirect3DDevice_->SetRenderState(D3DRS_NORMALIZENORMALS, true);
    
    /* Default camera range */
    RangeNear_  = 1.0f;
    RangeFar_   = 1000.0f;
    
    /* Clear & flip the video buffer */
    clearBuffers();
    flipBuffers();
}

void Direct3D9Driver::changeDisplaySettings(bool isFullScreen, dim::size2di Dimension, s32 ColorDepth)
{
    isFullScreen_ = isFullScreen;
}

bool Direct3D9Driver::initScreen(HWND hWnd, bool isAntiAlias, s32 MultiSamplePixelFormat, s32 ColorDepth)
{
    
    /* Create the Direct3D renderer */
    pDirect3D_ = Direct3DCreate9(D3D_SDK_VERSION);
    
    if (!pDirect3D_)
    {
        io::printError("could not create Direct3D9 interface");
        return false;
    }
    
    /* Get the display mode */
    D3DDISPLAYMODE d3dDisplayMode;
    pDirect3D_->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &d3dDisplayMode);
    
    /* Fill the Direct3D present parameter structure */
    if (isFullScreen_)
    {
        Presenter_.Windowed                     = FALSE;
        Presenter_.SwapEffect                   = D3DSWAPEFFECT_FLIP;
        Presenter_.FullScreen_RefreshRateInHz   = D3DPRESENT_RATE_DEFAULT;
        
        if (ColorDepth == 32)
            Presenter_.BackBufferFormat = D3DFMT_X8R8G8B8;
        else
            Presenter_.BackBufferFormat = D3DFMT_R5G6B5;
    }
    else
    {
        Presenter_.Windowed         = TRUE;
        Presenter_.SwapEffect       = D3DSWAPEFFECT_COPY;
        Presenter_.BackBufferFormat = d3dDisplayMode.Format;
    }
    
    Presenter_.BackBufferWidth          = __Screen_Width;
    Presenter_.BackBufferHeight         = __Screen_Height;
    Presenter_.BackBufferCount          = 1;
    Presenter_.EnableAutoDepthStencil   = TRUE;
    Presenter_.hDeviceWindow            = hWnd;
    Presenter_.PresentationInterval     = D3DPRESENT_INTERVAL_ONE; // if vsync
    
    Presenter_.AutoDepthStencilFormat   = D3DFMT_D24S8;//D3DFMT_D32;
    if (!checkDepthStencilFormat())
    {
        Presenter_.AutoDepthStencilFormat = D3DFMT_D24X4S4;//D3DFMT_D24X8;
        if (!checkDepthStencilFormat())
        {
            Presenter_.AutoDepthStencilFormat = D3DFMT_D15S1;//D3DFMT_D16;
            if (!checkDepthStencilFormat())
            {
                io::printError("could not found a suitable depth-stencil-format");
                return false;
            }
        }
    }
    
    if (isAntiAlias)
    {
        Presenter_.MultiSampleType      = (D3DMULTISAMPLE_TYPE)MultiSamplePixelFormat;
        Presenter_.SwapEffect           = D3DSWAPEFFECT_DISCARD;
    }
    
    /*
     * Create a device class using this information and
     * the info from the d3dpp stucture
     */
    HRESULT hResult = pDirect3D_->CreateDevice(
        D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd, D3DCREATE_HARDWARE_VERTEXPROCESSING, &Presenter_, &pDirect3DDevice_
    );
    
    if (hResult != D3D_OK)
    {
        hResult = pDirect3D_->CreateDevice(
            D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd, D3DCREATE_MIXED_VERTEXPROCESSING, &Presenter_, &pDirect3DDevice_
        );
    }
    
    if (hResult != D3D_OK)
    {
        hResult = pDirect3D_->CreateDevice(
            D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd, D3DCREATE_SOFTWARE_VERTEXPROCESSING, &Presenter_, &pDirect3DDevice_
        );
    }
    
    if (!pDirect3DDevice_)
    {
        io::printError("could not create Direct3D9 device");
        return false;
    }
    
    /*
     * Create the standard vertex buffer
     * used for 2d drawing operations
     * (drawing: rectangle, images etc.)
     */
    pDirect3DDevice_->CreateVertexBuffer(
        sizeof(SPrimitiveVertex)*4,
        0,
        FVF_VERTEX2D,
        D3DPOOL_DEFAULT,
        &pDirect3DVertexBuffer_,
        0
    );
    
    if (!pDirect3DVertexBuffer_)
    {
        io::printError("could not create Direct3D9 vertex buffer");
        return false;
    }
    
    /*
     * Create the flexible vertex buffer
     * used for 2d drawing operations
     * (drawing: polygon & other objects with undefined sizes)
     */
    pDirect3DDevice_->CreateVertexBuffer(
        sizeof(SPrimitiveVertex),
        0,
        FVF_VERTEX2D,
        D3DPOOL_DEFAULT,
        &pDirect3DFlexibleVertexBuffer_,
        0
    );
    
    if (!pDirect3DFlexibleVertexBuffer_)
    {
        io::printError("could not create Direct3D9 vertex buffer");
        return false;
    }
    
    return true;
    
}

void Direct3D9Driver::clearScreen(HWND &hWnd)
{
    
    if (hWnd)
    {
        if (!DestroyWindow(hWnd))
            io::printError("release of main window failed");
        hWnd = 0;
    }
    else
        io::printError("window handle is zero");
    
}

void Direct3D9Driver::initAntiAlias(HWND Wnd, bool &isAntiAlias, s32 &MultiSamplePixelFormat, s32 &MultiSampleFormat, s32 ColorDepth)
{
    
}


/*
 * ======= Rendering 3D scenes =======
 */

void Direct3D9Driver::setMaterialStatesTriangleList(SMeshMaterial &Material, dim::vector3df &Rotation)
{
    
    /* Temporary variables */
    D3DMATERIAL9 d3dMat;
    
    /* Cull facing */
    switch (Material.RenderFace)
    {
        case video::RENDER_FRONT:
            pDirect3DDevice_->SetRenderState(D3DRS_CULLMODE, isFrontFace_ ? D3DCULL_CCW : D3DCULL_CW);
            break;
        case video::RENDER_BACK:
            pDirect3DDevice_->SetRenderState(D3DRS_CULLMODE, isFrontFace_ ? D3DCULL_CW : D3DCULL_CCW);
            break;
        case video::RENDER_BOTH:
            pDirect3DDevice_->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);
            break;
    }
    
    /* Render states */
    pDirect3DDevice_->SetRenderState(D3DRS_ZENABLE          , Material.isZBuffer                    );
    pDirect3DDevice_->SetRenderState(D3DRS_LIGHTING         , Material.isLighting && __isLighting   );
    pDirect3DDevice_->SetRenderState(D3DRS_FOGENABLE        , Material.isFog && __isFog             );
    pDirect3DDevice_->SetRenderState(D3DRS_ALPHABLENDENABLE , Material.isBlending                   );
    pDirect3DDevice_->SetRenderState(D3DRS_COLORVERTEX      , Material.isColorMaterial              );
    
    /* Shininess */
    d3dMat.Power = Material.Shininess;
    
    /* Color: Diffuse */
    d3dMat.Diffuse = getD3DColor(Material.ColorDiffuse);
    
    /* Color: Ambient */
    d3dMat.Ambient = getD3DColor(Material.ColorAmbient);
    
    /* Color: Specular */
    d3dMat.Specular = getD3DColor(Material.ColorSpecular);
    
    /* Color: Emission */
    d3dMat.Emissive = getD3DColor(Material.ColorEmission);
    
    /* Set the material */
    pDirect3DDevice_->SetMaterial(&d3dMat);
    
    /* Depth functions */
    pDirect3DDevice_->SetRenderState(D3DRS_ZFUNC, D3DCompareList[Material.DepthMode]);
    
    /* Alpha functions */
    pDirect3DDevice_->SetRenderState(D3DRS_ALPHAFUNC, D3DCompareList[Material.AlphaMode]);
    pDirect3DDevice_->SetRenderState(D3DRS_ALPHAREF, s32(Material.AlphaReference * 255));
    
    /* Blending mode */
    pDirect3DDevice_->SetRenderState(D3DRS_SRCBLEND, D3DBlendingList[Material.BlendSource]);
    pDirect3DDevice_->SetRenderState(D3DRS_DESTBLEND, D3DBlendingList[Material.BlendTarget]);
    
    /* Polygon mode */
    pDirect3DDevice_->SetRenderState(D3DRS_FILLMODE, D3DFILL_POINT + Material.RenderModeFront);
    
    /* Flexible vertex format (FVF) */
    pDirect3DDevice_->SetFVF(FVF_VERTEX3D);
    
    /* Shader */
    if (Material.hShader)
        Material.hShader->bind();
    
}

void Direct3D9Driver::setMaterialStatesSprite(SMeshMaterial &Material)
{
    
    /* Cull facing */
    pDirect3DDevice_->SetRenderState(D3DRS_CULLMODE, D3DCULL_CCW);
    
    /* Orientation */
    pDirect3DDevice_->SetRenderState(D3DRS_ZENABLE, Material.isZBuffer);
    
    /* Material: lighting */
    pDirect3DDevice_->SetRenderState(D3DRS_LIGHTING, Material.isLighting && __isLighting);
    
    /* Blending mode */
    pDirect3DDevice_->SetRenderState(D3DRS_SRCBLEND, D3DBlendingList[Material.BlendSource]);
    pDirect3DDevice_->SetRenderState(D3DRS_DESTBLEND, D3DBlendingList[Material.BlendTarget]);
    
    /* Polygon mode */
    pDirect3DDevice_->SetRenderState(D3DRS_FILLMODE, D3DFILL_SOLID);
    
    /* Flexible vertex format (FVF) */
    pDirect3DDevice_->SetFVF(FVF_VERTEX3D);
    
    /* Transparency channel */
    pDirect3DDevice_->SetTextureStageState(0, D3DTSS_ALPHAOP, D3DTOP_MODULATE);
    
}

void Direct3D9Driver::drawTriangleList(
    u32 &BufferID, u32* IndicesList, u32 &IndicesCount, std::vector<SSurfaceTexture> &TextureList)
{
    
    /* Check if the object is empty */
    if (!IndicesCount)
        return;
    
    /* Temporary variables */
    s32 TextureLayer = 0;
    std::vector<SSurfaceTexture>::iterator itTex;
    
    /* Loop for each texture */
    for (itTex = TextureList.begin(); itTex != TextureList.end(); ++itTex, ++TextureLayer)
    {
        
        /* Bind the current texture */
        itTex->hTexture->bind(TextureLayer);
        
        /* Load texture matrix */
        pDirect3DDevice_->SetTransform(
            (D3DTRANSFORMSTATETYPE)(D3DTS_TEXTURE0 + TextureLayer), D3D_MATRIX(itTex->Matrix.getTextureMatrix())
        );
        
        /* Texture coordinate generation */
        pDirect3DDevice_->SetTextureStageState(
            TextureLayer,
            D3DTSS_TEXCOORDINDEX,
            itTex->TexMappingCoords != MAPGEN_NONE ? D3DMappingGenList[itTex->TexMappingGen] : TextureLayer
        );
        
        /* Texture stage states */
        pDirect3DDevice_->SetTextureStageState(TextureLayer, D3DTSS_COLOROP, D3DTextureEnvList[itTex->TexEnvType]);
        pDirect3DDevice_->SetTextureStageState(TextureLayer, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_COUNT2);
        
    }
    
    /* Render the triangles  */
    pDirect3DDevice_->DrawIndexedPrimitiveUP(
        D3DPT_TRIANGLELIST,
        0,
        VertexBufferList_[BufferID]->VerticesCount,
        IndicesCount/3,
        IndicesList,
        D3DFMT_INDEX32,
        VertexBufferList_[BufferID]->pVerticesList,
        sizeof(SMeshIndexedVertex3D)
    );
    
    /* Unbind the textures */
    for (itTex = TextureList.begin(), TextureLayer = 0; itTex != TextureList.end(); ++itTex, ++TextureLayer)
        itTex->hTexture->unbind(TextureLayer);
    
}

void Direct3D9Driver::drawTriangleList(u32 &BufferID, u32* IndicesList, u32 &IndicesCount)
{
    
}

void Direct3D9Driver::drawSprite(
    Texture* &hTexture, f32 &RotationZ, video::color &Color, dim::vector3df &Scale, dim::point2df &BasePosition)
{
    
    /* Bind textureing */
    hTexture->bind(0);
    
    /* Coordination */
    if (RotationZ)
        scene::spWorldMatrix.matrixRotateZ(RotationZ);
    
    /* Update the render matrix */
    updateModelviewMatrix();
    
    /* Temporary variables */
    u32 Clr = Color.getSingle();
    
    /* Set the vertex data */
    SMeshIndexedVertex3D VerticesList[4] = {
        SMeshIndexedVertex3D( -Scale.X + BasePosition.X,  Scale.Y + BasePosition.Y, 0.0f, Clr, 0.0f, 0.0f ),
        SMeshIndexedVertex3D(  Scale.X + BasePosition.X,  Scale.Y + BasePosition.Y, 0.0f, Clr, 1.0f, 0.0f ),
        SMeshIndexedVertex3D( -Scale.X + BasePosition.X, -Scale.Y + BasePosition.Y, 0.0f, Clr, 0.0f, 1.0f ),
        SMeshIndexedVertex3D(  Scale.X + BasePosition.X, -Scale.Y + BasePosition.Y, 0.0f, Clr, 1.0f, 1.0f )
    };
    
    u32 IndicesList[4] = { 0, 1, 2, 3 };
    
    /* Render the triangles  */
    pDirect3DDevice_->DrawIndexedPrimitiveUP(
        D3DPT_TRIANGLESTRIP,
        0,
        4,
        2,
        IndicesList,
        D3DFMT_INDEX32,
        VerticesList,
        sizeof(SMeshIndexedVertex3D)
    );
    
    /* Unbind textureing */
    hTexture->unbind(0);
    
}

void Direct3D9Driver::updateLight(
    u32 &LightID, scene::ELightTypes &LightingType, bool &isVolumetric,
    dim::vector3df Direction, f32 &NearSpot, f32 &FarSpot,
    f32 &AttenuationConstant, f32 &AttenuationLinear, f32 &AttenuationQuadratic)
{
    
    /* Get the light source */
    pDirect3DDevice_->GetLight(LightID, &CurLight_);
    
    /* Update the direction */
    Direction = scene::spWorldMatrix.getRotationMatrix() * Direction;
    
    /* Lighting location */
    CurLight_.Direction = *D3D_VECTOR(Direction);
    CurLight_.Position  = *D3D_VECTOR(scene::spWorldMatrix.getPosition());
    
    /* Spot light attributes */
    CurLight_.Theta = FarSpot * 2.0f * math::RAD;
    CurLight_.Phi   = NearSpot * 2.0f * math::RAD;
    
    /* Volumetric light attenuations */
    if (isVolumetric)
    {
        CurLight_.Attenuation0  = AttenuationConstant;
        CurLight_.Attenuation1  = AttenuationLinear;
        CurLight_.Attenuation2  = AttenuationQuadratic;
    }
    else
    {
        CurLight_.Attenuation0  = 0.0;
        CurLight_.Attenuation1  = 0.0;
        CurLight_.Attenuation2  = 0.0;
    }
    
    /* Set the light source */
    pDirect3DDevice_->SetLight(LightID, &CurLight_);
    
}

void Direct3D9Driver::createVertexBuffer(u32 &BufferID)
{
    SVertexBuffer* VertexBuffer = new SVertexBuffer();
    {
        BufferID = VertexBufferList_.size();
        VertexBuffer->pBufferID = &BufferID;
    }
    VertexBufferList_.push_back(VertexBuffer);
}

void Direct3D9Driver::deleteVertexBuffer(u32 &BufferID)
{
    for (std::vector<SVertexBuffer*>::iterator it = VertexBufferList_.begin(); it != VertexBufferList_.end(); ++it)
    {
        if (*(*it)->pBufferID == BufferID)
        {
            spDeletePointer(*it);
            
            it = VertexBufferList_.erase(it);
            
            for (; it != VertexBufferList_.end(); ++it)
                --*(*it)->pBufferID;
            
            break;
        }
    }
}

void Direct3D9Driver::updateVertexBuffer(u32 &BufferID, SMeshVertex3D* pVerticesList, s32 VerticesCount)
{
    VertexBufferList_[BufferID]->copyVerticesList(pVerticesList, VerticesCount);
}

void Direct3D9Driver::disableTriangleListStates()
{
    /* Default render functions */
    pDirect3DDevice_->SetRenderState(D3DRS_ALPHAFUNC, D3DCMP_ALWAYS);
    pDirect3DDevice_->SetRenderState(D3DRS_ALPHAREF, 0);
    pDirect3DDevice_->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
    pDirect3DDevice_->SetRenderState(D3DRS_ZFUNC, D3DCMP_LESSEQUAL);
}

void Direct3D9Driver::disable3DRenderStates()
{
    /* Disable all used states */
    pDirect3DDevice_->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
    pDirect3DDevice_->SetRenderState(D3DRS_LIGHTING, FALSE);
    pDirect3DDevice_->SetRenderState(D3DRS_FOGENABLE, FALSE);
}

void Direct3D9Driver::disableTexturing()
{
    
}

void Direct3D9Driver::setDefaultAlphaBlending()
{
    pDirect3DDevice_->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);
    pDirect3DDevice_->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA);
}

void Direct3D9Driver::enableBlending()
{
    pDirect3DDevice_->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
}

void Direct3D9Driver::disableBlending()
{
    pDirect3DDevice_->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
}


/*
 * ======= Lighting =======
 */

void Direct3D9Driver::addDynamicLightSource(
    u32 LightID, scene::ELightTypes Type,
    video::color &Diffuse, video::color &Ambient, video::color &Specular,
    f32 AttenuationConstant_, f32 AttenuationLinear_, f32 AttenuationQuadratic_)
{
    /* Set the lighting type */
    switch (Type)
    {
        case scene::LIGHT_DIRECTIONAL:
            CurLight_.Type = D3DLIGHT_DIRECTIONAL; break;
        case scene::LIGHT_POINT:
            CurLight_.Type = D3DLIGHT_POINT; break;
        case scene::LIGHT_SPOT:
            CurLight_.Type = D3DLIGHT_SPOT; break;
    }
    
    /* Default values */
    CurLight_.Range         = 1000.0f;
    CurLight_.Direction.z   = 1.0f;
    
    /* Lighting colors */
    CurLight_.Diffuse   = getD3DColor(Diffuse);
    CurLight_.Ambient   = getD3DColor(Ambient);
    CurLight_.Specular  = getD3DColor(Specular);
    
    /* Volumetric light attenuations */
    CurLight_.Attenuation0  = AttenuationConstant_;
    CurLight_.Attenuation1  = AttenuationLinear_;
    CurLight_.Attenuation2  = AttenuationQuadratic_;
    
    /* Set the light attributes */
    pDirect3DDevice_->SetLight(LightID, &CurLight_);
    
    /* Enable the light */
    pDirect3DDevice_->LightEnable(LightID, true);
}

void Direct3D9Driver::setLightStatus(u32 LightID, bool isEnable)
{
    pDirect3DDevice_->LightEnable(LightID, isEnable);
}

void Direct3D9Driver::setLightColor(u32 LightID, video::color &Diffuse, video::color &Ambient, video::color &Specular)
{
    /* Get the light attributes */
    pDirect3DDevice_->GetLight(LightID, &CurLight_);
    
    /* Diffuse color */
    CurLight_.Diffuse   = getD3DColor(Diffuse);
    
    /* Ambient color */
    CurLight_.Ambient   = getD3DColor(Ambient);
    
    /* Specular color */
    CurLight_.Specular  = getD3DColor(Specular);
    
    /* Set the light attributes */
    pDirect3DDevice_->SetLight(LightID, &CurLight_);
}


/*
 * ======= Fog effect =======
 */

void Direct3D9Driver::setFogMode(s32 FogMode)
{
    /* General settings */
    FogMode_ = FogMode;
    
    /* Select the fog mode */
    switch (FogMode)
    {
        case FOG_NONE:
        {
            __isFog = false;
        }
        break;
        
        case FOG_STATIC:
        {
            __isFog = true;
            
            /* Set fog type */
            switch (FogType_)
            {
                case FOG_PALE:
                    pDirect3DDevice_->SetRenderState(D3DRS_FOGTABLEMODE, D3DFOG_EXP); break;
                case FOG_THICK:
                    pDirect3DDevice_->SetRenderState(D3DRS_FOGTABLEMODE, D3DFOG_EXP2); break;
            }
            
            /* Range settings */
            pDirect3DDevice_->SetRenderState(D3DRS_FOGDENSITY, *(DWORD*)&FogRange_);
            pDirect3DDevice_->SetRenderState(D3DRS_FOGSTART, *(DWORD*)&FogNear_);
            pDirect3DDevice_->SetRenderState(D3DRS_FOGEND, *(DWORD*)&FogFar_);
        }
        break;
        
        case FOG_VOLUMETRIC:
        {
            __isFog = true;
            
            /* Renderer settings */
            pDirect3DDevice_->SetRenderState(D3DRS_FOGTABLEMODE, D3DFOG_LINEAR);
            pDirect3DDevice_->SetRenderState(D3DRS_FOGVERTEXMODE, D3DFOG_LINEAR); // ???
            pDirect3DDevice_->SetRenderState(D3DRS_FOGDENSITY, *(DWORD*)&FogRange_);
            pDirect3DDevice_->SetRenderState(D3DRS_FOGSTART, 0);
            pDirect3DDevice_->SetRenderState(D3DRS_FOGEND, 1);
        }
        break;
    }
}

void Direct3D9Driver::setFogColor(video::color Color)
{
    pDirect3DDevice_->SetRenderState(D3DRS_FOGCOLOR, Color.getSingle());
}

void Direct3D9Driver::setFogRange(f32 Range, f32 NearFog, f32 FarFog, s32 FogType)
{
    FogType_    = FogType;
    FogRange_   = Range;
    FogNear_    = NearFog;
    FogFar_     = FarFog;
    
    if (FogMode_ != FOG_VOLUMETRIC)
    {
        /* Set fog type */
        switch (FogType_)
        {
            case FOG_PALE:
                pDirect3DDevice_->SetRenderState(D3DRS_FOGTABLEMODE, D3DFOG_EXP); break;
            case FOG_THICK:
                pDirect3DDevice_->SetRenderState(D3DRS_FOGTABLEMODE, D3DFOG_EXP2); break;
        }
        
        /* Range settings */
        pDirect3DDevice_->SetRenderState(D3DRS_FOGDENSITY, *(DWORD*)&FogRange_);
        pDirect3DDevice_->SetRenderState(D3DRS_FOGSTART, *(DWORD*)&FogNear_);
        pDirect3DDevice_->SetRenderState(D3DRS_FOGEND, *(DWORD*)&FogFar_);
    }
}


/*
 * ======= Stencil buffer =======
 */

void Direct3D9Driver::drawStencilShadowVolume(
    const dim::vector3df* pTriangleList, s32 Count, bool ZFailMethod, bool VolumetricShadow)
{
    if (!pTriangleList || !Count)
        return;
    
    for (s32 i = 0; i < 4; ++i)
    {
        pDirect3DDevice_->SetTextureStageState(i, D3DTSS_COLOROP, D3DTOP_DISABLE);
        pDirect3DDevice_->SetTextureStageState(i, D3DTSS_ALPHAOP, D3DTOP_DISABLE);
    }
    
    pDirect3DDevice_->SetFVF(FVF_POSITION);
    
    pDirect3DDevice_->SetRenderState(D3DRS_ZWRITEENABLE, FALSE);
    pDirect3DDevice_->SetRenderState(D3DRS_STENCILENABLE, TRUE);
    
    if (ZFailMethod)
    {
        /* Set the render states */
        pDirect3DDevice_->SetRenderState(D3DRS_STENCILFUNC, D3DCMP_ALWAYS);
        pDirect3DDevice_->SetRenderState(D3DRS_STENCILZFAIL, D3DSTENCILOP_KEEP);
        pDirect3DDevice_->SetRenderState(D3DRS_STENCILFAIL, D3DSTENCILOP_KEEP);
        pDirect3DDevice_->SetRenderState(D3DRS_STENCILPASS, D3DSTENCILOP_KEEP);
        
        pDirect3DDevice_->SetRenderState(D3DRS_STENCILREF, 0x0);
        pDirect3DDevice_->SetRenderState(D3DRS_STENCILMASK, 0xFFFFFFFF);
        pDirect3DDevice_->SetRenderState(D3DRS_STENCILWRITEMASK, 0xFFFFFFFF);
        
        pDirect3DDevice_->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
        pDirect3DDevice_->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_ZERO);
        pDirect3DDevice_->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_ONE);
        
        /* Draw the stencil shadow volume */
        pDirect3DDevice_->SetRenderState(D3DRS_CULLMODE, D3DCULL_CW);
        pDirect3DDevice_->SetRenderState(D3DRS_STENCILZFAIL, D3DSTENCILOP_INCRSAT);
        pDirect3DDevice_->DrawPrimitiveUP(D3DPT_TRIANGLELIST, Count / 3, pTriangleList, sizeof(dim::vector3df));
        
        if (!VolumetricShadow)
        {
            pDirect3DDevice_->SetRenderState(D3DRS_CULLMODE, D3DCULL_CCW);
            pDirect3DDevice_->SetRenderState(D3DRS_STENCILZFAIL, D3DSTENCILOP_DECRSAT);
            pDirect3DDevice_->DrawPrimitiveUP(D3DPT_TRIANGLELIST, Count / 3, pTriangleList, sizeof(dim::vector3df));
        }
    }
    else
    {
        /* Set the render states */
        pDirect3DDevice_->SetRenderState(D3DRS_STENCILFUNC, D3DCMP_ALWAYS);
        pDirect3DDevice_->SetRenderState(D3DRS_STENCILZFAIL, D3DSTENCILOP_KEEP);
        pDirect3DDevice_->SetRenderState(D3DRS_STENCILFAIL, D3DSTENCILOP_KEEP);
        
        pDirect3DDevice_->SetRenderState(D3DRS_STENCILREF, 0x1);
        pDirect3DDevice_->SetRenderState(D3DRS_STENCILMASK, 0xFFFFFFFF);
        pDirect3DDevice_->SetRenderState(D3DRS_STENCILWRITEMASK, 0xFFFFFFFF);
        
        pDirect3DDevice_->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
        pDirect3DDevice_->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_ZERO);
        pDirect3DDevice_->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_ONE);
        
        /* Draw the stencil shadow volume */
        pDirect3DDevice_->SetRenderState(D3DRS_CULLMODE, D3DCULL_CCW);
        pDirect3DDevice_->SetRenderState(D3DRS_STENCILPASS, D3DSTENCILOP_INCRSAT);
        pDirect3DDevice_->DrawPrimitiveUP(D3DPT_TRIANGLELIST, Count / 3, pTriangleList, sizeof(dim::vector3df));
        
        if (!VolumetricShadow)
        {
            pDirect3DDevice_->SetRenderState(D3DRS_CULLMODE, D3DCULL_CW);
            pDirect3DDevice_->SetRenderState(D3DRS_STENCILPASS, D3DSTENCILOP_DECRSAT);
            pDirect3DDevice_->DrawPrimitiveUP(D3DPT_TRIANGLELIST, Count / 3, pTriangleList, sizeof(dim::vector3df));
        }
    }
}

void Direct3D9Driver::drawStencilShadow(video::color Color)
{
    /* Configure the stencil states */
    pDirect3DDevice_->SetRenderState(D3DRS_STENCILREF, 0x1);
    pDirect3DDevice_->SetRenderState(D3DRS_STENCILFUNC, D3DCMP_LESSEQUAL);
    pDirect3DDevice_->SetRenderState(D3DRS_STENCILPASS, D3DSTENCILOP_KEEP);
    pDirect3DDevice_->SetRenderState(D3DRS_STENCILZFAIL, D3DSTENCILOP_KEEP);
    pDirect3DDevice_->SetRenderState(D3DRS_STENCILMASK, 0xFFFFFFFF);
    pDirect3DDevice_->SetRenderState(D3DRS_STENCILWRITEMASK, 0xFFFFFFFF);
    pDirect3DDevice_->SetRenderState(D3DRS_CULLMODE, D3DCULL_CCW);
    
    /* Draw the rectangle */
    beginDrawing2D();
    draw2DRectangle(dim::rect2di(0, 0, __Screen_Width, __Screen_Height), Color);
    endDrawing2D();
    
    /* Clear the stencil buffer */
    pDirect3DDevice_->Clear(0, 0, D3DCLEAR_STENCIL, 0, 1.0, 0);
    
    /* Back settings */
    pDirect3DDevice_->SetRenderState(D3DRS_STENCILENABLE, FALSE);
    pDirect3DDevice_->SetRenderState(D3DRS_ZWRITEENABLE, TRUE);
}


/* === Clipping planes === */

void Direct3D9Driver::setClipPlane(u32 Index, const dim::plane3df &Plane, bool Enable)
{
    if (Index >= MaxClippingPlanes_)
        return;
    
    pDirect3DDevice_->SetClipPlane(Index, (const f32*)&Plane);
    
    DWORD State;
    
    pDirect3DDevice_->GetRenderState(D3DRS_CLIPPLANEENABLE, &State);
    
    if (Enable)
        State |= (1 << Index);
    else
        State &= ~(1 << Index);
    
    pDirect3DDevice_->SetRenderState(D3DRS_CLIPPLANEENABLE, State);
}


/* === Special effects actor === */

scene::effect::SpecialEffectsActor* Direct3D9Driver::createSpecialEffectsActor()
{
    return new scene::effect::SpecialEffectsActor(); // !!!
}



/*
 * ======= Shader effects =======
 */

Shader* Direct3D9Driver::createShader(EStandardShaderTypes Type)
{
    Shader* NewShader = new Direct3D9Shader(Type);
    
    ShaderList_.push_back(NewShader);
    
    return NewShader;
}

Shader* Direct3D9Driver::loadShader(io::stringc FilenameVertex, io::stringc FilenameFragment, bool UseHighLevel)
{
    const io::stringc ShaderType    = (UseHighLevel ? "HLSL shader" : "shader program");
    const io::stringc PrimaryFile   = (FilenameVertex != "" ? FilenameVertex : FilenameFragment);
    const io::stringc SecondaryFile = (FilenameFragment != "" && FilenameVertex != "" ? io::stringc("\" and \"") + FilenameFragment : "\"");
    
    io::printMessage(
        io::stringc("Load ") + ShaderType + io::stringc(": \"") + PrimaryFile + SecondaryFile
    );
    
    Shader* NewShader = new Direct3D9Shader(
        FilenameVertex, FilenameFragment, "", "", HLSL_VERTEX_1_1, HLSL_PIXEL_1_1, UseHighLevel
    );
    
    ShaderList_.push_back(NewShader);
    
    return NewShader;
}
Shader* Direct3D9Driver::loadShader(
    io::stringc FilenameVertex, io::stringc FilenamePixel,
    io::stringc EntryPointVertex, io::stringc EntryPointPixel,
    EHLSLVertexShaderTypes CompilerTargetVertex,
    EHLSLPixelShaderTypes CompilerTargetPixel,
    bool UseHighLevel)
{
    const io::stringc ShaderType    = (UseHighLevel ? "HLSL shader" : "shader program");
    const io::stringc PrimaryFile   = (FilenameVertex != "" ? FilenameVertex : FilenamePixel);
    const io::stringc SecondaryFile = (FilenamePixel != "" && FilenameVertex != "" ? io::stringc("\" and \"") + FilenamePixel : "\"");
    
    io::printMessage(
        io::stringc("Load ") + ShaderType + io::stringc(": \"") + PrimaryFile + SecondaryFile
    );
    
    Shader* NewShader = new Direct3D9Shader(
        FilenameVertex, FilenamePixel, EntryPointVertex, EntryPointPixel,
        CompilerTargetVertex, CompilerTargetPixel,
        UseHighLevel
    );
    
    ShaderList_.push_back(NewShader);
    
    return NewShader;
}

void Direct3D9Driver::unbindShader()
{
    pDirect3DDevice_->SetVertexShader(0);
    pDirect3DDevice_->SetPixelShader(0);
}


/*
 * ======= Stack operation functions =======
 */

void Direct3D9Driver::pushViewport()
{
    D3DVIEWPORT9 d3dViewport;
    pDirect3DDevice_->GetViewport(&d3dViewport);
    
    StackViewport_[0] = d3dViewport.X;
    StackViewport_[1] = d3dViewport.Y;
    StackViewport_[2] = d3dViewport.Width;
    StackViewport_[3] = d3dViewport.Height;
}

void Direct3D9Driver::popViewport()
{
    setViewport(StackViewport_[0], StackViewport_[1], StackViewport_[2], StackViewport_[3]);
}


/*
 * ======= Texturing =======
 */

void Direct3D9Driver::bindTexture(u32 ID)
{
    
}

void Direct3D9Driver::unbindTexture()
{
    
}


/*
 * ========== Drawing 2D objects ==========
 */



/* Drawing 2d */

void Direct3D9Driver::beginDrawing2D(bool isSettingViewport)
{
    /* Disable depth test (only 2d) */
    pDirect3DDevice_->SetRenderState(D3DRS_ZENABLE, FALSE);
    
    /* Alpha blending */
    pDirect3DDevice_->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);
    pDirect3DDevice_->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA);
    
    /* Enable alpha ablending */
    pDirect3DDevice_->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
    
    /* Unit matrices */
    dim::matrix4f Mat;
    Mat.matrixReset();
    setProjectionMatrix(Mat);
    setViewMatrix(Mat);
    setWorldMatrix(Mat);
    
    /* Disable 3d render states */
    pDirect3DDevice_->SetTextureStageState(0, D3DTSS_TEXTURETRANSFORMFLAGS, D3DTTFF_DISABLE);
    pDirect3DDevice_->SetTextureStageState(0, D3DTSS_TEXCOORDINDEX, 0);
    pDirect3DDevice_->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);
    
    /* Use no texture layer */
    pDirect3DDevice_->SetTexture(0, 0);
    
	if (isSettingViewport)
    	setViewport(0, 0, __Screen_Width, __Screen_Height);
}

void Direct3D9Driver::endDrawing2D()
{
    /* Enable depth test */
    pDirect3DDevice_->SetRenderState(D3DRS_ZENABLE, TRUE);
    
    /* Disable alpha ablending */
    pDirect3DDevice_->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
	
}

/* Drawing 3d */

void Direct3D9Driver::beginDrawing3D(bool isLighting, bool isFog)
{
    /* Lighting */
    pDirect3DDevice_->SetRenderState(D3DRS_LIGHTING, isLighting);
    
    /* Fog */
    pDirect3DDevice_->SetRenderState(D3DRS_FOGENABLE, isFog);
}

void Direct3D9Driver::endDrawing3D()
{
    /* Disable lighting & fog */
    pDirect3DDevice_->SetRenderState(D3DRS_LIGHTING, FALSE);
    pDirect3DDevice_->SetRenderState(D3DRS_FOGENABLE, FALSE);
}

/* Alpha blending */

void Direct3D9Driver::beginAlphaBlending(u8 Type)
{
    /* Blending function */
    if (Type == BLENDING_BRIGHTNESS)
    {
        pDirect3DDevice_->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);
        pDirect3DDevice_->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_ONE);
    }
    else if (Type == BLENDING_DARKNESS)
    {
        pDirect3DDevice_->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_DESTCOLOR);
        pDirect3DDevice_->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_ZERO);
    }
}

void Direct3D9Driver::endAlphaBlending()
{
    /* Blending function */
    pDirect3DDevice_->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);
    pDirect3DDevice_->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA);
}

/* Masking */

void Direct3D9Driver::beginMasking(u8 Type)
{
    /* Blending function */
    if (Type == MASKING_SOURCE)
    {
        pDirect3DDevice_->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_DESTCOLOR);
        pDirect3DDevice_->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_ZERO);
    }
    else if (Type == MASKING_TARGET)
    {
        pDirect3DDevice_->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);
        pDirect3DDevice_->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_ONE);
    }
}

void Direct3D9Driver::endMasking()
{
    /* Blending function */
    pDirect3DDevice_->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);
    pDirect3DDevice_->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA);
}

/* Texture drawing */

// !INCOMPLETE!
void Direct3D9Driver::beginTextureDrawing(Texture* hTexture, s32 RenderMode, u32 TextureDataType)
{
    pDirect3DDevice_->GetRenderTarget(0, &LastRenderTarget_);
    
    IDirect3DTexture9* d3dTexture = static_cast<Direct3D9Texture*>(hTexture)->getDirect3DTexture();
    IDirect3DSurface9* Surface = 0;
    
    d3dTexture->GetSurfaceLevel(0, &Surface);
    
    pDirect3DDevice_->SetRenderTarget(0, Surface);
}

void Direct3D9Driver::endTextureDrawing(bool isUpdateImageData)
{
    pDirect3DDevice_->SetRenderTarget(0, LastRenderTarget_);
    
    LastRenderTarget_->Release();
    LastRenderTarget_ = 0;
    
    pDirect3DDevice_->Clear(0, 0, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, ClearColor_.getSingle(), 1.0f, 0);
}

/* Scissors (drawing rectangle) */

void Direct3D9Driver::beginDrawingRect(dim::rect2di Dimension)
{
    /* Enable scissor */
    pDirect3DDevice_->SetRenderState(D3DRS_SCISSORTESTENABLE, TRUE);
    
    /* Scissor area */
    RECT rc;
    {
        rc.left     = Dimension.Left;
        rc.top      = Dimension.Top;
        rc.right    = Dimension.Right;
        rc.bottom   = Dimension.Bottom;
    }
    pDirect3DDevice_->SetScissorRect(&rc);
}

void Direct3D9Driver::beginDrawingRect(dim::point2di Position, dim::size2di Size)
{
    /* Enable scissor */
    pDirect3DDevice_->SetRenderState(D3DRS_SCISSORTESTENABLE, TRUE);
    
    /* Scissor area */
    RECT rc;
    {
        rc.left     = Position.X;
        rc.top      = Position.Y;
        rc.right    = Size.Width - Position.X;
        rc.bottom   = Size.Height - Position.Y;
    }
    pDirect3DDevice_->SetScissorRect(&rc);
}

void Direct3D9Driver::endDrawingRect()
{
    /* Disable scissor */
    pDirect3DDevice_->SetRenderState(D3DRS_SCISSORTESTENABLE, FALSE);
}

/* Viewports (drawing area) */

void Direct3D9Driver::setDrawingArea(dim::rect2di Dimension)
{
    setViewport(Dimension.Left, Dimension.Top, Dimension.Right, Dimension.Bottom);
}

void Direct3D9Driver::setDrawingArea(dim::point2di Position, dim::size2di Size)
{
    setViewport(Position.X, Position.Y, Size.Width, Size.Height);
}

/* Line & point size */

void Direct3D9Driver::setLineSize(s32 Size)
{
    
}
void Direct3D9Driver::setPointSize(s32 Size)
{
    
}

void Direct3D9Driver::setSolidMode(bool isSolidMode)
{
    
}


/* ======= Texture drawing ======= */


void Direct3D9Driver::draw2DImage(Texture* hTexture, dim::point2di Position, color Color, bool isTransFilterUsing)
{
    /* Bind the texture */
    hTexture->bind();
    
    /* Set the texture attributes */
    pDirect3DDevice_->SetTextureStageState(0, isTransFilterUsing ? D3DTSS_ALPHAOP : D3DTSS_COLOROP, D3DTOP_MODULATE);
    
    /* Temporary variables */
    s32 Width   = hTexture->getSize().Width;
    s32 Height  = hTexture->getSize().Height;
    u32 Clr     = Color.getSingle();
    
    /* Set the vertex data */
    SPrimitiveVertex VerticesList[4] = {
        SPrimitiveVertex( (f32)Position.X, (f32)Position.Y, 0.0f, Clr, 0.0f, 0.0f ),
        SPrimitiveVertex( (f32)Position.X + Width, (f32)Position.Y, 0.0f, Clr, 1.0f, 0.0f ),
        SPrimitiveVertex( (f32)Position.X + Width, (f32)Position.Y + Height, 0.0f, Clr, 1.0f, 1.0f ),
        SPrimitiveVertex( (f32)Position.X, (f32)Position.Y + Height, 0.0f, Clr, 0.0f, 1.0f )
    };
    
    /* Set the render states */
    pDirect3DDevice_->SetRenderState(D3DRS_FILLMODE, D3DFILL_SOLID);
    
    /* Update the primitive list */
    updatePrimitiveList(VerticesList, 4);
    
    /* Draw the rectangle */
    pDirect3DDevice_->DrawPrimitive(D3DPT_TRIANGLEFAN, 0, 2);
    
    /* Unbind the texture */
    hTexture->unbind();
}

void Direct3D9Driver::draw2DImage(
    Texture* hTexture, dim::rect2di Position, dim::rect2df Clipping,
    color Color, bool isTransFilterUsing)
{
    /* Bind the texture */
    hTexture->bind();
    
    /* Set the texture attributes */
    pDirect3DDevice_->SetTextureStageState(0, isTransFilterUsing ? D3DTSS_ALPHAOP : D3DTSS_COLOROP, D3DTOP_MODULATE);
    
    /* Temporary variables */
    u32 Clr = Color.getSingle();
    
    /* Correct the position */
    Position.Right += Position.Left;
    Position.Bottom += Position.Top;
    
    /* Set the vertex data */
    SPrimitiveVertex VerticesList[4] = {
        SPrimitiveVertex( (f32)Position.Left, (f32)Position.Top, 0.0f, Clr, Clipping.Left, Clipping.Top ),
        SPrimitiveVertex( (f32)Position.Right, (f32)Position.Top, 0.0f, Clr, Clipping.Right, Clipping.Top ),
        SPrimitiveVertex( (f32)Position.Right, (f32)Position.Bottom, 0.0f, Clr, Clipping.Right, Clipping.Bottom ),
        SPrimitiveVertex( (f32)Position.Left, (f32)Position.Bottom, 0.0f, Clr, Clipping.Left, Clipping.Bottom )
    };
    
    /* Set the render states */
    pDirect3DDevice_->SetRenderState(D3DRS_FILLMODE, D3DFILL_SOLID);
    
    /* Update the primitive list */
    updatePrimitiveList(VerticesList, 4);
    
    /* Draw the rectangle */
    pDirect3DDevice_->DrawPrimitive(D3DPT_TRIANGLEFAN, 0, 2);
    
    /* Unbind the texture */
    hTexture->unbind();
}

void Direct3D9Driver::draw2DImage(
    Texture* hTexture, dim::point2di Position,
    f32 Rotation, f32 Radius, color Color, bool isTransFilterUsing)
{
    /* Set the texture attributes */
    pDirect3DDevice_->SetTextureStageState(0, isTransFilterUsing ? D3DTSS_ALPHAOP : D3DTSS_COLOROP, D3DTOP_MODULATE);
    
    /* Bind the texture */
    hTexture->bind();
    
    /* Temporary variables */
    u32 Clr = Color.getSingle();
    
    /* Set the vertex data */
    Radius *= math::SQRT2;
    SPrimitiveVertex VerticesList[4] = {
        SPrimitiveVertex( (f32)Position.X + SIN(Rotation -  45)*Radius, (f32)Position.Y - COS(Rotation -  45)*Radius, 0.0f, Clr, 0.0f, 0.0f ),
        SPrimitiveVertex( (f32)Position.X + SIN(Rotation +  45)*Radius, (f32)Position.Y - COS(Rotation +  45)*Radius, 0.0f, Clr, 1.0f, 0.0f ),
        SPrimitiveVertex( (f32)Position.X + SIN(Rotation + 135)*Radius, (f32)Position.Y - COS(Rotation + 135)*Radius, 0.0f, Clr, 1.0f, 1.0f ),
        SPrimitiveVertex( (f32)Position.X + SIN(Rotation - 135)*Radius, (f32)Position.Y - COS(Rotation - 135)*Radius, 0.0f, Clr, 0.0f, 1.0f )
    };
    
    /* Set the render states */
    pDirect3DDevice_->SetRenderState(D3DRS_FILLMODE, D3DFILL_SOLID);
    
    /* Update the primitive list */
    updatePrimitiveList(VerticesList, 4);
    
    /* Draw the rectangle */
    pDirect3DDevice_->DrawPrimitive(D3DPT_TRIANGLEFAN, 0, 2);
    
    /* Unbind the texture */
    hTexture->unbind();
}

void Direct3D9Driver::draw2DImage(
    Texture* hTexture, dim::rect2di Position, dim::rect2df Clipping,
    bool isTransFilterUsing, f32 Rotation, dim::point2df RotationPoint,
    color lefttopColor, color righttopColor, color rightbottomColor, color leftbottomColor)
{
    /* Set the texture attributes */
    pDirect3DDevice_->SetTextureStageState(0, isTransFilterUsing ? D3DTSS_ALPHAOP : D3DTSS_COLOROP, D3DTOP_MODULATE);
    
    /* Bind the texture */
    hTexture->bind();
    
    /* Translation of location */
    if (Position.Right <= 0 && Position.Bottom <= 0)
    {
        Position.Right   = hTexture->getSize().Width;
        Position.Bottom  = hTexture->getSize().Height;
    }
    Position.Right   += Position.Left;
    Position.Bottom  += Position.Top;
    
    /* Temporary variables */
    dim::point2df lefttopPos, righttopPos;
    dim::point2df rightbottomPos, leftbottomPos;
    dim::point2df tmpRotPoint, tmpStartPoint;
    f32 tmpDegree, tmpRotDist;
    f32 tmpExtLine, tmpExpPosY;
    
    /* Computing the rotation */
    
    tmpRotPoint.X       = (Position.Right - Position.Left) * RotationPoint.X;
    tmpRotPoint.Y       = (Position.Bottom - Position.Top) * (1.0f - RotationPoint.Y);
    tmpExpPosY          = (Position.Bottom - Position.Top) * RotationPoint.Y;
    
    tmpExtLine          = (Position.Right-Position.Left) * RotationPoint.X;
    tmpRotDist          = math::getDistance((f32)Position.Left, (f32)Position.Top, Position.Left+tmpRotPoint.X, Position.Top+tmpRotPoint.Y);
    tmpDegree           = tmpRotDist == 0 ? 0 : ASIN( tmpExtLine / tmpRotDist );
    lefttopPos.X        = Position.Left + tmpRotPoint.X - SIN(tmpDegree - Rotation) * tmpRotDist;
    lefttopPos.Y        = Position.Top  + tmpExpPosY    - COS(tmpDegree - Rotation) * tmpRotDist;
    
    tmpRotDist          = math::getDistance((f32)Position.Left, (f32)Position.Bottom, Position.Left+tmpRotPoint.X, Position.Top+tmpRotPoint.Y);
    tmpDegree           = tmpRotDist == 0 ? 0 : ASIN( tmpExtLine / tmpRotDist );
    leftbottomPos.X     = Position.Left + tmpRotPoint.X - SIN(tmpDegree + Rotation) * tmpRotDist;
    leftbottomPos.Y     = Position.Top  + tmpExpPosY    + COS(tmpDegree + Rotation) * tmpRotDist;
    
    tmpExtLine          = (Position.Right-Position.Left) - (Position.Right-Position.Left) * RotationPoint.X;
    tmpRotDist          = math::getDistance((f32)Position.Right, (f32)Position.Top, Position.Left+tmpRotPoint.X, Position.Top+tmpRotPoint.Y);
    tmpDegree           = tmpRotDist == 0 ? 0 : ASIN( tmpExtLine / tmpRotDist );
    righttopPos.X       = Position.Left + tmpRotPoint.X + SIN(tmpDegree + Rotation) * tmpRotDist;
    righttopPos.Y       = Position.Top  + tmpExpPosY    - COS(tmpDegree + Rotation) * tmpRotDist;
    
    tmpRotDist          = math::getDistance((f32)Position.Right, (f32)Position.Bottom, Position.Left+tmpRotPoint.X, Position.Top+tmpRotPoint.Y);
    tmpDegree           = tmpRotDist == 0 ? 0 : ASIN( tmpExtLine / tmpRotDist );
    rightbottomPos.X    = Position.Left + tmpRotPoint.X + SIN(tmpDegree - Rotation) * tmpRotDist;
    rightbottomPos.Y    = Position.Top  + tmpExpPosY    + COS(tmpDegree - Rotation) * tmpRotDist;
    
    /* Set the vertex data */
    SPrimitiveVertex VerticesList[4] = {
        SPrimitiveVertex( lefttopPos.X, lefttopPos.Y, 0.0f, lefttopColor.getSingle(), Clipping.Left, Clipping.Bottom ),
        SPrimitiveVertex( righttopPos.X, righttopPos.Y, 0.0f, righttopColor.getSingle(), Clipping.Right, Clipping.Bottom ),
        SPrimitiveVertex( rightbottomPos.X, rightbottomPos.Y, 0.0f, rightbottomColor.getSingle(), Clipping.Right, Clipping.Top ),
        SPrimitiveVertex( leftbottomPos.X, leftbottomPos.Y, 0.0f, leftbottomColor.getSingle(), Clipping.Left, Clipping.Top )
    };
    
    /* Set the render states */
    pDirect3DDevice_->SetRenderState(D3DRS_FILLMODE, D3DFILL_SOLID);
    
    /* Update the primitive list */
    updatePrimitiveList(VerticesList, 4);
    
    /* Draw the rectangle */
    pDirect3DDevice_->DrawPrimitive(D3DPT_TRIANGLEFAN, 0, 2);
    
    /* Unbind the texture */
    hTexture->unbind();
}

void Direct3D9Driver::draw2DImage(
    Texture* hTexture,
    dim::point2di lefttopPosition, dim::point2di righttopPosition,
    dim::point2di rightbottomPosition, dim::point2di leftbottomPosition,
    dim::point2df lefttopClipping, dim::point2df righttopClipping,
    dim::point2df rightbottomClipping, dim::point2df leftbottomClipping,
    bool isTransFilterUsing,
    color lefttopColor, color righttopColor,
    color rightbottomColor, color leftbottomColor)
{
    /* Set the texture attributes */
    pDirect3DDevice_->SetTextureStageState(0, isTransFilterUsing ? D3DTSS_ALPHAOP : D3DTSS_COLOROP, D3DTOP_MODULATE);
    
    /* Bind the texture */
    hTexture->bind();
    
    /* Set the vertex data */
    SPrimitiveVertex VerticesList[4] = {
        SPrimitiveVertex( (f32)lefttopPosition.X, (f32)lefttopPosition.Y, 0.0f, lefttopColor.getSingle(), lefttopClipping.X, lefttopClipping.Y ),
        SPrimitiveVertex( (f32)righttopPosition.X, (f32)righttopPosition.Y, 0.0f, righttopColor.getSingle(), righttopClipping.X, righttopClipping.Y ),
        SPrimitiveVertex( (f32)rightbottomPosition.X, (f32)rightbottomPosition.Y, 0.0f, rightbottomColor.getSingle(), rightbottomClipping.X, rightbottomClipping.Y ),
        SPrimitiveVertex( (f32)leftbottomPosition.X, (f32)leftbottomPosition.Y, 0.0f, leftbottomColor.getSingle(), leftbottomClipping.X, leftbottomClipping.Y )
    };
    
    /* Set the render states */
    pDirect3DDevice_->SetRenderState(D3DRS_FILLMODE, D3DFILL_SOLID);
    
    /* Update the primitive list */
    updatePrimitiveList(VerticesList, 4);
    
    /* Draw the rectangle */
    pDirect3DDevice_->DrawPrimitive(D3DPT_TRIANGLEFAN, 0, 2);
    
    /* Unbind the texture */
    hTexture->unbind();
}


/* ======= Primitive drawing ======= */


color Direct3D9Driver::getPixel(dim::point2di Position)
{
    return color(0);
}


void Direct3D9Driver::draw2DPixel(dim::point2di Position, color Color)
{
    /* Set the vertex data */
    SPrimitiveVertex VerticesList[1] = {
        SPrimitiveVertex( (f32)Position.X, (f32)Position.Y, 0.0f, Color.getSingle() ),
    };
    
    /* Update the primitive list */
    updatePrimitiveList(VerticesList, 1);
    
    /* Draw the rectangle */
    pDirect3DDevice_->DrawPrimitive(D3DPT_POINTLIST, 0, 1);
}

void Direct3D9Driver::draw2DLine(dim::point2di Position1, dim::point2di Position2, color Color)
{
    draw2DLine(Position1, Position2, Color, Color);
}

void Direct3D9Driver::draw2DLine(dim::point2di Position1, dim::point2di Position2, color Color1, color Color2)
{
    /* Set the vertex data */
    SPrimitiveVertex VerticesList[2] = {
        SPrimitiveVertex( (f32)Position1.X, (f32)Position1.Y, 0.0f, Color1.getSingle() ),
        SPrimitiveVertex( (f32)Position2.X, (f32)Position2.Y, 0.0f, Color2.getSingle() ),
    };
    
    /* Update the primitive list */
    updatePrimitiveList(VerticesList, 2);
    
    /* Draw the rectangle */
    pDirect3DDevice_->DrawPrimitive(D3DPT_LINELIST, 0, 1);
}

void Direct3D9Driver::draw2DLine(dim::point2di Position1, dim::point2di Position2, color Color, s32 DotLength)
{
    
}

void Direct3D9Driver::draw2DRectangle(dim::rect2di Position, color Color, EDrawingModeTypes Type)
{
    draw2DRectangle(Position, Color, Color, Color, Color, Type);
}

void Direct3D9Driver::draw2DRectangle(
    dim::rect2di Position, color lefttopColor, color righttopColor,
    color rightbottomColor, color leftbottomColor, EDrawingModeTypes Type)
{
    /* Set the vertex data */
    SPrimitiveVertex VerticesList[4] = {
        SPrimitiveVertex( (f32)Position.Left, (f32)Position.Top, 0.0f, lefttopColor.getSingle() ),
        SPrimitiveVertex( (f32)Position.Right, (f32)Position.Top, 0.0f, righttopColor.getSingle() ),
        SPrimitiveVertex( (f32)Position.Right, (f32)Position.Bottom, 0.0f, rightbottomColor.getSingle() ),
        SPrimitiveVertex( (f32)Position.Left, (f32)Position.Bottom, 0.0f, leftbottomColor.getSingle() )
    };
    
    /* Set the render states */
    pDirect3DDevice_->SetRenderState(D3DRS_FILLMODE, Type == DRAW_FRAME ? D3DFILL_WIREFRAME : D3DFILL_SOLID);
    
    /* Update the primitive list */
    updatePrimitiveList(VerticesList, 4);
    
    /* Draw the rectangle */
    pDirect3DDevice_->DrawPrimitive(D3DPT_TRIANGLEFAN, 0, 2);
}

void Direct3D9Driver::draw2DRoundRectangle(dim::rect2di Position, dim::size2di CornerRadius, color Color)
{
    
}

void Direct3D9Driver::draw2DCircle(dim::point2di Position, s32 Radius, color Color)
{
    
}

void Direct3D9Driver::draw2DEllipse(dim::point2di Position, dim::size2di Radius, color Color)
{
    
}

void Direct3D9Driver::draw2DEllipse(
    dim::point2di Position, dim::size2di Radius, f32 RotationDegree,
    color Color, bool isSolid)
{
    
}

void Direct3D9Driver::draw2DPolygon(SPrimitiveVertex* VerticesList, u32 AryCount, EDrawingModeTypes Type)
{
    /* Set the render states */
    pDirect3DDevice_->SetRenderState(D3DRS_FILLMODE, Type == DRAW_FRAME ? D3DFILL_WIREFRAME : D3DFILL_SOLID);
    
    /* Update the primitive list */
    updatePrimitiveListFlexible(VerticesList, AryCount);
    
    /* Draw the rectangle */
    pDirect3DDevice_->DrawPrimitive(D3DPT_TRIANGLELIST, 0, AryCount/3);
}

void Direct3D9Driver::draw2DPolygonImage(
    Texture* hTexture, SPrimitiveVertex* VerticesList, u32 AryCount, EDrawingModeTypes Type)
{
    
}

/*
 * ======= Extra drawing functions =======
 */


void Direct3D9Driver::draw2DArc(
    dim::point2di Position, dim::size2di Radius, s32 RadialDegree1,
    s32 RadialDegree2, f32 RotationDegree, color Color)
{
    
}

void Direct3D9Driver::draw2DBezier(
    dim::point2di Position1, dim::point2di Position2,
    dim::point2di Radial1, dim::point2di Radial2,
    color Color, f32 Detail)
{
    dim::point2df Pos1, Pos2;
    Pos1 = math::getBezierValue(
        0.0f, Position1.cast<f32>(), Position2.cast<f32>(), Radial1.cast<f32>(), Radial2.cast<f32>()
    );
    
    for (f32 t = 0.0f; t < 1.0f; t += Detail)
    {
        Pos2 = math::getBezierValue(
            t+Detail, Position1.cast<f32>(), Position2.cast<f32>(), Radial1.cast<f32>(), Radial2.cast<f32>()
        );
        
        draw2DLine(Pos1.cast<s32>(), Pos2.cast<s32>(), Color);
        
        Pos1 = Pos2;
    }
}

void Direct3D9Driver::draw2DSpline(math::Spline2D &Spline, s32 Detail, dim::point2di Pos, color Color)
{
    
}

void Direct3D9Driver::draw2DFloodFill(dim::point2di Position, color FillColor, color BorderColor)
{
    
}


/* ======= 3D drawing functions ======= */


void Direct3D9Driver::draw3DPixel(dim::vector3df Position, color Color, scene::Camera* hCamera)
{
    
}

void Direct3D9Driver::draw3DLine(dim::vector3df Position1, dim::vector3df Position2, color Color, scene::Camera* hCamera)
{
    matrixWorldViewReset();
    
    if (hCamera)
        hCamera->updateModelviewMatrix();
    
    updateModelviewMatrix();
    
    /* Set the FVF (Flexible Vertex Format) */
    pDirect3DDevice_->SetFVF(FVF_VERTEX3D);
    
    /* Set the vertices data */
    SMeshIndexedVertex3D VerticesList[2] = {
        SMeshIndexedVertex3D( Position1.X, Position1.Y, Position1.Z, Color.getSingle() ),
        SMeshIndexedVertex3D( Position2.X, Position2.Y, Position2.Z, Color.getSingle() ),
    };
    
    /* Draw the primitive */
    pDirect3DDevice_->DrawPrimitiveUP(D3DPT_LINELIST, 1, VerticesList, sizeof(SMeshIndexedVertex3D));
}

void Direct3D9Driver::draw3DLine(
    dim::vector3df Position1, dim::vector3df Position2,
    color Color1, color Color2, scene::Camera* hCamera)
{
    matrixWorldViewReset();
    
    if (hCamera)
        hCamera->updateModelviewMatrix();
    
    updateModelviewMatrix();
    
    /* Set the FVF (Flexible Vertex Format) */
    pDirect3DDevice_->SetFVF(FVF_VERTEX3D);
    
    /* Set the vertices data */
    SMeshIndexedVertex3D VerticesList[2] = {
        SMeshIndexedVertex3D( Position1.X, Position1.Y, Position1.Z, Color1.getSingle() ),
        SMeshIndexedVertex3D( Position2.X, Position2.Y, Position2.Z, Color2.getSingle() ),
    };
    
    /* Draw the primitive */
    pDirect3DDevice_->DrawPrimitiveUP(D3DPT_LINELIST, 1, VerticesList, sizeof(SMeshIndexedVertex3D));
}

void Direct3D9Driver::draw3DEllipse(
    dim::vector3df Position, dim::vector3df Rotation, dim::size2df Radius,
    color Color, scene::Camera* hCamera)
{
    
}

void Direct3D9Driver::draw3DTriangle(
    Texture* hTexture, dim::triangle3df Triangle,
    dim::vector3df Position, dim::vector3df Rotation, dim::vector3df Scale,
    color Color, scene::Camera* hCamera)
{
    
}

void Direct3D9Driver::draw3DBox(
    dim::vector3df Corner1, dim::vector3df Corner2,
    dim::vector3df Position, dim::vector3df Rotation, dim::vector3df Scale,
    color Color, scene::Camera* hCamera)
{
    
}

void Direct3D9Driver::draw3DBone(scene::Entity* Obj, dim::matrix4f &BoneTransformation, scene::Camera* hCamera)
{
    matrixWorldViewReset();
    
    /* === Bone location === */
    
    if (hCamera)
        hCamera->updateModelviewMatrix();
    
    Obj->updateModelviewMatrix();
    
    scene::spWorldMatrix *= BoneTransformation;
    
    /* === Render the bone === */
    
}


/*
 * ========== Texture loading & creating ==========
 */


/*
 * Most important function, loads all the file-formats which are integrated
 * (BMP, DIB), (JPG, JPEG, JPE, JFIF), (TGA), (PNG), (PCX)
 */

Texture* Direct3D9Driver::loadTexture(io::stringc Filename)
{
    
    /* Information message */
    io::printMessage(io::stringc("Load texture: \"")+Filename+io::stringc("\""));
    
    /* Allocate some memory & temporary variables */
    
    Texture* LoadedTexture      = 0;
    textureRAWInfo* texture     = 0;
    TextureLoaderRAW* texraw    = 0;
    LPDIRECT3DTEXTURE9 d3dTex   = 0;
    
    io::stringc FileFormat4     = Filename.right(4).upper();
    io::stringc FileFormat5     = Filename.right(5).upper();
    
    u32 TexID                   = 0;
    s32 Type                    = TEXTURE_RGBA; // !!! (TEXTURE_RGB)
    s32 InternalFormat          = 4; // !!! (3)
    u8* ImageData               = 0;
    dim::size2di Size;
    
    /* === Check if the file does exist === */
    
    if (!io::FileLoader().doesFileExist(Filename))
    {
        /* Error message */
        io::printError("could not found texture file");
        
        /* Create a standard empty texture */
        return createTexture(DEF_TEXTURE_SIZE);
    }
    
    /* === Select format === */
    
    if (FileFormat4 == ".BMP" || FileFormat4 == ".DIB") // Windows Bitmap
        texraw = new TextureLoaderBMP(Filename);
    else if (FileFormat4 == ".JPG" || FileFormat5 == ".JPEG" || FileFormat4 == ".JPE" || FileFormat5 == ".JFIF") // Joint Photographic Experts
        texraw = new TextureLoaderJPG(Filename);
    else if (FileFormat4 == ".TGA") // Targa Image
        texraw = new TextureLoaderTGA(Filename);
    else if (FileFormat4 == ".PNG") // Portable Network Graphics
        texraw = new TextureLoaderPNG(Filename);
    else if (FileFormat4 == ".PCX") // Picture Exchange
        texraw = new TextureLoaderPCX(Filename);
    else
    {
        /* Error message */
        io::printError(io::stringc("texture file format: \"")+Filename.getFileExtension().upper()+io::stringc("\" is not supported"));
        
        /* Allocate the texture */
        LoadedTexture = createTexture(DEF_TEXTURE_SIZE);
        
        /* Add the texture to the texture list */
        TextureList_.push_back(LoadedTexture);
        
        /* Return the texture & exit the function */
        return LoadedTexture;
    }
    
    /* === Load the "RAW" texture === */
    
    /* Get the texture raw data */
    if (texraw)
        texture = texraw->loadImageData();
    
    /* Delete the raw texture loader */
    spDeletePointer(texraw);
    
    /* === Create the "Ray" texture === */
    
    if (texture)
    {
        /* Direct3D9 texture configurations */
        if (createRendererTexture(
            TexGenMipMapping_, TEXTURE_2D, dim::vector3di(texture->Width, texture->Height, 1),
            (ETextureImageDataTypes)Type, ImageData))
        {
            /* Settings */
            Type            = texture->Format;
            InternalFormat  = texture->InternalFormat;
            ImageData       = texture->ImageData;
            Size            = dim::size2di(texture->Width, texture->Height);
        }
        else
        {
            /* Delete the loaded raw texture */
            spDeletePointer(texture);
        }
    }
    else
    {
        /* Error message */
        io::printError("could not load texture");
    }
    
    /* === Create "SoftPixel" texture === */
    
    /* Allocate the texture */
    if (texture)
    {
        LoadedTexture = new Direct3D9Texture(
            CurD3DTexture_, Filename, Size, ImageData,
            TexGenRenderMode_, TexGenFilterMode_, InternalFormat, Type, TexGenEdgeSmooth_, TexGenMipMapping_
        );
        
        LoadedTexture->setAnisotropicSamples(TexGenAnisotropy_);
    }
    else
        LoadedTexture = createTexture(DEF_TEXTURE_SIZE);
    
    /* Delete the RAW texture */
    spDeletePointer(texture);
    
    /* Add the texture to the texture list */
    TextureList_.push_back(LoadedTexture);
    
    /* Return the texture & exit the function */
    return LoadedTexture;
    
}

bool Direct3D9Driver::isTexture(Texture* hTexture)
{
    return hTexture && static_cast<Direct3D9Texture*>(hTexture)->getDirect3DTexture();
}

Texture* Direct3D9Driver::copyTexture(Texture* hTexture)
{
    /* Temporary varibales */
    s32 InternalFormat          = hTexture->getFormat();
    s32 Type                    = hTexture->getType();
    dim::size2di Size           = hTexture->getOriginalSize();
    s32 ImageDataSize           = Size.Width * Size.Height * InternalFormat;
    
    Texture* NewTexture         = 0;
    
    /* Allocate new image memory */
    u8* NewImageData = new u8[ImageDataSize];
    
    /* Copy the image data buffers */
    memcpy(NewImageData, hTexture->getImageData(), ImageDataSize);
    
    /* Direct3D9 texture configurations */
    if (createRendererTexture(
        TexGenMipMapping_, hTexture->getDimensionType(),
        dim::vector3di(Size.Width, Size.Height, hTexture->getDepth()),
        (ETextureImageDataTypes)Type, NewImageData))
    {
        
        /* Allocate the new texture */
        NewTexture = new Direct3D9Texture(
            CurD3DTexture_, hTexture->getFilename(), Size, NewImageData, TexGenRenderMode_, TexGenFilterMode_,
            InternalFormat, Type, TexGenEdgeSmooth_, TexGenMipMapping_
        );
    }
    else
    {
        /* Error message */
        io::printError("could not create Direct3D9 texture");
        
        /* Allocate an empty texture */
        NewTexture = new Direct3D9Texture();
    }
    
    /* Add the texture to the texture list */
    TextureList_.push_back(NewTexture);
    
    /* Return the texture & exit the function */
    return NewTexture;
}

Texture* Direct3D9Driver::createTexture(dim::size2di Dimension, u32 TextureDataType, io::stringc Filename)
{
    /* Temporary vairables */
    u32 TexID;
    s32 InternalFormat  = TextureManager::getInternalFormat(TextureDataType);
    s32 ImageDataSize   = Dimension.Width * Dimension.Height * InternalFormat;
    
    register s32 x, y, i, j = 0;
    
    /* Direct3D9 texture configurations */
    if (!createRendererTexture(
        TexGenMipMapping_, TEXTURE_2D, dim::vector3di(Dimension.Width, Dimension.Height, 1),
        (ETextureImageDataTypes)TextureDataType, 0))
    {
        return 0;
    }
    
    /* Allocate new image memory */
    u8* ImageData = new u8[ImageDataSize];
    
    /* Fill the image buffer */
    for (y = 0; y < Dimension.Height; ++y)
    {
        for (x = 0; x < Dimension.Width; ++x)
        {
            for (i = 0; i < InternalFormat; ++i, ++j)
                ImageData[j] = StdFillColor_[i];
        }
    }
    
    /* Allocate an empty texture */
    Direct3D9Texture* NewTexture = new Direct3D9Texture(
        CurD3DTexture_, Filename, Dimension, ImageData, TexGenRenderMode_, TexGenFilterMode_, InternalFormat
    );
    
    /* Add the texture to the texture list */
    TextureList_.push_back(NewTexture);
    
    /* Return the texture & exit the function */
    return NewTexture;
}

Texture* Direct3D9Driver::createTexture(dim::size2di Dimension, u8* ImageData, u32 TextureDataType, io::stringc Filename)
{
    /* Temporary variables */
    Texture* LoadedTexture = 0;
    
    /* Get the internal format */
    s32 InternalFormat = TextureManager::getInternalFormat(TextureDataType);
    
    /* Direct3D9 texture configurations */
    if (createRendererTexture(
        TexGenMipMapping_, TEXTURE_2D, dim::vector3di(Dimension.Width, Dimension.Height, 1),
        (ETextureImageDataTypes)TextureDataType, 0))
    {
        LoadedTexture = new Direct3D9Texture(
            CurD3DTexture_, Filename, Dimension, ImageData,
            TexGenRenderMode_, TexGenFilterMode_, InternalFormat, TextureDataType, TexGenEdgeSmooth_, TexGenMipMapping_
        );
    }
    else
        LoadedTexture = new Direct3D9Texture();
    
    /* Add the texture to the texture list */
    TextureList_.push_back(LoadedTexture);
    
    /* Return the texture & exit the function */
    return LoadedTexture;
}

Texture* Direct3D9Driver::createScreenShot(dim::point2di Position, dim::size2di Size)
{
    
    // !TODO!
    
    /* Allocate the new texture */
    Texture* NewTexture = new Direct3D9Texture();
    
    /* Add the texture to the texture list */
    TextureList_.push_back(NewTexture);
    
    /* Return the texture & exit the function */
    return NewTexture;
}

void Direct3D9Driver::createScreenShot(Texture* hTexture, dim::point2di Position)
{
    /* Get the Direct3D texture handle */
    LPDIRECT3DTEXTURE9 d3dTex = static_cast<Direct3D9Texture*>(hTexture)->getDirect3DTexture();
    
    if (!d3dTex)
        return;
    
    IDirect3DSurface9* Surface;
    
    if (pDirect3DDevice_->CreateOffscreenPlainSurface(
        hTexture->getSize().Width, hTexture->getSize().Height,
        D3DFMT_A8R8G8B8, D3DPOOL_SCRATCH, &Surface, 0))
    {
        io::printError("could not create Direct3D9 offscreen plain surface");
        return;
    }
    
    if (pDirect3DDevice_->GetFrontBufferData(0, Surface))
    {
        Surface->Release();
        io::printError("could not get front buffer data from Direct3D9 surface");
        return;
    }
    
    Surface->Release();
}


/*
 * ========== Font loading & drawing ==========
 */


Font* Direct3D9Driver::loadFont(
    io::stringc FontName, dim::size2di FontSize, s32 TypeFlag, s32 RotationAngle, bool isSymbolUsing,
    bool is3DSpaceUsing, f32 FontDepth, f32 DetailDeviation, EDrawingModeTypes Wireframe)
{
    /* Temporary variables */
    HRESULT Result;
    
    bool isBold       = TypeFlag & FONT_BOLD;
    bool isItalic     = TypeFlag & FONT_ITALIC;
    bool isUnderlined = TypeFlag & FONT_UNDERLINED;
    bool isStrikeout  = TypeFlag & FONT_STRIKEOUT;
    
    /* Create the Direct3D font */
    
    #if D3DX_SDK_VERSION < 24
    
    #ifdef _MSC_VER
    #pragma comment (lib, "d3dx9.lib")
    #endif
    
    Result = D3DXCreateFont(
        pDirect3DDevice_, FontSize.Height, FontSize.Width,
        isBold ? FW_BOLD : 0, 0, isItalic,
        isSymbolUsing ? SYMBOL_CHARSET : ANSI_CHARSET,
        OUT_TT_ONLY_PRECIS, ANTIALIASED_QUALITY, FF_DONTCARE | DEFAULT_PITCH,
        FontName.c_str(), &CurFont_
    );
    
    #else
    
    typedef HRESULT (WINAPI *PFND3DXCREATEFONT)(
        LPDIRECT3DDEVICE9 pDevice, INT Height, UINT Width, UINT Weight, UINT MipLevels, BOOL Italic,
        DWORD CharSet, DWORD OutputPrecision, DWORD Quality, DWORD PitchAndFamily, LPCTSTR pFacename, LPD3DXFONT * ppFont
    );
    
    static PFND3DXCREATEFONT pFncCreateFont = 0;
    
    if (!pFncCreateFont)
    {
        HMODULE hModule = LoadLibrary(d3dDllFileName.c_str());
        
        if (hModule)
        {
            pFncCreateFont = (PFND3DXCREATEFONT)GetProcAddress(hModule, "D3DXCreateFontW");
            
            if (!pFncCreateFont)
            {
                io::printWarning(
                    io::stringc("could not load function \"D3DXCreateFontW\" from Direct3D9 library file: \"")+
                    d3dDllFileName+io::stringc("\", unicode is not supported")
                );
                
                pFncCreateFont = (PFND3DXCREATEFONT)GetProcAddress(hModule, "D3DXCreateFontA");
            }
            
            if (!pFncCreateFont)
            {
                io::printError(
                    io::stringc("could not load function \"D3DXCreateFontA\" from Direct3D9 library file: \"")+
                    d3dDllFileName+io::stringc("\"")
                );
            }
        }
    }
    
    if (pFncCreateFont)
    {
        Result = pFncCreateFont(
            pDirect3DDevice_, FontSize.Height, FontSize.Width,
            TypeFlag & FONT_BOLD ? FW_BOLD : 0, 0, TypeFlag & FONT_ITALIC,
            isSymbolUsing ? SYMBOL_CHARSET : ANSI_CHARSET,
            OUT_TT_ONLY_PRECIS, ANTIALIASED_QUALITY, FF_DONTCARE | DEFAULT_PITCH,
            FontName.c_str(), &CurFont_
        );
    }
    
    #endif
    
    /* Check for errors */
    if (Result != S_OK)
        io::printError("could not load font");
    
    /* Create the formated font */
    HFONT hFont = CreateFont(
        FontSize.Height,                                // Height of font
        FontSize.Width,                                 // Width of font
        RotationAngle,                                  // Angle of escapement
        0,                                              // Orientation angle
        isBold ? FW_BOLD : 0,                           // Font weight
        isItalic,                                       // Italic
        isUnderlined,                                   // Underline
        isStrikeout,                                    // Strikeout
        isSymbolUsing ? SYMBOL_CHARSET : ANSI_CHARSET,  // Character set identifier
        OUT_TT_PRECIS,                                  // Output precision
        CLIP_DEFAULT_PRECIS,                            // Clipping precision
        ANTIALIASED_QUALITY,                            // Output quality
        FF_DONTCARE | DEFAULT_PITCH,                    // Family and pitch
        FontName.c_str()                                // Font name
    );
    
    hDC_ = CurFont_->GetDC();
    
    /* Create a new font */
    return new Font(
        reinterpret_cast<u32>(CurFont_), FontName, FontSize, 0, (s32)hFont, is3DSpaceUsing
    );
}

Font* Direct3D9Driver::loadFont(
    io::stringc FontName, s32 FontSize, s32 TypeFlag, s32 RotationAngle, bool isSymbolUsing,
    bool is3DSpaceUsing, f32 FontDepth, f32 DetailDeviation, EDrawingModeTypes Wireframe)
{
    return loadFont(
        FontName, dim::size2di(0, FontSize), TypeFlag, RotationAngle, isSymbolUsing,
        is3DSpaceUsing, FontDepth, DetailDeviation, Wireframe
    );
}

Font* Direct3D9Driver::loadFont(Texture* hTextureFont, dim::size2di FontSize, dim::size2di ElementCount)
{
    
    return new Font();
}

void Direct3D9Driver::deleteFont(Font* hFont)
{
    /* Check if the font does exist */
    if (hFont)
    {
        /* Get the current font */
        CurFont_ = reinterpret_cast<LPD3DXFONT>(hFont->getID());
        
        /* Release the Direct3D9 font */
        releaseObject(CurFont_);
    }
    
    /* Delete the font */
    spDeletePointer(hFont);
}

void Direct3D9Driver::draw2DText(Font* hFont, dim::point2di Position, io::stringc Text, color Color)
{
    /* Get the font handle */
    LPD3DXFONT d3dFont = reinterpret_cast<LPD3DXFONT>(hFont->getID());
    
    if (!d3dFont)
        return;
    
    /* Temporary variabels */
    RECT rc;
    rc.left = Position.X, rc.top = Position.Y;
    rc.right = __Screen_Width, rc.bottom = __Screen_Height;
    
    /* Draw the text */
    d3dFont->DrawText(
        0, Text.c_str(), Text.size(), &rc, DT_LEFT, Color.getSingle()
    );
}


/*
 * ======= Matrix controll =======
 */

void Direct3D9Driver::updateModelviewMatrix()
{
    pDirect3DDevice_->SetTransform(D3DTS_VIEW, D3D_MATRIX(scene::spViewMatrix));
    pDirect3DDevice_->SetTransform(D3DTS_WORLD, D3D_MATRIX(scene::spWorldMatrix));
}

void Direct3D9Driver::setProjectionMatrix(const dim::matrix4f &Matrix)
{
    scene::spProjectionMatrix = Matrix;
    pDirect3DDevice_->SetTransform(D3DTS_PROJECTION, D3D_MATRIX(Matrix));
}
void Direct3D9Driver::setViewMatrix(const dim::matrix4f &Matrix)
{
    scene::spViewMatrix = Matrix;
    pDirect3DDevice_->SetTransform(D3DTS_VIEW, D3D_MATRIX(Matrix));
}
void Direct3D9Driver::setWorldMatrix(const dim::matrix4f &Matrix)
{
    scene::spWorldMatrix = Matrix;
    pDirect3DDevice_->SetTransform(D3DTS_WORLD, D3D_MATRIX(Matrix));
}
void Direct3D9Driver::setTextureMatrix(const dim::matrix4f &Matrix, u8 TextureLayer)
{
    scene::spTextureMatrix[TextureLayer] = Matrix;
    pDirect3DDevice_->SetTransform(
        (D3DTRANSFORMSTATETYPE)(D3DTS_TEXTURE0 + TextureLayer), D3D_MATRIX(Matrix.getTextureMatrix())
    );
}
void Direct3D9Driver::setColorMatrix(const dim::matrix4f &Matrix)
{
    scene::spColorMatrix = Matrix;
}


/*
 * ========== Other renderer option fuctions =======
 */


void Direct3D9Driver::setViewport(s32 X, s32 Y, s32 Width, s32 Height)
{
    D3DVIEWPORT9 d3dViewport;
    {
        d3dViewport.X       = X;
        d3dViewport.Y       = Y;
        d3dViewport.Width   = Width;
        d3dViewport.Height  = Height;
        d3dViewport.MinZ    = 0.0f;
        d3dViewport.MaxZ    = 1.0f;
    }
    pDirect3DDevice_->SetViewport(&d3dViewport);
}



/*
 * ========== Private functions =======
 */

void Direct3D9Driver::init()
{
    /* General settings */
    CurSamplerLevel_ = 0;
    
    /* Clear the presenter member */
    ZeroMemory(&Presenter_, sizeof(Presenter_));
    
    /* General settings */
    MaxClippingPlanes_ = 8; // !!!
}

void Direct3D9Driver::clear()
{
    /* Close and release the standard- & flexible vertex buffer */
    releaseObject(pDirect3DVertexBuffer_);
    releaseObject(pDirect3DFlexibleVertexBuffer_);
    
    /* Close and release the 3D device */
    releaseObject(pDirect3DDevice_);
    
    /* Close and release Direct3D */
    releaseObject(pDirect3D_);
}

void Direct3D9Driver::updatePrimitiveList(SPrimitiveVertex* pVerticesList, u32 Size)
{
    /* Fill the standard vertex buffer */
    VOID* pVoid;
    pDirect3DVertexBuffer_->Lock(0, sizeof(SPrimitiveVertex)*Size, (void**)&pVoid, 0);
    memcpy(pVoid, pVerticesList, sizeof(SPrimitiveVertex)*Size);
    pDirect3DVertexBuffer_->Unlock();
    
    /* Set the FVF (Flexible Vertex Format) */
    pDirect3DDevice_->SetFVF(FVF_VERTEX2D);
    
    /* Set the stream souce */
    pDirect3DDevice_->SetStreamSource(0, pDirect3DVertexBuffer_, 0, sizeof(SPrimitiveVertex));
}

void Direct3D9Driver::updatePrimitiveListFlexible(SPrimitiveVertex* pVerticesList, u32 Count)
{
    /* Delete the old vertex buffer */
    releaseObject(pDirect3DFlexibleVertexBuffer_);
    
    /* Create a new vertex buffer */
    pDirect3DDevice_->CreateVertexBuffer(
        sizeof(SPrimitiveVertex)*Count,
        0,
        FVF_VERTEX2D,
        D3DPOOL_DEFAULT,
        &pDirect3DFlexibleVertexBuffer_,
        0
    );
    
    if (!pDirect3DFlexibleVertexBuffer_)
    {
        io::printError("Could not create Direct3D9 vertex buffer");
        return;
    }
    
    /* Fill the standard vertex buffer */
    VOID* pVoid;
    pDirect3DFlexibleVertexBuffer_->Lock(0, sizeof(SPrimitiveVertex)*Count, (void**)&pVoid, 0);
    memcpy(pVoid, pVerticesList, sizeof(SPrimitiveVertex)*Count);
    pDirect3DFlexibleVertexBuffer_->Unlock();
    
    /* Set the FVF (Flexible Vertex Format) */
    pDirect3DDevice_->SetFVF(FVF_VERTEX2D);
    
    /* Set the stream souce */
    pDirect3DDevice_->SetStreamSource(0, pDirect3DFlexibleVertexBuffer_, 0, sizeof(SPrimitiveVertex));
}


bool Direct3D9Driver::checkDepthStencilFormat()
{
    return pDirect3D_->CheckDeviceFormat(
        D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, Presenter_.BackBufferFormat,
        D3DUSAGE_DEPTHSTENCIL, D3DRTYPE_SURFACE, Presenter_.AutoDepthStencilFormat) == D3D_OK;
}


void Direct3D9Driver::setTextureParameters(
    ETextureDimensions Dimension, s32 RenderMode, s32 FilterMode, f32 MaxAnisotropy, bool isMipMapping, bool isEdgeSmooth)
{
    /* Check if the edges are shall be smooth (e.g. for sky textrues) */
    if (isEdgeSmooth)
    {
        /* Clamp (smooth edges) */
        pDirect3DDevice_->SetSamplerState(CurSamplerLevel_, D3DSAMP_ADDRESSU, D3DTADDRESS_CLAMP);
        pDirect3DDevice_->SetSamplerState(CurSamplerLevel_, D3DSAMP_ADDRESSV, D3DTADDRESS_CLAMP);
    }
    else
    {
        /* Wrap (normal) */
        pDirect3DDevice_->SetSamplerState(CurSamplerLevel_, D3DSAMP_ADDRESSU, D3DTADDRESS_WRAP);
        pDirect3DDevice_->SetSamplerState(CurSamplerLevel_, D3DSAMP_ADDRESSV, D3DTADDRESS_WRAP);
    }
    
    /* Anisotropy */
    pDirect3DDevice_->SetSamplerState(CurSamplerLevel_, D3DSAMP_MAXANISOTROPY, (DWORD)MaxAnisotropy);
    
    /* Texture filter */
    if (isMipMapping)
    {
        switch (FilterMode)
        {
            case TEXTURE_BILINEAR:
                pDirect3DDevice_->SetSamplerState(CurSamplerLevel_, D3DSAMP_MIPFILTER, D3DTEXF_POINT); break;
            case TEXTURE_TRILINEAR:
                pDirect3DDevice_->SetSamplerState(CurSamplerLevel_, D3DSAMP_MIPFILTER, D3DTEXF_LINEAR); break;
            case TEXTURE_ANISOTROPIC:
                pDirect3DDevice_->SetSamplerState(CurSamplerLevel_, D3DSAMP_MIPFILTER, D3DTEXF_ANISOTROPIC); break;
        }
    }
    else
        pDirect3DDevice_->SetSamplerState(CurSamplerLevel_, D3DSAMP_MIPFILTER, D3DTEXF_NONE);
    
    /* Render mode */
    D3DTEXTUREFILTERTYPE Filter = D3DTEXF_NONE;
    
    if (RenderMode == TEXTURE_SMOOTH)
        Filter = (FilterMode == TEXTURE_ANISOTROPIC ? D3DTEXF_ANISOTROPIC : D3DTEXF_LINEAR);
    else if (RenderMode == TEXTURE_LINEAR)
        Filter = (FilterMode == TEXTURE_ANISOTROPIC ? D3DTEXF_ANISOTROPIC : D3DTEXF_POINT);
    
    pDirect3DDevice_->SetSamplerState(CurSamplerLevel_, D3DSAMP_MINFILTER, Filter);
    pDirect3DDevice_->SetSamplerState(CurSamplerLevel_, D3DSAMP_MAGFILTER, Filter);
}

bool Direct3D9Driver::createRendererTexture(
    bool MipMaps, ETextureDimensions Dimension, dim::vector3di Size, ETextureImageDataTypes Format, const u8* ImageData)
{
    D3DFORMAT d3dFormat = D3DFMT_A8R8G8B8; // !!! (D3DFMT_X8R8G8B8)
    
    /* Register a new Direct3D9 texture */
    HRESULT d3dError = pDirect3DDevice_->CreateTexture(
        Size.X,
        Size.Y,
        MipMaps ? 0 : 1,
        MipMaps ? D3DUSAGE_AUTOGENMIPMAP : 0,
        d3dFormat,
        D3DPOOL_MANAGED,
        &CurD3DTexture_,
        0
    );
    
    /* Check if an error has been detected */
    if (d3dError)
    {
        io::printError("could not create Direct3D9 texture");
        return false;
    }
    
    return true;
}


#endif


} // /namespace video

} // /namespace sp






// ================================================================================
