/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * Direct3D9 High Level Shading Language (HLSL)
 *
 */

#include "spDirect3D9Shader.hpp"
#include "spDirect3D9Driver.hpp"
#include "spMainDevice.hpp"


namespace sp
{

extern video::VideoDriver* __spVideoDriver;

namespace video
{


#if defined(WIN32)

const c8* const d3d9VertexShaderTypeNames[] =
{
    "vs_1_1",
    "vs_2_0",
    "vs_2_a",
    "vs_3_0",
    0
};

const c8* const d3d9PixelShaderTypeNames[] =
{
    "ps_1_1",
    "ps_1_2",
    "ps_1_3",
    "ps_1_4",
    "ps_2_0",
    "ps_2_a",
    "ps_2_b",
    "ps_3_0",
    0
};


Direct3D9Shader::Direct3D9Shader()
    : VertexShader_(0), PixelShader_(0), VSConstantsTable_(0), PSConstantsTable_(0)
{
    init();
}
Direct3D9Shader::Direct3D9Shader(EStandardShaderTypes Type)
    : VertexShader_(0), PixelShader_(0), VSConstantsTable_(0), PSConstantsTable_(0)
{
    init();
}
Direct3D9Shader::Direct3D9Shader(
    io::stringc FilenameVertex, io::stringc FilenamePixel,
    io::stringc EntryPointVertex, io::stringc EntryPointPixel,
    EHLSLVertexShaderTypes CompilerTargetVertex, EHLSLPixelShaderTypes CompilerTargetPixel,
    bool UseHighLevel)
    : VertexShader_(0), PixelShader_(0), VSConstantsTable_(0), PSConstantsTable_(0)
{
    init();
    
    createShader(
        FilenameVertex, FilenamePixel, EntryPointVertex, EntryPointPixel,
        d3d9VertexShaderTypeNames[CompilerTargetVertex], d3d9PixelShaderTypeNames[CompilerTargetPixel], UseHighLevel
    );
}
Direct3D9Shader::~Direct3D9Shader()
{
    clear();
}


void Direct3D9Shader::bind()
{
    if (pShaderCallbackProc_)
        pShaderCallbackProc_(this);
    if (VertexShader_)
        pD3D9Device_->SetVertexShader(VertexShader_);
    if (PixelShader_)
        pD3D9Device_->SetPixelShader(PixelShader_);
}

void Direct3D9Shader::unbind()
{
    pD3D9Device_->SetVertexShader(0);
    pD3D9Device_->SetPixelShader(0);
}


bool Direct3D9Shader::setConstant(io::stringc Name, const f32 Value)
{
    LPD3DXCONSTANTTABLE Table = (CurrentShader_ == SHADER_VERTEX ? VSConstantsTable_ : PSConstantsTable_);
    
    D3DXHANDLE Handle = Table->GetConstantByName(0, Name.c_str());
    
    if (!Handle)
    {
        io::printError(io::stringc("could not set HLSL constant: \"")+Name+io::stringc("\""));
        return false;
    }
    
    if (Table->SetFloat(pD3D9Device_, Handle, Value))
    {
        io::printError("setting HLSL constant 'float' failed");
        return false;
    }
    
    return true;
}

bool Direct3D9Shader::setConstant(io::stringc Name, dim::vector3df Position)
{
    return setConstant(Name, &Position.X, 3);
}

bool Direct3D9Shader::setConstant(io::stringc Name, video::color Color)
{
    f32 FloatColor[4] = { (f32)Color.Red/255, (f32)Color.Green/255, (f32)Color.Blue/255, (f32)Color.Alpha/255 };
    
    return setConstant(Name, FloatColor, 4);
}

bool Direct3D9Shader::setConstant(io::stringc Name, dim::matrix4f &Matrix)
{
    return setConstant(Name, Matrix.getArray16(), 16);
}

bool Direct3D9Shader::setConstant(io::stringc Name, const f32* pFloats, int Count)
{
    LPD3DXCONSTANTTABLE Table = (CurrentShader_ == SHADER_VERTEX ? VSConstantsTable_ : PSConstantsTable_);
    
    D3DXHANDLE Handle = Table->GetConstantByName(0, Name.c_str());
    
    if (!Handle)
    {
        io::printError(io::stringc("could not set HLSL constant: \"")+Name+io::stringc("\""));
        return false;
    }
    
    if (Table->SetFloatArray(pD3D9Device_, Handle, pFloats, Count))
    {
        io::printError("setting HLSL constant 'float array' failed");
        return false;
    }
    
    return true;
}


bool Direct3D9Shader::setConstant(const f32* pFloats, s32 StartRegister, s32 ConstAmount)
{
    return pD3D9Device_->SetVertexShaderConstantF(StartRegister, pFloats, ConstAmount) == D3D_OK;
}


/*
 * ========== Private: ==========
 */

void Direct3D9Shader::init()
{
    /* Get the Direct3D9 device */
    pD3D9Device_ = static_cast<video::Direct3D9Driver*>(__spVideoDriver)->getDirect3DDevice();
}
void Direct3D9Shader::clear()
{
    /* Release the shader objects */
    
    if (VertexShader_)
        VertexShader_->Release();
    if (PixelShader_)
        PixelShader_->Release();
    
    if (VSConstantsTable_)
        VSConstantsTable_->Release();
    if (PSConstantsTable_)
        PSConstantsTable_->Release();
}


void Direct3D9Shader::createShader(
    io::stringc FilenameVS, io::stringc FilenamePS,
    io::stringc EntryPointVS, io::stringc EntryPointPS,
    io::stringc CompilerTargetVS, io::stringc CompilerTargetPS,
    bool UseHighLevel)
{
    c8* pDataVS = 0, * pDataPS = 0;
    
    loadShaderFile(FilenameVS, pDataVS);
    loadShaderFile(FilenamePS, pDataPS);
    
    if (UseHighLevel)
    {
        createVertexShaderHLSL(pDataVS, EntryPointVS.c_str(), CompilerTargetVS.c_str());
        createPixelShaderHLSL(pDataVS, EntryPointPS.c_str(), CompilerTargetPS.c_str());
    }
    else
    {
        createVertexShader(pDataVS);
        createPixelShader(pDataVS);
    }
}

bool Direct3D9Shader::loadShaderFile(io::stringc Filename, c8* &pData)
{
    io::stringc str;
    
    io::FileLoader Loader;
    io::File* hFile = Loader.openFile(Filename, io::FILE_READ_BINARY);
    
    if (!hFile)
        return false;
    
    while (!Loader.isEOF(hFile))
        str += Loader.readString(hFile) + io::stringc("\n");
    
    pData = new c8[str.size()];
    
    memcpy(pData, str.c_str(), str.size());
    
    Loader.closeFile(hFile);
    
    return true;
}


bool Direct3D9Shader::createVertexShader(const c8* strShd)
{
    if (!strShd)
        return false;
    
    /* Temporary variables */
    LPD3DXBUFFER code   = 0;
    LPD3DXBUFFER errors = 0;
    
    /* Compile the shader */
    d3dAssembleShader(strShd, strlen(strShd), 0, 0, 0, &code, &errors);
    
    /* Check for errors */
    if (errors)
    {
        /* Print the error messages */
        io::printInfo("Direct3D9 vertex shader compilation failed:", io::INFO_WARNING);
        io::printInfo((c8*)errors->GetBufferPointer(), io::INFO_WARNING);
        
        /* Release all allocated objects */
        if (code)
            code->Release();
        errors->Release();
        
        return false;
    }
    
    if (!code)
        return false;
    
    /* Create the vertex shader */
    if (pD3D9Device_->CreateVertexShader((DWORD*)code->GetBufferPointer(), &VertexShader_))
    {
        code->Release();
        return false;
    }
    
    code->Release();
    
    return true;
}

bool Direct3D9Shader::createPixelShader(const c8* strShd)
{
    if (!strShd)
        return false;
    
    /* Temporary variables */
    LPD3DXBUFFER code   = 0;
    LPD3DXBUFFER errors = 0;
    
    /* Compile the shader */
    d3dAssembleShader(strShd, strlen(strShd), 0, 0, 0, &code, &errors);
    
    /* Check for errors */
    if (errors)
    {
        /* Print the error messages */
        io::printInfo("Direct3D9 pixel shader compilation failed:", io::INFO_WARNING);
        io::printInfo((c8*)errors->GetBufferPointer(), io::INFO_WARNING);
        
        /* Release all allocated objects */
        if (code)
            code->Release();
        errors->Release();
        
        return false;
    }
    
    if (!code)
        return false;
    
    /* Create the pixel shader */
    if (pD3D9Device_->CreatePixelShader((DWORD*)code->GetBufferPointer(), &PixelShader_))
    {
        code->Release();
        return false;
    }
    
    code->Release();
    
    return true;
}


bool Direct3D9Shader::createVertexShaderHLSL(const c8* strShd, const c8* shdEntryPoint, const c8* shdTargetName)
{
    if (!strShd)
        return true;
    
    /* Temporary variables */
    LPD3DXBUFFER buffer = 0;
    LPD3DXBUFFER errors = 0;
    
    /* Compile the shader */
    HRESULT Result = d3dCompileShader(
        strShd, strlen(strShd), 0, 0, shdEntryPoint, shdTargetName, 0, &buffer, &errors, &VSConstantsTable_
    );
    
    /* Check for errors */
    if (Result)
    {
        io::printInfo("Direct3D9 HLSL vertex shader compilation failed:", io::INFO_WARNING);
        
        if (errors)
        {
            /* Print the error messages */
            io::printInfo((c8*)errors->GetBufferPointer(), io::INFO_WARNING);
            
            /* Release all allocated objects */
            if (buffer)
                buffer->Release();
            errors->Release();
        }
        
        return false;
    }
    
    if (errors)
        errors->Release();
    
    if (buffer)
    {
        /* Create the vertex shader */
        if (pD3D9Device_->CreateVertexShader((DWORD*)buffer->GetBufferPointer(), &VertexShader_))
        {
            buffer->Release();
            return false;
        }
        
        buffer->Release();
        
        return true;
    }
    
    return false;
}

bool Direct3D9Shader::createPixelShaderHLSL(const c8* strShd, const c8* shdEntryPoint, const c8* shdTargetName)
{
    if (!strShd)
        return true;
    
    /* Temporary variables */
    LPD3DXBUFFER buffer = 0;
    LPD3DXBUFFER errors = 0;
    
    /* Compile the shader */
    HRESULT Result = d3dCompileShader(
        strShd, strlen(strShd), 0, 0, shdEntryPoint, shdTargetName, 0, &buffer, &errors, &PSConstantsTable_
    );
    
    /* Check for errors */
    if (Result)
    {
        io::printInfo("Direct3D9 HLSL pixel shader compilation failed:", io::INFO_WARNING);
        
        if (errors)
        {
            /* Print the error messages */
            io::printInfo((c8*)errors->GetBufferPointer(), io::INFO_WARNING);
            
            /* Release all allocated objects */
            if (buffer)
                buffer->Release();
            errors->Release();
        }
        
        return false;
    }
    
    if (errors)
        errors->Release();
    
    if (buffer)
    {
        /* Create the pixel shader */
        if (pD3D9Device_->CreatePixelShader((DWORD*)buffer->GetBufferPointer(), &PixelShader_))
        {
            buffer->Release();
            return false;
        }
        
        buffer->Release();
        
        return true;
    }
    
    return false;
}


HRESULT Direct3D9Shader::d3dAssembleShader(
    LPCSTR pSrcData, UINT SrcDataLen, CONST D3DXMACRO* pDefines,
    LPD3DXINCLUDE pInclude, DWORD Flags, LPD3DXBUFFER* ppShader, LPD3DXBUFFER* ppErrorMsgs)
{
    #if D3DX_SDK_VERSION < 24
    
    #ifdef _MSC_VER
    #pragma comment (lib, "d3dx9.lib")
    #endif
    
    return D3DXAssembleShader(
        pSrcData, SrcDataLen, pDefines, pInclude, Flags, ppShader, ppErrorMsgs
    );
    
    #else
    
    typedef HRESULT (WINAPI *PFND3DXASSEMBLESHADER)(
        LPCSTR pSrcData, UINT SrcDataLen, CONST D3DXMACRO* pDefines,
        LPD3DXINCLUDE pInclude, DWORD Flags, LPD3DXBUFFER* ppShader, LPD3DXBUFFER* ppErrorMsgs
    );
    
    static PFND3DXASSEMBLESHADER pFncAssembleShader = 0;
    
    if (!pFncAssembleShader)
    {
        HMODULE hModule = LoadLibrary(d3dDllFileName.c_str());
        
        if (hModule)
            pFncAssembleShader = (PFND3DXASSEMBLESHADER)GetProcAddress(hModule, "D3DXAssembleShader");
        
        if (!pFncAssembleShader)
        {
            io::printError(
                io::stringc("could not load function \"D3DXAssembleShader\" from Direct3D9 library file: \"")+
                d3dDllFileName+io::stringc("\"")
            );
        }
    }
    
    if (pFncAssembleShader)
    {
        return pFncAssembleShader(
            pSrcData, SrcDataLen, pDefines, pInclude, Flags, ppShader, ppErrorMsgs
        );
    }
    
    #endif
    
    return 0;
}

HRESULT Direct3D9Shader::d3dCompileShader(
    LPCSTR pSrcData, UINT SrcDataLen, CONST D3DXMACRO* pDefines,
    LPD3DXINCLUDE pInclude, LPCSTR pFunctionName,
    LPCSTR pProfile, DWORD Flags, LPD3DXBUFFER* ppShader,
    LPD3DXBUFFER* ppErrorMsgs, LPD3DXCONSTANTTABLE* ppConstantTable)
{
    #if D3DX_SDK_VERSION < 24
    
    #ifdef _MSC_VER
    #pragma comment (lib, "d3dx9.lib")
    #endif
    
    return D3DXCompileShader(
        pSrcData, SrcDataLen, pDefines, pInclude, pFunctionName, pProfile, Flags, ppShader, ppErrorMsgs, ppConstantTable
    );
    
    #else
    
    typedef HRESULT (WINAPI *PFND3DXCOMPILESHADER)(
        LPCSTR pSrcData, UINT SrcDataLen, CONST D3DXMACRO* pDefines,
        LPD3DXINCLUDE pInclude, LPCSTR pFunctionName,
        LPCSTR pProfile, DWORD Flags, LPD3DXBUFFER* ppShader,
        LPD3DXBUFFER* ppErrorMsgs, LPD3DXCONSTANTTABLE* ppConstantTable
    );
    
    static PFND3DXCOMPILESHADER pFncCompileShader = 0;
    
    if (!pFncCompileShader)
    {
        HMODULE hModule = LoadLibrary(d3dDllFileName.c_str());
        
        if (hModule)
            pFncCompileShader = (PFND3DXCOMPILESHADER)GetProcAddress(hModule, "D3DXCompileShader");
        
        if (!pFncCompileShader)
        {
            io::printError(
                io::stringc("could not load function \"D3DXCompileShader\" from Direct3D9 library file: \"")+
                d3dDllFileName+io::stringc("\"")
            );
        }
    }
    
    if (pFncCompileShader)
    {
        return pFncCompileShader(
            pSrcData, SrcDataLen, pDefines, pInclude, pFunctionName, pProfile, Flags, ppShader, ppErrorMsgs, ppConstantTable
        );
    }
    
    #endif
    
    return 0;
}




#endif


} // /namespace scene

} // /namespace sp








// ================================================================================
