/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * Math operation functions
 *
 */

#include "spMath.hpp"
#include "spInputOutputStandard.hpp"
#include "spDimension.hpp"


namespace sp
{
namespace math
{


/****************************************
 *                                      *
 *  Collision detection math functions  *
 *                                      *
 ****************************************/


SP_EXPORT dim::point2df getLinePointDistance(dim::point2df PointA, dim::point2df PointB, dim::point2df PointP, bool isOnlyNormal)
{
    #define GETNEARESTPOINT                                                             \
        if (math::getDistanceSq(PointP, PointA) < math::getDistanceSq(PointP, PointB))  \
            return PointA;                                                              \
        else                                                                            \
            return PointB;
    
    if (PointB.X - PointA.X == 0.0f)
    {
        if (PointP.Y < Min(PointA.Y, PointB.Y) || PointP.Y > Max(PointA.Y, PointB.Y))
            GETNEARESTPOINT;
        
        return dim::point2df(PointA.X, PointP.Y);
    }
    else if (PointB.Y - PointA.Y == 0.0f)
    {
        if (PointP.X < Min(PointA.X, PointB.X) || PointP.X > Max(PointA.X, PointB.X))
            GETNEARESTPOINT;
        
        return dim::point2df(PointP.X, PointA.Y);
    }
    
    f32 a, b, h, alpha;
    
    dim::point2df PointQ;
    
    a = PointP.Y - ( (PointB.Y - PointA.Y) * (PointP.X - PointA.X) / (PointB.X - PointA.X) ) - PointA.Y;
    b = PointP.X - ( (PointB.X - PointA.X) * (PointP.Y - PointA.Y) / (PointB.Y - PointA.Y) ) - PointA.X;
    
    alpha = ASIN( b / sqrt(a*a + b*b) );
    
    if (PointB.X > PointA.X)
        alpha = -alpha;
    if (PointB.Y > PointA.Y)
        alpha = -alpha;
    
    if ( ( h = COS(alpha)*b ) > 0.0f )
    {
        PointQ.X = PointP.X - SIN(90 - alpha) * h;
        PointQ.Y = PointP.Y + COS(90 - alpha) * h;
    }
    else
    {
        PointQ.X = PointP.X + SIN(alpha - 90) * h;
        PointQ.Y = PointP.Y - COS(alpha - 90) * h;
    }
    
    if ( ! isOnlyNormal &&
         ! ( ( ( PointQ.X >= PointA.X && PointQ.X <= PointB.X ) || ( PointQ.X >= PointB.X && PointQ.X <= PointA.X ) ) &&
             ( ( PointQ.Y >= PointA.Y && PointQ.Y <= PointB.Y ) || ( PointQ.Y >= PointB.Y && PointQ.Y <= PointA.Y ) ) ) )
    {
        GETNEARESTPOINT;
    }
    
    return PointQ;
    
    #undef GETNEARESTPOINT
}

SP_EXPORT dim::vector3df getLinePointDistance(dim::vector3df PointA, dim::vector3df PointB, dim::vector3df PointP, bool isOnlyNormal)
{
    dim::point2df Pos2dA = getLinePointDistance(
        dim::point2df(PointA.X, PointA.Y), dim::point2df(PointB.X, PointB.Y), dim::point2df(PointP.X, PointP.Y), isOnlyNormal
    );
    dim::point2df Pos2dB = getLinePointDistance(
        dim::point2df(PointA.Y, PointA.Z), dim::point2df(PointB.Y, PointB.Z), dim::point2df(PointP.Y, PointP.Z), isOnlyNormal
    );
    
    return dim::vector3df(Pos2dA.X, Pos2dA.Y, Pos2dB.Y);
}


SP_EXPORT dim::vector3df getClosestPoint(const dim::triangle3df &Triangle, const dim::vector3df &Point)
{
    dim::vector3df ab(Triangle.PointB - Triangle.PointA);
    dim::vector3df ac(Triangle.PointC - Triangle.PointA);
    
    /* Check if P is in vertex region outside A */
    dim::vector3df ap = Point - Triangle.PointA;
    f32 d1 = ab.getScalarProduct(ap);
    f32 d2 = ac.getScalarProduct(ap);
    if (d1 <= 0.0f && d2 <= 0.0f)
        return Triangle.PointA;
    
    /* Check if P is in vertex region outside B */
    dim::vector3df bp = Point - Triangle.PointB;
    f32 d3 = ab.getScalarProduct(bp);
    f32 d4 = ac.getScalarProduct(bp);
    if (d3 >= 0.0f && d4 <= d3)
        return Triangle.PointB;
    
    /* Check if P is in edge region of AB, of so return projection of P onto AB */
    f32 vc = d1*d4 - d3*d2;
    if (vc <= 0.0f && d1 >= 0.0f && d3 <= 0.0f)
    {
        f32 v = d1 / (d1-d3);
        return Triangle.PointA + ab * v;
    }
    
    /* Check if P is in vertex region outside C */
    dim::vector3df cp = Point - Triangle.PointC;
    f32 d5 = ab.getScalarProduct(cp);
    f32 d6 = ac.getScalarProduct(cp);
    if (d6 >= 0.0f && d5 <= d6)
        return Triangle.PointC;
    
    /* Check if P is in edge region of AC, if so return projection of P onto AC */
    f32 vb = d5*d2 - d1*d6;
    if (vb <= 0.0f && d2 >= 0.0f && d6 <= 0.0f)
    {
        f32 w = d2 / (d2-d6);
        return Triangle.PointA + ac * w;
    }
    
    /* Check if P is in edge region of BC, if so return projection of P onto BC */
    f32 va = d3*d6 - d5*d4;
    if (va <= 0.0f && (d4-d3) >= 0.0f && (d5-d6) >= 0.0f)
    {
        f32 w = (d4-d3) / ((d4-d3) + (d5-d6));
        return Triangle.PointB + (Triangle.PointC-Triangle.PointB) * w;
    }
    
    /* P is inside face region. Compute Q through its barycentric coordinates (u, v, w) */
    f32 denom = 1.0f / (va + vb + vc);
    f32 v = vb * denom;
    f32 w = vc * denom;
    
    return Triangle.PointA + ab * v + ac * w;
}


SP_EXPORT bool getNormalPoint(const dim::triangle3df &Triangle, const dim::vector3df &Point, dim::vector3df &NormalPoint)
{
    dim::vector3df ab(Triangle.PointB - Triangle.PointA);
    dim::vector3df ac(Triangle.PointC - Triangle.PointA);
    
    /* Check if P is in vertex region outside A */
    dim::vector3df ap = Point - Triangle.PointA;
    f32 d1 = ab.getScalarProduct(ap);
    f32 d2 = ac.getScalarProduct(ap);
    if (d1 <= 0.0f && d2 <= 0.0f)
        return false;
    
    /* Check if P is in vertex region outside B */
    dim::vector3df bp = Point - Triangle.PointB;
    f32 d3 = ab.getScalarProduct(bp);
    f32 d4 = ac.getScalarProduct(bp);
    if (d3 >= 0.0f && d4 <= d3)
        return false;
    
    /* Check if P is in edge region of AB, of so return projection of P onto AB */
    f32 vc = d1*d4 - d3*d2;
    if (vc <= 0.0f && d1 >= 0.0f && d3 <= 0.0f)
        return false;
    
    /* Check if P is in vertex region outside C */
    dim::vector3df cp = Point - Triangle.PointC;
    f32 d5 = ab.getScalarProduct(cp);
    f32 d6 = ac.getScalarProduct(cp);
    if (d6 >= 0.0f && d5 <= d6)
        return false;
    
    /* Check if P is in edge region of AC, if so return projection of P onto AC */
    f32 vb = d5*d2 - d1*d6;
    if (vb <= 0.0f && d2 >= 0.0f && d6 <= 0.0f)
        return false;
    
    /* Check if P is in edge region of BC, if so return projection of P onto BC */
    f32 va = d3*d6 - d5*d4;
    if (va <= 0.0f && (d4-d3) >= 0.0f && (d5-d6) >= 0.0f)
        return false;
    
    /* P is inside face region. Compute Q through its barycentric coordinates (u, v, w) */
    f32 denom = 1.0f / (va + vb + vc);
    f32 v = vb * denom;
    f32 w = vc * denom;
    
    NormalPoint = Triangle.PointA + ab * v + ac * w;
    
    return true;
}

/*

       |     |
        back 
       |     |
- - - -*-----*- - - -
       |     |      
 left  |top  | right
       |btm. |
- - - -*-----*- - - -
       |     |
        front
       |     |

*/

SP_EXPORT dim::vector3df getClosestPoint(dim::line3df &Box, dim::vector3df &Point, dim::vector3df &Normal)
{
    /* === Sides === */
    
    /* Check nearest point on side: left */
    if (Point.X <= Box.Corner1.X && Point.Y > Box.Corner1.Y && Point.Y < Box.Corner2.Y && Point.Z > Box.Corner1.Z && Point.Z < Box.Corner2.Z)
    {
        Normal.set(-1, 0, 0);
        return dim::vector3df(Box.Corner1.X, Point.Y, Point.Z);
    }
    
    /* Check nearest point on side: right */
    if (Point.X >= Box.Corner2.X && Point.Y > Box.Corner1.Y && Point.Y < Box.Corner2.Y && Point.Z > Box.Corner1.Z && Point.Z < Box.Corner2.Z)
    {
        Normal.set(1, 0, 0);
        return dim::vector3df(Box.Corner2.X, Point.Y, Point.Z);
    }
    
    /* Check nearest point on side: front */
    if (Point.Z <= Box.Corner1.Z && Point.Y > Box.Corner1.Y && Point.Y < Box.Corner2.Y && Point.X > Box.Corner1.X && Point.X < Box.Corner2.X)
    {
        Normal.set(0, 0, -1);
        return dim::vector3df(Point.X, Point.Y, Box.Corner1.Z);
    }
    
    /* Check nearest point on side: back */
    if (Point.Z >= Box.Corner2.Z && Point.Y > Box.Corner1.Y && Point.Y < Box.Corner2.Y && Point.X > Box.Corner1.X && Point.X < Box.Corner2.X)
    {
        Normal.set(0, 0, 1);
        return dim::vector3df(Point.X, Point.Y, Box.Corner2.Z);
    }
    
    /* Check nearest point on side: top */
    if (Point.Y >= Box.Corner2.Y && Point.X > Box.Corner1.X && Point.X < Box.Corner2.X && Point.Z > Box.Corner1.Z && Point.Z < Box.Corner2.Z)
    {
        Normal.set(0, 1, 0);
        return dim::vector3df(Point.X, Box.Corner2.Z, Point.Z);
    }
    
    /* Check nearest point on side: bottom */
    if (Point.Y <= Box.Corner1.Y && Point.X > Box.Corner1.X && Point.X < Box.Corner2.X && Point.Z > Box.Corner1.Z && Point.Z < Box.Corner2.Z)
    {
        Normal.set(0, -1, 0);
        return dim::vector3df(Point.X, Box.Corner1.Z, Point.Z);
    }
    
    if (Point.Y >= Box.Corner2.Y)
    {
        
        /* === Edges === */
        
        /* Check nearest point to edge: top-left */
        if (Point.X <= Box.Corner1.X && Point.Z > Box.Corner1.Z && Point.Z < Box.Corner2.Z)
        {
            Normal.make(Point.X - Box.Corner1.X, Point.Y - Box.Corner2.Y, 0).normalize();
            return dim::vector3df(Box.Corner1.X, Box.Corner2.Y, Point.Z);
        }
        
        /* Check nearest point to edge: top-right */
        if (Point.X >= Box.Corner2.X && Point.Z > Box.Corner1.Z && Point.Z < Box.Corner2.Z)
        {
            Normal.make(Point.X - Box.Corner2.X, Point.Y - Box.Corner2.Y, 0).normalize();
            return dim::vector3df(Box.Corner2.X, Box.Corner2.Y, Point.Z);
        }
        
        /* Check nearest point to edge: top-top */
        if (Point.Z >= Box.Corner2.Z && Point.X > Box.Corner1.X && Point.X < Box.Corner2.X)
        {
            Normal.make(0, Point.Y - Box.Corner2.Y, Point.Z - Box.Corner2.Z).normalize();
            return dim::vector3df(Point.X, Box.Corner2.Y, Box.Corner2.Z);
        }
        
        /* Check nearest point to edge: top-bottom */
        if (Point.Z <= Box.Corner1.Z && Point.X > Box.Corner1.X && Point.X < Box.Corner2.X)
        {
            Normal.make(0, Point.Y - Box.Corner2.Y, Point.Z - Box.Corner1.Z).normalize();
            return dim::vector3df(Point.X, Box.Corner2.Y, Box.Corner1.Z);
        }
        
        /* === Corners === */
        
        /* Check nearest point to corner: top-left-top */
        if (Point.X <= Box.Corner1.X && Point.Z >= Box.Corner2.Z)
        {
            Normal.make(Point.X - Box.Corner1.X, Point.Y - Box.Corner2.Y, Point.Z - Box.Corner2.Z).normalize();
            return dim::vector3df(Box.Corner1.X, Box.Corner2.Y, Box.Corner2.Z);
        }
        
        /* Check nearest point to corner: top-right-top */
        if (Point.X >= Box.Corner2.X && Point.Z >= Box.Corner2.Z)
        {
            Normal.make(Point.X - Box.Corner2.X, Point.Y - Box.Corner2.Y, Point.Z - Box.Corner2.Z).normalize();
            return dim::vector3df(Box.Corner2.X, Box.Corner2.Y, Box.Corner2.Z);
        }
        
        /* Check nearest point to corner: top-right-bottom */
        if (Point.X >= Box.Corner2.X && Point.Z <= Box.Corner1.Z)
        {
            Normal.make(Point.X - Box.Corner2.X, Point.Y - Box.Corner2.Y, Point.Z - Box.Corner1.Z).normalize();
            return dim::vector3df(Box.Corner2.X, Box.Corner2.Y, Box.Corner1.Z);
        }
        
        /* Check nearest point to corner: top-left-bottom */
        if (Point.X <= Box.Corner1.X && Point.Z <= Box.Corner1.Z)
        {
            Normal.make(Point.X - Box.Corner1.X, Point.Y - Box.Corner2.Y, Point.Z - Box.Corner1.Z).normalize();
            return dim::vector3df(Box.Corner1.X, Box.Corner2.Y, Box.Corner1.Z);
        }
        
    }
    else if (Point.Y <= Box.Corner1.Y)
    {
        
        /* === Edges === */
        
        /* Check nearest point to edge: bottom-left */
        if (Point.X <= Box.Corner1.X && Point.Z > Box.Corner1.Z && Point.Z < Box.Corner2.Z)
        {
            Normal.make(Point.X - Box.Corner1.X, Point.Y - Box.Corner1.Y, 0).normalize();
            return dim::vector3df(Box.Corner1.X, Box.Corner1.Y, Point.Z);
        }
        
        /* Check nearest point to edge: bottom-right */
        if (Point.X >= Box.Corner2.X && Point.Z > Box.Corner1.Z && Point.Z < Box.Corner2.Z)
        {
            Normal.make(Point.X - Box.Corner2.X, Point.Y - Box.Corner1.Y, 0).normalize();
            return dim::vector3df(Box.Corner2.X, Box.Corner1.Y, Point.Z);
        }
        
        /* Check nearest point to edge: bottom-top */
        if (Point.Z >= Box.Corner2.Z && Point.X > Box.Corner1.X && Point.X < Box.Corner2.X)
        {
            Normal.make(0, Point.Y - Box.Corner1.Y, Point.Z - Box.Corner2.Z).normalize();
            return dim::vector3df(Point.X, Box.Corner1.Y, Box.Corner2.Z);
        }
        
        /* Check nearest point to edge: bottom-bottom */
        if (Point.Z <= Box.Corner1.Z && Point.X > Box.Corner1.X && Point.X < Box.Corner2.X)
        {
            Normal.make(0, Point.Y - Box.Corner1.Y, Point.Z - Box.Corner1.Z).normalize();
            return dim::vector3df(Point.X, Box.Corner1.Y, Box.Corner1.Z);
        }
        
        /* === Corners === */
        
        /* Check nearest point to corner: bottom-left-top */
        if (Point.X <= Box.Corner1.X && Point.Z >= Box.Corner2.Z)
        {
            Normal.make(Point.X - Box.Corner1.X, Point.Y - Box.Corner1.Y, Point.Z - Box.Corner2.Z).normalize();
            return dim::vector3df(Box.Corner1.X, Box.Corner1.Y, Box.Corner2.Z);
        }
        
        /* Check nearest point to corner: bottom-right-top */
        if (Point.X >= Box.Corner2.X && Point.Z >= Box.Corner2.Z)
        {
            Normal.make(Point.X - Box.Corner2.X, Point.Y - Box.Corner1.Y, Point.Z - Box.Corner2.Z).normalize();
            return dim::vector3df(Box.Corner2.X, Box.Corner1.Y, Box.Corner2.Z);
        }
        
        /* Check nearest point to corner: bottom-right-bottom */
        if (Point.X >= Box.Corner2.X && Point.Z <= Box.Corner1.Z)
        {
            Normal.make(Point.X - Box.Corner2.X, Point.Y - Box.Corner1.Y, Point.Z - Box.Corner1.Z).normalize();
            return dim::vector3df(Box.Corner2.X, Box.Corner1.Y, Box.Corner1.Z);
        }
        
        /* Check nearest point to corner: bottom-left-bottom */
        if (Point.X <= Box.Corner1.X && Point.Z <= Box.Corner1.Z)
        {
            Normal.make(Point.X - Box.Corner1.X, Point.Y - Box.Corner1.Y, Point.Z - Box.Corner1.Z).normalize();
            return dim::vector3df(Box.Corner1.X, Box.Corner1.Y, Box.Corner1.Z);
        }
        
    }
    else
    {
        
        /* === Edges === */
        
        if (Point.Z <= Box.Corner1.Z)
        {
            
            /* Check nearest point to edge: front-left */
            if (Point.X <= Box.Corner1.X && Point.Y > Box.Corner1.Y && Point.Y < Box.Corner2.Y)
            {
                Normal.make(Point.X - Box.Corner1.X, 0, Point.Z - Box.Corner1.Z).normalize();
                return dim::vector3df(Box.Corner1.X, Point.Y, Box.Corner1.Z);
            }
            
            /* Check nearest point to edge: front-right */
            if (Point.X >= Box.Corner2.X && Point.Y > Box.Corner1.Y && Point.Y < Box.Corner2.Y)
            {
                Normal.make(Point.X - Box.Corner2.X, 0, Point.Z - Box.Corner1.Z).normalize();
                return dim::vector3df(Box.Corner2.X, Point.Y, Box.Corner1.Z);
            }
            
        }
        else if (Point.Z >= Box.Corner2.Z)
        {
            
            /* Check nearest point to edge: back-left */
            if (Point.X <= Box.Corner1.X && Point.Y > Box.Corner1.Y && Point.Y < Box.Corner2.Y)
            {
                Normal.make(Point.X - Box.Corner1.X, 0, Point.Z - Box.Corner2.Z).normalize();
                return dim::vector3df(Box.Corner1.X, Point.Y, Box.Corner2.Z);
            }
            
            /* Check nearest point to edge: back-right */
            if (Point.X >= Box.Corner2.X && Point.Y > Box.Corner1.Y && Point.Y < Box.Corner2.Y)
            {
                Normal.make(Point.X - Box.Corner2.X, 0, Point.Z - Box.Corner2.Z).normalize();
                return dim::vector3df(Box.Corner2.X, Point.Y, Box.Corner2.Z);
            }
            
        }
        
    }
    
    return dim::vector3df();
    
}



/*
 * ======= Line Line intersection =======
 */

SP_EXPORT dim::point2df getLineLineIntersection(dim::point2df A, dim::point2df B, dim::point2df C, dim::point2df D)
{
    return dim::point2df(
        (  ( A.X*B.Y - A.Y*B.X )*( C.X - D.X ) - ( A.X - B.X )*( C.X*D.Y - C.Y*D.X ) )  /
           ( ( A.X - B.X )*( C.Y - D.Y ) - ( A.Y - B.Y )*( C.X - D.X )  ),
        (  ( A.X*B.Y - A.Y*B.X )*( C.Y - D.Y ) - ( A.Y - B.Y )*( C.X*D.Y - C.Y*D.X ) )  /
           ( ( A.X - B.X )*( C.Y - D.Y ) - ( A.Y - B.Y )*( C.X - D.X )  )
    );
}

bool getLineLineIntersection(
    dim::vector3df A, dim::vector3df B, dim::vector3df C, dim::vector3df D, dim::vector3df &P1, dim::vector3df &P2)
{
    dim::vector3df p1(A), p2(B), p3(C), p4(D);
    dim::vector3df p13, p43, p21;
    f32 d1343, d4321, d1321, d4343, d2121;
    f32 numer, denom;
    f32 mua, mub;
    
    p13 = p1 - p3;
    p43 = p4 - p3;
    
    if (Abs(p43.X) < FLOAT_ROUND && Abs(p43.Y) < FLOAT_ROUND && Abs(p43.Z) < FLOAT_ROUND)
        return false;
    
    p21 = p2 - p1;
    
    if (Abs(p21.X) < FLOAT_ROUND && Abs(p21.Y) < FLOAT_ROUND && Abs(p21.Z) < FLOAT_ROUND)
        return false;
    
    d1343 = p13.X * p43.X + p13.Y * p43.Y + p13.Z * p43.Z;
    d4321 = p43.X * p21.X + p43.Y * p21.Y + p43.Z * p21.Z;
    d1321 = p13.X * p21.X + p13.Y * p21.Y + p13.Z * p21.Z;
    d4343 = p43.X * p43.X + p43.Y * p43.Y + p43.Z * p43.Z;
    d2121 = p21.X * p21.X + p21.Y * p21.Y + p21.Z * p21.Z;
    
    denom = d2121 * d4343 - d4321 * d4321;
    
    if (Abs(denom) < FLOAT_ROUND)
        return false;
    
    numer = d1343 * d4321 - d1321 * d4343;
    
    mua = numer / denom;
    mub = (d1343 + d4321 * mua) / d4343;
    
    P1 = p1 + p21 * mua;
    P2 = p3 + p43 * mub;
    
    return true;
}

// -> (http://softsurfer.com/Archive/algorithm_0106/algorithm_0106.htm)
SP_EXPORT f32 getLineLineDistance(
    dim::vector3df A, dim::vector3df B, dim::vector3df C, dim::vector3df D, dim::vector3df &P1, dim::vector3df &P2)
{
    // Temporary variables
    dim::vector3df u(B - A);
    dim::vector3df v(D - C);
    dim::vector3df w(C - A);
    
    // Scalar products
    f32 uu = u.getScalarProduct(u);
    f32 uv = u.getScalarProduct(v);
    f32 vv = v.getScalarProduct(v);
    f32 uw = u.getScalarProduct(w);
    f32 vw = v.getScalarProduct(w);
    
    f32 d = uu*vv - uv*uv;
    f32 sc, tc;
    
    // First calculations
    if (d > 0.0f)
    {
        sc = (uv*vw - vv*uw) / d;
        tc = (uu*vw - uv*uw) / d;
    }
    else
    {
        sc = 0.0f;
        tc = (uv > vv ? uw/uv : vw/vv);
    }
    
    // Compute the nearest points
    P1 = A + (B - A) * sc;
    P2 = C + (D - C) * tc;
    
    // Compute the distance
    dim::vector3df dP(w + (u * sc) - (v * tc));
    
    return sqrt(dP.getScalarProduct(dP));
}



/*********************************
 *                               *
 *  Intersection math functions  *
 *                               *
 *********************************/


/*
 * ======= Line triangle intersection =======
 */

SP_EXPORT bool checkLineTriangleIntersection(dim::triangle3df &Triangle, dim::line3df &Line, dim::vector3df &Intersection)
{
    dim::vector3df pq(Line.Corner2 - Line.Corner1);
    dim::vector3df pa(Triangle.PointA - Line.Corner1);
    dim::vector3df pb(Triangle.PointB - Line.Corner1);
    dim::vector3df pc(Triangle.PointC - Line.Corner1);
    
    /* Check if pq is inside the edges bc, ca and ab */
    Intersection.X = pb.getScalarProduct(pq.getCrossProduct(pc));
    if (Intersection.X < 0.0f)
        return false;
    Intersection.Y = pc.getScalarProduct(pq.getCrossProduct(pa));
    if (Intersection.Y < 0.0f)
        return false;
    Intersection.Z = pa.getScalarProduct(pq.getCrossProduct(pb));
    if (Intersection.Z < 0.0f)
        return false;
    
    return dim::plane3df(Triangle).getLineIntersection(Line.Corner1, Line.Corner2, Intersection);
}


/*
 * ======= Line rectangle intersection =======
 */

SP_EXPORT bool checkLineRectIntersection(dim::rect2df &Rect, dim::point2df &Pos1, dim::point2df &Pos2, dim::point2df &Intersection)
{
    /* Allocate memory */
    dim::point2df A, B, C, D;
    const f32 PickPrecision = 5.0f;
    
    /* First settings */
    f32 width = Pos1.X - Pos2.X;
    f32 height = Pos1.Y - Pos2.Y;
    f32 size;
    
    if (width == 0.0f)
        width = FLOAT_ROUND;
    if (height == 0.0f)
        height = FLOAT_ROUND;
    
    /* Calculate A */
    size = Pos1.Y - Rect.Bottom;
    A.set(Pos1.X - (width * size / height), Rect.Bottom);
    
    /* Calculate B */
    size = Pos1.Y - Rect.Top;
    B.set(Pos1.X - (width * size / height), Rect.Top);
    
    /* Calculate C */
    size = Pos1.X - Rect.Right;
    C.set(Rect.Right, Pos1.Y - (height * size / width));
    
    /* Calculate D */
    size = Pos1.X - Rect.Left;
    D.set(Rect.Left, Pos1.Y - (height * size / width));
    
    /* Checking the calculated points */
    if (height > 0 && Pos1.Y > Rect.Bottom &&
        A.X > Rect.Left - PickPrecision && A.X < Rect.Right + PickPrecision &&
        A.Y > Rect.Top - PickPrecision && A.Y < Rect.Bottom + PickPrecision)
    {
        Intersection = A;
        return true;
    }
    
    if (height < 0 && Pos1.Y < Rect.Top &&
        B.X > Rect.Left - PickPrecision && B.X < Rect.Right + PickPrecision &&
        B.Y > Rect.Top - PickPrecision && B.Y < Rect.Bottom + PickPrecision)
    {
        Intersection = B;
        return true;
    }
    
    if (width > 0 && Pos1.X > Rect.Right &&
        C.X > Rect.Left - PickPrecision && C.X < Rect.Right + PickPrecision &&
        C.Y > Rect.Top - PickPrecision && C.Y < Rect.Bottom + PickPrecision)
    {
        Intersection = C;
        return true;
    }
    
    if (width < 0 && Pos1.X < Rect.Left &&
        D.X > Rect.Left - PickPrecision && D.X < Rect.Right + PickPrecision &&
        D.Y > Rect.Top - PickPrecision && D.Y < Rect.Bottom + PickPrecision)
    {
        Intersection = D;
        return true;
    }
    
    return false;
}


/*
 * ======= Line sphere intersection =======
 */

SP_EXPORT bool checkLineSphereIntersection(dim::line3df &Line, dim::vector3df &Position, f32 Radius, dim::vector3df &Intersection)
{
    /* Temporary variables */
    dim::vector3df d(Line.getVector().normalize());
    dim::vector3df m(Line.Corner1 - Position);
    f32 t;
    
    /* Compute the scalar products */
    f32 b = m.getScalarProduct(d);
    f32 c = m.getScalarProduct(m) - Radius*Radius;
    
    /* Exit if r's origin outside s (c > 0) and r pointing away from s (b > 0) */
    if (c > 0.0f && b > 0.0f)
        return false;
    
    f32 discr = b*b - c;
    
    /* A negative discriminant corresponds to ray missing sphere */
    if (discr < 0.0f)
        return false;
    
    /* Ray now found to intersect sphere, compute smallest t value of intersection */
    t = -b - sqrt(discr);
    
    /* If t is negative, ray started inside sphere so clamp t to 0 (zero) */
    if (t < 0.0f)
        t = 0.0f;
    
    /* Compute the intersection point */
    Intersection = Line.Corner1 + d * t;
    
    /* An intersection has been detected */
    return true;
}


/*
 * ======= Line box intersection =======
 */

SP_EXPORT bool checkLineBoxIntersection(dim::line3df &Line, dim::line3df &Box, dim::vector3df &Intersection)
{
    /* Temporary variables */
    dim::vector3df Direction(Line.getVector().normalize());
    
    f32 p[3] = { Line.Corner1.X, Line.Corner1.Y, Line.Corner1.Z };
    f32 d[3] = { Direction.X, Direction.Y, Direction.Z };
    f32 min[3] = { Box.Corner1.X, Box.Corner1.Y, Box.Corner1.Z };
    f32 max[3] = { Box.Corner2.X, Box.Corner2.Y, Box.Corner2.Z };
    
    f32 tmin = 0.0f;
    f32 tmax = 1000000.0f;
    
    /* Loop for all three slabs */
    for (s32 i = 0; i < 3; ++i)
    {
        if (Abs(d[i]) < FLOAT_ROUND64)
        {
            /* Ray is parallel tp slab. No hit if origin not within slab */
            if (p[i] < min[i] || p[i] > max[i])
                return false;
        }
        else
        {
            /* Compute intersection t value of ray with near and far plane of slab */
            f32 ood = 1.0f / d[i];
            f32 t1 = (min[i] - p[i]) * ood;
            f32 t2 = (max[i] - p[i]) * ood;
            
            /* Make t1 be intersection with near plane, t2 with far plane */
            if (t1 > t2)
                Swap(t1, t2);
            
            /* Compute the intersection of slab intersection intervals */
            tmin = Max(tmin, t1);
            tmax = Min(tmax, t2);
            
            /* Exit with no collision as soon as slab intersection becomes empty */
            if (tmin > tmax)
                return false;
        }
    }
    
    /* Compute the intersection point */
    Intersection = Line.Corner1 + Direction * tmin;
    
    /* An intersection has been detected */
    return true;
}


/*
 * ======= Triangle triangle intersection =======
 */

SP_EXPORT bool checkTriangleTriangleIntersection(dim::triangle3df &TriangleA, dim::triangle3df &TriangleB, dim::line3df &Intersection)
{
    /* Internal macros */
    #define ADD                                             \
        if (TestIndex < 2 && !Point[0].isEqual(TempPoint))  \
        {                                                   \
            Point[TestIndex] = TempPoint;                   \
            ++TestIndex;                                    \
        }
    #define FINALCHECK                          \
        if (TestIndex > 1)                      \
        {                                       \
            Intersection.Corner1 = Point[0];    \
            Intersection.Corner2 = Point[1];    \
            return true;                        \
        }
    #define INTERSECT(t, a, b)                                      \
        Line = dim::line3df(a, b);                                  \
        if (checkLineTriangleIntersection(t, Line, TempPoint))      \
        {                                                           \
            ADD                                                     \
        }                                                           \
        else                                                        \
        {                                                           \
            Line = dim::line3df(b, a);                              \
            if (checkLineTriangleIntersection(t, Line, TempPoint))  \
                ADD                                                 \
        }
    #define INTERSECTA(a, b)                            \
        INTERSECT(TriangleA, TriangleB.a, TriangleB.b)
    #define INTERSECTB(a, b)                            \
        INTERSECT(TriangleB, TriangleA.a, TriangleA.b)
    
    /* Temporary variables */
    dim::line3df Line;
    
    dim::vector3df TempPoint;
    
    s32 TestIndex = 0;
    dim::vector3df Point[2];
    
    /* Get all intersections with the triangle A and the three edges of triangle B  */
    INTERSECTA(PointA, PointB);
    INTERSECTA(PointB, PointC);
    INTERSECTA(PointC, PointA);
    
    /* Check for intersection of triangle A */
    FINALCHECK;
    
    /* Get all intersections with the triangle B and the three edges of triangle A  */
    INTERSECTB(PointA, PointB);
    INTERSECTB(PointB, PointC);
    INTERSECTB(PointC, PointA);
    
    /* Check for intersection of triangle B */
    FINALCHECK;
    
    /* Terminate the macros */
    #undef ADD
    #undef FINALCHECK
    #undef INTERSECT
    #undef INTERSECTA
    #undef INTERSECTB
    
    /* No intersection has been detected */
    return false;
}


/*
 * ======= Line box intersection =======
 */

bool rayAABBIntersection1D(f32 start, f32 dir, f32 min, f32 max, f32 &enter, f32 &exit)
{
    /* Ray parallel to direction */
    if (Abs(dir) < FLOAT_ROUND64)
        return (start >= min && start <= max);
    
    /* Intersection parameters */
    f32 t0 = (min - start) / dir;
    f32 t1 = (max - start) / dir;
    
    /* Sort intersections */
    if (t0 > t1)
        Swap(t0, t1);
    
    /* Check if intervals are disjoint */
    if (t0 > exit || t1 < enter)
        return false;
    
    /* Reduce interval */
    if (t0 > enter)
        enter = t0;
    if (t1 < exit)
        exit = t1;
    
    return true;
}

SP_EXPORT bool checkLineBoxIntersection(dim::line3df &Line, dim::vector3df &Corner1, dim::vector3df &Corner2)
{
    /* Check if the line is in a box */
    if ( ( Line.Corner1.X >= Corner1.X && Line.Corner1.Y >= Corner1.Y && Line.Corner1.Z >= Corner1.Z && 
           Line.Corner1.X < Corner2.X && Line.Corner1.Y < Corner2.Y && Line.Corner1.Z < Corner2.Z ) ||
         ( Line.Corner2.X >= Corner1.X && Line.Corner2.Y >= Corner1.Y && Line.Corner2.Z >= Corner1.Z && 
           Line.Corner2.X < Corner2.X && Line.Corner2.Y < Corner2.Y && Line.Corner2.Z < Corner2.Z ) )
    {
        return true;
    }
    
    /* Check the ray-AABB intersection */
    
    f32 enter = 0.0f, exit = 1.0f;
    dim::vector3df dir(Line.getVector());
    
    if (!rayAABBIntersection1D(Line.Corner1.X, dir.X, Corner1.X, Corner2.X, enter, exit))
        return false;
    if (!rayAABBIntersection1D(Line.Corner1.Y, dir.Y, Corner1.Y, Corner2.Y, enter, exit))
        return false;
    if (!rayAABBIntersection1D(Line.Corner1.Z, dir.Z, Corner1.Z, Corner2.Z, enter, exit))
        return false;
    
    return true;
}


/*
 * ======= Triangle box intersection =======
 */

SP_EXPORT bool checkPlaneBoxIntersection(dim::plane3df &Plane, dim::vector3df &Corner1, dim::vector3df &Corner2)
{
    /* These two lines not necessary with a (center, extents) AABB representation */
    dim::vector3df c((Corner1 + Corner2) * 0.5f);   // Compute AABB center
    dim::vector3df e(Corner2 - c);                  // Compute positive extents
    
    /* Compute the projection interval readius of b onto L(t) = b.c + t * p.n */
    f32 r = e.X * Abs(Plane.Normal.X) + e.Y * Abs(Plane.Normal.Y) + e.Z * Abs(Plane.Normal.Z);
    
    /* Compute distance of box center from plane */
    f32 s = Plane.Normal.getScalarProduct(c) - Plane.Distance;
    
    /* Intersection occurs when distance s falls within [-r, +r] interval */
    return Abs(s) <= r;
}

// !!! INCOMPLETE !!!

#if 0

SP_EXPORT bool checkTriangleBoxIntersection(dim::triangle3df &Triangle, dim::vector3df &Corner1, dim::vector3df &Corner2)
{
    /* First checking the simple tests */
    if ( ( Triangle.PointA >= Corner1 && Triangle.PointA < Corner2 ) ||
         ( Triangle.PointB >= Corner1 && Triangle.PointB < Corner2 ) ||
         ( Triangle.PointC >= Corner1 && Triangle.PointC < Corner2 ) )
    {
        return true;
    }
    
    /* Allocate temporary memory */
    f32 p0, p1, p2, r;
    
    /* Compute box center and extents (if not already given in that format) */
    dim::vector3df c((Corner1 + Corner2) / 2), a;
    f32 e0 = (Corner2.X - Corner1.X) / 2;
    f32 e1 = (Corner2.Y - Corner1.Y) / 2;
    f32 e2 = (Corner2.Z - Corner1.Z) / 2;
    
    /* Translate triangle as conceptually moving AABB to origin */
    Triangle -= c;
    
    /* Compute edge vectors for triangle */
    dim::vector3df f0(Triangle.PointB - Triangle.PointA);
    dim::vector3df f1(Triangle.PointC - Triangle.PointB);
    dim::vector3df f2(Triangle.PointA - Triangle.PointC);
    
    #if 0
    /* Test axis a00 (category 3) */
    p0 = Triangle.PointA.Z * Triangle.PointB.Y - Triangle.PointA.Y * Triangle.PointB.Z;
    p2 = Triangle.PointC.Z * ( Triangle.PointB.Y - Triangle.PointA.Y ) - Triangle.PointC.Y * ( Triangle.PointB.Z - Triangle.PointA.Z );
    r = e1 * Abs(f0.Z) + e2 * Abs(f0.Y);
    if (Max(-Max(p0, p2), Min(p0, p2)) > r)
        return false;
    #endif
    
    #if 1
    /* Test axis a00 (category 3) */
    a.set(0, -f0.Z, f0.Y);
    p0 = Triangle.PointA.getScalarProduct(a);
    p1 = Triangle.PointB.getScalarProduct(a);
    p2 = Triangle.PointC.getScalarProduct(a);
    r = e0 * Abs(dim::vector3df(1, 0, 0).getScalarProduct(a)) +
        e1 * Abs(dim::vector3df(0, 1, 0).getScalarProduct(a)) +
        e2 * Abs(dim::vector3df(0, 0, 1).getScalarProduct(a));
    if (Max(-Max(p0, p1, p2), Min(p0, p1, p2)) > r)
        return false;
    
    /* Test axis a01 (category 3) */
    a.set(0, -f1.Z, f1.Y);
    p0 = Triangle.PointA.getScalarProduct(a);
    p1 = Triangle.PointB.getScalarProduct(a);
    p2 = Triangle.PointC.getScalarProduct(a);
    r = e0 * Abs(dim::vector3df(1, 0, 0).getScalarProduct(a)) +
        e1 * Abs(dim::vector3df(0, 1, 0).getScalarProduct(a)) +
        e2 * Abs(dim::vector3df(0, 0, 1).getScalarProduct(a));
    if (Max(-Max(p0, p1, p2), Min(p0, p1, p2)) > r)
        return false;
    
    /* Test axis a02 (category 3) */
    a.set(0, -f2.Z, f2.Y);
    p0 = Triangle.PointA.getScalarProduct(a);
    p1 = Triangle.PointB.getScalarProduct(a);
    p2 = Triangle.PointC.getScalarProduct(a);
    r = e0 * Abs(dim::vector3df(1, 0, 0).getScalarProduct(a)) +
        e1 * Abs(dim::vector3df(0, 1, 0).getScalarProduct(a)) +
        e2 * Abs(dim::vector3df(0, 0, 1).getScalarProduct(a));
    if (Max(-Max(p0, p1, p2), Min(p0, p1, p2)) > r)
        return false;
    
    /* Test axis a10 (category 3) */
    a.set(f0.Z, 0, -f0.X);
    p0 = Triangle.PointA.getScalarProduct(a);
    p1 = Triangle.PointB.getScalarProduct(a);
    p2 = Triangle.PointC.getScalarProduct(a);
    r = e0 * Abs(dim::vector3df(1, 0, 0).getScalarProduct(a)) +
        e1 * Abs(dim::vector3df(0, 1, 0).getScalarProduct(a)) +
        e2 * Abs(dim::vector3df(0, 0, 1).getScalarProduct(a));
    if (Max(-Max(p0, p1, p2), Min(p0, p1, p2)) > r)
        return false;
    
    /* Test axis a11 (category 3) */
    a.set(f1.Z, 0, -f1.X);
    p0 = Triangle.PointA.getScalarProduct(a);
    p1 = Triangle.PointB.getScalarProduct(a);
    p2 = Triangle.PointC.getScalarProduct(a);
    r = e0 * Abs(dim::vector3df(1, 0, 0).getScalarProduct(a)) +
        e1 * Abs(dim::vector3df(0, 1, 0).getScalarProduct(a)) +
        e2 * Abs(dim::vector3df(0, 0, 1).getScalarProduct(a));
    if (Max(-Max(p0, p1, p2), Min(p0, p1, p2)) > r)
        return false;
    
    /* Test axis a12 (category 3) */
    a.set(f2.Z, 0, -f2.X);
    p0 = Triangle.PointA.getScalarProduct(a);
    p1 = Triangle.PointB.getScalarProduct(a);
    p2 = Triangle.PointC.getScalarProduct(a);
    r = e0 * Abs(dim::vector3df(1, 0, 0).getScalarProduct(a)) +
        e1 * Abs(dim::vector3df(0, 1, 0).getScalarProduct(a)) +
        e2 * Abs(dim::vector3df(0, 0, 1).getScalarProduct(a));
    if (Max(-Max(p0, p1, p2), Min(p0, p1, p2)) > r)
        return false;
    
    /* Test axis a20 (category 3) */
    a.set(-f0.Y, f0.X, 0);
    p0 = Triangle.PointA.getScalarProduct(a);
    p1 = Triangle.PointB.getScalarProduct(a);
    p2 = Triangle.PointC.getScalarProduct(a);
    r = e0 * Abs(dim::vector3df(1, 0, 0).getScalarProduct(a)) +
        e1 * Abs(dim::vector3df(0, 1, 0).getScalarProduct(a)) +
        e2 * Abs(dim::vector3df(0, 0, 1).getScalarProduct(a));
    if (Max(-Max(p0, p1, p2), Min(p0, p1, p2)) > r)
        return false;
    
    /* Test axis a21 (category 3) */
    a.set(-f1.Y, f1.X, 0);
    p0 = Triangle.PointA.getScalarProduct(a);
    p1 = Triangle.PointB.getScalarProduct(a);
    p2 = Triangle.PointC.getScalarProduct(a);
    r = e0 * Abs(dim::vector3df(1, 0, 0).getScalarProduct(a)) +
        e1 * Abs(dim::vector3df(0, 1, 0).getScalarProduct(a)) +
        e2 * Abs(dim::vector3df(0, 0, 1).getScalarProduct(a));
    if (Max(-Max(p0, p1, p2), Min(p0, p1, p2)) > r)
        return false;
    
    /* Test axis a22 (category 3) */
    a.set(-f2.Y, f2.X, 0);
    p0 = Triangle.PointA.getScalarProduct(a);
    p1 = Triangle.PointB.getScalarProduct(a);
    p2 = Triangle.PointC.getScalarProduct(a);
    r = e0 * Abs(dim::vector3df(1, 0, 0).getScalarProduct(a)) +
        e1 * Abs(dim::vector3df(0, 1, 0).getScalarProduct(a)) +
        e2 * Abs(dim::vector3df(0, 0, 1).getScalarProduct(a));
    if (Max(-Max(p0, p1, p2), Min(p0, p1, p2)) > r)
        return false;
    #endif
    
    /* Test the three axes corresponding to the face normals of AABB b (category 1) */
    if (Max(Triangle.PointA.X, Triangle.PointB.X, Triangle.PointC.X) < -e0 || Min(Triangle.PointA.X, Triangle.PointB.X, Triangle.PointC.X) > e0)
        return false;
    if (Max(Triangle.PointA.Y, Triangle.PointB.Y, Triangle.PointC.Y) < -e1 || Min(Triangle.PointA.Y, Triangle.PointB.Y, Triangle.PointC.Y) > e1)
        return false;
    if (Max(Triangle.PointA.Z, Triangle.PointB.Z, Triangle.PointC.Z) < -e2 || Min(Triangle.PointA.Z, Triangle.PointB.Z, Triangle.PointC.Z) > e2)
        return false;
    
    /* Test separating axis correspondig to triangle face normal (category 2) */
    dim::plane3df p;
    p.Normal = f0.getCrossProduct(f1);
    p.Length = p.Normal.getScalarProduct(Triangle.PointA);
    
    return checkPlaneBoxIntersection(p, Corner1, Corner2);
}

#else

// !!! INCOMPLETE !!!

SP_EXPORT bool checkTriangleBoxIntersection(dim::triangle3df &Triangle, dim::vector3df &Corner1, dim::vector3df &Corner2)
{
    /* First checking the simple tests */
    if ( ( Triangle.PointA >= Corner1 && Triangle.PointA <= Corner2 ) ||
         ( Triangle.PointB >= Corner1 && Triangle.PointB <= Corner2 ) ||
         ( Triangle.PointC >= Corner1 && Triangle.PointC <= Corner2 ) )
    {
        return true;
    }
    
    /* Create a temporary triangle box */
    dim::line3df TriangleBox(Triangle.getBox());
    
    /* Check if the two boxes do intersect */
    return TriangleBox.Corner1 <= Corner2 && TriangleBox.Corner2 >= Corner1;
    /*
    return ( Corner1 >= TriangleBox.Corner1 && Corner1 <= TriangleBox.Corner2 ) ||
           ( Corner2 >= TriangleBox.Corner1 && Corner2 <= TriangleBox.Corner2 );*/
}

#endif


/*
 * ======= Other distance and normal-vector functions =======
 */

SP_EXPORT bool checkTriangleSide(dim::triangle3df &Triangle, dim::vector3df &Point)
{
    #if 1
    
    return dim::plane3df(Triangle).isPointFrontSide(Point);
    
    #endif
    
    /* Compute the both normals */
    dim::vector3df n( getNormalVectorSQ(Triangle.PointA, Triangle.PointB, Triangle.PointC) );
    dim::vector3df m( Point - dim::plane3df(Triangle).getClosestPoint(Point) );
    
    /* Check the side of the triangle (on this case: plane) */
    if ( ( ( n.X > 0.0 && m.X > 0.0 ) || ( n.X <= 0.0 && m.X <= 0.0 ) ) &&
         ( ( n.Y > 0.0 && m.Y > 0.0 ) || ( n.Y <= 0.0 && m.Y <= 0.0 ) ) &&
         ( ( n.Z > 0.0 && m.Z > 0.0 ) || ( n.Z <= 0.0 && m.Z <= 0.0 ) ) )
    {
        /* Front side */
        return true;
    }
    
    /* Back side */
    return false;
}


/*
 * ======= Other math functions =======
 */

SP_EXPORT void SeedRandom()
{
    srand( static_cast<u32>(time(0)) );
}


// Positive and negative random values
SP_EXPORT s32 Random(s32 Value)
{
    return Value < 0 ? - static_cast<s32>( rand()%(1 - Value) ) : static_cast<s32>( rand()%(1 + Value) );
}
SP_EXPORT s32 Random(s32 Value1, s32 Value2)
{
    if (Value1 >= Value2)
        return Value1;
    
    s32 RandomVal = 0;
    s32 Offset = Value1 < 0 ? -Value1 : 0;
    
    if (Value1)
        RandomVal = static_cast<s32>( Value1 + rand()%(Value2 - Value1 + 1) );
    else
        RandomVal = static_cast<s32>( Value1 - Offset + rand()%((Value2 - Offset) - (Value1 - Offset) + 1) );
    
    return RandomVal;
}


} // /namespace math

} // /namespace sp







// ================================================================================
