/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * Header for all math-operations (sin, cos ...)
 *
 */

#ifndef __SP_MATH_H__
#define __SP_MATH_H__


#include "spStandard.hpp"
#include "spDimensionPoint2D.hpp"
#include "spDimensionRect2D.hpp"
#include "spDimensionLine3D.hpp"
#include "spDimensionVector3D.hpp"
#include "spDimensionTriangle3D.hpp"
#include "spDimensionPlane3D.hpp"


#include <math.h>
#include <time.h>
#include <stdlib.h>


namespace sp
{
namespace math
{


/*
 * ======= Some standard S.P. math constants =======
 */

const f64 FLOAT_ROUND64 = 0.00000001f;  // 1.0e-8
const f32 FLOAT_ROUND   = 0.000001f;    // 1.0e-6

const f64 PI64      = 3.1415926535897932384626433832795028841971693993751;
const f32 PI        = 3.14159265359;

const f32 DEG       = PI / 180.0f;
const f32 RAD       = 180.0f / PI;
const f64 DEG64     = PI64 / 180.0;
const f64 RAD64     = 180.0 / PI64;

const f64 SQRT2     = sqrt(2.0);

const f64 STDASPECT = 640.0 / 480.0;

#define SIN(x)  sin((x)*M_PI/180.0)
#define COS(x)  cos((x)*M_PI/180.0)
#define TAN(x)  tan((x)*M_PI/180.0)
#define ASIN(x) asin(x)*180.0/M_PI
#define ACOS(x) acos(x)*180.0/M_PI
#define ATAN(x) atan(x)*180.0/M_PI


/*
 * ======= All the mathematical template inline functions which compute fast complexe equations and algorithms =======
 */

template <typename T> inline T Abs(const T Value)
{
    return Value > 0 ? Value : -Value;
}

template <typename T> inline T Max(const T A, const T B)
{
    return A > B ? A : B;
}
template <typename T> inline T Min(const T A, const T B)
{
    return A < B ? A : B;
}
template <typename T> inline T Max(const T A, const T B, const T C)
{
    return ( A >= B && A >= C ? A : ( B >= A && B >= C ? B : C ) );
}
template <typename T> inline T Min(const T A, const T B, const T C)
{
    return ( A <= B && A <= C ? A : ( B <= A && B <= C ? B : C ) );
}
template <typename T> inline T MinMax(const T Value, const T Min, const T Max)
{
    return Value > Max ? Max : Value < Min ? Min : Value;
}
template <typename T> inline void Clamp(T &Value, T Min, T Max)
{
    Value = (Value > Max ? Max : Value < Min ? Min : Value);
}
template <typename T> inline void Swap(T &A, T &B)
{
    T tmp = A; A = B; B = tmp;
}
template <typename T> inline T Sgn(const T Value)
{
    return (Value > 0) ? 1 : (Value < 0) ? -1 : 0; // Signum
}

template <typename T> inline T Round(T Value, s32 Precision)
{
    s32 exp = static_cast<s32>(pow(10, Precision));
    return (T)((s32)(Value*exp)) / exp;
}

template <typename T> inline T Pow2(const T Value)
{
    return Value*Value;
}

template <typename T> inline T Sinus(const T Value)
{
    return static_cast<T>(sin(static_cast<f64>(Value)*DEG64));
}
template <typename T> inline T Cosinus(const T Value)
{
    return static_cast<T>(cos(static_cast<f64>(Value)*DEG64));
}

template <typename T> inline T ArcSinus(const T Value)
{
    return static_cast<T>(asin(static_cast<f64>(Value)*RAD64));
}
template <typename T> inline T ArcCosinus(const T Value)
{
    return static_cast<T>(acos(static_cast<f64>(Value)*RAD64));
}

template <typename T> inline T getDistance(const T X1, const T Y1, const T X2, const T Y2)
{
    return static_cast<T>(sqrt(static_cast<f64>(Pow2(X2 - X1) + Pow2(Y2 - Y1))));
}
template <typename T> inline T getDistance(const dim::point2d<T> Position1, const dim::point2d<T> Position2)
{
    return static_cast<T>(sqrt(static_cast<f64>(Pow2(Position2.X - Position1.X) + Pow2(Position2.Y - Position1.Y))));
}
template <typename T> inline T getDistance(const T X1, const T Y1, const T Z1, const T X2, const T Y2, const T Z3)
{
    return static_cast<T>(sqrt(static_cast<f64>(Pow2(X2 - X1) + Pow2(Y2 - Y1) + Pow2(Y2 - Y1))));
}
template <typename T> inline T getDistance(const dim::vector3d<T> Position1, const dim::vector3d<T> Position2)
{
    return static_cast<T>(
        sqrt( static_cast<f64>(Pow2(Position2.X - Position1.X) + Pow2(Position2.Y - Position1.Y) + Pow2(Position2.Z - Position1.Z)) )
    );
}

template <typename T> inline T getDistanceSq(const T X1, const T Y1, const T X2, const T Y2)
{
    return Pow2(X2 - X1) + Pow2(Y2 - Y1);
}
template <typename T> inline T getDistanceSq(const dim::point2d<T> Position1, const dim::point2d<T> Position2)
{
    return Pow2(Position2.X - Position1.X) + Pow2(Position2.Y - Position1.Y);
}
template <typename T> inline T getDistanceSq(const T X1, const T Y1, const T Z1, const T X2, const T Y2, const T Z3)
{
    return Pow2(X2 - X1) + Pow2(Y2 - Y1) + Pow2(Y2 - Y1);
}
template <typename T> inline T getDistanceSq(const dim::vector3d<T> Position1, const dim::vector3d<T> Position2)
{
    return Pow2(Position2.X - Position1.X) + Pow2(Position2.Y - Position1.Y) + Pow2(Position2.Z - Position1.Z);
}

template <typename T> inline dim::point2d<T> getDifference(const dim::point2d<T> Position1, const dim::point2d<T> Position2)
{
    return dim::point2d<T>(Position1 - Position2).getAbs();
}
template <typename T> inline dim::vector3d<T> getDifference(const dim::vector3d<T> Position1, const dim::vector3d<T> Position2)
{
    return dim::vector3d<T>(Position1 - Position2).getAbs();
}

template <typename T> inline bool isPosInCircle(const T X1, const T Y1, const T X2, const T Y2, const T CircleRadius)
{
    return Pow2(X2 - X1) + Pow2(Y2 - Y1) < Pow2(CircleRadius);
}
template <typename T> inline bool isPosInCircle(const dim::point2d<T> Position1, const dim::point2d<T> Position2, const T CircleRadius)
{
    return Pow2(Position2.X - Position1.X) + Pow2(Position2.Y - Position1.Y) < Pow2(CircleRadius);
}
template <typename T> inline bool isPosInCircle(const T X1, const T Y1, const T Z1, const T X2, const T Y2, const T Z2, const T CircleRadius)
{
    return Pow2(X2 - X1) + Pow2(Y2 - Y1) + Pow2(Z2 - Z1) < Pow2(CircleRadius);
}
template <typename T> inline bool isPosInCircle(const dim::vector3d<T> Position1, const dim::vector3d<T> Position2, const T CircleRadius)
{
    return Pow2(Position2.X - Position1.X) + Pow2(Position2.Y - Position1.Y) + Pow2(Position2.Z - Position1.Z) < Pow2(CircleRadius);
}

template <typename T> inline T getDegree(const T X1, const T Y1, const T X2, const T Y2)
{
    T Degree;
    
    if (X1 != X2 || Y1 != Y2)
    {
        Degree = ASIN( (X2 - X1) / getDistance(X1, Y1, X2, Y2) );
        Degree = Y1 > Y2 ? (T)180 - Degree : Degree;
    }
    else
        Degree = 0;
    
    return Degree;
}
template <typename T> inline T getDegree(const dim::point2d<T> Position1, const dim::point2d<T> Position2)
{
    T Degree;
    
    if (Position1 != Position2)
    {
        Degree = ASIN( (Position2.X - Position1.X) / getDistance(Position1, Position2) );
        Degree = Position1.Y > Position2.Y ? (T)180 - Degree : Degree;
    }
    else
        Degree = 0;
    
    return Degree;
}

template <typename T> inline bool compareVector3DMin(const dim::vector3d<T> Vector1, const dim::vector3d<T> Vector2, const f32 MinValue)
{
    return Abs(
        ( Pow2(Vector1.X) + Pow2(Vector1.Y) + Pow2(Vector1.Z) ) - ( Pow2(Vector2.X) + Pow2(Vector2.Y) + Pow2(Vector2.Z) )
    ) <= MinValue;
}
template <typename T> inline bool compareVector3DMax(const dim::vector3d<T> Vector1, const dim::vector3d<T> Vector2, const f32 MaxValue)
{
    return Abs(
        ( Pow2(Vector1.X) + Pow2(Vector1.Y) + Pow2(Vector1.Z) ) - ( Pow2(Vector2.X) + Pow2(Vector2.Y) + Pow2(Vector2.Z) )
    ) >= MaxValue;
}

template <typename T> inline T getBezierValue(f32 t, T Pos1, T Pos2, T Radial1, T Radial2)
{
    f32 invt    = 1.0f - t;
    f32 invt2   = invt*invt;
    f32 invt3   = invt2*invt;
    f32 t2      = t*t;
    f32 t3      = t*t*t;
    
    return Pos1*invt3 + Radial1*3*t*invt2 + Radial2*3*t2*invt + Pos2*t3;
}

template <typename T> inline dim::vector3d<T> getPlaneEquation(
    const dim::vector3d<T> Vertex1, const dim::vector3d<T> Vertex2, const dim::vector3d<T> Vertex3)
{
    return Vertex1 + (Vertex2 - Vertex1) + (Vertex3 - Vertex1);
}

template <typename T> inline dim::vector3d<T> getNormalVectorSQ(
    const dim::vector3d<T> Vertex1, const dim::vector3d<T> Vertex2, const dim::vector3d<T> Vertex3)
{
    return (Vertex1 - Vertex2) % (Vertex2 - Vertex3);
}

template <typename T> inline dim::vector3d<T> getNormalVector(
    const dim::vector3d<T> Vertex1, const dim::vector3d<T> Vertex2, const dim::vector3d<T> Vertex3)
{
    dim::vector3d<T> Normal = (Vertex1 - Vertex2) % (Vertex2 - Vertex3);
    return Normal / dim::vector3d<T>(Normal.getLength());
}


/*
 * ======= Collision functions =======
 */

SP_EXPORT dim::point2df getLinePointDistance(
    dim::point2df PointA, dim::point2df PointB, dim::point2df PointP, bool isOnlyNormal = false
); // Line 'AB' to Point 'P'
SP_EXPORT dim::vector3df getLinePointDistance(
    dim::vector3df PointA, dim::vector3df PointB, dim::vector3df PointP, bool isOnlyNormal = false
); // Line 'AB' to Point 'P'

SP_EXPORT dim::vector3df getClosestPoint(
    const dim::triangle3df &Triangle, const dim::vector3df &Point
);

SP_EXPORT bool getNormalPoint(
    const dim::triangle3df &Triangle, const dim::vector3df &Point, dim::vector3df &NormalPoint
);

SP_EXPORT dim::vector3df getClosestPoint(dim::line3df &Box, dim::vector3df &Point, dim::vector3df &Normal);

SP_EXPORT dim::point2df getLineLineIntersection(dim::point2df A, dim::point2df B, dim::point2df C, dim::point2df D);
SP_EXPORT bool getLineLineIntersection(
    dim::vector3df A, dim::vector3df B, dim::vector3df C, dim::vector3df D, dim::vector3df &P1, dim::vector3df &P2
);

SP_EXPORT f32 getLineLineDistance(
    dim::vector3df A, dim::vector3df B, dim::vector3df C, dim::vector3df D, dim::vector3df &P1, dim::vector3df &P2
);


/*
 * ======= Intersection functions =======
 */

SP_EXPORT bool checkLineTriangleIntersection(dim::triangle3df &Triangle, dim::line3df &Line, dim::vector3df &Intersection);

SP_EXPORT bool checkLineSphereIntersection(dim::line3df &Line, dim::vector3df &Position, f32 Radius, dim::vector3df &Intersection);

SP_EXPORT bool checkLineBoxIntersection(dim::line3df &Line, dim::line3df &Box, dim::vector3df &Intersection);

SP_EXPORT bool checkTriangleTriangleIntersection(dim::triangle3df &TriangleA, dim::triangle3df &TriangleB, dim::line3df &Intersection);

/* Overlap tests */

SP_EXPORT bool checkLineRectIntersection(dim::rect2df &Rect, dim::point2df &Pos1, dim::point2df &Pos2, dim::point2di &Intersection);

SP_EXPORT bool checkLineBoxIntersection(dim::line3df &Line, dim::vector3df &Corner1, dim::vector3df &Corner2);

SP_EXPORT bool checkPlaneBoxIntersection(dim::plane3df &Plane, dim::vector3df &Corner1, dim::vector3df &Corner2);

SP_EXPORT bool checkTriangleBoxIntersection(dim::triangle3df &Triangle, dim::vector3df &Corner1, dim::vector3df &Corner2);

/*
 * ======= Other distance and normal-vector functions =======
 */

SP_EXPORT bool checkTriangleSide(dim::triangle3df &Triangle, dim::vector3df &Point);


/*
 * ======= Other math functions =======
 */

SP_EXPORT void SeedRandom();

// Only positive random values
inline u32 Rnd(const u32 Value)
{
    return rand()%(1 + Value);
}
inline u32 Rnd(const u32 Value1, const u32 Value2)
{
    return Value1 + rand()%(Value2 - Value1 + 1);
}

// Positive and negative random values
SP_EXPORT s32 Random(const s32 Value);
SP_EXPORT s32 Random(const s32 Value1, const s32 Value2);


inline s32 Pow2Pow(const s32 Count)
{
    register s32 base = 1;
    
    for (register s32 i = 0; i < Count; ++i)
        base <<= 1;
    
    return base;
}

inline s32 Round(const f32 Value)
{
    s32 Result = (s32)Value;
    
    if (f32(Value - Result) * 10.0 >= 5)
        ++Result;
    
    return Result;
}

inline s32 Round2(const s32 Value)
{
    s32 i;
    
    for (i = 1; i < Value; i <<= 1);
    
    if (i - Value <= Value - i/2)
        return i;
    
    return i/2;
}

inline bool isEqual(const f32 A, const f32 B, const f32 Precision = FLOAT_ROUND)
{
    return (A + Precision > B && A - Precision < B);
}




} // /namespace math

} // /namespace sp


#endif










// ================================================================================
