/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * Model loading 3DS-fileformat file
 *
 */

#include "spModelLoader3DS.hpp"
#include "spMainDevice.hpp"


namespace sp
{

extern video::VideoDriver* __spVideoDriver;

namespace scene
{


static const s32 DEFBUFFER_SIZE = 50000;


ModelLoader3DS::ModelLoader3DS()
    : hFile_(0), CurBone_(0), CurrentChunk_(0), TempChunk_(0)
{
    init();
}
ModelLoader3DS::~ModelLoader3DS()
{
    clear();
}

Entity* ModelLoader3DS::loadModel(io::stringc &Filename, io::stringc &TexturePath)
{
    Filename_       = Filename;
    TexturePath_    = TexturePath;
    
    hEntity_        = new Entity();
    
    if ( !( hFile_ = fopen(Filename.c_str(), "rb") ) )
    {
        io::printError("could not open model file");
        return hEntity_;
    }
    
    readChunk(CurrentChunk_);
    
    if (CurrentChunk_->ID != MODEL_3DS_PRIMARY)
    {
        io::printError(
            io::stringc("3DS model has incorrect magic number")
        );
        fclose(hFile_);
        return hEntity_;
    }
    
    processNextChunk(CurrentChunk_);
    
    fclose(hFile_);
    
    return buildModel();
}


/*
 * ========== Private: ==========
 */


void ModelLoader3DS::init()
{
    /* Gereral settings */
    MaterialCounter_    = 0;
    hasAnimation_       = false;
    ObjectCounter_      = 0;
    
    /* Allocate memories */
    CurrentChunk_       = new SChunk3DS;
    TempChunk_          = new SChunk3DS;
}
void ModelLoader3DS::clear()
{
    /* Delete memories */
    spDeletePointer(CurrentChunk_);
    spDeletePointer(TempChunk_);
    
    /* Delete the bone list */
    for (std::vector<SSkeletonBone3DS*>::iterator it = AnimBoneList_.begin(); it != AnimBoneList_.end(); ++it)
        spDeletePointer(*it);
}

s32 ModelLoader3DS::getString(FILE* CurFile, char* pBuffer)
{
    s32 i = 0;
    
    fread(pBuffer, 1, 1, CurFile);
    while (*(pBuffer + i++) != 0) 
        fread(pBuffer + i, 1, 1, CurFile);
    
    return strlen(pBuffer) + 1;
}

void ModelLoader3DS::readChunk(SChunk3DS* pChunk)
{
    pChunk->ReadedBytes = fread(&pChunk->ID, 1, 2, hFile_);
    pChunk->ReadedBytes += fread(&pChunk->Length, 1, 4, hFile_);
}

/*
 * ======= Object building function =======
 */

Entity* ModelLoader3DS::buildModel()
{
    
    /* Temporary variables */
    
    s32 i, j, k, v;
    
    dim::vector3df VertexCoord;
    dim::point2df TexCoord;
    
    /* Load all textures for the model */
    
    std::vector<video::Texture*> TextureList;
    video::Texture* hTexture;
    
    for (s32 i = 0, texi = 0; i < MaterialCounter_; ++i)
    {
        if (strlen(MaterialList_[i].strFile) > 0)
        {
            /* Load the current texture */
            hTexture = __spVideoDriver->loadTexture( TexturePath_ + MaterialList_[i].strFile );
            
            /* Add the texture to the list */
            TextureList.push_back(hTexture);
            MaterialList_[i].TextureID = texi++;
        }
        else
            MaterialList_[i].TextureID = -1;
    }
    
    /* Loop for all surfaces */
    
    SSurfaceObject3DS* CurObject;
    
    for (i = 0; i < ObjectCounter_; ++i)
    {
        
        /* Get the current surface */
        CurObject = &ObjectList_[i];
        
        /* Create new surface */
        hEntity_->createSurface();
        
        hEntity_->setSurfaceName(i, CurObject->strName);
        
        /* Add the texture to the surface */
        if (CurObject->hasTexture && MaterialList_[CurObject->MaterialID].TextureID != -1)
            hEntity_->addTexture( TextureList[ MaterialList_[CurObject->MaterialID].TextureID ] );
        
        /* Loop for all triangles */
        for (j = 0; j < CurObject->FacesCounter; ++j)
        {
            for (v = 0; v < 3; ++v)
            {
                /* Get the vertex information */
                k = CurObject->pFaces[j].VertexIndex[v];
                s8* Color = MaterialList_[CurObject->MaterialID].color;
                
                VertexCoord = CurObject->VerticesList[k];
                VertexCoord.Z = -VertexCoord.Z;
                
                if (CurObject->MapCoordList && k < CurObject->MapCoordCounter)
                    TexCoord = dim::point2df(CurObject->MapCoordList[k].X, -CurObject->MapCoordList[k].Y);
                
                /* Add new vertex */
                hEntity_->addVertex(
                    /*CurObject->TransformationMatrix * */VertexCoord,
                    TexCoord,
                    video::color(Color[0], Color[1], Color[2])
                );
            }
            
            /* Add new triangle */
            hEntity_->addTriangle(2, 1, 0);
            
            hEntity_->addVertexCounter(3);
        }
        
        /* Delete all allocated memories */
        spDeleteArray(CurObject->VerticesList);
        spDeleteArray(CurObject->MapCoordList);
        spDeleteArray(CurObject->pFaces);
        
    }
    
    /* Create the model finally */
    hEntity_->updateModeling();
    
    /* Build the animation */
    if (__spAnimator)
        buildAnimation();
    
    /* Return the object */
    return hEntity_;
}

/*
 * ======= Animation building function =======
 */

void ModelLoader3DS::buildAnimation()
{
    
    /* === Check if the object has a surface and a frame === */
    
    if (hEntity_->getMeshSurfaceCount() == 0 || AnimBoneList_.size() == 0)
        return;
    
    /* === Create a new animation === */
    
    Anim_ = __spAnimator->addAnimation(hEntity_);
    
    /* === Temporary variables === */
    
    std::vector<u32> VerticesIndices, VerticesSurfaces;
    u32 ParentBoneID = 0;
    s32 i, j, BoneID;
    
    /* === Loop for the frame bone list (creating the animation) === */
    
    for (std::vector<SSkeletonBone3DS*>::iterator it = AnimBoneList_.begin(); it != AnimBoneList_.end(); ++it)
    {
        
        /* Add a skeleton bone */
        BoneID = Anim_->addAnimSkeletonBone(
            (*it)->ParentBoneID,    // Parent bone identity number
            (*it)->Translation,     // Translation (location)
            (*it)->Quaternion,      // Quaternion
            (*it)->Scale,           // Scale
            (*it)->Name             // Bone name
        );
        
        /* Get the vertices list */
        /*VerticesIndices.resize((*it)->WeightsCount);
        VerticesSurfaces.resize((*it)->WeightsCount);
        
        for (i = 0; i < (*it)->WeightsCount; ++i)
        {
            VerticesIndices[i]  = (u32)(*it)->VerticesList[i].ID;
            VerticesSurfaces[i] = (*it)->VerticesList[i].Surface;
        }*/
        
        /* Add the vertices indices */
        Anim_->addSkeletonBoneVertices(BoneID, VerticesIndices, VerticesSurfaces);
        
        /* Create the animation keyframes */
        /*for (i = 0; i < AnimKeyframeCount_; ++i)
        {
            Anim_->addSkeletonBoneFrame(
                BoneID,                             // Bone identity number
                (*it)->KeyframeList[i].Position,    // Position keyframe
                (*it)->KeyframeList[i].Quaternion,  // Quaternion keyframe
                (*it)->KeyframeList[i].Scale,       // Scale keyframe
                1000.0f / AnimFPS_                  // Time (in ms.)
            );
        }*/
        
        /* Clear the vertices indices list */
        VerticesIndices.clear();
        
    }
    
    /* Update the skeleton bones */
    Anim_->updateSkeletonBones();
    
}


/*
 * ======= Main chunk =======
 */

void ModelLoader3DS::processNextChunk(SChunk3DS* pPreviousChunk)
{
    
    /* Temporary variables */
    
    u16 Version = 0;
    s32 pBuffer[DEFBUFFER_SIZE] = { 0 };
    SMaterial3DS Material = { 0 };
    SSurfaceObject3DS Object = { 0 };
    
    CurrentChunk_ = new SChunk3DS;
    
    /* Loop while reading the chunk */
    
    while (pPreviousChunk->ReadedBytes < pPreviousChunk->Length)
    {
        readChunk(CurrentChunk_);
        
        switch (CurrentChunk_->ID)
        {
            case MODEL_3DS_VERSION:
            {
                CurrentChunk_->ReadedBytes += fread(&Version, 1, CurrentChunk_->Length - CurrentChunk_->ReadedBytes, hFile_);
                
                if (Version > 0x03)
                    io::printWarning("version number of 3DS model is over 3 so it may load incorrect");
            }
            break;
            
            case MODEL_3DS_OBJECTINFO:
            break;
            
            case MODEL_3DS_OBJECT:
            {
                s32 i = ObjectCounter_++;
                ObjectList_.push_back(Object);
                
                // Initialize
                memset(&(ObjectList_[i]), 0, sizeof(SSurfaceObject3DS));
                ObjectList_[i].TransformationMatrix.matrixReset();
                
                CurrentChunk_->ReadedBytes += getString(hFile_, ObjectList_[i].strName);
                
                // Process next object chunk
                processNextObjectChunk(CurrentChunk_, &(ObjectList_[i]));
            }
            break;
            
            case MODEL_3DS_MATERIAL:
            {
                ++MaterialCounter_;
                
                MaterialList_.push_back(Material);
                
                // Process next material chunk
                processNextMaterialChunk(CurrentChunk_);
            }
            break;
            
            case MODEL_3DS_EDITKEYFRAME:
            {
                // Process next keyframe chunk
                processNextKeyframeChunk(CurrentChunk_);
            }
            break;
            
            default:
            {
                CurrentChunk_->ReadedBytes += fread(pBuffer, 1, CurrentChunk_->Length - CurrentChunk_->ReadedBytes, hFile_);
            }
            break;
        }
        
        pPreviousChunk->ReadedBytes += CurrentChunk_->ReadedBytes;
    }
    
    spDeletePointer(CurrentChunk_);
    CurrentChunk_ = pPreviousChunk;
    
}

/*
 * ======= Object chunk =======
 */

void ModelLoader3DS::processNextObjectChunk(SChunk3DS* pPreviousChunk, SSurfaceObject3DS* pObject)
{
    
    /* Temporary variables */
    
    s32 pBuffer[DEFBUFFER_SIZE] = { 0 };
    
    CurrentChunk_ = new SChunk3DS;
    
    /* Loop while reading the chunk */
    
    while (pPreviousChunk->ReadedBytes < pPreviousChunk->Length)
    {
        readChunk(CurrentChunk_);
        
        switch (CurrentChunk_->ID)
        {
            case MODEL_3DS_OBJECT_MESH:
                processNextObjectChunk(CurrentChunk_, pObject);
                break;
                
            case MODEL_3DS_OBJECT_VERTICES:
                readVertices(CurrentChunk_, pObject);
                break;
                
            case MODEL_3DS_OBJECT_FACES:
                readVerticesIndices(CurrentChunk_, pObject);
                break;
                
            case MODEL_3DS_OBJECT_MATERIAL:
                readObjectMaterial(CurrentChunk_, pObject);
                break;
                
            case MODEL_3DS_OBJECT_TEXCOORDS:
                readTexCoords(CurrentChunk_, pObject);
                break;
                
            case MODEL_3DS_OBJECT_MATRIX:
                readObjectMatrix(CurrentChunk_, pObject);
                break;
                
            default:
                CurrentChunk_->ReadedBytes += fread(pBuffer, 1, CurrentChunk_->Length - CurrentChunk_->ReadedBytes, hFile_);
                break;
        }
        
        pPreviousChunk->ReadedBytes += CurrentChunk_->ReadedBytes;
    }
    
    spDeletePointer(CurrentChunk_);
    CurrentChunk_ = pPreviousChunk;
    
}

/*
 * ======= Material chunk =======
 */

void ModelLoader3DS::processNextMaterialChunk(SChunk3DS* pPreviousChunk)
{
    
    /* Temporary variables */
    
    s32 pBuffer[DEFBUFFER_SIZE] = { 0 };
    
    CurrentChunk_ = new SChunk3DS;
    
    /* Loop while reading the chunk */
    
    while (pPreviousChunk->ReadedBytes < pPreviousChunk->Length)
    {
        readChunk(CurrentChunk_);
        
        switch (CurrentChunk_->ID)
        {
            case MODEL_3DS_MATNAME:
                CurrentChunk_->ReadedBytes += fread(
                    MaterialList_[MaterialCounter_ - 1].strName, 1,
                    CurrentChunk_->Length - CurrentChunk_->ReadedBytes, hFile_
                );
                break;
                
            case MODEL_3DS_MATDIFFUSE:
                readColorChunk(CurrentChunk_, &(MaterialList_[MaterialCounter_ - 1]));
                break;
                
            case MODEL_3DS_MATMAP:
                processNextMaterialChunk(CurrentChunk_);
                break;
                
            case MODEL_3DS_MATMAPFILE:
                CurrentChunk_->ReadedBytes += fread(
                    MaterialList_[MaterialCounter_ - 1].strFile, 1,
                    CurrentChunk_->Length - CurrentChunk_->ReadedBytes, hFile_
                );
                break;
                
            default:
                CurrentChunk_->ReadedBytes += fread(pBuffer, 1, CurrentChunk_->Length - CurrentChunk_->ReadedBytes, hFile_);
                break;
        }
        
        pPreviousChunk->ReadedBytes += CurrentChunk_->ReadedBytes;
    }
    
    spDeletePointer(CurrentChunk_);
    CurrentChunk_ = pPreviousChunk;
    
}

/*
 * ======= Keyframe chunk =======
 */

void ModelLoader3DS::processNextKeyframeChunk(SChunk3DS* pPreviousChunk)
{
    
    /* Temporary variables */
    
    u16 NodeID  = 0;
    u32 CurTime = 0;
    s32 Flags   = 0;
    s32 pBuffer[DEFBUFFER_SIZE] = { 0 };
    
    dim::vector3df PivotPoint, BoneVector;
    dim::quaternion BoneQuat;
    
    //std::vector<dim::vector3df> BonePositionList_, BoneScaleList_; // !!!
    //std::vector<dim::quaternion> BoneRotationList_;
    
    CurrentChunk_ = new SChunk3DS;
    
    /* Loop while reading the chunk */
    
    while (pPreviousChunk->ReadedBytes < pPreviousChunk->Length)
    {
        readChunk(CurrentChunk_);
        
        switch (CurrentChunk_->ID)
        {
            case MODEL_3DS_KEYF_TRACKINFO:
            {
                processNextKeyframeChunk(CurrentChunk_);
            }
            break;
            
            case MODEL_3DS_KEYF_NODE_ID:
            {
                pPreviousChunk->ReadedBytes += fread(&NodeID, 1, sizeof(u16), hFile_);
                
                animCreateNewBone(NodeID);
            }
            break;
            
            case MODEL_3DS_KEYF_BONENAME:
            {
                s8 pName[CurrentChunk_->Length - CurrentChunk_->ReadedBytes];
                
                CurrentChunk_->ReadedBytes += fread(pName, 1, CurrentChunk_->Length - CurrentChunk_->ReadedBytes, hFile_);
                
                animSetBoneName(io::stringc(pName));
            }
            break;
            
            case MODEL_3DS_KEYF_CURTIME:
            {
                CurrentChunk_->ReadedBytes += fread(&CurTime, 1, sizeof(u32), hFile_);
            }
            break;
            
            case MODEL_3DS_KEYF_PIVOTPOINT:
            {
                fread(&PivotPoint.X, 1, sizeof(f32), hFile_);
                fread(&PivotPoint.Y, 1, sizeof(f32), hFile_);
                fread(&PivotPoint.Z, 1, sizeof(f32), hFile_);
                
                CurrentChunk_->ReadedBytes += CurrentChunk_->Length - CurrentChunk_->ReadedBytes;
            }
            break;
            /*
            case MODEL_3DS_KEYF_BOUNDBOX:
            {
                
                
            }
            break;
            */
            case MODEL_3DS_KEYF_BONEPOSITION:
            {
                u16 FrameNumber, Number;
                f32 EmptyData;
                
                fseek(hFile_, 10, SEEK_CUR);
                fread(&Number, 1, sizeof(u16), hFile_);
                fseek(hFile_, 2, SEEK_CUR);
                
                while (Number-- > 0)
                {
                    fread(&FrameNumber, 1, sizeof(u16), hFile_);
                    fread(&Flags, 1, sizeof(s32), hFile_);
                    
                    for (s32 i = 0; i < 32; ++i)
                    {
                        if (Flags & (1 << i))
                            fread(&EmptyData, 1, sizeof(f32), hFile_);
                    }
                    
                    fread(&BoneVector.X, 1, sizeof(f32), hFile_);
                    fread(&BoneVector.Y, 1, sizeof(f32), hFile_);
                    fread(&BoneVector.Z, 1, sizeof(f32), hFile_);
                    
                    BonePositionList_.push_back(BoneVector);
                }
                
                /* Add the position list to the current bone */
                animAddBonePositionList();
                
                CurrentChunk_->ReadedBytes += CurrentChunk_->Length - CurrentChunk_->ReadedBytes;
            }
            break;
            
            case MODEL_3DS_KEYF_BONEROTATION:
            {
                u16 FrameNumber, Number;
                f32 EmptyData;
                
                fseek(hFile_, 10, SEEK_CUR);
                fread(&Number, 1, sizeof(u16), hFile_);
                fseek(hFile_, 2, SEEK_CUR);
                
                while (Number-- > 0)
                {
                    fread(&FrameNumber, 1, sizeof(u16), hFile_);
                    fread(&Flags, 1, sizeof(s32), hFile_);
                    
                    for (s32 i = 0; i < 32; ++i)
                    {
                        if (Flags & (1 << i))
                            fread(&EmptyData, 1, sizeof(f32), hFile_);
                    }
                    
                    fread(&BoneQuat.W, 1, sizeof(f32), hFile_);
                    fread(&BoneQuat.X, 1, sizeof(f32), hFile_);
                    fread(&BoneQuat.Y, 1, sizeof(f32), hFile_);
                    fread(&BoneQuat.Z, 1, sizeof(f32), hFile_);
                    
                    BoneRotationList_.push_back(BoneQuat);
                }
                
                /* Add the rotation list to the current bone */
                animAddBoneRotationList();
                
                CurrentChunk_->ReadedBytes += CurrentChunk_->Length - CurrentChunk_->ReadedBytes;
            }
            break;
            
            case MODEL_3DS_KEYF_BONESCALE:
            {
                u16 FrameNumber, Number;
                f32 EmptyData;
                
                fseek(hFile_, 10, SEEK_CUR);
                fread(&Number, 1, sizeof(u16), hFile_);
                fseek(hFile_, 2, SEEK_CUR);
                
                while (Number-- > 0)
                {
                    fread(&FrameNumber, 1, sizeof(u16), hFile_);
                    fread(&Flags, 1, sizeof(s32), hFile_);
                    
                    for (s32 i = 0; i < 32; ++i)
                    {
                        if (Flags & (1 << i))
                            fread(&EmptyData, 1, sizeof(f32), hFile_);
                    }
                    
                    fread(&BoneVector.X, 1, sizeof(f32), hFile_);
                    fread(&BoneVector.Y, 1, sizeof(f32), hFile_);
                    fread(&BoneVector.Z, 1, sizeof(f32), hFile_);
                    
                    BoneScaleList_.push_back(BoneVector);
                }
                
                /* Add the scale list to the current bone */
                animAddBoneScaleList();
                
                CurrentChunk_->ReadedBytes += CurrentChunk_->Length - CurrentChunk_->ReadedBytes;
            }
            break;
            
            default:
            {
                CurrentChunk_->ReadedBytes += fread(pBuffer, 1, CurrentChunk_->Length - CurrentChunk_->ReadedBytes, hFile_);
            }
            break;
        }
        
        pPreviousChunk->ReadedBytes += CurrentChunk_->ReadedBytes;
    }
    
    spDeletePointer(CurrentChunk_);
    CurrentChunk_ = pPreviousChunk;
    
}

/*
 * ======= Other chunks =======
 */

void ModelLoader3DS::readColorChunk(SChunk3DS* pPreviousChunk, SMaterial3DS* Material)
{
    readChunk(TempChunk_);
    
    TempChunk_->ReadedBytes += fread(Material->color, 1, TempChunk_->Length - TempChunk_->ReadedBytes, hFile_);
    
    pPreviousChunk->ReadedBytes += TempChunk_->ReadedBytes;
}

void ModelLoader3DS::readVerticesIndices(SChunk3DS* pPreviousChunk, SSurfaceObject3DS* Object)
{
    u16 index = 0;
    pPreviousChunk->ReadedBytes += fread(&Object->FacesCounter, 1, 2, hFile_);
    
    Object->pFaces = new SFace3DS[Object->FacesCounter];
    memset(Object->pFaces, 0, sizeof(SFace3DS) * Object->FacesCounter);
    
    for (s32 i = 0; i < Object->FacesCounter; ++i)
    {
        // Read the three vertex indices
        for (s32 j = 0; j < 3; ++j)
        {
            pPreviousChunk->ReadedBytes += fread(&index, 1, sizeof(u16), hFile_);
            Object->pFaces[i].VertexIndex[j] = index;
        }
        
        // Read the flags
        pPreviousChunk->ReadedBytes += fread(&Object->pFaces[i].Flags, 1, sizeof(u16), hFile_);
    }
}

void ModelLoader3DS::readTexCoords(SChunk3DS* pPreviousChunk, SSurfaceObject3DS* Object)
{
    pPreviousChunk->ReadedBytes += fread(&Object->MapCoordCounter, 1, 2, hFile_);
    
    Object->MapCoordList = new dim::point2df[Object->MapCoordCounter];
    
    pPreviousChunk->ReadedBytes += fread(Object->MapCoordList, 1, pPreviousChunk->Length - pPreviousChunk->ReadedBytes, hFile_);
}

void ModelLoader3DS::readVertices(SChunk3DS* pPreviousChunk, SSurfaceObject3DS* Object)
{
    pPreviousChunk->ReadedBytes += fread(&Object->VerticesCounter, 1, 2, hFile_);
    
    Object->VerticesList = new dim::vector3df[Object->VerticesCounter];
    memset(Object->VerticesList, 0, sizeof(dim::vector3df) * Object->VerticesCounter);
    
    pPreviousChunk->ReadedBytes += fread(Object->VerticesList, 1, pPreviousChunk->Length - pPreviousChunk->ReadedBytes, hFile_);
}

void ModelLoader3DS::readObjectMaterial(SChunk3DS* pPreviousChunk, SSurfaceObject3DS* Object)
{
    c8 strMaterial[255]         = { 0 };
    s32 pBuffer[DEFBUFFER_SIZE] = { 0 };
    
    pPreviousChunk->ReadedBytes += getString(hFile_, strMaterial);
    
    for (s32 i = 0; i < MaterialCounter_; ++i)
    {
        if (strcmp(strMaterial, MaterialList_[i].strName) == 0)
        {
            Object->MaterialID = i;
            if (strlen(MaterialList_[i].strFile) > 0)
                Object->hasTexture = true;
            break;
        }
    }
    
    pPreviousChunk->ReadedBytes += fread(pBuffer, 1, pPreviousChunk->Length - pPreviousChunk->ReadedBytes, hFile_);
}

void ModelLoader3DS::readObjectMatrix(SChunk3DS* pPreviousChunk, SSurfaceObject3DS* Object)
{
    f32 pMat[4][3];
    
    pPreviousChunk->ReadedBytes += fread(&pMat, 1, 12*sizeof(f32), hFile_);
    
    // Load the matrix identity
    Object->TransformationMatrix.matrixReset();
    
    // Set the matrix values
    for (s32 i = 0, j; i < 4; ++i)
    {
        for (j = 0; j < 3; ++j)
            Object->TransformationMatrix(i, j) = pMat[i][j];
    }
}

/*
 * ========== Animation functions ==========
 */

void ModelLoader3DS::animCreateNewBone(u32 ID)
{
    /* Allocate a new bone */
    SSkeletonBone3DS* BoneData = new SSkeletonBone3DS;
    
    /* Set the default bone values */
    BoneData->ID = ID;
    
    /* Add a new Bone */
    AnimBoneList_.push_back(BoneData);
    
    /* Set the current bone */
    CurBone_ = BoneData;
}

void ModelLoader3DS::animSetBoneName(io::stringc Name)
{
    if (CurBone_)
        CurBone_->Name = Name;
}

void ModelLoader3DS::animAddBonePositionList()
{
    /* Add the list to the bone */
    if (CurBone_)
    {
        CurBone_->PositionList = BonePositionList_;
        
        if (BonePositionList_.size())
            CurBone_->Translation = BonePositionList_[0];
    }
    
    /* Clear the old list */
    BonePositionList_.clear();
}

void ModelLoader3DS::animAddBoneRotationList()
{
    /* Add the list to the bone */
    if (CurBone_)
    {
        CurBone_->RotationList = BoneRotationList_;
        
        if (BoneRotationList_.size())
            CurBone_->Quaternion = BoneRotationList_[0];
    }
    
    /* Clear the old list */
    BoneRotationList_.clear();
}

void ModelLoader3DS::animAddBoneScaleList()
{
    /* Add the list to the bone */
    if (CurBone_)
    {
        CurBone_->ScaleList = BoneScaleList_;
        
        if (BoneScaleList_.size())
            CurBone_->Scale = BoneScaleList_[0];
    }
    
    /* Clear the old list */
    BoneScaleList_.clear();
}


} // /namespace scene

} // /namespace sp







// ================================================================================
