/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * Model loading B3D-fileformat (Blitz3D) file
 *
 */

#include "spModelLoaderB3D.hpp"
#include "spMainDevice.hpp"


namespace sp
{

extern video::VideoDriver* __spVideoDriver;

namespace scene
{


ModelLoaderB3D::ModelLoaderB3D()
{
    init();
}
ModelLoaderB3D::~ModelLoaderB3D()
{
    clear();
}

Entity* ModelLoaderB3D::loadModel(io::stringc &Filename, io::stringc &TexturePath)
{
    Filename_       = Filename;
    TexturePath_    = TexturePath;
    
    hEntity_        = new Entity();
    
    if ( !( hFile_ = hFileLoader_->openFile(Filename, io::FILE_READ_BINARY) ) )
        return hEntity_;
    
    if (!loadModelData())
        io::printError("loading B3D model failed");
    
    hFileLoader_->closeFile(hFile_);
    
    return hEntity_;
}


/*
 * ========== Private: ==========
 */


void ModelLoaderB3D::init()
{
    /* General settings */
    CurPos_     = 0;
    Surface_    = -1;
    
    /* Initialize stack memory */
    memset(Stack_, 0, sizeof(Stack_));
    
    /* Initialize building members */
    isExpactNewKeyframes_   = false;
    wasLastNodeMesh_        = false;
    CurBone_                = -1;
    AnimKeyframeCount_      = 0;
    AnimFPS_                = 1000.0f;
    
    /* Allocate all pointers */
    hFileLoader_     = new io::FileLoader();
}
void ModelLoaderB3D::clear()
{
    /* Delete all pointers */
    spDeletePointer(hFileLoader_);
    
    /* Delete the bone list */
    for (std::vector<SSkeletonBoneB3D*>::iterator it = AnimBoneList_.begin(); it != AnimBoneList_.end(); ++it)
        spDeletePointer(*it);
}

io::stringc ModelLoaderB3D::readChunk()
{
    // Read chunk name
    c8 ChunkName[5] = { 0 };
    hFileLoader_->readStream(hFile_, ChunkName, 4);
    
    // Read chunk size
    s32 Size = hFileLoader_->readValue<s32>(hFile_);
    ++CurPos_;
    Stack_[CurPos_] = hFileLoader_->getFileSeek(hFile_) + Size;
    
    // Returns chunk name
    return io::stringc(ChunkName);
}

void ModelLoaderB3D::breakChunk()
{
    hFileLoader_->setFileSeek(hFile_, Stack_[CurPos_]);
    --CurPos_;
}

s32 ModelLoaderB3D::getChunkSize()
{
    return Stack_[CurPos_] - hFileLoader_->getFileSeek(hFile_);
}

/*
 * Loads all chunks in the file
 */

/*
 * Texture flags:
 * 1: Color
 * 2: Alpha
 * 4: Masked
 * 8: Mipmapped
 * 16: Clamp U
 * 32: Clamp V
 * 64: Spherical reflection map
 */

void ModelLoaderB3D::dumpChunks(io::stringc Tab)
{
    
    // Temporary variables
    io::stringc ChunkName;
    Tab += io::stringc(' ');
    
    // Loop while there are chunk data
    while (getChunkSize())
    {
        
        ChunkName = readChunk();
        
        if (ChunkName == "ANIM") // Animation (general information)
            readChunkANIM();
        else if (ChunkName == "KEYS") // Keyframes (for bone information)
            readChunkKEYS();
        else if (ChunkName == "TEXS") // Textures
            readChunkTEXS();
        else if (ChunkName == "BRUS") // Brushes
            readChunkBRUS();
        else if (ChunkName == "VRTS") // Vertices
            readChunkVRTS();
        else if (ChunkName == "TRIS") // Triangles
            readChunkTRIS();
        else if (ChunkName == "MESH") // Mesh (texture information)
            readChunkMESH();
        else if (ChunkName == "BONE") // Bone (for skeletal animation)
            readChunkBONE(Tab);
        else if (ChunkName == "NODE") // Node (new mesh or new bone)
            readChunkNODE();
        
        dumpChunks(Tab);
        breakChunk();
        
    } // wend
    
}


/*
 * ========== Chunk functions ==========
 */

bool ModelLoaderB3D::readChunkANIM()
{
    
    s32 Flags           = hFileLoader_->readValue<s32>(hFile_);
    AnimKeyframeCount_  = hFileLoader_->readValue<s32>(hFile_);
    AnimFPS_            = hFileLoader_->readValue<f32>(hFile_);
    
    if (AnimFPS_ == 0.0f)
        AnimFPS_ = 1000.0f / 60.0f;
    
    return true;
    
}

bool ModelLoaderB3D::readChunkKEYS()
{
    
    #if 0
    if (CurBone_ < 0 || CurBone_ >= AnimBoneList_.size())
    {
        io::printError(
            io::stringc("B3D file loading problem, animation array index out of bounds ( ")+
            io::stringc(CurBone_) + io::stringc(" )")
        );
        return false;
    }
    #endif
    
    // Temporary variables
    s32 i, LastFrame, Frame = 0;
    s32 Size = 4;
    
    dim::vector3df KeyP, KeyS;
    dim::quaternion KeyR;
    
    s32 Flags = hFileLoader_->readValue<s32>(hFile_);
    
    if (Flags & 1) Size += 12;
    if (Flags & 2) Size += 12;
    if (Flags & 4) Size += 16;
    
    s32 KeysCount = getChunkSize() / Size;
    
    // Allocate new keyframe memory
    if (isExpactNewKeyframes_)
    {
        if (CurBone_ != -1)
        {
            // Allocate new keyframes
            AnimBoneList_[CurBone_]->KeyframeCount  = KeysCount;
            AnimBoneList_[CurBone_]->KeyframeList   = new SSkeletonBoneB3D::SBoneKeyframeInfo[AnimKeyframeCount_];
            
            // Default values
            for (i = 0; i < AnimKeyframeCount_; ++i)
                AnimBoneList_[CurBone_]->KeyframeList[i].Scale.set(1, 1, 1);
        }
        isExpactNewKeyframes_ = false;
    }
    
    if (KeysCount*Size == getChunkSize())
    {
        
        while (getChunkSize())
        {
            
            LastFrame = Frame;
            
            Frame = hFileLoader_->readValue<s32>(hFile_) - 1;
            
            /* === Keyframe === */
            
            // Keyframe - position
            if (Flags & 1)
            {
                KeyP.X = hFileLoader_->readValue<f32>(hFile_);
                KeyP.Y = hFileLoader_->readValue<f32>(hFile_);
                KeyP.Z = hFileLoader_->readValue<f32>(hFile_);
                
                if (CurBone_ != -1)
                {
                    AnimBoneList_[CurBone_]->KeyframeList[Frame].Position = KeyP;
                    
                    for (i = LastFrame + 1; i < Frame; ++i)
                    {
                        AnimBoneList_[CurBone_]->KeyframeList[i].Position =
                            AnimBoneList_[CurBone_]->KeyframeList[LastFrame].Position +
                            ( KeyP - AnimBoneList_[CurBone_]->KeyframeList[LastFrame].Position ) *
                            f32( i - LastFrame ) / f32( Frame - LastFrame );
                    }
                }
            }
            
            // Keyframe - scale
            if (Flags & 2)
            {
                KeyS.X = hFileLoader_->readValue<f32>(hFile_);
                KeyS.Y = hFileLoader_->readValue<f32>(hFile_);
                KeyS.Z = hFileLoader_->readValue<f32>(hFile_);
                
                if (CurBone_ != -1)
                {
                    AnimBoneList_[CurBone_]->KeyframeList[Frame].Scale = KeyS;
                    
                    for (i = LastFrame + 1; i < Frame; ++i)
                    {
                        AnimBoneList_[CurBone_]->KeyframeList[i].Scale =
                            AnimBoneList_[CurBone_]->KeyframeList[LastFrame].Scale +
                            ( KeyS - AnimBoneList_[CurBone_]->KeyframeList[LastFrame].Scale ) *
                            f32( i - LastFrame ) / f32( Frame - LastFrame );
                    }
                }
            }
            
            // Keyframe - rotation
            if (Flags & 4)
            {
                KeyR.W = hFileLoader_->readValue<f32>(hFile_);
                KeyR.X = -hFileLoader_->readValue<f32>(hFile_);
                KeyR.Y = -hFileLoader_->readValue<f32>(hFile_);
                KeyR.Z = -hFileLoader_->readValue<f32>(hFile_);
                
                if (CurBone_ != -1)
                {
                    AnimBoneList_[CurBone_]->KeyframeList[Frame].Quaternion = KeyR;
                    
                    for (i = LastFrame + 1; i < Frame; ++i)
                    {
                        AnimBoneList_[CurBone_]->KeyframeList[i].Quaternion =
                            AnimBoneList_[CurBone_]->KeyframeList[LastFrame].Quaternion +
                            ( KeyR - AnimBoneList_[CurBone_]->KeyframeList[LastFrame].Quaternion ) *
                            f32( i - LastFrame ) / f32( Frame - LastFrame );
                    }
                }
            }
            
        }
        
    }
    else
    {
        io::printWarning("illegal number of keyframes in B3D model");
        return false;
    }
    
    return true;
    
}

bool ModelLoaderB3D::readChunkTEXS()
{
    
    // Temporary variables
    io::stringc TexFilename;
    s32 Flags, Blend;
    f32 Rotation;
    dim::point2df Pos;
    dim::size2df Scale;
    
    while (getChunkSize())
    {
        // Reading
        TexFilename     = hFileLoader_->readStringANSIC(hFile_);
        Flags           = hFileLoader_->readValue<s32>(hFile_);
        Blend           = hFileLoader_->readValue<s32>(hFile_);
        Pos.X           = hFileLoader_->readValue<f32>(hFile_);
        Pos.Y           = hFileLoader_->readValue<f32>(hFile_);
        Scale.Width     = hFileLoader_->readValue<f32>(hFile_);
        Scale.Height    = hFileLoader_->readValue<f32>(hFile_);
        Rotation        = hFileLoader_->readValue<f32>(hFile_);
        
        // Loading
        STextureSurfaceB3D TextureSurfaceData;
        {
            // General settings
            TextureSurfaceData.hTexture         = __spVideoDriver->loadTexture(TexturePath_ + TexFilename);
            TextureSurfaceData.Pos              = Pos;
            TextureSurfaceData.Scale            = Scale;
            TextureSurfaceData.isSphereMapping  = (Flags & 64) ? true : false;
            
            // Flags
            if (Flags & 2) // Alpha
                TextureSurfaceData.hTexture->setTransFilterAlpha();
            else if (Flags & 4) // Masked
                TextureSurfaceData.hTexture->setTransFilter(video::color());
        }
        TextureList_.push_back(TextureSurfaceData);
    }
    
    return true;
    
}

bool ModelLoaderB3D::readChunkBRUS()
{
    
    // Temporary variables
    io::stringc Name;
    f32 Red, Green, Blue, Alpha, Shininess;
    s32 Blend, Effects, TexID;
    s32 TexCount = hFileLoader_->readValue<s32>(hFile_);
    
    while (getChunkSize())
    {
        Name        = hFileLoader_->readStringANSIC(hFile_);
        Red         = hFileLoader_->readValue<f32>(hFile_);
        Green       = hFileLoader_->readValue<f32>(hFile_);
        Blue        = hFileLoader_->readValue<f32>(hFile_);
        Alpha       = hFileLoader_->readValue<f32>(hFile_);
        Shininess   = hFileLoader_->readValue<f32>(hFile_);
        Blend       = hFileLoader_->readValue<s32>(hFile_);
        Effects     = hFileLoader_->readValue<s32>(hFile_);
        
        for (s32 i = 0; i < TexCount; ++i)
        {
            TexID = hFileLoader_->readValue<s32>(hFile_);
            BrushTextureList_.push_back(TexID);
        }
    }
    
    return true;
    
}

bool ModelLoaderB3D::readChunkVRTS()
{
    
    s32 Flags = hFileLoader_->readValue<s32>(hFile_);
    s32 tcSets = hFileLoader_->readValue<s32>(hFile_);
    s32 tcSize = hFileLoader_->readValue<s32>(hFile_);
    
    // Temporary variables
    s32 Size = 12 + tcSets*tcSize*4;
    f32 TexCoord[3];
    
    SVertexB3D Vertex;
    
    if (Flags & 1) Size += 12;
    if (Flags & 2) Size += 16;
    
    s32 VerticesCount = getChunkSize() / Size;
    
    if (VerticesCount*Size == getChunkSize())
    {
        
        while (getChunkSize())
        {
            // Standard values
            Vertex.Normal   = dim::vector3df(0, 0, 1);
            Vertex.Color    = video::color(255);
            
            // Reading
            Vertex.Coord.X  = hFileLoader_->readValue<f32>(hFile_);
            Vertex.Coord.Y  = hFileLoader_->readValue<f32>(hFile_);
            Vertex.Coord.Z  = hFileLoader_->readValue<f32>(hFile_);
            
            if (Flags & 1)
            {
                Vertex.Normal.X = hFileLoader_->readValue<f32>(hFile_);
                Vertex.Normal.Y = hFileLoader_->readValue<f32>(hFile_);
                Vertex.Normal.Z = hFileLoader_->readValue<f32>(hFile_);
            }
            if (Flags & 2)
            {
                Vertex.Color.Red    = (s32)(hFileLoader_->readValue<f32>(hFile_) * 255);
                Vertex.Color.Green  = (s32)(hFileLoader_->readValue<f32>(hFile_) * 255);
                Vertex.Color.Blue   = (s32)(hFileLoader_->readValue<f32>(hFile_) * 255);
                Vertex.Color.Alpha  = (s32)(hFileLoader_->readValue<f32>(hFile_) * 255);
            }
            for (s32 j = 1; j <= tcSets*tcSize; ++j)
            {
                TexCoord[j-1] = hFileLoader_->readValue<f32>(hFile_);
            }
            
            Vertex.TexCoord.X   = TexCoord[0];
            Vertex.TexCoord.Y   = TexCoord[1];
            
            Vertex.SurfVertexID = VerticesList_.size();
            
            // Add new vertex
            VerticesList_.push_back(Vertex);
        } // wend
        
    }
    else
    {
        io::printWarning("illegal number of vertices in B3D model");
        return false;
    }
    
    return true;
    
}

bool ModelLoaderB3D::readChunkTRIS()
{
    
    // Temporary vairables
    s32 BrushID = hFileLoader_->readValue<s32>(hFile_);
    s32 Size = 12;
    s32 TriangleCount = getChunkSize() / Size;
    s32 v0, v1, v2;
    
    std::list<s32> VertexIndices, VertexIndicesUnique;
    
    if (BrushID == -1)
        BrushID = CurBrushID_;
    
    if (BrushID >= 0 && BrushID < BrushTextureList_.size())
        BrushID = BrushTextureList_[BrushID];
    
    if (TriangleCount*Size == getChunkSize())
    {
        
        while (getChunkSize())
        {
            // Reading
            v0 = hFileLoader_->readValue<s32>(hFile_);
            v1 = hFileLoader_->readValue<s32>(hFile_);
            v2 = hFileLoader_->readValue<s32>(hFile_);
            
            // Save in list
            VertexIndices.push_back(v0);
            VertexIndices.push_back(v1);
            VertexIndices.push_back(v2);
        }
        
        // Copy the list
        VertexIndicesUnique = VertexIndices;
        
        // Remove multiple elements
        VertexIndicesUnique.sort();
        VertexIndicesUnique.unique();
        
        // Create a new surface
        Surface_ = (s32)hEntity_->createSurface();
        
        hEntity_->setSurfaceName(Surface_, CurName_);
        
        SBrushSurfaceB3D BrushSurfaceData;
        {
            BrushSurfaceData.BrushID    = BrushID;
            BrushSurfaceData.Surface    = Surface_;
        }
        BrushSurfaceList_.push_back(BrushSurfaceData);
        
        SVertexB3D Vertex;
        
        // Create all vertices
        for (std::list<s32>::iterator it = VertexIndicesUnique.begin(); it != VertexIndicesUnique.end(); ++it)
        {
            Vertex = VerticesList_[*it];
            
            VerticesList_[*it].SurfVertexID = hEntity_->addVertex(
                Vertex.Coord, Vertex.Normal, Vertex.TexCoord, Vertex.Color, 0.0f
            );
            VerticesList_[*it].Surface = Surface_;
        }
        
        // Create all triangles
        for (std::list<s32>::iterator it = VertexIndices.begin(); it != VertexIndices.end();)
        {
            // Get the indices
            v0 = VerticesList_[*it++].SurfVertexID;
            v1 = VerticesList_[*it++].SurfVertexID;
            v2 = VerticesList_[*it++].SurfVertexID;
            
            // Creating
            hEntity_->addTriangle(v0, v1, v2);
        }
        
    }
    else
    {
        io::printWarning("illegal number of triangles in B3D model");
        return false;
    }
    
    return true;
    
}

bool ModelLoaderB3D::readChunkMESH()
{
    
    // Read the current brush
    CurBrushID_ = hFileLoader_->readValue<s32>(hFile_);
    
    // Set the information: a new mesh has been created
    wasLastNodeMesh_ = true;
    
    // Clear the vertices list of the last mesh
    VerticesList_.clear();
    
    return true;
    
}

bool ModelLoaderB3D::readChunkBONE(io::stringc &Tab)
{
    
    s32 Size = 8;
    s32 WeightsCount = getChunkSize() / Size;
    s32 VertexID;
    f32 Weight;
    
    isExpactNewKeyframes_ = true;
    
    if (WeightsCount*Size == getChunkSize())
    {
        
        s32 i = 0;
        
        // Temporary bone memory
        SSkeletonBoneB3D* BoneData = new SSkeletonBoneB3D();
        {
            BoneData->TabSize       = Tab.size();
            BoneData->ParentBoneID  = 0;
            BoneData->Name          = CurName_;
            
            BoneData->Translation   = SurfPosition_;
            BoneData->Quaternion    = SurfQuaternion_;
            BoneData->Scale         = SurfScale_;
            
            BoneData->WeightsCount  = WeightsCount;
            BoneData->VerticesList  = new SSkeletonBoneB3D::SBoneVertexInfo[ WeightsCount ];
        }
        
        // Read the vertex identities
        while (getChunkSize())
        {
            VertexID    = hFileLoader_->readValue<s32>(hFile_);
            Weight      = hFileLoader_->readValue<f32>(hFile_);
            
            {
                BoneData->VerticesList[i].Surface   = VerticesList_[VertexID].Surface;
                BoneData->VerticesList[i].ID        = VerticesList_[VertexID].SurfVertexID;
                BoneData->VerticesList[i].Weight    = Weight;
            }
            
            ++i;
        }
        
        // Add a new bone
        AnimBoneList_.push_back(BoneData);
        
        CurBone_ = AnimBoneList_.size() - 1;
        
    }
    else
    {
        io::printWarning("illegal number of bone weights in B3D model");
        return false;
    }
    
    return true;
    
}

bool ModelLoaderB3D::readChunkNODE()
{
    
    // Temporary vairables
    dim::vector3df Position, Scale;
    dim::quaternion Quaternion;
    
    // Reading
    CurName_        = hFileLoader_->readStringANSIC(hFile_);
    
    // Reading - position
    Position.X      = hFileLoader_->readValue<f32>(hFile_);
    Position.Y      = hFileLoader_->readValue<f32>(hFile_);
    Position.Z      = hFileLoader_->readValue<f32>(hFile_);
    
    // Reading - scale
    Scale.X         = hFileLoader_->readValue<f32>(hFile_);
    Scale.Y         = hFileLoader_->readValue<f32>(hFile_);
    Scale.Z         = hFileLoader_->readValue<f32>(hFile_);
    
    // Reading - rotation
    Quaternion.W    = hFileLoader_->readValue<f32>(hFile_);
    Quaternion.X    = hFileLoader_->readValue<f32>(hFile_);
    Quaternion.Y    = hFileLoader_->readValue<f32>(hFile_);
    Quaternion.Z    = hFileLoader_->readValue<f32>(hFile_);
    
    // Transform the last surface
    updateSurfaceLocation();
    
    // Create a new surface
    SurfPosition_   = Position;
    SurfQuaternion_ = Quaternion;
    SurfScale_      = Scale;
    
}


/*
 * ========== Building functions ==========
 */

void ModelLoaderB3D::updateSurfaceLocation()
{
    if (Surface_ != -1 && wasLastNodeMesh_)
    {
        wasLastNodeMesh_ = false;
        
        dim::matrix4f Mat;
        Mat.matrixReset();
        
        Mat.matrixTranslate(SurfPosition_);
        Mat *= SurfQuaternion_.getMatrixTransposed();
        Mat.matrixScale(SurfScale_);
        
        hEntity_->doMeshTransform(Mat, (u32)Surface_);
    }
}

void ModelLoaderB3D::updateTexturing()
{
    for (std::vector<SBrushSurfaceB3D>::iterator it = BrushSurfaceList_.begin(); it != BrushSurfaceList_.end(); ++it)
    {
        if (it->Surface != -1 && it->BrushID < TextureList_.size() && it->BrushID >= 0)
        {
            hEntity_->selectSurface(it->Surface);
            
            hEntity_->addTexture(TextureList_[it->BrushID].hTexture);
            hEntity_->doTextureTransform(0, TextureList_[it->BrushID].Scale);
            hEntity_->doTextureTranslate(0, TextureList_[it->BrushID].Pos);
            
            if (TextureList_[it->BrushID].isSphereMapping)
                hEntity_->setMappingGen(0, video::MAPGEN_SPHERE_MAP);
        }
    }
}

void ModelLoaderB3D::buildAnimation()
{
    
    /* === Check if the object has a surface and a frame === */
    
    if (hEntity_->getMeshSurfaceCount() == 0 || AnimBoneList_.size() == 0)
        return;
    
    /* === Create a new animation === */
    
    Anim_ = __spAnimator->addAnimation(hEntity_);
    
    /* === Temporary variables === */
    
    std::vector<u32> VerticesIndices, VerticesSurfaces;
    u32 ParentBoneID = 0;
    s32 i = 0, j, BoneID;
    
    /* === Loop for the frame bone list (get the correct parent bones) === */
    
    for (std::vector<SSkeletonBoneB3D*>::iterator it = AnimBoneList_.begin() + 1, itsub; it != AnimBoneList_.end(); ++it, ++i)
    {
        j = i;
        for (itsub = it - 1; itsub >= AnimBoneList_.begin(); --itsub, --j)
        {
            if ((*it)->TabSize == (*itsub)->TabSize + 1)
            {
                (*it)->ParentBoneID = j + 1;
                break;
            }
        }
    }
    
    /* === Loop for the frame bone list (creating the animation) === */
    
    for (std::vector<SSkeletonBoneB3D*>::iterator it = AnimBoneList_.begin(); it != AnimBoneList_.end(); ++it)
    {
        
        // Add a skeleton bone
        BoneID = Anim_->addAnimSkeletonBone(
            (*it)->ParentBoneID,    // Parent bone identity number
            (*it)->Translation,     // Translation (location)
            (*it)->Quaternion,      // Quaternion
            (*it)->Scale,           // Scale
            (*it)->Name             // Bone name
        );
        
        // Get the vertices list
        VerticesIndices.resize((*it)->WeightsCount);
        VerticesSurfaces.resize((*it)->WeightsCount);
        
        for (i = 0; i < (*it)->WeightsCount; ++i)
        {
            VerticesIndices[i]  = (u32)(*it)->VerticesList[i].ID;
            VerticesSurfaces[i] = (*it)->VerticesList[i].Surface;
        }
        
        // Add the vertices indices
        Anim_->addSkeletonBoneVertices(BoneID, VerticesIndices, VerticesSurfaces);
        
        // Create the animation keyframes
        for (i = 0; i < AnimKeyframeCount_; ++i)
        {
            Anim_->addSkeletonBoneFrame(
                BoneID,                             // Bone identity number
                (*it)->KeyframeList[i].Position,    // Position keyframe
                (*it)->KeyframeList[i].Quaternion,  // Quaternion keyframe
                (*it)->KeyframeList[i].Scale,       // Scale keyframe
                1000.0f / AnimFPS_                  // Time (in ms.)
            );
        }
        
        // Clear the vertices indices list
        VerticesIndices.clear();
        
    }
    
    /* Update the skeleton bones */
    Anim_->updateSkeletonBones();
    
}

bool ModelLoaderB3D::loadModelData()
{
    
    // Identity
    if (readChunk() != "BB3D")
    {
        io::printError("B3D model has invalid identity");
        return false;
    }
    
    // Version
    if (hFileLoader_->readValue<s32>(hFile_) / 100 > 0)
    {
        io::printError("B3D model has invalid version");
        return false;
    }
    
    // Read all chunks
    dumpChunks();
    
    // Update the final surface
    updateSurfaceLocation();
    
    // Update texturing
    updateTexturing();
    
    // Update modeling
    hEntity_->updateModeling();
    
    // Create animation
    if (__spAnimator)
        buildAnimation();
    
    return true;
}






} // /namespace scene

} // /namespace sp








// ================================================================================
