/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * Model loading D3D-fileformat (Direct3D/ "*.x" DirectX) file
 *
 */

#include "spModelLoaderD3D.hpp"
#include "spMainDevice.hpp"


namespace sp
{

extern video::VideoDriver* __spVideoDriver;

namespace scene
{


const long D3D_MAGIC    = ((long)'x' + ((long)'o' << 8) + ((long)'f' << 16) + ((long)' ' << 24));
const long D3D_VERSION  = ((long)'0' + ((long)'3' << 8) + ((long)'0' << 16) + ((long)'2' << 24));
const long D3D_BINARY   = ((long)'b' + ((long)'i' << 8) + ((long)'n' << 16) + ((long)' ' << 24));
const long D3D_TEXT     = ((long)'t' + ((long)'x' << 8) + ((long)'t' << 16) + ((long)' ' << 24));
const long D3D_FLOAT_32 = ((long)'0' + ((long)'0' << 8) + ((long)'3' << 16) + ((long)'2' << 24));
const long D3D_FLOAT_64 = ((long)'0' + ((long)'0' << 8) + ((long)'6' << 16) + ((long)'4' << 24));

const long D3D_BIN_ZIP  = ((long)'b' + ((long)'z' << 8) + ((long)'i' << 16) + ((long)'p' << 24));
const long D3D_TEXT_ZIP = ((long)'t' + ((long)'z' << 8) + ((long)'i' << 16) + ((long)'p' << 24));


ModelLoaderD3D::ModelLoaderD3D()
{
    init();
}
ModelLoaderD3D::~ModelLoaderD3D()
{
    clear();
}

Entity* ModelLoaderD3D::loadModel(io::stringc &Filename, io::stringc &TexturePath)
{
    Filename_       = Filename;
    TexturePath_    = TexturePath;
    
    hEntity_        = new Entity();
    
    if ( !( hFile_ = FileLoader_->openFile(Filename, io::FILE_READ_BINARY) ) )
        return hEntity_;
    
    if (!loadModelData())
    {
        io::printError("loading D3D model failed");
        FileLoader_->closeFile(hFile_);
        return hEntity_;
    }
    
    FileLoader_->closeFile(hFile_);
    
    return buildModel();
}


/*
 * ========== Private: ==========
 */


void ModelLoaderD3D::init()
{
    /* General settings */
    isBinary_           = false;
    FloatSize_          = 0;
    FileLength_         = 0;
    TemplateLevel_      = 0;
    
    /* Building information */
    CurMaterialIndex_   = -1;
    CurTransformation_.matrixReset();
    CurFrameIndex_      = 0;
    CurBoneIndex_       = 0;
    HighestFrame_       = 0;
    
    /* Allocate all memories */
    FileLoader_         = new io::FileLoader();
}

void ModelLoaderD3D::clear()
{
    /* Delete all memories */
    
    // Delete fileloader
    spDeletePointer(FileLoader_);
    
    // Delete template value list
    spDeleteArray(ValueList_);
    
    // Delete token lists
    spDeleteArray(IntegerList_);
    spDeleteArray(FloatList_);
    spDeleteArray(DoubleList_);
}


/*
 * ======= Loading the X-file (D3D) =======
 */

bool ModelLoaderD3D::loadModelData()
{
    /* Read the header data */
    if (!loadModelDataHeader())
        return false;
    
    /* Read the body data */
    if (isBinary_)
        return loadModelDataBodyBinary();
    else
        return loadModelDataBodyText();
    
    return false;
}


/*
 * ======= Header loading =======
 */

bool ModelLoaderD3D::loadModelDataHeader()
{
    long lgHeader = 0;
    
    /* Check the file length */
    FileLength_ = hFile_->getLength();
    if (FileLength_ < 12)
    {
        io::printError("D3D model is too small");
        return false;
    }
    
    /* Check the magic number "xof " */
    lgHeader = FileLoader_->readValue<long>(hFile_);
    if (lgHeader != D3D_MAGIC)
    {
        io::printError("D3D model has invalid header");
        return false;
    }
    
    /* Check the version number */
    lgHeader = FileLoader_->readValue<long>(hFile_);
    if (lgHeader != D3D_VERSION)
    {
        io::printError("D3D model has invalid version");
        return false;
    }
    
    /* Check the format (binary/ text) */
    lgHeader = FileLoader_->readValue<long>(hFile_);
    if (lgHeader == D3D_BINARY)
    {
        isBinary_ = true;
    }
    else if (lgHeader == D3D_TEXT)
    {
        isBinary_ = false;
    }
    else
    {
        io::printError("D3D model has invalid format");
        return false;
    }
    
    /* Check the float type */
    lgHeader = FileLoader_->readValue<long>(hFile_);
    if (lgHeader == D3D_FLOAT_32)
    {
        FloatSize_ = 4;
    }
    else if (lgHeader == D3D_FLOAT_64)
    {
        FloatSize_ = 8;
    }
    else
    {
        io::printError("D3D model has invalid float size");
        return false;
    }
    
    return true;
}


/*
 * ======= Loading the body (binary) =======
 */

//#define DEBUG_D3D

#define __SP_D3D_TEMPLATETYPENAME_COUNT 19

io::stringc TemplateTypeNameList[__SP_D3D_TEMPLATETYPENAME_COUNT] = 
{
    "HEADER",
    "COORDS2D",
    "COLORRGBA",
    "COLORRGB",
    "MATERIAL",
    "TEXTUREFILENAME",
    "MESHFACE",
    "MESHFACEWRAPS",
    "MESHTEXTURECOORDS",
    "MESHNORMALS",
    "MESHVERTEXCOLORS",
    "MESHMATERIALLIST",
    "MESH",
    "FRAMETRANSFORMMATRIX",
    "FRAME",
    "ANIMATIONKEY",
    "ANIMATIONOPTIONS",
    "ANIMATION",
    "ANIMATIONSET",
};


#ifdef DEBUG_D3D
std::ofstream hDebugFile("DEBUG_FILE_D3D.txt");
#endif

bool ModelLoaderD3D::loadModelDataBodyText()
{
    
    /* Temporary variables */
    io::stringc TypeName, TemplateName, uTypeName;
    
    /* Read until the file is at the end */
    while (!FileLoader_->isEOF(hFile_))
    {
        
        /* Read the current line */
        CurLine_ = FileLoader_->readString(hFile_);
        TrimLine_ = CurLine_.ltrim();
        
        /* Check for a comment */
        if (TrimLine_.size() >= 2 && TrimLine_.left(1) != "#" && TrimLine_.left(2) != "//")
        {
            
            /* Read the next template */
            if (readTemplateTxt(TypeName, TemplateName))
            {
                
                uTypeName = TypeName.upper();
                
                #ifdef DEBUG_D3D
                io::stringc space;
                for (s32 i = 1; i < TemplateLevel_; ++i)
                    space += "\t";
                hDebugFile << space.c_str() << uTypeName.c_str() << ", " << TemplateName.c_str() << std::endl;
                #endif
                
                /* Select the template type name */
                if (uTypeName == "HEADER")
                    readTemplateTxtHeader();
                else if (uTypeName == "COORDS2D")
                    readTemplateTxtCoords2D();
                else if (uTypeName == "COLORRGBA")
                    readTemplateTxtColorRGBA();
                else if (uTypeName == "COLORRGB")
                    readTemplateTxtColorRGB();
                else if (uTypeName == "MATERIAL")
                    readTemplateTxtMaterial();
                else if (uTypeName == "TEXTUREFILENAME")
                    readTemplateTxtTextureFilename();
                else if (uTypeName == "MESHFACE")
                    readTemplateTxtMeshFace();
                else if (uTypeName == "MESHFACEWRAPS")
                    readTemplateTxtMeshFaceWraps();
                else if (uTypeName == "MESHTEXTURECOORDS")
                    readTemplateTxtMeshTextureCoords();
                else if (uTypeName == "MESHNORMALS")
                    readTemplateTxtMeshNormals();
                else if (uTypeName == "MESHVERTEXCOLORS")
                    readTemplateTxtMeshVertexColors();
                else if (uTypeName == "MESHMATERIALLIST")
                    readTemplateTxtMeshMaterialList();
                else if (uTypeName == "MESH")
                    readTemplateTxtMesh();
                else if (uTypeName == "FRAMETRANSFORMMATRIX")
                    readTemplateTxtFrameTransformMatrix();
                else if (uTypeName == "FRAME")
                    readTemplateTxtFrame();
                else if (uTypeName == "ANIMATIONKEY")
                    readTemplateTxtAnimationKey();
                else if (uTypeName == "ANIMATIONOPTIONS")
                    readTemplateTxtAnimationOptions();
                else if (uTypeName == "ANIMATION")
                    readTemplateTxtAnimation();
                else if (uTypeName == "ANIMATIONSET")
                    readTemplateTxtAnimationSet();
                
            } // fi
            
            if ((s32)CurLine_.find("}", 0) != -1 && (s32)CurLine_.find("{", 0) == -1)
            {
                --TemplateLevel_;
            }
            
        } // fi
        
    } // wend
    
    /* Exit the function */
    return true;
    
}


/*
 * ======= Loading the body (binary) =======
 */

bool ModelLoaderD3D::loadModelDataBodyBinary()
{
    
    /* Temporary variables */
    io::stringc TypeName, uTypeName;
    s16 Token;
    
    /* Read until the file is at the end */
    while (!FileLoader_->isEOF(hFile_))
    {
        
        /* Read the next template */
        if (readTemplateBin(TypeName))
        {
            
            uTypeName = TypeName.upper();
            
            /* Read the next token */
            Token = FileLoader_->readValue<s16>(hFile_);
            
            #ifdef DEBUG_D3D
            hDebugFile << "Token ID: " << Token << ", Name = " << TypeName.str() << std::endl;
            #endif
            
            /* Select the template type name */
            if (uTypeName == "HEADER")
                readTemplateBinHeader();
            else if (uTypeName == "COORDS2D")
                readTemplateBinCoords2D();
            else if (uTypeName == "COLORRGBA")
                readTemplateBinColorRGBA();
            else if (uTypeName == "COLORRGB")
                readTemplateBinColorRGB();
            else if (uTypeName == "MATERIAL")
                readTemplateBinMaterial();
            else if (uTypeName == "TEXTUREFILENAME")
                readTemplateBinTextureFilename();
            else if (uTypeName == "MESHFACE")
                readTemplateBinMeshFace();
            else if (uTypeName == "MESHFACEWRAPS")
                readTemplateBinMeshFaceWraps();
            else if (uTypeName == "MESHTEXTURECOORDS")
                readTemplateBinMeshTextureCoords();
            else if (uTypeName == "MESHNORMALS")
                readTemplateBinMeshNormals();
            else if (uTypeName == "MESHVERTEXCOLORS")
                readTemplateBinMeshVertexColors();
            else if (uTypeName == "MESHMATERIALLIST")
                readTemplateBinMeshMaterialList();
            else if (uTypeName == "MESH")
                readTemplateBinMesh();
            else if (uTypeName == "FRAMETRANSFORMMATRIX")
                readTemplateBinFrameTransformMatrix();
            else if (uTypeName == "FRAME")
                readTemplateBinFrame();
            else if (uTypeName == "ANIMATIONKEY")
                readTemplateBinAnimationKey();
            else if (uTypeName == "ANIMATIONOPTIONS")
                readTemplateBinAnimationOptions();
            else if (uTypeName == "ANIMATION")
                readTemplateBinAnimation();
            else if (uTypeName == "ANIMATIONSET")
                readTemplateBinAnimationSet();
            
        } // fi
        
    } // wend
    
    return true;
    
}


/*
 * ======= Reading values =======
 */

// +++ Internal function +++

template <typename T> inline void ModelLoaderD3D_readValue(T &Value, io::stringc &Line, s32 &LinePos)
{
    s32 tmp1, tmp2, tmp3, tmp4;
    s32 start = LinePos, end = LinePos;
    
    // Find the next pos
    tmp1 = (s32)Line.find(";", start + 1);
    tmp2 = (s32)Line.find(",", start + 1);
    tmp3 = (s32)Line.find(";,", start + 1);
    tmp4 = (s32)Line.find(";;", start + 1);
    
    if (tmp1 != -1)
        end = tmp1;
    if ( tmp2 != -1 && ( tmp1 == -1 || tmp2 < tmp1 ) )
        end = tmp2;
    if (tmp3 != -1 && tmp3 == tmp1 && tmp2 == -1)
        end = tmp3;
    if (tmp4 != -1 && tmp4 == tmp1 && tmp2 == -1)
        end = tmp4;
    
    // Get the value
    Value = io::stringc(Line.section(start, end).trim()).val<T>();
    
    if ( end >= Line.size() - 4 && ( (s32)Line.find(";;;", LinePos) || (s32)Line.find(";;,", LinePos) ) )
        ++end;
    
    // Add the linepos count
    LinePos = end + 1;
}


/*
 * ======= Blocks =======
 */

io::stringc ModelLoaderD3D::getNextValueLine()
{
    io::stringc Line;
    
    LinePos_ = 0;
    
    /* Read the next line */
    do
    {
        Line = FileLoader_->readString(hFile_);
    }
    while (Line.trim().size() < 2 && !FileLoader_->isEOF(hFile_));
    
    return Line;
}

void ModelLoaderD3D::readInt(s32 &Value, io::stringc Line)
{
    ModelLoaderD3D_readValue<s32>(Value, Line, LinePos_);
}

void ModelLoaderD3D::readFloat(f32 &Value, io::stringc Line)
{
    ModelLoaderD3D_readValue<f32>(Value, Line, LinePos_);
}

void ModelLoaderD3D::readBlock(f32* &ValueList, s32 Multiplicator, s32 Count)
{
    /* Temporary variables */
    io::stringc Line;
    s32 i = 0;
    f32 Value;
    
    Line = getNextValueLine();
    
    /* Get the count of elements of the block */
    if (Count == 0)
    {
        readInt(Count, Line);
        Count *= Multiplicator;
    }
    
    /* Allocate 'count' times elements for the list */
    ValueList = new f32[Count];
    ValueListSize_ = Count;
    
    /* Read and add the values to the list */
    while (true)
    {
        while (LinePos_ < Line.size() - 2)
        {
            readFloat(Value, Line);
            ValueList[i++] = Value;
        }
        
        if (i < Count)
        {
            Line = FileLoader_->readString(hFile_);
            LinePos_ = 0;
        }
        else
            break;
    }
    
    #ifdef DEBUG_D3D
    hDebugFile << "END-OF-BLOCK" << std::endl;
    #endif
}


void ModelLoaderD3D::readBlockKeyframe(f32* &ValueList, s32 &Count)
{
    /* Temporary variables */
    io::stringc Line;
    s32 i = 1, Number;
    f32 Value;
    
    Line = getNextValueLine();
    
    /* Get the number of elements of the block */
    readInt(Number, Line);
    
    /* Get the count of elements of the block */
    readInt(Count, Line);
    ++Count;
    
    /* Allocate 'count' times elements for the list */
    ValueList = new f32[Count];
    ValueListSize_ = Count;
    
    ValueList[0] = Number;
    
    /* Read and add the values to the list */
    while (1)
    {
        while (LinePos_ < Line.size() - 2)
        {
            readFloat(Value, Line);
            ValueList[i++] = Value;
        }
        
        if (i < Count)
        {
            Line = FileLoader_->readString(hFile_);
            LinePos_ = 0;
        }
        else
            break;
    }
    
    #ifdef DEBUG_D3D
    hDebugFile << "END-OF-BLOCK" << std::endl;
    #endif
}


/*
 * ======= Reading the templates - text =======
 */

bool ModelLoaderD3D::readTemplateTxt(io::stringc &TypeName, io::stringc &TemplateName)
{
    
    /* Temporary varibales */
    bool isTemplateFound = false;
    s32 start, end;
    
    /* Check if the line containes no numbers */
    if ((s32)CurLine_.find(",", 0) != -1 || (s32)CurLine_.find(";", 0) != -1)
        return false;
    
    if ( ( start = TrimLine_.find(" ", 0) ) == -1 )
        return false;
    
    if ( ( end = TrimLine_.find("{", start) ) == -1)
        end = TrimLine_.size()-1;
    
    /* Get a template type */
    TypeName = TrimLine_.left(start);
    
    /* Get the template name */
    TemplateName = TrimLine_.section(start, end).trim();
    
    io::stringc uTypeName(TypeName.upper());
    
    /* Check if an template type has been found */
    for (s32 i = 0; i < __SP_D3D_TEMPLATETYPENAME_COUNT; ++i)
    {
        if (TemplateTypeNameList[i] == uTypeName)
        {
            CurTemplate_.TypeName = uTypeName;
            CurTemplate_.TemplateName = TemplateName;
            TemplateList_.push_back(CurTemplate_);
            ++TemplateLevel_;
            isTemplateFound = true;
            break;
        }
    }
    
    if (!isTemplateFound)
        return false;
    
    return true;
    
}

void ModelLoaderD3D::readTemplateTxtHeader()
{
    
}

void ModelLoaderD3D::readTemplateTxtCoords2D()
{
    
}

void ModelLoaderD3D::readTemplateTxtColorRGBA()
{
    
}

void ModelLoaderD3D::readTemplateTxtColorRGB()
{
    
}

void ModelLoaderD3D::readTemplateTxtMaterial()
{
    
    /* Temporary vairbales */
    video::color FaceColor, SpecularColor, EmissiveColor;
    f32 Power;
    
    /* Save the material name */
    materialInfo MaterialData;
    {
        MaterialData.MaterialName       = CurTemplate_.TemplateName;
        MaterialData.TextureFilename    = "";
    }
    
    /* === Read the values === */
    
    /* Read color: face */
    readBlock(ValueList_, 1, 4);
    MaterialData.FaceColor.set(u8(ValueList_[0]*255), u8(ValueList_[1]*255), u8(ValueList_[2]*255), u8(ValueList_[3]*255));
    spDeleteArray(ValueList_);
    
    /* Read power */
    readBlock(ValueList_, 1, 1);
    MaterialData.Power = ValueList_[0];
    spDeleteArray(ValueList_);
    
    /* Read color: specular */
    readBlock(ValueList_, 1, 3);
    MaterialData.SpecularColor.set(u8(ValueList_[0]*255), u8(ValueList_[1]*255), u8(ValueList_[2]*255), u8(ValueList_[3]*255));
    spDeleteArray(ValueList_);
    
    /* Read color: emissive */
    readBlock(ValueList_, 1, 3);
    MaterialData.EmissiveColor.set(u8(ValueList_[0]*255), u8(ValueList_[1]*255), u8(ValueList_[2]*255), u8(ValueList_[3]*255));
    spDeleteArray(ValueList_);
    
    /* Add the material to the list */
    MaterialList_.push_back(MaterialData);
    CurMaterialIndex_ = MaterialList_.size() - 1;
    
}

void ModelLoaderD3D::readTemplateTxtTextureFilename()
{
    
    /* === Temporary varibales === */
    
    io::stringc Line;
    io::stringc Filename;
    s32 start, end;
    
    /* === Reading === */
    
    Line = FileLoader_->readString(hFile_);
    
    start = (s32)Line.find("\"", 0);
    if (start == -1)
        return;
    end = (s32)Line.find("\"", start + 1);
    if (end == -1)
        return;
    
    Filename = Line.section(start + 1, end);
    
    /* === Loading & adding === */
    
    if (hEntity_->getMeshSurfaceCount() > 0)
    {
        /* Add texture to the model */
        hEntity_->addTexture( __spVideoDriver->loadTexture(TexturePath_ + Filename) );
    }
    else
    {
        if (CurMaterialIndex_ != -1)
            MaterialList_[CurMaterialIndex_].TextureFilename = TexturePath_ + Filename;
    }
    
}

void ModelLoaderD3D::readTemplateTxtMeshFace()
{
    
}

void ModelLoaderD3D::readTemplateTxtMeshFaceWraps()
{
    
}

void ModelLoaderD3D::readTemplateTxtMeshTextureCoords()
{
    
    /* === Texture coordinates (u, v) === */
    
    readBlock(ValueList_, 2);
    
    // Set the vertex texture coords
    for (s32 i = 0; i < ValueListSize_/2; ++i)
    {
        hEntity_->setVertexTexCoord(i, dim::point2df( ValueList_[i*2+0], ValueList_[i*2+1] ));
    }
    
    spDeleteArray(ValueList_);
    
}

void ModelLoaderD3D::readTemplateTxtMeshNormals()
{
    
}

void ModelLoaderD3D::readTemplateTxtMeshVertexColors()
{
    
}

void ModelLoaderD3D::readTemplateTxtMeshMaterialList()
{
    
    /* === Temporary variables === */
    
    io::stringc MaterialName;
    s32 start, end, pos, i, j, c;
    
    /* === Read until the file is at the end === */
    
    while (!FileLoader_->isEOF(hFile_))
    {
        /* Read the current line */
        CurLine_ = FileLoader_->readString(hFile_);
        TrimLine_ = CurLine_.ltrim();
        
        /* Examine the current line */
        pos = (s32)CurLine_.find("{", 0);
        
        /* === Check if an material has been detected === */
        
        end = -1;
        
        if (pos != -1)
        {
            /* Get the material name */
            start = pos+1;
            end = (s32)CurLine_.find("}", start);
            if (end == -1)
                end = CurLine_.size()-1;
            
            MaterialName = CurLine_.section(start, end).trim();
            
            /* Loop the material list */
            for (i = 0; i < MaterialList_.size(); ++i)
            {
                if (MaterialList_[i].MaterialName == MaterialName)
                {
                    /* Check if the material has a texture */
                    if (MaterialList_[i].TextureFilename != "" && hEntity_->getMeshSurfaceCount() > 0)
                    {
                        /* Add texture to the model */
                        hEntity_->addTexture( __spVideoDriver->loadTexture(MaterialList_[i].TextureFilename) );
                    }
                    
                    /* Set the material's color */
                    for (j = 0, c = hEntity_->getMeshVerticesCount( hEntity_->getMeshSurfaceCount() - 1 ); j < c; ++j)
                    {
                        hEntity_->setVertexColor(j, MaterialList_[i].FaceColor);
                    }
                    
                    /* Break the loop  */
                    break;
                }
            } // next material
        } // fi
        
        /* === Check if the material block has been end === */
        
        if (end != -1)
            pos = (s32)CurLine_.find("}", end+1);
        else
            pos = (s32)CurLine_.find("}", 0);
        
        if (pos != -1)
            break;
        
    }
    
}

void ModelLoaderD3D::readTemplateTxtMesh()
{
    
    /* === Surface === */
    
    hEntity_->createSurface();
    
    /* === Vertices === */
    
    readBlock(ValueList_, 3);
    
    // Create the vertices
    for (s32 i = 0; i < ValueListSize_/3; ++i)
    {
        hEntity_->addVertex( CurTransformation_ * dim::vector3df( ValueList_[i*3+0], ValueList_[i*3+1], ValueList_[i*3+2] ) );
    }
    
    spDeleteArray(ValueList_);
    
    /* === Triangles === */
    
    readBlock(ValueList_, 4);
    
    // Create the triangles
    for (s32 i = 0; i < ValueListSize_/4; ++i)
    {
        hEntity_->addTriangle( (u32)ValueList_[i*4+1], (u32)ValueList_[i*4+2], (u32)ValueList_[i*4+3] );
    }
    
    spDeleteArray(ValueList_);
    
    /* === Back settings === */
    
    CurTransformation_.matrixReset();
    
    /* === Frame settings === */
    
    if (CurFrameIndex_ > 0)
    {
        // Add a frame
        --CurFrameIndex_;
        
        frameInfo FrameData;
        {
            FrameData.Name      = TempFrameNameList_[CurFrameIndex_];
            FrameData.Surface   = hEntity_->getMeshSurfaceCount() - 1;
        }
        FrameList_.push_back(FrameData);
        
        // Link the surface to the framebone
        for (s32 i = 0; i < FrameBoneList_.size(); ++i)
        {
            if (FrameBoneList_[i].FrameName == FrameData.Name)
            {
                FrameBoneList_[i].Surface = FrameData.Surface;
                break;
            }
        }
        
        // Delete the temporary frame name
        std::vector<io::stringc>::iterator it = TempFrameNameList_.begin() + CurFrameIndex_;
        TempFrameNameList_.erase(it);
    }
    
}

void ModelLoaderD3D::readTemplateTxtFrameTransformMatrix()
{
    
    /* === Vertices === */
    
    readBlock(ValueList_, 1, 16);
    
    // Transform to the read matrix
    for (s32 i = 0; i < 16; ++i)
        CurTransformation_[i] = ValueList_[i];
    
    spDeleteArray(ValueList_);
    
}

void ModelLoaderD3D::readTemplateTxtFrame()
{
    
    /* === Temporary frame name information === */
    
    TempFrameNameList_.push_back(CurTemplate_.TemplateName);
    
    ++CurFrameIndex_;
    
    /* === Frame bone information === */
    
    frameBoneInfo FrameBoneData;
    {
        FrameBoneData.FrameName     = CurTemplate_.TemplateName;
        FrameBoneData.Surface       = -1; // (at first: no surface)
        FrameBoneData.TemplateLevel = TemplateLevel_;
    }
    FrameBoneList_.push_back(FrameBoneData);
    
}

void ModelLoaderD3D::readTemplateTxtAnimationKey()
{
    
    /* === Temporary variables === */
    
    s32 AnimType, KeyframeCount, Count;
    dim::vector3df Vector;
    
    SFrameBone::SFrameSeq FrameSeqData;
    
    /* === Animation key information === */
    
    /* Read the animation key type */
    readInt(AnimType, getNextValueLine());
    
    /* Read the animation keyframe count */
    readInt(KeyframeCount, getNextValueLine());
    
    #ifdef DEBUG_D3D
    hDebugFile << "AnimationKey: AnimType = " << AnimType << ", KeyframeCount = " << KeyframeCount << std::endl;
    #endif
    
    /* Read the whole block */
    for (s32 i = 0, j; i < KeyframeCount; ++i)
    {
        readBlockKeyframe(ValueList_, Count);
        
        // Get the highest frame
        if ((s32)ValueList_[0] > HighestFrame_)
            HighestFrame_ = (s32)ValueList_[0];
        
        // Add the keyframe information
        switch (AnimType)
        {
            case D3D_ANIMKEY_ROT:
            {
                FrameSeqData.FrameNumber = (s32)ValueList_[0];
                FrameSeqData.Quat.set( ValueList_[1], ValueList_[2], ValueList_[3], ValueList_[4] );
                FrameBoneList_[CurBoneIndex_].KeyframeRotationList.push_back(FrameSeqData);
            }
            break;
            
            case D3D_ANIMKEY_SCL:
            {
                FrameSeqData.FrameNumber = (s32)ValueList_[0];
                FrameSeqData.Vector.make( ValueList_[1], ValueList_[2], ValueList_[3] );
                FrameBoneList_[CurBoneIndex_].KeyframeScaleList.push_back(FrameSeqData);
            }
            break;
            
            case D3D_ANIMKEY_POS:
            {
                FrameSeqData.FrameNumber = (s32)ValueList_[0];
                FrameSeqData.Vector.make( ValueList_[1], ValueList_[2], ValueList_[3] );
                FrameBoneList_[CurBoneIndex_].KeyframePositionList.push_back(FrameSeqData);
            }
            break;
        }
        
        spDeleteArray(ValueList_);
    }
    
}

void ModelLoaderD3D::readTemplateTxtAnimationOptions()
{
    
}

void ModelLoaderD3D::readTemplateTxtAnimation()
{
    
    /* === Temporary variables === */
    
    io::stringc Line, AnimName, BoneName;
    
    /* === Animation properties === */
    
    // Animation name
    AnimName = CurTemplate_.TemplateName;
    
    // Bone name
    Line = FileLoader_->readString(hFile_);
    
    BoneName = Line.section(Line.find("{", 0) + 1, Line.find("}", 0));
    
    // Find the bone index which fits
    for (u32 i = 0; i < FrameBoneList_.size(); ++i)
    {
        if (FrameBoneList_[i].FrameName == BoneName)
        {
            CurBoneIndex_ = i;
            break;
        }
    }
    
    #ifdef DEBUG_D3D
    hDebugFile << "Animation: Name = " << AnimName.str() << ", BoneName = " << BoneName.str() << std::endl;
    #endif
    
}

void ModelLoaderD3D::readTemplateTxtAnimationSet()
{
    
}


/*
 * ======= Reading the templates - binary =======
 */

bool ModelLoaderD3D::readTemplateBin(io::stringc &TypeName)
{
    
    /* Temporary variables */
    s16 Token, TokenOBRACE;
    io::stringc UUID;
    
    /* Read token: name */
    Token = FileLoader_->readValue<s16>(hFile_);
    
    if (Token != D3D_TOKEN_NAME)
        return false;
    
    /* Read the name */
    readTokenName(TypeName);
    
    /* Read the OBRACE token */
    TokenOBRACE = FileLoader_->readValue<s16>(hFile_);
    
    return TokenOBRACE == D3D_TOKEN_OBRACE;
    
}

void ModelLoaderD3D::readTemplateBinHeader()
{
    
    /* Temporary variables */
    s16 Major, Minor;
    s32 Flags;
    
    /* Read the data */
    Major = FileLoader_->readValue<s16>(hFile_);
    Minor = FileLoader_->readValue<s16>(hFile_);
    Flags = FileLoader_->readValue<s32>(hFile_);
    
    #ifdef DEBUG_D3D
    hDebugFile << "Template HEADER: Major = " << Major << ", Minor = " << Minor << ", Flags = " << Flags << std::endl;
    #endif
    
}

void ModelLoaderD3D::readTemplateBinCoords2D()
{
    
}

void ModelLoaderD3D::readTemplateBinColorRGBA()
{
    
}

void ModelLoaderD3D::readTemplateBinColorRGB()
{
    
}

void ModelLoaderD3D::readTemplateBinMaterial()
{
    
}

void ModelLoaderD3D::readTemplateBinTextureFilename()
{
    
    /* Temporary variables */
    io::stringc Filename;
    
    /* Read the data */
    readTokenString(Filename);
    
    #ifdef DEBUG_D3D
    hDebugFile << "Template TEXTUREFILENAME: " << Filename.str() << std::endl;
    #endif
    
}

void ModelLoaderD3D::readTemplateBinMeshFace()
{
    
}

void ModelLoaderD3D::readTemplateBinMeshFaceWraps()
{
    
}

void ModelLoaderD3D::readTemplateBinMeshTextureCoords()
{
    
}

void ModelLoaderD3D::readTemplateBinMeshNormals()
{
    
}

void ModelLoaderD3D::readTemplateBinMeshVertexColors()
{
    
}

void ModelLoaderD3D::readTemplateBinMeshMaterialList()
{
    
}

void ModelLoaderD3D::readTemplateBinMesh()
{
    
}

void ModelLoaderD3D::readTemplateBinFrameTransformMatrix()
{
    
}

void ModelLoaderD3D::readTemplateBinFrame()
{
    
}

void ModelLoaderD3D::readTemplateBinAnimationKey()
{
    
}

void ModelLoaderD3D::readTemplateBinAnimationOptions()
{
    
}

void ModelLoaderD3D::readTemplateBinAnimation()
{
    
}

void ModelLoaderD3D::readTemplateBinAnimationSet()
{
    
}


/*
 * ======= Tokens =======
 */

void ModelLoaderD3D::readTokenName(io::stringc &Name)
{
    
    // Length of string
    s32 Count = FileLoader_->readValue<s32>(hFile_);
    
    // Array
    c8* strArray = new c8[Count];
    
    // Read string
    FileLoader_->readStream(hFile_, strArray, Count);
    
    strArray[Count] = 0;
    Name = strArray;
    
    // Delete array
    delete [] strArray;
    
}

void ModelLoaderD3D::readTokenString(io::stringc &String)
{

    // Length of string
    s32 Count = FileLoader_->readValue<s32>(hFile_);
    
    // Array
    c8* strArray = new c8[Count];
    
    // Read string
    FileLoader_->readStream(hFile_, strArray, Count);
    
    strArray[Count] = 0;
    String = strArray;
    
    // Delete array
    delete [] strArray;
    
    // Terminator ( D3D_TOKEN_SEMICOLON / D3D_TOKEN_COMMA )
    s16 Terminator = FileLoader_->readValue<s16>(hFile_);
    
}

void ModelLoaderD3D::readTokenInteger(s32 &Value)
{
    
    // Single interger
    Value = FileLoader_->readValue<s32>(hFile_);
    
}

void ModelLoaderD3D::readTokenGuid(io::stringc &UUID)
{
    
    // UUID data fields
    s32 data1 = FileLoader_->readValue<s32>(hFile_);
    s16 data2 = FileLoader_->readValue<s16>(hFile_);
    s16 data3 = FileLoader_->readValue<s16>(hFile_);
    
    // Array
    s8 data4[8];
    
    FileLoader_->readStream(hFile_, data4, 8);
    
    // Get the UUID
    UUID = io::getHex(data1) + io::stringc("-") + io::getHex(data2) + io::stringc("-") + io::getHex(data3) +
           io::getHex(data4[0]) + io::getHex(data4[1]) + io::stringc("-") + io::getHex(data4[2]) +
           io::getHex(data4[3]) + io::getHex(data4[4]) + io::getHex(data4[5]) + io::getHex(data4[6]) +io::getHex(data4[8]);
    
}

void ModelLoaderD3D::readTokenIntegerList(s32* &ValueList)
{

    // Length of string
    s32 Count = FileLoader_->readValue<s32>(hFile_);
    
    // Allocate memory
    ValueList = new s32[Count];
    
    // Array
    FileLoader_->readStream(hFile_, ValueList, Count, sizeof(s32));
    
    #ifdef DEBUG_D3D
    hDebugFile << "Token INTEGERLIST: [ " << Count << " ] = { ";
    for (s32 i = 0; i < Count-1; ++i)
        hDebugFile << ValueList[i] << ", ";
    hDebugFile << ValueList[Count-1] << " }" << std::endl;
    delete [] ValueList; // !!!
    #endif
    
}

void ModelLoaderD3D::readTokenFloat32List(f32* &ValueList)
{

    // Length of string
    s32 Count = FileLoader_->readValue<s32>(hFile_);
    
    // Allocate memory
    ValueList = new f32[Count];
    
    // Array
    FileLoader_->readStream(hFile_, ValueList, Count, sizeof(f32));
    
    #ifdef DEBUG_D3D
    hDebugFile << "Token FLOAT32LIST: [ " << Count << " ] = { ";
    for (s32 i = 0; i < Count-1; ++i)
        hDebugFile << ValueList[i] << ", ";
    hDebugFile << ValueList[Count-1] << " }" << std::endl;
    delete [] ValueList; // !!!
    #endif
    
}

void ModelLoaderD3D::readTokenFloat64List(f64* &ValueList)
{

    // Length of string
    s32 Count = FileLoader_->readValue<s32>(hFile_);
    
    // Allocate memory
    ValueList = new f64[Count];
    
    // Array
    FileLoader_->readStream(hFile_, ValueList, Count, sizeof(f64));
    
}

void ModelLoaderD3D::readTokenFloatList(f32* &FloatList, f64* &DoubleList)
{
    if (FloatSize_ == 4)
        readTokenFloat32List(FloatList);
    else if (FloatSize_ == 8)
        readTokenFloat64List(DoubleList);
}


/*
 * ======= Build the model finally =======
 */

Entity* ModelLoaderD3D::buildModel()
{
    
    /* building the model */
    
    hEntity_->updateModeling();
    
    if (__spAnimator)
        buildAnimation();
    
    return hEntity_;
}

void ModelLoaderD3D::buildAnimation()
{
    
    #ifdef DEBUG_D3D
    for (s32 i = 0; i < FrameBoneList_.size(); ++i)
    {
        hDebugFile << "FRAMEBONE: FrameName = " << FrameBoneList_[i].FrameName.str()
                   << ",\tSurface = " << FrameBoneList_[i].Surface
                   << ",\tTemplateLevel = " << FrameBoneList_[i].TemplateLevel << std::endl;
    }
    #endif
    
    #if 0
    
    /* === Check if the object has a surface and a frame === */
    
    if (hEntity_->getMeshSurfaceCount() == 0 || FrameBoneList_.size() == 0)
        return;
    
    /* === Create a new animation === */
    
    Anim_ = __spAnimator->addAnimation(hEntity_);
    
    /* === Temporary variables === */
    
    std::vector<u32> VerticesIndices, VerticesSurfaces;
    dim::vector3df Translation, Scale;
    dim::quaternion Quat;
    
    u32 Surface, IndicesCount, ParentBoneID;
    s32 i, j, k;
    
    /* === Make an identified frame list === */
    
    //HighestFrame_
    u32 LastFrame = 0;
    
    std::vector<SFrameBone::SFrameSeq> NewPositionList;
    std::vector<SFrameBone::SFrameSeq> NewRotationList;
    std::vector<SFrameBone::SFrameSeq> NewScaleList;
    
    for (i = 0; i < FrameBoneList_.size(); ++i)
    {
        
        /* Identify positions */
        for (j = 0; j < FrameBoneList_[i].KeyframePositionList.size(); ++j)
        {
            for (k = LastFrame + 1; k <= FrameBoneList_[i].KeyframePositionList[j].FrameNumber; ++k)
            {
                NewPositionList.push_back(FrameBoneList_[i].KeyframePositionList[k]);
            }
            
            LastFrame = FrameBoneList_[i].KeyframePositionList[j].FrameNumber;
        }
        
        FrameBoneList_[i].KeyframePositionList = NewPositionList;
        NewPositionList.clear();
        
        /* Identify quaternions */
        for (j = 0; j < FrameBoneList_[i].KeyframeRotationList.size(); ++j)
        {
            for (k = LastFrame + 1; k <= FrameBoneList_[i].KeyframeRotationList[j].FrameNumber; ++k)
            {
                NewRotationList.push_back(FrameBoneList_[i].KeyframeRotationList[k]);
            }
            
            LastFrame = FrameBoneList_[i].KeyframeRotationList[j].FrameNumber;
        }
        
        FrameBoneList_[i].KeyframePositionList = NewPositionList;
        NewPositionList.clear();
        
        /* Identify sizes */
        for (j = 0; j < FrameBoneList_[i].KeyframePositionList.size(); ++j)
        {
            for (k = LastFrame + 1; k <= FrameBoneList_[i].KeyframePositionList[j].FrameNumber; ++k)
            {
                NewPositionList.push_back(FrameBoneList_[i].KeyframePositionList[k]);
            }
            
            LastFrame = FrameBoneList_[i].KeyframePositionList[j].FrameNumber;
        }
        
        FrameBoneList_[i].KeyframePositionList = NewPositionList;
        NewPositionList.clear();
        
    }
    
    /* === Loop for the frame bone list === */
    
    for (i = 0; i < FrameBoneList_.size(); ++i)
    {
        
        // First settings
        Surface         = (FrameBoneList_[i].Surface != -1 ? FrameBoneList_[i].Surface : 0);
        Translation     = (FrameBoneList_[i].KeyframePositionList.size() > 0 ? FrameBoneList_[i].KeyframePositionList[0].Vector : 0.0f);
        Quat            = (FrameBoneList_[i].KeyframePositionList.size() > 0 ? FrameBoneList_[i].KeyframePositionList[0].Quat : dim::quaternion());
        Scale           = (FrameBoneList_[i].KeyframeScaleList.size() > 0 ? FrameBoneList_[i].KeyframeScaleList[0].Vector : 1.0f);
        ParentBoneID    = 0;
        
        // Get the parent bone ID
        if (i > 0)
        {
            for (j = i - 1; j >= 0; --j)
            {
                if (FrameBoneList_[i].TemplateLevel == FrameBoneList_[j].TemplateLevel - 1)
                {
                    ParentBoneID = j + 1;
                    break;
                }
            }
        }
        
        // Add a skeleton bone
        Anim_->addAnimSkeletonBone(
            ParentBoneID,               // Parent bone identity number
            Translation,                // Translation (location)
            Quat,                       // Quaternion
            Scale,                      // Bone scale
            FrameBoneList_[i].FrameName // Bone name
        );
        
        // Add the vertices indices if the bone is linkes to a surface
        if (FrameBoneList_[i].Surface != -1)
        {
            // Select the current surface
            IndicesCount = hEntity_->getMeshVerticesCount(FrameBoneList_[i].Surface);
            
            // Fill the vertices indices container
            VerticesIndices.resize(IndicesCount);
            VerticesSurfaces.resize(IndicesCount);
            
            for (j = 0; j < IndicesCount; ++j)
            {
                VerticesIndices[j] = j;
                VerticesSurfaces[j] = FrameBoneList_[i].Surface;
            }
            
            // Add the vertices indices
            Anim_->addSkeletonBoneVertices(i+1, VerticesIndices, VerticesSurfaces);
            
            VerticesIndices.clear();
        }
        
        // Create the animation keyframes
        for (j = 0; j < FrameBoneList_[i].KeyframeRotationList.size(); ++j)
        {
            
            /*for (k = FrameBoneList_[i].KeyframePositionList.size(); k < FrameBoneList_[i].KeyframeRotationList.size(); ++k)
            {
                FrameBoneList_[i].KeyframePositionList.push_back(
                    FrameBoneList_[i].KeyframePositionList[ FrameBoneList_[i].KeyframePositionList.size() ]
                );
            } // ?!?*/
            
            Anim_->addSkeletonBoneFrame(
                i+1,                                                // Bone identity number
                FrameBoneList_[i].KeyframePositionList[j].Vector,   // Translation keyframe
                FrameBoneList_[i].KeyframeRotationList[j].Quat,     // Quaternion keyframe
                1.0f,//FrameBoneList_[i].KeyframeScaleList[j].Vector,      // Scale keyframe
                1000                                                // Time (in ms.)
            );
        }
        
    }
    
    /* Update the skeleton bones */
    Anim_->updateSkeletonBones();
    
    #endif
    
}


// Other reading functions





} // /namespace scene

} // /namespace sp








// ================================================================================
