/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * Model loading MD2-fileformat (Quake II) file
 *
 */

#include "spModelLoaderMD2.hpp"


namespace sp
{
namespace scene
{


/*
 * Macros
 */

static const s32 MD2_IDENTITY       = (('2'<<24) + ('P'<<16) + ('D'<<8) + 'I'); // magic number must be "IDP2" or 844121161
static const s32 MD2_VERSION        = 8;
static const s32 MD2_MAX_VERTICES   = 1000;


/*
 * ModelLoaderMD2 class
 */

ModelLoaderMD2::ModelLoaderMD2()
{
    init();
}
ModelLoaderMD2::~ModelLoaderMD2()
{
    clear();
}

Entity* ModelLoaderMD2::loadModel(io::stringc &Filename, io::stringc &TexturePath)
{
    Filename_       = Filename;
    TexturePath_    = TexturePath;
    
    hEntity_        = new Entity();
    
    hFile_.open(Filename.c_str(), std::ios::in | std::ios::binary);
    
    if (!hFile_ || hFile_.fail())
    {
        io::printError("could not open MD2 file");
        return hEntity_;
    }
    
    if (!loadModelData(Filename))
    {
        io::printError("loading MD2 model failed");
        hFile_.close();
        return hEntity_;
    }
    
    hFile_.close();
    
    return buildModel();
}

void ModelLoaderMD2::getAnimationSeq(s32 Type, u32 &Start, u32 &End)
{
    Start   = AnimList_[Type].FirstFrame;
    End     = AnimList_[Type].LastFrame;
}


// ========== Private: ==========


void ModelLoaderMD2::init()
{
    Scale_          = 1.0f;
    
    KeyFramesCount_ = 0;
    VerticesCount_  = 0;
    CommandsCount_  = 0;
    
    pVertices_      = 0;
    pCommands_      = 0;
    pLightNormals_  = 0;
    
    /* Initialize animation memory */
    memset(&Animation_, 0, sizeof(md2AnimStateInfo));
}
void ModelLoaderMD2::clear()
{
    /* Delete all arrays */
    spDeleteArray(pVertices_);
    spDeleteArray(pCommands_);
    spDeleteArray(pLightNormals_);
}

bool ModelLoaderMD2::loadModelData(io::stringc Filename)
{
    headerMD2Info Header;
    md2KeyFrameInfo* pFrame;
    dim::vector3df* pVertices;
    s32* pNormals;
    s8* pBuffer;
    
    hFile_.read((char*)&Header, sizeof(headerMD2Info));
    
    if (Header.ident != MD2_IDENTITY)
    {
        io::printError("MD2 model has wrong identity, must be 'IDP2'");
        return false;
    }
    
    if (Header.version != MD2_VERSION)
    {
        io::printError(io::stringc("MD2 model has unsupported version (")+io::stringc(Header.version)+io::stringc(")"));
        return false;
    }
    
    // Initialize member variables
    KeyFramesCount_     = Header.num_frames;
    VerticesCount_      = Header.num_xyz;
    CommandsCount_      = Header.num_glcmds;
    
    // Allocate memory
    pVertices_          = new dim::vector3df[ VerticesCount_ * KeyFramesCount_ ];
    pCommands_          = new s32[ CommandsCount_ ];
    pLightNormals_      = new s32[ VerticesCount_ * KeyFramesCount_ ];
    pBuffer             = new s8[ KeyFramesCount_ * Header.framesize ];
    
    // Read frame data ...
    hFile_.seekg(Header.ofs_frames, std::ios::beg);
    hFile_.read((char*)pBuffer, KeyFramesCount_ * Header.framesize);
    
    // Read OpenGL commands ...
    hFile_.seekg(Header.ofs_glcmds, std::ios::beg);
    hFile_.read((char*)pCommands_, CommandsCount_ * sizeof(s32));
    
    // Vertex array initialization
    for (s32 i = 0; i < KeyFramesCount_; ++i)
    {
        // Adjust pointers
        pFrame      = (md2KeyFrameInfo*)&pBuffer[ Header.framesize * i ];
        pVertices   = &pVertices_[ VerticesCount_ * i ];
        pNormals    = &pLightNormals_[ VerticesCount_ * i ];
        
        for (s32 j = 0; j < VerticesCount_; ++j)
        {
            pVertices[j].X = (pFrame->Vertices[j].Vertex.X * pFrame->Scale.X) + pFrame->Translate.X;
            pVertices[j].Y = (pFrame->Vertices[j].Vertex.Y * pFrame->Scale.Y) + pFrame->Translate.Y;
            pVertices[j].Z = (pFrame->Vertices[j].Vertex.Z * pFrame->Scale.Z) + pFrame->Translate.Z;
            
            pNormals[j] = pFrame->Vertices[j].LightNormalIndex;
        }
    }
    
    spDeleteArray(pBuffer);
    
    return true;
}

Entity* ModelLoaderMD2::buildModel()
{
    dim::vector3df VerticesList[MD2_MAX_VERTICES];
    s32* pTriangleCmds = pCommands_;
    bool isTriangleFan, isStrip = false;
    u32 cnt1 = 0;
    
    interpolate(VerticesList);
    
    hEntity_->createSurface();
    
    while (s32 i = *(pTriangleCmds++))
    {
        
        isTriangleFan = (i < 0);
        if (i < 0)
            i = -i;
        hEntity_->addVertexCounter(cnt1);
        cnt1 = 0;
        isStrip = true;
        
        for (; i > 0; --i, pTriangleCmds += 3)
        {
            // pTriangleCmds[0] texture coordinate u
            // pTriangleCmds[1] texture coordinate v
            // pTriangleCmds[2] vertex index to render
            
            hEntity_->addVertex(
                VerticesList[ pTriangleCmds[2] ],
                dim::point2df( ((f32*)pTriangleCmds)[0], ((f32*)pTriangleCmds)[1] )
            );
            
            if (++cnt1 >= 3)
            {
                if (!isTriangleFan)
                {
                    isStrip = !isStrip;
                    if (isStrip)
                        hEntity_->addTriangle(cnt1 - 1, cnt1 - 2, cnt1 - 3);
                    else
                        hEntity_->addTriangle(cnt1 - 3, cnt1 - 2, cnt1 - 1);
                }
                else
                    hEntity_->addTriangle(cnt1 - 2, cnt1 - 1, 0);
            }
        }
        
    }
    
    hEntity_->updateModeling();
    
    if (__spAnimator)
        buildAnimation();
    
    return hEntity_;
}

void ModelLoaderMD2::interpolate(dim::vector3df* VerticesList)
{
    for (s32 i = 0; i < VerticesCount_; ++i) // z, x, y (fastest rotation)
    {
        VerticesList[i].Z = - pVertices_[ i + (VerticesCount_*Animation_.CurFrame) ].X * Scale_;
        VerticesList[i].X =   pVertices_[ i + (VerticesCount_*Animation_.CurFrame) ].Y * Scale_;
        VerticesList[i].Y =   pVertices_[ i + (VerticesCount_*Animation_.CurFrame) ].Z * Scale_;
    }
}

void ModelLoaderMD2::buildAnimation()
{
    dim::vector3df VerticesList[MD2_MAX_VERTICES];
    s32* pTriangleCmds;
    std::vector<dim::vector3df> VerticesCoordList;
    
    Anim_ = __spAnimator->addAnimation(hEntity_);
    
    for (s32 frame = 0; frame < KeyFramesCount_; ++frame)
    {
        pTriangleCmds = pCommands_;
        Animation_.CurFrame = frame;
        interpolate(VerticesList);
        
        while (s32 i = *(pTriangleCmds++))
        {
            if (i < 0)
                i = -i;
            for (; i > 0; --i, pTriangleCmds += 3)
                VerticesCoordList.push_back( VerticesList[ pTriangleCmds[2] ] );
        }
        
        Anim_->addAnimSeqMesh(VerticesCoordList);
        VerticesCoordList.clear();
    }
    
    Animation_.CurFrame = 0;
}

ModelLoaderMD2::md2AnimInfo ModelLoaderMD2::AnimList_[21] =
{
    // first, last, fps
    {   0,  39,  9 },   // STAND
    {  40,  45, 10 },   // RUN
    {  46,  53, 10 },   // ATTACK
    {  54,  57,  7 },   // PAIN_A
    {  58,  61,  7 },   // PAIN_B
    {  62,  65,  7 },   // PAIN_C
    {  66,  71,  7 },   // JUMP
    {  72,  83,  7 },   // FLIP
    {  84,  94,  7 },   // SALUTE
    {  95, 111, 10 },   // FALLBACK
    { 112, 122,  7 },   // WAVE
    { 123, 134,  6 },   // POINT
    { 135, 153, 10 },   // CROUCH_STAND
    { 154, 159,  7 },   // CROUCH_WALK
    { 160, 168, 10 },   // CROUCH_ATTACK
    { 196, 172,  7 },   // CROUCH_PAIN
    { 173, 177,  5 },   // CROUCH_DEATH
    { 178, 183,  7 },   // DEATH_FALLBACK
    { 184, 189,  7 },   // DEATH_FALLFORWARD
    { 190, 197,  7 },   // DEATH_FALLBACKSLOW
    { 198, 198,  5 },   // BOOM
};




} // /namespace scene

} // /namespace sp








// ================================================================================
