/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * Model loading MS3D-fileformat (Milkshape 3D) file
 *
 */

#include "spModelLoaderMS3D.hpp"
#include "spMainDevice.hpp"


namespace sp
{

extern video::VideoDriver* __spVideoDriver;

namespace scene
{


// Packed structures

#if defined(_MSC_VER)
#   pragma pack(push, packing)
#   pragma pack(1)
#   define __SP_PACK_STRUCT
#elif defined(__GNUC__)
#   define __SP_PACK_STRUCT __attribute__((packed))
#else
#   define __SP_PACK_STRUCT
#endif


// File header
typedef struct SMS3DHeader
{
    c8 m_ID[10];
    s32 m_version;
}
__SP_PACK_STRUCT ms3dHeaderInfo;

// Vertex information
typedef struct SMS3DVertex
{
    u8 m_flags;
    f32 m_vertex[3];
    s8 m_boneID;
    u8 m_refCount;
}
__SP_PACK_STRUCT ms3dVertexInfo;

// Triangle information
typedef struct SMS3DTriangle
{
    u16 m_flags;
    u16 m_vertexIndices[3];
    f32 m_vertexNormals[3][3];
    f32 m_s[3], m_t[3];
    u8 m_smoothingGroup;
    u8 m_groupIndex;
}
__SP_PACK_STRUCT ms3dTriangleInfo;

// Material information
typedef struct SMS3DMaterial
{
    s8 m_name[32];
    f32 m_ambient[4];
    f32 m_diffuse[4];
    f32 m_specular[4];
    f32 m_emissive[4];
    f32 m_shininess;    // 0.0f - 128.0f
    f32 m_transparency; // 0.0f - 1.0f
    u8 m_mode; // 0, 1, 2 is unused now
    c8 m_texture[128];
    s8 m_alphamap[128];
}
__SP_PACK_STRUCT ms3dMaterialInfo;

//	Joint information
typedef struct SMS3DJoint
{
    u8 m_flags;
    c8 m_name[32];
    c8 m_parentName[32];
    f32 m_rotation[3];
    f32 m_translation[3];
    u16 m_numRotationKeyframes;
    u16 m_numTranslationKeyframes;
}
__SP_PACK_STRUCT ms3dJointInfo;

// Keyframe data
typedef struct SMS3DKeyframe
{
    f32 m_time;
    f32 m_parameter[3];
}
__SP_PACK_STRUCT ms3dKeyframeInfo;

// Default alignment
#ifdef _MSC_VER
#   pragma pack(pop, packing)
#endif

#undef __SP_PACK_STRUCT


/*
 * ModelLoaderMS3D class
 */

ModelLoaderMS3D::ModelLoaderMS3D()
{
    init();
}
ModelLoaderMS3D::~ModelLoaderMS3D()
{
    clear();
}

Entity* ModelLoaderMS3D::loadModel(io::stringc &Filename, io::stringc &TexturePath)
{
    Filename_       = Filename;
    TexturePath_    = TexturePath;
    
    hEntity_        = new Entity();
    
    if (!loadModelData(Filename))
    {
        io::printError("loading MS3D model failed");
        return hEntity_;
    }
    
    return buildModel();
}


/* ========== Private: ========== */


bool ModelLoaderMS3D::loadModelData(io::stringc Filename)
{
    
    s32 i;
    
    /* Open the file */
    std::ifstream inputFile(Filename.c_str(), std::ios::in | std::ios::binary);
    
    /* Check if the file was opened successful */
    if (inputFile.fail())
        return false;
    
    inputFile.seekg(0, std::ios::end);
    long fileSize = inputFile.tellg();
    inputFile.seekg(0, std::ios::beg);
    
    /*
     * Reading buffer
     */
    
    u8* pBuffer = new u8[fileSize];
    inputFile.read((char*)pBuffer, fileSize);
    inputFile.close();
    
    const u8* pPtr = pBuffer;
    ms3dHeaderInfo* pHeader = (ms3dHeaderInfo*)pPtr;
    pPtr += sizeof(ms3dHeaderInfo);
    
    if (strncmp(pHeader->m_ID, "MS3D000000", 10) != 0)
    {
        io::printError("MS3D model has wrong identity, must be 'MS3D000000'");
        return false; // "Not a valid Milkshape3D model file."
    }
    
    if (pHeader->m_version < 3 || pHeader->m_version > 4)
    {
        io::printError(
            io::stringc("MS3D model has unsupported version (")+io::stringc(pHeader->m_version)+
            io::stringc("), only Milkshape3D 1.3 and 1.4 are supported")
        );
        return false; // "Unhandled file version. Only Milkshape3D Version 1.3 and 1.4 is supported." );
    }
    
    /*
     * Reading vertices
     */
    
    s32 nVertices       = *(u16*)pPtr;
    CountOfVertcies_    = nVertices;
    pVertices_          = new SVertexMS3D[nVertices];
    pPtr += sizeof(u16);
    
    for (i = 0; i < nVertices; ++i)
    {
        ms3dVertexInfo* pVertex = (ms3dVertexInfo*)pPtr;
        pVertices_[i].BoneID    = pVertex->m_boneID;
        
        memcpy(pVertices_[i].Location, pVertex->m_vertex, sizeof(f32)*3);
        
        pPtr += sizeof(ms3dVertexInfo);
    }
    
    /*
     * Reading triangle indices
     */
    
    s32 nTriangles      = *(u16*)pPtr;
    CountOfTriangles_   = nTriangles;
    pTriangles_         = new STriangleMS3D[nTriangles];
    pPtr += sizeof(u16);
    
    for (i = 0; i < nTriangles; ++i)
    {
        ms3dTriangleInfo *pTriangle = (ms3dTriangleInfo*)pPtr;
        s32 vertexIndices[3] = { pTriangle->m_vertexIndices[0], pTriangle->m_vertexIndices[1], pTriangle->m_vertexIndices[2] };
        f32 t[3] = { 1.0f - pTriangle->m_t[0], 1.0f - pTriangle->m_t[1], 1.0f - pTriangle->m_t[2] };
        
        memcpy(pTriangles_[i].VertexNormals, pTriangle->m_vertexNormals, sizeof(f32)*3*3);
        memcpy(pTriangles_[i].TexCoordU, pTriangle->m_s, sizeof(f32)*3);
        memcpy(pTriangles_[i].TexCoordV, t, sizeof(f32)*3);
        memcpy(pTriangles_[i].VertexIndices, vertexIndices, sizeof(s32)*3);
        
        pPtr += sizeof(ms3dTriangleInfo);
    }
    
    /*
     * Reading groups
     */
    
    s32 nGroups     = *(u16*)pPtr;
    CountOfMeshes_  = nGroups;
    pMeshes_        = new SMeshMS3D[nGroups];
    pPtr += sizeof(u16);
    
    for (i = 0; i < nGroups; ++i)
    {
        pPtr += sizeof(u8); // flags
        pPtr += 32;         // name
        
        u16 nTriangles = *(u16*)pPtr;
        pPtr += sizeof(u16);
        s32* pTriangleIndices = new s32[nTriangles];
        
        for (s32 j = 0; j < nTriangles; ++j)
        {
            pTriangleIndices[j] = *(u16*)pPtr;
            pPtr += sizeof(u16);
        }
        
        s8 materialIndex = *(s8*)pPtr;
        pPtr += sizeof(s8);
        
        pMeshes_[i].MaterialIndex       = materialIndex;
        pMeshes_[i].CountOfTriangles    = nTriangles;
        pMeshes_[i].pTriangleIndices    = pTriangleIndices;
    }
    
    /*
     * Reading materials (textures etc.)
     */
    
    s32 nMaterials      = *(u16*)pPtr;
    CountOfMaterials_   = nMaterials;
    pMaterials_         = new SMaterialMS3D[nMaterials];
    pPtr += sizeof(u16);
    
    for (i = 0; i < nMaterials; ++i)
    {
        ms3dMaterialInfo *pMaterial = (ms3dMaterialInfo*)pPtr;
        
        memcpy(pMaterials_[i].Ambient, pMaterial->m_ambient, sizeof(f32)*4);
        memcpy(pMaterials_[i].Diffuse, pMaterial->m_diffuse, sizeof(f32)*4);
        memcpy(pMaterials_[i].Specular, pMaterial->m_specular, sizeof(f32)*4);
        memcpy(pMaterials_[i].Emissive, pMaterial->m_emissive, sizeof(f32)*4);
        
        pMaterials_[i].Shininess         = pMaterial->m_shininess;
        pMaterials_[i].TextureFilename   = new c8[strlen(pMaterial->m_texture)+1];
        
        strcpy(pMaterials_[i].TextureFilename, pMaterial->m_texture);
        pPtr += sizeof(ms3dMaterialInfo);
    }
    
    /*
     * Reading joints (bone/ skeletal animation)
     */
    
    f32 animFPS = *(f32*)pPtr;
    pPtr += sizeof(f32)*2;
    
    s32 totalFrames = *(s32*)pPtr;
    pPtr += sizeof(s32);
    
    TotalFrames_    = totalFrames;
    FPS_            = animFPS;
    TotalTime_      = totalFrames*1000.0f/animFPS;
    
    CountOfJoints_  = *(s16*)pPtr;
    pJoints_        = new SJointMS3D[CountOfJoints_];
    pPtr += sizeof(s16);
    
    struct SJointNameListRec
    {
        s32 JointIndex;
        const char* Name;
    };
    
    const u8* pTempPtr = pPtr;
    
    SJointNameListRec* pNameList = new SJointNameListRec[CountOfJoints_];
    
    for (i = 0; i < CountOfJoints_; ++i)
    {
        ms3dJointInfo* pJoint = (ms3dJointInfo*)pTempPtr;
        
        pTempPtr += sizeof(ms3dJointInfo);
        pTempPtr += sizeof(ms3dKeyframeInfo) * (pJoint->m_numRotationKeyframes + pJoint->m_numTranslationKeyframes);
        
        pNameList[i].JointIndex = i;
        pNameList[i].Name       = pJoint->m_name;
    }
    
    for (i = 0; i < CountOfJoints_; ++i)
    {
        ms3dJointInfo* pJoint = (ms3dJointInfo*)pPtr;
        pPtr += sizeof(ms3dJointInfo);
        
        s32 j, parentIndex = -1;
        if (strlen(pJoint->m_parentName) > 0)
        {
            for (j = 0; j < CountOfJoints_; ++j)
            {
                if (stricmp(pNameList[j].Name, pJoint->m_parentName) == 0)
                {
                    parentIndex = pNameList[j].JointIndex;
                    break;
                }
            }
            if (parentIndex == -1)
            {
                io::printError("could not find parent bone in MS3D model");
                return false;
            }
        }
        
        memcpy(&pJoints_[i].Rotation.X, pJoint->m_rotation, sizeof(dim::vector3df));
        memcpy(&pJoints_[i].Translation.X, pJoint->m_translation, sizeof(dim::vector3df));
        pJoints_[i].Parent = parentIndex;
        
        // Rotation keyframes
        SKeyframeMS3D* pTempKeyFrames;
        
        pJoints_[i].RotationKeyframes.resize(pJoint->m_numRotationKeyframes);
        pTempKeyFrames = new SKeyframeMS3D[ pJoint->m_numRotationKeyframes ];
        memcpy(&(pJoints_[i].RotationKeyframes)[0], pTempKeyFrames, sizeof(SKeyframeMS3D));
        
        spDeleteArray(pTempKeyFrames);
        
        // Translation keyframes
        pJoints_[i].TranslationKeyframes.resize(pJoint->m_numTranslationKeyframes);
        pTempKeyFrames = new SKeyframeMS3D[ pJoint->m_numTranslationKeyframes ];
        memcpy(&(pJoints_[i].TranslationKeyframes)[0], pTempKeyFrames, sizeof(SKeyframeMS3D));
        
        spDeleteArray(pTempKeyFrames);
        
        // Proces the joint keyframes
        for (j = 0; j < pJoint->m_numRotationKeyframes; ++j)
        {
            ms3dKeyframeInfo* pKeyframe = (ms3dKeyframeInfo*)pPtr;
            pPtr += sizeof(ms3dKeyframeInfo);
            //setJointKeyframe(i, j, pKeyframe->m_time*1000.0f, pKeyframe->m_parameter, true);
        }
        
        for (j = 0; j < pJoint->m_numTranslationKeyframes; ++j)
        {
            ms3dKeyframeInfo* pKeyframe = (ms3dKeyframeInfo*)pPtr;
            pPtr += sizeof(ms3dKeyframeInfo);
            //setJointKeyframe(i, j, pKeyframe->m_time*1000.0f, pKeyframe->m_parameter, false);
        }
        
    }
    
    /* Delete the memories */
    
    if (pNameList)
        delete [] pNameList;
    
    spDeleteArray(pBuffer);
    
    return true;
    
}

Entity* ModelLoaderMS3D::buildModel()
{
    
    s32 triangleIndex, index;
    
    /* Load textures */
    
    for (s32 i = 0; i < CountOfMaterials_; ++i)
    {
        if (strlen(pMaterials_[i].TextureFilename ) > 0)
        {
            pMaterials_[i].hTexture    = __spVideoDriver->loadTexture(
                TexturePath_ + io::stringc(pMaterials_[i].TextureFilename)
            );
            pMaterials_[i].TextureID   = pMaterials_[i].hTexture->getID();
        }
        else
            pMaterials_[i].TextureID   = 0;
    }
    
    /* Create model */
    
	for (s32 i = 0, j, k; i < CountOfMeshes_; ++i)
    {
        
        hEntity_->createSurface();
        
        s32 materialIndex = pMeshes_[i].MaterialIndex;
        if (materialIndex >= 0 && pMaterials_[materialIndex].TextureID > 0)
        {
            hEntity_->addTexture(pMaterials_[materialIndex].hTexture);
            hEntity_->setShininess(128.0f - pMaterials_[materialIndex].Shininess/128);
        }
        
        for (j = 0; j < pMeshes_[i].CountOfTriangles; ++j)
        {
            triangleIndex = pMeshes_[i].pTriangleIndices[j];
            const STriangleMS3D* pTri = &pTriangles_[triangleIndex];
            
            for (k = 0; k < 3; ++k)
            {
                index = pTri->VertexIndices[k];
                hEntity_->addVertex(
                    dim::vector3df(pVertices_[index].Location[0], pVertices_[index].Location[1], -pVertices_[index].Location[2]),
                    dim::vector3df(pTri->VertexNormals[k][0], pTri->VertexNormals[k][1], -pTri->VertexNormals[k][2]),
                    dim::point2df(pTri->TexCoordU[k], -pTri->TexCoordV[k])
                );
            }
            
            hEntity_->addTriangle(2, 1, 0);
            hEntity_->addVertexCounter(3);
        }
        
    }
    
    hEntity_->updateModeling(false);
	
    if (__spAnimator)
        buildAnimation();
    
    return hEntity_;
}

// ! incomplete ! (TEST AREA)

void ModelLoaderMS3D::buildAnimation()
{
    
    //anim::Animation* hAnim = __spAnimator->addAnimation(hEntity_);
    
    #if 0
    for (s32 i = 0; i < CountOfJoints_; ++i)
    {
        io::printMessage(
            io::stringc("Name = ")+pJoints_[i].Name+
            io::stringc(", rotX = ")+io::stringc(pJoints_[i].Rotation.X)+
            io::stringc(", rotY = ")+io::stringc(pJoints_[i].Rotation.Y)+
            io::stringc(", rotZ = ")+io::stringc(pJoints_[i].Rotation.Z)
        );
    }
    #endif
    
    //hAnim->updateSkeletonBones();
    
}


void ModelLoaderMS3D::init()
{
    /* Initialization */
    CountOfMeshes_      = 0;
    CountOfMaterials_   = 0;
    CountOfTriangles_   = 0;
    CountOfVertcies_    = 0;
    
    pMeshes_            = 0;
    pMaterials_         = 0;
    pTriangles_         = 0;
    pVertices_          = 0;
    pJoints_            = 0;
}

void ModelLoaderMS3D::clear()
{
    /* Delete all members */
    for (s32 i = 0; i < CountOfMeshes_; ++i)
        spDeleteArray(pMeshes_[i].pTriangleIndices);
    for (s32 i = 0; i < CountOfMaterials_; ++i)
        spDeleteArray(pMaterials_[i].TextureFilename);
    
    spDeleteArray(pMeshes_);
    spDeleteArray(pMaterials_);
    spDeleteArray(pTriangles_);
    spDeleteArray(pVertices_);
    spDeleteArray(pJoints_);
}




} // /namespace scene

} // /namespace sp








// ================================================================================
