/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * File for model loading SPM-fileformat
 *
 */

#include "spModelLoaderSPM.hpp"
#include "spMainDevice.hpp"


namespace sp
{

extern video::VideoDriver* __spVideoDriver;

namespace scene
{


ModelLoaderSPM::ModelLoaderSPM() : FileLoader_(0), hFile_(0)
{
    init();
}
ModelLoaderSPM::~ModelLoaderSPM()
{
    clear();
}

Entity* ModelLoaderSPM::loadModel(io::stringc &Filename, io::stringc &TexturePath)
{
    Filename_       = Filename;
    TexturePath_    = TexturePath;
    
    hEntity_        = new Entity();
    
    if ( !( hFile_ = FileLoader_->openFile(Filename, io::FILE_READ_BINARY) ) )
        return hEntity_;
    
    readChunk(CurrentChunk_);
    
    if (CurrentChunk_->ID != MODEL_SPM_PRIMARY)
    {
        io::printError("SPM file has incorrect header id");
        FileLoader_->closeFile(hFile_);
        return hEntity_;
    }
    
    processAllChunks();
    
    FileLoader_->closeFile(hFile_);
    
    return hEntity_;
}


/* ========== Private: ========== */


void ModelLoaderSPM::init()
{
    /* Allocate all memories */
    CurrentChunk_   = new chunkInfo;
    FileLoader_     = new io::FileLoader();
}
void ModelLoaderSPM::clear()
{
    /* Delete all memories */
    spDeletePointer(CurrentChunk_);
    spDeletePointer(FileLoader_);
}

void ModelLoaderSPM::readChunk(chunkInfo* pChunk)
{
    pChunk->ID      = FileLoader_->readValue<u16>(hFile_);
    pChunk->Length  = FileLoader_->readValue<u32>(hFile_);
}

/*
 * Main chunk
 */

void ModelLoaderSPM::processAllChunks()
{
    
    /* === Temporary variables === */
    
    f32 VertexData[8]   = { 0.0f }; // x, y, z, nx, ny, nz, u, v
    u8 Color[4]         = { 0 };    // r, g, b, a
    u32 Vertices[3]     = { 0 };    // v0, v1, v2
    f32 NodeAnimSeq[10] = { 0.0f }; // pos(x|y|z); rot(x|y|z|w); scl(x|y|z);
    
    /* === Read until the end-of-file === */
    
    while (!FileLoader_->isEOF(hFile_))
    {
        
        readChunk(CurrentChunk_);
        
        switch (CurrentChunk_->ID)
        {
            
            /* ======= Identity ======= */
            
            case MODEL_SPM_VERSION:
            {
                u16 Version = FileLoader_->readValue<u16>(hFile_);
                
                if (Version < SPM_VERSION_NUMBER)
                    io::printWarning("SPM model version is less then 3 so it may load incorrect");
            }
            break;
            
            case MODEL_SPM_OBJECTNAME:
            {
                if (CurrentChunk_->Length > 0)
                {
                    
                    c8* ObjectName = new c8[ CurrentChunk_->Length + 1 ];
                    
                    FileLoader_->readStream(hFile_, ObjectName, sizeof(c8), CurrentChunk_->Length);
                    ObjectName[CurrentChunk_->Length] = 0;
                    hEntity_->setName(ObjectName);
                    
                    spDeleteArray(ObjectName);
                }
            }
            break;
            
            /* ======= Surface & material ======= */
            
            case MODEL_SPM_NEWSURFACE:
            {
                hEntity_->createSurface();
            }
            break;
            
            case MODEL_SPM_NEWMATERIAL:
            {
                // Material name
                u32 MatNameLength = FileLoader_->readValue<u32>(hFile_);
                
                if (MatNameLength > 0)
                {
                    c8* MatName = new c8[ MatNameLength + 1 ];
                    FileLoader_->readStream(hFile_, MatName, sizeof(c8), MatNameLength);
                    spDeleteArray(MatName);
                }
                
                // Material filename
                u32 TexFileNameLength = FileLoader_->readValue<u32>(hFile_);
                c8* TexFileName = new c8[ TexFileNameLength + 1 ];
                
                if (TexFileNameLength > 0)
                {
                    FileLoader_->readStream(hFile_, TexFileName, sizeof(c8), TexFileNameLength);
                    TexFileName[TexFileNameLength] = 0;
                    hEntity_->addTexture(__spVideoDriver->loadTexture(TexFileName));
                }
                
                spDeleteArray(TexFileName);
                
                // Material properties
                hEntity_->setMappingGen(
                    0, FileLoader_->readValue<video::EMappingGenTypes>(hFile_)
                );
            }
            break;
            
            /* ======= Mesh ======= */
            
            case MODEL_SPM_NEWVERTEX:
            {
                // Read all vertex data
                FileLoader_->readStream(hFile_, VertexData, sizeof(f32), 8);
                FileLoader_->readStream(hFile_, Color, sizeof(u8), 4);
                
                // Add a vertex
                hEntity_->addVertex(
                    dim::vector3df(VertexData[0], VertexData[1], VertexData[2]),    // x, y, z
                    dim::vector3df(VertexData[3], VertexData[4], VertexData[5]),    // nx, ny, nz
                    dim::point2df(VertexData[6], VertexData[7]),                    // u, v
                    video::color(Color[0], Color[1], Color[2], Color[3]),           // r, g, b, a
                    0.0f                                                            // fog
                );
            }
            break;
            
            case MODEL_SPM_NEWTRIANGLE:
            {
                // Read vertices data
                FileLoader_->readStream(hFile_, Vertices, sizeof(u32), 3);
                
                // Add a triangle
                hEntity_->addTriangle(Vertices[0], Vertices[1], Vertices[2]);
            }
            break;
            
            /* ======= Animation ======= */
            
            case MODEL_SPM_ANIMATION:
            {
                // Create a new animation
                if (__spAnimator)
                    Anim_ = __spAnimator->addAnimation(hEntity_);
            }
            break;
            
            case MODEL_SPM_NEWNODE:
            {
                // Read the three vectors
                FileLoader_->readStream(hFile_, NodeAnimSeq, sizeof(f32), 10);
                
                // Add a new node-animation sequence
                if (Anim_)
                {
                    Anim_->addAnimSeqNode(
                        dim::vector3df(NodeAnimSeq[0], NodeAnimSeq[1], NodeAnimSeq[2]),
                        dim::quaternion(NodeAnimSeq[3], NodeAnimSeq[4], NodeAnimSeq[5], NodeAnimSeq[6]),
                        dim::vector3df(NodeAnimSeq[7], NodeAnimSeq[8], NodeAnimSeq[9])
                    );
                }
            }
            break;
            
            case MODEL_SPM_NEWKEYFRAME:
            {
                // Vertices coordination list
                u32 VerticesCount = hEntity_->getMeshVerticesCount(0);
                std::vector<dim::vector3df> VerticesCoord(VerticesCount);
                
                // Loop for the vertices count
                for (u32 i = 0; i < VerticesCount; ++i)
                {
                    // Read the vertices coords
                    FileLoader_->readStream(hFile_, &VerticesCoord[i].X, sizeof(f32), 3);
                }
                
                // Add a new keyframe-animation sequence
                if (Anim_)
                    Anim_->addAnimSeqMesh(VerticesCoord);
            }
            break;
            
            /* ======= Skeletal animation ======= */
            
            case MODEL_SPM_NEWBONE:
            {
                // Temporary vairables
                u32 ParentBoneID;
                io::stringc BoneName;
                dim::vector3df Translation, Rotation, Scale;
                u32 VerticesIndicesCount;
                
                /* === General bone information === */
                
                // Read parent bone identity number
                ParentBoneID = FileLoader_->readValue<u32>(hFile_);
                
                // Read translation
                FileLoader_->readStream(hFile_, &Translation.X, sizeof(f32), 3);
                
                // Read rotation
                FileLoader_->readStream(hFile_, &Rotation.X, sizeof(f32), 3);
                
                // Read scale
                FileLoader_->readStream(hFile_, &Scale.X, sizeof(f32), 3);
                
                // Read the bone name
                u32 BoneNameLength = FileLoader_->readValue<u32>(hFile_);
                c8* TmpStr = new c8[ BoneNameLength + 1 ];
                
                if (BoneNameLength > 0)
                {
                    FileLoader_->readStream(hFile_, TmpStr, sizeof(c8), BoneNameLength);
                    TmpStr[BoneNameLength] = 0;
                    BoneName = TmpStr;
                }
                
                spDeleteArray(TmpStr);
                
                /* === Linked vertices === */
                
                // Read vertices indices count
                VerticesIndicesCount = FileLoader_->readValue<u32>(hFile_);
                
                // Read all vertices indices
                std::vector<u32> VerticesIndices(VerticesIndicesCount), VerticesSurfaces(VerticesIndicesCount);
                
                for (u32 i = 0; i < VerticesIndicesCount; ++i)
                {
                    VerticesIndices [i] = FileLoader_->readValue<u32>(hFile_);
                    VerticesSurfaces[i] = FileLoader_->readValue<u32>(hFile_);
                }
                
                /* === Creating the bone === */
                
                // Check if the animation has been created
                if (Anim_)
                {
                    // Add a skeleton bone
                    u32 BoneID = Anim_->addAnimSkeletonBone(
                        ParentBoneID,   // Parent bone identity number
                        Translation,    // Translation (location)
                        Rotation,       // Rotation (identity)
                        Scale,          // Scale (identity)
                        BoneName        // Bone name
                    );
                    
                    // Add all vertices indices
                    Anim_->addSkeletonBoneVertices(BoneID, VerticesIndices, VerticesSurfaces);
                }
            }
            break;
            
            case MODEL_SPM_NEWBONEFRAME:
            {
                // Temporary variables
                u32 BoneID;
                f32 Speed;
                
                dim::vector3df Pos, Scl;
                dim::quaternion Quat;
                
                // Read the bone identity number
                BoneID = FileLoader_->readValue<u32>(hFile_);
                
                // Read the frame: translation
                FileLoader_->readStream(hFile_, &Pos.X, sizeof(f32), 3);
                
                // Read the frame: rotation
                FileLoader_->readStream(hFile_, &Quat.X, sizeof(f32), 4);
                
                // Read the frame: scale
                FileLoader_->readStream(hFile_, &Scl.X, sizeof(f32), 3);
                
                // Read the frame speed
                Speed = FileLoader_->readValue<f32>(hFile_);
                
                // Add a new bone frame
                if (Anim_)
                    Anim_->addSkeletonBoneFrame(BoneID, Pos, Quat, Scl, Speed);
            }
            break;
            
        }
        
    } // /while
    
    hEntity_->updateModeling(false);
    
}


} // /namespace scene

} // /namespace sp








// ================================================================================
