/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * OpenGL video driver
 *
 */

#include "spOpenGLDriver.hpp"
#include "spSceneManagerCamera.hpp"
#include "spOpenGLSpecialEffectsActor.hpp"
#include "spOpenGLFunctionsARB.hpp"
#include "spInternalDeclerations.hpp"


namespace sp
{

extern UserControl* __spUserControl;

namespace video
{


/*
 * ========== Internal members ==========
 */

f32 __spLightPosition[4] = { 0.0f, 0.0f, 0.0f, 1.0f };

const GLenum GLTexDimensions[4] = {
    GL_TEXTURE_1D, GL_TEXTURE_2D, GL_TEXTURE_3D, GL_TEXTURE_CUBE_MAP
};

const s32 GLCompareList[] = {
    GL_NEVER, GL_EQUAL, GL_NOTEQUAL, GL_LESS, GL_LEQUAL, GL_GREATER, GL_GEQUAL, GL_ALWAYS,
};

const s32 GLMappingGenList[] = {
    GL_OBJECT_LINEAR, GL_OBJECT_LINEAR, GL_EYE_LINEAR, GL_SPHERE_MAP, GL_NORMAL_MAP, GL_REFLECTION_MAP,
};

const s32 GLTextureEnvList[] = {
    GL_MODULATE, GL_REPLACE, GL_ADD, GL_ADD_SIGNED, GL_SUBTRACT, GL_INTERPOLATE, GL_DOT3_RGB,
};

const s32 GLBlendingList[] = {
    GL_ZERO, GL_ONE, GL_SRC_COLOR, GL_ONE_MINUS_SRC_COLOR, GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA,
    GL_DST_COLOR, GL_ONE_MINUS_DST_COLOR, GL_DST_ALPHA, GL_ONE_MINUS_DST_ALPHA,
};


/*
 * ========== Internal static functions ==========
 */

static GLenum getGLTexDim(s32 DimType)
{
    return GLTexDimensions[DimType - TEXTURE_1D];
}


/*
 * ========== Constructors & destructor ==========
 */


OpenGLDriver::OpenGLDriver() :
    VideoDriver(DRIVER_OPENGL), hRenderTexture_(0), isCullFace_(false), isSolidMode_(false), MultiTextureCount_(0)
{
    init();
}
OpenGLDriver::~OpenGLDriver()
{
}


/*
 * ========== Video driver identity ==========
 */

io::stringc OpenGLDriver::getRenderer() const
{
    return io::stringc(glGetString(GL_RENDERER));
}
io::stringc OpenGLDriver::getVersion() const
{
    return io::stringc("OpenGL ") + glGetString(GL_VERSION);
}
io::stringc OpenGLDriver::getVendor() const
{
    return io::stringc(glGetString(GL_VENDOR));
}

bool OpenGLDriver::supportAntiAlias() const
{
    return isExtensionSupported("WGL_ARB_multisample");
}

bool OpenGLDriver::supportMultitex() const
{
    return getMultitexCount() > 1;
}

bool OpenGLDriver::supportVBO() const
{
    return isExtensionSupported("GL_ARB_vertex_buffer_object");
}

bool OpenGLDriver::supportAnisotropicFiltering() const
{
    return isExtensionSupported("GL_EXT_texture_filter_anisotropic");
}

s32 OpenGLDriver::getMultitexCount() const
{
    s32 MultiTextureCount;
    glGetIntegerv(GL_MAX_TEXTURE_UNITS_ARB, &MultiTextureCount);
    return MultiTextureCount;
}

s32 OpenGLDriver::getMaxAnisotropicFilter() const
{
    GLfloat MaxAnisotropy;
    glGetFloatv(GL_MAX_TEXTURE_MAX_ANISOTROPY_EXT, &MaxAnisotropy);
    return (s32)MaxAnisotropy;
}

dim::EMatrixProjectionTypes OpenGLDriver::getProjectionMatrixType() const
{
    return dim::MATRIX_RIGHTHANDED;
}

void OpenGLDriver::printWarning()
{
    if (io::stringc(glGetString(GL_VERSION)).left(1).val<s32>() < 2)
        io::printWarning("OpenGL version is less than 2.0, 3D scenes are maybe not available");
}


/*
 * ========== User controll functions (clear buffers, flip buffers, 2d drawing etc.) ==========
 */

void OpenGLDriver::clearBuffers()
{
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}

void OpenGLDriver::flipBuffers()
{
    #if defined(WIN32)
    SwapBuffers(hDC_);
    #elif defined(LINUX)
    
    #endif
}


/*
 * ========== Setting-/ getting functions ==========
 */


void OpenGLDriver::setShadeMode(EShadeModeTypes ShadeMode)
{
    switch (ShadeMode)
    {
        case SHADEMODE_SMOOTH:
            glShadeModel(GL_SMOOTH);
            break;
        case SHADEMODE_FLAT:
            glShadeModel(GL_FLAT);
            break;
    }
}

void OpenGLDriver::setRenderMode(EDriverRenderModes RenderMode)
{
    switch (RenderMode)
    {
        case RENDERMODE_LINE_LINEAR:
            glDisable(GL_LINE_SMOOTH);
            break;
        case RENDERMODE_LINE_SMOOTH:
            glEnable(GL_LINE_SMOOTH);
            break;
            
        case RENDERMODE_POINT_LINEAR:
            glDisable(GL_POINT_SMOOTH);
            break;
        case RENDERMODE_POINT_SMOOTH:
            glEnable(GL_POINT_SMOOTH);
            break;
    }
}

void OpenGLDriver::setPolygonMode(EDriverRenderModes RenderFace, EDriverRenderModes RenderMode)
{
    s32 Face = 0, Mode = 0;
    
    switch (RenderFace)
    {
        case RENDERMODE_FRONT:
            Face = GL_BACK; break;
        case RENDERMODE_BACK:
            Face = GL_FRONT; break;
        case RENDERMODE_FRONT_AND_BACK:
            Face = GL_FRONT_AND_BACK; break;
    }
    switch (RenderMode)
    {
        case RENDERMODE_SOLID:
            Mode = GL_FILL; break;
        case RENDERMODE_LINES:
            Mode = GL_LINE; break;
        case RENDERMODE_POINTS:
            Mode = GL_POINT; break;
    }
    
    if (Face && Mode)
        glPolygonMode(Face, Mode);
}

void OpenGLDriver::setFrontFace(bool isFrontFace)
{
    isFrontFace_ = isFrontFace;
    glFrontFace(isFrontFace ? GL_CCW : GL_CW);
}

void OpenGLDriver::setClearColor(color Color)
{
    glClearColor(f32(Color.Red)/255.0f, f32(Color.Green)/255.0f, f32(Color.Blue)/255.0f, f32(Color.Alpha)/255.0f);
}

void OpenGLDriver::setColorMask(bool isRed, bool isGreen, bool isBlue, bool isAlpha)
{
    glColorMask(isRed, isGreen, isBlue, isAlpha);
}

#if defined(WIN32)

void OpenGLDriver::setVsync(bool isVsync)
{
    if (wglSwapIntervalEXT)
        wglSwapIntervalEXT(isVsync ? 1 : 0);
}

#elif defined(LINUX)

#endif

void OpenGLDriver::setAntiAlias(bool isAntiAlias)
{
    if (isAntiAlias)
        glEnable(GL_MULTISAMPLE_ARB);
    else
        glDisable(GL_MULTISAMPLE_ARB);
}


/*
 * ======= Context functions (for initializing the engine) =======
 */

void OpenGLDriver::initRenderer()
{
    /* Load all OpenGL extensions */
    loadExtensions();
    
    /* Default settings */
    glDepthFunc(GL_LEQUAL);
    glEnable(GL_ALPHA_TEST);
    glEnable(GL_NORMALIZE);
    
    /* Hints for the renderer */
    glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
    glHint(GL_GENERATE_MIPMAP_HINT, GL_NICEST);
    
    /* Set the coloring mode with lighting */
    glColorMaterial(GL_FRONT_AND_BACK, GL_DIFFUSE);
    
    /* Get multi texture count information */
    MultiTextureCount_ = getMultitexCount();
    
    defaultTextureGenMode();
    
    #if 0
    //glTexEnvi(GL_TEXTURE_FILTER_CONTROL, GL_TEXTURE_LOD_BIAS, 2);
    #endif
    
    /* Clear & flip the video buffer */
    clearBuffers();
    flipBuffers();
}

void OpenGLDriver::changeDisplaySettings(bool isFullScreen, dim::size2di Dimension, s32 ColorDepth)
{
    if (isFullScreen)
    {
        DEVMODE dmScreenSettings;
        memset(&dmScreenSettings, 0, sizeof(dmScreenSettings));
        
        dmScreenSettings.dmSize         = sizeof(dmScreenSettings);
        dmScreenSettings.dmPelsWidth    = Dimension.Width;
        dmScreenSettings.dmPelsHeight   = Dimension.Height;
        dmScreenSettings.dmBitsPerPel   = ColorDepth;
        dmScreenSettings.dmFields       = DM_BITSPERPEL | DM_PELSWIDTH | DM_PELSHEIGHT;
        
        /* Change the display settings for the screen */
        ChangeDisplaySettings(&dmScreenSettings, CDS_FULLSCREEN);
    }
}

bool OpenGLDriver::initScreen(HWND hWnd, bool isAntiAlias, s32 MultiSamplePixelFormat, s32 ColorDepth)
{
    
    /* Device context */
    if ( !( hDC_ = GetDC(hWnd) ) ) // Device-Context
    {
        io::printError("could not create device context");
        return false;
    }
    
    /* Pixel format attributes */
    static PIXELFORMATDESCRIPTOR pfd =
    {
        sizeof(PIXELFORMATDESCRIPTOR),  // Size Of This Pixel Format Descriptor
        1,                              // Version Number
        PFD_DRAW_TO_WINDOW |            // Format Must Support Window
        PFD_SUPPORT_OPENGL |            // Format Must Support OpenGL
        PFD_DOUBLEBUFFER,               // Must Support Double Buffering
        PFD_TYPE_RGBA,                  // Request An RGBA Format
        ColorDepth,                     // Select Our Color Depth
        0, 0, 0, 0, 0, 0,               // Color Bits Ignored
        0,                              // No Alpha Buffer
        0,                              // Shift Bit Ignored
        0,                              // No Accumulation Buffer
        0, 0, 0, 0,                     // Accumulation Bits Ignored
        24,                             // Z-Buffer bits (Depth Buffer)
        1,                              // Stencil Buffer
        0,                              // No Auxiliary Buffer
        PFD_MAIN_PLANE,                 // Main Drawing Layer
        0,                              // Reserved
        0, 0, 0                         // Layer Masks Ignored
    };
    
    /* Choos a pixel format */
    if (isAntiAlias && MultiSamplePixelFormat)
    {
        PixelFormat_ = MultiSamplePixelFormat;
    }
    else
    {
        if ( !( PixelFormat_ = ChoosePixelFormat(hDC_, &pfd) ) )
        {
            io::printError("could not find a suitable pixelformat");
            return false;
        }
    }
    
    /* Set the pixel format */
    if (!SetPixelFormat(hDC_, PixelFormat_, &pfd))
    {
        io::printError("could not set the pixelformat");
        return false;
    }
    
    /* Create a device context */
    if ( !( hRC_ = wglCreateContext(hDC_) ) ) // Rendering-Context
    {
        io::printError("could not create 'OpenGL' rendering context");
        return false;
    }
    
    /* Make current & activate the OpenGL context */
    if (!wglMakeCurrent(hDC_, hRC_))
    {
        io::printError("could not activate 'OpenGL' rendering context");
        return false;
    }
    
    return true;
    
}

void OpenGLDriver::clearScreen(HWND &hWnd)
{
    
    if (hWnd)
    {
        if (hDC_)
        {
            if (!wglMakeCurrent(hDC_, 0))
                io::printError("release of 'OpenGL' device- and resource context failed");
            
            if (hRC_)
            {
                if (!wglDeleteContext(hRC_))
                    io::printError("release of 'OpenGL' rendering context failed");
                hRC_ = 0;
            }
            else
                io::printError("resource-context handle is zero");
            
            if (!ReleaseDC(hWnd, hDC_))
                io::printError("release of device context failed");
            hDC_ = 0;
        }
        else
            io::printError("device-contect handle is zero");
        
        if (!DestroyWindow(hWnd))
            io::printError("release of main window failed");
        hWnd = 0;
    }
    else
        io::printError("window handle is zero");
    
}

void OpenGLDriver::initAntiAlias(HWND Wnd, bool &isAntiAlias, s32 &MultiSamplePixelFormat, s32 &MultiSampleFormat, s32 ColorDepth)
{
    if (supportAntiAlias())
    {
        // Load the pixel-format-ARB function
        if (!wglChoosePixelFormatARB)
        {
            wglChoosePixelFormatARB = (PFNWGLCHOOSEPIXELFORMATARBPROC)wglGetProcAddress("wglChoosePixelFormatARB");
            if (!wglChoosePixelFormatARB)
            {
                io::printError("could not load OpenGL function 'wglChoosePixelFormatARB'");
                return;
            }
        }
        
        // Temporary varibales
        s32 valid;
        u32 numFormats;
        f32 fAttributes[] = { 0, 0 };
        
        s32 iAttributes[] =
        {
            WGL_DRAW_TO_WINDOW_ARB, GL_TRUE,
            WGL_SUPPORT_OPENGL_ARB, GL_TRUE,
            WGL_ACCELERATION_ARB, WGL_FULL_ACCELERATION_ARB,
            WGL_COLOR_BITS_ARB, ColorDepth,
            WGL_ALPHA_BITS_ARB, 8,
            WGL_DEPTH_BITS_ARB, 24,
            WGL_STENCIL_BITS_ARB, 1,
            WGL_DOUBLE_BUFFER_ARB, GL_TRUE,
            WGL_SAMPLE_BUFFERS_ARB, isAntiAlias ? GL_TRUE : GL_FALSE,
            WGL_SAMPLES_ARB, MultiSampleFormat,
            0, 0
        };
        
        // Set the multisample attributes
        valid = wglChoosePixelFormatARB(hDC_, iAttributes, fAttributes, 1, &MultiSamplePixelFormat, &numFormats);
        
        if (!valid || numFormats < 1)
            io::printWarning(io::stringc("multisample: ")+io::stringc(MultiSampleFormat)+io::stringc(" operation failed"));
        
        // Activating
        if (isAntiAlias)
            glEnable(GL_MULTISAMPLE_ARB);
        else
            glDisable(GL_MULTISAMPLE_ARB);
        
        // Recreation of the screen
        HWND hWndParent = GetParent(Wnd);
        
    }
    else
    {
        io::printWarning("multisample is not supported, antialias can not be used");
        isAntiAlias = false;
        MultiSampleFormat = 0;
    }
}


/*
 * ======= Rendering 3D scenes =======
 */

void OpenGLDriver::setMaterialStatesTriangleList(SMeshMaterial &Material, dim::vector3df &Rotation)
{
    
    /* Cull facing */
    switch (Material.RenderFace)
    {
        case video::RENDER_FRONT:
            glEnable(GL_CULL_FACE);
            glCullFace(GL_FRONT);
            break;
        case video::RENDER_BACK:
            glEnable(GL_CULL_FACE);
            glCullFace(GL_BACK);
            break;
        case video::RENDER_BOTH:
            glDisable(GL_CULL_FACE);
            break;
    }
    
    /* Render states */
    setGlRenderState(GL_DEPTH_TEST      , Material.isZBuffer                    );
    setGlRenderState(GL_LIGHTING        , Material.isLighting && __isLighting   );
    setGlRenderState(GL_FOG             , Material.isFog && __isFog             );
    setGlRenderState(GL_BLEND           , Material.isBlending                   );
    setGlRenderState(GL_COLOR_MATERIAL  , Material.isColorMaterial              );
    
    /* Shininess */
    glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, Material.Shininess);
    
    /* Color: Diffuse */
    Material.ColorDiffuse.getFloatArray(TempColor_);
    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, TempColor_);
    
    /* Color: Ambient */
    Material.ColorAmbient.getFloatArray(TempColor_);
    glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, TempColor_);
    
    /* Color: Specular */
    Material.ColorSpecular.getFloatArray(TempColor_);
    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, TempColor_);
    
    /* Color: Emission */
    Material.ColorEmission.getFloatArray(TempColor_);
    glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, TempColor_);
    
    /* Depth function */
    glDepthFunc(GLCompareList[Material.DepthMode]);
    
    /* Alpha function */
    glAlphaFunc(GLCompareList[Material.AlphaMode], Material.AlphaReference);
    
    /* Blending function */
    glBlendFunc(GLBlendingList[Material.BlendSource], GLBlendingList[Material.BlendTarget]);
    
    /* Polygon mode */
    glPolygonMode(GL_BACK, GL_POINT + Material.RenderModeFront);
    glPolygonMode(GL_FRONT, GL_POINT + Material.RenderModeBack);
    
    /* Shader */
    if (Material.hShader)
        Material.hShader->bind();
    
}

void OpenGLDriver::setMaterialStatesSprite(SMeshMaterial &Material)
{
    
    /* Orientation */
    if (Material.isZBuffer)
        glEnable(GL_DEPTH_TEST);
    else
        glDisable(GL_DEPTH_TEST);
    
    /* Material: lighting */
    if (Material.isLighting && __isLighting)
        glEnable(GL_LIGHTING);
    else
        glDisable(GL_LIGHTING);
    
    /* Blending mode */
    glBlendFunc(GLBlendingList[Material.BlendSource], GLBlendingList[Material.BlendTarget]);
    
}

void OpenGLDriver::drawTriangleList(
    u32 &BufferID, u32* IndicesList, u32 &IndicesCount, std::vector<SSurfaceTexture> &TextureList)
{
    
    /* Check if the object is empty */
    if (!IndicesCount)
        return;
    
    /* Temporary variables */
    s32 TextureLayer = 0;
    std::vector<SSurfaceTexture>::iterator itTex;
    
    /*
     * ===== Set the buffer pointers location =====
     */
    
    /* Bind the VBO buffer */
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, BufferID);
    
    /* Vertex coordination pointer */
    glVertexPointer(3, GL_FLOAT, sizeof(SMeshVertex3D), VBO_OFFSET_VERTEX);
    
    /* Color pointer */
    glColorPointer(4, GL_UNSIGNED_BYTE, sizeof(SMeshVertex3D), VBO_OFFSET_COLOR);
    
    /* Normal vector pointer */
    glNormalPointer(GL_FLOAT, sizeof(SMeshVertex3D), VBO_OFFSET_NORMAL);
    
    /* Fog coordination pointer */
    glFogCoordPointer(GL_FLOAT, sizeof(SMeshVertex3D), VBO_OFFSET_FOG);
    
    /*
     * ===== Set the texture layers =====
     */
    
    /* Use texture matrix */
    glMatrixMode(GL_TEXTURE);
    
    /* Loop for each texture */
    for (itTex = TextureList.begin(); itTex != TextureList.end(); ++itTex, ++TextureLayer)
    {
        
        /* Bind the texture */
        itTex->hTexture->bind(TextureLayer);
        
        /* Enable texture VBO pointer */
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);
        
        /* Load texture matrix */
        glLoadMatrixf(itTex->Matrix.getArray16());
        
        /* Texture coordination pointer */
        glTexCoordPointer(3, GL_FLOAT, sizeof(SMeshVertex3D), VBO_OFFSET_TEXCOORD(TextureLayer));
        
        /* Select the texture projection */
        setGlRenderState(GL_TEXTURE_GEN_S, itTex->TexMappingCoords & MAPGEN_S);
        setGlRenderState(GL_TEXTURE_GEN_T, itTex->TexMappingCoords & MAPGEN_T);
        setGlRenderState(GL_TEXTURE_GEN_R, itTex->TexMappingCoords & MAPGEN_R);
        setGlRenderState(GL_TEXTURE_GEN_Q, itTex->TexMappingCoords & MAPGEN_Q);
        
        /* Texture coordinate generation */
        glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GLMappingGenList[itTex->TexMappingGen]);
        glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GLMappingGenList[itTex->TexMappingGen]);
        glTexGeni(GL_R, GL_TEXTURE_GEN_MODE, GLMappingGenList[itTex->TexMappingGen]);
        glTexGeni(GL_Q, GL_TEXTURE_GEN_MODE, GLMappingGenList[itTex->TexMappingGen]);
        
        /* Textrue environment */
        if (itTex->TexEnvType != TEXENV_MODULATE)
        {
            glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_COMBINE);
            glTexEnvi(GL_TEXTURE_ENV, GL_COMBINE_RGB, GLTextureEnvList[itTex->TexEnvType]);
        }
        else
            glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
        
    }
    
    /*
     * ===== Draw the buffer elements =====
     */
    
    /* Render the triangles */
    glDrawElements(GL_TRIANGLES, IndicesCount, GL_UNSIGNED_INT, IndicesList);
    
    /*
     * ===== Deactivate the textures =====
     */
    
    /* Deactivate multi texturing */
    TextureLayer = 0;
    for (std::vector<SSurfaceTexture>::iterator itTex = TextureList.begin();
         itTex != TextureList.end(); ++itTex, ++TextureLayer)
    {
        /* Unbind the texture */
        itTex->hTexture->unbind(TextureLayer);
        
        /* Disable texture VBO pointer */
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    }
    
}

void OpenGLDriver::drawTriangleList(u32 &BufferID, u32* IndicesList, u32 &IndicesCount)
{
    /* Bing the VBO buffer */
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, BufferID);
    
    /* Vertex coordination pointer */
    glVertexPointer(3, GL_FLOAT, sizeof(SMeshVertex3D), VBO_OFFSET_VERTEX);
    
    /* Final rendering */
    if (IndicesCount)
        glDrawElements(GL_TRIANGLES, IndicesCount, GL_UNSIGNED_INT, IndicesList);
}

void OpenGLDriver::drawSprite(
    Texture* &hTexture, f32 &RotationZ, video::color &Color, dim::vector3df &Scale, dim::point2df &BasePosition)
{
    
    /* Bind textureing */
    hTexture->bind(0);
    
    /* Coordination */
    if (!math::isEqual(RotationZ, 0))
        scene::spWorldMatrix.matrixRotateZ(RotationZ);
    
    /* Update the render matrix */
    updateModelviewMatrix();
    
    /* Coloring */
    glColor4ub(Color.Red, Color.Green, Color.Blue, Color.Alpha);
    
    /* Rendering */
    glBegin(GL_QUADS);
    {
        glTexCoord2f(0.0f, 0.0f); glVertex2f(-Scale.X + BasePosition.X,  Scale.Y + BasePosition.Y);
        glTexCoord2f(1.0f, 0.0f); glVertex2f( Scale.X + BasePosition.X,  Scale.Y + BasePosition.Y);
        glTexCoord2f(1.0f, 1.0f); glVertex2f( Scale.X + BasePosition.X, -Scale.Y + BasePosition.Y);
        glTexCoord2f(0.0f, 1.0f); glVertex2f(-Scale.X + BasePosition.X, -Scale.Y + BasePosition.Y);
    }
    glEnd();
    
    if (!math::isEqual(RotationZ, 0))
        scene::spWorldMatrix.matrixReset();
    
    /* Unbind textureing */
    hTexture->unbind(0);
    
}

void OpenGLDriver::updateLight(
    u32 &LightID, scene::ELightTypes &LightingType, bool &isVolumetric,
    dim::vector3df Direction, f32 &NearSpot, f32 &FarSpot,
    f32 &AttenuationConstant, f32 &AttenuationLinear, f32 &AttenuationQuadratic)
{
    
    /* Use an OpenGL light */
    u32 ID = LightID + GL_LIGHT0;
    
    /* Lighting location */
    if (LightingType == scene::LIGHT_DIRECTIONAL)
    {
        f32 LightPos[4] = { -Direction.X, -Direction.Y, -Direction.Z, 0.0 };
        glLightfv(ID, GL_POSITION, LightPos);
    }
    else
        glLightfv(ID, GL_POSITION, __spLightPosition);
    
    /* Spot light attributes */
    if (LightingType == scene::LIGHT_SPOT)
    {
        glLightfv(ID, GL_SPOT_DIRECTION, &Direction.X);
        glLightf(ID, GL_SPOT_CUTOFF, NearSpot);
        glLightf(ID, GL_SPOT_EXPONENT, FarSpot);
    }
    
    /* Volumetric light attenuations */
    if (isVolumetric)
    {
        glLightf(ID, GL_CONSTANT_ATTENUATION, AttenuationConstant);
        glLightf(ID, GL_LINEAR_ATTENUATION, AttenuationLinear);
        glLightf(ID, GL_QUADRATIC_ATTENUATION, AttenuationQuadratic);
    }
    
}

void OpenGLDriver::enableVertexBuffer()
{
    /* Enable all needed VBO pointer */
    glEnableClientState(GL_VERTEX_ARRAY);
    glEnableClientState(GL_NORMAL_ARRAY);
    glEnableClientState(GL_FOG_COORDINATE_ARRAY);
    glEnableClientState(GL_COLOR_ARRAY);
}

void OpenGLDriver::disableVertexBuffer()
{
    /* Disable all used VBO pointer */
    glDisableClientState(GL_VERTEX_ARRAY);
    glDisableClientState(GL_NORMAL_ARRAY);
    glDisableClientState(GL_FOG_COORDINATE_ARRAY);
    glDisableClientState(GL_COLOR_ARRAY);
}

void OpenGLDriver::enableVertexArray()
{
    /* Enable only the vertex-array */
    glEnableClientState(GL_VERTEX_ARRAY);
}

void OpenGLDriver::disableVertexArray()
{
    /* Disable only the vertex-array */
    glDisableClientState(GL_VERTEX_ARRAY);
}

void OpenGLDriver::createVertexBuffer(u32 &BufferID)
{
    /* Generate a new vertex buffer */
    glGenBuffersARB(1, &BufferID);
}

void OpenGLDriver::deleteVertexBuffer(u32 &BufferID)
{
    /* Delete the vertex buffer */
    glDeleteBuffersARB(1, &BufferID);
}

void OpenGLDriver::updateVertexBuffer(u32 &BufferID, SMeshVertex3D* pVerticesList, s32 VerticesCount)
{
    /* Bind the vertex buffer */
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, BufferID);
    
    /* Fill the vertex buffer */
    glBufferDataARB(GL_ARRAY_BUFFER_ARB, VerticesCount*sizeof(SMeshVertex3D), pVerticesList, GL_STATIC_DRAW_ARB);
}

void OpenGLDriver::disableTriangleListStates()
{
    /* Back settings - texture */
    glActiveTextureARB(GL_TEXTURE0);
    glEnable(GL_TEXTURE_2D);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    
    /* Back settings - polygons */
    glPolygonMode(GL_BACK, GL_FILL);
    glEnable(GL_CULL_FACE);
    glCullFace(GL_FRONT);
    
    /* Disable mapping projection */
    glDisable(GL_TEXTURE_GEN_S);
    glDisable(GL_TEXTURE_GEN_T);
    glDisable(GL_TEXTURE_GEN_R);
    glDisable(GL_TEXTURE_GEN_Q);
    
    /* Default render functions */
    glAlphaFunc(GL_ALWAYS, 0.0);
    glDepthFunc(GL_LEQUAL);
    glEnable(GL_BLEND);
}

void OpenGLDriver::disable3DRenderStates()
{
    /* Disable all used states */
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_BLEND);
    glDisable(GL_LIGHTING);
    glDisable(GL_FOG);
}

void OpenGLDriver::disableTexturing()
{
    glDisable(GL_TEXTURE_2D);
}

void OpenGLDriver::setDefaultAlphaBlending()
{
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}

void OpenGLDriver::enableBlending()
{
    glEnable(GL_BLEND);
}

void OpenGLDriver::disableBlending()
{
    glDisable(GL_BLEND);
}


/*
 * ======= Lighting =======
 */

void OpenGLDriver::addDynamicLightSource(
    u32 LightID, scene::ELightTypes Type,
    video::color &Diffuse, video::color &Ambient, video::color &Specular,
    f32 AttenuationConstant_, f32 AttenuationLinear_, f32 AttenuationQuadratic_)
{
    /* Enable the light */
    glEnable(GL_LIGHT0 + LightID);
    
    /* Set the light color */
    setLightColor(LightID, Diffuse, Ambient, Specular);
}

void OpenGLDriver::setLightStatus(u32 LightID, bool isEnable)
{
    if (isEnable)
        glEnable(GL_LIGHT0 + LightID);
    else
        glDisable(GL_LIGHT0 + LightID);
}

void OpenGLDriver::setLightColor(u32 LightID, video::color &Diffuse, video::color &Ambient, video::color &Specular)
{
    /* Use an OpenGL light */
    LightID += GL_LIGHT0;
    
    /* Temporary memories */
    f32 pDiffuse[4], pAmbient[4], pSpecular[4];
    
    /* Color: Diffuse */
    Diffuse.getFloatArray(pDiffuse);
    glLightfv(LightID, GL_DIFFUSE, pDiffuse);
    
    /* Color: Ambient */
    Ambient.getFloatArray(pAmbient);
    glLightfv(LightID, GL_AMBIENT, pAmbient);
    
    /* Color: Specular */
    Specular.getFloatArray(pSpecular);
    glLightfv(LightID, GL_SPECULAR, pSpecular);
}


/*
 * ======= Fog effect =======
 */

void OpenGLDriver::setFogMode(s32 FogMode)
{
    /* General settings */
    FogMode_ = FogMode;
    
    /* Select the fog mode */
    switch (FogMode)
    {
        case FOG_NONE:
        {
            __isFog = false;
        }
        break;
        
        case FOG_STATIC:
        {
            __isFog = true;
            
            /* Set fog type */
            switch (FogType_)
            {
                case FOG_PALE:
                    glFogi(GL_FOG_MODE, GL_EXP); break;
                case FOG_THICK:
                    glFogi(GL_FOG_MODE, GL_EXP2); break;
            }
            
            /* Range settings */
            glFogf(GL_FOG_DENSITY, FogRange_);
            glFogf(GL_FOG_START, FogNear_);
            glFogf(GL_FOG_END, FogFar_);
        }
        break;
        
        case FOG_VOLUMETRIC:
        {
            __isFog = true;
            
            /* Renderer settings */
            glFogi(GL_FOG_MODE, GL_LINEAR);
            glFogi(GL_FOG_COORDINATE_SOURCE_EXT, GL_FOG_COORDINATE_EXT);
            glFogf(GL_FOG_DENSITY, FogRange_);
            glFogf(GL_FOG_START, 0.0f);
            glFogf(GL_FOG_END, 1.0f);
        }
        break;
    }
}

void OpenGLDriver::setFogColor(video::color Color)
{
    f32 pColor[4];
    Color.getFloatArray(pColor);
    glFogfv(GL_FOG_COLOR, pColor);
}

void OpenGLDriver::setFogRange(f32 Range, f32 NearFog, f32 FarFog, s32 FogType)
{
    FogType_    = FogType;
    FogRange_   = Range;
    FogNear_    = NearFog;
    FogFar_     = FarFog;
    
    if (FogMode_ != FOG_VOLUMETRIC)
    {
        /* Set fog type */
        switch (FogType_)
        {
            case FOG_PALE:
                glFogi(GL_FOG_MODE, GL_EXP); break;
            case FOG_THICK:
                glFogi(GL_FOG_MODE, GL_EXP2); break;
        }
        
        /* Set range */
        glFogf(GL_FOG_DENSITY, FogRange_);
        glFogf(GL_FOG_START, FogNear_);
        glFogf(GL_FOG_END, FogFar_);
    }
}


/*
 * ======= Stencil buffer =======
 */

void OpenGLDriver::clearStencilBuffer()
{
    glClearStencil(0);
}

void OpenGLDriver::drawStencilShadowVolume(
    const dim::vector3df* pTriangleList, s32 Count, bool ZFailMethod, bool VolumetricShadow)
{
    if (!pTriangleList || !Count)
        return;
    
    /* Store the current OpenGL states */
    glPushAttrib(
        GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_ENABLE_BIT | GL_POLYGON_BIT | GL_STENCIL_BUFFER_BIT
    );
    
    /* Configure the stencil states */
    glDisable(GL_LIGHTING);
    glDisable(GL_FOG);
    
    glDepthFunc(GL_LEQUAL);
    
    glColorMask(GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE);
    glDepthMask(GL_FALSE);
    
    glEnable(GL_STENCIL_TEST);
    glEnable(GL_CULL_FACE);
    glEnable(GL_DEPTH_CLAMP_NV);
    //glEnable(GL_POLYGON_OFFSET_FILL);
    //glPolygonOffset(0.0f, 1.0f);
    
    glBindBufferARB(GL_ARRAY_BUFFER_ARB, 0);
    glEnableClientState(GL_VERTEX_ARRAY);
    glVertexPointer(3, GL_FLOAT, sizeof(dim::vector3df), pTriangleList);
    
    glStencilMask(~0);
    glStencilFunc(GL_ALWAYS, 0, ~0);
    
    /* Draw the stencil shadow */
    if (ZFailMethod)
    {
        glStencilOp(GL_KEEP, GL_INCR_WRAP_EXT, GL_KEEP);
        glCullFace(GL_FRONT);
        glDrawArrays(GL_TRIANGLES, 0, Count);
        
        if (!VolumetricShadow)
        {
            glStencilOp(GL_KEEP, GL_DECR_WRAP_EXT, GL_KEEP);
            glCullFace(GL_BACK);
            glDrawArrays(GL_TRIANGLES, 0, Count);
        }
    }
    else
    {
        glCullFace(GL_BACK);
        glStencilOp(GL_KEEP, GL_KEEP, GL_INCR_WRAP_EXT);
        glDrawArrays(GL_TRIANGLES, 0, Count);
        
        if (!VolumetricShadow)
        {
            glCullFace(GL_FRONT);
            glStencilOp(GL_KEEP, GL_KEEP, GL_DECR_WRAP_EXT);
            glDrawArrays(GL_TRIANGLES, 0, Count);
        }
    }
    
    /* Reset the OpenGL states */
    glDisableClientState(GL_VERTEX_ARRAY);
    glPopAttrib();
}

void OpenGLDriver::drawStencilShadow(video::color Color)
{
    /* Store the current OpenGL states */
    glPushAttrib(
        GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_ENABLE_BIT | GL_POLYGON_BIT | GL_STENCIL_BUFFER_BIT
    );
    
    /* Configure the stencil states */
    glColorMask(GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE);
    glDepthMask(GL_FALSE);
    
    glEnable(GL_STENCIL_TEST);
    glStencilFunc(GL_NOTEQUAL, 0, 0xFFFFFFFF);//~0);
    glStencilOp(GL_KEEP, GL_KEEP, GL_KEEP);
    
    /* Draw the rectangle */
    beginDrawing2D();
    draw2DRectangle(dim::rect2di(0, 0, __Screen_Width, __Screen_Height), Color);
    endDrawing2D();
    
    /* Clear the stencil buffer */
    glClear(GL_STENCIL_BUFFER_BIT);
    
    /* Reset the OpenGL states */
    glPopMatrix();
    glPopAttrib();
}


/* === Clipping planes === */

void OpenGLDriver::setClipPlane(u32 Index, const dim::plane3df &Plane, bool Enable)
{
    if (Index >= MaxClippingPlanes_)
        return;
    
    setGlRenderState(GL_CLIP_PLANE0 + Index, Enable);
    
    ClippingPlanes_[Index] = SClipPlane(Index, Plane, Enable);
}


/* === Special effects actor === */

scene::effect::SpecialEffectsActor* OpenGLDriver::createSpecialEffectsActor()
{
    return new scene::effect::OpenGLSpecialEffectsActor();
}



/*
 * ======= Shader effects =======
 */

Shader* OpenGLDriver::createShader(EStandardShaderTypes Type)
{
    Shader* NewShader = new OpenGLShader(Type);
    
    ShaderList_.push_back(NewShader);
    
    return NewShader;
}

Shader* OpenGLDriver::loadShader(io::stringc FilenameVertex, io::stringc FilenameFragment, bool UseHighLevel)
{
    const io::stringc ShaderType    = (UseHighLevel ? "GLSL shader" : "shader program");
    const io::stringc PrimaryFile   = (FilenameVertex != "" ? FilenameVertex : FilenameFragment);
    const io::stringc SecondaryFile = (FilenameFragment != "" && FilenameVertex != "" ? io::stringc("\" and \"") + FilenameFragment : "\"");
    
    io::printMessage(
        io::stringc("Load ") + ShaderType + io::stringc(": \"") + PrimaryFile + SecondaryFile
    );
    
    Shader* NewShader = new OpenGLShader(FilenameVertex, FilenameFragment, UseHighLevel);
    
    ShaderList_.push_back(NewShader);
    
    return NewShader;
}
Shader* OpenGLDriver::loadShader(
    io::stringc FilenameVertex, io::stringc FilenamePixel,
    io::stringc EntryPointVertex, io::stringc EntryPointPixel,
    EHLSLVertexShaderTypes CompilerTargetVertex,
    EHLSLPixelShaderTypes CompilerTargetPixel,
    bool UseHighLevel)
{
    const io::stringc ShaderType    = (UseHighLevel ? "GLSL shader" : "shader program");
    const io::stringc PrimaryFile   = (FilenameVertex != "" ? FilenameVertex : FilenamePixel);
    const io::stringc SecondaryFile = (FilenamePixel != "" && FilenameVertex != "" ? io::stringc("\" and \"") + FilenamePixel : "\"");
    
    io::printMessage(
        io::stringc("Load ") + ShaderType + io::stringc(": \"") + PrimaryFile + SecondaryFile
    );
    
    Shader* NewShader = new OpenGLShader(FilenameVertex, FilenamePixel, UseHighLevel);
    
    ShaderList_.push_back(NewShader);
    
    return NewShader;
}

void OpenGLDriver::unbindShader()
{
    /* Unbind the shader program */
    glUseProgramObjectARB(0);
}




/*
 * ======= Stack operation functions =======
 */

void OpenGLDriver::pushViewport()
{
    /* Get the current viewport settings */
    glGetIntegerv(GL_VIEWPORT, StackViewport_);
}

void OpenGLDriver::popViewport()
{
    glViewport(StackViewport_[0], StackViewport_[1], StackViewport_[2], StackViewport_[3]);
}


/*
 * ======= Texturing =======
 */

void OpenGLDriver::bindTexture(u32 ID)
{
    glBindTexture(GL_TEXTURE_2D, ID);
}

void OpenGLDriver::unbindTexture()
{
    glBindTexture(GL_TEXTURE_2D, 0);
}

void OpenGLDriver::setTextureSubImage(u32 ID, s32 X, s32 Y, s32 Width, s32 Height, const u8* ImageData)
{
    /* Bind the texture */
    glBindTexture(GL_TEXTURE_2D, ID);
    
    /* Set the sub image data */
    glTexSubImage2D(GL_TEXTURE_2D, 0, X, Y, Width, Height, GL_BGR, GL_UNSIGNED_BYTE, ImageData);
    //glTexSubImage2D(GL_TEXTURE_2D, 0, X, Y, Width, Height, GL_RGB, GL_UNSIGNED_BYTE, ImageData);
}


/*
 * ========== Drawing 2D objects ==========
 */



/* Drawing 2d */

void OpenGLDriver::beginDrawing2D(bool isSettingViewport)
{
    /* Disable depth test (only 2d) */
	glDisable(GL_DEPTH_TEST);
	
	/* Alpha blending */
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    
    /* Enable alpha ablending */
	glEnable(GL_BLEND);
    
    /* Projection matrix */
    glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	
	/* Modelview matrix */
    glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
    Matrix2D_.make2Dimensional(__Screen_Width, -__Screen_Height);
    
    /* Disable lighting & fog */
    glDisable(GL_LIGHTING);
    glDisable(GL_FOG);
    
    /* Use first texture layer */
    glActiveTextureARB(GL_TEXTURE0);
    
    /* Normal polygon mode */
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
    
    /* Other modes & options */
    
	if (isSettingViewport)
    	glViewport(0, 0, __Screen_Width, __Screen_Height);
	
    if (isSolidMode_)
    {
        glGetBooleanv(GL_CULL_FACE, &isCullFace_);
        glDisable(GL_CULL_FACE);
    }
}

void OpenGLDriver::endDrawing2D()
{
	/* Enable depth test */
    glEnable(GL_DEPTH_TEST);
    
    /* Disable alpha ablending */
	glDisable(GL_BLEND);
	
    /* Projection matrix */
    glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	
	/* Modelview matrix */
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
	
    /* Other settings */
    if (isSolidMode_ && isCullFace_)
        glEnable(GL_CULL_FACE);
}

/* Drawing 3d */

void OpenGLDriver::beginDrawing3D(bool isLighting, bool isFog)
{
    /* Matrix reset */
	glLoadIdentity();
	
    /* Lighting */
    if (isLighting)
        glEnable(GL_LIGHTING);
    else
        glDisable(GL_LIGHTING);
    
    /* Fog */
    if (isFog)
        glEnable(GL_FOG);
    else
        glDisable(GL_FOG);
    
    /* Texture 0 (zero) */
    glActiveTextureARB(GL_TEXTURE0);
    glDisable(GL_TEXTURE_2D);
    
    /* Normal polygon mode */
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
}

void OpenGLDriver::endDrawing3D()
{
    /* Matrix reset */
    glLoadIdentity();
    
    /* Disable lighting & fog */
    glDisable(GL_LIGHTING);
    glDisable(GL_FOG);
}

/* Alpha blending */

void OpenGLDriver::beginAlphaBlending(u8 Type)
{
    /* Blending function */
    if (Type == BLENDING_BRIGHTNESS)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE);
    else if (Type == BLENDING_DARKNESS)
        glBlendFunc(GL_DST_COLOR, GL_ZERO);
}

void OpenGLDriver::endAlphaBlending()
{
    /* Blending function */
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}

/* Masking */

void OpenGLDriver::beginMasking(u8 Type)
{
    /* Blending function */
    if (Type == MASKING_SOURCE)
        glBlendFunc(GL_DST_COLOR, GL_ZERO);
    else if (Type == MASKING_TARGET)
        glBlendFunc(GL_SRC_ALPHA, GL_ONE);
}

void OpenGLDriver::endMasking()
{
    /* Blending function */
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}

/* Texture drawing */

void OpenGLDriver::beginTextureDrawing(Texture* hTexture, s32 RenderMode, u32 TextureDataType)
{
    /* First settings */
    hRenderTexture_         = hTexture;
    RenderTextureMode_      = RenderMode;
    //FilterTextureMode_      = FilterMode;
    RenderTextureDataType_  = TextureDataType;
    
    /* Clear the drawing buffers */
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
    
    /* Save the current matrix */
    glPushMatrix();
    
    /* Load the matrix's identity */
    glLoadIdentity();
    
    /* Get the current viewport */
    pushViewport();
    
    /* Set the new viewport */
    setViewport(0, 0, hRenderTexture_->getSize().Width, hRenderTexture_->getSize().Height);
}

void OpenGLDriver::endTextureDrawing(bool isUpdateImageData)
{
    /* Copy to texture */
    if (hRenderTexture_)
    {
        glEnable(GL_TEXTURE_2D);
        
        /* Bind the texture */
        glBindTexture(GL_TEXTURE_2D, hRenderTexture_->getID());
        
        /* Set the texture paremter */
        setTextureParameters(
            hRenderTexture_->getDimensionType(), RenderTextureMode_, FilterTextureMode_, TexGenAnisotropy_, false, false
        );
        
        /* Copy the pixels into the texture */
        glCopyTexImage2D(
            GL_TEXTURE_2D, 0, RenderTextureDataType_, 0, 0,
            hRenderTexture_->getSize().Width, hRenderTexture_->getSize().Height, 0
        );
        
        /* Update the image buffer */
        if (isUpdateImageData)
            hRenderTexture_->updateImageData(GET_IMAGEDATA);
        
        /* Back settings */
        hRenderTexture_ = 0;
        
        glDisable(GL_TEXTURE_2D);
    }
    
    /* Clear the buffers */
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
    
    /* Use the old matrix */
    glPopMatrix();
    
    /* Set the old viewport back */
    popViewport();
}

/* Scissors (drawing rectangle) */

void OpenGLDriver::beginDrawingRect(dim::rect2di Dimension)
{
    /* Enable scissor */
    glEnable(GL_SCISSOR_TEST);
    
    /* Scissor area */
    glScissor(
        Dimension.Left, __Screen_Height - Dimension.Top - Dimension.Bottom,
        Dimension.Right - Dimension.Left, Dimension.Bottom - Dimension.Top
    );
}

void OpenGLDriver::beginDrawingRect(dim::point2di Position, dim::size2di Size)
{
    /* Enable scissor */
    glEnable(GL_SCISSOR_TEST);
    
    /* Scissor area */
    glScissor(
        Position.X, __Screen_Height - Position.Y - Size.Height,
        Size.Width, Size.Height
    );
}

void OpenGLDriver::endDrawingRect()
{
    /* Disable scissor */
    glDisable(GL_SCISSOR_TEST);
}

/* Viewports (drawing area) */

void OpenGLDriver::setDrawingArea(dim::rect2di Dimension)
{
    setViewport(
        Dimension.Left, Dimension.Top,
        Dimension.Right - Dimension.Left, Dimension.Bottom - Dimension.Top
    );
}

void OpenGLDriver::setDrawingArea(dim::point2di Position, dim::size2di Size)
{
    setViewport(Position.X, Position.Y, Size.Width, Size.Height);
}

/* Shader */

void OpenGLDriver::beginDrawingShader(Shader* hShader)
{
    /* Enable multi texturing */
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    
    /* Bind the shader */
    hShader->bind();
}

void OpenGLDriver::endDrawingShader()
{
    /* Unbind the shader */
    unbindShader();
    
    /* Disable multi texturing */
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
}

/* Manual binding & unbinding textures */

void OpenGLDriver::bindTexture(Texture* hTexture, s32 Level)
{
    glActiveTextureARB(GL_TEXTURE0 + Level);
    glBindTexture(GL_TEXTURE_2D, hTexture->getID());
    glEnable(GL_TEXTURE_2D);
}
void OpenGLDriver::unbindTexture(s32 Level)
{
    glActiveTextureARB(GL_TEXTURE0 + Level);
    glDisable(GL_TEXTURE_2D);
}

/* Line & point size */

void OpenGLDriver::setLineSize(s32 Size)
{
    glLineWidth(Size);
}
void OpenGLDriver::setPointSize(s32 Size)
{
    glPointSize(Size);
}

void OpenGLDriver::setSolidMode(bool isSolidMode)
{
    isSolidMode_ = isSolidMode;
}


/* ======= Texture drawing ======= */


void OpenGLDriver::draw2DImage(Texture* hTexture, dim::point2di Position, color Color, bool isTransFilterUsing)
{
    /* Blending */
    if (!isTransFilterUsing)
        glDisable(GL_BLEND);
    
    /* Coloring */
    glColor4ub(Color.Red, Color.Green, Color.Blue, Color.Alpha);
    
    /* Binding the texture */
    glBindTexture(GL_TEXTURE_2D, hTexture->getID());
    glEnable(GL_TEXTURE_2D);
    
    /* Load 2dimensional matrix */
    glLoadMatrixf(Matrix2D_.getArray16());
    
    /* Drawing the quad */
    glBegin(GL_QUADS);
    {
        glTexCoord2f(0.0f, 0.0f);
        glVertex2i(Position.X, Position.Y);
        
        glTexCoord2f(1.0f, 0.0f);
        glVertex2i(Position.X + hTexture->getSize().Width, Position.Y);
        
        glTexCoord2f(1.0f, 1.0f);
        glVertex2i(Position.X + hTexture->getSize().Width, Position.Y + hTexture->getSize().Height);
        
        glTexCoord2f(0.0f, 1.0f);
        glVertex2i(Position.X, Position.Y + hTexture->getSize().Height);
    }
    glEnd();
    
    /* Back settings */
    glDisable(GL_TEXTURE_2D);
    glEnable(GL_BLEND);
}

void OpenGLDriver::draw2DImage(
    Texture* hTexture, dim::rect2di Position, dim::rect2df Clipping,
    color Color, bool isTransFilterUsing)
{
    /* Blending */
    if (!isTransFilterUsing)
        glDisable(GL_BLEND);
    
    /* Coloring */
    glColor4ub(Color.Red, Color.Green, Color.Blue, Color.Alpha);
    
    /* Binding the texture */
    glBindTexture(GL_TEXTURE_2D, hTexture->getID());
    glEnable(GL_TEXTURE_2D);
    
    /* Load 2dimensional matrix */
    glLoadMatrixf(Matrix2D_.getArray16());
    
    /* Drawing the quad */
    glBegin(GL_QUADS);
    {
        glTexCoord2f(Clipping.Left, Clipping.Top);
        glVertex2i(Position.Left, Position.Top);
        
        glTexCoord2f(Clipping.Right, Clipping.Top);
        glVertex2i(Position.Left + Position.Right, Position.Top);
        
        glTexCoord2f(Clipping.Right, Clipping.Bottom);
        glVertex2i(Position.Left + Position.Right, Position.Top + Position.Bottom);
        
        glTexCoord2f(Clipping.Left, Clipping.Bottom);
        glVertex2i(Position.Left, Position.Top + Position.Bottom);
    }
    glEnd();
    
    /* Back settings */
    glDisable(GL_TEXTURE_2D);
    glEnable(GL_BLEND);
}

void OpenGLDriver::draw2DImage(
    Texture* hTexture, dim::point2di Position,
    f32 Rotation, f32 Radius, color Color, bool isTransFilterUsing)
{
    /* Temporary variables */
    Radius *= math::SQRT2;
    dim::point2df lefttopPos    (f32(Position.X + SIN(Rotation -  45)*Radius), f32(Position.Y - COS(Rotation -  45)*Radius));
    dim::point2df righttopPos   (f32(Position.X + SIN(Rotation +  45)*Radius), f32(Position.Y - COS(Rotation +  45)*Radius));
    dim::point2df rightbottomPos(f32(Position.X + SIN(Rotation + 135)*Radius), f32(Position.Y - COS(Rotation + 135)*Radius));
    dim::point2df leftbottomPos (f32(Position.X + SIN(Rotation - 135)*Radius), f32(Position.Y - COS(Rotation - 135)*Radius));
    
    /* Blending */
    if (!isTransFilterUsing)
        glDisable(GL_BLEND);
    
    /* Coloring */
    glColor4ub(Color.Red, Color.Green, Color.Blue, Color.Alpha);
    
    /* Binding the texture */
    glBindTexture(GL_TEXTURE_2D, hTexture->getID());
    glEnable(GL_TEXTURE_2D);
    
    /* Load 2dimensional matrix */
    glLoadMatrixf(Matrix2D_.getArray16());
    
    /* Drawing the quad */
    glBegin(GL_QUADS);
    {
        glTexCoord2f(0.0f, 0.0f);
        glVertex2f(lefttopPos.X, lefttopPos.Y);
        
        glTexCoord2f(1.0f, 0.0f);
        glVertex2f(righttopPos.X, righttopPos.Y);
        
        glTexCoord2f(1.0f, 1.0f);
        glVertex2f(rightbottomPos.X, rightbottomPos.Y);
        
        glTexCoord2f(0.0f, 1.0f);
        glVertex2f(leftbottomPos.X, leftbottomPos.Y);
    }
    glEnd();
    
    /* Back settings */
    glDisable(GL_TEXTURE_2D);
    glEnable(GL_BLEND);
}

void OpenGLDriver::draw2DImage(
    Texture* hTexture, dim::rect2di Position, dim::rect2df Clipping,
    bool isTransFilterUsing, f32 Rotation, dim::point2df RotationPoint,
    color lefttopColor, color righttopColor, color rightbottomColor, color leftbottomColor)
{
    /* Translation of location */
    if (Position.Right <= 0 && Position.Bottom <= 0)
    {
        Position.Right   = hTexture->getSize().Width;
        Position.Bottom  = hTexture->getSize().Height;
    }
    Position.Right   += Position.Left;
    Position.Bottom  += Position.Top;
    
    /* Temporary variables */
    dim::point2df lefttopPos, righttopPos;
    dim::point2df rightbottomPos, leftbottomPos;
    dim::point2df tmpRotPoint, tmpStartPoint;
    f32 tmpDegree, tmpRotDist;
    f32 tmpExtLine, tmpExpPosY;
    
    /* Computing the rotation */
    
    tmpRotPoint.X       = (Position.Right - Position.Left) * RotationPoint.X;
    tmpRotPoint.Y       = (Position.Bottom - Position.Top) * (1.0f - RotationPoint.Y);
    tmpExpPosY          = (Position.Bottom - Position.Top) * RotationPoint.Y;
    
    tmpExtLine          = (Position.Right - Position.Left) * RotationPoint.X;
    tmpRotDist          = math::getDistance((f32)Position.Left, (f32)Position.Top, Position.Left + tmpRotPoint.X, Position.Top + tmpRotPoint.Y);
    tmpDegree           = tmpRotDist == 0 ? 0 : ASIN( tmpExtLine / tmpRotDist );
    lefttopPos.X        = Position.Left + tmpRotPoint.X - SIN(tmpDegree - Rotation) * tmpRotDist;
    lefttopPos.Y        = Position.Top  + tmpExpPosY    - COS(tmpDegree - Rotation) * tmpRotDist;
    
    tmpRotDist          = math::getDistance((f32)Position.Left, (f32)Position.Bottom, Position.Left + tmpRotPoint.X, Position.Top + tmpRotPoint.Y);
    tmpDegree           = tmpRotDist == 0 ? 0 : ASIN( tmpExtLine / tmpRotDist );
    leftbottomPos.X     = Position.Left + tmpRotPoint.X - SIN(tmpDegree + Rotation) * tmpRotDist;
    leftbottomPos.Y     = Position.Top  + tmpExpPosY    + COS(tmpDegree + Rotation) * tmpRotDist;
    
    tmpExtLine          = (Position.Right - Position.Left) - (Position.Right - Position.Left) * RotationPoint.X;
    tmpRotDist          = math::getDistance((f32)Position.Right, (f32)Position.Top, Position.Left + tmpRotPoint.X, Position.Top + tmpRotPoint.Y);
    tmpDegree           = tmpRotDist == 0 ? 0 : ASIN( tmpExtLine / tmpRotDist );
    righttopPos.X       = Position.Left + tmpRotPoint.X + SIN(tmpDegree + Rotation) * tmpRotDist;
    righttopPos.Y       = Position.Top  + tmpExpPosY    - COS(tmpDegree + Rotation) * tmpRotDist;
    
    tmpRotDist          = math::getDistance((f32)Position.Right, (f32)Position.Bottom, Position.Left + tmpRotPoint.X, Position.Top + tmpRotPoint.Y);
    tmpDegree           = tmpRotDist == 0 ? 0 : ASIN( tmpExtLine / tmpRotDist );
    rightbottomPos.X    = Position.Left + tmpRotPoint.X + SIN(tmpDegree - Rotation) * tmpRotDist;
    rightbottomPos.Y    = Position.Top  + tmpExpPosY    + COS(tmpDegree - Rotation) * tmpRotDist;
    
    /* Blending */
    if (!isTransFilterUsing)
        glDisable(GL_BLEND);
    
    /* Texture binding */
    glBindTexture(GL_TEXTURE_2D, hTexture->getID());
    glEnable(GL_TEXTURE_2D);
    
    /* Load 2dimensional matrix */
    glLoadMatrixf(Matrix2D_.getArray16());
    
    /* Drawing the quad */
    glBegin(GL_QUADS);
    {
        glColor4ub(lefttopColor.Red, lefttopColor.Green, lefttopColor.Blue, lefttopColor.Alpha);
        glTexCoord2f(Clipping.Left, Clipping.Bottom);
        glVertex2f(lefttopPos.X, lefttopPos.Y);
        
        glColor4ub(righttopColor.Red, righttopColor.Green, righttopColor.Blue, righttopColor.Alpha);
        glTexCoord2f(Clipping.Right, Clipping.Bottom);
        glVertex2f(righttopPos.X, righttopPos.Y);
        
        glColor4ub(rightbottomColor.Red, rightbottomColor.Green, rightbottomColor.Blue, rightbottomColor.Alpha);
        glTexCoord2f(Clipping.Right, Clipping.Top);
        glVertex2f(rightbottomPos.X, rightbottomPos.Y);
        
        glColor4ub(leftbottomColor.Red, leftbottomColor.Green, leftbottomColor.Blue, leftbottomColor.Alpha);
        glTexCoord2f(Clipping.Left, Clipping.Top);
        glVertex2f(leftbottomPos.X, leftbottomPos.Y);
    }
    glEnd();
    
    /* Back settings */
    glDisable(GL_TEXTURE_2D);
    glEnable(GL_BLEND);
}

void OpenGLDriver::draw2DImage(
    Texture* hTexture,
    dim::point2di lefttopPosition, dim::point2di righttopPosition,
    dim::point2di rightbottomPosition, dim::point2di leftbottomPosition,
    dim::point2df lefttopClipping, dim::point2df righttopClipping,
    dim::point2df rightbottomClipping, dim::point2df leftbottomClipping,
    bool isTransFilterUsing,
    color lefttopColor, color righttopColor,
    color rightbottomColor, color leftbottomColor)
{
    /* Blending */
    if (!isTransFilterUsing)
        glDisable(GL_BLEND);
    
    /* Texture binding */
    glBindTexture(GL_TEXTURE_2D, hTexture->getID());
    glEnable(GL_TEXTURE_2D);
    
    /* Load 2dimensional matrix */
    glLoadMatrixf(Matrix2D_.getArray16());
    
    /* Drawing the quad */
    glBegin(GL_QUADS);
    {
        glColor4ub(lefttopColor.Red, lefttopColor.Green, lefttopColor.Blue, lefttopColor.Alpha);
        glTexCoord2f(lefttopClipping.X, lefttopClipping.Y);
        glVertex2i(lefttopPosition.X, lefttopPosition.Y);
        
        glColor4ub(righttopColor.Red, righttopColor.Green, righttopColor.Blue, righttopColor.Alpha);
        glTexCoord2f(righttopClipping.X, righttopClipping.Y);
        glVertex2i(righttopPosition.X, righttopPosition.Y);
        
        glColor4ub(rightbottomColor.Red, rightbottomColor.Green, rightbottomColor.Blue, rightbottomColor.Alpha);
        glTexCoord2f(rightbottomClipping.X, rightbottomClipping.Y);
        glVertex2i(rightbottomPosition.X, rightbottomPosition.Y);
        
        glColor4ub(leftbottomColor.Red, leftbottomColor.Green, leftbottomColor.Blue, leftbottomColor.Alpha);
        glTexCoord2f(leftbottomClipping.X, leftbottomClipping.Y);
        glVertex2i(leftbottomPosition.X, leftbottomPosition.Y);
    }
    glEnd();
    
    /* Back settings */
    glDisable(GL_TEXTURE_2D);
    glEnable(GL_BLEND);
}


/* ======= Primitive drawing ======= */


color OpenGLDriver::getPixel(dim::point2di Position)
{
    u8 pPixel[3];
    
    glReadPixels(Position.X, __Screen_Height - Position.Y, 1, 1, GL_RGB, GL_UNSIGNED_BYTE, pPixel);
    
    return color(pPixel[0], pPixel[1], pPixel[2]);
}


void OpenGLDriver::draw2DPixel(dim::point2di Position, color Color)
{
    glColor4ub(Color.Red, Color.Green, Color.Blue, Color.Alpha);
    
    /* Load 2dimensional matrix */
    glLoadMatrixf(Matrix2D_.getArray16());
    
    glBegin(GL_POINTS);
    {
        glVertex2i(Position.X, Position.Y);
    }
    glEnd();
}

void OpenGLDriver::draw2DLine(dim::point2di Position1, dim::point2di Position2, color Color)
{
    glColor4ub(Color.Red, Color.Green, Color.Blue, Color.Alpha);
    
    /* Load 2dimensional matrix */
    glLoadMatrixf(Matrix2D_.getArray16());
    
    glBegin(GL_LINES);
    {
        glVertex2i(Position1.X, Position1.Y);
        glVertex2i(Position2.X, Position2.Y);
    }
    glEnd();
}

void OpenGLDriver::draw2DLine(dim::point2di Position1, dim::point2di Position2, color Color1, color Color2)
{
    /* Load 2dimensional matrix */
    glLoadMatrixf(Matrix2D_.getArray16());
    
    glBegin(GL_LINES);
    {
        glColor4ub(Color1.Red, Color1.Green, Color1.Blue, Color1.Alpha);
        glVertex2i(Position1.X, Position1.Y);
        
        glColor4ub(Color2.Red, Color2.Green, Color2.Blue, Color2.Alpha);
        glVertex2i(Position2.X, Position2.Y);
    }
    glEnd();
}

void OpenGLDriver::draw2DLine(dim::point2di Position1, dim::point2di Position2, color Color, s32 DotLength)
{
    s32 dx = Position2.X - Position1.X;
    s32 dy = Position2.Y - Position1.Y;
    s32 Raster = 0;
    
    glColor4ub(Color.Red, Color.Green, Color.Blue, Color.Alpha);
    
    /* Load 2dimensional matrix */
    glLoadMatrixf(Matrix2D_.getArray16());
    
    glBegin(GL_POINTS);
    
    draw2DPixelFast(Position1.X, Position1.Y);
    
    if (math::Abs(dx) > math::Abs(dy))
    {
        f32 m = (f32)dy / (f32)dx;
        f32 b = Position1.Y - m*Position1.X;
        dx = dx < 0 ? -1 : 1;
        while (Position1.X != Position2.X)
        {
            Position1.X += dx;
            ++Raster;
            if (DotLength == 0 || Raster <= DotLength)
                draw2DPixelFast(Position1.X, s32(m*Position1.X+b));
            if (DotLength != 0 && Raster >= DotLength*2)
                Raster = 0;
        }
    }
    else if (dy != 0)
    {
        f32 m = (f32)dx / (f32)dy;
        f32 b = Position1.X - m*Position1.Y;
        dy = dy < 0 ? -1 : 1;
        while (Position1.Y != Position2.Y)
        {
            Position1.Y += dy;
            ++Raster;
            if (DotLength == 0 || Raster < DotLength)
                draw2DPixelFast(s32(m*Position1.Y+b), Position1.Y);
            if (DotLength != 0 && Raster >= DotLength*2)
                Raster = 0;
        }
    }
    
    glEnd();
}

void OpenGLDriver::draw2DRectangle(dim::rect2di Position, color Color, EDrawingModeTypes Type)
{
    dim::point2df lefttopPos(f32(Position.Left), f32(Position.Top));
    dim::point2df rightbottomPos(f32(Position.Right), f32(Position.Bottom));
    
    glColor4ub(Color.Red, Color.Green, Color.Blue, Color.Alpha);
    
    /* Load 2dimensional matrix */
    glLoadMatrixf(Matrix2D_.getArray16());
    
    glBegin(Type == DRAW_FRAME ? GL_LINE_LOOP : GL_QUADS);
    {
        glVertex2f(lefttopPos.X, lefttopPos.Y);
        glVertex2f(rightbottomPos.X, lefttopPos.Y);
        glVertex2f(rightbottomPos.X, rightbottomPos.Y);
        glVertex2f(lefttopPos.X, rightbottomPos.Y);
    }
    glEnd();
}

void OpenGLDriver::draw2DRectangle(
    dim::rect2di Position, color lefttopColor, color righttopColor,
    color rightbottomColor, color leftbottomColor, EDrawingModeTypes Type)
{
    dim::point2df lefttopPos(f32(Position.Left), f32(Position.Top));
    dim::point2df rightbottomPos(f32(Position.Right), f32(Position.Bottom));
    
    /* Load 2dimensional matrix */
    glLoadMatrixf(Matrix2D_.getArray16());
    
    glBegin(Type == DRAW_FRAME ? GL_LINE_LOOP : GL_QUADS);
    {
        glColor4ub(lefttopColor.Red, lefttopColor.Green, lefttopColor.Blue, lefttopColor.Alpha);
        glVertex2f(lefttopPos.X, lefttopPos.Y);
        
        glColor4ub(righttopColor.Red, righttopColor.Green, righttopColor.Blue, righttopColor.Alpha);
        glVertex2f(rightbottomPos.X, lefttopPos.Y);
        
        glColor4ub(rightbottomColor.Red, rightbottomColor.Green, rightbottomColor.Blue, rightbottomColor.Alpha);
        glVertex2f(rightbottomPos.X, rightbottomPos.Y);
        
        glColor4ub(leftbottomColor.Red, leftbottomColor.Green, leftbottomColor.Blue, leftbottomColor.Alpha);
        glVertex2f(lefttopPos.X, rightbottomPos.Y);
    }
    glEnd();
}

void OpenGLDriver::draw2DRoundRectangle(dim::rect2di Position, dim::size2di CornerRadius, color Color)
{
    /* Settings */
    if (CornerRadius.Width < 0)
        CornerRadius.Width = 0;
    if (CornerRadius.Height < 0)
        CornerRadius.Height = 0;
    if (CornerRadius.Width > (Position.Right-Position.Left)/2)
        CornerRadius.Width = (Position.Right-Position.Left)/2;
    if (CornerRadius.Height > (Position.Bottom-Position.Top)/2)
        CornerRadius.Height = (Position.Bottom-Position.Top)/2;
    
    /* Lines */
    draw2DLine(dim::point2di(Position.Left+CornerRadius.Width, Position.Top),
               dim::point2di(Position.Right-CornerRadius.Width, Position.Top), Color);
    draw2DLine(dim::point2di(Position.Left+CornerRadius.Width, Position.Bottom),
               dim::point2di(Position.Right-CornerRadius.Width, Position.Bottom), Color);
    draw2DLine(dim::point2di(Position.Left, Position.Top+CornerRadius.Height),
               dim::point2di(Position.Left, Position.Bottom-CornerRadius.Height), Color);
    draw2DLine(dim::point2di(Position.Right, Position.Top+CornerRadius.Height),
               dim::point2di(Position.Right, Position.Bottom-CornerRadius.Height), Color);
    
    /* Arcs */
    draw2DArc(dim::point2di(Position.Left+CornerRadius.Width+1, Position.Top+CornerRadius.Height),
              CornerRadius+dim::size2di(0, 1), -90, 180, 0.0f, Color);
    draw2DArc(dim::point2di(Position.Right-CornerRadius.Width, Position.Top+CornerRadius.Height),
              CornerRadius+dim::size2di(0, 1), 0, 270, 0.0f, Color);
    draw2DArc(dim::point2di(Position.Left+CornerRadius.Width+1, Position.Bottom-CornerRadius.Height-1),
              CornerRadius, -180, 90, 0.0f, Color);
    draw2DArc(dim::point2di(Position.Right-CornerRadius.Width, Position.Bottom-CornerRadius.Height-1),
              CornerRadius+dim::size2di(1, 0), -270, 0, 0.0f, Color);
}

void OpenGLDriver::draw2DCircle(dim::point2di Position, s32 Radius, color Color)
{
    Radius = math::Abs(Radius);
    
    s32 f = 1 - Radius;
    s32 ddF_x = 0;
    s32 ddF_y = -2 * Radius;
    s32 x = 0;
    s32 y = Radius;
    
    /* ===== Drawing ===== */
    
    glColor4ub(Color.Red, Color.Green, Color.Blue, Color.Alpha);
    
    /* Load 2dimensional matrix */
    glLoadMatrixf(Matrix2D_.getArray16());
    
    glBegin(GL_POINTS);
    
    draw2DPixelFast(Position.X, Position.Y + Radius);
    draw2DPixelFast(Position.X, Position.Y - Radius);
    draw2DPixelFast(Position.X + Radius, Position.Y);
    draw2DPixelFast(Position.X - Radius, Position.Y);
    
    while (x < y)
    {
        if (f >= 0)
        {
            --y;
            ddF_y += 2;
            f += ddF_y;
        }
        ++x;
        ddF_x += 2;
        f += ddF_x + 1;
        
        draw2DPixelFast(Position.X + x, Position.Y + y);
        draw2DPixelFast(Position.X - x, Position.Y + y);
        draw2DPixelFast(Position.X + x, Position.Y - y);
        draw2DPixelFast(Position.X - x, Position.Y - y);
        draw2DPixelFast(Position.X + y, Position.Y + x);
        draw2DPixelFast(Position.X - y, Position.Y + x);
        draw2DPixelFast(Position.X + y, Position.Y - x);
        draw2DPixelFast(Position.X - y, Position.Y - x);
    }
    
    glEnd();
}

void OpenGLDriver::draw2DEllipse(dim::point2di Position, dim::size2di Radius, color Color)
{
    Radius.Width = math::Abs(Radius.Width);
    Radius.Height = math::Abs(Radius.Height);
    
    s32 a = Radius.Width, b = Radius.Height;
    s32 xc = Position.X, yc = Position.Y;
    s32 a2 = a*a, b2 = b*b, fa2 = 4*a2, fb2 = 4*b2;
    s32 x, y, sigma;
    
    /* ===== Drawing ===== */
    
    glColor4ub(Color.Red, Color.Green, Color.Blue, Color.Alpha);
    
    /* Load 2dimensional matrix */
    glLoadMatrixf(Matrix2D_.getArray16());
    
    glBegin(GL_POINTS);
    
    for (x = 0, y = b, sigma = 2*b2+a2*(1-2*b); b2*x <= a2*y; ++x)
    {
        draw2DPixelFast(xc+x,yc+y);
        draw2DPixelFast(xc-x,yc+y);
        draw2DPixelFast(xc+x,yc-y);
        draw2DPixelFast(xc-x,yc-y);
        
        if ( sigma >= 0 )
        {
            sigma += fa2*(1-y);
            --y;
        }
        
        sigma += b2*(4*x+6);
    }
    
    for (x = a, y = 0, sigma = 2*a2+b2*(1-2*a); a2*y <= b2*x; ++y)
    {
        draw2DPixelFast(xc+x,yc+y);
        draw2DPixelFast(xc-x,yc+y);
        draw2DPixelFast(xc+x,yc-y);
        draw2DPixelFast(xc-x,yc-y);
        
        if ( sigma >= 0 )
        {
            sigma += fb2*(1-x);
            --x;
        }
        
        sigma += a2*(4*y+6);
    }
    
    glEnd();
}

void OpenGLDriver::draw2DEllipse(
    dim::point2di Position, dim::size2di Radius, f32 RotationDegree,
    color Color, bool isSolid)
{
    if (Radius.Width < 1)   Radius.Width = 1;
    if (Radius.Height < 1)  Radius.Height = 1;
    
    s32 Detail = Radius.Width > Radius.Height ? 800 / Radius.Width : 800 / Radius.Height;
    if (Detail < 2)
        Detail = 2;
    else if (Detail > 30)
        Detail = 30;
    const s32 Counter = 360 % Detail == 0 ? 360/Detail : 360/Detail+1;
    
    /* ===== Drawing ===== */
    
    glColor4ub(Color.Red, Color.Green, Color.Blue, Color.Alpha);
    
    /* Load 2dimensional matrix */
    glLoadMatrixf(Matrix2D_.getArray16());
    
    glBegin(GL_LINES);
    
    if (isSolid) // posx,posy,radiusx,radiusy,angle
    {
        
        s32 MaxRadius = Radius.Width > Radius.Height ? Radius.Width : Radius.Height;
        f32 xcos = COS(RotationDegree);
        f32 xsin = SIN(RotationDegree);
        f32 tmpx, tmpy;
        s32 x2, y2, x, y, oldx, oldy, diffx, diffy;
        
        const s32 FillSize = MaxRadius*2;
        s32 FillMin[FillSize], FillMax[FillSize];
        
        for (s32 i = 0; i <= FillSize; ++i)
        {
            FillMax[i] = 0;
            FillMin[i] = __Screen_Width;
        }
        
        for (s32 i = 0; i <= Counter; ++i)
        {
            tmpx = COS(i*Detail) * Radius.Width;
            tmpy = SIN(i*Detail) * -Radius.Height;
            x = s32(Position.X + tmpx * xcos + tmpy * -xsin);
            y = s32(tmpx * xsin + tmpy * xcos);
            
            if (i > 0)
            {
                draw2DLineFast(x, Position.Y+y, oldx, Position.Y+oldy);
                
                diffx = x-oldx;
                diffy = y-oldy;
                
                if (diffy != 0)
                {
                    const f32 Counter2 = math::Abs(diffy);
                    for (s32 i2 = 1; i2 <= Counter2; ++i2)
                    {
                        if (diffy > 0)
                            y2 = oldy+i2;
                        else if (diffy < 0)
                            y2 = oldy-i2;
                        x2 = oldx + diffx*i2 / math::Abs(diffy);
                        if (FillMax[y2+MaxRadius] < x2)
                            FillMax[y2+MaxRadius] = x2;
                        if (FillMin[y2+MaxRadius] > x2)
                            FillMin[y2+MaxRadius] = x2;
                    }
                }
            }
            
            oldx = x;
            oldy = y;
        }
        
        for (s32 i = 0; i <= FillSize; ++i)
        {
            if (FillMin[i] <= FillMax[i])
            {
                draw2DLineFast(
                    FillMin[i], i-MaxRadius+Position.Y,
                    FillMax[i]+1, i-MaxRadius+Position.Y
                );
            }
        }
        
    }
    else
    {
        
        f32 xcos = COS(RotationDegree);
        f32 xsin = SIN(RotationDegree);
        f32 tmpx, tmpy, x, y, oldx, oldy;
        
        for (s32 i = 0; i <= Counter; ++i)
        {
            tmpx = COS(i*Detail)*Radius.Width;
            tmpy = SIN(i*Detail)*(-Radius.Height);
            x = Position.X + tmpx*xcos + tmpy*(-xsin);
            y = Position.Y + tmpx*xsin + tmpy*xcos;
            
            if (i > 0)
                draw2DLineFast(x, y, oldx, oldy);
            
            oldx = x;
            oldy = y;
        }
        
    }
    
    glEnd();
}

void OpenGLDriver::draw2DPolygon(SPrimitiveVertex* VerticesList, u32 AryCount, EDrawingModeTypes Type)
{
    /* Load 2dimensional matrix */
    glLoadMatrixf(Matrix2D_.getArray16());
    
    if (Type == DRAW_SOLID)
    {
        glBegin(GL_TRIANGLES);
        {
            for (s32 i = 0; i < AryCount; ++i)
            {
                glColor4ub(
                    getRed  (VerticesList[i].Color),
                    getGreen(VerticesList[i].Color),
                    getBlue (VerticesList[i].Color),
                    getAlpha(VerticesList[i].Color)
                );
                
                glVertex2f(VerticesList[i].Coord.X, VerticesList[i].Coord.Y);
            }
        }
        glEnd();
    }
    else if (Type == DRAW_FRAME)
    {
        glBegin(GL_LINES);
        {
            for (s32 i = 0, j, k, l; i < AryCount; i += 3)
            {
                for (j = 0; j < 3; ++j)    
                {
                    for (k = 0; k < 2; ++k)
                    {
                        l = (j == 2 && k == 1 ? i : i+j+k);
                        
                        glColor4ub(
                            getRed  (VerticesList[l].Color),
                            getGreen(VerticesList[l].Color),
                            getBlue (VerticesList[l].Color),
                            getAlpha(VerticesList[l].Color)
                        );
                        
                        glVertex2f(VerticesList[l].Coord.X, VerticesList[l].Coord.Y);
                    } // next vertex
                } // next line
            } // next triangle
        }
        glEnd();
    }
}

void OpenGLDriver::draw2DPolygonImage(
    Texture* hTexture, SPrimitiveVertex* VerticesList, u32 AryCount, EDrawingModeTypes Type)
{
    
}

/*
 * ======= Extra drawing functions =======
 */


void OpenGLDriver::draw2DArc(
    dim::point2di Position, dim::size2di Radius, s32 RadialDegree1,
    s32 RadialDegree2, f32 RotationDegree, color Color)
{
    if (Radius.Width < 1)   Radius.Width = 1;
    if (Radius.Height < 1)  Radius.Height = 1;
    
    s32 Detail = Radius.Width > Radius.Height ? 800 / Radius.Width : 800 / Radius.Height;
    if (Detail < 2)
        Detail = 2;
    else if (Detail > 30)
        Detail = 30;
    const s32 Counter = 360 % Detail == 0 ? 360/Detail : 360/Detail+1;
    const s32 Degree1 = -RadialDegree1 * Counter / 360; // !!!
    const s32 Degree2 = RadialDegree2 * Counter / 360;
    
    /* Drawing */
    
    glColor4ub(Color.Red, Color.Green, Color.Blue, Color.Alpha);
    
    /* Load 2dimensional matrix */
    glLoadMatrixf(Matrix2D_.getArray16());
    
    glBegin(GL_LINES);
    
    f32 xcos = COS(RotationDegree);
    f32 xsin = SIN(RotationDegree);
    f32 tmpx, tmpy, x, y, oldx, oldy;
    
    for (s32 i = Degree1; i <= Counter-Degree2; ++i)
    {
        tmpx = COS(i*Detail)*Radius.Width;
        tmpy = SIN(i*Detail)*(-Radius.Height);
        x = Position.X+tmpx*xcos+tmpy*-xsin;
        y = Position.Y+tmpx*xsin+tmpy*xcos;
        
        if (i > Degree1)
            draw2DLineFast(x, y, oldx, oldy);
        
        oldx = x;
        oldy = y;
    }
    
    glEnd();
}

void OpenGLDriver::draw2DBezier(
    dim::point2di Position1, dim::point2di Position2,
    dim::point2di Radial1, dim::point2di Radial2,
    color Color, f32 Detail)
{
    dim::point2df Pos1, Pos2;
    Pos1 = math::getBezierValue(
        0.0f, Position1.cast<f32>(), Position2.cast<f32>(), Radial1.cast<f32>(), Radial2.cast<f32>()
    );
    
    glColor4ub(Color.Red, Color.Green, Color.Blue, Color.Alpha);
    
    /* Load 2dimensional matrix */
    glLoadMatrixf(Matrix2D_.getArray16());
    
    glBegin(GL_LINES);
    
    for (f32 t = 0.0f; t < 1.0f; t += Detail)
    {
        Pos2 = math::getBezierValue(
            Detail + t, Position1.cast<f32>(), Position2.cast<f32>(), Radial1.cast<f32>(), Radial2.cast<f32>()
        );
        
        draw2DLineFast((s32)Pos1.X, (s32)Pos1.Y, (s32)Pos2.X, (s32)Pos2.Y);
        
        Pos1 = Pos2;
    }
    
    glEnd();
}

void OpenGLDriver::draw2DSpline(math::Spline2D &Spline, s32 Detail, dim::point2di Pos, color Color)
{
    /* Temporary variables */
    f64 step    = 1.0 / Detail;
    f64 t       = 0.0;
    
    dim::point2df TmpPoint;
    
    /* Set color */
    glColor4ub(Color.Red, Color.Green, Color.Blue, Color.Alpha);
    
    /* Load 2dimensional matrix */
    glLoadMatrixf(Matrix2D_.getArray16());
    
    /* Draw the spline */
    for (s32 i = 0; i < Spline.getPointCount(); ++i, t = 0.0)
    {
        glBegin(GL_LINE_STRIP);
        
        for (s32 j = 0; j <= Detail; ++j, t += step)
        {
            TmpPoint = Spline.getPolynom(i).calc(t);
            
            glVertex2f(TmpPoint.X, TmpPoint.Y);
        }
        
        glEnd();
    }
}

void OpenGLDriver::draw2DFloodFill(dim::point2di Position, color FillColor, color BorderColor)
{
    std::vector<dim::point2di> PointList;
    std::vector<dim::point2di>::iterator it;
    color stdColor = getPixel(Position);
    
    PointList.push_back(Position);
    
    glColor4ub(FillColor.Red, FillColor.Green, FillColor.Blue, FillColor.Alpha);
    
    /* Load 2dimensional matrix */
    glLoadMatrixf(Matrix2D_.getArray16());
    
    glBegin(GL_POINTS);
    
    if (BorderColor.Alpha == 0) // Someone border
    {
        while (PointList.size() > 0)
        {
            // Setting
            it = PointList.end();
            --it;
            Position = *it;
            PointList.erase(it);
            
            // Checking
            if (getPixel(Position) == stdColor)
            {
                draw2DPixelFast(Position.X, Position.Y);
                PointList.push_back(Position + dim::point2di(0, 1));
                PointList.push_back(Position - dim::point2di(0, 1));
                PointList.push_back(Position + dim::point2di(1, 0));
                PointList.push_back(Position - dim::point2di(1, 0));
            }
        }
    }
    else // Specified border
    {
        while (PointList.size() > 0)
        {
            // Setting
            it = PointList.end();
            --it;
            Position = *it;
            PointList.erase(it);
            
            // Checking
            if (getPixel(Position) != BorderColor)
            {
                draw2DPixelFast(Position.X, Position.Y);
                PointList.push_back(Position + dim::point2di(0, 1));
                PointList.push_back(Position - dim::point2di(0, 1));
                PointList.push_back(Position + dim::point2di(1, 0));
                PointList.push_back(Position - dim::point2di(1, 0));
            }
        }
    }
    
    glEnd();
}


/* ======= 3D drawing functions ======= */


void OpenGLDriver::draw3DPixel(dim::vector3df Position, color Color, scene::Camera* hCamera)
{
    matrixWorldViewReset();
    
    if (hCamera)
        hCamera->updateModelviewMatrix();
    
    updateModelviewMatrix();
    
    glColor4ub(Color.Red, Color.Green, Color.Blue, Color.Alpha);
    glBegin(GL_POINTS);
    {
        glVertex3f(Position.X, Position.Y, Position.Z);
    }
    glEnd();
}

void OpenGLDriver::draw3DLine(dim::vector3df Position1, dim::vector3df Position2, color Color, scene::Camera* hCamera)
{
    matrixWorldViewReset();
    
    if (hCamera)
        hCamera->updateModelviewMatrix();
    
    updateModelviewMatrix();
    
    glColor4ub(Color.Red, Color.Green, Color.Blue, Color.Alpha);
    glBegin(GL_LINES);
    {
        glVertex3f(Position1.X, Position1.Y, Position1.Z);
        glVertex3f(Position2.X, Position2.Y, Position2.Z);
    }
    glEnd();
}

void OpenGLDriver::draw3DLine(
    dim::vector3df Position1, dim::vector3df Position2,
    color Color1, color Color2, scene::Camera* hCamera)
{
    matrixWorldViewReset();
    
    if (hCamera)
        hCamera->updateModelviewMatrix();
    
    updateModelviewMatrix();
    
    glBegin(GL_LINES);
    {
        glColor4ub(Color1.Red, Color1.Green, Color1.Blue, Color1.Alpha);
        glVertex3f(Position1.X, Position1.Y, Position1.Z);
        
        glColor4ub(Color2.Red, Color2.Green, Color2.Blue, Color2.Alpha);
        glVertex3f(Position2.X, Position2.Y, Position2.Z);
    }
    glEnd();
}

void OpenGLDriver::draw3DEllipse(
    dim::vector3df Position, dim::vector3df Rotation, dim::size2df Radius,
    color Color, scene::Camera* hCamera)
{
    matrixWorldViewReset();
    
    if (hCamera)
        hCamera->updateModelviewMatrix();
    
    updateModelviewMatrix();
    
    glTranslatef(Position.X, Position.Y, Position.Z);
    glRotatef(Rotation.Y, 0.0f, 1.0f, 0.0f);
    glRotatef(Rotation.X, 1.0f, 0.0f, 0.0f);
    glRotatef(Rotation.Z, 0.0f, 0.0f, 1.0f);
    
    glColor4ub(Color.Red, Color.Green, Color.Blue, Color.Alpha);
    glBegin(GL_LINE_STRIP);
    {
        for (s32 i = 0; i <= 36; ++i)
            glVertex2f(SIN(i*10)*Radius.Width, COS(i*10)*Radius.Height);
    }
    glEnd();
}

void OpenGLDriver::draw3DTriangle(
    Texture* hTexture, dim::triangle3df Triangle,
    dim::vector3df Position, dim::vector3df Rotation, dim::vector3df Scale,
    color Color, scene::Camera* hCamera)
{
    matrixWorldViewReset();
    
    if (hCamera)
        hCamera->updateModelviewMatrix();
    
    updateModelviewMatrix();
    
    glTranslatef(Position.X, Position.Y, Position.Z);
    glRotatef(Rotation.Y, 0.0f, 1.0f, 0.0f);
    glRotatef(Rotation.X, 1.0f, 0.0f, 0.0f);
    glRotatef(Rotation.Z, 0.0f, 0.0f, 1.0f);
    glScalef(Scale.X, Scale.Y, Scale.Z);
    
    if (hTexture)
    {
        glEnable(GL_TEXTURE_2D);
        glBindTexture(GL_TEXTURE_2D, hTexture->getID());
    }
    
    glColor4ub(Color.Red, Color.Green, Color.Blue, Color.Alpha);
    
    glBegin(GL_TRIANGLES);
    {
        glVertex3f(Triangle.PointA.X, Triangle.PointA.Y, Triangle.PointA.Z);
        glVertex3f(Triangle.PointB.X, Triangle.PointB.Y, Triangle.PointB.Z);
        glVertex3f(Triangle.PointC.X, Triangle.PointC.Y, Triangle.PointC.Z);
    }
    glEnd();
    
    glDisable(GL_TEXTURE_2D);
}

void OpenGLDriver::draw3DBox(
    dim::vector3df Corner1, dim::vector3df Corner2,
    dim::vector3df Position, dim::vector3df Rotation, dim::vector3df Scale,
    color Color, scene::Camera* hCamera)
{
    matrixWorldViewReset();
    
    if (hCamera)
        hCamera->updateModelviewMatrix();
    
    updateModelviewMatrix();
    
    glTranslatef(Position.X, Position.Y, Position.Z);
    glRotatef(Rotation.Y, 0.0f, 1.0f, 0.0f);
    glRotatef(Rotation.X, 1.0f, 0.0f, 0.0f);
    glRotatef(Rotation.Z, 0.0f, 0.0f, 1.0f);
    glScalef(Scale.X, Scale.Y, Scale.Z);
    
    glColor4ub(Color.Red, Color.Green, Color.Blue, Color.Alpha);
    
    glBegin(GL_LINE_LOOP);
    {
        glVertex3f(Corner1.X, Corner1.Y, Corner1.Z);
        glVertex3f(Corner1.X, Corner2.Y, Corner1.Z);
        glVertex3f(Corner2.X, Corner2.Y, Corner1.Z);
        glVertex3f(Corner2.X, Corner1.Y, Corner1.Z);
    }
    glEnd();
    glBegin(GL_LINE_LOOP);
    {
        glVertex3f(Corner1.X, Corner1.Y, Corner2.Z);
        glVertex3f(Corner1.X, Corner2.Y, Corner2.Z);
        glVertex3f(Corner2.X, Corner2.Y, Corner2.Z);
        glVertex3f(Corner2.X, Corner1.Y, Corner2.Z);
    }
    glEnd();
    glBegin(GL_LINES);
    {
        glVertex3f(Corner1.X, Corner2.Y, Corner1.Z);
        glVertex3f(Corner1.X, Corner2.Y, Corner2.Z);
        glVertex3f(Corner2.X, Corner2.Y, Corner1.Z);
        glVertex3f(Corner2.X, Corner2.Y, Corner2.Z);
        glVertex3f(Corner1.X, Corner1.Y, Corner1.Z);
        glVertex3f(Corner1.X, Corner1.Y, Corner2.Z);
        glVertex3f(Corner2.X, Corner1.Y, Corner1.Z);
        glVertex3f(Corner2.X, Corner1.Y, Corner2.Z);
    }
    glEnd();
}

void OpenGLDriver::draw3DBone(scene::Entity* Obj, dim::matrix4f &BoneTransformation, scene::Camera* hCamera)
{
    matrixWorldViewReset();
    
    /* === Bone location === */
    
    if (hCamera)
        hCamera->updateModelviewMatrix();
    
    Obj->updateModelviewMatrix();
    
    scene::spWorldMatrix *= BoneTransformation;
    
    updateModelviewMatrix();
    
    /* === Render the bone === */
    
    /* Bone frame */
    glColor3ub(0, 0, 255);
    glBegin(GL_LINES);
    {
        for (s32 i = 0; i < 24; ++i)
            glVertex3f(-__spBoneBodyCoords[i][1], __spBoneBodyCoords[i][2], __spBoneBodyCoords[i][0]);
    }
    glEnd();
    
    /* X-Axis */
    glColor3ub(255, 0, 0);
    glBegin(GL_LINE_STRIP);
    {
        for (s32 i = 0; i <= 36; ++i)
            glVertex3f(0.0f, SIN(i*10)*0.05f, COS(i*10)*0.05f);
    }
    glEnd();
    
    /* Y-Axis */
    glColor3ub(0, 255, 0);
    glBegin(GL_LINE_STRIP);
    {
        for (s32 i = 0; i <= 36; ++i)
            glVertex3f(SIN(i*10)*0.05f, 0.0f, COS(i*10)*0.05f);
    }
    glEnd();
    
    /* Z-Axis */
    glColor3ub(255, 255, 0);
    glBegin(GL_LINE_STRIP);
    {
        for (s32 i = 0; i <= 36; ++i)
            glVertex3f(SIN(i*10)*0.05f, COS(i*10)*0.05f, 0.0f);
    }
    glEnd();
}


/*
 * ========== Texture loading & creating ==========
 */


/*
 * Most important function, loads all the file-formats which are integrated
 * (BMP, DIB), (JPG, JPEG, JPE, JFIF), (TGA), (PNG), (PCX)
 */

Texture* OpenGLDriver::loadTexture(io::stringc Filename)
{
    
    /* Information message */
    io::printMessage(io::stringc("Load texture: \"")+Filename+io::stringc("\""));
    
    /* Allocate some memory & temporary variables */
    
    Texture* LoadedTexture      = 0;
    textureRAWInfo* texture     = 0;
    TextureLoaderRAW* texraw    = 0;
    
    io::stringc FileFormat4     = Filename.right(4).upper();
    io::stringc FileFormat5     = Filename.right(5).upper();
    
    u32 TexID                   = 0;
    s32 Type                    = TEXTURE_RGB;
    s32 InternalFormat          = 3;
    u8* ImageData               = 0;
    dim::size2di Size;
    
    /* === Check if the file does exist === */
    
    if (!io::FileLoader().doesFileExist(Filename))
    {
        /* Error message */
        io::printError("could not found texture file");
        
        /* Create a standard empty texture */
        return createTexture(DEF_TEXTURE_SIZE);
    }
    
    /* === Select format === */
    
    if (FileFormat4 == ".BMP" || FileFormat4 == ".DIB") // Windows Bitmap
        texraw = new TextureLoaderBMP(Filename);
    else if (FileFormat4 == ".JPG" || FileFormat5 == ".JPEG" || FileFormat4 == ".JPE" || FileFormat5 == ".JFIF") // Joint Photographic Experts
        texraw = new TextureLoaderJPG(Filename);
    else if (FileFormat4 == ".TGA") // Targa Image
        texraw = new TextureLoaderTGA(Filename);
    else if (FileFormat4 == ".PNG") // Portable Network Graphics
        texraw = new TextureLoaderPNG(Filename);
    else if (FileFormat4 == ".PCX") // Picture Exchange
        texraw = new TextureLoaderPCX(Filename);
    else
    {
        /* Error message */
        if (FileFormat4 == ".WAD")
            io::printError(io::stringc("texture file format: \"")+Filename.getFileExtension().upper()+io::stringc("\" must be load as a texture list"));
        else
            io::printError(io::stringc("texture file format: \"")+Filename.getFileExtension().upper()+io::stringc("\" is not supported"));
        
        /* Allocate the texture */
        LoadedTexture = createTexture(DEF_TEXTURE_SIZE);
        
        /* Unbinding 2d textures */
        glBindTexture(GL_TEXTURE_2D, 0);
        
        /* Add the texture to the texture list */
        TextureList_.push_back(LoadedTexture);
        
        /* Return the texture & exit the function */
        return LoadedTexture;
    }
    
    /* === Load the "RAW" texture === */
    
    /* Get the texture raw data */
    if (texraw)
        texture = texraw->loadImageData();
    
    /* Delete the raw texture loader */
    spDeletePointer(texraw);
    
    /* === Create the "OpenGL" texture === */
    
    if (texture)
    {
        /* Register a new OpenGL texture */
        glGenTextures(1, &TexID);
        glBindTexture(GL_TEXTURE_2D, TexID);
        
        /* OpenGL texture configurations */
        defaultTextureConfig();
        createRendererTexture(
            TexGenMipMapping_, TEXTURE_2D, dim::vector3di(texture->Width, texture->Height, 1),
            (ETextureImageDataTypes)texture->Format, texture->ImageData
        );
        
        /* Settings */
        Type            = texture->Format;
        InternalFormat  = texture->InternalFormat;
        ImageData       = texture->ImageData;
        Size            = dim::size2di(texture->Width, texture->Height);
    }
    else
    {
        /* Error message */
        io::printError("could not load texture");
    }
    
    /* === Create "SoftPixel" texture === */
    
    /* Unbinding 2d textures */
    glBindTexture(GL_TEXTURE_2D, 0);
    
    /* Allocate the texture */
    if (texture)
    {
        LoadedTexture = new OpenGLTexture(
            TexID, Filename, Size, ImageData,
            TexGenRenderMode_, TexGenFilterMode_, InternalFormat, Type, TexGenEdgeSmooth_, TexGenMipMapping_
        );
    }
    else
        LoadedTexture = createTexture(DEF_TEXTURE_SIZE);
    
    /* Delete the RAW texture */
    spDeletePointer(texture);
    
    /* Add the texture to the texture list */
    TextureList_.push_back(LoadedTexture);
    
    /* Return the texture & exit the function */
    return LoadedTexture;
    
}

bool OpenGLDriver::isTexture(Texture* hTexture)
{
    return hTexture && glIsTexture(hTexture->getID());
}

Texture* OpenGLDriver::copyTexture(Texture* hTexture)
{
    /* Temporary vairables */
    u32 TexID;
    s32 InternalFormat  = hTexture->getFormat();
    s32 Type            = hTexture->getType();
    dim::size2di Size   = hTexture->getOriginalSize();
    s32 ImageDataSize   = Size.Width * Size.Height * InternalFormat;
    GLenum TexDim       = getGLTexDim(hTexture->getDimensionType());
    
    /* Get the source image data */
    u8* pData = hTexture->getImageData();
    
    /* Allocate new image memory */
    u8* NewImageData = new u8[ImageDataSize];
    
    /* Get the format */
    s32 Format = hTexture->getType();
    checkFormat(Format);
    
    /* Copy the image data buffers */
    memcpy(NewImageData, pData, ImageDataSize);
    
    /* Register a new OpenGL texture */
    glGenTextures(1, &TexID);
    
    /* Bind the texture */
    glBindTexture(TexDim, TexID);
    
    /* Set the texture parameters */
    setTextureParameters(
        hTexture->getDimensionType(), hTexture->getRenderMode(), hTexture->getFilterMode(),
        hTexture->getAnisotropicSamples(), hTexture->getMipMapping(), TexGenEdgeSmooth_
    );
    
    /* Create the OpenGL texture */
    createRendererTexture(
        hTexture->getMipMapping(), hTexture->getDimensionType(), dim::vector3di(Size.Width, Size.Height, hTexture->getDepth()),
        (ETextureImageDataTypes)Format, NewImageData
    );
    
    /* Unbind texturing */
    glBindTexture(TexDim, 0);
    
    /* Allocate the new texture */
    Texture* NewTexture = new OpenGLTexture(
        TexID, hTexture->getFilename(), Size, NewImageData, TexGenRenderMode_, TexGenFilterMode_,
        InternalFormat, Type, TexGenEdgeSmooth_, TexGenMipMapping_
    );
    
    /* Add the texture to the texture list */
    TextureList_.push_back(NewTexture);
    
    /* Return the texture & exit the function */
    return NewTexture;
}

Texture* OpenGLDriver::createTexture(dim::size2di Dimension, u32 TextureDataType, io::stringc Filename)
{
    /* Temporary vairables */
    u32 TexID;
    s32 InternalFormat  = TextureManager::getInternalFormat(TextureDataType);
    s32 ImageDataSize   = Dimension.Width * Dimension.Height * InternalFormat;
    
    register s32 x, y, i, j = 0;
    
    /* Register a new OpenGL texture */
    glGenTextures(1, &TexID);
    
    /* Bind the texture */
    glBindTexture(GL_TEXTURE_2D, TexID);
    
    /* Allocate new image memory */
    u8* ImageData = new u8[ImageDataSize];
    
    /* Fill the image buffer */
    for (y = 0; y < Dimension.Height; ++y)
    {
        for (x = 0; x < Dimension.Width; ++x)
        {
            for (i = 0; i < InternalFormat; ++i, ++j)
                ImageData[j] = StdFillColor_[i];
        }
    }
    
    /* Set the texture parameter */
    defaultTextureConfig();
    createRendererTexture(
        TexGenMipMapping_, TEXTURE_2D, dim::vector3di(Dimension.Width, Dimension.Height, 1),
        (ETextureImageDataTypes)TextureDataType, ImageData
    );
    
    /* Unbind texturing */
    glBindTexture(GL_TEXTURE_2D, 0);
    
    /* Allocate the new texture */
    Texture* NewTexture = new OpenGLTexture(
        TexID, Filename, Dimension, ImageData, TexGenRenderMode_, TexGenFilterMode_,
        InternalFormat, TextureDataType, TexGenEdgeSmooth_, TexGenMipMapping_
    );
    
    /* Add the texture to the texture list */
    TextureList_.push_back(NewTexture);
    
    /* Return the texture & exit the function */
    return NewTexture;
}

Texture* OpenGLDriver::createTexture(dim::size2di Dimension, u8* ImageData, u32 TextureDataType, io::stringc Filename)
{
    /* Temporary vairables */
    u32 TexID;
    s32 InternalFormat = TextureManager::getInternalFormat(TextureDataType);
    
    /* Register a new OpenGL texture */
    glGenTextures(1, &TexID);
    
    /* Bind the texture */
    glBindTexture(GL_TEXTURE_2D, TexID);
    
    /* Set the texture parameter */
    defaultTextureConfig();
    createRendererTexture(
        TexGenMipMapping_, TEXTURE_2D, dim::vector3di(Dimension.Width, Dimension.Height, 1),
        (ETextureImageDataTypes)TextureDataType, ImageData
    );
    
    /* Unbind texturing */
    glBindTexture(GL_TEXTURE_2D, 0);
    
    /* Allocate & return the new texture */
    Texture* NewTexture = new OpenGLTexture(
        TexID, Filename, Dimension, ImageData, TexGenRenderMode_, TexGenFilterMode_,
        InternalFormat, TextureDataType, TexGenEdgeSmooth_, TexGenMipMapping_
    );
    
    /* Add the texture to the texture list */
    TextureList_.push_back(NewTexture);
    
    /* Return the texture & exit the function */
    return NewTexture;
}

Texture* OpenGLDriver::createScreenShot(dim::point2di Position, dim::size2di Size)
{
    /* Temporary vairables */
    u32 TexID;
    
    /* Use the standard dimension */
    if (Size == dim::size2di())
        Size = dim::size2di(__Screen_Width, __Screen_Height);
    
    /* Get the current viewport */
    pushViewport();
    
    /* Set the new viewport */
    setViewport(Position.X, Position.Y, Size.Width, Size.Height);
    
    /* Register a new OpenGL texture */
    glGenTextures(1, &TexID);
    
    /* Bind the texture */
    glBindTexture(GL_TEXTURE_2D, TexID);
    
    /* Set the texture parameter */
    setTextureParameters(
        TEXTURE_2D, TexGenRenderMode_, TexGenFilterMode_, TexGenAnisotropy_, false, TexGenEdgeSmooth_
    );
    
    /* Create OpenGL texture and copy the pixels into the texture */
    glCopyTexImage2D(
        GL_TEXTURE_2D, 0, GL_DEPTH_COMPONENT,
        Position.X, __Screen_Height - Size.Height - Position.Y,
        Size.Width, Size.Height, 0
    );
    
    /* Unbind texturing */
    glBindTexture(GL_TEXTURE_2D, 0);
    
    /* Set the old viewport back */
    popViewport();
    
    /* Allocate the new texture */
    Texture* NewTexture = new OpenGLTexture(
        TexID, "", Size, 0, TexGenRenderMode_, TexGenFilterMode_, 3, TEXTURE_RGB, false, false
    );
    
    /* Add the texture to the texture list */
    TextureList_.push_back(NewTexture);
    
    /* Return the texture & exit the function */
    return NewTexture;
}

void OpenGLDriver::createScreenShot(Texture* hTexture, dim::point2di Position)
{
    if (!hTexture || hTexture->getDimensionType() != TEXTURE_2D)
        return;
    
    /* Bind the texture */
    hTexture->bind();
    
    /* Get the OpenGL format */
    s32 GLFormat = hTexture->getType();
    checkFormat(GLFormat);
    
    /* Copy the pixels into the texture */
    glCopyTexImage2D(
        GL_TEXTURE_2D, 0, GLFormat,
        Position.X, __Screen_Height - hTexture->getSize().Height - Position.Y,
        hTexture->getSize().Width, hTexture->getSize().Height, 0
    );
    
    /* Undbing texturing */
    hTexture->unbind();
}

void OpenGLDriver::updateCubeMapDirectionTexture(ECubeMapDirections Direction, Texture* CubeMapTexture)
{
    /* Bind the texture */
    CubeMapTexture->bind();
    
    /* Copy the pixels into the texture */
    glCopyTexSubImage2D(
        GL_TEXTURE_CUBE_MAP_POSITIVE_X + (s32)Direction, 0, 0, 0,
        0, __Screen_Height - CubeMapTexture->getSize().Height / 6,
        CubeMapTexture->getSize().Width, CubeMapTexture->getSize().Height / 6
    );
    
    /* Undbing texturing */
    CubeMapTexture->unbind();
    
    /* Clear the old buffers */
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}


/*
 * ========== Font loading & drawing ==========
 */


#if defined(WIN32)

Font* OpenGLDriver::loadFont(
    io::stringc FontName, dim::size2di FontSize, s32 TypeFlag, s32 RotationAngle, bool isSymbolUsing,
    bool is3DSpaceUsing, f32 FontDepth, f32 DetailDeviation, EDrawingModeTypes Wireframe)
{
    /* Temporary variables */
    GLYPHMETRICSFLOAT gmf[256];
    HFONT hFont;
    HFONT hOldFont;
    
    bool isBold       = TypeFlag & FONT_BOLD;
    bool isItalic     = TypeFlag & FONT_ITALIC;
    bool isUnderlined = TypeFlag & FONT_UNDERLINED;
    bool isStrikeout  = TypeFlag & FONT_STRIKEOUT;
    
    s32 Space3DMode;
    
    switch (Wireframe)
    {
        case DRAW_SOLID:
            Space3DMode = WGL_FONT_POLYGONS; break;
        case DRAW_FRAME:
            Space3DMode = WGL_FONT_LINES; break;
    }
    
    /* Register 256 new OpenGL lists */
    u32 FontID = glGenLists(256);
    
    if (!FontSize.Height)
        FontSize.Height = DEF_FONT_SIZE;
    
    /* Create the formated font */
    hFont = CreateFont(
        FontSize.Height,                                // Height of font
        FontSize.Width,                                 // Width of font
        RotationAngle,                                  // Angle of escapement
        0,                                              // Orientation angle
        isBold ? FW_BOLD : 0,                           // Font weight
        isItalic,                                       // Italic
        isUnderlined,                                   // Underline
        isStrikeout,                                    // Strikeout
        isSymbolUsing ? SYMBOL_CHARSET : ANSI_CHARSET,  // Character set identifier
        OUT_TT_PRECIS,                                  // Output precision
        CLIP_DEFAULT_PRECIS,                            // Clipping precision
        ANTIALIASED_QUALITY,                            // Output quality
        FF_DONTCARE | DEFAULT_PITCH,                    // Family and pitch
        FontName.c_str()                                // Font name
    );
    
    hOldFont = (HFONT)SelectObject(hDC_, hFont);
    
    /* Create the bitmap font */
    if (is3DSpaceUsing)
        wglUseFontOutlines(hDC_, 0, 255, FontID, DetailDeviation, FontDepth, Space3DMode, gmf);
    else
        wglUseFontBitmaps(hDC_, 0, 255, FontID);
    
    SelectObject(hDC_, hOldFont);
    
    /* Create a new font */
    return new Font(FontID, FontName, FontSize, 0, (s32)hFont, is3DSpaceUsing);
}

#elif defined(LINUX)

#endif

Font* OpenGLDriver::loadFont(
    io::stringc FontName, s32 FontSize, s32 TypeFlag, s32 RotationAngle, bool isSymbolUsing,
    bool is3DSpaceUsing, f32 FontDepth, f32 DetailDeviation, EDrawingModeTypes Wireframe)
{
    return loadFont(
        FontName, dim::size2di(0, FontSize), TypeFlag, RotationAngle, isSymbolUsing,
        is3DSpaceUsing, FontDepth, DetailDeviation, Wireframe
    );
}

Font* OpenGLDriver::loadFont(Texture* hTextureFont, dim::size2di FontSize, dim::size2di ElementCount)
{
    u32 FontID = glGenLists(256);
    
    dim::size2df Size;
    Size.Width  = (f32)FontSize.Width  * f32(1.0f/(__Screen_Width /2));
    Size.Height = (f32)FontSize.Height * f32(1.0f/(__Screen_Height/2));
    
    glBindTexture(GL_TEXTURE_2D, hTextureFont->getID());
    glEnable(GL_TEXTURE_2D);
    
    const f32 u = (f32)ElementCount.Width / 256;
    const f32 v = (f32)ElementCount.Height / 256;
    f32 x, y;
    
    for (s32 i = 0; i < 256; ++i)
    {
        x = f32(i % ElementCount.Width) / ElementCount.Width;
        y = f32(i / ElementCount.Height) / ElementCount.Height;
        
        glNewList(FontID + i, GL_COMPILE);
        {
            glBegin(GL_QUADS);
            {
                glTexCoord2f(x, y - v*2);
                glVertex2f(0, 0);
                
                glTexCoord2f(x + u, y - v*2);
                glVertex2f(Size.Width, 0);
                
                glTexCoord2f(x + u, y - v);
                glVertex2f(Size.Width, -Size.Height);
                
                glTexCoord2f(x, y - v);
                glVertex2f(0, -Size.Height);
            }
            glEnd();
            glTranslatef(Size.Width, 0, 0);
        }
        glEndList();
    }
    
    return new Font(FontID, "", hTextureFont->getSize() / 16, hTextureFont, 0, false);
}

void OpenGLDriver::deleteFont(Font* hFont)
{
    /* Delete the OpenGL lists */
    glDeleteLists(hFont->getID(), 256);
    
    /* Delete the font */
    spDeletePointer(hFont);
}

void OpenGLDriver::lockTexture(Texture* hTexture)
{
    /* Activate object linear */
    glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
    glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_OBJECT_LINEAR);
    
    /* Binding texture & enabling */
    glBindTexture(GL_TEXTURE_2D, hTexture->getID());
    glEnable(GL_TEXTURE_2D);
    glEnable(GL_TEXTURE_GEN_S);
    glEnable(GL_TEXTURE_GEN_T);
}
void OpenGLDriver::unlockTexture()
{
    defaultTextureGenMode();
    
    /* Disbaling & unbinding texture */
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_TEXTURE_GEN_S);
    glDisable(GL_TEXTURE_GEN_T);
    glBindTexture(GL_TEXTURE_2D, 0);
}

void OpenGLDriver::draw2DText(Font* hFont, dim::point2di Position, io::stringc Text, color Color)
{
    const char* cStr = Text.c_str();
    
    glLoadIdentity();
    
    /* Coloring (before loacting raster position) */
    glColor4ub(Color.Red, Color.Green, Color.Blue, Color.Alpha);
    
    /* Check if the font is a textured font */
    if (hFont->getTexture())
    {
        /* Bind the texture */
        glBindTexture(GL_TEXTURE_2D, hFont->getTexture()->getID());
        glEnable(GL_TEXTURE_2D);
    }
    
    /* Locate raster position */
    glRasterPos2f(0.0f, 0.0f);
    
    /* Additional move for the raster position */
    glBitmap(
        0, 0, 0, 0,
        Position.X - __Screen_Width/2,                              // x-moved
        -Position.Y + __Screen_Height/2 - hFont->getSize().Height,  // y-moved
        0
    );
    
    /* Draw the text */
    glPushAttrib(GL_LIST_BIT);                              // Pushes The Display List Bits
    {
        glListBase(hFont->getID());                         // Sets The Base Character
        glCallLists(Text.size(), GL_UNSIGNED_BYTE, cStr);   // Draws The Display List Text
    }
    glPopAttrib();                                          // Pops The Display List Bits
    
    glDisable(GL_TEXTURE_2D);
}

void OpenGLDriver::draw3DText(
    Font* hFont, io::stringc Text, dim::vector3df Position, dim::vector3df Rotation,
    color Color, scene::Camera* hCamera)
{
    const char* cStr = Text.c_str();
    
    glLoadIdentity();
    
    /* Coloring (before loacting raster position) */
    glColor4ub(Color.Red, Color.Green, Color.Blue, Color.Alpha);
    
    /* Check if a camera is used */
    if (hCamera)
    {
        /* Update the modelview matrix of the camera */
        hCamera->updateModelviewMatrix();
        
        /* Update the matrix */
        updateModelviewMatrix();
    }
    
    /* Locate raster position */
    if (hFont->is3DSpaceUsing())
        glTranslatef(Position.X, Position.Y, Position.Z);
    else
        glRasterPos3f(Position.X, Position.Y, Position.Z);
    
    glRotatef(Rotation.Y, 0.0f, 1.0f, 0.0f);
    glRotatef(Rotation.X, 1.0f, 0.0f, 0.0f);
    glRotatef(Rotation.Z, 0.0f, 0.0f, 1.0f);
    
    /* Draw the text */
    glPushAttrib(GL_LIST_BIT);                              // Pushes The Display List Bits
    {
        glListBase(hFont->getID());                         // Sets The Base Character to 32
        glCallLists(Text.size(), GL_UNSIGNED_BYTE, cStr);   // Draws The Display List Text
    }
    glPopAttrib();                                          // Pops The Display List Bits
}


/*
 * ======= Matrix controll =======
 */

void OpenGLDriver::updateModelviewMatrix()
{
    dim::matrix4f ModelviewMatrix(scene::spViewMatrix * scene::spWorldMatrix);
    
    #if 0
    
    glMatrixMode(GL_MODELVIEW);
    glLoadMatrixf(ModelviewMatrix.getArray16());
    
    #else
    
    dim::matrix4f Mat;
    Mat.matrixReset();
    Mat[10] = -1;
    Mat *= ModelviewMatrix;
    
    glMatrixMode(GL_MODELVIEW);
    glLoadMatrixf(Mat.getArray16());
    
    #endif
}

void OpenGLDriver::setProjectionMatrix(const dim::matrix4f &Matrix)
{
    scene::spProjectionMatrix = Matrix;
    
    glMatrixMode(GL_PROJECTION);
    glLoadMatrixf(Matrix.getArray16());
}
void OpenGLDriver::setViewMatrix(const dim::matrix4f &Matrix)
{
    scene::spViewMatrix = Matrix;
    
    #if 0
    
    glMatrixMode(GL_MODELVIEW);
    glLoadMatrixf(Matrix.getArray16());
    
    #else
    
    dim::matrix4f Mat;
    Mat.matrixReset();
    Mat[10] = -1;
    Mat *= Matrix;
    glMatrixMode(GL_MODELVIEW);
    glLoadMatrixf(Mat.getArray16());
    
    #endif
    
    /* Update each clipping plane */
    for (std::vector<SClipPlane>::iterator it = ClippingPlanes_.begin(); it != ClippingPlanes_.end(); ++it)
    {
        if (it->Enable)
            glClipPlane(GL_CLIP_PLANE0 + it->Index, it->PlaneEquation);
    }
}
void OpenGLDriver::setWorldMatrix(const dim::matrix4f &Matrix)
{
    scene::spWorldMatrix = Matrix;
}
void OpenGLDriver::setTextureMatrix(const dim::matrix4f &Matrix, u8 TextureLayer)
{
    scene::spTextureMatrix[TextureLayer] = Matrix;
    
    glActiveTextureARB(GL_TEXTURE0 + TextureLayer);
    glClientActiveTextureARB(GL_TEXTURE0 + TextureLayer);
    
    glMatrixMode(GL_TEXTURE);
    glLoadMatrixf(Matrix.getArray16());
}
void OpenGLDriver::setColorMatrix(const dim::matrix4f &Matrix)
{
    scene::spColorMatrix = Matrix;
    
    glMatrixMode(GL_COLOR);
    glLoadMatrixf(Matrix.getArray16());
}


/*
 * ========== Other renderer option fuctions ==========
 */


void OpenGLDriver::setViewport(s32 X, s32 Y, s32 Width, s32 Height)
{
    glViewport(X, __Screen_Height - Height - Y, Width, Height);
}



/*
 * ========== Wrapping empty functions (when the OpenGL version is to low) ==========
 */

void WrapEmptyProc_glFogCoordPointer(GLenum type, GLsizei stride, const GLvoid* pointer)
{
}


/*
 * ========== Private functions ==========
 */


void OpenGLDriver::init()
{
    /* Shademodel: smooth */
    glShadeModel(GL_SMOOTH);
    
    /* Standard clear color */
    glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
    
    /* Init fog range */
    setFogRange(0.1f);
    
    /* General settings */
    MaxClippingPlanes_ = GL_MAX_CLIP_PLANES;
    ClippingPlanes_.resize(MaxClippingPlanes_);
}


#if defined(WIN32)

bool OpenGLDriver::isExtensionSupported(io::stringc TargetExtension) const
{
    if (TargetExtension.left(1).upper() == "W") // WGL extensions
    {
        
        const s32 extlen = TargetExtension.size();
        const c8* supported = 0;
        
        // Try to use wglGetExtensionStringARB on current device-context, if possible
        PROC wglGetExtString = wglGetProcAddress("wglGetExtensionsStringARB");
        
        if (wglGetExtString)
            supported = ((char*(__stdcall*)(HDC))wglGetExtString)(wglGetCurrentDC());
        
        // If that gailed, try standard OpenGL extensions string
        if (!supported)
            supported = (char*)glGetString(GL_EXTENSIONS);
        
        // If that failed too, must be no extensions supported
        if (!supported)
            return false;
        
        // Begin Examination At Start Of String, Increment By 1 On False Match
        for (const c8* p = supported;; p++)
        {
            p = strstr(p, TargetExtension.c_str());
            
            if (!p)
                return false;
            
            if ( ( p == supported || p[-1] == ' ' ) && ( p[extlen] == '\0' || p[extlen] == ' ' ) )
                return true;
        }
        
    }
    else // GL extensions
    {
        
        const u8* pszExtensions = 0;
        const u8* pszStart;
        c8* szTargetExtension = TargetExtension.cstr();
        u8* pszWhere, * pszTerminator;
        
        // Extension names should not have spaces
        pszWhere = (unsigned char*)strchr(szTargetExtension, ' ');
        if (pszWhere || *szTargetExtension == '\0')
            return false;
        
        // Get Extensions String
        pszExtensions = glGetString(GL_EXTENSIONS);
        
        // Search The Extensions String For An Exact Copy
        pszStart = pszExtensions;
        
        for (;;)
        {
            pszWhere = (unsigned char*)strstr((const char*)pszStart, szTargetExtension);
            
            if (!pszWhere)
                break;
            
            pszTerminator = pszWhere + strlen(szTargetExtension);
            if (pszWhere == pszStart || *(pszWhere-1) == ' ')
            {
                if (*pszTerminator == ' ' || *pszTerminator == '\0')
                    return true;
            }
            
            pszStart = pszTerminator;
        }
        
    }
    
    return false;
}

void OpenGLDriver::loadExtensions()
{
    
    #define LOADOPENGLPROC(o, t, n)                                 \
        o = (t)wglGetProcAddress(n);                                \
        if (!o)                                                     \
            io::printError(                                         \
                io::stringc("could not load OpenGL function \"")+   \
                io::stringc(n)+io::stringc("\"")                    \
            );
    
    /* Load OpenGL multi-texturing functions */
    
    if (supportMultitex())
    {
        LOADOPENGLPROC(glActiveTextureARB,          PFNGLACTIVETEXTUREARBPROC,          "glActiveTextureARB"        );
        LOADOPENGLPROC(glMultiTexCoord2fARB,        PFNGLMULTITEXCOORD2FARBPROC,        "glMultiTexCoord2fARB"      );
        LOADOPENGLPROC(glClientActiveTextureARB,    PFNGLCLIENTACTIVETEXTUREARBPROC,    "glClientActiveTextureARB"  );
    }
    else
        io::printError("multi-texturing is not supported");
    
    /* Load OpenGL vertex-buffer-object (VBO) functions */
    
    if (supportVBO())
    {
        LOADOPENGLPROC(glGenBuffersARB,     PFNGLGENBUFFERSARBPROC,     "glGenBuffersARB"   );
        LOADOPENGLPROC(glBindBufferARB,     PFNGLBINDBUFFERARBPROC,     "glBindBufferARB"   );
        LOADOPENGLPROC(glBufferDataARB,     PFNGLBUFFERDATAARBPROC,     "glBufferDataARB"   );
        LOADOPENGLPROC(glDeleteBuffersARB,  PFNGLDELETEBUFFERSARBPROC,  "glDeleteBuffersARB");
    }
    else
        io::printError("vertex-buffer-objects (VBOs) are not supported");
    
    /* Load OpenGL shader (GLSL) functions */
    
    LOADOPENGLPROC(glBindProgramARB,                PFNGLBINDPROGRAMARBPROC,                "glBindProgramARB"              );
    LOADOPENGLPROC(glProgramStringARB,              PFNGLPROGRAMSTRINGARBPROC,              "glProgramStringARB"            );
    LOADOPENGLPROC(glGenProgramsARB,                PFNGLGENPROGRAMSARBPROC,                "glGenProgramsARB"              );
    LOADOPENGLPROC(glDeleteProgramsARB,             PFNGLDELETEPROGRAMSARBPROC,             "glDeleteProgramsARB"           );
    LOADOPENGLPROC(glProgramLocalParameter4fvARB,   PFNGLPROGRAMLOCALPARAMETER4FVARBPROC,   "glProgramLocalParameter4fvARB" );
    
    LOADOPENGLPROC(glDeleteProgram,             PFNGLDELETEPROGRAMPROC,             "glDeleteProgram"           );
    LOADOPENGLPROC(glDeleteShader,              PFNGLDELETESHADERPROC,              "glDeleteShader"            );
    LOADOPENGLPROC(glCreateProgramObjectARB,    PFNGLCREATEPROGRAMOBJECTARBPROC,    "glCreateProgramObjectARB"  );
    LOADOPENGLPROC(glCreateShaderObjectARB,     PFNGLCREATESHADEROBJECTARBPROC,     "glCreateShaderObjectARB"   );
    LOADOPENGLPROC(glShaderSourceARB,           PFNGLSHADERSOURCEARBPROC,           "glShaderSourceARB"         );
    LOADOPENGLPROC(glCompileShaderARB,          PFNGLCOMPILESHADERARBPROC,          "glCompileShaderARB"        );
    LOADOPENGLPROC(glAttachObjectARB,           PFNGLATTACHOBJECTARBPROC,           "glAttachObjectARB"         );
    LOADOPENGLPROC(glDeleteObjectARB,           PFNGLDELETEOBJECTARBPROC,           "glDeleteObjectARB"         );
    LOADOPENGLPROC(glLinkProgramARB,            PFNGLLINKPROGRAMARBPROC,            "glLinkProgramARB"          );
    LOADOPENGLPROC(glUseProgramObjectARB,       PFNGLUSEPROGRAMOBJECTARBPROC,       "glUseProgramObjectARB"     );
    LOADOPENGLPROC(glGetObjectParameterivARB,   PFNGLGETOBJECTPARAMETERIVARBPROC,   "glGetObjectParameterivARB" );
    LOADOPENGLPROC(glGetInfoLogARB,             PFNGLGETINFOLOGARBPROC,             "glGetInfoLogARB"           );
    LOADOPENGLPROC(glDetachObjectARB,           PFNGLDETACHOBJECTARBPROC,           "glDetachObjectARB"         );
    LOADOPENGLPROC(glGetActiveUniformARB,       PFNGLGETACTIVEUNIFORMARBPROC,       "glGetActiveUniformARB"     );
    LOADOPENGLPROC(glUniform1fARB,              PFNGLUNIFORM1FARBPROC,              "glUniform1fARB"            );
    LOADOPENGLPROC(glUniform3fARB,              PFNGLUNIFORM3FARBPROC,              "glUniform3fARB"            );
    LOADOPENGLPROC(glUniform4fARB,              PFNGLUNIFORM4FARBPROC,              "glUniform4fARB"            );
    LOADOPENGLPROC(glUniform1iARB,              PFNGLUNIFORM1IARBPROC,              "glUniform1iARB"            );
    LOADOPENGLPROC(glUniform1ivARB,             PFNGLUNIFORM1IVARBPROC,             "glUniform1ivARB"           );
    LOADOPENGLPROC(glUniform1fvARB,             PFNGLUNIFORM1FVARBPROC,             "glUniform1fvARB"           );
    LOADOPENGLPROC(glUniform2fvARB,             PFNGLUNIFORM2FVARBPROC,             "glUniform2fvARB"           );
    LOADOPENGLPROC(glUniform3fvARB,             PFNGLUNIFORM3FVARBPROC,             "glUniform3fvARB"           );
    LOADOPENGLPROC(glUniform4fvARB,             PFNGLUNIFORM4FVARBPROC,             "glUniform4fvARB"           );
    LOADOPENGLPROC(glUniformMatrix2fvARB,       PFNGLUNIFORMMATRIX2FVARBPROC,       "glUniformMatrix2fvARB"     );
    LOADOPENGLPROC(glUniformMatrix3fvARB,       PFNGLUNIFORMMATRIX3FVARBPROC,       "glUniformMatrix3fvARB"     );
    LOADOPENGLPROC(glUniformMatrix4fvARB,       PFNGLUNIFORMMATRIX4FVARBPROC,       "glUniformMatrix4fvARB"     );
    
    /* Load OpenGL extendet functions */
    
    LOADOPENGLPROC(glFogCoordPointer,       PFNGLFOGCOORDPOINTERPROC,       "glFogCoordPointer"     );
    LOADOPENGLPROC(wglSwapIntervalEXT,      PFNWGLSWAPINTERVALFARPROC,      "wglSwapIntervalEXT"    );
    //LOADOPENGLPROC(glActiveStencilFaceEXT,  PFNGLACTIVESTENCILFACEEXTPROC,  "glActiveStencilFaceEXT");
    LOADOPENGLPROC(glTexImage3DEXT,         PFNGLTEXIMAGE3DEXTPROC,         "glTexImage3DEXT"       );
    
    /* Special cases operations */
    
    if (!glFogCoordPointer)
        glFogCoordPointer = (PFNGLFOGCOORDPOINTERPROC)WrapEmptyProc_glFogCoordPointer;
    
    #undef LOADOPENGLPROC
    
}

#elif defined(LINUX)

#endif


/* ========== Drawing 2D - private functions ========== */

void OpenGLDriver::draw2DPixelFast(s32 X, s32 Y)
{
    glVertex2i(X, Y);
}

void OpenGLDriver::draw2DLineFast(f32 X1, f32 Y1, f32 X2, f32 Y2)
{
    glVertex2f(X1, Y1);
    glVertex2f(X2, Y2);
}


/*
 * ========== Texture loading & creating - private functions ==========
 */

void OpenGLDriver::setTextureParameters(
    ETextureDimensions Dimension, s32 RenderMode, s32 FilterMode, f32 MaxAnisotropy, bool isMipMapping, bool isEdgeSmooth)
{
    GLenum TexDim = getGLTexDim(Dimension);
    
    /* Clamp to edges */
    if (isEdgeSmooth)
    {
        glTexParameteri(TexDim, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
        glTexParameteri(TexDim, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    }
    
    /* MIP-mapping */
    glTexParameteri(TexDim, GL_GENERATE_MIPMAP, isMipMapping ? GL_TRUE : GL_FALSE);
    
    /* Anisotropy */
    if (FilterMode == TEXTURE_ANISOTROPIC)
        glTexParameterf(TexDim, GL_TEXTURE_MAX_ANISOTROPY_EXT, MaxAnisotropy);
    
    /* Render mode */
    if (RenderMode == TEXTURE_SMOOTH)
    {
        if (isMipMapping)
        {
            glTexParameteri(
                TexDim, GL_TEXTURE_MIN_FILTER,
                FilterMode == TEXTURE_BILINEAR ? GL_LINEAR_MIPMAP_NEAREST : GL_LINEAR_MIPMAP_LINEAR
            );
        }
        else
            glTexParameteri(TexDim, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
        
        glTexParameteri(TexDim, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    }
    else if (RenderMode == TEXTURE_LINEAR)
    {
        if (isMipMapping)
        {
            glTexParameteri(
                TexDim, GL_TEXTURE_MIN_FILTER,
                FilterMode == TEXTURE_BILINEAR ? GL_NEAREST_MIPMAP_NEAREST : GL_NEAREST_MIPMAP_LINEAR
            );
        }
        else
            glTexParameteri(TexDim, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
        
        glTexParameteri(TexDim, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
    }
}

void OpenGLDriver::defaultTextureGenMode()
{
    for (s32 i = 0; i < MultiTextureCount_; ++i)
    {
        /* Active current texture layer */
        glActiveTextureARB(GL_TEXTURE0 + i);
        
        /* Default sphete mapping generation */
        glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
        glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
    }
}

void OpenGLDriver::defaultTextureConfig()
{
    /* Selecting the texture configuration */
    setTextureParameters(
        TEXTURE_2D, TexGenRenderMode_, TexGenFilterMode_, TexGenAnisotropy_, TexGenMipMapping_, TexGenEdgeSmooth_
    );
}

void OpenGLDriver::createRendererTexture(
    bool MipMaps, ETextureDimensions Dimension, dim::vector3di Size, ETextureImageDataTypes Format, const u8* ImageData)
{
    /* First settings */
    s32 GLFormat = Format;
    checkFormat(GLFormat);
    
    s32 InternalFormat = GLFormat;
    checkInternalFormat(InternalFormat);
    
    #if 0
    
    /* Texture type creation (with/ without mipmaps) */
    if (!MipMaps)
    {
        Size.Y /= Size.Z;
        createRendererTextureNormal(Dimension, Size, InternalFormat, GLFormat, ImageData);
    }
    else
        createRendererTextureMipmaps(Dimension, Size, InternalFormat, GLFormat, ImageData);
    
    #else
    
    Size.Y /= Size.Z;
    
    /* Texture type creation */
    createRendererTextureNormal(Dimension, Size, InternalFormat, GLFormat, ImageData);
    
    #endif
}

void OpenGLDriver::createRendererTextureNormal(
    ETextureDimensions Dimension, dim::vector3di Size, s32 InternalFormat, GLenum GLFormat, const u8* ImageData, s32 Level)
{
    switch (Dimension)
    {
        case TEXTURE_1D:
        {
            glTexImage1D(
                GL_TEXTURE_1D, Level, InternalFormat, Size.X, 0, GLFormat, GL_UNSIGNED_BYTE, ImageData
            );
        }
        break;
        
        case TEXTURE_2D:
        {
            glTexImage2D(
                GL_TEXTURE_2D, Level, InternalFormat, Size.X, Size.Y, 0, GLFormat, GL_UNSIGNED_BYTE, ImageData
            );
        }
        break;
        
        case TEXTURE_3D:
        {
            glTexImage3DEXT(
                GL_TEXTURE_3D, Level, InternalFormat, Size.X, Size.Y, Size.Z, 0, GLFormat, GL_UNSIGNED_BYTE, ImageData
            );
        }
        break;
        
        case TEXTURE_CUBEMAP:
        {
            GLenum pos;
            const s32 OffsetSize = Size.X*Size.Y*InternalFormat;
            
            /* Reflection map attributes */
            glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
            glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
            glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);
            
            for (s32 i = 0; i < 6; ++i)
            {
                switch (i)
                {
                    case 0: pos = GL_TEXTURE_CUBE_MAP_POSITIVE_X; break;
                    case 1: pos = GL_TEXTURE_CUBE_MAP_NEGATIVE_X; break;
                    case 2: pos = GL_TEXTURE_CUBE_MAP_POSITIVE_Y; break;
                    case 3: pos = GL_TEXTURE_CUBE_MAP_NEGATIVE_Y; break;
                    case 4: pos = GL_TEXTURE_CUBE_MAP_POSITIVE_Z; break;
                    case 5: pos = GL_TEXTURE_CUBE_MAP_NEGATIVE_Z; break;
                }
                
                glTexImage2D(
                    pos, Level, InternalFormat, Size.X, Size.Y, 0, GLFormat, GL_UNSIGNED_BYTE,
                    (const u8*)(ImageData + i*OffsetSize)
                );
            }
        }
        break;
    }
}

void OpenGLDriver::createRendererTextureMipmaps(
    ETextureDimensions Dimension, dim::vector3di Size, s32 InternalFormat, GLenum GLFormat, const u8* ImageData)
{
    s32 Width   = Size.X;
    s32 Height  = Size.Y;
    
    /* Temporary memories */
    s32 i                   = 0;
    u8* pData               = 0;
    const s32 ImageDataSize = Width*Height*4;//InternalFormat;
    
    /* Allocate new image data memory */
    pData = new u8[ImageDataSize];
    
    /* Copy the image data */
    #if 1
    switch (InternalFormat)
    {
        case 1:
        {
            for (s32 i = 0; i < Width*Height; ++i)
            {
                pData[i*4+0] = ImageData[i];
                pData[i*4+1] = ImageData[i];
                pData[i*4+2] = ImageData[i];
                pData[i*4+3] = 255;
            }
        }
        break;
        
        case 2:
        {
            for (s32 i = 0; i < Width*Height; ++i)
            {
                pData[i*4+0] = ImageData[(i << 1)+0];
                pData[i*4+1] = ImageData[(i << 1)+0];
                pData[i*4+2] = ImageData[(i << 1)+0];
                pData[i*4+3] = ImageData[(i << 1)+1];
            }
        }
        break;
        
        case 3:
        {
            for (s32 i = 0; i < Width*Height; ++i)
            {
                pData[i*4+0] = ImageData[i*3+0];
                pData[i*4+1] = ImageData[i*3+1];
                pData[i*4+2] = ImageData[i*3+2];
                pData[i*4+3] = 255;
            }
        }
        break;
        
        case 4:
        {
            for (s32 i = 0; i < ImageDataSize; ++i)
                pData[i] = ImageData[i];
        }
        break;
    }
    InternalFormat = 4;
    GLFormat = GL_RGBA;
    #else
    memcpy(pData, ImageData, ImageDataSize);
    #endif
    
    /* Get the optimized size & scale if the input size is not correct */
    dim::size2di Size2D(Size.X, Size.Y);
    TextureManager::checkImageSize(Size2D);
    Size.X = Size2D.Width, Size.Y = Size2D.Height;
    
    /* Get the maximal size */
    GLint MaxSize;
    glGetIntegerv(GL_MAX_TEXTURE_SIZE, &MaxSize);
    
    /* Check the size of maxima */
    if (Size.X > MaxSize) Size.X = MaxSize;
    if (Size.Y > MaxSize) Size.Y = MaxSize;
    
    /* Clamp the image data size, must be power of to (2^n) */
    if (Size.X != Width || Size.Y != Height)
        TextureManager::scaleImageData(pData, Width, Height, Size.X, Size.Y, InternalFormat);
    
    Size.Y /= Size.Z;
    
    /* Loop for all mipmap levels */
    while (1)
    {
        /* Build a new mipmap level */
        createRendererTextureNormal(Dimension, Size, InternalFormat, GLFormat, pData, i++);
        
        /* Check if the last mipmap level has been achieved */
        if (Size.X == 1 && Size.Y == 1 && Size.Z == 1)
            break;
        
        /* Scale the image data to the half size */
        TextureManager::scaleHalfImageData(pData, Size.X, Size.Y * Size.Z, InternalFormat);
        
        /* Half the size */
        if (Size.X > 1) Size.X >>= 1;
        if (Size.Y > 1) Size.Y >>= 1;
        if (Size.Z > 1) Size.Z >>= 1;
    }
    
    /* Delete temporary memory */
    spDeleteArray(pData);
}


void OpenGLDriver::checkInternalFormat(s32 &InternalFormat)
{
    /* Check if the correction is not needed */
    if (InternalFormat >= 1 && InternalFormat <= 4)
        return;
    
    /* Change the value */
    switch (InternalFormat)
    {
        case GL_RGBA:
            InternalFormat = 4; break;
        case GL_RGB:
            InternalFormat = 3; break;
        case GL_LUMINANCE_ALPHA:
            InternalFormat = 2; break;
        default:
            InternalFormat = 1; break;
    }
}

void OpenGLDriver::checkFormat(s32 &Format)
{
    /* Check if the correction is not needed */
    if (Format < TEXTURE_INDEX || Format > TEXTURE_RGBA)
        return;
    
    /* Change the value */
    switch (Format)
    {
        case TEXTURE_INDEX:
            Format = GL_COLOR_INDEX; break;
        case TEXTURE_STENCIL:
            Format = GL_STENCIL_INDEX; break;
        case TEXTURE_DEPTH:
            Format = GL_DEPTH_COMPONENT; break;
        case TEXTURE_RED:
            Format = GL_RED; break;
        case TEXTURE_GREEN:
            Format = GL_GREEN; break;
        case TEXTURE_BLUE:
            Format = GL_BLUE; break;
        case TEXTURE_ALPHA:
            Format = GL_ALPHA; break;
        case TEXTURE_GRAY:
            Format = GL_LUMINANCE; break;
        case TEXTURE_GRAYALPHA:
            Format = GL_LUMINANCE_ALPHA; break;
        case TEXTURE_RGB:
            Format = GL_RGB; break;
        case TEXTURE_RGBA:
            Format = GL_RGBA; break;
    }
}


} // /namespace video

} // /namespace sp






// ================================================================================
