/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * OpenGL Shading Language (GLSL) file
 *
 */

#include "spOpenGLShader.hpp"
#include "spOpenGLFunctionsARB.hpp"


namespace sp
{
namespace video
{


OpenGLShader::OpenGLShader()
    : Shader(), ProgramObject_(0), VertexShaderObject_(0), FragmentShaderObject_(0)
{
}

OpenGLShader::OpenGLShader(EStandardShaderTypes Type)
    : Shader(), ProgramObject_(0), VertexShaderObject_(0), FragmentShaderObject_(0)
{
    /* Allocate temporary varibales */
    u32 LinesCountVert, LinesCountFrag;
    c8 ** pDataVert, ** pDataFrag;
    
    HighLevelLanguage_ = true;
    
    /* Fill the data */
    if (!switchShaderTypes(Type, pDataVert, pDataFrag, LinesCountVert, LinesCountFrag))
        return;
    
    /* Initialize the shader */
    createShader(pDataVert, pDataFrag, LinesCountVert, LinesCountFrag);
}

OpenGLShader::OpenGLShader(io::stringc &FilenameVertex, io::stringc &FilenameFragment, bool UseHighLevel)
    : Shader(), ProgramObject_(0), VertexShaderObject_(0), FragmentShaderObject_(0)
{
    /* Allocate temporary varibales */
    u32 LinesCountVert, LinesCountFrag;
    c8 ** pDataVert, ** pDataFrag;
    
    HighLevelLanguage_ = UseHighLevel;
    
    /* Read the files data */
    if (!loadShaderFile(FilenameVertex, pDataVert, LinesCountVert))
        return;
    if (!loadShaderFile(FilenameFragment, pDataFrag, LinesCountFrag))
        return;
    
    /* Initialize the shader */
    createShader(pDataVert, pDataFrag, LinesCountVert, LinesCountFrag);
}

OpenGLShader::~OpenGLShader()
{
    /* Delete the program & shader objects */
    if (HighLevelLanguage_)
    {
        glDeleteProgram(ProgramObject_);
        glDeleteShader(VertexShaderObject_);
        glDeleteShader(FragmentShaderObject_);
    }
    else
    {
        glDeleteProgramsARB(1, &VertexShader_);
        glDeleteProgramsARB(1, &PixelShader_);
    }
}

/* Binds & unbinds the shader */

void OpenGLShader::bind()
{
    if (pShaderCallbackProc_)
        pShaderCallbackProc_(this);
    
    if (HighLevelLanguage_)
    {
        glUseProgramObjectARB(ProgramObject_);
    }
    else
    {
        if (VertexShader_)
        {
            glBindProgramARB(GL_VERTEX_PROGRAM_ARB, VertexShader_);
            glEnable(GL_VERTEX_PROGRAM_ARB);
        }
        if (PixelShader_)
        {
            glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB, PixelShader_);
            glEnable(GL_FRAGMENT_PROGRAM_ARB);
        }
    }
}

void OpenGLShader::unbind()
{
    if (HighLevelLanguage_)
    {
        glUseProgramObjectARB(0);
    }
    else
    {
        if (VertexShader_)
            glDisable(GL_VERTEX_PROGRAM_ARB);
        if (PixelShader_)
            glDisable(GL_FRAGMENT_PROGRAM_ARB);
    }
}

/* Set the constants (by number) */

void OpenGLShader::setConstant(s32 Number, EConstantTypes Type, const f32 Value)
{
    setConstant(Number, Type, &Value, 1);
}

void OpenGLShader::setConstant(s32 Number, EConstantTypes Type, dim::vector3df Position)
{
    f32 pFloat[3] = { Position.X, Position.Y, -Position.Z };
    setConstant(Number, Type, pFloat, 3);
}

void OpenGLShader::setConstant(s32 Number, EConstantTypes Type, video::color Color)
{
    f32 pFloat[4] = { (f32)Color.Red / 255.0f, (f32)Color.Blue / 255.0f, (f32)Color.Green / 255.0f, (f32)Color.Alpha / 255.0f };
    setConstant(Number, Type, pFloat, 4);
}

void OpenGLShader::setConstant(s32 Number, EConstantTypes Type, dim::matrix4f &Matrix)
{
    setConstant(Number, Type, Matrix.getArray16(), 16);
}

void OpenGLShader::setConstant(s32 Number, EConstantTypes Type, const f32* pFloats, int Count)
{
    /* Use the current shader program */
    glUseProgramObjectARB(ProgramObject_);
    
    /* Check the correct constant type */
    GLenum UniformType = (GLenum)Type;
    checkConstantType(UniformType);
    
    /* Select the type */
    switch (UniformType)
    {
        case GL_FLOAT:
            glUniform1fvARB(Number, Count, pFloats);
            break;
        case GL_FLOAT_VEC2_ARB:
            glUniform2fvARB(Number, Count/2, pFloats);
            break;
        case GL_FLOAT_VEC3_ARB:
            glUniform3fvARB(Number, Count/3, pFloats);
            break;
        case GL_FLOAT_VEC4_ARB:
            glUniform4fvARB(Number, Count/4, pFloats);
            break;
        case GL_FLOAT_MAT2_ARB:
            glUniformMatrix2fvARB(Number, Count/4, false, pFloats);
            break;
        case GL_FLOAT_MAT3_ARB:
            glUniformMatrix3fvARB(Number, Count/9, false, pFloats);
            break;
        case GL_FLOAT_MAT4_ARB:
            glUniformMatrix4fvARB(Number, Count/16, false, pFloats);
            break;
        default:
            glUniform1ivARB(Number, Count, (GLint*)pFloats);
            break;
    }
    
    /* Unbind the current shader program */
    glUseProgramObjectARB(0);
}

/* Set the constants (by name) */

bool OpenGLShader::setConstant(io::stringc Name, const f32 Value)
{
    s32 i, c = (s32)UniformList_.size();
    
    // Loop all uniforms
    for (i = 0; i < c; ++i)
    {
        if (UniformList_[i].Name == Name)
            break;
    }
    
    if (i == c)
        return false;
    
    /* Use the current shader program */
    glUseProgramObjectARB(ProgramObject_);
    
    /* Select the type */
    if (UniformList_[i].Type == GL_FLOAT)
        glUniform1fARB(i, Value);
    else
        glUniform1iARB(i, (GLint)Value);
    
    /* Unbind the current shader program */
    glUseProgramObjectARB(0);
    
    return true;
}

bool OpenGLShader::setConstant(io::stringc Name, dim::vector3df Position)
{
    s32 i, c = (s32)UniformList_.size();
    
    /* Loop all uniforms */
    for (i = 0; i < c; ++i)
    {
        if (UniformList_[i].Name == Name)
            break;
    }
    
    if (i == c)
        return false;
    
    /* Use the current shader program */
    glUseProgramObjectARB(ProgramObject_);
    
    /* Select the type */
    if (UniformList_[i].Type == GL_FLOAT_VEC3_ARB)
        glUniform3fARB(i, Position.X, Position.Y, -Position.Z);
    else if (UniformList_[i].Type == GL_FLOAT_VEC4_ARB)
        glUniform4fARB(i, Position.X, Position.Y, -Position.Z, 1.0f);
    
    /* Unbind the current shader program */
    glUseProgramObjectARB(0);
    
    return true;
}

bool OpenGLShader::setConstant(io::stringc Name, video::color Color)
{
    s32 i, c = (s32)UniformList_.size();
    
    /* Loop all uniforms */
    for (i = 0; i < c; ++i)
    {
        if (UniformList_[i].Name == Name)
            break;
    }
    
    if (i == c)
        return false;
    
    /* Use the current shader program */
    glUseProgramObjectARB(ProgramObject_);
    
    /* Select the type */
    if (UniformList_[i].Type == GL_FLOAT_VEC3_ARB)
        glUniform3fARB(i, (f32)Color.Red / 255.0f, (f32)Color.Green / 255.0f, (f32)Color.Blue / 255.0f);
    else if (UniformList_[i].Type == GL_FLOAT_VEC4_ARB)
        glUniform4fARB(i, (f32)Color.Red / 255.0f, (f32)Color.Green / 255.0f, (f32)Color.Blue / 255.0f, (f32)Color.Alpha / 255.0f);
    
    /* Unbind the current shader program */
    glUseProgramObjectARB(0);
    
    return true;
}

bool OpenGLShader::setConstant(io::stringc Name, dim::matrix4f &Matrix)
{
    s32 i, c = (s32)UniformList_.size();
    
    /* Loop all uniforms */
    for (i = 0; i < c; ++i)
    {
        if (UniformList_[i].Name == Name)
            break;
    }
    
    if (i == c)
        return false;
    
    /* Use the current shader program */
    glUseProgramObjectARB(ProgramObject_);
    
    /* Select the type */
    if (UniformList_[i].Type == GL_FLOAT_MAT4_ARB)
        glUniformMatrix4fvARB(i, 1, false, Matrix.getArray16());
    
    /* Unbind the current shader program */
    glUseProgramObjectARB(0);
    
    return true;
}

bool OpenGLShader::setConstant(io::stringc Name, const f32* pFloats, int Count)
{
    s32 i, c = (s32)UniformList_.size();
    
    /* Loop all uniforms */
    for (i = 0; i < c; ++i)
    {
        if (UniformList_[i].Name == Name)
            break;
    }
    
    if (i == c)
        return false;
    
    /* Use the current shader program */
    glUseProgramObjectARB(ProgramObject_);
    
    /* Select the type */
    switch (UniformList_[i].Type)
    {
        case GL_FLOAT:
            glUniform1fvARB(i, Count, pFloats);
            break;
        case GL_FLOAT_VEC2_ARB:
            glUniform2fvARB(i, Count/2, pFloats);
            break;
        case GL_FLOAT_VEC3_ARB:
            glUniform3fvARB(i, Count/3, pFloats);
            break;
        case GL_FLOAT_VEC4_ARB:
            glUniform4fvARB(i, Count/4, pFloats);
            break;
        case GL_FLOAT_MAT2_ARB:
            glUniformMatrix2fvARB(i, Count/4, false, pFloats);
            break;
        case GL_FLOAT_MAT3_ARB:
            glUniformMatrix3fvARB(i, Count/9, false, pFloats);
            break;
        case GL_FLOAT_MAT4_ARB:
            glUniformMatrix4fvARB(i, Count/16, false, pFloats);
            break;
        default:
            glUniform1ivARB(i, Count, (GLint*)pFloats);
            break;
    }
    
    /* Unbind the current shader program */
    glUseProgramObjectARB(0);
    
    return true;
}


bool OpenGLShader::setConstant(const f32* pFloats, s32 StartRegister, s32 ConstAmount)
{
    s32 ShaderType = (CurrentShader_ == SHADER_VERTEX ? GL_VERTEX_PROGRAM_ARB : GL_FRAGMENT_PROGRAM_ARB);
    
    for (s32 i = 0; i < ConstAmount; ++i)
        glProgramLocalParameter4fvARB(ShaderType, StartRegister + i, &pFloats[i*4]);
}


/*
 * ========== Private functions: ==========
 */

bool OpenGLShader::loadShaderFile(io::stringc Filename, c8** &pData, u32 &LinesCount)
{
    spDeleteArray(pData);
    LinesCount = 0;
    
    io::FileLoader fLoader;
    io::File* hFile = fLoader.openFile(Filename, io::FILE_READ_BINARY);
    
    std::vector<io::stringc> TexList;
    
    if (hFile)
    {
        while (!fLoader.isEOF(hFile))
        {
            TexList.push_back( fLoader.readString(hFile) + "\n" );
            ++LinesCount;
        }
        
        pData = new c8*[LinesCount];
        
        for (u32 i = 0; i < LinesCount; ++i)
        {
            pData[i] = new c8[TexList[i].size()+1];
            memcpy(pData[i], TexList[i].c_str(), TexList[i].size());
            pData[i][TexList[i].size()] = 0;
        }
        
        fLoader.closeFile(hFile);
        
        return true;
    }
    
    return false;
}


void OpenGLShader::createShader(c8** pDataVert, c8** pDataFrag, u32 LinesCountVert, u32 LinesCountFrag)
{
    if (HighLevelLanguage_)
    {
        createShaderGLSL(pDataVert, pDataFrag, LinesCountVert, LinesCountFrag);
    }
    else
    {
        
        
        
    }
}

void OpenGLShader::createShaderGLSL(c8** pDataVert, c8** pDataFrag, u32 LinesCountVert, u32 LinesCountFrag)
{
    /* Temporary memory */
    c8 * pDataSourceVert, * pDataSourceFrag;
    
    /* Create the program & shader objects */
    ProgramObject_          = glCreateProgramObjectARB();
    VertexShaderObject_     = glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB);
    FragmentShaderObject_   = glCreateShaderObjectARB(GL_FRAGMENT_SHADER_ARB);
    
    /* Convert the data into single strings */
    convertShaderText((const c8**)pDataVert, pDataSourceVert, LinesCountVert);
    convertShaderText((const c8**)pDataFrag, pDataSourceFrag, LinesCountFrag);
    
    /* Initialize the data */
    glShaderSourceARB(VertexShaderObject_, 1, (const c8**)&pDataSourceVert, 0);
    glShaderSourceARB(FragmentShaderObject_, 1, (const c8**)&pDataSourceFrag, 0);
    
    /* Compile shader */
    glCompileShaderARB(VertexShaderObject_);
    glCompileShaderARB(FragmentShaderObject_);
    
    /* Attach object */
    glAttachObjectARB(ProgramObject_, VertexShaderObject_);
    glAttachObjectARB(ProgramObject_, FragmentShaderObject_);
    
    /* Link program */
    glLinkProgramARB(ProgramObject_);
    
    /* Check for errors */
    checkForErrorsGLSL();
    
    /* Receive all uniforms */
    receiveAllUniforms();
    
    /* Free all memories */
    deleteShaderText(pDataVert, LinesCountVert);
    deleteShaderText(pDataFrag, LinesCountFrag);
}

void OpenGLShader::createShaderVertex(c8** pDataShd, u32 LinesCount)
{
    /* Convert the data into a single string */
    c8* pDataSourceShd;
    convertShaderText((const c8**)pDataShd, pDataSourceShd, LinesCount);
    
    /* Generate a new shader program */
    glGenProgramsARB(1, &VertexShader_);
    
    /* Bind the shader program */
    glBindProgramARB(GL_VERTEX_PROGRAM_ARB, VertexShader_);
    
    /* Set the assemly shader string */
    glProgramStringARB(GL_VERTEX_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB, strlen(pDataSourceShd), pDataSourceShd);
    
    /* Check for errors */
    if (checkForErrorsShader("Vertex"))
    {
        /* Delete the shader program */
        glDeleteProgramsARB(1, &VertexShader_);
        VertexShader_ = 0;
    }
    
    /* Free all memories */
    deleteShaderText(pDataShd, LinesCount);
}

void OpenGLShader::createShaderPixel(c8** pDataShd, u32 LinesCount)
{
    /* Convert the data into a single string */
    c8* pDataSourceShd;
    convertShaderText((const c8**)pDataShd, pDataSourceShd, LinesCount);
    
    /* Generate a new shader program */
    glGenProgramsARB(1, &VertexShader_);
    
    /* Bind the shader program */
    glBindProgramARB(GL_FRAGMENT_PROGRAM_ARB, PixelShader_);
    
    /* Set the assemly shader string */
    glProgramStringARB(GL_FRAGMENT_PROGRAM_ARB, GL_PROGRAM_FORMAT_ASCII_ARB, strlen(pDataSourceShd), pDataSourceShd);
    
    /* Check for errors */
    if (checkForErrorsShader("Fragment"))
    {
        /* Delete the shader program */
        glDeleteProgramsARB(1, &PixelShader_);
        VertexShader_ = 0;
    }
    
    /* Free all memories */
    deleteShaderText(pDataShd, LinesCount);
}


bool OpenGLShader::checkForErrorsGLSL()
{
    s32 LogLength = 0;
    
    // Get the error log information length
    glGetObjectParameterivARB(ProgramObject_, GL_OBJECT_INFO_LOG_LENGTH_ARB, &LogLength);
    
    if (LogLength > 1)
    {
        s32 CharsWritten = 0;
        c8* InfoLog = new c8[LogLength];
        
        // Get the error log information
        glGetInfoLogARB(ProgramObject_, LogLength, &CharsWritten, InfoLog);
        
        // Print the warning
        io::printInfo(io::stringc("\n")+io::stringc(InfoLog), io::INFO_WARNING);
        
        delete [] InfoLog;
        
        return true;
    }
    
    return false;
}

bool OpenGLShader::checkForErrorsShader(const io::stringc ShaderName)
{
    GLenum error = glGetError();
    
    if (error != GL_NO_ERROR)
    {
        s32 ErrorPos;
        glGetIntegerv(GL_PROGRAM_ERROR_POSITION_ARB, &ErrorPos);
        
        const c8* ErrorStr = (const c8*)glGetString(GL_PROGRAM_ERROR_STRING_ARB);
        
        io::printInfo(
            ShaderName + io::stringc(" shader ompilation failed at position ") + io::stringc(ErrorPos) +
            io::stringc(":\n") + io::stringc(ErrorStr),
            io::INFO_WARNING
        );
        
        return true;
    }
    
    return false;
}


void OpenGLShader::receiveAllUniforms()
{
    
    s32 Count = 0, MaxLen = 0;
    
    // Get the count of uniforms
    glGetObjectParameterivARB(ProgramObject_, GL_OBJECT_ACTIVE_UNIFORMS_ARB, &Count);
    
    if (Count == 0)
        return;
    
    // Get the maximal uniform length
    glGetObjectParameterivARB(ProgramObject_, GL_OBJECT_ACTIVE_UNIFORM_MAX_LENGTH_ARB, &MaxLen);
    
    if (MaxLen == 0)
        return;
    
    c8* pBuffer = new c8[MaxLen];
    
    UniformList_.clear();
    
    SUniformInfo UniformData;
    s32 Size;
    
    // Receive the uniform information
    for (s32 i = 0; i < Count; ++i)
    {
        // Clear the buffer
        memset(pBuffer, 0, MaxLen);
        
        // Get the active uniform
        glGetActiveUniformARB(ProgramObject_, i, MaxLen, 0, &Size, &UniformData.Type, (GLcharARB*)pBuffer);
        
        // Add the element to the list
        UniformData.Name = io::stringc(pBuffer);
        UniformList_.push_back(UniformData);
    }
    
    delete [] pBuffer;
}

/* Standard shader */

const c8* spStandardShaderVertex[] = 
{
    "void main(void)\n",
    "{\n",
    "gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;\n",
    "gl_TexCoord[0] = gl_MultiTexCoord0;\n",
    "}\n"
};

const c8* spStandardShaderFragment[] = 
{
    "uniform sampler2D Texture0;\n",
    "uniform float Value = 0.5;\n",
    "void main(void)\n",
    "{\n",
    "vec4 RGB = texture2D(Texture0, vec2(gl_TexCoord[0]));\n",
    "\n",
    "}\n"
};

bool OpenGLShader::switchShaderTypes(EStandardShaderTypes Type, char** &pDataVert, char** &pDataFrag, u32 &LinesCountVert, u32 &LinesCountFrag)
{
    /* General settings */
    LinesCountVert = 5;
    LinesCountFrag = 7;
    
    /* Get the lines count */
    switch (Type)
    {
        case SHADER_MONOCHROME:
            spStandardShaderFragment[5] = "if ((RGB.r + RGB.g + RGB.b)/3.0 >= Value) gl_FragColor = vec4(1.0, 1.0, 1.0, RGB.a); else gl_FragColor = vec4(0.0, 0.0, 0.0, RGB.a);";
            break;
            
        case SHADER_NEGATIVE:
            spStandardShaderFragment[5] = "gl_FragColor = vec4(1.0 - RGB.r, 1.0 - RGB.g, 1.0 - RGB.b, RGB.a);\n";
            break;
            
        case SHADER_GREY:
            spStandardShaderFragment[5] = "gl_FragColor = vec4((RGB.r + RGB.g + RGB.b)/3.0, (RGB.r + RGB.g + RGB.b)/3.0, (RGB.r + RGB.g + RGB.b)/3.0, RGB.a);\n";
            break;
            
        case SHADER_BGR:
            spStandardShaderFragment[5] = "gl_FragColor = vec4(RGB.b, RGB.g, RGB.r, RGB.a);\n";
            break;
            
        default:
            return false;
    }
    
    /* Text data */
    pDataVert = new char*[LinesCountVert];
    pDataFrag = new char*[LinesCountFrag];
    
    /* Text lines data */
    for (u32 i = 0, len; i < LinesCountVert; ++i)
    {
        len = strlen(spStandardShaderVertex[i])+1;
        pDataVert[i] = new char[len];
        memcpy(pDataVert[i], spStandardShaderVertex[i], sizeof(char)*len);
    }
    for (u32 i = 0, len; i < LinesCountFrag; ++i)
    {
        len = strlen(spStandardShaderFragment[i])+1;
        pDataFrag[i] = new char[len];
        memcpy(pDataFrag[i], spStandardShaderFragment[i], sizeof(char)*len);
    }
    
    return true;
}

void OpenGLShader::convertShaderText(const c8** pData, c8* &pDataSource, u32 LinesCount)
{
    /* Temporary varibales */
    s32 i, j, c, len;
    c = len = 0;
    
    /* Loop the original data for the first time */    
    for (i = 0; i < LinesCount; ++i)
        len += strlen(pData[i]);
    
    /* Allocate new memory */
    pDataSource = new c8[len];
    
    /* Convert copy the data */
    for (i = 0; i < LinesCount; ++i)
    {
        len = strlen(pData[i]);
        
        for (j = 0; j < len; ++j)
            pDataSource[c+j] = pData[i][j];
        c += len;
    }
    
    /* Define the end of line of the string */
    pDataSource[c] = 0;
}    

void OpenGLShader::deleteShaderText(c8** &pData, u32 LinesCount)
{
    for (s32 i = 0; i < LinesCount; ++i)
        spDeleteArray(pData[i]);
    spDeleteArray(pData);
}

void OpenGLShader::checkConstantType(GLenum &Type)
{
    switch (Type)
    {
        case CONSTANT_INT:
            Type = GL_INT; break;
        case CONSTANT_FLOAT:
            Type = GL_FLOAT; break;
        case CONSTANT_VECTOR2:
            Type = GL_FLOAT_VEC2_ARB; break;
        case CONSTANT_VECTOR3:
            Type = GL_FLOAT_VEC3_ARB; break;
        case CONSTANT_VECTOR4:
            Type = GL_FLOAT_VEC4_ARB; break;
        case CONSTANT_MATRIX2:
            Type = GL_FLOAT_MAT2_ARB; break;
        case CONSTANT_MATRIX3:
            Type = GL_FLOAT_MAT3_ARB; break;
        case CONSTANT_MATRIX4:
            Type = GL_FLOAT_MAT4_ARB; break;
    }
}


} // /namespace scene

} // /namespace sp







// ================================================================================
