/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * OpenGL Textures
 *
 */

#include "spOpenGLTexture.hpp"
#include "spTextureManager.hpp"
#include "spMainDevice.hpp"
#include "spOpenGLFunctionsARB.hpp"
#include "spOpenGLDriver.hpp"


namespace sp
{

extern video::VideoDriver* __spVideoDriver;

namespace video
{


OpenGLTexture::OpenGLTexture() : Texture()
{
}
OpenGLTexture::OpenGLTexture(
    u32 ID, io::stringc Filename, dim::size2di Size, u8* ImageData,
    s32 RenderMode, s32 FilterMode, s32 Format, s32 Type, bool isEdgeSmooth, bool isMipMapping)
    : Texture(ID, Filename, Size, ImageData, RenderMode, FilterMode, Format, Type, isEdgeSmooth, isMipMapping)
{
    /* Initialization */
    updateFormat();
    
    if (OriginalImageData_)
    {
        /* Update the image data */
        makeIdentityImageData();
        
        /* Update the original image data formats */
        OrigInternalFormat_ = Format;
        OrigType_           = Type;
    }
}
OpenGLTexture::~OpenGLTexture()
{
    clear();
}

void OpenGLTexture::changeAttributes(s32 RenderMode, s32 FilterMode, bool isEdgeSmooth)
{
    /* Check if the image is not empty */
    if (!ImageData_ || hasAnimation())
        return;
    
    /* Set the new render mode */
    RenderMode_     = RenderMode;
    
    /* Set the new render mode */
    FilterMode_     = FilterMode;
    
    /* Set the new edge type */
    isEdgeSmooth_   = isEdgeSmooth;
    
    /* Recreate the image */
    recreateImageData();
}

void OpenGLTexture::setColorIntensity(f32 Red, f32 Green, f32 Blue)
{
    /* Check if the image is not empty */
    if (!ImageData_ || hasAnimation())
        return;
    
    /* Use this texture */
    glBindTexture(Dimension_, ID_);
    
    /* Settings */
    glPixelTransferf(GL_RED_SCALE, Red);
    glPixelTransferf(GL_GREEN_SCALE, Green);
    glPixelTransferf(GL_BLUE_SCALE, Blue);
    
    /* Recreate the image */
    recreateImageData();
    
    /* Back setting */
    glPixelTransferf(GL_RED_SCALE, 1.0f);
    glPixelTransferf(GL_GREEN_SCALE, 1.0f);
    glPixelTransferf(GL_BLUE_SCALE, 1.0f);
}

void OpenGLTexture::flipColors()
{
    /* Check if the image is not empty */
    if (!ImageData_ || hasAnimation())
        return;
    
    /* Determine the image data size */
    s32 ImageDataSize = Size_.Width * Size_.Height * InternalFormat_;
    
    /* Flip the image data */
    for (s32 i = 0; i < ImageDataSize; ++i)
        ImageData_[i] = 255 - ImageData_[i];
    
    /* Recreate the image */
    recreateImageData();
}

void OpenGLTexture::setMipMapFilter(s32 MinFilter, s32 MagFilter, bool isMipMapping)
{
    /* Check if the image is not empty */
    if (!ImageData_ || hasAnimation())
        return;
    
    /* Clear the image data */
    makeEmptyImageData();
    
    /* Set new render modes */
    RenderMode_     = MagFilter;
    isMipMapping_   = isMipMapping;
    
    /* Update the correct format */
    updateFormat();
    
    /* Set the texture parameter */
    if (isMipMapping_)
    {
        glTexParameteri(
            Dimension_, GL_TEXTURE_MIN_FILTER,
            MinFilter == TEXTURE_SMOOTH ? GL_LINEAR_MIPMAP_LINEAR : GL_NEAREST_MIPMAP_NEAREST
        );
        glTexParameteri(
            Dimension_, GL_TEXTURE_MAG_FILTER,
            MagFilter == TEXTURE_SMOOTH ? GL_LINEAR_MIPMAP_LINEAR : GL_NEAREST
        );
    }
    else
    {
        glTexParameteri(
            Dimension_, GL_TEXTURE_MIN_FILTER,
            MinFilter == TEXTURE_SMOOTH ? GL_LINEAR : GL_NEAREST
        );
        glTexParameteri(
            Dimension_, GL_TEXTURE_MAG_FILTER,
            MagFilter == TEXTURE_SMOOTH ? GL_LINEAR : GL_NEAREST
        );
    }
    
    /* Create the OpenGL texture */
    static_cast<OpenGLDriver*>(__spVideoDriver)->createRendererTexture(
        isMipMapping_, DimensionType_, dim::vector3di(Size_.Width, Size_.Height, Depth_), (ETextureImageDataTypes)Type_, ImageData_
    );
}

void OpenGLTexture::updateImageData(EUpdateDataFunc UpdateDataFunc)
{
    /* Check if the image is not empty */
    if (ImageData_)
    {
        if (UpdateDataFunc == GET_IMAGEDATA)
            receiveImageData();
        else if (UpdateDataFunc == SET_IMAGEDATA)
            recreateImageData();
    }
}

void OpenGLTexture::bind(s32 Level)
{
    /* Set the texture level */
    glActiveTextureARB(GL_TEXTURE0 + Level);
    glClientActiveTextureARB(GL_TEXTURE0 + Level);
    
    /* Bind the texture */
    glBindTexture(Dimension_, ID_);
    
    /* Enable the texture */
    glEnable(Dimension_);
}

void OpenGLTexture::unbind(s32 Level)
{
    /* Set the texture level */
    glActiveTextureARB(GL_TEXTURE0 + Level);
    glClientActiveTextureARB(GL_TEXTURE0 + Level);
    
    /* Unbind the texture */
    glBindTexture(Dimension_, 0);
    
    /* Disable the texture */
    glDisable(Dimension_);
}


/*
 * ======= Private: =======
 */


void OpenGLTexture::updateFormat()
{
    /* Use the default 1 component */
    InternalFormat_ = 1;
    
    /* Determine the SoftPixel type -> and set the renderer format */
    switch (Type_)
    {
        /* 1 component */
        case TEXTURE_INDEX:
            Format_ = GL_COLOR_INDEX; break;
        case TEXTURE_STENCIL:
            Format_ = GL_STENCIL_INDEX; break;
        case TEXTURE_DEPTH:
            Format_ = GL_DEPTH_COMPONENT; break;
        case TEXTURE_RED:
            Format_ = GL_RED; break;
        case TEXTURE_GREEN:
            Format_ = GL_GREEN; break;
        case TEXTURE_BLUE:
            Format_ = GL_BLUE; break;
        case TEXTURE_ALPHA:
            Format_ = GL_ALPHA; break;
        case TEXTURE_GRAY:
            Format_ = GL_LUMINANCE; break;
            
        /* 2 components */
        case TEXTURE_GRAYALPHA:
            Format_         = GL_LUMINANCE_ALPHA;
            InternalFormat_ = 2;
            break;
            
        /* 3 components */
        case TEXTURE_RGB:
            Format_         = GL_RGB;
            InternalFormat_ = 3;
            break;
            
        /* 4 components */
        case TEXTURE_RGBA:
            Format_         = GL_RGBA;
            InternalFormat_ = 4;
            break;
    }
    
    /* Determine which dimension is used */
    switch (DimensionType_)
    {
        case TEXTURE_1D:
            Dimension_ = GL_TEXTURE_1D; break;
        case TEXTURE_2D:
            Dimension_ = GL_TEXTURE_2D; break;
        case TEXTURE_3D:
            Dimension_ = GL_TEXTURE_3D; break;
        case TEXTURE_CUBEMAP:
            Dimension_ = GL_TEXTURE_CUBE_MAP; break;
    }
}

void OpenGLTexture::makeEmptyImageData()
{
    /* Delete the OpenGL texture */
    clear();
    
    /* Register a new OpenGL texture */
    glGenTextures(1, &ID_);
    
    /* Bind the texture */
    glBindTexture(Dimension_, ID_);
}

void OpenGLTexture::receiveImageData()
{
    /* Delete the old memories */
    spDeleteArray(OriginalImageData_);
    spDeleteArray(ImageData_);
    
    /* Bind this texture */
    glBindTexture(Dimension_, ID_);
    
    /* Get the dimension */
    glGetTexLevelParameteriv(Dimension_, 0, GL_TEXTURE_WIDTH, &OriginalSize_.Width);
    glGetTexLevelParameteriv(Dimension_, 0, GL_TEXTURE_HEIGHT, &OriginalSize_.Height);
    
    /* Get the internal format */
    glGetTexLevelParameteriv(Dimension_, 0, GL_TEXTURE_INTERNAL_FORMAT, &InternalFormat_);
    
    /* Update the correct format */
    updateFormat();
    
    /* Update the second size */
    Size_ = OriginalSize_;
    
    /* Allocate new memory */
    OriginalImageData_  = new u8[ OriginalSize_.Width * OriginalSize_.Height * InternalFormat_ ];
    
    /* Get the image data */
    glGetTexImage(Dimension_, 0, Format_, GL_UNSIGNED_BYTE, OriginalImageData_);
    
    /* Update the current image data */
    makeIdentityImageData();
    
    /* Update the original image data formats */
    OrigInternalFormat_ = InternalFormat_;
    OrigType_           = Type_;
}

void OpenGLTexture::recreateImageData()
{
    /* Clear the image data */
    makeEmptyImageData();
    
    /* Set the texture parameter */
    static_cast<OpenGLDriver*>(__spVideoDriver)->setTextureParameters(
        DimensionType_, RenderMode_, FilterMode_, AnisotropicSamples_, isMipMapping_, isEdgeSmooth_
    );
    
    /* Update the correct format */
    updateFormat();
    
    /* Create the OpenGL texture */
    static_cast<OpenGLDriver*>(__spVideoDriver)->createRendererTexture(
        isMipMapping_, DimensionType_, dim::vector3di(Size_.Width, Size_.Height, Depth_), (ETextureImageDataTypes)Type_, ImageData_
    );
}

void OpenGLTexture::clear()
{
    /* Check if the texture is a registered OpenGL texture */
    if (glIsTexture(ID_))
    {
        /* Delete the old texture */
        glDeleteTextures(1, &ID_);
    }
}


} // /namespace video

} // /namespace sp







// ================================================================================
