/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * File for scene loading BSP-scene file format (Quake3 map)
 *
 */

#include "spSceneLoaderBSP3.hpp"
#include "spMainDevice.hpp"


namespace sp
{

extern video::VideoDriver* __spVideoDriver;
extern scene::SceneManager* __spSceneManager;

namespace scene
{


SceneLoaderBSP3::SceneLoaderBSP3()
{
}
SceneLoaderBSP3::~SceneLoaderBSP3()
{
}

scene::Entity* SceneLoaderBSP3::loadScene(SceneManager* SceneMngr, io::stringc &Filename, io::stringc &TexturePath, s32 Tesselation)
{
    
    /* First settings */
    Filename_       = Filename;
    TexturePath_    = TexturePath;
    Tesselation_    = Tesselation;
    SceneMngr_      = SceneMngr;
    
    /* Open a new file */
    hFile_          = hFileLoader_->openFile(Filename, io::FILE_READ_BINARY);
    
    /* Create an empty entity */
    hEntityBase_    = new Entity();
    
    /* Create an empty entity for transparent textures */
    hEntityTrans_   = __spSceneManager->createPivot();
    hEntityTrans_->getThisMaterial().isBlending = false;
    hEntityTrans_->setAlphaMode(video::CMPSIZE_GREATER, 0.5);
    hEntityTrans_->setRenderFace(video::RENDER_BOTH);
    hEntityTrans_->setParent(hEntityBase_);
    
    /* Check if the file has been opened successful */
    if (!hFile_)
        return hEntityBase_;
    
    /* Read the header */
    readHeader();
    
    /* Read the lumps */
    readLumps();
    
    /* Close the file */
    hFileLoader_->closeFile(hFile_);
    
    /* Build the model */
    buildModel();
    
    /* Return the entity */
    return hEntityBase_;
    
}

bool SceneLoaderBSP3::readHeader()
{
    
    /* Read the header */
    hFileLoader_->readStream(hFile_, &Header_, sizeof(Header_));
    
    /* Check the magic identity */
    c8 Magic[5] = { 0 };
    memcpy(Magic, Header_.Magic, 4);
    
    if (io::stringc(Magic) != "IBSP")
    {
        /* Print an error message */
        io::printError("BSP (Quake III Arena) file has incorrect magic identity");
        
        /* Exit the function and return false for a failure */
        return false;
    }
    
    /* Check the version */
    if (Header_.Version != 0x2E)
    {
        /* Print an error message */
        io::printError("BSP (Quake III Arena) file has incorrect version number");
        
        /* Exit the function and return false for a failure */
        return false;
    }
    
    /* Exit the function and return true for success */
    return true;
    
}

void SceneLoaderBSP3::readLumps()
{
    
    /* Read all lumps */
    readLumpEntities();
    readLumpTextures();
    readLumpPlanes();
    readLumpNodes();
    readLumpLeafs();
    readLumpLeafFaces();
    readLumpLeafBrushes();
    readLumpModels();
    readLumpBrushes();
    readLumpBrushSides();
    readLumpVertexes();
    readLumpMeshVerts();
    readLumpEffecs();
    readLumpFaces();
    readLumpLightMaps();
    readLumpLightVols();
    readLumpVisData();
    
}

void SceneLoaderBSP3::readLumpEntities()
{
    
    /* Temporary vairables */
    std::vector<io::stringc> ScriptData;
    
    /* Set the offset */
    hFileLoader_->setFileSeek(hFile_, Header_.DirEntries[BSP_LUMP_ENTITIES].Offset);
    
    /* Allocate temporary memory */
    c8* EntDesc = new c8[ Header_.DirEntries[BSP_LUMP_ENTITIES].Length + 1 ];
    
    /* Read the entity descriptions */
    hFileLoader_->readStream(hFile_, EntDesc, Header_.DirEntries[BSP_LUMP_ENTITIES].Length);
    
    /* Set the end of the string */
    EntDesc[Header_.DirEntries[BSP_LUMP_ENTITIES].Length] = 0;
    
    /* Create a script */
    BSPLoaderExtensions::createScript(ScriptData, EntDesc);
    
    /* Examine the script */
    examineScript(ScriptData);
    
    /* Delete the temporary memory */
    spDeleteArray(EntDesc);
    
}

void SceneLoaderBSP3::readLumpTextures()
{
    
    /* Temporary variables */
    const s32 ChunkSize     = 64 + 2 * sizeof(s32);
    const s32 ChunkCount    = Header_.DirEntries[BSP_LUMP_TEXTURES].Length / ChunkSize;
    
    c8 TextureName[64];
    s32 Flags, Contents;
    
    io::stringc TexFilename;
    io::stringc FileExtension;
    
    /* Set the offset */
    hFileLoader_->setFileSeek(hFile_, Header_.DirEntries[BSP_LUMP_TEXTURES].Offset);
    
    /* Loop for all chunks */
    for (s32 i = 0; i < ChunkCount; ++i)
    {
        
        /* Read the texture name */
        hFileLoader_->readStream(hFile_, TextureName, sizeof(TextureName));
        
        /* Read the flags */
        Flags = hFileLoader_->readValue<s32>(hFile_);
        
        /* Read the contents */
        Contents = hFileLoader_->readValue<s32>(hFile_);
        
        /* Set the texture filename */
        TexFilename = TexturePath_ + io::stringc(TextureName);
        
        /* Check for the file extension */
        if (hFileLoader_->doesFileExist(TexFilename+".jpg"))
            FileExtension = ".jpg";
        else
            FileExtension = ".tga";
        
        /* Load the texture */
        TextureList_.push_back( __spVideoDriver->loadTexture(TexFilename + FileExtension) );
        
        /* Swap the image on the y-axis if it is a tga image */
        if (FileExtension == ".tga")
            TextureList_[TextureList_.size()-1]->flipImageDataY();
        
    } // next chunk
    
}

void SceneLoaderBSP3::readLumpPlanes()
{
    
    /* Temporary variables */
    const s32 ChunkSize     = 4 * sizeof(f32);
    const s32 ChunkCount    = Header_.DirEntries[BSP_LUMP_PLANES].Length / ChunkSize;
    
    dim::vector3df Normal;
    f32 Dist;
    
    /* Set the offset */
    hFileLoader_->setFileSeek(hFile_, Header_.DirEntries[BSP_LUMP_PLANES].Offset);
    
    /* Loop for all chunks */
    for (s32 i = 0; i < ChunkCount; ++i)
    {
        
        /* Read the normal */
        Normal.set(
            hFileLoader_->readValue<f32>(hFile_),
            hFileLoader_->readValue<f32>(hFile_),
            hFileLoader_->readValue<f32>(hFile_)
        );
        
        /* Read the distance */
        Dist = hFileLoader_->readValue<f32>(hFile_);
        
        /* Create a new plane */
        PlaneList_.push_back(dim::plane3df(Normal, Dist));
        
    } // next chunk
    
}

void SceneLoaderBSP3::readLumpNodes()
{
    
    /* Temporary variables */
    const s32 ChunkSize     = sizeof(SNodeBSP);
    const s32 ChunkCount    = Header_.DirEntries[BSP_LUMP_NODES].Length / ChunkSize;
    
    SNodeBSP ChunkData;
    
    /* Set the offset */
    hFileLoader_->setFileSeek(hFile_, Header_.DirEntries[BSP_LUMP_NODES].Offset);
    
    /* Loop for all chunks */
    for (s32 i = 0; i < ChunkCount; ++i)
    {
        
        /* Read the node data */
        hFileLoader_->readStream(hFile_, &ChunkData, sizeof(ChunkData));
        
        /* ... */
        
    } // next chunk
    
}

void SceneLoaderBSP3::readLumpLeafs()
{
    
    /* Temporary variables */
    const s32 ChunkSize     = sizeof(SLeafBSP);
    const s32 ChunkCount    = Header_.DirEntries[BSP_LUMP_LEAFS].Length / ChunkSize;
    
    SLeafBSP ChunkData;
    
    /* Set the offset */
    hFileLoader_->setFileSeek(hFile_, Header_.DirEntries[BSP_LUMP_LEAFS].Offset);
    
    /* Loop for all chunks */
    for (s32 i = 0; i < ChunkCount; ++i)
    {
        
        /* Read the leaf data */
        hFileLoader_->readStream(hFile_, &ChunkData, sizeof(ChunkData));
        
        /* ... */
        
    } // next chunk
    
}

void SceneLoaderBSP3::readLumpLeafFaces()
{
    
    /* Temporary variables */
    const s32 ChunkSize     = sizeof(s32);
    const s32 ChunkCount    = Header_.DirEntries[BSP_LUMP_LEAFFACES].Length / ChunkSize;
    
    s32 Face;
    
    /* Set the offset */
    hFileLoader_->setFileSeek(hFile_, Header_.DirEntries[BSP_LUMP_LEAFFACES].Offset);
    
    /* Loop for all chunks */
    for (s32 i = 0; i < ChunkCount; ++i)
    {
        
        /* Read the leaf face data */
        Face = hFileLoader_->readValue<s32>(hFile_);
        
        /* ... */
        
    } // next chunk
    
}

void SceneLoaderBSP3::readLumpLeafBrushes()
{
    
    /* Temporary variables */
    const s32 ChunkSize     = sizeof(s32);
    const s32 ChunkCount    = Header_.DirEntries[BSP_LUMP_LEAFBRUSHES].Length / ChunkSize;
    
    s32 Brush;
    
    /* Set the offset */
    hFileLoader_->setFileSeek(hFile_, Header_.DirEntries[BSP_LUMP_LEAFBRUSHES].Offset);
    
    /* Loop for all chunks */
    for (s32 i = 0; i < ChunkCount; ++i)
    {
        
        /* Read the leaf brush data */
        Brush = hFileLoader_->readValue<s32>(hFile_);
        
        /* ... */
        
    } // next chunk
    
}

void SceneLoaderBSP3::readLumpModels()
{
    
    /* Temporary variables */
    const s32 ChunkSize     = sizeof(SModelBSP);
    const s32 ChunkCount    = Header_.DirEntries[BSP_LUMP_MODELS].Length / ChunkSize;
    
    SModelBSP ChunkData;
    
    /* Set the offset */
    hFileLoader_->setFileSeek(hFile_, Header_.DirEntries[BSP_LUMP_MODELS].Offset);
    
    /* Loop for all chunks */
    for (s32 i = 0; i < ChunkCount; ++i)
    {
        
        /* Read the model data */
        hFileLoader_->readStream(hFile_, &ChunkData, sizeof(ChunkData));
        
        /* ... */
        
    } // next chunk
    
}

void SceneLoaderBSP3::readLumpBrushes()
{
    
    /* Temporary variables */
    const s32 ChunkSize     = sizeof(SBrushBSP);
    const s32 ChunkCount    = Header_.DirEntries[BSP_LUMP_BRUSHES].Length / ChunkSize;
    
    SBrushBSP ChunkData;
    
    /* Set the offset */
    hFileLoader_->setFileSeek(hFile_, Header_.DirEntries[BSP_LUMP_BRUSHES].Offset);
    
    /* Loop for all chunks */
    for (s32 i = 0; i < ChunkCount; ++i)
    {
        
        /* Read the brush data */
        hFileLoader_->readStream(hFile_, &ChunkData, sizeof(ChunkData));
        
        /* ... */
        
    } // next chunk
    
}

void SceneLoaderBSP3::readLumpBrushSides()
{
    
    /* Temporary variables */
    const s32 ChunkSize     = sizeof(SBrushSideBSP);
    const s32 ChunkCount    = Header_.DirEntries[BSP_LUMP_BRUSHSIDES].Length / ChunkSize;
    
    SBrushSideBSP ChunkData;
    
    /* Set the offset */
    hFileLoader_->setFileSeek(hFile_, Header_.DirEntries[BSP_LUMP_BRUSHSIDES].Offset);
    
    /* Loop for all chunks */
    for (s32 i = 0; i < ChunkCount; ++i)
    {
        
        /* Read the brush side data */
        hFileLoader_->readStream(hFile_, &ChunkData, sizeof(ChunkData));
        
        /* ... */
        
    } // next chunk
    
}

void SceneLoaderBSP3::readLumpVertexes()
{
    
    /* Temporary variables */
    const s32 ChunkSize     = sizeof(SVertexBSP);
    const s32 ChunkCount    = Header_.DirEntries[BSP_LUMP_VERTEXES].Length / ChunkSize;
    
    SVertexBSP ChunkData;
    
    /* Set the offset */
    hFileLoader_->setFileSeek(hFile_, Header_.DirEntries[BSP_LUMP_VERTEXES].Offset);
    
    /* Loop for all chunks */
    for (s32 i = 0; i < ChunkCount; ++i)
    {
        
        /* Read the vertex data */
        hFileLoader_->readStream(hFile_, &ChunkData, sizeof(ChunkData));
        
        /* Swap the vertex coord */
        dim::vector3df Vector;
        
        Vector = ChunkData.Position / 64;
        
        ChunkData.Position.X = Vector.X;
        ChunkData.Position.Y = Vector.Z;
        ChunkData.Position.Z = Vector.Y;
        
        /* Swap the vertex normal */
        Vector = ChunkData.Normal;
        
        ChunkData.Normal.X = Vector.X;
        ChunkData.Normal.Y = Vector.Z;
        ChunkData.Normal.Z = Vector.Y;
        
        /* Add the vertex to the list */
        VerticesList_.push_back(ChunkData);
        
    } // next chunk
    
}

void SceneLoaderBSP3::readLumpMeshVerts()
{
    
    /* Temporary variables */
    const s32 ChunkSize     = sizeof(s32);
    const s32 ChunkCount    = Header_.DirEntries[BSP_LUMP_MESHVERTS].Length / ChunkSize;
    
    /* Set the offset */
    hFileLoader_->setFileSeek(hFile_, Header_.DirEntries[BSP_LUMP_MESHVERTS].Offset);
    
    /* Loop for all chunks */
    for (s32 i = 0; i < ChunkCount; ++i)
    {
        
        /* Read the offset */
        MeshVertOffsetList_.push_back( hFileLoader_->readValue<s32>(hFile_) );
        
    } // next chunk
    
}

void SceneLoaderBSP3::readLumpEffecs()
{
    
    /* Temporary variables */
    const s32 ChunkSize     = sizeof(SEffectBSP);
    const s32 ChunkCount    = Header_.DirEntries[BSP_LUMP_EFFECTS].Length / ChunkSize;
    
    SEffectBSP ChunkData;
    
    /* Set the offset */
    hFileLoader_->setFileSeek(hFile_, Header_.DirEntries[BSP_LUMP_EFFECTS].Offset);
    
    /* Loop for all chunks */
    for (s32 i = 0; i < ChunkCount; ++i)
    {
        
        /* Read the vertex data */
        hFileLoader_->readStream(hFile_, &ChunkData, sizeof(ChunkData));
        
        /* ... */
        
    } // next chunk
    
}

void SceneLoaderBSP3::readLumpFaces()
{
    
    /* Temporary variables */
    const s32 ChunkSize     = sizeof(SFaceBSP);
    const s32 ChunkCount    = Header_.DirEntries[BSP_LUMP_FACES].Length / ChunkSize;
    
    SFaceBSP ChunkData;
    
    /* Set the offset */
    hFileLoader_->setFileSeek(hFile_, Header_.DirEntries[BSP_LUMP_FACES].Offset);
    
    /* Loop for all chunks */
    for (s32 i = 0; i < ChunkCount; ++i)
    {
        
        /* Read the vertex data */
        hFileLoader_->readStream(hFile_, &ChunkData, sizeof(ChunkData));
        
        /* Add the face to the list */
        FacesList_.push_back(ChunkData);
        
    } // next chunk
    
}

void SceneLoaderBSP3::readLumpLightMaps()
{
    
    /* Temporary variables */
    const s32 ChunkSize     = sizeof(SLightMapBSP);
    const s32 ChunkCount    = Header_.DirEntries[BSP_LUMP_LIGHTMAPS].Length / ChunkSize;
    
    SLightMapBSP ChunkData;
    
    const s32 ImageDataSize = 128 * 128 * 3;
    
    const f32 Gamma = 2.5f;
    const f32 ScaleGamma = Gamma / 255;
    
    s32 x, y, j;
    f32 r, g, b;
    f32 scale, temp;
    
    video::Texture* LightMapTexture;
    u8* ImageData;
    
    /* Set the offset */
    hFileLoader_->setFileSeek(hFile_, Header_.DirEntries[BSP_LUMP_LIGHTMAPS].Offset);
    
    /* Loop for all chunks */
    for (s32 i = 0; i < ChunkCount; ++i)
    {
        
        /* Read the vertex data */
        hFileLoader_->readStream(hFile_, &ChunkData, sizeof(ChunkData));
        
        /*
         * Allocate new image data.
         * it will not deleted in this function,
         * because the Texture-class proceeds this
         */
        ImageData = new u8[ImageDataSize];
        
        /* Fill the image data */
        for (y = 0, j = 0; y < 128; ++y)
        {
            for (x = 0; x < 128; ++x, j += 3)
            {
                /* Get the lightmap raw data */
                r = ChunkData.LightMap[y][x][0];
                g = ChunkData.LightMap[y][x][1];
                b = ChunkData.LightMap[y][x][2];
                
                /* Scale the temporary color values */
                r *= ScaleGamma;
                g *= ScaleGamma;
                b *= ScaleGamma;
                
                scale = 1.0f;
                
                /* Check for to highest values */
                if (r > 1.0f && (temp = (1.0f/r)) < scale) scale = temp;
                if (g > 1.0f && (temp = (1.0f/g)) < scale) scale = temp;
                if (b > 1.0f && (temp = (1.0f/b)) < scale) scale = temp;
                
                scale *= 255.0f;
                
                /* Scale the temporary color values */
                r *= scale;
                g *= scale;
                b *= scale;
                
                /* Set the final color */
                ImageData[j+0] = (u8)r;
                ImageData[j+1] = (u8)g;
                ImageData[j+2] = (u8)b;
            }
        }
        
        /* Create a new light map */
        LightMapTexture = __spVideoDriver->createTexture(dim::size2di(128, 128), ImageData);
        
        /* Add the light map to the list */
        LightMapList_.push_back(LightMapTexture);
        
    } // next chunk
    
}

void SceneLoaderBSP3::readLumpLightVols()
{
    
    /* Temporary variables */
    const s32 ChunkSize     = sizeof(SLightVolsBSP);
    const s32 ChunkCount    = Header_.DirEntries[BSP_LUMP_LIGHTVOLS].Length / ChunkSize;
    
    SLightVolsBSP ChunkData;
    
    /* Set the offset */
    hFileLoader_->setFileSeek(hFile_, Header_.DirEntries[BSP_LUMP_LIGHTVOLS].Offset);
    
    /* Loop for all chunks */
    for (s32 i = 0; i < ChunkCount; ++i)
    {
        
        /* Read the vertex data */
        hFileLoader_->readStream(hFile_, &ChunkData, sizeof(ChunkData));
        
        /* ... */
        
    } // next chunk
    
}

void SceneLoaderBSP3::readLumpVisData()
{
    
    /* Temporary variables */
    SVisDataBSP ChunkData;
    s32 VisDataSize;
    
    /* Set the offset */
    hFileLoader_->setFileSeek(hFile_, Header_.DirEntries[BSP_LUMP_VISDATA].Offset);
    
    /* Read the number of vectors */
    ChunkData.CountVectors = hFileLoader_->readValue<s32>(hFile_);
    
    /* Read the size of vectors */
    ChunkData.VectorSize = hFileLoader_->readValue<s32>(hFile_);
    
    /* Compute the visbility data size */
    VisDataSize = ChunkData.CountVectors * ChunkData.VectorSize;
    
    /* Allocate enough visibility memory */
    ChunkData.VisibilityData = new u8[VisDataSize];
    
    /* Read the visibility data */
    hFileLoader_->readStream(hFile_, ChunkData.VisibilityData, VisDataSize);
    
    /* ... */
    
    /* Delete the visibility data */
    delete [] ChunkData.VisibilityData;
    
}

void SceneLoaderBSP3::createNewVertex(SVertexBSP &Vertex, SFaceBSP &Face)
{
    /* Add a new vertex */
    s32 CurVertexNr = hEntity_->addVertex(
        Vertex.Position,
        Vertex.Normal,
        dim::point2df(Vertex.TexCoord[0].X, Vertex.TexCoord[0].Y),
        255,//Vertex.Color
        0.0f
    );
    
    /* Set the second texture map coordinate */
    hEntity_->setVertexTexCoord(
        CurVertexNr,
        dim::point2df(Vertex.TexCoord[1].X, Vertex.TexCoord[1].Y),
        1
    );
}

void SceneLoaderBSP3::buildModel()
{
    
    /* Temporary variables */
    SVertexBSP Vertex;
    u32 surf;
    
    /* Loop for all faces */
    for (std::vector<SFaceBSP>::iterator it = FacesList_.begin(); it != FacesList_.end(); ++it)
    {
        
        /* Determine which entity must be used */
        hEntity_ = (TextureList_[it->Texture]->getFormat() == 4 ? hEntityTrans_ : hEntityBase_);
        
        /* Create a new surface */
        surf = hEntity_->createSurface();
        
        /* Add the texture */
        hEntity_->addTexture( TextureList_[it->Texture] );
        
        /* Add the lightmap */
        if (it->LightMap >= 0 && it->LightMap < LightMapList_.size())
            hEntity_->addTexture( LightMapList_[it->LightMap] );
        
        /* Compute the count of vertices/ control points */
        s32 Count = (it->Type == BSP_FACE_MESH ? it->CountMeshVerts : it->CountVertices);
        
        /* Check the type */
        if (it->Type != BSP_FACE_PATCH)
        {
            /* Loop for all vertices for the surface */
            for (s32 i = 0; i < Count; ++i)
            {
                /* Create a new vertex */
                createNewVertex(VerticesList_[it->FirstVertex + i], *it);
            } // next vertex
        }
        
        /* Determine which face type is used */
        switch (it->Type)
        {
            case BSP_FACE_POLYGON:
            {
                
                /* Loop for all traingles for the surface */
                for (s32 i = 1; i < it->CountVertices - 1; ++i)
                {
                    
                    /* Add a new triangle */
                    hEntity_->addTriangle(0, i, i+1);
                    
                } // next triangle
                
            }
            break;
            
            case BSP_FACE_PATCH:
            {
                
                /* Temporary vairbales */
                s32 x, y, s, p;
                s32 i, j, k, c;
                
                BiQuadraticPatch Patch;
                
                std::vector<SVertexBSP> Vertices(Count);
                
                dim::size2di biquadSize(
                    (it->PatchSize.Width  - 1) / 2,
                    (it->PatchSize.Height - 1) / 2
                );
                
                /* Translate the created vertices in the temporary vertices list */
                for (i = 0; i < Count; ++i)
                {
                    /* Copy the vertex data */
                    Vertices[i] = VerticesList_[it->FirstVertex + i];
                }
                
                /* Loop for all patches */
                for (y = 0; y < biquadSize.Height; ++y)
                {
                    for (x = 0; x < biquadSize.Width; ++x)
                    {
                        /* Loop for all control points */
                        for (s32 row = 0; row < 3; ++row)
                        {
                            for (s32 point = 0; point < 3; ++point)
                            {
                                /* Get the control points */
                                Patch.ControlPoints[row*3+point] = 
                                    Vertices[ ( y*2*it->PatchSize.Width+x*2 ) + row*it->PatchSize.Width+point ];
                            }
                        }
                        
                        /* Compute the tesselation */
                        Patch.doTesselate(Tesselation_);
                    }
                }
                
                /* Loop for all vertices */
                for (std::vector<SVertexBSP>::iterator itPatch = Patch.VerticesList.begin(); itPatch != Patch.VerticesList.end(); ++itPatch)
                {
                    /* Create a new vertex */
                    createNewVertex(*itPatch, *it);
                } // next vertex
                
                /* Loop for all patches */
                for (p = 0, c = 2*(Tesselation_+1); p < biquadSize.Height * biquadSize.Width; ++p)
                {
                    /* Loop for all triangles */
                    for (i = 0; i < Tesselation_; ++i)
                    {
                        for (j = 0, s = 0; j < c - 2; ++j)
                        {
                            
                            /* Compute the first index */
                            k = i*c;
                            
                            /* Swap the triangle strip side */
                            s = 1 - s;
                            
                            /* Create a new triangle */
                            if (s)
                                hEntity_->addTriangle(Patch.IndicesList[p][k+j+0], Patch.IndicesList[p][k+j+1], Patch.IndicesList[p][k+j+2]);
                            else
                                hEntity_->addTriangle(Patch.IndicesList[p][k+j+2], Patch.IndicesList[p][k+j+1], Patch.IndicesList[p][k+j+0]);
                            
                        } // next triangle
                    } // next tesseleation
                } // next patch
                
                /* Update the normals for the current surface */
                hEntity_->updateNormals(surf);
                
            }
            break;
            
            case BSP_FACE_MESH:
            {
                
                /* Loop for all traingles for the surface */
                for (s32 i = 0; i < it->CountMeshVerts; i += 3)
                {
                    
                    /* Add a new triangle */
                    hEntity_->addTriangle(
                        MeshVertOffsetList_[it->FirstMeshVerts + i+0],
                        MeshVertOffsetList_[it->FirstMeshVerts + i+1],
                        MeshVertOffsetList_[it->FirstMeshVerts + i+2]
                    );
                    
                } // next triangle
                
            }
            break;
            
            case BSP_FACE_BILLBOARD:
            {
                
            }
            break;
        }
        
    } // next face
    
    /* Update the model and build it finally */
    hEntityBase_->updateModeling(false);
    hEntityTrans_->updateModeling(false);
    
}

void SceneLoaderBSP3::examineScript(std::vector<io::stringc> &ScriptData)
{
    
    /* Temporary variables */
    io::stringc Type, Value, ClassName;
    dim::vector3df Vector;
    
    /* Loop for all lines */
    for (std::vector<io::stringc>::iterator it = ScriptData.begin(); it != ScriptData.end(); ++it)
    {
        /* Check special symboles */
        if (*it == "{")
            continue;
        
        /* Check if an block has been end */
        if (*it == "}")
        {
            /* Check which classname has the block */
            if (ClassName.left(7) == "weapon_" || ClassName.left(5) == "item_" || ClassName.left(5) == "ammo_")
            {
                scene::Node* joint = SceneMngr_->createJoint();
                joint->setParent(hEntityBase_);
                joint->setPosition(dim::vector3df(Vector.X, Vector.Z, Vector.Y) / 64);
                joint->setName(ClassName);
            }
            
            /* Reset the classname */
            ClassName = "";
            
            /* Continue with the next line */
            continue;
        }
        
        /* Get the type */
        Type = getScriptLineType(*it);
        
        /* Get the value */
        Value = getScriptLineValue(*it);
        
        /* First check of the type */
        if (Type == "classname")
            ClassName = Value;
        else if (Type == "origin")
            Vector = getScriptLineVector(*it, Value);
        
    } // next line
    
}

io::stringc SceneLoaderBSP3::getScriptLineType(io::stringc &Line)
{
    return Line.section(1, Line.find("\"", 1));
}

io::stringc SceneLoaderBSP3::getScriptLineValue(io::stringc &Line)
{
    return Line.section(Line.find("\"", 2)+3, Line.size() - 1);
}

std::vector<io::stringc> SceneLoaderBSP3::getScriptLineValueList(io::stringc &Line)
{
    
}

dim::vector3df SceneLoaderBSP3::getScriptLineVector(io::stringc &Line, io::stringc Value)
{
    /* Temporary variables */
    dim::vector3df Vector;
    
    /* Get the blank positions */
    s32 pos1 = Value.find(" ", 1);
    
    if (pos1 == -1)
        return dim::vector3df();
    
    s32 pos2 = Value.find(" ", pos1 + 1);
    
    /* Get the component values */
    Vector.X = Value.left(pos1).val<f32>();
    Vector.Y = Value.section(pos1, pos2).val<f32>();
    Vector.Z = Value.right(Value.size() - pos2).val<f32>();
    
    /* Return the vector */
    return Vector;
}


/*
 * ========== BSP Loader internal class - BiQuadraticPatch ==========
 */

SceneLoaderBSP3::BiQuadraticPatch::BiQuadraticPatch()
{
}
SceneLoaderBSP3::BiQuadraticPatch::~BiQuadraticPatch()
{
}

void SceneLoaderBSP3::BiQuadraticPatch::doTesselate(s32 Level)
{
    
    /* Temporary variables */
    f32 a, b;
    s32 i, j, k, p, c;
    
    SVertexBSP TempVertex[3];
    SVertexBSP Vertex;
    
    /* Loop for the first tesselation */
    for (i = 0; i <= Level; ++i)
    {
        /* Compute the patch x axis */
        a = (f32)i / Level;
        b = 1.0f - a;
        
        /* Set the current vertex data */
        Vertex = 
            ControlPoints[0] * (b * b) +
            ControlPoints[3] * (2 * b * a) +
            ControlPoints[6] * (a * a);
        
        /* Add a new vertex */
        VerticesList.push_back(Vertex);
    }
    
    /* Loop for the second tesselation */
    for (i = 1; i <= Level; ++i)
    {
        /* Compute the patch y axis */
        a = (f32)i / Level;
        b = 1.0f - a;
        
        /* Loop for the three temporary vertices */
        for (j = 0; j < 3; ++j)
        {
            /* Compute the index for the temporary vertices */
            k = j * 3;
            
            /* Compute the temporary vertices */
            TempVertex[j] =
                ControlPoints[k + 0] * (b * b) +
                ControlPoints[k + 1] * (2 * b * a) +
                ControlPoints[k + 2] * (a * a);
        }
        
        /* Loop for the third tesselation */
        for (j = 0; j <= Level; ++j)
        {
            /* Compute the patch x axis */
            a = (f32)j / Level;
            b = 1.0f - a;
            
            /* Set the current vertex data */
            Vertex = 
                TempVertex[0] * (b * b) +
                TempVertex[1] * (2 * b * a) +
                TempVertex[2] * (a * a);
            
            /* Add a new vertex */
            VerticesList.push_back(Vertex);
        }
    }
    
    /* Temporary indices list */
    std::vector<u32> TempList;
    
    /* Compute the last count */
    c = IndicesList.size() * ( ( Level + 1 ) * (Level + 1) );
    
    /* Loop for all indices */
    for (s32 row = 0; row < Level; ++row)
    {
        for (s32 point = 0; point <= Level; ++point)
        {
            /* Compute the next two indices */
            TempList.push_back( c + (row + 1) * (Level + 1) + point );
            TempList.push_back( c + row * (Level + 1) + point );
        }
    }
    
    /* Add a new indices list */
    IndicesList.push_back(TempList);
    
}


/*
 * BSPLoaderExtensions class
 */

BSPLoaderExtensions::BSPLoaderExtensions()
{
}
BSPLoaderExtensions::~BSPLoaderExtensions()
{
}

void BSPLoaderExtensions::createScript(std::vector<io::stringc> &ScriptData, c8* EntityDescription)
{
    
    /* Temporary variables */
    s32 begin = 0, end = 0, i;
    
    /* Loop while search for new lines */
    do
    {
        /* Check if the end of a line has been detected */
        if (EntityDescription[end] == '\n')
        {
            /* Allocate a new line */
            std::string Line(end - begin, ' ');
            
            /* Loop for the detected part */
            for (i = begin; i < end; ++i)
                Line[i - begin] = EntityDescription[i];
            
            /* Add the new line */
            ScriptData.push_back(Line);
            
            /* Set the new start point */
            begin = end + 1;
        }
    }
    while (EntityDescription[end++]);
    
}


} // /namespace scene

} // /namespace sp








// ==========================================================================
