/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * Scene management
 *
 */

#include "spSceneManager.hpp"
#include "spMainDevice.hpp"
#include "spInternalDeclerations.hpp"


namespace sp
{

extern UserControl* __spUserControl;
extern video::VideoDriver* __spVideoDriver;

namespace scene
{


/*
 * Internal functions
 */

bool sortSprites(Sprite* &obj1, Sprite* &obj2)
{
    return obj1->Location_.Z > obj2->Location_.Z;
}

bool sortEntitiesDefault(Entity* &obj1, Entity* &obj2)
{
    return obj1->getOrder() > obj2->getOrder() || ( obj1->getOrder() == obj2->getOrder() &&
        ( obj1->getThisMaterial().ColorDiffuse.Alpha > obj2->getThisMaterial().ColorDiffuse.Alpha ||
        ( obj1->getThisMaterial().ColorDiffuse.Alpha == obj2->getThisMaterial().ColorDiffuse.Alpha &&
        obj1->PerspectiveDistance_ > obj2->PerspectiveDistance_ ) ) );
}

bool sortEntitiesFast(Entity* &obj1, Entity* &obj2)
{
    return obj1->getOrder() > obj2->getOrder() || ( obj1->getOrder() == obj2->getOrder() &&
        ( obj1->getThisMaterial().ColorDiffuse.Alpha > obj2->getThisMaterial().ColorDiffuse.Alpha ||
        ( obj1->getThisMaterial().ColorDiffuse.Alpha == obj2->getThisMaterial().ColorDiffuse.Alpha &&
        obj1->PerspectiveDistance_ < obj2->PerspectiveDistance_ ) ) );
}


/*
 * Internal members
 */

StencilManager* __spStencilManager = 0;
anim::Animator* __spAnimator = 0;
effect::SpecialEffectsActor* __spSpecialEffectsActor = 0;


/*
 * SceneManager class
 */

SceneManager::SceneManager() :
    ActiveCamera_(0), isLighting_(false), isStencilEffects_(false), isDepthSortFast_(false)
{
    /* Init light ID list */
    memset(__spLightIDList, 0, sizeof(__spLightIDList));
}
SceneManager::~SceneManager()
{
    /* Delete all 3d objects */
    for (std::list<Entity*>::iterator it = EntityList_.begin(); it != EntityList_.end(); ++it)
        spDeletePointer(*it);
    for (std::list<Light*>::iterator it = LightList_.begin(); it != LightList_.end(); ++it)
        spDeletePointer(*it);
    for (std::list<Sprite*>::iterator it = SpriteList_.begin(); it != SpriteList_.end(); ++it)
        spDeletePointer(*it);
    for (std::list<Camera*>::iterator it = CameraList_.begin(); it != CameraList_.end(); ++it)
        spDeletePointer(*it);
    for (std::list<Terrain*>::iterator it = TerrainList_.begin(); it != TerrainList_.end(); ++it)
        spDeletePointer(*it);
    for (std::list<Node*>::iterator it = NodeList_.begin(); it != NodeList_.end(); ++it)
        spDeletePointer(*it);
    
    /* Delete all pointers */
    spDeletePointer(__spStencilManager);
    spDeletePointer(__spAnimator);
    spDeletePointer(__spSpecialEffectsActor);
}

StencilManager* SceneManager::getStencilManager()
{
    return __spStencilManager ? __spStencilManager : __spStencilManager = new scene::StencilManager();
}
anim::Animator* SceneManager::getAnimator()
{
    return __spAnimator ? __spAnimator : __spAnimator = new anim::Animator();
}
effect::SpecialEffectsActor* SceneManager::getSpecialEffectsActor()
{
    return __spSpecialEffectsActor ? __spSpecialEffectsActor : __spSpecialEffectsActor = __spVideoDriver->createSpecialEffectsActor();
}

/* Create a simple node */

Node* SceneManager::createJoint()
{
    Node* NewJoint = new Node(NODE_JOINT);
    NodeList_.push_back(NewJoint);
    return NewJoint;
}

/* Entities (empty, primitives, models, supershapes, skyboxes, highmaps) */

Entity* SceneManager::createPivot()
{
    Entity* NewEntity = new Entity();
    EntityList_.push_back(NewEntity);
    return NewEntity;
}

Entity* SceneManager::createModel(u32 Type, s32 Detail, f32 Radius1, f32 Radius2, bool isCap)
{
    ModelCreator Creator;
    Entity* NewEntity = Creator.createModel(Type, Detail, Radius1, Radius2, 360.0f, 1.0f, isCap);
    EntityList_.push_back(NewEntity);
    return NewEntity;
}

Entity* SceneManager::createModel(u32 Type, s32 Detail, f32 Radius1, f32 Radius2, f32 DegreeLength, f32 Size, bool isCap)
{
    ModelCreator Creator;
    Entity* NewEntity = Creator.createModel(Type, Detail, Radius1, Radius2, DegreeLength, Size, isCap);
    EntityList_.push_back(NewEntity);
    return NewEntity;
}

Entity* SceneManager::createSuperShape(f32 ValueList[12], s32 Detail)
{
    ModelCreator Creator;
    Entity* NewEntity = Creator.createSuperShape(ValueList, Detail);
    EntityList_.push_back(NewEntity);
    return NewEntity;
}

/*
 * Load model function:
 * (supports several 3d file formats)
 */

Entity* SceneManager::loadModel(io::stringc Filename, io::stringc TexturePath, EModelTypes FileType)
{
    
    /* Information message */
    io::printMessage(io::stringc("Load model: \"")+Filename+io::stringc("\""));
    
    /* Temporary variables */
    io::stringc FileFormat = Filename.right(4).upper();
    Entity* NewEntity = 0;
    
    if (TexturePath == "?")
        TexturePath = Filename.getFilePath();
    
    ModelLoaderRAW* ModelLoader = 0;
    
    /* === Check if the file does exist === */
    
    if (!io::FileLoader().doesFileExist(Filename))
    {
        /* Error message */
        io::printError("could not found model file");
        
        /* Return an empty model */
        return createPivot();
    }
    
    /* === Select format === */
    
    if ( ( FileType == MODEL_AUTODETECT && FileFormat == ".SPM" ) || FileType == MODEL_SPM ) // "SoftPixel Model"
        ModelLoader = new ModelLoaderSPM();
    else if ( ( FileType == MODEL_AUTODETECT && FileFormat == ".3DS" ) || FileType == MODEL_3DS ) // "3D Studio"
        ModelLoader = new ModelLoader3DS();
    else if ( ( FileType == MODEL_AUTODETECT && Filename.right(5).upper() == ".MS3D" ) || FileType == MODEL_MS3D ) // "Milkshape 3D"
        ModelLoader = new ModelLoaderMS3D();
    else if ( ( FileType == MODEL_AUTODETECT && Filename.right(2).upper() == ".X" ) || FileType == MODEL_D3D ) // "Direct X"
        ModelLoader = new ModelLoaderD3D();
    else if ( ( FileType == MODEL_AUTODETECT && FileFormat == ".B3D" ) || FileType == MODEL_B3D ) // "Blitz3D"
        ModelLoader = new ModelLoaderB3D();
    else if ( ( FileType == MODEL_AUTODETECT && FileFormat == ".MD2" ) || FileType == MODEL_MD2 ) // "Quake 2"
        ModelLoader = new ModelLoaderMD2();
    else if ( ( FileType == MODEL_AUTODETECT && FileFormat == ".OBJ" ) || FileType == MODEL_OBJ ) // "Wavefront object"
        ModelLoader = new ModelLoaderOBJ();
    else if ( ( FileType == MODEL_AUTODETECT && FileFormat == ".TMF" ) || FileType == MODEL_TMF ) // "ToxyVision"
        ModelLoader = new ModelLoaderTMF();
    else
    {
        /* Print an error message */
        io::printError(io::stringc("model file format: \"")+Filename.getFileExtension().upper()+io::stringc("\" is not supported"));
        
        /* Return an empty model */
        return createPivot();
    }
    
    /* Load the model with the determined model loader */
    NewEntity = ModelLoader->loadModel(Filename, TexturePath);
    
    /* Add the entity to the list */
    EntityList_.push_back(NewEntity);
    
    /* Delete the temporary model loader */
    spDeletePointer(ModelLoader);
    
    /* Return the new entity */
    return NewEntity;
    
}

bool SceneManager::saveModel(Entity* hEntity, io::stringc Filename, EModelTypes FileType)
{
    
    /* Information message */
    io::printMessage(io::stringc("Save model: \"")+Filename+io::stringc("\""));
    
    io::stringc FileFormat = Filename.right(4).upper();
    
    ModelSaverRAW* ModelSaver = 0;
    
    /* Determine the correct 3d file format */
    if ( ( FileType == MODEL_AUTODETECT && FileFormat == ".SPM" ) || FileType == MODEL_SPM ) // "SoftPixel Model"
        ModelSaver = new ModelSaverSPM();
    else if ( ( FileType == MODEL_AUTODETECT && FileFormat == ".B3D" ) || FileType == MODEL_B3D ) // "Blitz3D"
        ModelSaver = new ModelSaverB3D();
    else if ( ( FileType == MODEL_AUTODETECT && FileFormat == ".OBJ" ) || FileType == MODEL_OBJ ) // "Wavefront object"
        ModelSaver = new ModelSaverOBJ();
    else
    {
        /* Print an error message */
        io::printError(io::stringc("model file format: \"")+Filename.getFileExtension().upper()+io::stringc("\" is not supported"));
        
        /* Exit with a failure */
        return false;
    }
    
    /* Save the model with the determined model saver */
    return ModelSaver && ModelSaver->saveModel(hEntity, Filename);
    
}


/*
 * Load scene function:
 * (supports several 3d level formats)
 */

Entity* SceneManager::loadScene(io::stringc Filename, io::stringc TexturePath, ESceneTypes FileType, s32 Tesselation)
{
    /* Information message */
    io::printMessage(io::stringc("Load scene: \"")+Filename+io::stringc("\""));
    
    /* Temporary variables */
    io::stringc FileFormat = Filename.right(4).upper();
    Entity* NewEntity = 0;
    
    if (TexturePath == "?")
        TexturePath = Filename.getFilePath();
    
    SceneLoaderRAW* SceneLoader = 0;
    
    /* Determine the correct 3d file format */
    if ( ( FileType == SCENE_AUTODETECT && FileFormat == ".SPS" ) || FileType == SCENE_SPS )
        SceneLoader = new SceneLoaderSPS();
    else if ( ( FileType == SCENE_AUTODETECT && FileFormat == ".BSP" ) || FileType == SCENE_BSP )
    {
        /*
         * Get the magic number of the BSP file,
         * because there are more than one types of this file format
         */
        io::FileLoader TempLoader;
        io::File* TempFile = TempLoader.openFile(Filename);
        s32 MagicNumber = TempLoader.readValue<s32>(TempFile);
        TempLoader.closeFile(TempFile);
        
        /* Determine the format type */
        if (MagicNumber == 0x17 || MagicNumber == 0x1E)
            SceneLoader = new SceneLoaderBSP1(); // Quake1/ Half-Life 1 level
        else
            SceneLoader = new SceneLoaderBSP3(); // Quake III Arena level
    }
    else
    {
        /* Print an error message */
        io::printError(io::stringc("scene file format: \"")+Filename.getFileExtension().upper()+io::stringc("\" is not supported"));
        
        /* Allocate a new & empty entity */
        NewEntity = new Entity();
        
        /* Add the entity to the list */
        EntityList_.push_back(NewEntity);
        
        /* Return the new entity */
        return NewEntity;
    }
    
    /* Load the scene with the determined scene loader */
    NewEntity = SceneLoader->loadScene(this, Filename, TexturePath, Tesselation);
    
    /* Add the entity to the list */
    EntityList_.push_back(NewEntity);
    
    /* Delete the temporary scene loader */
    spDeletePointer(SceneLoader);
    
    /* Return the new entity */
    return NewEntity;
}

bool SceneManager::saveScene(io::stringc Filename)
{
    if (Filename.right(4).upper() == ".SPW")
    {
        ;
    }
    return false;
}

Entity* SceneManager::createSkyBox(video::Texture* TextureList[6], f32 Radius, u8 Type)
{
    Entity* NewEntity = new Entity(TextureList, Radius, Type);
    EntityList_.push_back(NewEntity);
    return NewEntity;
}

Entity* SceneManager::createHeightField(video::Texture* hTextureHighMap, s32 Detail)
{
    Entity* NewEntity = new Entity(hTextureHighMap, Detail);
    EntityList_.push_back(NewEntity);
    return NewEntity;
}

Entity* SceneManager::createModelList(std::vector<Entity*> ModelList, bool isOldDelete)
{
    /* Allocate a new entity */
    Entity* NewEntity = new Entity();
    
    /* Temporary variables */
    std::list<video::Texture*> TexList;
    u32 TriangleVertices[3];
    u32 Surface = 0, surf = 0, i = 0;
    dim::matrix4f NormalMatrix;
    
    // Loop all entities in the list
    for (std::vector<Entity*>::iterator it = ModelList.begin(); it != ModelList.end(); ++it)
    {
        /* Matrix transformation */
        spWorldMatrix.matrixReset();
        
        (*it)->updateModelviewMatrix();
        
        NormalMatrix = spWorldMatrix;
        NormalMatrix.setPosition(0);
        
        /* Loop the surfaces of the current entity */
        for (surf = 0; surf < (*it)->getMeshSurfaceCount(); ++surf)
        {
            /* Create a new surface */
            Surface = NewEntity->createSurface();
            
            /* Get the texture list */
            TexList = (*it)->getTextureList(surf);
            
            /* Add all textures */
            for (std::list<video::Texture*>::iterator itTex = TexList.begin(); itTex != TexList.end(); ++itTex)
                NewEntity->addTexture(*itTex);
            
            /* Add all vertices */
            for (i = 0; i < (*it)->getMeshVerticesCount(surf); ++i)
            {
                NewEntity->addVertex(
                    spWorldMatrix       * (*it)->getVertexCoord(i),
                    (NormalMatrix       * (*it)->getVertexNormal(i)).normalize(),
                    spTextureMatrix[0]  * (*it)->getVertexTexCoord(i).vec3<f32>(),
                    (*it)->getVertexColor(i),
                    (*it)->getVertexFog(i)
                );
            }
            
            /* Add all triangles */
            for (i = 0; i < (*it)->getMeshTriangleCount(surf); ++i)
            {
                (*it)->getTriangleVertices(i, TriangleVertices);
                NewEntity->addTriangle(TriangleVertices[0], TriangleVertices[1], TriangleVertices[2]);
            }
        }
        
        /* Delete the original entity if the parameter 'isOldDelete' is 'true' */
        if (isOldDelete)
            deleteNode(*it);
    }
    
    /* Create the new entity finally */
    NewEntity->updateModeling(false);
    
    /* Add the new entity to the list */
    EntityList_.push_back(NewEntity);
    
    /* Return the new entity */
    return NewEntity;
}

Entity* SceneManager::createModelSurface(Entity* Model, u32 Surface)
{
    /* Check if surface is not out of range */
    if (Surface >= Model->getMeshSurfaceCount())
        return 0;
    
    /* Allocate a new entity */
    Entity* NewEntity = new Entity();
    
    /* Temporary variables */
    s32 i;
    u32 v[3];
    video::Texture* tex;
    
    /* Create a new surface */
    NewEntity->createSurface();
    
    /* Select the current surface */
    Model->selectSurface(Surface);
    
    /* Loop for all vertices of the specified surface */
    for (i = 0; i < Model->getMeshVerticesCount(Surface); ++i)
    {
        NewEntity->addVertex(
            Model->getVertexCoord(i),
            Model->getVertexNormal(i),
            Model->getVertexTexCoord(i),
            Model->getVertexColor(i),
            Model->getVertexFog(i)
        );
    }
    
    /* Loop for all triangles of the specified surface */
    for (i = 0; i < Model->getMeshTriangleCount(Surface); ++i)
    {
        Model->getTriangleVertices(i, v);
        NewEntity->addTriangle(v[0], v[1], v[2]);
    }
    
    /* Add all textures of the specified surface */
    for (i = 0; i < video::MAX_COUNT_OF_TEXTURES; ++i)
    {
        if (tex = Model->getTexture(i))
            NewEntity->addTexture(tex, i);
    }
    
    /* Create the new entity finally */
    NewEntity->updateModeling(false);
    
    /* Add the new entity to the list */
    EntityList_.push_back(NewEntity);
    
    /* Return the new entity */
    return NewEntity;
}

/* Lights, sprites, cameras, terrains */

Light* SceneManager::createLight(ELightTypes Type)
{
    Light* hLight = new Light(Type);
    LightList_.push_back(hLight);
    return hLight;
}
Sprite* SceneManager::createSprite(video::Texture* hTexture)
{
    Sprite* hSprite = new Sprite(hTexture);
    SpriteList_.push_back(hSprite);
    return hSprite;
}
Camera* SceneManager::createCamera()
{
    Camera* hCamera = new Camera();
    CameraList_.push_back(hCamera);
    return hCamera;
}
Terrain* SceneManager::createTerrain(video::Texture* hTextureHighMap, video::Texture* hTexture, f32 ChangeDist)
{
    Terrain* hTerrain = new Terrain(hTextureHighMap, hTexture, ChangeDist);
    TerrainList_.push_back(hTerrain);
    return hTerrain;
}

/*
 * Render whole scene/ world (you can render objects manual but this function is the normal way to render scenes)
 */

void SceneManager::renderScene(Camera* hCamera)
{
    
    setActiveCamera(hCamera);
    
    /* Temporary variables */
    
    dim::vector3df CamPosition(hCamera->getPosition(true));
    
    /* Special effects (stencil clipping mirrors) */
    
    if (isStencilEffects_)
        __spStencilManager->renderStencilMirrors(hCamera);
    
    /* Update the 3d sound effect */
    if (sound::__ssDevice)
        sound::__ssDevice->updateSounds3D(CamPosition, hCamera->getRotation(true).Y);
    
    /* ===== Camera ===== */
    
    /* Camera matrix transformation */
    
    hCamera->updateViewmode();
    hCamera->updateModelviewMatrix();
    
    /* ===== Lights ===== */
    
    for (std::list<Light*>::iterator itLight = LightList_.begin(); itLight != LightList_.end(); ++itLight)
    {
        if ((*itLight)->getVisible())
        {
            spWorldMatrix.matrixReset();
            (*itLight)->update();
        }
    }
    
    /* ===== Entities ===== */
    
    __spVideoDriver->enableVertexBuffer();
    
    for (std::list<Entity*>::iterator itEntity = EntityList_.begin(); itEntity != EntityList_.end(); ++itEntity)
    {
        if ((*itEntity)->getVisible())
        {
            spWorldMatrix.matrixReset();
            (*itEntity)->render();
        }
    }
    
    /* ===== Terrains ===== */
    
    for (std::list<Terrain*>::iterator itTerrain = TerrainList_.begin(); itTerrain != TerrainList_.end(); ++itTerrain)
    {
        if ((*itTerrain)->getVisible())
        {
            spWorldMatrix.matrixReset();
            (*itTerrain)->render(CamPosition);
        }
    }
    
    __spVideoDriver->disableVertexBuffer();
    
    /* ===== Sprites ===== */
    
    /* Back settings of triangle list states */
    __spVideoDriver->disableTriangleListStates();
    
    /* Update coordination */
    
    for (std::list<Sprite*>::iterator itSprite = SpriteList_.begin(); itSprite != SpriteList_.end(); ++itSprite)
    {
        if ((*itSprite)->getVisible())
        {
            spWorldMatrix = spViewMatrix;
            
            (*itSprite)->updateParent();
            spWorldMatrix.matrixTranslate((*itSprite)->getPosition());
            
            (*itSprite)->Location_ = spWorldMatrix.getPosition();
        }
    }
    
    SpriteList_.sort(sortSprites);
    
    /* Render sprites */
    
    spWorldMatrix.matrixReset();
    spViewMatrix.matrixReset();
    
    for (std::list<Sprite*>::iterator itSprite = SpriteList_.begin(); itSprite != SpriteList_.end(); ++itSprite)
    {
        if ((*itSprite)->getVisible())
            (*itSprite)->render();
    }
    
    __spVideoDriver->disableTexturing();
    __spVideoDriver->setDefaultAlphaBlending();
    
    /* Special effects (stencil shadows) */
    if (isStencilEffects_)
        __spStencilManager->renderStencilShadows(hCamera);
    
    /* Sorting entities */
    if (isDepthSortFast_)
        EntityList_.sort(sortEntitiesFast);
    else
        EntityList_.sort(sortEntitiesDefault);
    
    /* Disable all 3d rendering states */
    __spVideoDriver->disable3DRenderStates();
    
}

/*
 * Render whole scene/ world for all cameras
 */

void SceneManager::renderScene()
{
    
    /* Loop all cameras */
    for (std::list<Camera*>::iterator itCam = CameraList_.begin(); itCam != CameraList_.end(); ++itCam)
    {
        
        /* Check if the camera is hide */
        if (!(*itCam)->getVisible())
            continue;
        
        /* Render the whole scene for the current camera */
        renderScene(*itCam);
        
    } // next camera
    
}

/*
 * Renders the whole scene in a 'virtual reality'. You can use a '3d glaces' because this scene will renderd two time: red and green!
 */

void SceneManager::renderSceneVirtualReality(Camera* hCamera, f32 CamDegree, f32 CamDist)
{
    static video::Texture* spacetex1 = 0, * spacetex2 = 0;
    
    /* Red color scene */
    
    __spVideoDriver->setColorMask(false, true, false, true);
    {
        hCamera->doMove(dim::vector3df(CamDist, 0, 0));
        hCamera->doTurn(dim::vector3df(0, CamDegree, 0));
    }
    renderScene();
    
    __spVideoDriver->deleteTexture(spacetex1);
    spacetex1 = __spVideoDriver->createScreenShot(dim::point2di(), dim::size2di(1024, 1024));
    
    /* Green color scene */
    
    __spVideoDriver->clearBuffers();
    __spVideoDriver->setColorMask(true, false, false, true);
    {
        hCamera->doTurn(dim::vector3df(0, -CamDegree, 0));
        hCamera->doMove(dim::vector3df(-CamDist*2, 0, 0));
        hCamera->doTurn(dim::vector3df(0, -CamDegree, 0));
    }
    renderScene();
    
    __spVideoDriver->deleteTexture(spacetex2);
    spacetex2 = __spVideoDriver->createScreenShot(dim::point2di(), dim::size2di(1024, 1024));
    
    __spVideoDriver->clearBuffers();
    {
        hCamera->doTurn(dim::vector3df(0, CamDegree, 0));
        hCamera->doMove(dim::vector3df(CamDist, 0, 0));
    }
    __spVideoDriver->setColorMask(true, true, true, true);
    
    /* Drawing */
    
    __spVideoDriver->beginDrawing2D();
    {
        __spVideoDriver->draw2DImage(spacetex1, dim::rect2di(0, __Screen_Height-1024, 1024, 1024), dim::rect2df(0, 1, 1, 0), video::color(255, 255, 255, 128), true);
        __spVideoDriver->draw2DImage(spacetex2, dim::rect2di(0, __Screen_Height-1024, 1024, 1024), dim::rect2df(0, 1, 1, 0), video::color(255, 255, 255, 128), true);
    }
    __spVideoDriver->endDrawing2D();
}

/* Clears the whole world */

void SceneManager::clearScene(
    bool isDeleteEntitis, bool isDeleteCameras, bool isDeleteLights,
    bool isDeleteSprites, bool isDeleteTerrains)
{
    if (isDeleteEntitis)
    {
        for (std::list<Entity*>::iterator it = EntityList_.begin(); it != EntityList_.end(); ++it)
            spDeletePointer(*it);
        EntityList_.clear();
        memset(&EntityList_, 0, sizeof(EntityList_));
    }
    if (isDeleteCameras)
    {
        for (std::list<Camera*>::iterator it = CameraList_.begin(); it != CameraList_.end(); ++it)
            spDeletePointer(*it);
        CameraList_.clear();
        memset(&EntityList_, 0, sizeof(CameraList_));
    }
    if (isDeleteLights)
    {
        for (std::list<Light*>::iterator it = LightList_.begin(); it != LightList_.end(); ++it)
            spDeletePointer(*it);
        LightList_.clear();
        memset(&EntityList_, 0, sizeof(LightList_));
    }
    if (isDeleteSprites)
    {
        for (std::list<Sprite*>::iterator it = SpriteList_.begin(); it != SpriteList_.end(); ++it)
            spDeletePointer(*it);
        SpriteList_.clear();
        memset(&EntityList_, 0, sizeof(SpriteList_));
    }
    if (isDeleteTerrains)
    {
        for (std::list<Terrain*>::iterator it = TerrainList_.begin(); it != TerrainList_.end(); ++it)
            spDeletePointer(*it);
        TerrainList_.clear();
        memset(&EntityList_, 0, sizeof(TerrainList_));
    }
}

/* Sort the entity list by its depths */

void SceneManager::sortDepthEntityList(Camera* hCamera)
{
    hCamera->updateModelviewMatrix();
    
    for (std::list<Entity*>::iterator itEntity = EntityList_.begin(); itEntity != EntityList_.end(); ++itEntity)
    {
        if ((*itEntity)->getVisible())
        {
            spWorldMatrix.matrixReset();
            
            (*itEntity)->updateGlobalLocation();
            
            (*itEntity)->PerspectiveDistance_ = (spViewMatrix * spWorldMatrix.getPosition()).Z;
        }
    }
    
    if (isDepthSortFast_)
        EntityList_.sort(sortEntitiesFast);
    else
        EntityList_.sort(sortEntitiesDefault);
}

Entity* SceneManager::copyNode(Entity* hEntity)
{
    Entity* NewEntity = hEntity->copy();
    EntityList_.push_back(NewEntity);
    return NewEntity;
}
Light* SceneManager::copyNode(Light* hLight)
{
    
}
Sprite* SceneManager::copyNode(Sprite* hSprite)
{
    
}
Camera* SceneManager::copyNode(Camera* hCamera)
{
    
}
Terrain* SceneManager::copyNode(Terrain* hTerrain)
{
    
}

bool SceneManager::deleteNode(Node* hNode)
{
    if (!hNode)
        return false;
    
    switch (hNode->getType())
    {
        
        case NODE_ENTITY:
        {
            Entity* hEntity = (Entity*)hNode;
            for (std::list<Entity*>::iterator it = EntityList_.begin(); it != EntityList_.end(); ++it)
            {
                if (*it == hEntity)
                {
                    spDeletePointer(hEntity);
                    EntityList_.erase(it);
                    return true;
                }
            }
        }
        break;
        
        case NODE_CAMERA:
        {
            Camera* hCamera = (Camera*)hNode;
            for (std::list<Camera*>::iterator it = CameraList_.begin(); it != CameraList_.end(); ++it)
            {
                if (*it == hCamera)
                {
                    spDeletePointer(hCamera);
                    CameraList_.erase(it);
                    return true;
                }
            }
        }
        break;
        
        case NODE_LIGHT:
        {
            Light* hLight = (Light*)hNode;
            for (std::list<Light*>::iterator it = LightList_.begin(); it != LightList_.end(); ++it)
            {
                if (*it == hLight)
                {
                    spDeletePointer(hLight);
                    LightList_.erase(it);
                    return true;
                }
            }
        }
        break;
        
        case NODE_SPRITE:
        {
            Sprite* hSprite = (Sprite*)hNode;
            for (std::list<Sprite*>::iterator it = SpriteList_.begin(); it != SpriteList_.end(); ++it)
            {
                if (*it == hSprite)
                {
                    spDeletePointer(hSprite);
                    SpriteList_.erase(it);
                    return true;
                }
            }
        }
        break;
        
        case NODE_TERRAIN:
        {
            Terrain* hTerrain = (Terrain*)hNode;
            for (std::list<Terrain*>::iterator it = TerrainList_.begin(); it != TerrainList_.end(); ++it)
            {
                if (*it == hTerrain)
                {
                    spDeletePointer(hTerrain);
                    TerrainList_.erase(it);
                    return true;
                }
            }
        }
        break;
        
        case NODE_JOINT:
        {
            for (std::list<Node*>::iterator it = NodeList_.begin(); it != NodeList_.end(); ++it)
            {
                if (*it == hNode)
                {
                    spDeletePointer(hNode);
                    NodeList_.erase(it);
                    return true;
                }
            }
        }
        break;
        
    }
    
    return false;
}

std::list<Node*> SceneManager::findChildren(Node* RootNode)
{
    std::list<Node*> NodeList;
    
    /* Entities */
    for (std::list<Entity*>::iterator it = EntityList_.begin(); it != EntityList_.end(); ++it)
        if ((*it)->getParent() == RootNode)
            NodeList.push_back(*it);
    
    /* Cameras */
    for (std::list<Light*>::iterator it = LightList_.begin(); it != LightList_.end(); ++it)
        if ((*it)->getParent() == RootNode)
            NodeList.push_back(*it);
    
    /* Lights */
    for (std::list<Sprite*>::iterator it = SpriteList_.begin(); it != SpriteList_.end(); ++it)
        if ((*it)->getParent() == RootNode)
            NodeList.push_back(*it);
    
    /* Sprites */
    for (std::list<Camera*>::iterator it = CameraList_.begin(); it != CameraList_.end(); ++it)
        if ((*it)->getParent() == RootNode)
            NodeList.push_back(*it);
    
    /* Terrains */
    for (std::list<Terrain*>::iterator it = TerrainList_.begin(); it != TerrainList_.end(); ++it)
        if ((*it)->getParent() == RootNode)
            NodeList.push_back(*it);
    
    return NodeList;
}

Node* SceneManager::findChild(Node* RootNode, io::stringc Name)
{
    /* Entities */
    for (std::list<Entity*>::iterator it = EntityList_.begin(); it != EntityList_.end(); ++it)
        if ((*it)->getParent() == RootNode && (*it)->getName() == Name)
            return *it;
    
    /* Cameras */
    for (std::list<Light*>::iterator it = LightList_.begin(); it != LightList_.end(); ++it)
        if ((*it)->getParent() == RootNode && (*it)->getName() == Name)
            return *it;
    
    /* Lights */
    for (std::list<Sprite*>::iterator it = SpriteList_.begin(); it != SpriteList_.end(); ++it)
        if ((*it)->getParent() == RootNode && (*it)->getName() == Name)
            return *it;
    
    /* Sprites */
    for (std::list<Camera*>::iterator it = CameraList_.begin(); it != CameraList_.end(); ++it)
        if ((*it)->getParent() == RootNode && (*it)->getName() == Name)
            return *it;
    
    /* Terrains */
    for (std::list<Terrain*>::iterator it = TerrainList_.begin(); it != TerrainList_.end(); ++it)
        if ((*it)->getParent() == RootNode && (*it)->getName() == Name)
            return *it;
    
    return 0;
}

void SceneManager::updateErasedNodes()
{
    /* Entities */
    for (std::list<Entity*>::iterator it = EntityList_.begin(); it != EntityList_.end(); ++it)
        if (*it == 0)
            EntityList_.erase(it);
    
    /* Cameras */
    for (std::list<Camera*>::iterator it = CameraList_.begin(); it != CameraList_.end(); ++it)
        if (*it == 0)
            CameraList_.erase(it);
    
    /* Lights */
    for (std::list<Light*>::iterator it = LightList_.begin(); it != LightList_.end(); ++it)
        if (*it == 0)
            LightList_.erase(it);
    
    /* Sprites */
    for (std::list<Sprite*>::iterator it = SpriteList_.begin(); it != SpriteList_.end(); ++it)
        if (*it == 0)
            SpriteList_.erase(it);
    
    /* Terrains */
    for (std::list<Terrain*>::iterator it = TerrainList_.begin(); it != TerrainList_.end(); ++it)
        if (*it == 0)
            TerrainList_.erase(it);
}


/*
 * ========== Extra functions: ==========
 */


void SceneManager::setWireframe(s32 Type)
{
    /* Object wireframe */
    for (std::list<Entity*>::iterator it = EntityList_.begin(); it != EntityList_.end(); ++it)
        (*it)->setWireframe(Type);
    for (std::list<Terrain*>::iterator it = TerrainList_.begin(); it != TerrainList_.end(); ++it)
        (*it)->setWireframe(Type);
}

void SceneManager::setWireframe(s32 TypeFront, s32 TypeBack)
{
    for (std::list<Entity*>::iterator it = EntityList_.begin(); it != EntityList_.end(); ++it)
        (*it)->setWireframe(TypeFront, TypeBack);
}

void SceneManager::setRenderFace(s32 Face)
{
    for (std::list<Entity*>::iterator it = EntityList_.begin(); it != EntityList_.end(); ++it)
        (*it)->setRenderFace(Face);
}

void SceneManager::setLighting(bool isLighting)
{
    /* Lighting setting */
    __isLighting = isLighting_ = isLighting;
}

u32 SceneManager::getSceneVerticesCount() const
{
    u32 VerticesCount = 0;
    for (std::list<Entity*>::const_iterator it = EntityList_.begin(); it != EntityList_.end(); ++it)
        VerticesCount += (*it)->getMeshVerticesCount();
    return VerticesCount;
}

u32 SceneManager::getSceneTriangleCount() const
{
    u32 TriangleCount = 0;
    for (std::list<Entity*>::const_iterator it = EntityList_.begin(); it != EntityList_.end(); ++it)
        TriangleCount += (*it)->getMeshTriangleCount();
    return TriangleCount;
}

u32 SceneManager::getSceneObjectsCount() const
{
    u32 Counter = 0;
    Counter += EntityList_.size();
    Counter += LightList_.size();
    Counter += SpriteList_.size();
    Counter += TerrainList_.size();
    return Counter;
}

// !!! UNTESTED !!!

std::vector<u32> SceneManager::getTriangleIndicesList(std::vector<u32> &IndicesList, ETriangleListTypes Type)
{
    /* Check if the types are equal */
    if (Type == TRIANGLE_LIST)
        return IndicesList;
    
    /* Temporary variables */
    std::vector<u32> NewList;
    
    /* Change to listed type */
    switch (Type)
    {
        case TRIANGLE_STRIP:
        {
            /* Loop all indices */
            for (s32 i = 0, s = 1; i < IndicesList.size() - 2; ++i)
            {
                /* Swap the side */
                s = 1 - s;
                
                /* Check which is the current side */
                if (s)
                {
                    /* Add the three indices */
                    NewList.push_back( IndicesList[i+0] );
                    NewList.push_back( IndicesList[i+1] );
                    NewList.push_back( IndicesList[i+2] );
                }
                else
                {
                    /* Add the three indices */
                    NewList.push_back( IndicesList[i+2] );
                    NewList.push_back( IndicesList[i+1] );
                    NewList.push_back( IndicesList[i+0] );
                }
            } // next index
            
        }
        break;
        
        case TRIANGLE_FAN:
        {
            /* Loop all indices */
            for (s32 i = 0, s = 1; i < IndicesList.size() - 2; ++i)
            {
                /* Add the three indices */
                NewList.push_back( IndicesList[0] );
                NewList.push_back( IndicesList[i+1] );
                NewList.push_back( IndicesList[i+2] );
            } // next index
        }
        break;
    }
    
    /* Return the new list */
    return NewList;
}


/*
 * Static functions
 */

SP_EXPORT void controlMovement(Node* hObj, f32 MoveSpeed, f32 TurnSpeed, bool isCamLike, bool isControlShooterLike)
{
    static f32 Pitch, Yaw;
    
    if (isControlShooterLike)
    {
        if (__spUserControl->isKeyPressed(KEY_D))
            hObj->doMove(dim::vector3df(MoveSpeed, 0, 0));
        if (__spUserControl->isKeyPressed(KEY_A))
            hObj->doMove(dim::vector3df(-MoveSpeed, 0, 0));
        if (__spUserControl->isKeyPressed(KEY_W))
            hObj->doMove(dim::vector3df(0, 0, MoveSpeed));
        if (__spUserControl->isKeyPressed(KEY_S))
            hObj->doMove(dim::vector3df(0, 0, -MoveSpeed));
    }
    else
    {
        if (__spUserControl->isKeyPressed(KEY_RIGHT))
            hObj->doMove(dim::vector3df(MoveSpeed, 0, 0));
        if (__spUserControl->isKeyPressed(KEY_LEFT))
            hObj->doMove(dim::vector3df(-MoveSpeed, 0, 0));
        if (__spUserControl->isKeyPressed(KEY_UP))
            hObj->doMove(dim::vector3df(0, 0, MoveSpeed));
        if (__spUserControl->isKeyPressed(KEY_DOWN))
            hObj->doMove(dim::vector3df(0, 0, -MoveSpeed));
    }
    
    Pitch += (f32)__spUserControl->getCursorSpeed().Y * TurnSpeed;
    Yaw   += (f32)__spUserControl->getCursorSpeed().X * TurnSpeed;
    
    if (isCamLike)
    {
        if (Pitch > 90)
            Pitch = 90;
        else if (Pitch < -90)
            Pitch = -90;
    }
    
    hObj->setRotation(dim::vector3df(Pitch, Yaw, 0));
    __spUserControl->setCursorPosition(dim::point2di(__Screen_Width/2, __Screen_Height/2));
}


} // /namespace scene

} // /namespace sp








// ================================================================================
