/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * Header for all scene nodes (nodes, entities, cameras, lights, sprites)
 *
 */

#ifndef __SP_SCENEMANAGER_H__
#define __SP_SCENEMANAGER_H__


#include "spStandard.hpp"
#include "spSceneManagerEntity.hpp"
#include "spSceneManagerCamera.hpp"
#include "spSceneManagerLight.hpp"
#include "spSceneManagerSprite.hpp"
#include "spSceneManagerTerrain.hpp"
#include "spSceneManagerModelCreator.hpp"
#include "spSceneManagerCollisionDetector.hpp"
#include "spSceneManagerAnimator.hpp"
#include "spSceneManagerSpecialEffectsActor.hpp"
#include "spVideoDriverShader.hpp"
#include "spInputOutputString.hpp"
#include "spSensitiveSound.hpp"
#include "spStencilManager.hpp"
#include "spModelManager.hpp"

#include <list>
#include <io.h>


namespace sp
{
namespace scene
{


static const s32 DEF_MODEL_DETAIL       = 9;
static const s32 DEF_MODEL_TESSELATION  = 8;

enum EModelTypes
{
    MODEL_AUTODETECT = 0,
    
    MODEL_SPM,
    MODEL_3DS,
    MODEL_MS3D,
    MODEL_D3D,
    MODEL_B3D,
    MODEL_MD2,
    MODEL_OBJ,
    MODEL_TMF,
};

enum ESceneTypes
{
    SCENE_AUTODETECT = 0,
    
    SCENE_SPS,
    SCENE_BSP,
};

enum ETriangleListTypes
{
    TRIANGLE_LIST,
    TRIANGLE_STRIP,
    TRIANGLE_FAN
};


class SP_EXPORT SceneManager
{
    
    public:
        
        SceneManager();
        ~SceneManager();
        
        /* Stencil manager for real-time shadow and clipping mirrors */
        StencilManager* getStencilManager();
        
        /* Animator for node-/ keyframe-/ and skeletal animations */
        anim::Animator* getAnimator();
        
        /* Special effects actor for special effects: Blur effect, Fur effect, Bump mapping, Cel (cartoon) shading */
        effect::SpecialEffectsActor* getSpecialEffectsActor();
        
        /* Create simple node */
        Node* createJoint();
        
        /* Create a pivot (empty entity) */
        Entity* createPivot(); // Empty 3d model
        
        /* Create a standard model */
        Entity* createModel(u32 Type, s32 Detail = DEF_MODEL_DETAIL, f32 Radius1 = 0.5f, f32 Radius2 = 0.25f, bool isCap = true); // Cube ... Teapot
        
        /* Create a standard model (for spiral with more options) */
        Entity* createModel(u32 Type, s32 Detail, f32 Radius1, f32 Radius2, f32 DegreeLength, f32 Size, bool isCap); // Spiral
        
        /* Create a supershape (such as a shell, sphere ...) */
        Entity* createSuperShape(f32 ValueList[12], s32 Detail = DEF_MODEL_DETAIL);
        
        /* Create a skybox (back, front, top, bottom, left, right) */
        Entity* createSkyBox(video::Texture* TextureList[6], f32 Radius = 100.0f, u8 Type = SKYBOX_NORMAL);
        
        /* Create a highmap (such as a terrain) */
        Entity* createHeightField(video::Texture* hTextureHighMap, s32 Detail = DEF_MODEL_DETAIL);
        
        /* Create a model of a model list (e.g. for map loading) */
        Entity* createModelList(std::vector<Entity*> ModelList, bool isOldDelete = true);
        
        /* Create a model of the specified model surface */
        Entity* createModelSurface(Entity* Model, u32 Surface);
        
        /* Load & save a model */
        Entity* loadModel(io::stringc Filename, io::stringc TexturePath = "?", EModelTypes FileType = MODEL_AUTODETECT);
        bool saveModel(Entity* hEntity, io::stringc Filename, EModelTypes FileType = MODEL_AUTODETECT);
        
        /* Load & save a scene */
        Entity* loadScene(
            io::stringc Filename, io::stringc TexturePath = "?", ESceneTypes FileType = SCENE_AUTODETECT,
            s32 Tesselation = DEF_MODEL_TESSELATION
        );
        bool saveScene(io::stringc Filename);
        
        /* Cameras (to see the scene at least one camera is needed) */
        Camera* createCamera();
        
        /* Lights (static, volumetric, spot) */
        Light* createLight(ELightTypes Type = LIGHT_DIRECTIONAL);
        
        /* Sprites (particles for water-/ dash-/ sparks-/ cloud effects) */
        Sprite* createSprite(video::Texture* hTexture);
        
        /* Terrains (high mapping based entities) */
        Terrain* createTerrain(video::Texture* hTextureHighMap, video::Texture* hTexture, f32 ChangeDist = __SP_TERRAIN_STD_CHANGEDIST);
        
        /* Nodes (entities, lights, sprites, cameras, terrains) */
        Entity* copyNode(Entity* hEntity);
        Light* copyNode(Light* hLight);
        Sprite* copyNode(Sprite* hSprite);
        Camera* copyNode(Camera* hCamera);
        Terrain* copyNode(Terrain* hTerrain);
        
        /* Delete nodes */
        bool deleteNode(Node* hNode);
        
        /* Children control */
        std::list<Node*> findChildren(Node* RootNode);
        Node* findChild(Node* RootNode, io::stringc Name);
        
        /* After deleting a node manual */
        void updateErasedNodes();
        
        /* Renders the whole scene */
        void renderScene(Camera* hCamera);
        
        /* Renders the whole scene (for all cameras) */
        void renderScene();
        
        /* Virtual reality (use a glasses with a red-/ and a green glass) */
        void renderSceneVirtualReality(Camera* hCamera, f32 CamDegree, f32 CamDist = 0.25f); // Renders the scene two times (makes the program slower)
        
        /* Clears all entities, sprites and lights etc. */
        void clearScene(
            bool isDeleteEntitis = true, bool isDeleteCameras = true,
            bool isDeleteLights = true, bool isDeleteSprites = true,
            bool isDeleteTerrains = true
        );
        
        /* Sort the object lists */
        void sortDepthEntityList(Camera* hCamera);
        
        /* ===== Extra functions ===== */
        
        /* Render modes */
        void setWireframe(s32 Type);
        void setWireframe(s32 TypeFront, s32 TypeBack);
        void setRenderFace(s32 Face);
        
        /* Lighting */
        void setLighting(bool isLighting = true);
        
        inline void setStencilEffects(bool isStencilEffects = true)
        {
            isStencilEffects_ = isStencilEffects;
        }
        
        /* Get lists - objects */
        inline std::list<Entity*> getEntityList() const
        {
            return EntityList_;
        }
        inline std::list<Light*> getLightList() const
        {
            return LightList_;
        }
        inline std::list<Sprite*> getSpriteList() const
        {
            return SpriteList_;
        }
        inline std::list<Camera*> getCameraList() const
        {
            return CameraList_;
        }
        inline std::list<Terrain*> getTerrainList() const
        {
            return TerrainList_;
        }
        inline std::list<Node*> getJointList() const
        {
            return NodeList_;
        }
        
        /* Active camera */
        inline void setActiveCamera(Camera* ActiveCamera)
        {
            ActiveCamera_ = ActiveCamera;
        }
        inline Camera* getActiveCamera() const
        {
            return ActiveCamera_;
        }
        
        // "Useful for sortware-rendering"
        inline void setDepthSortFast(bool UseFastMethod)
        {
            isDepthSortFast_ = UseFastMethod;
        }
        inline bool getDepthSortFast() const
        {
            return isDepthSortFast_;
        }
        
        /* Count lists */
        u32 getSceneVerticesCount() const; // (Entities, Terrains)
        u32 getSceneTriangleCount() const; // (Entities, Terrains)
        u32 getSceneObjectsCount() const; // (Entities, Lights, Sprites, Terrains)
        
        std::vector<u32> getTriangleIndicesList(std::vector<u32> &IndicesList, ETriangleListTypes Type);
        
    private:
        
        /* ===== Members ===== */
        
        /* All object lists (3d scene objects, lights etc.) */
        std::list<Entity*> EntityList_;
        std::list<Light*> LightList_;
        std::list<Sprite*> SpriteList_;
        std::list<Camera*> CameraList_;
        std::list<Terrain*> TerrainList_;
        std::list<Node*> NodeList_;
        
        Camera* ActiveCamera_;
        
        /* Information members */
        bool isLighting_, isStencilEffects_;
        bool isDepthSortFast_;
        
};

extern StencilManager* __spStencilManager;
extern anim::Animator* __spAnimator;
extern effect::SpecialEffectsActor* __spSpecialEffectsActor;


/*
 * Static functions
 */

/* Object movement simple control (for debugging) */
SP_EXPORT void controlMovement(
    Node* hObj, f32 MoveSpeed = 0.25f, f32 TurnSpeed = 0.25f, bool isCamLike = true, bool isControlShooterLike = false
);


} // /namespace scene

} // /namespace sp


#endif







// ================================================================================
