/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * Animations
 *
 */

#include "spSceneManagerAnimation.hpp"
#include "spSceneManagerAnimator.hpp"
#include "spModelLoaderMD2.hpp"
#include "spMainDevice.hpp"


namespace sp
{

extern video::VideoDriver* __spVideoDriver;

namespace scene
{
namespace anim
{


/*
 * ========== Animation skeleton bone class ==========
 */

SAnimSkeletonBone::SAnimSkeletonBone(
    u32 &BoneIDCounter, SAnimSkeletonBone* hParent,
    dim::vector3df BoneTranslation, dim::quaternion BoneQuaternion, dim::vector3df BoneScale,
    io::stringc BoneName)
    : ParentBone(0)
{
    ID          = ++BoneIDCounter;
    Name        = BoneName;
    
    Translation = BoneTranslation;
    Quat        = BoneQuaternion;
    Scale       = BoneScale;
    
    ParentBone  = hParent;
    
    MatrixOrigin.matrixReset();
}
SAnimSkeletonBone::~SAnimSkeletonBone()
{
}

void SAnimSkeletonBone::transformMatrix(dim::matrix4f &Mat, dim::matrix4f &MatNormal)
{
    /* Parent bone */
    if (ParentBone)
        ParentBone->transformMatrix(Mat, MatNormal);
    
    /* Translation */
    Mat.matrixTranslate(Translation);
    
    /* Quaternion (for linear interpolation) */
    Mat *= QuatMatrix;
    
    /* Normal - rotation */
    MatNormal *= QuatMatrix;
    
    /* Scale */
    Mat.matrixScale(Scale);
}

void SAnimSkeletonBone::getBoneTranslation(dim::vector3df &BoneTranslation)
{
    dim::matrix4f Mat, MatNormal;
    
    Mat.matrixReset();
    transformMatrix(Mat, MatNormal);
    
    BoneTranslation = Mat.getPosition();
}


/*
 * ========== Animation class ==========
 */

Animation::Animation()
{
    init();
}
Animation::Animation(Node* hNode)
{
    init();
    
    hNode_      = hNode;
}
Animation::Animation(Entity* hEntity)
{
    init();
    
    hNode_      = hEntity;
    hEntity_    = hEntity;
}
Animation::~Animation()
{
}

u32 Animation::addAnimSeqNode(dim::vector3df Position, dim::quaternion Rotation, dim::vector3df Scale)
{
    SAnimNode AnimSeqData;
    {
        AnimSeqData.Index       = (AnimNodeList_.size() ? AnimNodeList_[AnimNodeList_.size()-1].Index + 1 : 1);
        AnimSeqData.Position    = Position;
        AnimSeqData.Rotation    = Rotation;
        AnimSeqData.Scale       = Scale;
    }
    AnimNodeList_.push_back(AnimSeqData);
    
    return AnimSeqData.Index;
}
void Animation::removeAnimSeqNode(u32 SeqIndex)
{
    for (std::vector<SAnimNode>::iterator it = AnimNodeList_.begin(); it != AnimNodeList_.end(); ++it)
    {
        if (it->Index == SeqIndex)
        {
            AnimNodeList_.erase(it);
            break;
        }
    }
}

u32 Animation::addAnimSeqMesh(std::vector<dim::vector3df> VerticesCoord)
{
    SAnimMesh AnimInfoData;
    {
        AnimInfoData.VerticesCoord  = VerticesCoord;
        LastMeshVertices_           = VerticesCoord;
        NextMeshVertices_           = VerticesCoord;
    }
    AnimMeshList_.push_back(AnimInfoData);
    
    return AnimMeshList_.size() - 1;
}
void Animation::removeAnimSeqMesh(u32 SeqIndex)
{
    std::vector<SAnimMesh>::iterator itSeq = AnimMeshList_.begin() + SeqIndex;
    AnimMeshList_.erase(itSeq);
}

u32 Animation::addAnimSkeletonBone(
    u32 ParentBoneID,
    dim::vector3df Translation, dim::quaternion Quaternion, dim::vector3df Scale,
    io::stringc Name)
{
    SAnimSkeletonBone* ParentBone = 0;
    
    for (std::vector<SAnimSkeleton>::iterator it = AnimSkeletonList_.begin(); it != AnimSkeletonList_.end(); ++it)
    {
        if (it->SkeletonBone->ID == ParentBoneID)
        {
            ParentBone = it->SkeletonBone;
            break;
        }
    }
    
    SAnimSkeleton AnimBoneData;
    {
        AnimBoneData.SkeletonBone = new SAnimSkeletonBone(SkeletonBoneID_, ParentBone, Translation, Quaternion, Scale, Name);
    }
    AnimSkeletonList_.push_back(AnimBoneData);
    
    return AnimBoneData.SkeletonBone->ID;
}
void Animation::removeAnimSkeletonBone(u32 BoneID)
{
    for (std::vector<SAnimSkeleton>::iterator it = AnimSkeletonList_.begin(); it != AnimSkeletonList_.end(); ++it)
    {
        if (it->SkeletonBone->ID == BoneID)
        {
            AnimSkeletonList_.erase(it);
            break;
        }
    }
}

void Animation::doAnimate(s8 AnimType, s32 AnimTime, s32 Start, s32 End)
{
    doAnimateNode(AnimType, AnimTime, Start, End);
    doAnimateMesh(AnimType, AnimTime, Start, End);
    doAnimateSkeleton(AnimType, (f32)AnimTime / DEF_ANIMMESH_SPEED, Start, End);
}

void Animation::doAnimateNode(s8 AnimType, s32 AnimTime, s32 Start, s32 End)
{
    const f32 CurFPS = 60.0f; // !!!
    
    if (AnimNodeList_.size() > 0)
    {
        // General settings
        AnimType_           = (s8)AnimType;
        AnimModeNode_       = ANIM_FOREWARDS;
        AnimSpeed_          = 1000.0f / (CurFPS*AnimTime);
        AnimSeekNode_       = 0.0f;
        NodeStart_          = Start;
        NodeEnd_            = End ? End : AnimNodeList_.size() - 1;
        AnimSeekSeqNode_    = Start + 1;
        
        // Process next sequence
        updateNextSeqNode();
    }
}

void Animation::doAnimateMesh(s8 AnimType, s32 AnimTime, s32 Start, s32 End)
{
    const f32 CurFPS = 60.0f; // !!!
    
    if (AnimMeshList_.size() > 0)
    {
        // General settings
        AnimType_           = (s8)AnimType;
        AnimModeMesh_       = ANIM_FOREWARDS;
        AnimSpeed_          = 1000.0f / (CurFPS*AnimTime);
        AnimSeekMesh_       = 0.0f;
        MeshStart_          = Start;
        MeshEnd_            = End ? End : AnimMeshList_.size() - 1;
        AnimSeekSeqMesh_    = Start + 1;
        
        // Process next sequence
        updateNextSeqMesh();
    }
}

void Animation::doAnimateSkeleton(s8 AnimType, f32 AnimSpeed, s32 Start, s32 End)
{
    if (AnimSkeletonList_.size() > 0)
    {
        std::vector<SAnimSkeleton>::iterator it = AnimSkeletonList_.begin();
        
        // General settings
        AnimTypeSkeleton_   = AnimType;
        SkeletonStart_      = Start;
        SkeletonEnd_        = End ? End : it->FrameList.size() - 1;
        AnimSpeedSkeleton_  = AnimSpeed;
        
        // Reset all bones
        for (; it != AnimSkeletonList_.end(); ++it)
        {
            it->Interpolation   = 0.0f;
            it->Sequence        = Start;
        }
        
        // Process next sequence
        updateNextSeqSkeleton();
    }
}

void Animation::setAnimSeekNode(f32 AnimSeek, u32 AnimSeekSeq)
{
    AnimSeekNode_       = AnimSeek;
    AnimSeekSeqNode_    = AnimSeekSeq;
}
void Animation::setAnimSeekMesh(f32 AnimSeek, u32 AnimSeekSeq)
{
    AnimSeekMesh_       = AnimSeek;
    AnimSeekSeqMesh_    = AnimSeekSeq;
}
void Animation::setAnimSeekSkeleton(f32 AnimSeek, u32 AnimSeekSeq)
{
    for (std::vector<SAnimSkeleton>::iterator it = AnimSkeletonList_.begin(); it != AnimSkeletonList_.end(); ++it)
    {
        it->Interpolation   = AnimSeek;
        it->Sequence        = AnimSeekSeq;
    }
}

void Animation::getAnimSeekNode(f32 &AnimSeek, s32 &AnimSeekSeq)
{
    AnimSeek    = AnimSeekNode_;
    AnimSeekSeq = AnimSeekSeqNode_;
}
void Animation::getAnimSeekMesh(f32 &AnimSeek, s32 &AnimSeekSeq)
{
    AnimSeek    = AnimSeekMesh_;
    AnimSeekSeq = AnimSeekSeqMesh_;
}
void Animation::getAnimSeekSkeleton(f32 &AnimSeek, s32 &AnimSeekSeq)
{
    if (AnimSkeletonList_.size() > 0)
    {
        std::vector<SAnimSkeleton>::iterator it = AnimSkeletonList_.begin();
        AnimSeek    = it->Interpolation;
        AnimSeekSeq = it->Sequence;
    }
}

void Animation::pauseAnim(bool isPaused)
{
    if (isPaused)
    {
        /* Node animation */
        if (AnimModeNode_ != ANIM_STOP)
            PausedAnimModeNode_ = AnimModeNode_;
        AnimModeNode_ = ANIM_STOP;
        
        /* Keyframe animation */
        if (AnimModeMesh_ != ANIM_STOP)
            PausedAnimModeMesh_ = AnimModeMesh_;
        AnimModeMesh_ = ANIM_STOP;
        
        /* Bone/ Skeletal animation */
        if (AnimTypeSkeleton_ != ANIM_SKELETON_STOP)
            PausedAnimModeSkeleton_ = AnimTypeSkeleton_;
        AnimTypeSkeleton_ = ANIM_SKELETON_STOP;
    }
    else
    {
        /* Node animation */
        if (PausedAnimModeNode_ != ANIM_STOP)
            AnimModeNode_ = PausedAnimModeNode_;
        
        /* Keyframe animation */
        if (PausedAnimModeMesh_ != ANIM_STOP)
            AnimModeMesh_ = PausedAnimModeMesh_;
        
        /* Bone/ Skeletal animation */
        if (PausedAnimModeSkeleton_ != ANIM_SKELETON_STOP)
            AnimTypeSkeleton_ = PausedAnimModeSkeleton_;
    }
}
void Animation::pauseNode(bool isPaused)
{
    if (isPaused)
    {
        if (AnimModeNode_ != ANIM_STOP)
            PausedAnimModeNode_ = AnimModeNode_;
        AnimModeNode_ = ANIM_STOP;
    }
    else
        AnimModeNode_ = PausedAnimModeNode_;
}
void Animation::pauseMesh(bool isPaused)
{
    if (isPaused)
    {
        if (AnimModeMesh_ != ANIM_STOP)
            PausedAnimModeMesh_ = AnimModeMesh_;
        AnimModeMesh_ = ANIM_STOP;
    }
    else
        AnimModeMesh_ = PausedAnimModeMesh_;
}
void Animation::pauseSkeleton(bool isPaused)
{
    if (isPaused)
    {
        if (AnimTypeSkeleton_ != ANIM_SKELETON_STOP)
            PausedAnimModeSkeleton_ = AnimTypeSkeleton_;
        AnimTypeSkeleton_ = ANIM_SKELETON_STOP;
    }
    else
        AnimTypeSkeleton_ = PausedAnimModeSkeleton_;
}

void Animation::stopAnim()
{
    // Node & Keyframe animation
    AnimModeNode_ = AnimModeMesh_ = ANIM_STOP;
    // Bone/ Skeletal animation
    AnimTypeSkeleton_ = ANIM_SKELETON_STOP;
}
void Animation::stopNode()
{
    AnimModeNode_ = ANIM_STOP;
}
void Animation::stopMesh()
{
    AnimModeMesh_ = ANIM_STOP;
}
void Animation::stopSkeleton()
{
    AnimTypeSkeleton_ = ANIM_SKELETON_STOP;
}


/*
 * =================== Updates the animation ===================
 * Node animation, Keyframe animation, Bone-/ Skeletal animation
 */

void Animation::updateAnimation()
{
    
    /* Node animation */
    if (AnimModeNode_ != ANIM_STOP)
        updateAnimationNode();
    
    /* Mesh animation (keyframe/ morph-target animation) */
    if (AnimModeMesh_ != ANIM_STOP)
        updateAnimationMesh();
    
    /* Skeletal animation (bone animation) */
    if (AnimTypeSkeleton_ != ANIM_SKELETON_STOP)
        updateAnimationSkeleton();
    
}


/*
 * ========== Private: ==========
 */


void Animation::init()
{
    /* Initialize the pointers */
    hNode_                  = 0;
    hEntity_                = 0;
    
    /* General settings */
    AnimType_               = ANIM_ONESHOT;
    AnimModeNode_           = ANIM_STOP;
    AnimModeMesh_           = ANIM_STOP;
    PausedAnimModeNode_     = ANIM_STOP;
    PausedAnimModeMesh_     = ANIM_STOP;
    PausedAnimModeSkeleton_ = ANIM_SKELETON_STOP;
    AnimTypeSkeleton_       = ANIM_SKELETON_STOP;
    AnimSpeed_              = 0.0f;
    AnimSeekNode_           = 0.0f;
    AnimSeekMesh_           = 0.0f;
    SkeletonBoneID_         = 0;
    
    AnimSeekSeqNode_        = 0;
    AnimSeekSeqMesh_        = 0;
    NodeStart_              = 0;
    NodeEnd_                = 0;
    MeshStart_              = 0;
    MeshEnd_                = 0;
}

void Animation::updateNextSeqNode()
{
    /* Gets the latest position */
    if ( ( AnimType_ == ANIM_LOOP || AnimType_ == ANIM_ONESHOT ) && AnimSeekSeqNode_ == 0 )
    {
        LastNodeLocation_.Position  = AnimNodeList_[AnimSeekSeqNode_].Position;
        LastNodeLocation_.Rotation  = AnimNodeList_[AnimSeekSeqNode_].Rotation;
        LastNodeLocation_.Scale     = AnimNodeList_[AnimSeekSeqNode_].Scale;
        ++AnimSeekSeqNode_;
    }
    else if (AnimSeekSeqNode_ > 0)
    {
        LastNodeLocation_.Position  = AnimNodeList_[AnimSeekSeqNode_ - 1].Position;
        LastNodeLocation_.Rotation  = AnimNodeList_[AnimSeekSeqNode_ - 1].Rotation;
        LastNodeLocation_.Scale     = AnimNodeList_[AnimSeekSeqNode_ - 1].Scale;
    }
    
    /* Set the next sequence position */
    NextNodeLocation_.Position  = AnimNodeList_[AnimSeekSeqNode_].Position - LastNodeLocation_.Position;
    NextNodeLocation_.Rotation  = AnimNodeList_[AnimSeekSeqNode_].Rotation;
    NextNodeLocation_.Scale     = AnimNodeList_[AnimSeekSeqNode_].Scale    - LastNodeLocation_.Scale;
}

void Animation::updateNextSeqMesh()
{
    u32 i = 0;
    
    /* Gets the latest position */
    hEntity_->getVertexCoordArray( &(LastMeshVertices_)[0] );
    
    /* Set the next sequence position */
    for (std::vector<dim::vector3df>::iterator it = NextMeshVertices_.begin(); it != NextMeshVertices_.end(); ++it, ++i)
        *it = AnimMeshList_[AnimSeekSeqMesh_].VerticesCoord[i] - LastMeshVertices_[i];
}


/*
 * ========== Node animation ==========
 */

void Animation::updateAnimationNode()
{
    
    // Temporary variables
    dim::matrix4f Mat;
    dim::quaternion Quat;
    
    if (AnimModeNode_ == ANIM_FOREWARDS)
    {
        
        if ( ( AnimSeekNode_ += AnimSpeed_ ) >= 1.0f )
        {
            updateSeekStatus(AnimSeekNode_, AnimSeekSeqNode_);
            
            if (AnimSeekSeqNode_ > NodeEnd_)
            {
                switch (AnimType_)
                {
                    case ANIM_ONESHOT:
                    {
                        AnimModeNode_       = ANIM_STOP;
                        AnimSeekSeqNode_    = 0;
                    }
                    break;
                    
                    case ANIM_PINGPONG:
                    case ANIM_LOOP_PINGPONG:
                    {
                        AnimSeekSeqNode_    = NodeEnd_ - 1;
                        AnimModeNode_       = ANIM_BACKWARDS;
                    }
                    break;
                    
                    case ANIM_LOOP:
                    {
                        AnimSeekSeqNode_ = 0;
                    }
                    break;
                }
            }
            
            updateNextSeqNode();
        } // fi
        
        // Interpolate the rotation
        Quat.slerp(LastNodeLocation_.Rotation, NextNodeLocation_.Rotation, AnimSeekNode_);
        
        // Transform the matrix
        Mat.matrixReset();
        Mat.matrixTranslate(LastNodeLocation_.Position + NextNodeLocation_.Position * AnimSeekNode_);
        Mat *= Quat.getMatrix();
        Mat.matrixScale(LastNodeLocation_.Scale + NextNodeLocation_.Scale * AnimSeekNode_);
        
        // Set the matrix
        hNode_->setMatrix(Mat);
        
    }
    else if (AnimModeNode_ == ANIM_BACKWARDS)
    {
        
        if ( ( AnimSeekNode_ -= AnimSpeed_ ) <= -1.0f )
        {
            updateSeekStatus(AnimSeekNode_, AnimSeekSeqNode_);
            
            if (AnimSeekSeqNode_ < NodeStart_)
            {
                switch (AnimType_)
                {
                    case ANIM_PINGPONG:
                    {
                        AnimModeNode_       = ANIM_STOP;
                        AnimSeekSeqNode_    = 0;
                    }
                    break;
                    
                    case ANIM_LOOP_PINGPONG:
                    {
                        AnimSeekSeqNode_    = NodeStart_ + 1;
                        AnimModeNode_       = ANIM_FOREWARDS;
                        AnimSeekNode_       = 0.0f;
                    }
                    break;
                }
            }
            
            updateNextSeqNode();
        } // fi
        
        // Interpolate the rotation
        Quat.slerp(NextNodeLocation_.Rotation, LastNodeLocation_.Rotation, -AnimSeekNode_);
        
        // Transform the matrix
        Mat.matrixReset();
        Mat.matrixTranslate(LastNodeLocation_.Position - NextNodeLocation_.Position * AnimSeekNode_);
        Mat *= Quat.getMatrix();
        Mat.matrixScale(LastNodeLocation_.Scale - NextNodeLocation_.Scale * AnimSeekNode_);
        
        // Set the matrix
        hNode_->setMatrix(Mat);
        
    }
    
}


/*
 * ========== Keyframe animation ==========
 */

void Animation::updateAnimationMesh()
{
    
    if (AnimModeMesh_ == ANIM_FOREWARDS)
    {
        
        if ( ( AnimSeekMesh_ += AnimSpeed_ ) >= 1.0f )
        {
            updateSeekStatus(AnimSeekMesh_, AnimSeekSeqMesh_);
            
            if (AnimSeekSeqMesh_ > MeshEnd_)
            {
                switch (AnimType_)
                {
                    case ANIM_ONESHOT:
                    {
                        AnimModeMesh_       = ANIM_STOP;
                        AnimSeekSeqMesh_    = 0;
                    }
                    break;
                    
                    case ANIM_PINGPONG:
                    case ANIM_LOOP_PINGPONG:
                    {
                        AnimSeekSeqMesh_    = MeshEnd_ - 1;
                        AnimModeMesh_       = ANIM_BACKWARDS;
                    }
                    break;
                    
                    case ANIM_LOOP:
                    {
                        AnimSeekSeqMesh_    = MeshStart_;
                    }
                    break;
                }
            }
            
            updateNextSeqMesh();
        } // fi
        
    }
    else if (AnimModeMesh_ == ANIM_BACKWARDS)
    {
        
        /* Update sequence */
        if ( ( AnimSeekMesh_ -= AnimSpeed_ ) <= -1.0f )
        {
            updateSeekStatus(AnimSeekMesh_, AnimSeekSeqMesh_);
            
            if (AnimSeekSeqMesh_ < MeshStart_)
            {
                switch (AnimType_)
                {
                    case ANIM_PINGPONG:
                    {
                        AnimModeMesh_       = ANIM_STOP;
                        AnimSeekSeqMesh_    = 0;
                    }
                    break;
                    
                    case ANIM_LOOP_PINGPONG:
                    {
                        AnimSeekSeqMesh_    = MeshStart_ + 1;
                        AnimModeMesh_       = ANIM_FOREWARDS;
                        AnimSeekMesh_       = 0.0f;
                    }
                    break;
                }
            }
            
            updateNextSeqMesh();
        } // fi
        
    } // fi
    
    /* Update vertices */
    for (register u32 i = 0, j, k = 0; i < hEntity_->getMeshSurfaceCount(); ++i)
    {
        hEntity_->selectSurface(i);
        for (j = 0; j < hEntity_->getMeshVerticesCount(i); ++j, ++k)
            hEntity_->setVertexCoord( j, LastMeshVertices_[k] + NextMeshVertices_[k] * AnimSeekMesh_ );
    }
    
    /* Update modeling */
    hEntity_->updateModeling(false);
    
}


/*
 * ========== Bone/ Skeleton animation ==========
 */

void Animation::updateAnimationSkeleton()
{
    
    register std::vector<SAnimSkeleton>::iterator it;
    register std::vector<SAnimSkeletonBone::SAnimSkeletonBoneVertex>::iterator itv;
    
    /* 
     * Loop through all bones of the skeleton
     */
    for (it = AnimSkeletonList_.begin(); it != AnimSkeletonList_.end(); ++it)
    {
        
        /*
         * Update sequence:
         * Next step for the seek status until the next frame has been arived
         */
        if ( ( it->Interpolation += it->FrameList[ it->Sequence ].AnimSpeed * AnimSpeedSkeleton_ ) >= 1.0f )
        {
            /*
             * Subtract the seek status in dependet of the out-of-range factor
             */
            updateSeekStatus(it->Interpolation, it->Sequence);
            
            /* Check the end of animation */
            if (it->Sequence >= SkeletonEnd_ + 1)
            {
                while (it->Sequence >= SkeletonEnd_ + 1)
                    it->Sequence = SkeletonStart_ + (it->Sequence - SkeletonEnd_ - 1);
                
                if (AnimTypeSkeleton_ == ANIM_SKELETON_ONELOOP)
                {
                    AnimTypeSkeleton_ = ANIM_SKELETON_STOP;
                    return;
                }
                
                updateNextSeqSkeleton(it, it->Sequence, it->Sequence + 1);
            }
            else if (it->Sequence >= SkeletonEnd_)
            {
                if (AnimTypeSkeleton_ == ANIM_SKELETON_ONESHOT)
                {
                    AnimTypeSkeleton_ = ANIM_SKELETON_STOP;
                    return;
                }
                
                updateNextSeqSkeleton(it, it->Sequence, SkeletonStart_);
            }
            else
                updateNextSeqSkeleton(it, it->Sequence, it->Sequence + 1);
        }
        
        /* Initialize rotation matrix */
        SkeletonMatrix_.matrixReset();
        SkeletonNormalMatrix_.matrixReset();
        
        /* Next translation & scale keyframe transformation */
        it->SkeletonBone->Translation   = it->FrameList[it->Sequence].Translation  + it->NextTranslation   * it->Interpolation;
        it->SkeletonBone->Scale         = it->FrameList[it->Sequence].Scale        + it->NextScale         * it->Interpolation;
        
        /* Next quaternion & quat-matrix keyframe transformation */
        it->SkeletonBone->Quat.slerp(
            it->FrameList[it->Sequence].CurQuat,
            it->NextQuat,
            it->Interpolation
        );
        it->SkeletonBone->Quat.getMatrix(it->SkeletonBone->QuatMatrix);
        
        /* Transform the bone matrix */
        it->SkeletonBone->transformMatrix(SkeletonMatrix_, SkeletonNormalMatrix_);
        
        /* Multiply the bone matrix by the origin matrix */
        SkeletonMatrix_ *= it->SkeletonBone->MatrixOrigin;
        SkeletonNormalMatrix_ *= it->SkeletonBone->MatrixNormalOrigin;
        
        /* Update vertices & normals */
        for (itv = it->SkeletonBone->Vertices.begin(); itv != it->SkeletonBone->Vertices.end(); ++itv)
        {
            hEntity_->selectSurface( itv->Surface );
            hEntity_->setVertexCoord( itv->Index, SkeletonMatrix_ * itv->Coord );
            hEntity_->setVertexNormal( itv->Index, SkeletonNormalMatrix_ * itv->Normal );
        } // next vertex
        
    } // next bone
    
    /* Update modeling */
    hEntity_->updateModeling(false);
    
}


void Animation::updateSeekStatus(f32 &Seek, s32 &SeekSeq)
{
    s32 SeeqOutOfRange = (s32)floor(Seek);
    Seek    -= SeeqOutOfRange;
    SeekSeq += SeeqOutOfRange;
}


/*
 * ========== Pure skeleton functions ==========
 */

void Animation::updateNextSeqSkeleton()
{
    for (std::vector<SAnimSkeleton>::iterator it = AnimSkeletonList_.begin(); it != AnimSkeletonList_.end(); ++it)
    {
        if (it->FrameList.size() >= 2)
        {
            it->NextTranslation = it->FrameList[it->Sequence + 1].Translation  - it->FrameList[it->Sequence].Translation;
            it->NextScale       = it->FrameList[it->Sequence + 1].Scale        - it->FrameList[it->Sequence].Scale;
            it->NextQuat        = it->FrameList[it->Sequence + 1].CurQuat;
        }
    } // next bone
}

void Animation::updateNextSeqSkeleton(std::vector<SAnimSkeleton>::iterator it, s32 IndexFrom, s32 IndexTo)
{
    it->NextTranslation = it->FrameList[IndexTo].Translation - it->FrameList[IndexFrom].Translation;
    it->NextScale       = it->FrameList[IndexTo].Scale       - it->FrameList[IndexFrom].Scale;
    it->NextQuat        = it->FrameList[IndexTo].CurQuat;
}

/*
 * Calculates the origin matrxi transformatino for all bones
 * (Call this function after all bone creations and settings)
 */
void Animation::updateSkeletonBones()
{
    /* Loop through all bones */
    for (std::vector<SAnimSkeleton>::iterator it = AnimSkeletonList_.begin(); it != AnimSkeletonList_.end(); ++it)
    {
        /* Update rotation matrix */
        it->SkeletonBone->MatrixOrigin.matrixReset();
        it->SkeletonBone->MatrixNormalOrigin.matrixReset();
        
        it->SkeletonBone->Quat.getMatrix(it->SkeletonBone->QuatMatrix);
        
        /* Get the inverse matrix for the origin matrices */
        it->SkeletonBone->QuatMatrix.setInverse();
        
        it->SkeletonBone->transformMatrix(it->SkeletonBone->MatrixOrigin, it->SkeletonBone->MatrixNormalOrigin);
        
        it->SkeletonBone->MatrixOrigin.setInverse();
        it->SkeletonBone->MatrixNormalOrigin.setInverse();
    } // next bone
}

SAnimSkeleton Animation::getSkeleton(u32 BoneID)
{
    for (std::vector<SAnimSkeleton>::iterator it = AnimSkeletonList_.begin(); it != AnimSkeletonList_.end(); ++it)
    {
        if (it->SkeletonBone->ID == BoneID)
            return *it;
    }
    
    return SAnimSkeleton();
}

SAnimSkeletonBone* Animation::getSkeletonBone(u32 BoneID)
{
    for (std::vector<SAnimSkeleton>::iterator it = AnimSkeletonList_.begin(); it != AnimSkeletonList_.end(); ++it)
    {
        if (it->SkeletonBone->ID == BoneID)
            return it->SkeletonBone;
    }
    
    return 0;
}

// !!! MUST BE UPDATED !!!
void Animation::setSkeletonBoneFrame(
    u32 BoneID,
    std::vector<dim::vector3df> FrameList,
    std::vector<f32> SpeedList)
{
    const f32 tFPS = 60.0f; // !!!
    for (std::vector<SAnimSkeleton>::iterator it = AnimSkeletonList_.begin(); it != AnimSkeletonList_.end(); ++it)
    {
        if (it->SkeletonBone->ID == BoneID)
        {
            it->FrameList.clear();
            
            SAnimSkeleton::SAnimSkeletonFrame FrameData;
            
            std::vector<dim::vector3df>::iterator itf1;
            std::vector<f32>::iterator itf2;
            
            for (itf1 = FrameList.begin(), itf2 = SpeedList.begin();
                 itf1 != FrameList.end() && itf2 != SpeedList.end(); ++itf1, ++itf2)
            {
                //FrameData.Rotation = *itf1;
                FrameData.AnimSpeed = 1000.0f / (tFPS*(*itf2));
                it->FrameList.push_back(FrameData);
            }
            
            break;
        }
    } // next bone
}
u32 Animation::addSkeletonBoneFrame(
    u32 BoneID,
    dim::vector3df Translation, dim::quaternion Quaternion, dim::vector3df Scale,
    f32 Speed)
{
    const f32 tFPS = 60.0f; // !!!
    
    for (std::vector<SAnimSkeleton>::iterator it = AnimSkeletonList_.begin(); it != AnimSkeletonList_.end(); ++it)
    {
        if (it->SkeletonBone->ID == BoneID)
        {
            SAnimSkeleton::SAnimSkeletonFrame FrameData;
            {
                FrameData.Translation   = Translation;
                FrameData.CurQuat       = Quaternion;
                FrameData.Scale         = Scale;
                
                FrameData.AnimSpeed     = 1000.0f / (tFPS*Speed);
            }
            it->FrameList.push_back(FrameData);
            
            return it->FrameList.size() - 1;
        }
    } // next bone
    
    return 0;
}
void Animation::removeSkeletonBoneFrame(u32 BoneID, u32 FrameIndex)
{
    for (std::vector<SAnimSkeleton>::iterator it = AnimSkeletonList_.begin(); it != AnimSkeletonList_.end(); ++it)
    {
        if (it->SkeletonBone->ID == BoneID)
        {
            std::vector<SAnimSkeleton::SAnimSkeletonFrame>::iterator itf = it->FrameList.begin() + FrameIndex;
            it->FrameList.erase(itf);
            break;
        }
    } // next bone
}

void Animation::addSkeletonBoneVertices(u32 BoneID, std::vector<u32> VerticesIndices, std::vector<u32> VerticesSurfaces)
{
    SAnimSkeletonBone::SAnimSkeletonBoneVertex AnimVertexData;
    
    /* Add all vertices which are dependent of the specified bone */
    for (std::vector<SAnimSkeleton>::iterator it = AnimSkeletonList_.begin(); it != AnimSkeletonList_.end(); ++it)
    {
        if (it->SkeletonBone->ID == BoneID)
        {
            /* Loop for all vertices */
            for (std::vector<u32>::iterator itv = VerticesIndices.begin(), its = VerticesSurfaces.begin();
                 itv != VerticesIndices.end(); ++itv, ++its)
            {
                hEntity_->selectSurface(*its);
                
                AnimVertexData.Index    = *itv;
                AnimVertexData.Surface  = *its;
                AnimVertexData.Coord    = hEntity_->getVertexCoord(*itv);
                AnimVertexData.Normal   = hEntity_->getVertexNormal(*itv);
                
                it->SkeletonBone->Vertices.push_back( AnimVertexData );
            } // next vertex
            
            break;
        } // fi
    } // next bone
}

void Animation::updateSkeletonSequence(bool isUpdateSeek)
{
    
    /* Loop through all bones of the skeleton */
    for (std::vector<SAnimSkeleton>::iterator it = AnimSkeletonList_.begin(); it != AnimSkeletonList_.end(); ++it)
    {
        
        /* Check if the seek states shall be updates */
        if (isUpdateSeek)
        {
            if (it->Sequence == SkeletonEnd_)
            {
                it->NextTranslation = it->FrameList[SkeletonStart_].Translation - it->FrameList[it->Sequence].Translation;
                it->NextScale       = it->FrameList[SkeletonStart_].Scale       - it->FrameList[it->Sequence].Scale;
                it->NextQuat        = it->FrameList[SkeletonStart_].CurQuat;
            }
            else
            {
                it->NextTranslation = it->FrameList[it->Sequence + 1].Translation    - it->FrameList[it->Sequence].Translation;
                it->NextScale       = it->FrameList[it->Sequence + 1].Scale          - it->FrameList[it->Sequence].Scale;
                it->NextQuat        = it->FrameList[it->Sequence + 1].CurQuat;
            }
            
            /* Next translation & scale keyframe transformation */
            it->SkeletonBone->Translation   = it->FrameList[it->Sequence].Translation  + it->NextTranslation   * it->Interpolation;
            it->SkeletonBone->Scale         = it->FrameList[it->Sequence].Scale        + it->NextScale         * it->Interpolation;
            
            /* Next quaternion & quat-matrix keyframe transformation */
            it->SkeletonBone->Quat.slerp(
                it->FrameList[it->Sequence].CurQuat,
                it->NextQuat,
                it->Interpolation
            );
            
        }
        
        /* Initialize rotation matrix */
        SkeletonMatrix_.matrixReset();
        SkeletonNormalMatrix_.matrixReset();
        
        /* Next quaternion & quat-matrix keyframe transformation */
        it->SkeletonBone->Quat.getMatrix(it->SkeletonBone->QuatMatrix);
        
        /* Transform the bone matrix */
        it->SkeletonBone->transformMatrix(SkeletonMatrix_, SkeletonNormalMatrix_);
        
        /* Multiply the bone matrix by the origin matrix */
        SkeletonMatrix_ *= it->SkeletonBone->MatrixOrigin;
        SkeletonNormalMatrix_ *= it->SkeletonBone->MatrixNormalOrigin;
        
        /* Update vertices */
        for (std::vector<SAnimSkeletonBone::SAnimSkeletonBoneVertex>::iterator itv = it->SkeletonBone->Vertices.begin();
             itv != it->SkeletonBone->Vertices.end(); ++itv)
        {
            hEntity_->selectSurface( itv->Surface );
            hEntity_->setVertexCoord( itv->Index, SkeletonMatrix_ * itv->Coord );
            hEntity_->setVertexNormal( itv->Index, SkeletonNormalMatrix_ * itv->Normal );
        } // next vertex
        
    } // next bone
    
    hEntity_->updateModeling(false);
    
}

void Animation::renderSkeleton(Camera* hCamera)
{
    dim::matrix4f NormalMatrix;
    
    for (std::vector<SAnimSkeleton>::iterator it = AnimSkeletonList_.begin(); it != AnimSkeletonList_.end(); ++it)
    {
        /* Initialize rotation matrix */
        SkeletonMatrix_.matrixReset();
        
        /* Transform the bone matrix */
        it->SkeletonBone->transformMatrix( SkeletonMatrix_, NormalMatrix );
        
        /* Draw the bone */
        __spVideoDriver->draw3DBone(hEntity_, SkeletonMatrix_, hCamera);
    }
}







} // /namespace anim

} // /namespace scene

} // /namespace sp







// ================================================================================
