/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * Header for animations
 *
 */

#ifndef __SP_SCENEMANAGER_ANIMATION_H__
#define __SP_SCENEMANAGER_ANIMATION_H__


#include "spStandard.hpp"
#include "spSceneManagerEntity.hpp"
#include "spSceneManagerCamera.hpp"
#include "spInputOutputString.hpp"
#include "spDimension.hpp"
#include "spMath.hpp"

#include <list>
#include <vector>


namespace sp
{
namespace scene
{
namespace anim
{


/*
 * Macros
 */

static const s32 DEF_ANIMMESH_SPEED = 100;


/*
 * Enumerations
 */

enum EAnimTypes
{
    ANIM_ONESHOT = 0,
    ANIM_PINGPONG,
    ANIM_LOOP,
    ANIM_LOOP_PINGPONG
};

enum EAnimSkeletonTypes
{
    ANIM_SKELETON_STOP = 0,
    ANIM_SKELETON_ONESHOT,
    ANIM_SKELETON_ONELOOP,
    ANIM_SKELETON_LOOP,
};

enum EAnimModeType
{
    ANIM_BACKWARDS  = -1,
    ANIM_STOP       = 0,
    ANIM_FOREWARDS  = 1
};

/*
 * Node animation
 */

struct SAnimNode
{
    u32 Index;
    dim::vector3df Position;
    dim::quaternion Rotation;
    dim::vector3df Scale;
};

/*
 * Morph-target (keyframe) animation
 */

struct SAnimMesh
{
    std::vector<dim::vector3df> VerticesCoord;
};

/*
 * Skeletal animation
 */

struct SP_EXPORT SAnimSkeletonBone
{
    
    SAnimSkeletonBone(
        u32 &BoneIDCounter, SAnimSkeletonBone* hParent,
        dim::vector3df BoneTranslation, dim::quaternion BoneQuaternion, dim::vector3df BoneScale,
        io::stringc BoneName
    );
    ~SAnimSkeletonBone();
    
    /* Structures */
    
    struct SAnimSkeletonBoneVertex
    {
        u32 Index;
        u32 Surface;
        dim::vector3df Coord;
        dim::vector3df Normal;
    };
    
    /* Functions */
    
    void transformMatrix(dim::matrix4f &Mat, dim::matrix4f &MatNormal);
    void getBoneTranslation(dim::vector3df &BoneTranslation);
    
    /* Members */
    
    u32 ID;
    io::stringc Name;
    
    dim::vector3df Translation, Scale;
    dim::matrix4f MatrixOrigin, MatrixNormalOrigin;
    
    dim::quaternion Quat; // Quaternion (for linear interpolation)
    dim::matrix4f QuatMatrix; // Quaternion matrix
    
    std::vector<SAnimSkeletonBoneVertex> Vertices;
    
    SAnimSkeletonBone* ParentBone;
    
};

struct SAnimSkeleton
{
    struct SAnimSkeletonFrame
    {
        f32 AnimSpeed;
        
        dim::vector3df Translation;
        dim::quaternion CurQuat;
        dim::vector3df Scale;
    };
    
    /* Members */
    
    f32 Interpolation;
    s32 Sequence;
    
    dim::vector3df NextTranslation;
    dim::quaternion NextQuat;
    dim::vector3df NextScale;
    
    SAnimSkeletonBone* SkeletonBone;
    
    std::vector<SAnimSkeletonFrame> FrameList;
};


/*
 * Animation class
 * function for update/ manipulation etc. for the following animations type:
 * node animation; morph-target animation; skeletal animation
 */

class SP_EXPORT Animation
{
    
    public:
        
        /* === Constructor & destructor === */
        
        Animation();
        Animation(Node* hNode);
        Animation(Entity* hEntity);
        ~Animation();
        
        /* Adding an animation sequence */
        u32 addAnimSeqNode(
            dim::vector3df Position, dim::quaternion Rotation = dim::quaternion(), dim::vector3df Scale = 1.0f
        );
        void removeAnimSeqNode(u32 SeqIndex);
        
        u32 addAnimSeqMesh(std::vector<dim::vector3df> VerticesCoord);
        void removeAnimSeqMesh(u32 SeqIndex);
        
        u32 addAnimSkeletonBone(
            u32 ParentBoneID,
            dim::vector3df Translation = 0.0f, dim::quaternion Quaternion = dim::quaternion(), dim::vector3df Scale = 1.0f,
            io::stringc Name = io::stringc()
        );
        void removeAnimSkeletonBone(u32 BoneID);
        
        /* Animating */
        void doAnimate(s8 AnimType = ANIM_ONESHOT, s32 AnimTime = DEF_ANIMMESH_SPEED, s32 Start = 0, s32 End = 0);
        void doAnimateNode(s8 AnimType = ANIM_ONESHOT, s32 AnimTime = DEF_ANIMMESH_SPEED, s32 Start = 0, s32 End = 0);
        void doAnimateMesh(s8 AnimType = ANIM_ONESHOT, s32 AnimTime = DEF_ANIMMESH_SPEED, s32 Start = 0, s32 End = 0);
        void doAnimateSkeleton(s8 AnimType = ANIM_SKELETON_ONESHOT, f32 AnimSpeed = 1.0f, s32 Start = 0, s32 End = 0);
        
        /* Checking for animating */
        inline bool isAnimate() const
        {
            return AnimModeNode_ != ANIM_STOP || AnimModeMesh_ != ANIM_STOP || AnimTypeSkeleton_ != ANIM_SKELETON_STOP;
        }
        inline bool isAnimateNode() const
        {
            return AnimModeNode_ != ANIM_STOP;
        }
        inline bool isAnimateMesh() const
        {
            return AnimModeMesh_ != ANIM_STOP;
        }
        inline bool isAnimateSkeleton() const
        {
            return AnimTypeSkeleton_ != ANIM_SKELETON_STOP;
        }
        
        /* Getting the animation length */
        inline u32 getAnimLengthNode() const
        {
            return AnimNodeList_.size() - 1;
        }
        inline u32 getAnimLengthMesh() const
        {
            return AnimMeshList_.size() - 1;
        }
        
        /* Getting the animation types */
        inline bool isNodeAnimation() const
        {
            return AnimNodeList_.size() > 0;
        }
        inline bool isMeshAnimation() const
        {
            return AnimMeshList_.size() > 0;
        }
        inline bool isSkeletonAnimation() const
        {
            return AnimSkeletonList_.size() > 0;
        }
        
        /* Setting the animation seek */
        void setAnimSeekNode(f32 AnimSeek, u32 AnimSeekSeq);
        void setAnimSeekMesh(f32 AnimSeek, u32 AnimSeekSeq);
        void setAnimSeekSkeleton(f32 AnimSeek, u32 AnimSeekSeq);
        
        /* Getting the animation seek */
        void getAnimSeekNode(f32 &AnimSeek, s32 &AnimSeekSeq);
        void getAnimSeekMesh(f32 &AnimSeek, s32 &AnimSeekSeq);
        void getAnimSeekSkeleton(f32 &AnimSeek, s32 &AnimSeekSeq);
        
        /* Pause/ resume the animation */
        void pauseAnim(bool isPaused = true);
        void pauseNode(bool isPaused = true);
        void pauseMesh(bool isPaused = true);
        void pauseSkeleton(bool isPaused = true);
        
        /* Stopping the animation */
        void stopAnim();
        void stopNode();
        void stopMesh();
        void stopSkeleton();
        
        /* Updating the animation */
        void updateAnimation();
        
        /* Set/ get node/ mesh */
        inline void setNode(Node* hNode)
        {
            hNode_ = hNode;
        }
        inline Node* getNode() const
        {
            return hNode_;
        }
        
        inline void setMesh(Entity* hEntity)
        {
            hEntity_ = hEntity;
        }
        inline Entity* getMesh() const
        {
            return hEntity_;
        }
        
        /* Set/ get skeletal animation speed */
        inline void setSkeletonAnimSpeed(f32 Speed)
        {
            AnimSpeedSkeleton_ = Speed;
        }
        inline f32 getSkeletonAnimSpeed() const
        {
            return AnimSpeedSkeleton_;
        }
        
        /* ===== Pure skeleton functions ===== */
        
        inline u32 getSkeletonBoneCount() const
        {
            return AnimSkeletonList_.size();
        }
        
        SAnimSkeleton getSkeleton(u32 BoneID);
        SAnimSkeletonBone* getSkeletonBone(u32 BoneID);
        
        inline std::vector<SAnimSkeleton> getSkeletonList() const
        {
            return AnimSkeletonList_;
        }
        
        /* Create an skeleton animation */
        
        u32 addSkeletonBoneFrame(
            u32 BoneID,
            dim::vector3df Translation, dim::quaternion Quaternion, dim::vector3df Scale,
            f32 Speed
        );
        void removeSkeletonBoneFrame(u32 BoneID, u32 FrameIndex);
        
        // !!!
        void setSkeletonBoneFrame(u32 BoneID, std::vector<dim::vector3df> FrameList, std::vector<f32> SpeedList);
        
        void addSkeletonBoneVertices(u32 BoneID, std::vector<u32> VerticesIndices, std::vector<u32> VerticesSurfaces);
        
        /* Manual manipulation */
        
        void updateSkeletonSequence(bool isUpdateSeek = true); // (For manual bone rotation)
        
        /*
         * Calculates the translations of all bones (with a 4*4 matrix)
         * Call this function after all bone creations and settings
         */
        void updateSkeletonBones();
        
        /* Renders the whole skeleton */
        void renderSkeleton(Camera* hCamera = 0);
        
    private:
        
        /* === Functions === */
        
        void init();
        
        void updateNextSeqNode();
        void updateNextSeqMesh();
        void updateNextSeqSkeleton();
        
        void updateAnimationNode();
        void updateAnimationMesh();
        void updateAnimationSkeleton();
        
        void updateSeekStatus(f32 &Seek, s32 &SeekSeq);
        
        void updateNextSeqSkeleton(std::vector<SAnimSkeleton>::iterator it, s32 IndexFrom, s32 IndexTo);
        
        /* === Member === */
        
        Node* hNode_;
        Entity* hEntity_;
        
        s8 AnimType_, AnimTypeSkeleton_;
        s8 PausedAnimModeNode_, PausedAnimModeMesh_, PausedAnimModeSkeleton_;
        s8 AnimModeNode_, AnimModeMesh_;
        f32 AnimSpeed_, AnimSpeedSkeleton_;//AnimSpeedNode_, AnimSpeedMesh_;
        f32 AnimSeekNode_, AnimSeekMesh_;
        s32 AnimSeekSeqNode_, AnimSeekSeqMesh_;
        s32 NodeStart_, NodeEnd_;
        s32 MeshStart_, MeshEnd_;
        s32 SkeletonStart_, SkeletonEnd_;
        u32 SkeletonBoneID_;
        
        /* Node animation */
        SAnimNode LastNodeLocation_, NextNodeLocation_;
        std::vector<SAnimNode> AnimNodeList_;
        
        /* Keyframe animation */
        std::vector<dim::vector3df> LastMeshVertices_, NextMeshVertices_;
        std::vector<SAnimMesh> AnimMeshList_;
        
        /* Bone/ Skeletal animation */
        dim::matrix4f SkeletonMatrix_, SkeletonNormalMatrix_;
        std::vector<SAnimSkeleton> AnimSkeletonList_;
        
};


} // /namespace anim

} // /namespace scene

} // /namespace sp


#endif







// ================================================================================
