/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * Header for all collision detections
 *
 */

#include "spSceneManagerCollisionDetector.hpp"


namespace sp
{
namespace scene
{


/*
 * ======= CollisionDetector =======
 */


Collision::Collision()
{
    ID_         = 0;
}
Collision::Collision(u32 ID, std::vector<s32> &TypeList, std::vector<u32> &DestinationIDList, std::list<Collision*> &CollisionTypeList)
{
    ID_         = ID;
    TypeList_   = TypeList;
    init(DestinationIDList, CollisionTypeList);
}
Collision::~Collision()
{
}

void Collision::init(std::vector<u32> &DestinationIDList, std::list<Collision*> &CollisionTypeList)
{
    for (std::vector<u32>::iterator it = DestinationIDList.begin(); it != DestinationIDList.end(); ++it)
    {
        /* Check this collision */
        if (*it == ID_)
            DestinationCollisionList_.push_back(this);
        
        /* Check all created collisions */
        for (std::list<Collision*>::iterator itColl = CollisionTypeList.begin(); itColl != CollisionTypeList.end(); ++itColl)
        {
            if (*it == (*itColl)->ID_)
                DestinationCollisionList_.push_back(*itColl);
        }
    }
}




/*
 * ======= CollisionDetector =======
 */


CollisionDetector::CollisionDetector()
{
}
CollisionDetector::~CollisionDetector()
{
    /* Delete all collisions */
    for (std::list<Collision*>::iterator it = CollisionTypeList_.begin(); it != CollisionTypeList_.end(); ++it)
        spDeletePointer(*it);
    CollisionTypeList_.clear();
    
    /* Delete all collision objects */
    for (std::list<SCollisionObject*>::iterator it = CollisionEntityList_.begin(); it != CollisionEntityList_.end(); ++it)
        spDeletePointer(*it);
    CollisionEntityList_.clear();
    
    /* Delete all picking objects */
    for (std::list<SPickingObject*>::iterator it = PickingEntityList_.begin(); it != PickingEntityList_.end(); ++it)
        spDeletePointer(*it);
    PickingEntityList_.clear();
}


/*
 * All the collision detections
 * (sphere-to-sphere collision; sphere-to-box collision; sphere-to-polygon collision)
 */

void CollisionDetector::updateScene()
{
    
    /* Allocate temporary memory */
    
    register std::list<SCollisionObject*>::iterator it, itColl;
    register std::vector<Collision*>::iterator itDest;
    register u32 i = 0;
    
    /* Loop all collision entities */
    
    for (it = CollisionEntityList_.begin(); it != CollisionEntityList_.end(); ++it)
    {
        
        (*it)->Object->resetCollision();
        
        if (!(*it)->Object->getCollisionUpdate())
            continue;
        
        for (itDest = (*it)->hColl->getDestinationList().begin(), i = 0; itDest != (*it)->hColl->getDestinationList().end(); ++itDest, ++i)
        {
            
            switch ((*it)->hColl->getType(i))
            {
                /* Sphere-to-sphere collision */
                case COLLISION_SPHERE_TO_SPHERE:
                {
                    for (itColl = (*itDest)->getEntityList().begin(); itColl != (*itDest)->getEntityList().end(); ++itColl)
                    {
                        if (it != itColl && (*itColl)->Object->getVisible() && (*it)->Object != (*itColl)->Object)
                            checkCollisionSphereToSphere(*it, *itColl);
                    }
                }
                break;
                
                /* Sphere-to-box collision */
                case COLLISION_SPHERE_TO_BOX:
                {
                    for (itColl = (*itDest)->getEntityList().begin(); itColl != (*itDest)->getEntityList().end(); ++itColl)
                    {
                        if (it != itColl && (*itColl)->Object->getVisible() && (*it)->Object != (*itColl)->Object)
                            checkCollisionSphereToBox(*it, *itColl);
                    }
                }
                break;
                
                /* Sphere-to-polygon collision */
                case COLLISION_SPHERE_TO_POLYGON:
                {
                    for (itColl = (*itDest)->getEntityList().begin(); itColl != (*itDest)->getEntityList().end(); ++itColl)
                    {
                        if (it != itColl && (*itColl)->Object->getVisible() && (*it)->Object != (*itColl)->Object)
                        {
                            if ((*itColl)->Object->getOctTreeRoot())
                                checkCollisionSphereToPolygonOptimized(*it, *itColl);
                            else
                                checkCollisionSphereToPolygon(*it, *itColl);
                        }
                    }
                }
                break;
            } // /switch
            
        } // next destination collision object
        
        (*it)->LastPosition = (*it)->Object->getPosition(true);
        
    } // next collision object
    
}


/*
 * ========== Private: ==========
 */


/*
 * Sphere-to-sphere collision
 */

void CollisionDetector::checkCollisionSphereToSphere(SCollisionObject* CheckingObject, SCollisionObject* CurrentObject)
{
    
    /* First settings */
    
    register dim::vector3df CheckPosition(CheckingObject->Object->getPosition(true)); // Colliding object
    register dim::vector3df CurPosition(CurrentObject->Object->getPosition(true));
    
    register dim::vector3df Distance = CurPosition - CheckPosition;
    
    /* Sphere-sphere collision */
    
    if (Distance.getLength() <= CheckingObject->Radius + CurrentObject->Radius)
    {
        dim::vector3df DistNormalized = Distance.normalize();
        SCollisionData CollInfo;
        {
            CollInfo.ID         = CurrentObject->hColl->getID();
            CollInfo.Triangle   = dim::triangle3df();
            CollInfo.Point      = CurPosition - DistNormalized * CheckingObject->Radius;
            CollInfo.Normal     = DistNormalized;
            CollInfo.Object     = CurrentObject->Object;
        }
        
        /* Simulate a collision */
        CheckingObject->Object->simulateCollision(CollInfo);
        CheckPosition = CurPosition - DistNormalized * ( CheckingObject->Radius + CurrentObject->Radius );
        CheckingObject->Object->setPosition(CheckPosition, true);
    }
    
}


/*
 * Sphere-to-box collision
 */

void CollisionDetector::checkCollisionSphereToBox(SCollisionObject* CheckingObject, SCollisionObject* CurrentObject)
{
    
    /* First settings */
    
    register dim::vector3df CheckPosition(CheckingObject->Object->getPosition(true)); // Colliding object
    
    register dim::vector3df CurPosition(CurrentObject->Object->getPosition(true));
    register dim::vector3df CurRotation(CurrentObject->Object->getRotation(true));
    register dim::vector3df CurScale(CurrentObject->Object->getScale(true));
    
    register dim::line3df Box(CurrentObject->Box.Corner1 * CurScale, CurrentObject->Box.Corner2 * CurScale);
    register dim::vector3df Distance, CollisionPoint, Normal;
    
    /* Transform the collision matrix (such as the tree matrix) */
    
    #if 0
    CollMatrix_ = CurrentObject->Object->getGlobalLocationInvers();
    #else
    CollMatrix_.matrixReset();
    CollMatrix_.matrixScale(dim::vector3df(1.0) / CurScale);
    CollMatrix_.matrixRotateZ(CurRotation.Z);
    CollMatrix_.matrixRotateX(CurRotation.X);
    CollMatrix_.matrixRotateY(CurRotation.Y);
    CollMatrix_.matrixTranslate(-CurPosition);
    #endif
    
    CheckPosition = CollMatrix_ * CheckPosition;
    
    /* Check if the sphere is inside the box-area */
    
    if (CheckPosition.X > CurrentObject->Box.Corner1.X / CurScale.X - CheckingObject->Radius && 
        CheckPosition.Y > CurrentObject->Box.Corner1.Y / CurScale.Y - CheckingObject->Radius && 
        CheckPosition.Z > CurrentObject->Box.Corner1.Z / CurScale.Z - CheckingObject->Radius && 
        CheckPosition.X < CurrentObject->Box.Corner2.X / CurScale.X + CheckingObject->Radius &&
        CheckPosition.Y < CurrentObject->Box.Corner2.Y / CurScale.Y + CheckingObject->Radius &&
        CheckPosition.Z < CurrentObject->Box.Corner2.Z / CurScale.Z + CheckingObject->Radius)
    {
        
        /* Get the closet point to the box and check the collsion */
        
        CollisionPoint = math::getClosestPoint(CurrentObject->Box, CheckPosition, Normal);
        
        Distance = (CollisionPoint - CheckPosition) * CurScale;
        
        if (Distance.getLength() <= CheckingObject->Radius)
        {
            /* Transform the collision matrix */
            CollMatrix_.matrixReset();
            CollMatrix_.matrixRotateY(-CurRotation.Y);
            CollMatrix_.matrixRotateX(-CurRotation.X);
            CollMatrix_.matrixRotateZ(-CurRotation.Z);
            
            /* Generate a collision information */
            SCollisionData CollInfo;
            {
                CollInfo.ID         = CurrentObject->hColl->getID();
                CollInfo.Triangle   = dim::triangle3df();
                CollInfo.Point      = CollisionPoint;
                CollInfo.Normal     = CollMatrix_ * Normal;
                CollInfo.Object     = CurrentObject->Object;
            }
            
            /* Simulate a collision */
            CheckingObject->Object->simulateCollision(CollInfo);
            CheckPosition = CollisionPoint - Distance.normalize() * CheckingObject->Radius / CurScale;
            
            /* Transform the collision matrix */
            #if 0
            
            #else
            CollMatrix_.matrixReset();
            CollMatrix_.matrixTranslate(CurPosition);
            CollMatrix_.matrixRotateY(-CurRotation.Y);
            CollMatrix_.matrixRotateX(-CurRotation.X);
            CollMatrix_.matrixRotateZ(-CurRotation.Z);
            CollMatrix_.matrixScale(CurScale);
            #endif
            
            /* Locate the checking object */
            CheckingObject->Object->setPosition(CollMatrix_ * CheckPosition, true);
        } // fi
        
    } // fi
    
}


/*
 * Sphere-to-polygon collision
 */

void CollisionDetector::checkCollisionSphereToPolygon(SCollisionObject* CheckingObject, SCollisionObject* CurrentObject)
{
    
    /* Allocate temporary memory */
    
    register u32 i;
    
    /* First settings */
    
    CollMatrix_ = CurrentObject->Object->getGlobalLocation();
    
    register dim::vector3df CheckPosition(CheckingObject->Object->getPosition(true)); // Colliding object
    
    register dim::vector3df Distance, CollisionPoint;
    register dim::triangle3df Triangle;
    register dim::line3df Line;
    
    /* Loop all triangles (only normals check) */
    
    for (i = 0; i < CurrentObject->TriangleCount; ++i)
    {
        
        /* Get the current triangle */
        Triangle = CollMatrix_ * CurrentObject->TriangleList[i];
        
        /* Set the line data */
        Line.Corner1 = CheckingObject->LastPosition;
        Line.Corner2 = CheckPosition;
        
        /* Get the line-triangle intersection (for robustness) */
        if (math::checkLineTriangleIntersection(Triangle, Line, CollisionPoint))
        {
            SCollisionData CollInfo;
            {
                CollInfo.ID         = CurrentObject->hColl->getID();
                CollInfo.Triangle   = Triangle;
                CollInfo.Point      = CollisionPoint;
                CollInfo.Object     = CurrentObject->Object;
            }
            
            /* Simulate a collision */
            CheckingObject->Object->simulateCollision(CollInfo);
            CheckPosition = CollisionPoint + Triangle.getNormal() * CheckingObject->Radius;
            CheckingObject->Object->setPosition(CheckPosition, true);
            
            /* Next triangle */
            continue;
        }
        
        /* Check the triangle side */
        if ( CurrentObject->FaceType != video::FACE_BOTH && 
             ( ( CurrentObject->FaceType == video::FACE_FRONT && !dim::plane3df(Triangle).isPointFrontSide(CheckPosition) ) ||
              CurrentObject->FaceType == video::FACE_BACK ) )
        {
            continue;
        }
        
        /* Get the closest normal the current triangle */
        if (math::getNormalPoint(Triangle, CheckPosition, CollisionPoint))
        {
            Distance = CollisionPoint - CheckPosition;
            
            if (Distance.getLength() <= CheckingObject->Radius)
            {
                SCollisionData CollInfo;
                {
                    CollInfo.ID         = CurrentObject->hColl->getID();
                    CollInfo.Triangle   = Triangle;
                    CollInfo.Point      = CollisionPoint;
                    CollInfo.Object     = CurrentObject->Object;
                }
                
                /* Simulate a collision */
                CheckingObject->Object->simulateCollision(CollInfo);
                CheckPosition = CollisionPoint - Distance.normalize() * CheckingObject->Radius;
                CheckingObject->Object->setPosition(CheckPosition, true);
            }
        }
        
    }
    
    /* Loop all triangles (complete distance check) */
    
    for (i = 0; i < CurrentObject->TriangleCount; ++i)
    {
        
        /* Get the current triangle */
        Triangle = CollMatrix_ * CurrentObject->TriangleList[i];
        
        /* Check the triangle side */
        if ( CurrentObject->FaceType != video::FACE_BOTH && 
             ( ( CurrentObject->FaceType == video::FACE_FRONT && !dim::plane3df(Triangle).isPointFrontSide(CheckPosition) ) ||
              CurrentObject->FaceType == video::FACE_BACK ) )
        {
            continue;
        }
        
        /* Get the closest point to the current triangle */
        CollisionPoint = math::getClosestPoint(Triangle, CheckPosition);
        
        Distance = CollisionPoint - CheckPosition;
        
        if (Distance.getLength() <= CheckingObject->Radius)
        {
            SCollisionData CollInfo;
            {
                CollInfo.ID         = CurrentObject->hColl->getID();
                CollInfo.Triangle   = Triangle;
                CollInfo.Point      = CollisionPoint;
                CollInfo.Object     = CurrentObject->Object;
            }
            
            /* Simulate a collision */
            CheckingObject->Object->simulateCollision(CollInfo);
            CheckPosition = CollisionPoint - Distance.normalize() * CheckingObject->Radius;
            CheckingObject->Object->setPosition(CheckPosition, true);
        }
        
    }
    
    /* Get the last position of the colliding object */
    
    //CheckingObject->LastPosition = CheckPosition;
    
}


/*
 * Sphere-to-polygon collision (optimized)
 */

void CollisionDetector::checkCollisionSphereToPolygonOptimized(SCollisionObject* CheckingObject, SCollisionObject* CurrentObject)
{
    
    /* Allocate temporary memory */
    
    register std::vector<OctTree*>::iterator itTree;
    register std::vector<dim::ptriangle3df>::iterator it;
    
    /* First settings */
    
    CollMatrix_ = CurrentObject->Object->getGlobalLocation();
    TreeMatrix_ = CurrentObject->Object->getGlobalLocationInverse();
    
    register dim::vector3df CheckPosition(CheckingObject->Object->getPosition(true)); // Colliding object
    
    register dim::vector3df CurScale(CurrentObject->Object->getScale());
    
    register dim::vector3df Distance, CollisionPoint;
    register dim::triangle3df Triangle;
    register dim::line3df Line;
    
    /* Search for the current oct-tree-nodes */
    
    register dim::vector3df TreeNodePos(TreeMatrix_ * CheckPosition);
    register dim::vector3df TreeNodeRadius(dim::vector3df(CheckingObject->Radius) / CurScale);
    
    OctTree::clearTreeList();
    
    CurrentObject->Object->getOctTreeRoot()->searchTreeNodes(TreeNodePos, TreeNodeRadius);
    
    /* Loop all triangles (only normals check) */
    
    for (itTree = OctTree::getTreeList().begin(); itTree != OctTree::getTreeList().end(); ++itTree)
    {
        for (it = (*itTree)->getTriangleList().begin(); it != (*itTree)->getTriangleList().end(); ++it)
        {
            
            /* Get the current triangle */
            Triangle = CollMatrix_ * (*it);
            
            /* Set the line data */
            Line.Corner1 = CheckingObject->LastPosition;
            Line.Corner2 = CheckPosition;
            
            /* Get the line-triangle intersection (for robustness) */
            if (math::checkLineTriangleIntersection(Triangle, Line, CollisionPoint))
            {
                SCollisionData CollInfo;
                {
                    CollInfo.ID         = CurrentObject->hColl->getID();
                    CollInfo.Triangle   = Triangle;
                    CollInfo.Point      = CollisionPoint;
                    CollInfo.Object     = CurrentObject->Object;
                }
                
                /* Simulate a collision */
                CheckingObject->Object->simulateCollision(CollInfo);
                CheckPosition = CollisionPoint + Triangle.getNormal() * CheckingObject->Radius;
                CheckingObject->Object->setPosition(CheckPosition, true);
                
                /* Next triangle */
                continue;
            }
            
            /* Check the triangle side */
            if ( CurrentObject->FaceType != video::FACE_BOTH && 
                 ( ( CurrentObject->FaceType == video::FACE_FRONT && !dim::plane3df(Triangle).isPointFrontSide(CheckPosition) ) ||
                  CurrentObject->FaceType == video::FACE_BACK ) )
            {
                continue;
            }
            
            /* Get the closest normal the current triangle */
            if (math::getNormalPoint(Triangle, CheckPosition, CollisionPoint))
            {
                Distance = CollisionPoint - CheckPosition;
                
                if (Distance.getLength() <= CheckingObject->Radius)
                {
                    SCollisionData CollInfo;
                    {
                        CollInfo.ID         = CurrentObject->hColl->getID();
                        CollInfo.Triangle   = Triangle;
                        CollInfo.Point      = CollisionPoint;
                        CollInfo.Object     = CurrentObject->Object;
                    }
                    
                    /* Simulate a collision */
                    CheckingObject->Object->simulateCollision(CollInfo);
                    CheckPosition = CollisionPoint - Distance.normalize() * CheckingObject->Radius;
                    CheckingObject->Object->setPosition(CheckPosition, true);
                }
            }
            
        }
    }
    
    /* Loop all triangles (complete distance check) */
    
    for (itTree = OctTree::getTreeList().begin(); itTree != OctTree::getTreeList().end(); ++itTree)
    {
        for (it = (*itTree)->getTriangleList().begin(); it != (*itTree)->getTriangleList().end(); ++it)
        {
            
            /* Get the current triangle */
            Triangle = CollMatrix_ * (*it);
            
            /* Check the triangle side */
            if ( CurrentObject->FaceType != video::FACE_BOTH && 
                 ( ( CurrentObject->FaceType == video::FACE_FRONT && !dim::plane3df(Triangle).isPointFrontSide(CheckPosition) ) ||
                  CurrentObject->FaceType == video::FACE_BACK ) )
            {
                continue;
            }
            
            /* Get the closest point the current triangle */
            CollisionPoint = math::getClosestPoint(Triangle, CheckPosition);
            
            Distance = CollisionPoint - CheckPosition;
            
            if (Distance.getLength() <= CheckingObject->Radius)
            {
                SCollisionData CollInfo;
                {
                    CollInfo.ID         = CurrentObject->hColl->getID();
                    CollInfo.Triangle   = Triangle;
                    CollInfo.Point      = CollisionPoint;
                    CollInfo.Object     = CurrentObject->Object;
                }
                
                /* Simulate a collision */
                CheckingObject->Object->simulateCollision(CollInfo);
                CheckPosition = CollisionPoint - Distance.normalize() * CheckingObject->Radius;
                CheckingObject->Object->setPosition(CheckPosition, true);
            }
            
        }
    }
    
    /* Get the last position of the colliding object */
    
    //CheckingObject->LastPosition = CheckPosition;
    
}


/*
 * Line-sphere intersection
 */

void CollisionDetector::checkIntersectionLineSphere(SPickingObject* CurrentObject, dim::line3df &Line)
{
    
    /* First settings */
    
    register dim::vector3df CurPosition = CurrentObject->Object->getPosition(true);
    
    /* Line-sphere intersection */
    
    if (math::checkLineSphereIntersection(Line, CurPosition, CurrentObject->Radius, CurrentObject->Intersection))
    {
        CurrentObject->Distance         = math::getDistanceSq(Line.Corner1, CurrentObject->Intersection);
        CurrentObject->TriangleIndex    = -1;
        
        PickedObjectList_.push_back(*CurrentObject);
    }
    
}


/*
 * Line-box intersection
 */

void CollisionDetector::checkIntersectionLineBox(SPickingObject* CurrentObject, dim::line3df &Line)
{
    
    /* First settings */
    
    register dim::line3df Ray(Line);
    
    /* Rotate the ray inverse the object location */
    
    dim::matrix4f Mat(CurrentObject->Object->getGlobalLocationInverse());
    Ray.Corner1 = Mat * Ray.Corner1;
    Ray.Corner2 = Mat * Ray.Corner2;
    
    /* Line-sphere intersection */
    
    if (math::checkLineBoxIntersection(Ray, CurrentObject->Box, CurrentObject->Intersection))
    {
        CurrentObject->Intersection     = CurrentObject->Object->getGlobalLocation() * CurrentObject->Intersection;
        CurrentObject->Distance         = math::getDistanceSq(Line.Corner1, CurrentObject->Intersection);
        CurrentObject->TriangleIndex    = -1;
        
        PickedObjectList_.push_back(*CurrentObject);
    }
    
}


/*
 * Line-triangle intersection
 */

void CollisionDetector::checkIntersectionLinePolygon(SPickingObject* CurrentObject, dim::line3df &Line)
{
    
    /* Allocate temporary memory */
    
    register u32 i;
    
    /* First settings */
    
    CollMatrix_ = CurrentObject->Object->getGlobalLocation();
    
    /* Loop all triangles (only normals check) */
    
    for (i = 0; i < CurrentObject->TriangleCount; ++i)
    {
        
        CurrentObject->Triangle = CurrentObject->TriangleList[i];
        
        CurrentObject->Triangle.PointA = CollMatrix_ * CurrentObject->Triangle.PointA;
        CurrentObject->Triangle.PointB = CollMatrix_ * CurrentObject->Triangle.PointB;
        CurrentObject->Triangle.PointC = CollMatrix_ * CurrentObject->Triangle.PointC;
        
        /* Get the closest normal the current triangle */
        if (math::checkLineTriangleIntersection(CurrentObject->Triangle, Line, CurrentObject->Intersection))
        {
            CurrentObject->Distance         = math::getDistanceSq(Line.Corner1, CurrentObject->Intersection);
            CurrentObject->TriangleIndex    = i;
            
            PickedObjectList_.push_back(*CurrentObject);
        }
        
    }
    
}


/*
 * Line-triangle intersection (optimized)
 */

void CollisionDetector::checkIntersectionLinePolygonOptimized(SPickingObject* CurrentObject, dim::line3df &Line)
{
    
    /* Allocate temporary memory */
    
    register std::vector<OctTree*>::iterator itTree;
    register std::vector<dim::ptriangle3df>::iterator it;
    
    /* First settings */
    
    CollMatrix_ = CurrentObject->Object->getGlobalLocation();
    TreeMatrix_ = CurrentObject->Object->getGlobalLocationInverse();
    
    /* Search for the current oct-tree-nodes */
    
    register dim::line3df TreeNodeLine(TreeMatrix_ * Line.Corner1, TreeMatrix_ * Line.Corner2);
    
    OctTree::clearTreeList();
    
    CurrentObject->Object->getOctTreeRoot()->searchTreeNodes(TreeNodeLine);
    
    /* Loop all triangles (only normals check) */
    
    for (itTree = OctTree::getTreeList().begin(); itTree != OctTree::getTreeList().end(); ++itTree)
    {
        for (it = (*itTree)->getTriangleList().begin(); it != (*itTree)->getTriangleList().end(); ++it)
        {
            
            CurrentObject->Triangle = *it;
            
            CurrentObject->Triangle.PointA = CollMatrix_ * CurrentObject->Triangle.PointA;
            CurrentObject->Triangle.PointB = CollMatrix_ * CurrentObject->Triangle.PointB;
            CurrentObject->Triangle.PointC = CollMatrix_ * CurrentObject->Triangle.PointC;
            
            /* Get the closest normal the current triangle */
            if (math::checkLineTriangleIntersection(CurrentObject->Triangle, Line, CurrentObject->Intersection))
            {
                CurrentObject->Distance         = math::getDistanceSq(Line.Corner1, CurrentObject->Intersection);
                CurrentObject->TriangleIndex    = (*itTree)->getTriangleNrList()[ u32( it - (*itTree)->getTriangleList().begin() ) ];
                
                PickedObjectList_.push_back(*CurrentObject);
            }
            
        }
    }
    
}



/*
 * Line-sphere intersection
 */

bool CollisionDetector::checkInviewLineSphere(SPickingObject* CurrentObject, dim::line3df &Line)
{
    /* First settings */
    
    register dim::vector3df CurPosition = CurrentObject->Object->getPosition(true);
    
    /* Line-sphere intersection */
    
    return math::checkLineSphereIntersection(Line, CurPosition, CurrentObject->Radius, CurrentObject->Intersection);
}


/*
 * Line-box intersection
 */

bool CollisionDetector::checkInviewLineBox(SPickingObject* CurrentObject, dim::line3df &Line)
{
    
    /* First settings */
    
    register dim::line3df Ray(Line);
    
    /* Rotate the ray inverse the object location */
    
    dim::matrix4f Mat(CurrentObject->Object->getGlobalLocationInverse());
    Ray.Corner1 = Mat * Ray.Corner1;
    Ray.Corner2 = Mat * Ray.Corner2;
    
    /* Line-sphere intersection */
    
    return math::checkLineBoxIntersection(Ray, CurrentObject->Box, CurrentObject->Intersection);
    
}


/*
 * Line-triangle intersection
 */

bool CollisionDetector::checkInviewLinePolygon(SPickingObject* CurrentObject, dim::line3df &Line)
{
    
    /* Allocate temporary memory */
    
    register u32 i;
    
    /* First settings */
    
    CollMatrix_ = CurrentObject->Object->getGlobalLocation();
    
    /* Loop all triangles (only normals check) */
    
    for (i = 0; i < CurrentObject->TriangleCount; ++i)
    {
        
        CurrentObject->Triangle = CurrentObject->TriangleList[i];
        
        CurrentObject->Triangle.PointA = CollMatrix_ * CurrentObject->Triangle.PointA;
        CurrentObject->Triangle.PointB = CollMatrix_ * CurrentObject->Triangle.PointB;
        CurrentObject->Triangle.PointC = CollMatrix_ * CurrentObject->Triangle.PointC;
        
        /* Get the closest normal the current triangle */
        if (math::checkLineTriangleIntersection(CurrentObject->Triangle, Line, CurrentObject->Intersection))
            return true;
        
    }
    
    return false;
    
}


/*
 * Line-triangle intersection (optimized)
 */

bool CollisionDetector::checkInviewLinePolygonOptimized(SPickingObject* CurrentObject, dim::line3df &Line)
{
    
    /* Allocate temporary memory */
    
    register std::vector<OctTree*>::iterator itTree;
    register std::vector<dim::ptriangle3df>::iterator it;
    
    /* First settings */
    
    CollMatrix_ = CurrentObject->Object->getGlobalLocation();
    TreeMatrix_ = CurrentObject->Object->getGlobalLocationInverse();
    
    /* Search for the current oct-tree-nodes */
    
    register dim::line3df TreeNodeLine(TreeMatrix_ * Line.Corner1, TreeMatrix_ * Line.Corner2);
    
    OctTree::clearTreeList();
    
    CurrentObject->Object->getOctTreeRoot()->searchTreeNodes(TreeNodeLine);
    
    /* Loop all triangles (only normals check) */
    
    for (itTree = OctTree::getTreeList().begin(); itTree != OctTree::getTreeList().end(); ++itTree)
    {
        for (it = (*itTree)->getTriangleList().begin(); it != (*itTree)->getTriangleList().end(); ++it)
        {
            
            CurrentObject->Triangle = *it;
            
            CurrentObject->Triangle.PointA = CollMatrix_ * CurrentObject->Triangle.PointA;
            CurrentObject->Triangle.PointB = CollMatrix_ * CurrentObject->Triangle.PointB;
            CurrentObject->Triangle.PointC = CollMatrix_ * CurrentObject->Triangle.PointC;
            
            /* Get the closest normal the current triangle */
            if (math::checkLineTriangleIntersection(CurrentObject->Triangle, Line, CurrentObject->Intersection))
                return true;
            
        }
    }
    
    return false;
    
}


/*
 * ========== /Private ==========
 */



Collision* CollisionDetector::createCollision(u32 CollisionID, std::vector<s32> &CollisionTypeList, std::vector<u32> &DestinationIDList)
{
    Collision* hColl = new Collision(CollisionID, CollisionTypeList, DestinationIDList, CollisionTypeList_);
    CollisionTypeList_.push_back(hColl);
    return hColl;
}

void CollisionDetector::deleteCollision(Collision* hColl)
{
    for (std::list<Collision*>::iterator it = CollisionTypeList_.begin(); it != CollisionTypeList_.end(); ++it)
    {
        if (*it == hColl)
        {
            spDeletePointer(hColl);
            CollisionTypeList_.erase(it);
            break;
        }
    }
}

void CollisionDetector::addCollisionEntity(
    Entity* hEntity, Collision* hColl, f32 Radius, dim::line3df Box, video::EFaceTypes FaceType)
{
    SCollisionObject* ObjectData = new SCollisionObject();
    {
        /* General settings */
        ObjectData->Object          = hEntity;
        ObjectData->hColl           = hColl;
        ObjectData->Radius          = Radius;
        ObjectData->Box             = Box;
        ObjectData->LastPosition    = hEntity->getPosition(true);
        ObjectData->FaceType        = FaceType;
        
        /* Allocate memory for the triangles */
        ObjectData->TriangleList = new dim::ptriangle3df[ ObjectData->Object->getMeshTriangleCount() ];
        
        /* Get the triganle list */
        ObjectData->Object->getTriangleList(ObjectData->TriangleList, ObjectData->TriangleCount);
    }
    CollisionEntityList_.push_back(ObjectData);
    
    /* Add to the collisions */
    for (std::list<Collision*>::iterator it = CollisionTypeList_.begin(); it != CollisionTypeList_.end(); ++it)
    {
        if (*it == hColl)
            hColl->getEntityList().push_back(ObjectData);
    }
}

void CollisionDetector::removeCollisionEntity(Entity* hEntity)
{
    for (std::list<SCollisionObject*>::iterator it = CollisionEntityList_.begin(); it != CollisionEntityList_.end(); ++it)
    {
        if ((*it)->Object == hEntity)
        {
            
            /* Remove this entity from the collions list */
            for (std::list<Collision*>::iterator itColl = CollisionTypeList_.begin(); itColl != CollisionTypeList_.end(); ++itColl)
            {
                if (*itColl == (*it)->hColl)
                {
                    for (std::list<SCollisionObject*>::iterator itDest = (*it)->hColl->getEntityList().begin(); itDest != (*it)->hColl->getEntityList().end(); ++itDest)
                    {
                        if (itDest == it)
                        {
                            (*it)->hColl->getEntityList().erase(itDest);
                            break;
                        }
                    }
                    break;
                }
            }
            
            /* Delete the triangle pointer and the oct-tree-root-node */
            spDeletePointer(*it);
            CollisionEntityList_.erase(it);
            
            /* Break the loop and exit the function */
            break;
        }
    }
}

void CollisionDetector::setCollisionEntityData(
    Entity* hEntity, Collision* hColl, f32 Radius, dim::line3df Box, video::EFaceTypes FaceType)
{
    for (std::list<SCollisionObject*>::iterator it = CollisionEntityList_.begin(); it != CollisionEntityList_.end(); ++it)
    {
        if ((*it)->Object == hEntity)
        {
            (*it)->Object   = hEntity;
            (*it)->hColl    = hColl;
            (*it)->Radius   = Radius;
            (*it)->Box      = Box;
            (*it)->FaceType = FaceType;
            break;
        }
    }
}


/*
 * ======= Picking intersection =======
 */


void CollisionDetector::addPickingEntity(Entity* hEntity, s32 PickMode, void* UserData)
{
    SPickingObject* PickObjectData = new SPickingObject();
    {
        /* General settings */
        PickObjectData->Object      = hEntity;
        PickObjectData->UserData    = UserData;
        PickObjectData->PickMode    = PickMode;
        
        /* Sphere & bounding box information */
        PickObjectData->Radius      = 1.0f;
        PickObjectData->Box         = dim::line3df(-0.5f, 0.5f);
        
        /* Polygon information */
        if (PickMode == PICKMODE_POLYGON)
        {
            /* Allocate memory for the triangles */
            PickObjectData->TriangleList = new dim::ptriangle3df[ PickObjectData->Object->getMeshTriangleCount() ];
            
            /* Get the triganle list */
            PickObjectData->Object->getTriangleList(PickObjectData->TriangleList, PickObjectData->TriangleCount);
        }
    }
    PickingEntityList_.push_back(PickObjectData);
}

void CollisionDetector::addPickingEntity(Entity* hEntity, f32 Radius, dim::line3df Box, s32 PickMode, void* UserData)
{
    SPickingObject* PickObjectData = new SPickingObject();
    {
        /* General settings */
        PickObjectData->Object      = hEntity;
        PickObjectData->UserData    = UserData;
        PickObjectData->PickMode    = PickMode;
        
        /* Sphere & bounding box information */
        PickObjectData->Radius      = Radius;
        PickObjectData->Box         = Box;
        
        /* Polygon information */
        if (PickMode == PICKMODE_POLYGON)
        {
            /* Allocate memory for the triangles */
            PickObjectData->TriangleList = new dim::ptriangle3df[ PickObjectData->Object->getMeshTriangleCount() ];
            
            /* Get the triganle list */
            PickObjectData->Object->getTriangleList(PickObjectData->TriangleList, PickObjectData->TriangleCount);
        }
    }
    PickingEntityList_.push_back(PickObjectData);
}

void CollisionDetector::removePickingEntity(Entity* hEntity)
{
    for (std::list<SPickingObject*>::iterator it = PickingEntityList_.begin(); it != PickingEntityList_.end(); ++it)
    {
        if ((*it)->Object == hEntity)
        {
            /* Delete triangle list */
            spDeleteArray((*it)->TriangleList);
            
            /* Delete the pointer of this object */
            spDeletePointer(*it);
            
            /* Remove this object from the list */
            PickingEntityList_.erase(it);
            
            /* Break the loop and exit the function */
            break;
        }
    }
}

bool sortPickedObjects(SPickingObject &obj1, SPickingObject &obj2)
{
    return obj1.Distance < obj2.Distance;
}

std::list<SPickingObject> CollisionDetector::pickLineIntersection(dim::line3df Line, bool isAutoSort)
{
    /* Allocate solution intersection list */
    PickedObjectList_.clear();
    
    /* Loop all pickable objects */
    
    for (std::list<SPickingObject*>::iterator it = PickingEntityList_.begin(); it != PickingEntityList_.end(); ++it)
    {
        
        if (!(*it)->Object->getVisible())
            continue;
        
        switch ((*it)->PickMode)
        {
            /* Line-sphere intersection */
            case PICKMODE_SPHERE:
            {
                checkIntersectionLineSphere(*it, Line);
            }
            break;
            
            /* Line-box intersection */
            case PICKMODE_BOX:
            {
                checkIntersectionLineBox(*it, Line);
            }
            break;
            
            /* Line-polygon intersection */
            case PICKMODE_POLYGON:
            {
                if ((*it)->Object->getOctTreeRoot())
                    checkIntersectionLinePolygonOptimized(*it, Line);
                else
                    checkIntersectionLinePolygon(*it, Line);
            }
            break;
        }
        
    } // next picking object
    
    /* Sort intersection list */
    if (isAutoSort)
        PickedObjectList_.sort(sortPickedObjects);
    
    /* Return the intersection list */
    return PickedObjectList_;
}

bool CollisionDetector::areJointsInview(const dim::vector3df PosA, const dim::vector3df PosB)
{
    /* Picking line */
    dim::line3df Line(PosA, PosB);
    
    /* Loop all pickable obejcts */
    
    for (std::list<SPickingObject*>::iterator it = PickingEntityList_.begin(); it != PickingEntityList_.end(); ++it)
    {
        
        if (!(*it)->Object->getVisible())
            continue;
        
        switch ((*it)->PickMode)
        {
            /* Line-sphere inview */
            case PICKMODE_SPHERE:
            {
                if (checkInviewLineSphere(*it, Line))
                    return false;
            }
            break;
            
            /* Line-box inview */
            case PICKMODE_BOX:
            {
                if (checkInviewLineBox(*it, Line))
                    return false;
            }
            break;
            
            /* Line-polygon inview */
            case PICKMODE_POLYGON:
            {
                if ((*it)->Object->getOctTreeRoot())
                {
                    if (checkInviewLinePolygonOptimized(*it, Line))
                        return false;
                }
                else
                {
                    if (checkInviewLinePolygon(*it, Line))
                        return false;
                }
            }
            break;
        }
        
    } // next picking object
    
    /* The joints are inview */
    return true;
}

bool CollisionDetector::areNodesInview(Node* Object1, Node* Object2)
{
    return areJointsInview(Object1->getPosition(true), Object2->getPosition(true));
}


/*
 * ======= OctTrees =======
 */


std::vector<OctTree*> OctTreeList;


OctTree::OctTree() : TempTriangleList_(0)
{
    for (s32 i = 0; i < 8; ++i)
        Children_[i] = 0;
    
    Index_              = 1;
    Length_             = 0;
    hasChildren_        = false;
    TempTriangleCount_  = 0;
}
OctTree::OctTree(SOctTreeTransportData Transporter)
    : TempTriangleList_(Transporter.pTriangleList), TempTriangleCount_(Transporter.TriangleListCount)
{
    /* General settings */
    for (s32 i = 0; i < 8; ++i)
        Children_[i] = 0;
    
    Index_              = Transporter.Index;
    Length_             = Transporter.Length;
    Pos_                = Transporter.Pos;
    Size_               = Transporter.Size;
    hasChildren_        = false;
    
    Transporter.Corner1 = Transporter.Pos;
    Transporter.Corner2 = Transporter.Pos + Transporter.Size;
    
    /* Create new children or add all needed triangles */
    if (Transporter.doCreateChildren)
        createChildren(Transporter.Object, Transporter.Corner1, Transporter.Corner2);
    else
        fillTriangleList(Transporter.Object, Transporter.Corner1, Transporter.Corner2);
}
OctTree::~OctTree()
{
    for (s32 i = 0; i < 8; ++i)
        spDeletePointer(Children_[i]);
}

void OctTree::beginTriangleList(Entity* hEntity)
{
    /* Allocate temporary memory for the triangles */
    TempTriangleCount_  = hEntity->getMeshTriangleCount();
    TempTriangleList_   = new dim::ptriangle3df[TempTriangleCount_];
    
    /* Get the triganle list */
    hEntity->getTriangleList(TempTriangleList_, TempTriangleCount_);
}

void OctTree::endTriangleList()
{
    /* Delete the temporary memory */
    spDeleteArray(TempTriangleList_);
}

void OctTree::createTree(Entity* hEntity, u32 Length)
{
    /* Allocate a temporary trianlge list */
    beginTriangleList(hEntity);
    
    /* Allocate temporary memory */
    dim::vector3df Corner1, Corner2;
    
    /* Get the dimension of the entity */
    hEntity->getBoundingBox(Corner1, Corner2);
    Pos_    = Corner1;
    Size_   = Corner2 - Corner1;
    Length_ = Length;
    
    /* Create the tree */
    createChildren(hEntity, Corner1, Corner2);
    
    /* Clear up the triangle list */
    endTriangleList();
}

void OctTree::createTree(Entity* hEntity, dim::vector3df Pos, dim::vector3df Size, u32 Length)
{
    /* Allocate a temporary trianlge list */
    beginTriangleList(hEntity);
    
    /* Allocate temporary memory */
    dim::vector3df Corner1(Pos), Corner2(Pos + Size);
    
    Length_ = Length;
    
    /* Create the tree */
    createChildren(hEntity, Corner1, Corner2);
    
    /* Clear up the triangle list */
    endTriangleList();
}

void OctTree::searchTreeNodes(dim::vector3df &Pos, dim::vector3df &Radius)
{
    /* Loop the eight children */
    for (s32 i = 0; i < 8; ++i)
    {
        /* Check if the position is inside the current child */
        if (Pos.X >= (Children_[i]->Pos_.X - Radius.X) &&
            Pos.Y >= (Children_[i]->Pos_.Y - Radius.Y) &&
            Pos.Z >= (Children_[i]->Pos_.Z - Radius.Z) && 
            Pos.X <= (Children_[i]->Pos_.X + Children_[i]->Size_.X + Radius.X) &&
            Pos.Y <= (Children_[i]->Pos_.Y + Children_[i]->Size_.Y + Radius.Y) &&
            Pos.Z <= (Children_[i]->Pos_.Z + Children_[i]->Size_.Z + Radius.Z))
        {
            /*
             * Check if the current child has also children
             * Otherwise add this node to the tree-node list
             */
            if (Children_[i]->hasChildren_)
                Children_[i]->searchTreeNodes(Pos, Radius);
            else
                OctTreeList.push_back(Children_[i]);
        } // fi
    } // next oct-tree node
}

void OctTree::searchTreeNodes(dim::line3df &Line)
{
    dim::vector3df tmpCorner;
    
    /* Loop the eight children */
    for (s32 i = 0; i < 8; ++i)
    {
        /* Check if the line intersects with the current child */
        tmpCorner = Children_[i]->Pos_ + Children_[i]->Size_;
        
        if (math::checkLineBoxIntersection(Line, Children_[i]->Pos_, tmpCorner))
        {
            /*
             * Check if the current child has also children
             * Otherwise add this node to the tree-node list
             */
            if (Children_[i]->hasChildren_)
                Children_[i]->searchTreeNodes(Line);
            else
                OctTreeList.push_back(Children_[i]);
        } // fi
    } // next oct-tree node
}

void OctTree::clearTreeList()
{
    OctTreeList.clear();
}

std::vector<OctTree*>& OctTree::getTreeList()
{
    return OctTreeList;
}

/*
       -------`
      / E / F /|
     ---/---/||
     | A | B |/|
G -> |---|---||| <- H
     | C | D |/
     `-------
*/

void OctTree::createChildren(Entity* hEntity, dim::vector3df &Corner1, dim::vector3df &Corner2)
{
    hasChildren_ = true;
    
    SOctTreeTransportData Transporter;
    {
        Transporter.Size                = Size_ / 2;
        Transporter.Index               = Index_ + 1;
        Transporter.Length              = Length_;
        Transporter.Object              = hEntity;
        Transporter.Corner1             = Corner1;
        Transporter.Corner2             = Corner2;
        Transporter.doCreateChildren    = Transporter.Index <= Length_;
        Transporter.pTriangleList       = TempTriangleList_;
        Transporter.TriangleListCount   = TempTriangleCount_;
    }
    
    // A
    Transporter.Pos = Pos_ + dim::vector3df(0, Transporter.Size.Y, 0);
    Children_[0] = new OctTree(Transporter);
    
    // B
    Transporter.Pos = Pos_ + dim::vector3df(Transporter.Size.X, Transporter.Size.Y, 0);
    Children_[1] = new OctTree(Transporter);
    
    // C
    Transporter.Pos = Pos_;
    Children_[2] = new OctTree(Transporter);
    
    // D
    Transporter.Pos = Pos_ + dim::vector3df(Transporter.Size.X, 0, 0);
    Children_[3] = new OctTree(Transporter);
    
    // E
    Transporter.Pos = Pos_ + dim::vector3df(0, Transporter.Size.Y, Transporter.Size.Z);
    Children_[4] = new OctTree(Transporter);
    
    // F
    Transporter.Pos = Pos_ + Transporter.Size;
    Children_[5] = new OctTree(Transporter);
    
    // G
    Transporter.Pos = Pos_ + dim::vector3df(0, 0, Transporter.Size.Z);
    Children_[6] = new OctTree(Transporter);
    
    // H
    Transporter.Pos = Pos_ + dim::vector3df(Transporter.Size.X, 0, Transporter.Size.Z);
    Children_[7] = new OctTree(Transporter);
}

void OctTree::fillTriangleList(Entity* hEntity, dim::vector3df &Corner1, dim::vector3df &Corner2)
{
    dim::triangle3df Triangle;
    
    /* Loop all triangles */
    for (register u32 i = 0; i < TempTriangleCount_; ++i)
    {
        /* Use the current triangle */
        Triangle = TempTriangleList_[i];
        
        /* Check if the current triangle collides with this tree-node-box */
        if (math::checkTriangleBoxIntersection(Triangle, Corner1, Corner2))
        {
            TriangleList_.push_back(TempTriangleList_[i]);
            TriangleNrList_.push_back(i);
        }
    }
}



} // /namespace scene

} // /namespace sp







// ================================================================================
