/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * Header for all collision detections
 *
 */

#ifndef __SP_SCENEMANAGER_COLLISIONS_H__
#define __SP_SCENEMANAGER_COLLISIONS_H__


#include "spStandard.hpp"
#include "spSceneManagerEntity.hpp"
#include "spDimension.hpp"
#include "spMath.hpp"

#include <list>
#include <vector>


namespace sp
{
namespace scene
{


/*
 * ======= OctTrees =======
 */

class SP_EXPORT OctTree
{
    
    public:
        
        OctTree();
        ~OctTree();
        
        void createTree(Entity* hEntity, u32 Length = DEF_OCTTREE_LENGTH);
        void createTree(Entity* hEntity, dim::vector3df Pos, dim::vector3df Size, u32 Length = DEF_OCTTREE_LENGTH);
        
        void searchTreeNodes(dim::vector3df &Pos, dim::vector3df &Radius);
        
        void searchTreeNodes(dim::line3df &Line);
        
        /* Inline functions */
        
        inline std::vector<dim::ptriangle3df>& getTriangleList()
        {
            return TriangleList_;
        }
        inline std::vector<u32> getTriangleNrList() const
        {
            return TriangleNrList_;
        }
        
        /* Static functions */
        
        static void clearTreeList();
        
        static std::vector<OctTree*>& getTreeList();
        
    private:
        
        /* Structures */
        
        struct SOctTreeTransportData
        {
            dim::vector3df Pos, Size;
            dim::vector3df Corner1, Corner2;
            
            u32 Index, Length;
            Entity* Object;
            bool doCreateChildren;
            
            dim::ptriangle3df* pTriangleList;
            u32 TriangleListCount;
        };
        
        /* Functions */
        
        OctTree(SOctTreeTransportData Transporter);
        
        void createChildren(Entity* hEntity, dim::vector3df &Corner1, dim::vector3df &Corner2);
        
        void fillTriangleList(Entity* hEntity, dim::vector3df &Corner1, dim::vector3df &Corner2);
        
        void beginTriangleList(Entity* hEntity);
        void endTriangleList();
        
        /* Members */
        
        OctTree* Children_[8];
        dim::vector3df Pos_, Size_;
        u32 Index_, Length_;
        bool hasChildren_;
        
        dim::ptriangle3df* TempTriangleList_;
        u32 TempTriangleCount_;
        
        std::vector<dim::ptriangle3df> TriangleList_;
        std::vector<u32> TriangleNrList_;
        
};


/*
 * ======= Collisions =======
 */

enum ECollisionTypes
{
    COLLISION_NONE = 0,
    COLLISION_SPHERE_TO_SPHERE,
    COLLISION_SPHERE_TO_BOX,
    COLLISION_SPHERE_TO_POLYGON
};

class Collision;

struct SCollisionObject
{
    SCollisionObject() : hColl(0), Object(0), TriangleList(0), TriangleCount(0)
    {
    }
    ~SCollisionObject()
    {
        spDeleteArray(TriangleList);
    }
    
    /* Collision */
    Collision* hColl;
    
    /* Collision object */
    Entity* Object;
    
    /* Collision dimension information */
    f32 Radius;
    dim::line3df Box;
    dim::vector3df LastPosition;
    video::EFaceTypes FaceType;
    
    /* Triangle information */
    dim::ptriangle3df* TriangleList;
    u32 TriangleCount;
};


class SP_EXPORT Collision
{
    
    public:
        
        Collision();
        Collision(u32 ID, std::vector<s32> &TypeList, std::vector<u32> &DestinationIDList, std::list<Collision*> &CollisionTypeList);
        ~Collision();
        
        inline u32 getID()
        {
            return ID_;
        }
        inline s32 getType(u32 i)
        {
            return TypeList_[i];
        }
        inline void setType(u32 i, s32 Type)
        {
            TypeList_[i] = Type;
        }
        
        inline std::list<SCollisionObject*>& getEntityList()
        {
            return CollisionEntityList_;
        }
        inline std::vector<Collision*>& getDestinationList()
        {
            return DestinationCollisionList_;
        }
        
    private:
        
        /* Functions */
        
        void init(std::vector<u32> &DestinationIDList, std::list<Collision*> &CollisionTypeList);
        
        /* Members */
        
        u32 ID_;
        
        // List of all entities which are using this collision
        std::list<SCollisionObject*> CollisionEntityList_;
        
        std::vector<Collision*> DestinationCollisionList_;
        
        std::vector<s32> TypeList_;
        
};


/*
 * ======= CollisionDetector =======
 */

enum EPickModeTypes
{
    PICKMODE_NONE = 0,
    PICKMODE_SPHERE,
    PICKMODE_BOX,
    PICKMODE_POLYGON
};

struct SPickingObject
{
    SPickingObject() : Object(0), UserData(0), TriangleList(0), TriangleCount(0)
    {
        Distance = 0.0f;
    }
    ~SPickingObject()
    {
    }
    
    /* Object */
    Entity* Object;
    void* UserData; // !NO LONGER NEEDED -> Node class has user data!
    
    /* Pickable members */
    s32 PickMode;
    dim::triangle3df Triangle;
    dim::vector3df Intersection;
    
    s32 TriangleIndex;
    
    /* Square distance */
    f32 Distance;
    
    /* Sphere & bounding box information */
    f32 Radius;
    dim::line3df Box;
    
    /* Triangle information (not useful for users) */
    dim::ptriangle3df* TriangleList;
    u32 TriangleCount;
};


class SP_EXPORT CollisionDetector
{
    
    public:
        
        CollisionDetector();
        ~CollisionDetector();
        
        /* === Collision detection === */
        
        void updateScene(); // All collision detections
        
        Collision* createCollision(u32 CollisionID, std::vector<s32> &CollisionTypeList, std::vector<u32> &DestinationIDList);
        void deleteCollision(Collision* hColl);
        
        void addCollisionEntity(
            Entity* hEntity, Collision* hColl, f32 Radius = 0.5f,
            dim::line3df Box = dim::line3df(-0.5f, 0.5f),
            video::EFaceTypes FaceType = video::FACE_FRONT
        );
        void removeCollisionEntity(Entity* hEntity);
        
        void setCollisionEntityData(
            Entity* hEntity, Collision* hColl, f32 Radius = 0.5f,
            dim::line3df Box = dim::line3df(-0.5f, 0.5f),
            video::EFaceTypes FaceType = video::FACE_FRONT
        );
        
        /* === Picking intersection === */
        
        void addPickingEntity(Entity* hEntity, s32 PickMode, void* UserData = 0);
        void addPickingEntity(Entity* hEntity, f32 Radius, dim::line3df Box, s32 PickMode, void* UserData = 0);
        void removePickingEntity(Entity* hEntity);
        
        std::list<SPickingObject> pickLineIntersection(dim::line3df Line, bool isAutoSort = true);
        
        bool areJointsInview(const dim::vector3df PosA, const dim::vector3df PosB);
        
        bool areNodesInview(Node* Object1, Node* Object2);
        
    private:
        
        /* === Functions === */
        
        /* Collision detection */
        void checkCollisionSphereToSphere           (SCollisionObject* CheckingObject, SCollisionObject* CurrentObject);
        void checkCollisionSphereToBox              (SCollisionObject* CheckingObject, SCollisionObject* CurrentObject);
        void checkCollisionSphereToPolygon          (SCollisionObject* CheckingObject, SCollisionObject* CurrentObject);
        void checkCollisionSphereToPolygonOptimized (SCollisionObject* CheckingObject, SCollisionObject* CurrentObject);
        
        /* Intersection tests */
        void checkIntersectionLineSphere            (SPickingObject* CurrentObject, dim::line3df &Line);
        void checkIntersectionLineBox               (SPickingObject* CurrentObject, dim::line3df &Line);
        void checkIntersectionLinePolygon           (SPickingObject* CurrentObject, dim::line3df &Line);
        void checkIntersectionLinePolygonOptimized  (SPickingObject* CurrentObject, dim::line3df &Line);
        
        /* Inview intersection tests */
        bool checkInviewLineSphere                  (SPickingObject* CurrentObject, dim::line3df &Line);
        bool checkInviewLineBox                     (SPickingObject* CurrentObject, dim::line3df &Line);
        bool checkInviewLinePolygon                 (SPickingObject* CurrentObject, dim::line3df &Line);
        bool checkInviewLinePolygonOptimized        (SPickingObject* CurrentObject, dim::line3df &Line);
        
        /* === Members === */
        
        std::list<SCollisionObject*> CollisionEntityList_;
        std::list<SPickingObject*> PickingEntityList_;
        std::list<SPickingObject> PickedObjectList_;
        
        std::list<Collision*> CollisionTypeList_;
        
        dim::matrix4f CollMatrix_, TreeMatrix_;
        
};


} // /namespace scene

} // /namespace sp


#endif





// ================================================================================
