/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * Entity file
 *
 */

#include "spSceneManager.hpp"
#include "spMainDevice.hpp"


namespace sp
{

extern video::VideoDriver* __spVideoDriver;

namespace scene
{


/*
 * Internal comparision functions
 */

bool cmpTextureLayers(const video::SSurfaceTexture &tex1, const video::SSurfaceTexture &tex2)
{
    return tex1.TextureLayer < tex2.TextureLayer;
}

bool cmpVertexCoords(const SNormalCoord &obj1, const SNormalCoord &obj2)
{
    const dim::vector3df Pos1(obj1.Position), Pos2(obj2.Position);
    
    return
        Pos1.X > Pos2.X ||
        ( math::isEqual(Pos1.X, Pos2.X) && ( Pos1.Y > Pos2.Y ||
        ( math::isEqual(Pos1.Y, Pos2.Y) && Pos1.Z > Pos2.Z ) ) );
}


/*
 * ======= Constructors =======
 */

Entity::Entity() : Node(NODE_ENTITY), OctTreeRoot_(0)
{
    init();
}
Entity::Entity(video::Texture* TextureList[6], f32 Radius, u8 Type) : Node(NODE_ENTITY), OctTreeRoot_(0)
{
    init();
    createSkyBox(TextureList, Radius, Type);
}
Entity::Entity(video::Texture* hTextureHighMap, s32 Detail) : Node(NODE_ENTITY), OctTreeRoot_(0)
{
    init();
    createHighMap(hTextureHighMap, Detail);
}
Entity::~Entity()
{
    clear();
}


/*
 * ======= Textureing =======
 */


void Entity::addTexture(video::Texture* hTexture, u8 Layer)
{
    if (!SurfaceList_.size())
        return;
    
    SMeshSurface3D* Surf = &SurfaceList_[CurSurface_];
    
    if (Surf->TextureList.size() >= video::MAX_COUNT_OF_TEXTURES)
    {
        io::printWarning(
            io::stringc("3d model can not hold more then ") + io::stringc(video::MAX_COUNT_OF_TEXTURES) + io::stringc(" textures")
        );
        return;
    }
    
    /* Fill the surface texture data */
    video::SSurfaceTexture TextureData(hTexture);
    
    if (Layer == video::TEXTURE_IGNORE)
    {
        Layer = 0;
        
        if (Surf->TextureList.size())
            Layer = (Surf->TextureList.end() - 1)->TextureLayer + 1;
        
        TextureData.TextureLayer = Layer;
    }
    else
    {
        if (Surf->TextureList.size())
        {
            if (Layer < Surf->TextureList.begin()->TextureLayer + 1)
            {
                for (std::vector<video::SSurfaceTexture>::iterator itTex = Surf->TextureList.begin() + Layer;
                     itTex != Surf->TextureList.end(); ++itTex)
                {
                    ++itTex->TextureLayer;
                }
            }
            else if (Layer > Surf->TextureList.begin()->TextureLayer + 1)
                TextureData.TextureLayer = Surf->TextureList.begin()->TextureLayer + 1;
        }
        else
            TextureData.TextureLayer = 0;
    }
    
    /* Add the texture */
    Surf->TextureList.push_back(TextureData);
    
    /* Sort the texture list */
    sort(Surf->TextureList.begin(), Surf->TextureList.end(), cmpTextureLayers);
}

void Entity::removeTexture(u8 Layer)
{
    if (!SurfaceList_.size())
        return;
    
    SMeshSurface3D* Surf = &SurfaceList_[CurSurface_];
    
    if (Layer >= Surf->TextureList.size())
        return;
    
    std::vector<video::SSurfaceTexture>::iterator itTex = Surf->TextureList.begin() + Layer;
    
    /* Remove the texture */
    SurfaceList_[CurSurface_].TextureList.erase(itTex);
    
    if (Layer >= Surf->TextureList.size())
        return;
    
    for (std::vector<video::SSurfaceTexture>::iterator itTex = Surf->TextureList.begin() + Layer;
         itTex != Surf->TextureList.end(); ++itTex)
    {
        --itTex->TextureLayer;
    }
}

void Entity::doTextureTranslate(u8 Layer, dim::point2df Direction)
{
    for (std::vector<video::SMeshVertex3D>::iterator it = SurfaceList_[CurSurface_].VerticesList.begin();
         it != SurfaceList_[CurSurface_].VerticesList.end(); ++it)
    {
        it->TexCoord[Layer].X += Direction.X;
        it->TexCoord[Layer].Y += Direction.Y;
    }
    
    updateBuffersVBO(CurSurface_);
}

void Entity::doTextureTransform(u8 Layer, dim::size2df Size)
{
    for (std::vector<video::SMeshVertex3D>::iterator it = SurfaceList_[CurSurface_].VerticesList.begin();
         it != SurfaceList_[CurSurface_].VerticesList.end(); ++it)
    {
        it->TexCoord[Layer].X *= Size.Width;
        it->TexCoord[Layer].Y *= Size.Height;
    }
    
    updateBuffersVBO(CurSurface_);
}

void Entity::doTextureTurn(u8 Layer, f32 Rotation)
{
    dim::matrix4f Matrix;
    Matrix.matrixReset();
    Matrix.matrixRotateZ(Rotation);
    
    for (std::vector<video::SMeshVertex3D>::iterator it = SurfaceList_[CurSurface_].VerticesList.begin();
         it != SurfaceList_[CurSurface_].VerticesList.end(); ++it)
    {
        it->TexCoord[Layer] = Matrix * it->TexCoord[Layer];
    }
    
    updateBuffersVBO(CurSurface_);
}

void Entity::setTexture(u8 Layer, video::Texture* hTexture)
{
    if (Layer >= SurfaceList_[CurSurface_].TextureList.size())
        return;
    
    SurfaceList_[CurSurface_].TextureList[Layer].hTexture = hTexture;
}
video::Texture* Entity::getTexture(u8 Layer) const
{
    if (Layer >= SurfaceList_[CurSurface_].TextureList.size())
        return 0;
    
    return SurfaceList_[CurSurface_].TextureList[Layer].hTexture;
}

void Entity::setTextureMatrix(u8 Layer, const dim::matrix4f &Matrix)
{
    if (Layer >= SurfaceList_[CurSurface_].TextureList.size())
        return;
    
    SurfaceList_[CurSurface_].TextureList[Layer].Matrix = Matrix;
}
dim::matrix4f Entity::getTextureMatrix(u8 Layer) const
{
    if (Layer >= SurfaceList_[CurSurface_].TextureList.size())
        return dim::matrix4f();
    
    return SurfaceList_[CurSurface_].TextureList[Layer].Matrix;
}

void Entity::setTextureEnv(u8 Layer, video::ETextureEnvTypes Type)
{
    if (Layer >= SurfaceList_[CurSurface_].TextureList.size())
        return;
    
    SurfaceList_[CurSurface_].TextureList[Layer].TexEnvType = Type;
}
video::ETextureEnvTypes Entity::getTextureEnv(u8 Layer) const
{
    if (Layer >= SurfaceList_[CurSurface_].TextureList.size())
        return video::TEXENV_MODULATE;
    
    return SurfaceList_[CurSurface_].TextureList[Layer].TexEnvType;
}

void Entity::setMappingGen(u8 Layer, video::EMappingGenTypes Type)
{
    if (Layer >= SurfaceList_[CurSurface_].TextureList.size())
        return;
    
    /* Select the mapping generation */
    switch (Type)
    {
        case video::MAPGEN_DISABLE:
            setMappingGenCoords(Layer, video::MAPGEN_NONE);
            break;
        case video::MAPGEN_OBJECT_LINEAR:
        case video::MAPGEN_EYE_LINEAR:
        case video::MAPGEN_SPHERE_MAP:
        case video::MAPGEN_NORMAL_MAP:
            setMappingGenCoords(Layer, video::MAPGEN_S | video::MAPGEN_T);
            break;
        case video::MAPGEN_REFLECTION_MAP:
            setMappingGenCoords(Layer, video::MAPGEN_S | video::MAPGEN_T | video::MAPGEN_R);
            break;
    }
    
    SurfaceList_[CurSurface_].TextureList[Layer].TexMappingGen  = Type;
}
video::EMappingGenTypes Entity::getMappingGen(u8 Layer) const
{
    if (Layer >= SurfaceList_[CurSurface_].TextureList.size())
        return video::MAPGEN_OBJECT_LINEAR;
    
    return SurfaceList_[CurSurface_].TextureList[Layer].TexMappingGen;
}

void Entity::setMappingGenCoords(u8 Layer, s32 Coords)
{
    if (Layer > SurfaceList_[CurSurface_].TextureList.size())
        return;
    
    SurfaceList_[CurSurface_].TextureList[Layer].TexMappingCoords = Coords;
}
s32 Entity::getMappingGenCoords(u8 Layer) const
{
    if (Layer >= SurfaceList_[CurSurface_].TextureList.size())
        return video::MAPGEN_NONE;
    
    return SurfaceList_[CurSurface_].TextureList[Layer].TexMappingCoords;
}

std::list<video::Texture*> Entity::getTextureList(u32 Surface) const
{
    std::list<video::Texture*> TexList;
    std::vector<video::SSurfaceTexture>::const_iterator itTex;
    
    if (Surface == SURFACE_IGNORE)
    {
        for (std::vector<SMeshSurface3D>::const_iterator itSurf = SurfaceList_.begin(); itSurf != SurfaceList_.end(); ++itSurf)
        {
            for (itTex = itSurf->TextureList.begin();
                 itTex != itSurf->TextureList.end(); ++itTex)
            {
                TexList.push_back(itTex->hTexture);
            }
        }
    }
    else
    {
        for (itTex = SurfaceList_[Surface].TextureList.begin();
             itTex != SurfaceList_[Surface].TextureList.end(); ++itTex)
        {
            TexList.push_back(itTex->hTexture);
        }
    }
    
    return TexList;
}


/*
 * ======= Mesh control =======
 */


void Entity::updateNormals(u32 Surface)
{
    
    VertexCounter_ = 0;
    
    if (Surface == SURFACE_IGNORE)
    {
        for (u32 s = 0; s < SurfaceList_.size(); ++s)
            updateNormals(s);
        return;
    }
    
    switch (NormalsType_)
    {
        case NORMALS_SIMPLE:
        {
            dim::vector3df TriangleNormal;
            
            for (u32 i = 0; i < SurfaceList_[Surface].TriangleCount; ++i)
            {
                TriangleNormal = math::getNormalVector(
                    SurfaceList_[Surface].VerticesList[ SurfaceList_[Surface].TriangleIndicesList[i*3+0] ].Position,
                    SurfaceList_[Surface].VerticesList[ SurfaceList_[Surface].TriangleIndicesList[i*3+1] ].Position,
                    SurfaceList_[Surface].VerticesList[ SurfaceList_[Surface].TriangleIndicesList[i*3+2] ].Position
                );
                
                SurfaceList_[Surface].VerticesList[ SurfaceList_[Surface].TriangleIndicesList[i*3+0] ].Normal = TriangleNormal;
                SurfaceList_[Surface].VerticesList[ SurfaceList_[Surface].TriangleIndicesList[i*3+1] ].Normal = TriangleNormal;
                SurfaceList_[Surface].VerticesList[ SurfaceList_[Surface].TriangleIndicesList[i*3+2] ].Normal = TriangleNormal;
            }
        }
        break;
        
        case NORMALS_SMOOTH:
        {
            calcSurfaceNoramls(Surface);
        }
        break;
    }
    
    updateBuffersVBO(Surface);
    
}


void Entity::doMeshTranslate(dim::vector3df Direction, u32 Surface)
{
    if (Surface == SURFACE_IGNORE)
    {
        for (u32 s = 0; s < SurfaceList_.size(); ++s)
            doMeshTranslate(Direction, s);
    }
    else if (Surface < getMeshSurfaceCount())
    {
        for (std::vector<video::SMeshVertex3D>::iterator itVert = SurfaceList_[Surface].VerticesList.begin();
             itVert != SurfaceList_[Surface].VerticesList.end(); ++itVert)
        {
            itVert->Position += Direction;
        }
        updateBuffersVBO(Surface);
    }
}

void Entity::doMeshTransform(dim::vector3df Size, u32 Surface)
{
    if (Surface == SURFACE_IGNORE)
    {
        for (u32 s = 0; s < SurfaceList_.size(); ++s)
            doMeshTransform(Size, s);
    }
    else if (Surface < getMeshSurfaceCount())
    {
        for (std::vector<video::SMeshVertex3D>::iterator itVert = SurfaceList_[Surface].VerticesList.begin();
             itVert != SurfaceList_[Surface].VerticesList.end(); ++itVert)
        {
            itVert->Position *= Size;
        }
        updateBuffersVBO(Surface);
    }
}

void Entity::doMeshTransform(dim::matrix4f Matrix, u32 Surface)
{
    if (Surface == SURFACE_IGNORE)
    {
        for (u32 s = 0; s < SurfaceList_.size(); ++s)
            doMeshTransform(Matrix, s);
    }
    else if (Surface < getMeshSurfaceCount())
    {
        for (std::vector<video::SMeshVertex3D>::iterator itVert = SurfaceList_[Surface].VerticesList.begin();
             itVert != SurfaceList_[Surface].VerticesList.end(); ++itVert)
        {
            itVert->Position    = Matrix * itVert->Position;
            itVert->Normal      = Matrix * itVert->Normal;
            itVert->Normal.normalize();
        }
        updateBuffersVBO(Surface);
    }
}

void Entity::doMeshTurn(dim::vector3df Rotation, u32 Surface)
{
    /* Matrix rotation (for more performance) */
    dim::matrix4f RotationMatrix;
    RotationMatrix.matrixReset();
    RotationMatrix.matrixRotateYXZ(Rotation);
    
    /* Updating vertex coordinates */
    if (Surface == SURFACE_IGNORE)
    {
        for (u32 s = 0; s < SurfaceList_.size(); ++s)
            doMeshTurn(Rotation, s);
    }
    else if (Surface < getMeshSurfaceCount())
    {
        for (std::vector<video::SMeshVertex3D>::iterator itVert = SurfaceList_[Surface].VerticesList.begin();
             itVert != SurfaceList_[Surface].VerticesList.end(); ++itVert)
        {
            itVert->Position    = RotationMatrix * itVert->Position;
            itVert->Normal      = RotationMatrix * itVert->Normal;
            itVert->Normal.normalize();
        }
        
        updateBuffersVBO(Surface);
    }
}

void Entity::flipMesh(u32 Surface)
{
    if (Surface == SURFACE_IGNORE)
    {
        for (u32 s = 0; s < SurfaceList_.size(); ++s)
            flipMesh(s);
    }
    else if (Surface < getMeshSurfaceCount())
    {
        for (std::vector<video::SMeshVertex3D>::iterator itVert = SurfaceList_[Surface].VerticesList.begin();
             itVert != SurfaceList_[Surface].VerticesList.end(); ++itVert)
        {
            itVert->Position.setInvert();
        }
        updateBuffersVBO(Surface);
    }
}

void Entity::flipMesh(bool isXAxis, bool isYAxis, bool isZAxis, u32 Surface)
{
    if (Surface == SURFACE_IGNORE)
    {
        for (u32 s = 0; s < SurfaceList_.size(); ++s)
            flipMesh(isXAxis, isYAxis, isZAxis, s);
    }
    else if (Surface < getMeshSurfaceCount())
    {
        for (std::vector<video::SMeshVertex3D>::iterator itVert = SurfaceList_[Surface].VerticesList.begin();
             itVert != SurfaceList_[Surface].VerticesList.end(); ++itVert)
        {
            if (isXAxis)
                itVert->Position.X = -itVert->Position.X;
            if (isYAxis)
                itVert->Position.Y = -itVert->Position.Y;
            if (isZAxis)
                itVert->Position.Z = -itVert->Position.Z;
        }
        
        updateBuffersVBO(Surface);
    }
}

void Entity::fitMesh(dim::vector3df Position, dim::vector3df Size)
{
    const f32 Max = 999999.0f;
    
    u32 surf = 0;
    
    dim::vector3df MinPos(Max, Max, Max);
    dim::vector3df MaxPos(-Max, -Max, -Max);
    
    /* Find the dimensions */
    for (std::vector<SMeshSurface3D>::iterator itSurf = SurfaceList_.begin(); itSurf != SurfaceList_.end(); ++itSurf)
    {
        for (std::vector<video::SMeshVertex3D>::iterator it = itSurf->VerticesList.begin(); it != itSurf->VerticesList.end(); ++it)
        {
            if (it->Position.X < MinPos.X) MinPos.X = it->Position.X;
            if (it->Position.Y < MinPos.Y) MinPos.Y = it->Position.Y;
            if (it->Position.Z < MinPos.Z) MinPos.Z = it->Position.Z;
        }
        for (std::vector<video::SMeshVertex3D>::iterator it = itSurf->VerticesList.begin(); it != itSurf->VerticesList.end(); ++it)
        {
            if (it->Position.X > MaxPos.X) MaxPos.X = it->Position.X;
            if (it->Position.Y > MaxPos.Y) MaxPos.Y = it->Position.Y;
            if (it->Position.Z > MaxPos.Z) MaxPos.Z = it->Position.Z;
        }
    }
    
    /* Fit the mesh */
    for (std::vector<SMeshSurface3D>::iterator itSurf = SurfaceList_.begin(); itSurf != SurfaceList_.end(); ++itSurf, ++surf)
    {
        for (std::vector<video::SMeshVertex3D>::iterator it = itSurf->VerticesList.begin(); it != itSurf->VerticesList.end(); ++it)
        {
            it->Position -= MinPos;
            it->Position /= (MaxPos - MinPos);
            it->Position *= (Size - Position);
            it->Position += Position;
        }
        updateBuffersVBO(surf);
    }
}

void Entity::flipTriangles(u32 Surface)
{
    if (Surface == SURFACE_IGNORE)
    {
        for (u32 s = 0; s < SurfaceList_.size(); ++s)
            flipTriangles(s);
    }
    else if (Surface < getMeshSurfaceCount())
    {
        u32* Index = &(SurfaceList_[Surface].TriangleIndicesList[0]);
        
        for (s32 i = 0, c = SurfaceList_[Surface].TriangleIndicesList.size(); i < c; i += 3, Index += 3)
            math::Swap(Index[0], Index[2]);
    }
}


/*
 * ======= Surfaces =======
 */

void Entity::updateCopyBuffersVBO()
{
    u32 surf = 0;
    for (std::vector<SMeshSurface3D>::iterator itSurf = SurfaceList_.begin(); itSurf != SurfaceList_.end(); ++itSurf, ++surf)
    {
        /* Generate a new vertex buffer */
        __spVideoDriver->createVertexBuffer(itSurf->ID);
        
        /* Update the vertex buffer */
        updateBuffersVBO(surf);
    }
}

u32 Entity::createSurface()
{
    SMeshSurface3D SurfaceData;
    {
        /* Generate a new vertex buffer */
        __spVideoDriver->createVertexBuffer(SurfaceData.ID);
    }
    
    SurfaceList_.push_back(SurfaceData);
    VertexCounter_ = 0;
    
    return CurSurface_ = SurfaceList_.size() - 1;
}

void Entity::removeSurface(u32 Surface)
{
    std::vector<SMeshSurface3D>::iterator itSurf = SurfaceList_.begin() + Surface;
    
    /* Delete vertex buffer */
    __spVideoDriver->deleteVertexBuffer(itSurf->ID);
    
    SurfaceList_.erase(itSurf);
}

// !TODO: optimize speed with sorting and not (surface count) ^ 2!
void Entity::optimizeSurfaces()
{
    
    /* Temporary variables */
    std::vector<SMeshSurface3D> NewSurfaceList = SurfaceList_, TempSurfaceList;
    std::vector< std::vector<SMeshSurface3D> > TempSurfacesLists;
    
    SMeshSurface3D NewSurface;
    
    u32 VerticesIndicesStart;
    
    /* === Find surfaces with equal texture lists === */
    
    for (std::vector<SMeshSurface3D>::iterator it = SurfaceList_.begin(), itsnd; it != SurfaceList_.end(); ++it)
    {
        TempSurfaceList.clear();
        
        for (itsnd = NewSurfaceList.begin(); itsnd != NewSurfaceList.end();)
        {
            if (it->TextureList == itsnd->TextureList)
            {
                TempSurfaceList.push_back(*itsnd);
                itsnd = NewSurfaceList.erase(itsnd);
            }
            else
                ++itsnd;
        }
        
        TempSurfacesLists.push_back(TempSurfaceList);
    }
    
    /* === Connect the surfaces lists === */
    
    /* Delete the old surfaces */
    for (std::vector<SMeshSurface3D>::iterator it = SurfaceList_.begin(); it != SurfaceList_.end(); ++it)
        __spVideoDriver->deleteVertexBuffer(it->ID);
    SurfaceList_.clear();
    
    for (std::vector< std::vector<SMeshSurface3D> >::iterator it = TempSurfacesLists.begin(); it != TempSurfacesLists.end(); ++it)
    {
        if (!it->size())
            continue;
        
        /* Reset the new surface */
        VerticesIndicesStart        = 0;
        NewSurface.Name             = "";
        NewSurface.IndicesCount     = 0;
        NewSurface.TriangleCount    = 0;
        NewSurface.TextureList      = it->begin()->TextureList;
        NewSurface.VerticesList.clear();
        NewSurface.TriangleIndicesList.clear();
        
        /* Add the surface */
        for (std::vector<SMeshSurface3D>::iterator itsurf = it->begin(); itsurf != it->end(); ++itsurf)
        {
            NewSurface.IndicesCount     += itsurf->IndicesCount;
            NewSurface.TriangleCount    += itsurf->TriangleCount;
            
            for (std::vector<video::SMeshVertex3D>::iterator itvert = itsurf->VerticesList.begin();
                 itvert != itsurf->VerticesList.end(); ++itvert)
            {
                NewSurface.VerticesList.push_back(*itvert);
            }
            
            for (std::vector<u32>::iterator itface = itsurf->TriangleIndicesList.begin();
                 itface != itsurf->TriangleIndicesList.end(); ++itface)
            {
                NewSurface.TriangleIndicesList.push_back(VerticesIndicesStart + *itface);
            }
            
            VerticesIndicesStart += itsurf->VerticesList.size();
        }
        
        /* Generate a new vertex buffer */
        __spVideoDriver->createVertexBuffer(NewSurface.ID);
        
        SurfaceList_.push_back(NewSurface);
    }
    
    updateModeling(false);
    
}

void Entity::clearTriangleList(u32 Surface)
{
    SurfaceList_[Surface].IndicesCount = 0;
    SurfaceList_[Surface].TriangleCount = 0;
    
    SurfaceList_[Surface].TriangleIndicesList.clear();
    SurfaceList_[Surface].VerticesList.clear();
    
    updateBuffersVBO(Surface);
}


/*
 * ======= Vertices =======
 */


u32 Entity::getMeshVerticesCount(u32 Surface) const
{
    if (Surface == SURFACE_IGNORE)
    {
        u32 VerticesCounter = 0;
        
        for (std::vector<SMeshSurface3D>::const_iterator itSurf = SurfaceList_.begin(); itSurf != SurfaceList_.end(); ++itSurf)
            VerticesCounter += itSurf->VerticesList.size();
        
        return VerticesCounter;
    }
    
    return SurfaceList_[Surface].VerticesList.size();
}
u32 Entity::getMeshTriangleCount(u32 Surface) const
{
    if (Surface == SURFACE_IGNORE)
    {
        u32 TriangleCounter = 0;
        
        for (std::vector<SMeshSurface3D>::const_iterator itSurf = SurfaceList_.begin(); itSurf != SurfaceList_.end(); ++itSurf)
            TriangleCounter += itSurf->TriangleCount;
        
        return TriangleCounter;
    }
    
    return SurfaceList_[Surface].TriangleCount;
}

void Entity::getMeshSurfaceByTriangle(u32 &TriangleIndex, u32 &SurfaceIndex) const
{
    for (SurfaceIndex = 0; SurfaceIndex < SurfaceList_.size(); ++SurfaceIndex)
    {
        if (TriangleIndex >= SurfaceList_[SurfaceIndex].TriangleCount)
            TriangleIndex -= SurfaceList_[SurfaceIndex].TriangleCount;
        else
            break;
    }
}

void Entity::setVertexTexCoord(u32 VertexIndex, STexCoord UVMapping, u8 Layer)
{
    if (Layer == video::TEXTURE_IGNORE)
    {
        for (s32 i = 0; i < video::MAX_COUNT_OF_TEXTURES; ++i)
            SurfaceList_[CurSurface_].VerticesList[VertexIndex + VertexCounter_].TexCoord[i] = UVMapping.vec3<f32>();
    }
    else
        SurfaceList_[CurSurface_].VerticesList[VertexIndex + VertexCounter_].TexCoord[Layer] = UVMapping.vec3<f32>();
}

u32 Entity::addVertex(dim::vector3df Position, STexCoord TexCoord, video::color Color)
{
    video::SMeshVertex3D VertexData;
    {
        VertexData.Position = Position;
        VertexData.Color    = Color;
        
        for (s32 i = 0; i < video::MAX_COUNT_OF_TEXTURES; ++i)
            VertexData.TexCoord[i] = TexCoord.vec3<f32>();
    }
    SurfaceList_[CurSurface_].VerticesList.push_back(VertexData);
    
    return SurfaceList_[CurSurface_].VerticesList.size() - 1;
}

u32 Entity::addVertex(
    dim::vector3df Position, dim::vector3df Normal, STexCoord TexCoord, video::color Color, f32 Fog)
{
    video::SMeshVertex3D VertexData;
    {
        VertexData.Position = Position;
        VertexData.Color    = Color;
        VertexData.Normal   = Normal;
        VertexData.Fog      = Fog;
        
        for (s32 i = 0; i < video::MAX_COUNT_OF_TEXTURES; ++i)
            VertexData.TexCoord[i] = TexCoord.vec3<f32>();
    }
    SurfaceList_[CurSurface_].VerticesList.push_back(VertexData);
    
    return SurfaceList_[CurSurface_].VerticesList.size() - 1;
}

void Entity::removeVertex(u32 VertexIndex)
{
    /* Check if the index is to high */
    if (VertexIndex >= SurfaceList_[CurSurface_].VerticesList.size())
        return;
    
    std::vector<u32>::iterator it;
    
    /* Erase vertex */
    SurfaceList_[CurSurface_].VerticesList.erase(
        SurfaceList_[CurSurface_].VerticesList.begin() + VertexIndex
    );
    
    std::vector<u32>* Indices = &(SurfaceList_[CurSurface_].TriangleIndicesList);
    
    u32 i, j, k;
    
    /* Erase triangle which lose the vertex */
    for (i = 0; i < SurfaceList_[CurSurface_].TriangleCount; ++i)
    {
        for (j = 0; j < 3; ++j)
        {
            k = i*3+j;
            
            if ((*Indices)[k] == VertexIndex)
            {
                removeTriangle(i--);
                break;
            }
            else if ((*Indices)[k] > VertexIndex)
                --(*Indices)[k];
        }
    }
}

u32 Entity::addTriangle(u32 Vertex1, u32 Vertex2, u32 Vertex3)
{
    const u32 MaxIndex = SurfaceList_[CurSurface_].VerticesList.size();
    
    if (Vertex1 < MaxIndex && Vertex2 < MaxIndex && Vertex3 < MaxIndex)
    {
        SurfaceList_[CurSurface_].TriangleIndicesList.push_back(Vertex1 + VertexCounter_);
        SurfaceList_[CurSurface_].TriangleIndicesList.push_back(Vertex2 + VertexCounter_);
        SurfaceList_[CurSurface_].TriangleIndicesList.push_back(Vertex3 + VertexCounter_);
        
        SurfaceList_[CurSurface_].IndicesCount += 3;
        
        return ++SurfaceList_[CurSurface_].TriangleCount;
    }
    
    return 0;
}

void Entity::removeTriangle(u32 TriangleIndex)
{
    if (TriangleIndex < SurfaceList_[CurSurface_].TriangleIndicesList.size())
    {
        std::vector<u32>::iterator it = SurfaceList_[CurSurface_].TriangleIndicesList.begin() + TriangleIndex*3;
        
        SurfaceList_[CurSurface_].TriangleIndicesList.erase(it, it + 3);
        
        SurfaceList_[CurSurface_].IndicesCount  -= 3;
        SurfaceList_[CurSurface_].TriangleCount -= 1;
    }
}

void Entity::getTriangleVertices(u32 TriangleIndex, u32 Vertices[3]) const
{
    Vertices[0] = SurfaceList_[CurSurface_].TriangleIndicesList[TriangleIndex*3+0];
    Vertices[1] = SurfaceList_[CurSurface_].TriangleIndicesList[TriangleIndex*3+1];
    Vertices[2] = SurfaceList_[CurSurface_].TriangleIndicesList[TriangleIndex*3+2];
}

void Entity::getBoundingBox(dim::vector3df &LeftBottomFrontCorner, dim::vector3df &RightTopBackCorner) const
{
    dim::vector3df Pos;
    
    LeftBottomFrontCorner   = 999999;
    RightTopBackCorner      = -999999;
    
    /* Loop all surfaces */
    for (std::vector<SMeshSurface3D>::const_iterator itSurf = SurfaceList_.begin(); itSurf != SurfaceList_.end(); ++itSurf)
    {
        /* Loop all vertices */
        for (std::vector<video::SMeshVertex3D>::const_iterator it = itSurf->VerticesList.begin(); it != itSurf->VerticesList.end(); ++it)
        {
            /* Get the current vertex coord */
            Pos = it->Position;
            
            /* LeftBottomFrontCorner */
            if (Pos.X < LeftBottomFrontCorner.X) LeftBottomFrontCorner.X = Pos.X;
            if (Pos.Y < LeftBottomFrontCorner.Y) LeftBottomFrontCorner.Y = Pos.Y;
            if (Pos.Z < LeftBottomFrontCorner.Z) LeftBottomFrontCorner.Z = Pos.Z;
            
            /* RightTopBackCorner */
            if (Pos.X > RightTopBackCorner.X) RightTopBackCorner.X = Pos.X;
            if (Pos.Y > RightTopBackCorner.Y) RightTopBackCorner.Y = Pos.Y;
            if (Pos.Z > RightTopBackCorner.Z) RightTopBackCorner.Z = Pos.Z;
        }
    }
}

dim::line3df Entity::getBoundingBox() const
{
    dim::line3df BoundingBox;
    
    getBoundingBox(BoundingBox.Corner1, BoundingBox.Corner2);
    
    return BoundingBox;
}

void Entity::setVertexCoordArray(dim::vector3df* pVector)
{
    u32 i = 0;
    for (std::vector<SMeshSurface3D>::iterator itSurf = SurfaceList_.begin(); itSurf != SurfaceList_.end(); ++itSurf)
    {
        for (std::vector<video::SMeshVertex3D>::iterator it = itSurf->VerticesList.begin(); it != itSurf->VerticesList.end(); ++it, ++i)
            it->Position = pVector[i];
    }
}

void Entity::getVertexCoordArray(dim::vector3df* pVector)
{
    u32 i = 0;
    for (std::vector<SMeshSurface3D>::iterator itSurf = SurfaceList_.begin(); itSurf != SurfaceList_.end(); ++itSurf)
    {
        for (std::vector<video::SMeshVertex3D>::iterator it = itSurf->VerticesList.begin(); it != itSurf->VerticesList.end(); ++it, ++i)
            pVector[i] = it->Position;
    }
}

void Entity::getTriangleArray(u32* pTriangle)
{
    u32 i = 0, j = 0;
    for (std::vector<SMeshSurface3D>::iterator itSurf = SurfaceList_.begin(); itSurf != SurfaceList_.end(); ++itSurf)
    {
        for (i = 0; i < itSurf->IndicesCount; ++i, ++j)
            pTriangle[j] = itSurf->TriangleIndicesList[i];
    }
}

void Entity::getTriangleList(dim::triangle3df* pTriangle, u32 &TriangelCount)
{
    u32 i = 0, j = 0;
    
    for (std::vector<SMeshSurface3D>::iterator itSurf = SurfaceList_.begin(); itSurf != SurfaceList_.end(); ++itSurf)
    {
        for (i = 0; i < itSurf->TriangleCount; ++i, ++j)
        {
            pTriangle[j] = dim::triangle3df(
                itSurf->VerticesList[ itSurf->TriangleIndicesList[i*3+0] ].Position,
                itSurf->VerticesList[ itSurf->TriangleIndicesList[i*3+1] ].Position,
                itSurf->VerticesList[ itSurf->TriangleIndicesList[i*3+2] ].Position
            );
        }
    }
    
    TriangelCount = j;
}

void Entity::getTriangleList(dim::ptriangle3df* pTriangle, u32 &TriangelCount)
{
    u32 i = 0, j = 0;
    
    dim::ptriangle3df Triangle;
    
    for (std::vector<SMeshSurface3D>::iterator itSurf = SurfaceList_.begin(); itSurf != SurfaceList_.end(); ++itSurf)
    {
        for (i = 0; i < itSurf->TriangleCount; ++i, ++j)
        {
            Triangle.PointA = &itSurf->VerticesList[ itSurf->TriangleIndicesList[i*3+0] ].Position;
            Triangle.PointB = &itSurf->VerticesList[ itSurf->TriangleIndicesList[i*3+1] ].Position;
            Triangle.PointC = &itSurf->VerticesList[ itSurf->TriangleIndicesList[i*3+2] ].Position;
            
            pTriangle[j] = Triangle;
        }
    }
    
    TriangelCount = j;
}

void Entity::updateModeling(bool isNormalsUpdate)
{
    VertexCounter_ = 0;
    if (isNormalsUpdate)
    {
        updateNormals();
    }
    else
    {
        u32 surf = 0;
        for (std::vector<SMeshSurface3D>::iterator itSurf = SurfaceList_.begin(); itSurf != SurfaceList_.end(); ++itSurf, ++surf)
            updateBuffersVBO(surf);
    }
}

void Entity::updateBuffersVBO(u32 Surface)
{
    /* Update the vertex buffer */
    __spVideoDriver->updateVertexBuffer(
        SurfaceList_[Surface].ID,
        &SurfaceList_[Surface].VerticesList[0],
        SurfaceList_[Surface].VerticesList.size()
    );
}


/*
 * ======= Collision detection =======
 */

void Entity::resetCollision()
{
    CollisionList_.clear();
}

void Entity::simulateCollision(SCollisionData CollisionInformation)
{
    CollisionList_.push_back(CollisionInformation);
}

bool Entity::detectCollision(u32 CollisionID)
{
    for (std::list<SCollisionData>::iterator itColl = CollisionList_.begin(); itColl != CollisionList_.end(); ++itColl)
    {
        if (itColl->ID == CollisionID)
        {
            itCurColl_ = itColl;
            return true;
        }
    }
    return false;
}

#if 0

bool Entity::detectIntersection(Entity* other)
{
    /* Temporary variables */
    dim::triangle3df TriangleA, TriangleB;
    u32 i, j;
    
    dim::matrix4f MatA(getGlobalLocation()), MatB(other->getGlobalLocation());
    
    /* Loop for each triangles of this object */
    for (std::vector<SMeshSurface3D>::iterator itSurfA = SurfaceList_.begin(), itSurfB; itSurfA != SurfaceList_.end(); ++itSurfA)
    {
        for (i = 0; i < itSurfA->TriangleCount; ++i)
        {
            
            /* Get the first triangle */
            TriangleA = dim::triangle3df(
                MatA * itSurfA->VerticesList[ itSurfA->TriangleIndicesList[i*3+0] ].Coord,
                MatA * itSurfA->VerticesList[ itSurfA->TriangleIndicesList[i*3+1] ].Coord,
                MatA * itSurfA->VerticesList[ itSurfA->TriangleIndicesList[i*3+2] ].Coord
            );
            
            /* Loop for each triangles of this object */
            for (itSurfB = other->SurfaceList_.begin(); itSurfB != other->SurfaceList_.end(); ++itSurfB)
            {
                for (j = 0; j < itSurfB->TriangleCount; ++j)
                {
                    
                    /* Get the second triangle */
                    TriangleB = dim::triangle3df(
                        MatB * itSurfB->VerticesList[ itSurfB->TriangleIndicesList[j*3+0] ].Coord,
                        MatB * itSurfB->VerticesList[ itSurfB->TriangleIndicesList[j*3+1] ].Coord,
                        MatB * itSurfB->VerticesList[ itSurfB->TriangleIndicesList[j*3+2] ].Coord
                    );
                    
                    /* Check for an intersection */
                    if (math::checkTriangleTriangleIntersection(TriangleA, TriangleB))
                        return true;
                    
                } // next triangle (ObjB)
            } // next surface (ObjB)
            
        } // next triangle (ObjA)
    } // next surface (ObjA)
    
    /* No intersection found */
    return false;
}

#endif

dim::vector3df Entity::getCollisionNormal()
{
    return (itCurColl_->Triangle.isEmpty()) ? itCurColl_->Normal : itCurColl_->Normal = itCurColl_->Triangle.getNormal();
}


/*
 * ======= Oct-tree optimization =======
 */


void Entity::createOctTree(u32 Length)
{
    // Delete the old octtree
    spDeletePointer(OctTreeRoot_);
    
    // Create a new octtree
    OctTreeRoot_ = new OctTree();
    
    // Set the octtree parameters
    OctTreeRoot_->createTree(this, Length);
}

void Entity::deleteOctTree()
{
    // Delete the old octtree
    spDeletePointer(OctTreeRoot_);
}


/*
 * ======= Something else =======
 */

void Entity::setWireframe(s32 TypeFront, s32 TypeBack)
{
    Material_.RenderModeFront   = TypeFront;
    Material_.RenderModeBack    = TypeBack;
}

void Entity::setBlendingMode(s32 BlendSource, s32 BlendTarget)
{
    Material_.BlendSource = BlendSource;
    Material_.BlendTarget = BlendTarget;
}


void Entity::setDepthMode(video::ESizeComparisionTypes Mode)
{
    Material_.DepthMode = Mode;
}

void Entity::setAlphaMode(video::ESizeComparisionTypes Mode, f32 AlphaReference)
{
    if (AlphaReference >= 0.0f)
    {
        math::Clamp<f32>(AlphaReference, 0.0f, 1.0f);
        Material_.AlphaReference = AlphaReference;
    }
    
    Material_.AlphaMode = Mode;
}


void Entity::setColor(video::color Color, u32 Surface, bool isForce)
{
    /* For the material lighting color */
    Material_.ColorDiffuse = Color_ = Color;
    
    /* For all surfaces */
    if (Surface == SURFACE_IGNORE)
    {
        u32 surf = 0;
        if (isForce)
        {
            for (std::vector<SMeshSurface3D>::iterator itSurf = SurfaceList_.begin(); itSurf != SurfaceList_.end(); ++itSurf, ++surf)
            {
                for (std::vector<video::SMeshVertex3D>::iterator it = itSurf->VerticesList.begin(); it != itSurf->VerticesList.end(); ++it)
                    it->Color = Color;
                updateBuffersVBO(surf);
            }
        }
        else
        {
            for (std::vector<SMeshSurface3D>::iterator itSurf = SurfaceList_.begin(); itSurf != SurfaceList_.end(); ++itSurf, ++surf)
            {
                for (std::vector<video::SMeshVertex3D>::iterator it = itSurf->VerticesList.begin(); it != itSurf->VerticesList.end(); ++it)
                {
                    it->Color.Red   = (u8)((s32)it->Color.Red   * (s32)Color.Red    / 255);
                    it->Color.Green = (u8)((s32)it->Color.Green * (s32)Color.Green  / 255);
                    it->Color.Blue  = (u8)((s32)it->Color.Blue  * (s32)Color.Blue   / 255);
                    it->Color.Alpha = (u8)((s32)it->Color.Alpha * (s32)Color.Alpha  / 255);
                }
                updateBuffersVBO(surf);
            }
        }
    }
    else
    {
        if (isForce)
        {
            for (std::vector<video::SMeshVertex3D>::iterator it = SurfaceList_[Surface].VerticesList.begin();
                it != SurfaceList_[Surface].VerticesList.end(); ++it)
            {
                it->Color = Color;
            }
            updateBuffersVBO(Surface);
        }
        else
        {
            for (std::vector<video::SMeshVertex3D>::iterator it = SurfaceList_[Surface].VerticesList.begin();
                it != SurfaceList_[Surface].VerticesList.end(); ++it)
            {
                it->Color.Red   = (u8)((s32)it->Color.Red   * (s32)Color.Red    / 255);
                it->Color.Green = (u8)((s32)it->Color.Green * (s32)Color.Green  / 255);
                it->Color.Blue  = (u8)((s32)it->Color.Blue  * (s32)Color.Blue   / 255);
                it->Color.Alpha = (u8)((s32)it->Color.Alpha * (s32)Color.Alpha  / 255);
            }
            updateBuffersVBO(Surface);
        }
    }
}


void Entity::updateModelviewMatrix()
{
    /* Compute the normal location */
    updateGlobalLocation();
}

Entity* Entity::copy()
{
    /* Allocate a new model */
    Entity* NewEntity = new Entity();
    
    /* Copy the attributes */
    NewEntity->copyAttributes(this);
    
    /* Copy the rest members */
    NewEntity->Type_          = Type_;
    
    NewEntity->SurfaceList_   = SurfaceList_;
    
    NewEntity->RenderMode_    = RenderMode_;
    NewEntity->VertexCounter_ = VertexCounter_;
    
    NewEntity->Material_      = Material_;
    
    /* Update the VBO buffers */
    NewEntity->updateCopyBuffersVBO();
    
    /* Return the new entity */
    return NewEntity;
}


/*
 * ========== Rendering ==========
 */


void Entity::render()
{
    
    /* Check if the entity has any surfaces */
    if (!SurfaceList_.size())
        return;
    
    /* Material */
    __spVideoDriver->setMaterialStatesTriangleList(Material_, Rotation_);
    
    /* Matrix transformation */
    updateModelviewMatrix();
    
    /* Process the possible user matrial begin procedure */
    if (Material_.UserMaterialBeginProc)
        Material_.UserMaterialBeginProc(this);
    
    /* Update the render matrix */
    __spVideoDriver->updateModelviewMatrix();
    
    /* Draw the current vertex-buffer-object */
    for (std::vector<SMeshSurface3D>::iterator itSurf = SurfaceList_.begin(); itSurf != SurfaceList_.end(); ++itSurf)
    {
        __spVideoDriver->drawTriangleList(
            itSurf->ID, &itSurf->TriangleIndicesList[0], itSurf->IndicesCount, itSurf->TextureList
        );
    }
    
    /* Process the possible user matrial end procedure */
    if (Material_.UserMaterialEndProc)
        Material_.UserMaterialEndProc(this);
    
    /* Unbinding the shader */
    __spVideoDriver->unbindShader();
    
    /* Cam perspective distance */
    PerspectiveDistance_ = (spViewMatrix * spWorldMatrix.getPosition()).Z;
    
}

// ! UNFINISHED !
void Entity::renderCelShading(dim::vector3df LightDirection)
{
    dim::matrix4f Matrix(getGlobalLocation());
    dim::vector3df Normal;
    f32 Shadow;
    
    for (std::vector<SMeshSurface3D>::iterator itSurf = SurfaceList_.begin(); itSurf != SurfaceList_.end(); ++itSurf)
    {
        for (std::vector<video::SMeshVertex3D>::iterator itVert = itSurf->VerticesList.begin(); itVert != itSurf->VerticesList.end(); ++itVert)
        {
            Normal = Matrix * itVert->Normal;
            Normal.normalize();
            
            Shadow = -Normal.getScalarProduct(LightDirection);
            
            if (Shadow < 0.0f)
                Shadow = 0.0f;
            
            itVert->TexCoord[0].X = Shadow;
        }
    }
    
    updateModeling(false);
}

void Entity::drawNormals(const video::color Color, const f32 NormalLength, Camera* hCamera)
{
    const dim::matrix4f Mat(getGlobalLocation());
    
    for (std::vector<SMeshSurface3D>::iterator itSurf = SurfaceList_.begin(); itSurf != SurfaceList_.end(); ++itSurf)
    {
        for (std::vector<video::SMeshVertex3D>::iterator it = itSurf->VerticesList.begin(); it != itSurf->VerticesList.end(); ++it)
        {
            __spVideoDriver->draw3DLine(
                Mat * it->Position, Mat * (it->Position + it->Normal * NormalLength), Color, hCamera
            );
        }
    }
}


/*
 * =========================== Private: ===========================
 */


// Skybox

void Entity::createSkyBox(video::Texture* TextureList[6], f32 Radius, u8 Type)
{
    
    f32 UVMap1 = 0.0f, UVMap2 = 0.0f;
    if (Type == SKYBOX_NORMAL)
        UVMap1 = 1.0f;
    else if (Type == SKYBOX_SWAPED)
        UVMap2 = 1.0f;
    
    // Back
    createSurface();
    if (Type == SKYBOX_SWAPED)
        addTexture(TextureList[1]);
    else
        addTexture(TextureList[0]);
    addVertex(dim::vector3df(-Radius,  Radius,  Radius), dim::point2df(UVMap1, UVMap1));
    addVertex(dim::vector3df( Radius,  Radius,  Radius), dim::point2df(UVMap2, UVMap1));
    addVertex(dim::vector3df( Radius, -Radius,  Radius), dim::point2df(UVMap2, UVMap2));
    addVertex(dim::vector3df(-Radius, -Radius,  Radius), dim::point2df(UVMap1, UVMap2));
    addTriangle(0, 1, 2); addTriangle(0, 2, 3);
    
    // Front
    createSurface();
    if (Type == SKYBOX_SWAPED)
        addTexture(TextureList[0]);
    else
        addTexture(TextureList[1]);
    addVertex(dim::vector3df(-Radius,  Radius, -Radius), dim::point2df(UVMap2, UVMap1));
    addVertex(dim::vector3df( Radius,  Radius, -Radius), dim::point2df(UVMap1, UVMap1));
    addVertex(dim::vector3df( Radius, -Radius, -Radius), dim::point2df(UVMap1, UVMap2));
    addVertex(dim::vector3df(-Radius, -Radius, -Radius), dim::point2df(UVMap2, UVMap2));
    addTriangle(2, 1, 0); addTriangle(3, 2, 0);
    
    // Top
    createSurface();
    addTexture(TextureList[2]);
    addVertex(dim::vector3df(-Radius,  Radius,  Radius), dim::point2df(UVMap2, UVMap2));
    addVertex(dim::vector3df( Radius,  Radius,  Radius), dim::point2df(UVMap2, UVMap1));
    addVertex(dim::vector3df( Radius,  Radius, -Radius), dim::point2df(UVMap1, UVMap1));
    addVertex(dim::vector3df(-Radius,  Radius, -Radius), dim::point2df(UVMap1, UVMap2));
    addTriangle(2, 1, 0); addTriangle(3, 2, 0);
    
    // Bottom
    createSurface();
    addTexture(TextureList[3]);
    addVertex(dim::vector3df(-Radius, -Radius,  Radius), dim::point2df(UVMap2, UVMap1));
    addVertex(dim::vector3df( Radius, -Radius,  Radius), dim::point2df(UVMap2, UVMap2));
    addVertex(dim::vector3df( Radius, -Radius, -Radius), dim::point2df(UVMap1, UVMap2));
    addVertex(dim::vector3df(-Radius, -Radius, -Radius), dim::point2df(UVMap1, UVMap1));
    addTriangle(0, 1, 2); addTriangle(0, 2, 3);
    
    // Left
    createSurface();
    addTexture(TextureList[5]);
    addVertex(dim::vector3df(-Radius,  Radius,  Radius), dim::point2df(UVMap2, UVMap1));
    addVertex(dim::vector3df(-Radius,  Radius, -Radius), dim::point2df(UVMap1, UVMap1));
    addVertex(dim::vector3df(-Radius, -Radius, -Radius), dim::point2df(UVMap1, UVMap2));
    addVertex(dim::vector3df(-Radius, -Radius,  Radius), dim::point2df(UVMap2, UVMap2));
    addTriangle(2, 1, 0); addTriangle(3, 2, 0);
    
    // Right
    createSurface();
    addTexture(TextureList[4]);
    addVertex(dim::vector3df( Radius,  Radius,  Radius), dim::point2df(UVMap1, UVMap1));
    addVertex(dim::vector3df( Radius,  Radius, -Radius), dim::point2df(UVMap2, UVMap1));
    addVertex(dim::vector3df( Radius, -Radius, -Radius), dim::point2df(UVMap2, UVMap2));
    addVertex(dim::vector3df( Radius, -Radius,  Radius), dim::point2df(UVMap1, UVMap2));
    addTriangle(0, 1, 2); addTriangle(0, 2, 3);
    
    updateNormals();
    
    Order_ = ORDER_BACKGROUND;
    Material_.isZBuffer = false;
    Material_.isLighting = false;
    
}

// Highmap (alternate of a terrain)

void Entity::createHighMap(video::Texture* hTextureHighMap, s32 Detail)
{
    
    /* Temporary variables */
    
    const f32 size = 1.0f / Detail;
    
    s32 x, z;
    s32 v0, v1, v2, v3;
    f32 y;
    
    s32 Width   = hTextureHighMap->getSize().Width;
    s32 Height  = hTextureHighMap->getSize().Height;
    
    createSurface();
    
    /* Create all vertices */
    for (z = 0; z <= Detail; ++z)
    {
        for (x = 0; x <= Detail; ++x)
        {
            y = hTextureHighMap->getPixelColor(
                dim::point2di(s32(size*x*Width), s32(size*z*Height))
            ).getBrightness<f32>() / 255;
            
            addVertex(
                dim::vector3df((f32)x*size - 0.5f, y, -(f32)z*size + 0.5f), dim::point2df((f32)x*size, (f32)z*size)
            );
        }
    }
    
    /* Create all triangles */
    for (z = 0; z < Detail; ++z)
    {
        for (x = 0; x < Detail; ++x)
        {
            /* Compute the vertices indices */
            v0 = z * ( Detail + 1 ) + x;
            v1 = z * ( Detail + 1 ) + x + 1;
            v2 = ( z + 1 ) * ( Detail + 1 ) + x + 1;
            v3 = ( z + 1 ) * ( Detail + 1 ) + x;
            
            /* Add the triangel */
            addTriangle(v0, v1, v2);
            addTriangle(v0, v2, v3);
        }
    }
    
    updateNormals();
    
}


/*
 * ======= Other private functions =======
 */


void Entity::init()
{
    /* Gerneral settings */
    RenderMode_     = video::WIREFRAME_SOLID;
    VertexCounter_  = 0;
    CurSurface_     = 0;
    Order_          = ORDER_NORMAL;
    NormalsType_    = NORMALS_SIMPLE;
    isCollUpdate_   = true;
}

void Entity::clear()
{
    /* Delete all vertex buffers */
    for (std::vector<SMeshSurface3D>::iterator itSurf = SurfaceList_.begin(); itSurf != SurfaceList_.end(); ++itSurf)
    {
        __spVideoDriver->deleteVertexBuffer(itSurf->ID);
    }
    
    /* Clear the surface list */
    SurfaceList_.clear();
    
    /* Delete the OctTreeRoot */
    spDeletePointer(OctTreeRoot_);
}

void Entity::calcSurfaceNoramls(u32 Surface)
{
    
    /* Temporary variables */
    std::vector<SNormalCoord> NormalsList;
    SNormalCoord NormalCoordData;
    
    /* Calculate normals */
    #if 1
    
    for (s32 i = 0; i < SurfaceList_[Surface].TriangleCount; ++i)
    {
        NormalCoordData.Normal = 
            math::getNormalVector(
                SurfaceList_[Surface].VerticesList[ SurfaceList_[Surface].TriangleIndicesList[i*3+0] ].Position,
                SurfaceList_[Surface].VerticesList[ SurfaceList_[Surface].TriangleIndicesList[i*3+1] ].Position,
                SurfaceList_[Surface].VerticesList[ SurfaceList_[Surface].TriangleIndicesList[i*3+2] ].Position
            );
        
        NormalCoordData.Index       = SurfaceList_[Surface].TriangleIndicesList[i*3+0];
        NormalCoordData.Position    = SurfaceList_[Surface].VerticesList[ NormalCoordData.Index ].Position;
        NormalsList.push_back(NormalCoordData);
        
        NormalCoordData.Index       = SurfaceList_[Surface].TriangleIndicesList[i*3+1];
        NormalCoordData.Position    = SurfaceList_[Surface].VerticesList[ NormalCoordData.Index ].Position;
        NormalsList.push_back(NormalCoordData);
        
        NormalCoordData.Index       = SurfaceList_[Surface].TriangleIndicesList[i*3+2];
        NormalCoordData.Position    = SurfaceList_[Surface].VerticesList[ NormalCoordData.Index ].Position;
        NormalsList.push_back(NormalCoordData);
    }
    
    #else
    
    dim::vector3df TempNormal;
    
    for (s32 i = 0; i < SurfaceList_[Surface].TriangleCount; ++i)
    {
        TempNormal =
            math::getNormalVector(
                SurfaceList_[Surface].VerticesList[ SurfaceList_[Surface].TriangleIndicesList[i*3+0] ].Position,
                SurfaceList_[Surface].VerticesList[ SurfaceList_[Surface].TriangleIndicesList[i*3+1] ].Position,
                SurfaceList_[Surface].VerticesList[ SurfaceList_[Surface].TriangleIndicesList[i*3+2] ].Position
            );
        SurfaceList_[Surface].VerticesList[ SurfaceList_[Surface].TriangleIndicesList[i*3+0] ].Normal = TempNormal;
        SurfaceList_[Surface].VerticesList[ SurfaceList_[Surface].TriangleIndicesList[i*3+1] ].Normal = TempNormal;
        SurfaceList_[Surface].VerticesList[ SurfaceList_[Surface].TriangleIndicesList[i*3+2] ].Normal = TempNormal;
    }
    
    NormalCoordData.Index = 0;
    for (std::vector<video::SMeshVertex3D>::iterator it = SurfaceList_[Surface].VerticesList.begin();
         it != SurfaceList_[Surface].VerticesList.end(); ++it, ++NormalCoordData.Index)
    {
        NormalCoordData.Position    = it->Position;
        NormalCoordData.Normal      = it->Normal;
        NormalsList.push_back(NormalCoordData);
    }
    
    #endif
    
    if (!NormalsList.size())
        return;
    
    /* Sort listed coordinations */
    sort(NormalsList.begin(), NormalsList.end(), cmpVertexCoords);
    
    /* Temporary variables */
    s32 curNormalCount = 0;
    
    dim::vector3df curCoord = NormalsList.begin()->Position;
    dim::vector3df curNormal;
    
    std::vector<SNormalCoord>::iterator it = NormalsList.begin(), start = NormalsList.begin(), tmpit;
    
    /* Find all normal groups */
    do
    {
        curNormal += it->Normal;
        ++curNormalCount;
        
        if (it == NormalsList.end() - 1 || !curCoord.isEqual((it+1)->Position))
        {
            curNormal /= curNormalCount;
            
            for (tmpit = start; tmpit <= it; ++tmpit)
                SurfaceList_[Surface].VerticesList[tmpit->Index].Normal = curNormal;
            
            ++it;
            
            curNormalCount  = 0;
            curNormal       = 0;
            curCoord        = it->Position;
            start           = it;
        }
        else
            ++it;
    }
    while (it != NormalsList.end());
    
}


} // /namespace scene

} // /namespace sp










// ================================================================================
