/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * lights
 *
 */

#include "spSceneManagerLight.hpp"
#include "spMainDevice.hpp"


namespace sp
{

extern video::VideoDriver* __spVideoDriver;

namespace scene
{


bool __spLightIDList[MAX_COUNT_OF_LIGHTS];

Light::Light() : Node(NODE_LIGHT), LightID_(0)
{
    /* General settings */
    AmbientColor_   = SpecularColor_    = video::color();
    LightingType_   = LIGHT_DIRECTIONAL;
    NearSpot_       = LIGHTDEF_RANGE_NEAR;
    FarSpot_        = LIGHTDEF_RANGE_FAR;
    isVolumetric_   = false;
    Direction_      = dim::vector3df(0, 0, 1);
    
    /* Volumetric light */
    AttenuationConstant_    = 1.0f;
    AttenuationLinear_      = 0.1f;
    AttenuationQuadratic_   = 0.4f;
}

Light::Light(ELightTypes Type) : Node(NODE_LIGHT), LightID_(0)
{
    /* General settings */
    LightingType_   = Type;
    
    /* Spot light settings */
    NearSpot_       = LIGHTDEF_RANGE_NEAR;
    FarSpot_        = LIGHTDEF_RANGE_FAR;
    Direction_      = dim::vector3df(0, 0, 1);
    
    /* Color settings */
    Color_          = 200;
    AmbientColor_   = 255;
    SpecularColor_  = 0;
    
    /* Volumetric light settings */
    isVolumetric_           = false;
    AttenuationConstant_    = 1.0f;
    AttenuationLinear_      = 0.1f;
    AttenuationQuadratic_   = 0.4f;
    
    u32 i;
    
    /* Check if the there is still enough space for an additional light */
    for (i = 0; i < MAX_COUNT_OF_LIGHTS; ++i)
    {
        if (!__spLightIDList[i])
        {
            __spLightIDList[i]  = true;
            LightID_            = i;
            break;
        }
    }
    
    /* Check if the light has an identity */
    if (i < MAX_COUNT_OF_LIGHTS)
    {
        /* Add a dynamic light soruce */
        __spVideoDriver->addDynamicLightSource(
            LightID_, LightingType_,
            Color_, AmbientColor_, SpecularColor_,
            AttenuationConstant_, AttenuationLinear_, AttenuationQuadratic_
        );
    }
    else
    {
        io::printWarning(
            io::stringc("to many lights in use, only ")+io::stringc(MAX_COUNT_OF_LIGHTS)+io::stringc(" lights available")
        );
    }
}

Light::~Light()
{
    /* Enable the light */
    __spVideoDriver->setLightStatus(LightID_, false);
    
    /* Remove the light from the list */
    __spLightIDList[LightID_] = false;
}

/* Lighting type (diffuse, spot) */

void Light::setVisible(bool isVisible)
{
    /* Set the light stauts */
    __spVideoDriver->setLightStatus(LightID_, isVisible_ = isVisible);
}

/* Lighting color */

void Light::setLightingColor(video::color Color, video::color AmbientColor, video::color SpecularColor)
{
    /* General settings */
    Color_          = Color;
    AmbientColor_   = AmbientColor;
    SpecularColor_  = SpecularColor;
    
    /* Update lighting colors */
    __spVideoDriver->setLightColor(LightID_, Color_, AmbientColor_, SpecularColor_);
}
void Light::getLightingColor(video::color &Color, video::color &AmbientColor, video::color &SpecularColor)
{
    Color           = Color_;
    AmbientColor    = AmbientColor_;
    SpecularColor   = SpecularColor_;
}

/* Spot range */

void Light::setSpotRange(f32 Near, f32 Far)
{
    NearSpot_   = Near;
    FarSpot_    = Far;
}

void Light::getSpotRange(f32 &Near, f32 &Far)
{
    Near    = NearSpot_;
    Far     = FarSpot_;
}

/* Volumetric */

void Light::setVolumetric(bool isVolumetric)
{
    if (!isVolumetric)
    {
        f32 tmp1 = 1.0f, tmp2 = 0.0f;
        
        /* Update the renderer for the light */
        __spVideoDriver->updateLight(
            LightID_, LightingType_, isVolumetric_, Direction_, NearSpot_, FarSpot_, tmp1, tmp2, tmp2
        );
    }
    
    isVolumetric_ = isVolumetric;
}

void Light::setVolumetricRadius(f32 Radius)
{
    AttenuationConstant_    = 1.0;
    AttenuationLinear_      = 1.0 / Radius;
    AttenuationQuadratic_   = 1.0 / Radius;
}
f32 Light::getVolumetricRadius() const
{
    return 1.0 / AttenuationLinear_;
}

void Light::setVolumetricRange(f32 Constant, f32 Linear, f32 Quadratic)
{
    AttenuationConstant_    = Constant;
    AttenuationLinear_      = Linear;
    AttenuationQuadratic_   = Quadratic;
}
void Light::getVolumetricRange(f32 &Constant, f32 &Linear, f32 &Quadratic) const
{
    Constant    = AttenuationConstant_;
    Linear      = AttenuationLinear_;
    Quadratic   = AttenuationQuadratic_;
}


/*
 * ======= Updateing =======
 */

void Light::update()
{
    
    /* Matrix transformation */
    updateGlobalLocation();
    
    /* Update the render matrix */
    __spVideoDriver->updateModelviewMatrix();
    
    /* Update the renderer for the light */
    __spVideoDriver->updateLight(
        LightID_, LightingType_, isVolumetric_, Direction_, NearSpot_, FarSpot_,
        AttenuationConstant_, AttenuationLinear_, AttenuationQuadratic_
    );
    
}


} // /namespace scene

} // /namespace sp







// ================================================================================
 
