/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * nodes
 *
 */

#include "spSceneManagerNode.hpp"
#include "spMainDevice.hpp"


namespace sp
{

extern video::VideoDriver* __spVideoDriver;

namespace scene
{


/*
 * ======= Internal matrices =======
 */

dim::matrix4f spProjectionMatrix;
dim::matrix4f spViewMatrix;
dim::matrix4f spWorldMatrix;
dim::matrix4f spTextureMatrix[video::MAX_COUNT_OF_TEXTURES];
dim::matrix4f spColorMatrix;


/*
 * ======= Constructor & destructor =======
 */

Node::Node(ENodeTypes Type) : ParentNode_(0), UserData_(0)
{
    Type_           = Type;
    Name_           = "";
    Position_       = 0.f;
    Rotation_       = 0.f;
    Scale_          = 1.f;
    Color_          = 255;
    isVisible_      = true;
    isAutoMatrix_   = true;
    ParentNode_     = 0;
    Matrix_.matrixReset();
}
Node::~Node()
{
}

void Node::setPosition(dim::vector3df Position, bool isGlobal)
{
    if ( ( isGlobal && !ParentNode_) || !isGlobal )
        Position_ = Position;
    else if (isGlobal && ParentNode_)
        Position_ = Position - ParentNode_->getPosition(true); // !!!
}
dim::vector3df Node::getPosition(bool isGlobal) const
{
    if (isGlobal)
        return getGlobalLocation().getPosition();
    return Position_;
}

void Node::setRotation(dim::vector3df Rotation, bool isGlobal)
{
    if ( ( isGlobal && !ParentNode_ ) || !isGlobal )
        Rotation_ = Rotation;
    else if (isGlobal && ParentNode_)
        Rotation_ = Rotation - ParentNode_->getRotation(true); // !!!
}
dim::vector3df Node::getRotation(bool isGlobal) const
{
    if (isGlobal)
        return getGlobalLocation().getRotation();
    return Rotation_;
}

void Node::setScale(dim::vector3df Scale, bool isGlobal)
{
    if ( ( isGlobal && !ParentNode_ ) || !isGlobal )
        Scale_ = Scale;
    else if (isGlobal && ParentNode_)
        Scale_ = Scale - ParentNode_->getScale(true); // !!!
}
dim::vector3df Node::getScale(bool isGlobal) const
{
    if (isGlobal)
        return getGlobalLocation().getScale();
    return Scale_;
}

void Node::doMove(dim::vector3df Direction)
{
    dim::matrix4f Mat;
    Mat.matrixReset();
    Mat.matrixRotateYXZ(Rotation_);
    Position_ += Mat * Direction;
}

void Node::lookAt(dim::vector3df Position, bool isGlobal)
{
    /* Gettings */
    dim::vector3df Pos(getPosition(isGlobal));
    dim::vector3df Rot(getRotation(isGlobal));
    
    /* Calculate rotation */
    Rot.X =   ASIN( (Position.Y - Pos.Y) / math::getDistance(Pos, Position) );
    Rot.Y = - ASIN( (Position.X - Pos.X) / math::getDistance(Pos.X, Pos.Z, Position.X, Position.Z) );
    Rot.Z = 0.0f;
    if (Pos.Z < Position.Z)
        Rot.Y = 180.0f - Rot.Y;
    
    /* Settings */
    setRotation(Rot, isGlobal);
}

void Node::updateParent() const
{
    if (ParentNode_)
    {
        /* Update the parent of this node */
        ParentNode_->updateParent();
        
        /* Check what matrix is in use */
        if (ParentNode_->getAutoMatrix())
        {
            /* Matrix transformation */
            spWorldMatrix.matrixTranslate(ParentNode_->Position_);
            spWorldMatrix.matrixRotateYXZ(ParentNode_->Rotation_);
            spWorldMatrix.matrixScale(ParentNode_->Scale_);
        }
        else
            ParentNode_->updateMatrix();
    }
}

dim::matrix4f Node::getGlobalLocation() const
{
    /* Work with an identified matrix */
    spWorldMatrix.matrixReset();
    
    /* Compute the global location */
    updateGlobalLocation();
    
    /* Return the calculated matrix and exit the function */
    return spWorldMatrix;
}

dim::matrix4f Node::getGlobalLocationInverse() const
{
    /* Work with an identified matrix */
    spWorldMatrix.matrixReset();
    
    /* Compute the global inverse location */
    updateGlobalLocationInverse();
    
    /* Return the calculated matrix and exit the function */
    return spWorldMatrix;
}

void Node::updateGlobalLocation() const
{
    /* Parent location */
    updateParent();
    
    /* Check what matrix is in use */
    if (getAutoMatrix())
    {
        /* Matrix transformation */
        spWorldMatrix.matrixTranslate(Position_);
        spWorldMatrix.matrixRotateYXZ(Rotation_);
        spWorldMatrix.matrixScale(Scale_);
    }
    else
        updateMatrix();
}

void Node::updateGlobalLocationInverse() const
{
    /* Compute the global location */
    updateGlobalLocation();
    
    /* Inverse the matrix */
    spWorldMatrix.setInverse();
}    

void Node::updateMatrix() const
{
    spWorldMatrix *= Matrix_;
}

dim::vector3df Node::getDirection(dim::vector3df upVector) const
{
    dim::matrix4f Matrix;
    
    Matrix.matrixReset();
    Matrix.matrixRotateY(Rotation_.Y);
    Matrix.matrixRotateX(Rotation_.X);
    Matrix.matrixRotateZ(Rotation_.Z);
    
    return Matrix * upVector;
}

void Node::copyAttributes(Node* OtherNode)
{
    Name_           = OtherNode->Name_;
    
    Matrix_         = OtherNode->Matrix_;
    
    Position_       = OtherNode->Position_;
    Rotation_       = OtherNode->Rotation_;
    Scale_          = OtherNode->Scale_;
    
    Color_          = OtherNode->Color_;
    isVisible_      = OtherNode->isVisible_;
    isAutoMatrix_   = OtherNode->isAutoMatrix_;
    
    ParentNode_     = OtherNode->ParentNode_;
}

void Node::updateModelviewMatrix()
{
    
}


/*
 * ======= Protected functions =======
 */

void Node::setBlending(s32 Mode, s32 &Blend1, s32 &Blend2)
{
    switch (Mode)
    {
        case BLENDING_NORMAL:
            Blend1 = video::BLEND_SRCALPHA;
            Blend2 = video::BLEND_INVSRCALPHA;
            break;
        case BLENDING_BRIGHTNESS:
            Blend1 = video::BLEND_SRCALPHA;
            Blend2 = video::BLEND_ONE;
            break;
        case BLENDING_DARKNESS:
            Blend1 = video::BLEND_DESTCOLOR;
            Blend2 = video::BLEND_ZERO;
            break;
    }
}


} // /namespace scene

} // /namespace sp







// ================================================================================
