/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * SpecialEffectsActor (for special effects implementation)
 *
 */

#include "spSceneManagerSpecialEffectsActor.hpp"
#include "spMainDevice.hpp"


namespace sp
{

extern scene::SceneManager* __spSceneManager;
extern video::VideoDriver* __spVideoDriver;

namespace scene
{
namespace effect
{


SpecialEffectsActor::SpecialEffectsActor() : 
    CSLineWidth_(__SP_CS_STDLINESIZE), CSBorderColor_(video::color(0)), BEBlurTexture_(0)
{
    /* General settings */
    setBlurTextureSize(__SP_BE_STDTEXSIZE);
}
SpecialEffectsActor::~SpecialEffectsActor()
{
    /* Delete the blur texture */
    __spVideoDriver->deleteTexture(BEBlurTexture_);
}

void SpecialEffectsActor::registerCamera(Camera* hCamera)
{
    for (std::list<Camera*>::iterator it = RegisteredCameras_.begin(); it != RegisteredCameras_.end(); ++it)
    {
        if (*it == hCamera)
        {
            io::printWarning("Warning: multiple camera registration in 'SpecialEffectsActor'!");
            return;
        }
    }
    RegisteredCameras_.push_back(hCamera);
}
void SpecialEffectsActor::unregisterCamera(Camera* hCamera)
{
    for (std::list<Camera*>::iterator it = RegisteredCameras_.begin(); it != RegisteredCameras_.end(); ++it)
    {
        if (*it == hCamera)
        {
            RegisteredCameras_.erase(it);
            return;
        }
    }
}


/*
 * ========== Bump mapping ==========
 */


/* ======= Private: ======= */

void SpecialEffectsActor::setUpBumps(
    f32* VertexCoord, f32* Normal, f32* LightPosition,
    f32* TexCoordVectorS, f32* TexCoordVectorT, f32 BumpHeight)
{
    f32 CurCoord[3]; // Vertex From Current Position To Light
    f32 LenSq; // Used To Normalize
    
    // Calculate v From Current Vector c To Lightposition And Normalize v
    CurCoord[0] = LightPosition[0] - Normal[0];
    CurCoord[1] = LightPosition[1] - Normal[1];
    CurCoord[2] = LightPosition[2] - Normal[2];
    
    LenSq = static_cast<f32>( sqrt( CurCoord[0]*CurCoord[0] + CurCoord[1]*CurCoord[1] + CurCoord[2]*CurCoord[2] ) );
    
    CurCoord[0] /= LenSq;
    CurCoord[1] /= LenSq;
    CurCoord[2] /= LenSq;
    
    // Project v Such That We Get Two Values Along Each Texture-Coordinat Axis.
    Normal[0] = ( TexCoordVectorS[0]*CurCoord[0] + TexCoordVectorS[1]*CurCoord[1] + TexCoordVectorS[2]*CurCoord[2] ) * __SP_BM_MAX_EMBOSS * BumpHeight;
    Normal[1] = ( TexCoordVectorT[0]*CurCoord[0] + TexCoordVectorT[1]*CurCoord[1] + TexCoordVectorT[2]*CurCoord[2] ) * __SP_BM_MAX_EMBOSS * BumpHeight;	
}


/*
 * ========== Fur effect ==========
 */

void SpecialEffectsActor::createFurObject(
    Entity* hEntity, video::Texture* FurTexture,
    s32 LayerCount, f32 HairLength, s32 HairCloseness)
{
    
    /* Check if the scene manager has allready created */
    
    if (!__spSceneManager)
        return;
    
    /* ===== Create the hair texture (for layers) ===== */
    
    /* Copy the fur texture */
    video::Texture* HairTexture = __spVideoDriver->copyTexture(FurTexture);
    
    /* Change the type for the alpha-channel */
    HairTexture->setType(video::TEXTURE_RGBA);
    
    /* Loop for the image data */
    for (s32 y = 0, x; y < HairTexture->getSize().Height; ++y)
    {
        for (x = 0; x < HairTexture->getSize().Width; ++x)
        {
            if (math::Rnd(HairCloseness) != 1)
            {
                /* Change the pixel transparency */
                video::color Color( HairTexture->getPixelColor(dim::point2di(x, y)) );
                Color.Alpha = 0;
                HairTexture->setPixelColor(dim::point2di(x, y), Color);
            }
        }
    }
    
    /* Update the image data */
    HairTexture->updateImageData(video::SET_IMAGEDATA);
    
    /* ===== Create the hair layers ===== */
    
    /* Allocate a layer list */
    Entity** layer = new Entity*[LayerCount];
    
    /* Temporary variables */
    dim::vector3df Coord, Normal;
    
    /* Loop for all layers */
    for (s32 i = 0, j; i < LayerCount; ++i)
    {
        layer[i] = __spSceneManager->copyNode(hEntity);
        layer[i]->setParent(hEntity);
        layer[i]->setPosition(0);
        layer[i]->setScale(1);
        
        layer[i]->setColorMaterial(false);
        layer[i]->getThisMaterial().ColorDiffuse.Alpha = 255 - 255 * (i+1) / LayerCount;
        
        for (j = 0; j < layer[i]->getMeshVerticesCount(); ++j)
        {
            Coord   = hEntity->getVertexCoord(j);
            Normal  = hEntity->getVertexNormal(j);
            
            Coord += Normal * (HairLength * (f32)(i+1) / LayerCount);
            
            layer[i]->setVertexCoord(j, Coord);
            layer[i]->setTexture(0, HairTexture);
        }
        
        layer[i]->updateModeling(false);
    } // next layer
    
    /* Delete the layer list */
    spDeleteArray(layer);
    
}


/*
 * ========== Cel shading ==========
 */

// CelShadow struct

CelShadow::CelShadow()
{
    ID_ = 0;
}
CelShadow::CelShadow(u32 ID)
{
    ID_ = ID;
}
CelShadow::~CelShadow()
{
}

CelShadow* SpecialEffectsActor::createCelShadow(s32 LevelsCount, video::color ShadowColor, video::color DarknessColor) { }
CelShadow* SpecialEffectsActor::createCelShadow(s32 LevelsCount, video::color* LevelsColor) { }

void SpecialEffectsActor::deleteCelShadow(CelShadow* &hCelShadow) { }


/*
 * ========== Blur effect ==========
 */

void SpecialEffectsActor::setBlurTextureSize(s32 TextureSize)
{
    __spVideoDriver->deleteTexture(BEBlurTexture_);
    BEBlurTexture_ = __spVideoDriver->createTexture(dim::size2di(TextureSize));
}

/*
 * ========== Render functions ==========
 */

void SpecialEffectsActor::renderCelShading(Entity* hEntity, CelShadow* hCelShadow, Light* hLight) { }
        
void SpecialEffectsActor::renderBlurEffect(Entity* hEntity, s32 LevelsCount, f32 LevelsDist, u8 ColorAlpha) { }


} // /namespace effect

} // /namespace scene

} // /namespace sp







// ================================================================================
