/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * Header for all sound & music operations (sound, music, video-to-texture)
 *
 */

#include "spSensitiveSound.hpp"

#if defined(WIN32)
#   include <windows.h>
#   include <mmsystem.h>
#endif


namespace sp
{
namespace sound
{


#define __SP_MCI_STRLEN             256
#define __SP_MUSIC_VOLUME_HEIGHT    1000.0f
#define __SP_MUSIC_SPEED_HEIGHT     1000.0f
#define __SP_MUSIC_SEEK_HEIGHT      1000.0f


#if defined(WIN32)

SensitiveSoundDevice* __ssDevice = 0;

SP_EXPORT SensitiveSoundDevice* createListener(f32 NearHearing, f32 FarHearing, f32 NearVolume, f32 FarVolume)
{
    __ssDevice = new SensitiveSoundDevice(NearHearing, FarHearing, NearVolume, FarVolume);
    
    if (__ssDevice)
    {
        io::printMessage("SensitiveSound system - v.1.2");
        io::printMessage("Copyright (c) 2008 - Lukas Hermanns");
        io::printMessage("Using Microsoft(c) Windows(c) multimedia system");
        io::printMessage("");
    }
    else
        io::printError("could not create 'SensitiveSound' system");
    
    return __ssDevice;
}

SensitiveSoundDevice::SensitiveSoundDevice(f32 NearHearing, f32 FarHearing, f32 NearVolume, f32 FarVolume)
{
    /* General settings */
    NearHearing_    = math::Pow2(NearHearing);
    FarHearing_     = math::Pow2(FarHearing);
    NearVolume_     = NearVolume;
    FarVolume_      = FarVolume;
    
    /* Middle values */
    VolumeMiddle_   = NearVolume_ - FarVolume_;
    HearingMiddle_  = FarHearing_ - NearHearing_;
}
SensitiveSoundDevice::~SensitiveSoundDevice()
{
}

void SensitiveSoundDevice::changeAttributes(f32 NearHearing, f32 FarHearing, f32 NearVolume, f32 FarVolume)
{
    /* General settings */
    NearHearing_    = math::Pow2(NearHearing);
    FarHearing_     = math::Pow2(FarHearing);
    NearVolume_     = NearVolume;
    FarVolume_      = FarVolume;
    
    /* Middle values */
    VolumeMiddle_   = NearVolume_-FarVolume_;
    HearingMiddle_  = FarHearing_-NearHearing_;
}

/*
 * ========== Sounds: ==========
 */

Sound* SensitiveSoundDevice::loadSound(io::stringc Filename, u8 Quality, u8 Count)
{
    /* Information message */
    io::printMessage(io::stringc("Load sound: \"")+Filename+io::stringc("\""));
    
    /* Check if the file does exist */
    if (!io::FileLoader().doesFileExist(Filename))
    {
        /* Error message */
        io::printError("could not found sound file");
        
        /* Return nothing */
        return 0;
    }
    
    /* Allocate a new sound */
    Sound* NewSound = new Sound(Filename, Quality, Count);
    
    /* Add the sound to the list */
    SoundList_.push_back(NewSound);
    
    io::stringc QualityName;
    
    /* Select sound quality */
    switch (Quality)
    {
        case SOUNDQUALITY_LOW:
            QualityName = "LOW"; break;
        case SOUNDQUALITY_MEDIUM:
            QualityName = "MEDIUM"; break;
        case SOUNDQUALITY_HIGH:
            QualityName = "HIGH"; break;
    }
    
    /* Load the sound */
    for (u32 i = 0, id = NewSound->getID(); i < Count; ++i, ++id)
    {
        SndMciMsg(
            io::stringc("OPEN \"") + Filename + io::stringc("\" TYPE MpegVideo ALIAS MMF") + io::stringc(id)
        );
        
        if (Quality != SOUNDQUALITY_IGNORE)
        {
            SndMciMsg(
                io::stringc("SETAUDIO MMF") + io::stringc(id) + io::stringc(" QUALITY ") + QualityName
            );
        }
    }
    
    /* Return the new sound */
    return NewSound;
}

void SensitiveSoundDevice::closeSound(Sound* hSound)
{
    for (std::list<Sound*>::iterator it = SoundList_.begin(); it != SoundList_.end(); ++it)
    {
        if ((*it)->getID() == hSound->getID())
        {
            /* Close the sound */
            for (u32 i = 0; i < hSound->getCount(); ++i)
                SndMciMsg(io::stringc("CLOSE MMF")+io::stringc(hSound->getID()+i));
            
            /* Remove the sound of the list */
            SoundList_.erase(it);
            
            /* Delete the sound */
            spDeletePointer(hSound);
            
            /* Exit the loop */
            break;
        }
    }
}

void SensitiveSoundDevice::playSound(Sound* hSound, bool isLooped, bool isVolumetric)
{
    hSound->isLooped_       = isLooped;
    hSound->isVolumetric_   = isVolumetric;
    hSound->addIDCount();
    
    SndMciMsg(io::stringc("PLAY MMF")+io::stringc(hSound->getCurID())+io::stringc(" from 0"));
}

void SensitiveSoundDevice::stopSound(Sound* hSound)
{
    if (hSound == DEF_SOUND_ALL)
    {
        for (std::list<Sound*>::iterator it = SoundList_.begin(); it != SoundList_.end(); ++it)
        {
            (*it)->isLooped_ = false;
            for (u32 i = 0; i < hSound->getCount(); ++i)
                SndMciMsg(io::stringc("STOP MMF")+io::stringc((*it)->getID()+i));
        }
    }
    else
    {
        hSound->isLooped_ = false;
        for (u32 i = 0; i < hSound->getCount(); ++i)
            SndMciMsg(io::stringc("STOP MMF")+io::stringc(hSound->getID()+i));
    }
}

void SensitiveSoundDevice::pauseSound(Sound* hSound, bool isPaused)
{
    if (hSound == DEF_SOUND_ALL)
    {
        if (isPaused)
        {
            for (std::list<Sound*>::iterator it = SoundList_.begin(); it != SoundList_.end(); ++it)
            {
                for (u32 i = 0; i < hSound->getCount(); ++i)
                    SndMciMsg(io::stringc("PAUSE MMF")+io::stringc((*it)->getID()+i));
                (*it)->isPaused_ = true;
            }
        }
        else
        {
            for (std::list<Sound*>::iterator it = SoundList_.begin(); it != SoundList_.end(); ++it)
            {
                if ((*it)->isPaused_)
                {
                    for (u32 i = 0; i < hSound->getCount(); ++i)
                        SndMciMsg(io::stringc("RESUME MMF")+io::stringc((*it)->getID()+i));
                    (*it)->isPaused_ = false;
                }
            }
        }
    }
    else
    {
        if (hSound->isPaused_ = isPaused)
        {
            for (u32 i = 0; i < hSound->getCount(); ++i)
                SndMciMsg(io::stringc("PAUSE MMF")+io::stringc(hSound->getID()+i));
        }
        else
        {
            for (u32 i = 0; i < hSound->getCount(); ++i)
                SndMciMsg(io::stringc("RESUME MMF")+io::stringc(hSound->getID()+i));
        }
    }
}

bool SensitiveSoundDevice::isSound(Sound* hSound)
{
    for (std::list<Sound*>::iterator it = SoundList_.begin(); it != SoundList_.end(); ++it)
    {
        if ((*it)->getID() == hSound->getID())
            return true;
    }
    return false;
}

/* ===== Volume ===== */

void SensitiveSoundDevice::setSoundVolume(Sound* hSound, f32 Volume, bool isOnlyCurrent)
{
    if (!isOnlyCurrent)
    {
        for (u32 i = 0; i < hSound->getCount(); ++i)
        {
            SndMciMsg(
                io::stringc("SETAUDIO MMF") + io::stringc(hSound->getID()+i) + io::stringc(" VOLUME to ") + io::stringc(s32(Volume*__SP_MUSIC_VOLUME_HEIGHT))
            );
        }
    }
    else
    {
        SndMciMsg(
            io::stringc("SETAUDIO MMF") + io::stringc(hSound->getCurID()) + io::stringc(" VOLUME to ") + io::stringc(s32(Volume*__SP_MUSIC_VOLUME_HEIGHT))
        );
    }
}
f32 SensitiveSoundDevice::getSoundVolume(Sound* hSound)
{
    return io::stringc(SndMciMsg(io::stringc("STATUS MMF")+io::stringc(hSound->getID())+io::stringc(" VOLUME"))).val<f32>() / __SP_MUSIC_VOLUME_HEIGHT;
}

/* ===== Seek (status) ===== */

void SensitiveSoundDevice::setSoundSeek(Sound* hSound, f32 SeekPosition, bool isOnlyCurrent)
{
    if (!isOnlyCurrent)
    {
        for (u32 i = 0; i < hSound->getCount(); ++i)
        {
            SndMciMsg(
                io::stringc("PLAY MMF") + io::stringc(hSound->getID()+i) + io::stringc(" from ") + io::stringc(getSoundLength(hSound)*SeekPosition).val<s32>()
            );
        }
    }
    else
    {
        SndMciMsg(
            io::stringc("PLAY MMF") + io::stringc(hSound->getCurID()) + io::stringc(" from ") + io::stringc(getSoundLength(hSound)*SeekPosition).val<s32>()
        );
    }
}
f32 SensitiveSoundDevice::getSoundSeek(Sound* hSound)
{
    return io::stringc(
        SndMciMsg(io::stringc("STATUS MMF")+io::stringc(hSound->getCurID())+io::stringc(" POSITION"))).val<f32>() / getSoundLength(hSound
    );
}

/* ===== Speed (frequency) ===== */

void SensitiveSoundDevice::setSoundSpeed(Sound* hSound, f32 Speed, bool isOnlyCurrent)
{
    if (!isOnlyCurrent)
    {
        for (u32 i = 0; i < hSound->getCount(); ++i)
        {
            SndMciMsg(
                io::stringc("SET MMF")+io::stringc(hSound->getID()+i)+io::stringc(" SPEED ")+io::stringc((Speed+1.0f)*__SP_MUSIC_SPEED_HEIGHT)
            );
        }
    }
    else
    {
        SndMciMsg(
            io::stringc("SET MMF")+io::stringc(hSound->getCurID())+io::stringc(" SPEED ")+io::stringc((Speed+1.0f)*__SP_MUSIC_SPEED_HEIGHT)
        );
    }
}
f32 SensitiveSoundDevice::getSoundSpeed(Sound* hSound)
{
    return io::stringc(
        SndMciMsg(io::stringc("STATUS MMF")+io::stringc(hSound->getID())+io::stringc(" SPEED"))
    ).val<f32>() / __SP_MUSIC_SPEED_HEIGHT - 1.0f;
}

/* ===== Pan (balance) ===== */

void SensitiveSoundDevice::setSoundPan(Sound* hSound, f32 Balance, f32 Volume, bool isOnlyCurrent)
{
    if (!isOnlyCurrent)
    {
        for (u32 i = 0; i < hSound->getCount(); ++i)
        {
            SndMciMsg(
                io::stringc("SETAUDIO MMF")+io::stringc(hSound->getID()+i)+io::stringc(" LEFT VOLUME to ")+
                io::stringc(s32(math::MinMax<f32>(-Balance+1.0f, 0.0f, 1.0f)*__SP_MUSIC_VOLUME_HEIGHT*Volume))
            );
            SndMciMsg(
                io::stringc("SETAUDIO MMF")+io::stringc(hSound->getID()+i)+io::stringc(" RIGHT VOLUME to ")+
                io::stringc(s32(math::MinMax<f32>(Balance+1.0f, 0.0f, 1.0f)*__SP_MUSIC_VOLUME_HEIGHT*Volume))
            );
        }
    }
    else
    {
        SndMciMsg(
            io::stringc("SETAUDIO MMF")+io::stringc(hSound->getCurID())+io::stringc(" LEFT VOLUME to ")+
            io::stringc(s32(math::MinMax<f32>(-Balance+1.0f, 0.0f, 1.0f)*__SP_MUSIC_VOLUME_HEIGHT*Volume))
        );
        SndMciMsg(
            io::stringc("SETAUDIO MMF")+io::stringc(hSound->getCurID())+io::stringc(" RIGHT VOLUME to ")+
            io::stringc(s32(math::MinMax<f32>(Balance+1.0f, 0.0f, 1.0f)*__SP_MUSIC_VOLUME_HEIGHT*Volume))
        );
    }
}
f32 SensitiveSoundDevice::getSoundPan(Sound* hSound)
{
    const f32 VolumeLeft = io::stringc(
        SndMciMsg(io::stringc("STATUS MMF")+io::stringc(hSound->getID())+io::stringc(" LEFT VOLUME"))
    ).val<f32>() / __SP_MUSIC_VOLUME_HEIGHT;
    
    const f32 VolumeRight = io::stringc(
        SndMciMsg(io::stringc("STATUS MMF")+io::stringc(hSound->getID())+io::stringc(" RIGHT VOLUME"))
    ).val<f32>() / __SP_MUSIC_VOLUME_HEIGHT;
    
    return VolumeRight - VolumeLeft;
}

/* ===== State ===== */

bool SensitiveSoundDevice::isSoundFinished(Sound* hSound)
{
    return getSoundSeek(hSound) >= 1.0f;
}
f32 SensitiveSoundDevice::getSoundLength(Sound* hSound)
{
    return io::stringc(SndMciMsg(io::stringc("STATUS MMF")+io::stringc(hSound->getID())+io::stringc(" LENGTH"))).val<f32>();
}


/*
 * ========== Recordings: ==========
 */

Record* SensitiveSoundDevice::createRecord()
{
    return new Record();
}

void SensitiveSoundDevice::closeRecord(Record* hRecord)
{
    if (hRecord)
    {
        /* Check if the record is still running */
        if (hRecord->isRecording())
            hRecord->endRecording();
        
        /* Close the record */
        SndMciMsg(io::stringc("CLOSE RECORD")+io::stringc(hRecord->getID()));
        
        /* Delete the record */
        spDeletePointer(hRecord);
    }
}

/*
 * ========== Sensitive sound system: ==========
 */

void SensitiveSoundDevice::deleteDevice()
{
    /* Delete each sound */
    for (std::list<Sound*>::iterator itSnd = SoundList_.begin(); itSnd != SoundList_.end(); ++itSnd)
    {
        for (u32 i = 0; i < (*itSnd)->getCount(); ++i)
            SndMciMsg(io::stringc("CLOSE MMF")+io::stringc((*itSnd)->getID()+i));
        spDeletePointer(*itSnd);
    }
    
    /* Self destruction: terminate your self */
    delete this;
}

void SensitiveSoundDevice::updateSounds3D(dim::vector3df CamPos, f32 CamYaw)
{
    
    /* Temporary variables */
    f32 Distance, Volume;
    
    /* Loop for all sounds */
    for (std::list<Sound*>::iterator itSnd = SoundList_.begin(); itSnd != SoundList_.end(); ++itSnd)
    {
        if ((*itSnd)->isVolumetric_)
        {
            /* Calculate the distance (for correct volume) */
            Distance = math::getDistanceSq(CamPos, (*itSnd)->getPosition());
            
            /* Compare the distance with the near & far volumes */
            if (Distance <= NearHearing_)
                Volume = NearVolume_;
            else if (Distance <= FarHearing_)
                Volume = FarVolume_ + ( 1.0f - math::Abs<f32>(Distance - NearHearing_) / HearingMiddle_ ) * VolumeMiddle_;
            else
                Volume = FarVolume_;
            
            /* Calculate the pan (for 3d sound effect) */
            f32 PanDegree = 180.0 - math::getDegree(
                CamPos.X, CamPos.Z, (*itSnd)->getPosition().X, (*itSnd)->getPosition().Z
            );
            
            /* Set the new volume & balance values */
            setSoundVolume(*itSnd, Volume);
            setSoundPan(*itSnd, SIN(PanDegree - CamYaw), Volume);
        }
    }
    
}

void SensitiveSoundDevice::updateSounds()
{
    for (std::list<Sound*>::iterator itSnd = SoundList_.begin(); itSnd != SoundList_.end(); ++itSnd)
    {
        if ((*itSnd)->isLooped_ && getSoundSeek(*itSnd) >= 1.0f)
        {
            (*itSnd)->addIDCount();
            SndMciMsg(io::stringc("PLAY MMF")+io::stringc((*itSnd)->getCurID())+io::stringc(" from 0"));
        }
    }
}

/*
 * ========== Private: ==========
 */

std::string SensitiveSoundDevice::SndMciMsg(io::stringc Message)
{
    char Result[__SP_MCI_STRLEN];
    mciSendString(Message.c_str(), Result, __SP_MCI_STRLEN, 0);
    return std::string(Result);
}

#elif defined(LINUX)

/* The whole "SensitiveSound System" should be complete new for linux version */

#endif



} // /namespace sound

} // /namespace sp








// ================================================================================
