/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * "Ray" Software video driver
 *
 */

#include "spSoftwareDriver.hpp"
#include "spSceneManagerSpecialEffectsActor.hpp"
#include "spSceneManagerCamera.hpp"
#include "spMainDevice.hpp"
#include "spInternalDeclerations.hpp"


namespace sp
{

extern video::VideoDriver* __spVideoDriver;

namespace video
{


/*
 * ========== Internal members ==========
 */

const s32 RAYMappingGenList[] = {
    RAY_OBJECT_LINEAR, RAY_OBJECT_LINEAR, RAY_EYE_LINEAR, RAY_SPHERE_MAP, RAY_OBJECT_LINEAR, RAY_OBJECT_LINEAR
};


/*
 * ========== Constructors & destructor ==========
 */


SoftwareDriver::SoftwareDriver() : VideoDriver(DRIVER_SOFTWARE)
{
}
SoftwareDriver::~SoftwareDriver()
{
}


/*
 * ========== Video driver identity ==========
 */

io::stringc SoftwareDriver::getRenderer() const
{
    return io::stringc("SoftwareRenderer");
}
io::stringc SoftwareDriver::getVersion() const
{
    return io::stringc(rayGetString(RAY_RENDERER)) + io::stringc(" ") + io::stringc(rayGetString(RAY_VERSION));
}
io::stringc SoftwareDriver::getVendor() const
{
    return io::stringc(rayGetString(RAY_VENDOR));
}

bool SoftwareDriver::supportAntiAlias() const
{
    return false;
}

bool SoftwareDriver::supportMultitex() const
{
    return getMultitexCount() > 1;
}

bool SoftwareDriver::supportVBO() const
{
    return true;
}

bool SoftwareDriver::supportAnisotropicFiltering() const
{
    return false;
}

s32 SoftwareDriver::getMultitexCount() const
{
    return 2;
}

s32 SoftwareDriver::getMaxAnisotropicFilter() const
{
    return 0;
}


/*
 * ========== User controll functions (clear buffers, flip buffers, 2d drawing etc.) ==========
 */

void SoftwareDriver::clearBuffers()
{
    rayClear(RAY_CLEAR_COLOR | RAY_CLEAR_DEPTH);
}

void SoftwareDriver::flipBuffers()
{
    rayFlipBuffers();
    
    #if defined(WIN32)
    SwapBuffers(hDC_);
    #elif defined(LINUX)
    
    #endif
}


/*
 * ========== Setting-/ getting functions ==========
 */

void SoftwareDriver::setShadeMode(EShadeModeTypes ShadeMode)
{
    switch (ShadeMode)
    {
        case SHADEMODE_SMOOTH:
            rayEnable(RAY_COLOR_SMOOTH);
            break;
        case SHADEMODE_FLAT:
            rayDisable(RAY_COLOR_SMOOTH);
            break;
    }
}

void SoftwareDriver::setClearColor(color Color)
{
    rayClearColor(Color.Red, Color.Green, Color.Blue);
}


/*
 * ======= Context functions (for initializing the engine) =======
 */

void SoftwareDriver::initRenderer()
{
    /* Create the "Ray" device */
    rayCreateDevice(__Screen_Width, __Screen_Height);
    
    /* Default settings */
    rayEnable(RAY_NORMALIZE);
    rayEnable(RAY_COLOR_SMOOTH);
    
    /* Flip the video buffer */
    flipBuffers();
}

#if defined(WIN32)

void SoftwareDriver::changeDisplaySettings(bool isFullScreen, dim::size2di Dimension, s32 ColorDepth)
{
    if (isFullScreen)
    {
        DEVMODE dmScreenSettings;
        memset(&dmScreenSettings, 0, sizeof(dmScreenSettings));
        dmScreenSettings.dmSize         = sizeof(dmScreenSettings);
        dmScreenSettings.dmPelsWidth    = Dimension.Width;
        dmScreenSettings.dmPelsHeight   = Dimension.Height;
        dmScreenSettings.dmBitsPerPel   = ColorDepth;
        dmScreenSettings.dmFields       = DM_BITSPERPEL | DM_PELSWIDTH | DM_PELSHEIGHT;
        
        /* Change the display settings for the screen */
        ChangeDisplaySettings(&dmScreenSettings, CDS_FULLSCREEN);
    }
}

bool SoftwareDriver::initScreen(HWND hWnd, bool isAntiAlias, s32 MultiSamplePixelFormat, s32 ColorDepth)
{
    
    /* Device context */
    if ( !( hDC_ = GetDC(hWnd) ) ) // Device-Context
    {
        io::printError("could not create device context");
        return false;
    }
    
    /* Pixel format attributes */
    static PIXELFORMATDESCRIPTOR pfd =
    {
        sizeof(PIXELFORMATDESCRIPTOR),  // Size Of This Pixel Format Descriptor
        1,                              // Version Number
        PFD_DRAW_TO_WINDOW |            // Format Must Support Window
        PFD_SUPPORT_OPENGL |            // Format Must Support OpenGL
        PFD_DOUBLEBUFFER,               // Must Support Double Buffering
        PFD_TYPE_RGBA,                  // Request An RGBA Format
        ColorDepth,                     // Select Our Color Depth
        0, 0, 0, 0, 0, 0,               // Color Bits Ignored
        0,                              // No Alpha Buffer
        0,                              // Shift Bit Ignored
        0,                              // No Accumulation Buffer
        0, 0, 0, 0,                     // Accumulation Bits Ignored
        24,                             // Z-Buffer bits (Depth Buffer)
        0,                              // Stencil Buffer
        0,                              // No Auxiliary Buffer
        PFD_MAIN_PLANE,                 // Main Drawing Layer
        0,                              // Reserved
        0, 0, 0                         // Layer Masks Ignored
    };
    
    /* Choos a pixel format */
    if (isAntiAlias && MultiSamplePixelFormat)
    {
        PixelFormat_ = MultiSamplePixelFormat;
    }
    else
    {
        if ( !( PixelFormat_ = ChoosePixelFormat(hDC_, &pfd) ) )
        {
            io::printError("could not find a suitable pixelformat");
            return false;
        }
    }
    
    /* Set the pixel format */
    if (!SetPixelFormat(hDC_, PixelFormat_, &pfd))
    {
        io::printError("could not set the pixelformat");
        return false;
    }
    
    /* Create a device context */
    if ( !( hRC_ = wglCreateContext(hDC_) ) ) // Rendering-Context
    {
        io::printError("could not create 'OpenGL' rendering context");
        return false;
    }
    
    /* Make current & activate the OpenGL context */
    if (!wglMakeCurrent(hDC_, hRC_))
    {
        io::printError("could not activate 'OpenGL' rendering context");
        return false;
    }
    
    return true;
    
}

void SoftwareDriver::clearScreen(HWND &hWnd)
{
    
    /* Delete the "Ray" device */
    rayDeleteDevice();
    
    if (hWnd)
    {
        if (hDC_)
        {
            if (!wglMakeCurrent(hDC_, 0))
                io::printError("release of 'OpenGL' device- and resource context failed");
            
            if (hRC_)
            {
                if (!wglDeleteContext(hRC_))
                    io::printError("release of 'OpenGL' rendering context failed");
                hRC_ = 0;
            }
            else
                io::printError("resource-context handle is zero");
            
            if (!ReleaseDC(hWnd, hDC_))
                io::printError("release of device context failed");
            hDC_ = 0;
        }
        else
            io::printError("device-contect handle is zero");
        
        if (!DestroyWindow(hWnd))
            io::printError("release of main window failed");
        hWnd = 0;
    }
    else
        io::printError("window handle is zero");
    
}

#elif defined(LINUX)

#endif


/*
 * ======= Rendering 3D scenes =======
 */

void SoftwareDriver::setMaterialStatesTriangleList(SMeshMaterial &Material, dim::vector3df &Rotation)
{
    
    /* Cull facing */
    switch (Material.RenderFace)
    {
        case video::RENDER_FRONT:
            rayRenderMode(isFrontFace_ ? RAY_RENDER_FRONT : RAY_RENDER_BACK);
            break;
        case video::RENDER_BACK:
            rayRenderMode(isFrontFace_ ? RAY_RENDER_BACK : RAY_RENDER_FRONT);
            break;
        case video::RENDER_BOTH:
            rayRenderMode(RAY_RENDER_BOTH);
            break;
    }
    
    /* Render states */
    raySetStatus(RAY_DEPTH_TEST , (s32)Material.isZBuffer                       );
    raySetStatus(RAY_LIGHTING   , (s32)(Material.isLighting && __isLighting)    );
    raySetStatus(RAY_BLEND      , (s32)Material.isBlending                      );
    
    /* Blending mode */
    //rayBlendFunc(Material.BlendSource, Material.BlendTarget);
    
    /* Polygon mode */
    Wireframe_ = Material.RenderModeFront;
    
}

void SoftwareDriver::setMaterialStatesSprite(SMeshMaterial &Material)
{
    
    /* Render states */
    raySetStatus(RAY_DEPTH_TEST, Material.isZBuffer);
    
    /* Blending mode */
    //rayBlendFunc(Material.BlendSource, Material.BlendTarget);
    
}

void SoftwareDriver::drawTriangleList(
    u32 &BufferID, u32* IndicesList, u32 &IndicesCount, std::vector<SSurfaceTexture> &TextureList)
{
    
    /* Temporary variables */
    register s32 i, j;
    
    u32 BufferIndex     = BufferID - 1;
    s32 TextureLayer    = 0;
    
    /* Update the modelview matrix */
    updateModelviewMatrix();
    
    rayUpdateNormalMatrix();
    
    const s32 CountOfTextureLayer = getMultitexCount();
    
    /* Loop for all texture layers */
    for (i = 0; i < TextureList.size() && i < CountOfTextureLayer; ++i)
    {
        /* Bind the textures */
        rayUseTextureLayer(i);
        rayBindTexture(TextureList[i].hTexture->getID());
        
        /* Load texture matrix */
        setTextureMatrix(TextureList[i].Matrix, i);
        
        /* Texture coordinate generation */
        raySetStatus(RAY_TEXGEN, TextureList[i].TexMappingCoords != MAPGEN_NONE);
        rayTexGen(RAY_TEXTURE_GEN_U, RAYMappingGenList[TextureList[i].TexMappingGen]);
        rayTexGen(RAY_TEXTURE_GEN_V, RAYMappingGenList[TextureList[i].TexMappingGen]);
    }
    
    /* Temporary variables */
    dim::vector3df Pos1, Pos2, Pos3;
    dim::vector3df Normal1, Normal2, Normal3;
    dim::vector3df Map1[CountOfTextureLayer], Map2[CountOfTextureLayer], Map3[CountOfTextureLayer];
    video::color Color1, Color2, Color3;
    
    PFNWIREFRAMEPROC pWireframeProc = rayDrawTriangle;
    
    /*
     * Loop for all vertices
     */
    
    switch (Wireframe_)
    {
        
        case WIREFRAME_LINES:
            pWireframeProc = rayDrawLineStrip;
        case WIREFRAME_SOLID:
        {
            for (i = 0; i < IndicesCount;)
            {
                
                /* Get the render information */
                Pos1    = (VertexBufferList_[BufferIndex])[ IndicesList[i] ].Position;
                Normal1 = (VertexBufferList_[BufferIndex])[ IndicesList[i] ].Normal;
                Color1  = (VertexBufferList_[BufferIndex])[ IndicesList[i] ].Color;
                Map1[0] = (VertexBufferList_[BufferIndex])[ IndicesList[i] ].TexCoord[0];
                Map1[1] = (VertexBufferList_[BufferIndex])[ IndicesList[i] ].TexCoord[1];
                ++i;
                
                Pos2    = (VertexBufferList_[BufferIndex])[ IndicesList[i] ].Position;
                Normal2 = (VertexBufferList_[BufferIndex])[ IndicesList[i] ].Normal;
                Color2  = (VertexBufferList_[BufferIndex])[ IndicesList[i] ].Color;
                Map2[0] = (VertexBufferList_[BufferIndex])[ IndicesList[i] ].TexCoord[0];
                Map2[1] = (VertexBufferList_[BufferIndex])[ IndicesList[i] ].TexCoord[1];
                ++i;
                
                Pos3    = (VertexBufferList_[BufferIndex])[ IndicesList[i] ].Position;
                Normal3 = (VertexBufferList_[BufferIndex])[ IndicesList[i] ].Normal;
                Color3  = (VertexBufferList_[BufferIndex])[ IndicesList[i] ].Color;
                Map3[0] = (VertexBufferList_[BufferIndex])[ IndicesList[i] ].TexCoord[0];
                Map3[1] = (VertexBufferList_[BufferIndex])[ IndicesList[i] ].TexCoord[1];
                ++i;
                
                /* Set the render states */
                rayColor4i(Color1.Red, Color1.Green, Color1.Blue, Color1.Alpha);
                rayTexCoord2f(0, Map1[0].X, Map1[0].Y);
                rayTexCoord2f(1, Map1[1].X, Map1[1].Y);
                rayNormal3f(Normal1.X, Normal1.Y, Normal1.Z);
                rayVertex3f(Pos1.X, Pos1.Y, Pos1.Z);
                
                rayColor4i(Color2.Red, Color2.Green, Color2.Blue, Color2.Alpha);
                rayTexCoord2f(0, Map2[0].X, Map2[0].Y);
                rayTexCoord2f(1, Map2[1].X, Map2[1].Y);
                rayNormal3f(Normal2.X, Normal2.Y, Normal2.Z);
                rayVertex3f(Pos2.X, Pos2.Y, Pos2.Z);
                
                rayColor4i(Color3.Red, Color3.Green, Color3.Blue, Color3.Alpha);
                rayTexCoord2f(0, Map3[0].X, Map3[0].Y);
                rayTexCoord2f(1, Map3[1].X, Map3[1].Y);
                rayNormal3f(Normal3.X, Normal3.Y, Normal3.Z);
                rayVertex3f(Pos3.X, Pos3.Y, Pos3.Z);
                
                /* Render the triangle/ line strip */
                pWireframeProc();
                
            } // next triangle
        }
        break;
        
        case WIREFRAME_POINTS:
        {
            for (i = 0; i < IndicesCount; ++i)
            {
                
                /* Get the render information */
                Pos1    = (VertexBufferList_[BufferIndex])[ IndicesList[i] ].Position;
                Normal1 = (VertexBufferList_[BufferIndex])[ IndicesList[i] ].Normal;
                Color1  = (VertexBufferList_[BufferIndex])[ IndicesList[i] ].Color;
                Map1[0] = (VertexBufferList_[BufferIndex])[ IndicesList[i] ].TexCoord[0];
                Map1[1] = (VertexBufferList_[BufferIndex])[ IndicesList[i] ].TexCoord[1];
                
                /* Set the render states */
                rayColor4i(Color1.Red, Color1.Green, Color1.Blue, Color1.Alpha);
                rayTexCoord2f(0, Map1[0].X, Map1[0].Y);
                rayTexCoord2f(1, Map1[1].X, Map1[1].Y);
                rayNormal3f(Normal1.X, Normal1.Y, Normal1.Z);
                rayVertex3f(Pos1.X, Pos1.Y, Pos1.Z);
                
                /* Render the point */
                rayDrawPoint();
                
            } // next vertex
        }
        break;
        
    }
    
    /* Unbind all textures */
    for (i = 0; i < TextureList.size() && i < CountOfTextureLayer; ++i)
    {
        rayUseTextureLayer(i);
        rayUnbindTexture();
    }
    
    rayUseTextureLayer(0);
    
}

void SoftwareDriver::drawTriangleList(u32 &BufferID, u32* IndicesList, u32 &IndicesCount)
{
    
}

void SoftwareDriver::drawSprite(
    Texture* &hTexture, f32 &RotationZ, video::color &Color, dim::vector3df &Scale, dim::point2df &BasePosition)
{
    
    /* Bind textureing */
    hTexture->bind(0);
    
    /* Coordination */
    if (RotationZ)
        scene::spWorldMatrix.matrixRotateZ(RotationZ);
    
    /* Update the render matrix */
    updateModelviewMatrix();
    
    dim::vector3df Pos1(-Scale.X + BasePosition.X,  Scale.Y + BasePosition.Y, 0);
    dim::vector3df Pos2( Scale.X + BasePosition.X,  Scale.Y + BasePosition.Y, 0);
    dim::vector3df Pos3( Scale.X + BasePosition.X, -Scale.Y + BasePosition.Y, 0);
    dim::vector3df Pos4(-Scale.X + BasePosition.X, -Scale.Y + BasePosition.Y, 0);
    
    rayColor4i(Color.Red, Color.Green, Color.Blue, Color.Alpha);
    rayTexCoord2i(0, 0, 0);
    rayVertex3f(Pos1.X, Pos1.Y, Pos1.Z);
    
    rayColor4i(Color.Red, Color.Green, Color.Blue, Color.Alpha);
    rayTexCoord2i(0, 1, 0);
    rayVertex3f(Pos2.X, Pos2.Y, Pos2.Z);
    
    rayColor4i(Color.Red, Color.Green, Color.Blue, Color.Alpha);
    rayTexCoord2i(0, 1, 1);
    rayVertex3f(Pos3.X, Pos3.Y, Pos3.Z);
    
    /* First triangle */
    rayDrawTriangle();
    
    rayColor4i(Color.Red, Color.Green, Color.Blue, Color.Alpha);
    rayTexCoord2i(0, 0, 0);
    rayVertex3f(Pos1.X, Pos1.Y, Pos1.Z);
    
    rayColor4i(Color.Red, Color.Green, Color.Blue, Color.Alpha);
    rayTexCoord2i(0, 1, 1);
    rayVertex3f(Pos3.X, Pos3.Y, Pos3.Z);
    
    rayColor4i(Color.Red, Color.Green, Color.Blue, Color.Alpha);
    rayTexCoord2i(0, 0, 1);
    rayVertex3f(Pos4.X, Pos4.Y, Pos4.Z);
    
    /* Second triangle */
    rayDrawTriangle();
    
    /* Unbind textureing */
    hTexture->unbind(0);
    
}

void SoftwareDriver::updateLight(
    u32 &LightID, scene::ELightTypes &LightingType, bool &isVolumetric,
    dim::vector3df Direction, f32 &NearSpot, f32 &FarSpot,
    f32 &AttenuationConstant, f32 &AttenuationLinear, f32 &AttenuationQuadratic)
{
    
    
    
}

void SoftwareDriver::createVertexBuffer(u32 &BufferID)
{
    SMeshVertex3D* p;
    
    VertexBufferList_.push_back(p);
    VertexBufferIDList_.push_back(&BufferID);
    
    BufferID = VertexBufferList_.size();
}
void SoftwareDriver::deleteVertexBuffer(u32 &BufferID)
{
    /*
    if (BufferID > 0)
    {
        for (std::vector<u32*>::iterator it = VertexBufferIDList_.begin() + BufferID; it != VertexBufferIDList_.end(); ++it)
            --**it;
        
        std::vector<SVertexInfo*>::iterator it1 = VertexBufferList_.begin() + (BufferID - 1);
        VertexBufferList_.erase(it1);
        
        std::vector<u32*>::iterator it2 = VertexBufferIDList_.begin() + (BufferID - 1);
        VertexBufferIDList_.erase(it2);
    }
    */
}
void SoftwareDriver::updateVertexBuffer(u32 &BufferID, SMeshVertex3D* pVerticesList, s32 VerticesCount)
{
    /* Update the vertex buffer pointer */
    VertexBufferList_[BufferID - 1] = pVerticesList;
}

void SoftwareDriver::disableTriangleListStates()
{
    rayEnable(RAY_BLEND_TEST);
    rayEnable(RAY_BLEND);
}

void SoftwareDriver::disable3DRenderStates()
{
    rayDisable(RAY_BLEND_TEST);
    rayDisable(RAY_BLEND);
    rayDisable(RAY_LIGHTING);
}

void SoftwareDriver::disableTexturing()
{
    rayUnbindTexture();
}

void SoftwareDriver::setDefaultAlphaBlending()
{
    rayBlendFunc(BLEND_NORMAL);
}

void SoftwareDriver::enableBlending()
{
    rayEnable(RAY_BLEND_TEST);
}

void SoftwareDriver::disableBlending()
{
    rayDisable(RAY_BLEND_TEST);
}


/*
 * ======= Special effects actor =======
 */

scene::effect::SpecialEffectsActor* SoftwareDriver::createSpecialEffectsActor()
{
    return new scene::effect::SpecialEffectsActor();
}


/*
 * ======= Stack operation functions =======
 */

void SoftwareDriver::pushViewport()
{
    s32 ViewportRect[4];
    
    rayGetValue(RAY_VIEWPORT, ViewportRect);
    
    StackViewport_[0] = ViewportRect[0];
    StackViewport_[1] = ViewportRect[1];
    StackViewport_[2] = ViewportRect[2];
    StackViewport_[3] = ViewportRect[3];
}

void SoftwareDriver::popViewport()
{
    rayViewport(StackViewport_[0], StackViewport_[1], StackViewport_[2], StackViewport_[3]);
}


/*
 * ======= Texturing =======
 */

void SoftwareDriver::bindTexture(u32 ID)
{
    rayBindTexture(ID);
}

void SoftwareDriver::unbindTexture()
{
    rayUnbindTexture();
}

void SoftwareDriver::setTextureSubImage(u32 ID, s32 X, s32 Y, s32 Width, s32 Height, const u8* ImageData)
{
    rayTexSubImage2D(ID, 0, 3, RAY_UNSIGNED_BYTE, ImageData);
}


/*
 * ========== Drawing 2D objects ==========
 */



/* Drawing 2d */

void SoftwareDriver::beginDrawing2D(bool isSettingViewport)
{
    /* Disable depth test (only 2d) */
	rayDisable(RAY_DEPTH_TEST);
	
	/* Enable alpha test */
	rayEnable(RAY_ALPHA_TEST);
	
    /* Projection matrix */
    rayMatrixMode(RAY_PROJECTION_MATRIX);
	rayLoadIdentity();
	
    /* Modelview matrix */
    rayMatrixMode(RAY_MODELVIEW_MATRIX);
    Matrix2D_.make2Dimensional(__Screen_Width, __Screen_Height);
	
	/* Disable lighting */
	rayDisable(RAY_LIGHTING);
    
    /* Use first texture layer */
    rayUseTextureLayer(0);
}

void SoftwareDriver::endDrawing2D()
{
	/* Enable depth test */
    rayEnable(RAY_DEPTH_TEST);
    
    /* Disable alpha test */
	rayDisable(RAY_ALPHA_TEST);
}

/* Drawing 3d */

void SoftwareDriver::beginDrawing3D(bool isLighting, bool isFog)
{
    rayLoadIdentity();
}

void SoftwareDriver::endDrawing3D()
{
    rayLoadIdentity();
}

/* Alpha blending */

void SoftwareDriver::beginAlphaBlending(u8 Type)
{
    /* Enable blending */
    rayEnable(RAY_BLEND_TEST);
    
    /* Blending function */
    if (Type == BLENDING_BRIGHTNESS)
        rayBlendFunc(BLEND_BRIGHT);
    else if (Type == BLENDING_DARKNESS)
        rayBlendFunc(BLEND_DARK);
}

void SoftwareDriver::endAlphaBlending()
{
    /* Disable blending */
    rayDisable(RAY_BLEND_TEST);
}

/* Masking */

void SoftwareDriver::beginMasking(u8 Type)
{
    
}

void SoftwareDriver::endMasking()
{
    
}

/* Texture drawing */

void SoftwareDriver::beginTextureDrawing(Texture* hTexture, s32 RenderMode, u32 TextureDataType)
{
    
}

void SoftwareDriver::endTextureDrawing(bool isUpdateImageData)
{
    
}

/* Scissors (drawing rectangle) */

void SoftwareDriver::beginDrawingRect(dim::rect2di Dimension)
{
    /* Enable scissor */
    rayEnable(RAY_SCISSOR_TEST);
    
    /* Scissor area */
    rayScissor(Dimension.Left, Dimension.Top, Dimension.Right - Dimension.Left, Dimension.Bottom - Dimension.Top);
}

void SoftwareDriver::beginDrawingRect(dim::point2di Position, dim::size2di Size)
{
    /* Enable scissor */
    rayEnable(RAY_SCISSOR_TEST);
    
    /* Scissor area */
    rayScissor(Position.X, Position.Y, Size.Width, Size.Height);
}

void SoftwareDriver::endDrawingRect()
{
    /* Disable scissor */
    rayDisable(RAY_SCISSOR_TEST);
}

/* Viewports (drawing area) */

void SoftwareDriver::setDrawingArea(dim::rect2di Dimension)
{
    
}

void SoftwareDriver::setDrawingArea(dim::point2di Position, dim::size2di Size)
{
    
}

/* Line & point size */

void SoftwareDriver::setLineSize(s32 Size)
{
    
}
void SoftwareDriver::setPointSize(s32 Size)
{
    
}

void SoftwareDriver::setSolidMode(bool isSolidMode)
{
    
}


/* ======= Texture drawing ======= */


void SoftwareDriver::draw2DImage(Texture* hTexture, dim::point2di Position, color Color, bool isTransFilterUsing)
{
    
    if (!isTransFilterUsing && Color == 255)
    {
        rayDrawImage2DFast(hTexture->getID(), Position.X, Position.Y);
    }
    else
    {
        
        
    }
    
}

// !INCOMPLETE!
void SoftwareDriver::draw2DImage(
    Texture* hTexture, dim::rect2di Position, dim::rect2df Clipping, color Color, bool isTransFilterUsing)
{
    
    rayBindTexture(hTexture->getID());
    
    /* Load 2dimensional matrix */
    //rayLoadMatrix(Matrix2D_.getArray16());
    rayLoadIdentity();
    
    /* Draw the first triangle of the rectangle */
    rayColor4i(Color.Red, Color.Green, Color.Blue, Color.Alpha);
    rayTexCoord2f(0, Clipping.Left, Clipping.Top);
    rayVertex3i(Position.Left, Position.Top, 3);
    
    rayColor4i(Color.Red, Color.Green, Color.Blue, Color.Alpha);
    rayTexCoord2f(0, Clipping.Right, Clipping.Top);
    rayVertex3i(Position.Right, Position.Top, 3);
    
    rayColor4i(Color.Red, Color.Green, Color.Blue, Color.Alpha);
    rayTexCoord2f(0, Clipping.Right, Clipping.Bottom);
    rayVertex3i(Position.Right, Position.Bottom, 3);
    
    rayDrawTriangle();
    
    /* Draw the second triangle of the rectangle */
    rayColor4i(Color.Red, Color.Green, Color.Blue, Color.Alpha);
    rayTexCoord2f(0, Clipping.Left, Clipping.Top);
    rayVertex2i(Position.Left, Position.Top);
    
    rayColor4i(Color.Red, Color.Green, Color.Blue, Color.Alpha);
    rayTexCoord2f(0, Clipping.Right, Clipping.Bottom);
    rayVertex2i(Position.Right, Position.Bottom);
    
    rayColor4i(Color.Red, Color.Green, Color.Blue, Color.Alpha);
    rayTexCoord2f(0, Clipping.Left, Clipping.Bottom);
    rayVertex2i(Position.Left, Position.Bottom);
    
    rayDrawTriangle();
}

void SoftwareDriver::draw2DImage(
    Texture* hTexture, dim::point2di Position,
    f32 Rotation, f32 Radius, color Color, bool isTransFilterUsing)
{
    
}

void SoftwareDriver::draw2DImage(
    Texture* hTexture, dim::rect2di Position, dim::rect2df Clipping,
    bool isTransFilterUsing, f32 Rotation, dim::point2df RotationPoint,
    color lefttopColor, color righttopColor, color rightbottomColor, color leftbottomColor)
{
    
}

void SoftwareDriver::draw2DImage(
    Texture* hTexture,
    dim::point2di lefttopPosition, dim::point2di righttopPosition,
    dim::point2di rightbottomPosition, dim::point2di leftbottomPosition,
    dim::point2df lefttopClipping, dim::point2df righttopClipping,
    dim::point2df rightbottomClipping, dim::point2df leftbottomClipping,
    bool isTransFilterUsing,
    color lefttopColor, color righttopColor,
    color rightbottomColor, color leftbottomColor)
{
    
}


/* ======= Primitive drawing ======= */


color SoftwareDriver::getPixel(dim::point2di Position)
{
    RAYcolor_t r, g, b, a;
    
    rayRead2DPixel(Position.X, Position.Y, r, g, b, a);
    
    return color(r, g, b, a);
}


void SoftwareDriver::draw2DPixel(dim::point2di Position, color Color)
{
    rayWrite2DPixel(Position.X, Position.Y, Color.Red, Color.Green, Color.Blue, Color.Alpha);
}

void SoftwareDriver::draw2DLine(dim::point2di Position1, dim::point2di Position2, color Color)
{
    
}

void SoftwareDriver::draw2DLine(dim::point2di Position1, dim::point2di Position2, color Color1, color Color2)
{
    
}

void SoftwareDriver::draw2DLine(dim::point2di Position1, dim::point2di Position2, color Color, s32 DotLength)
{
    
}

void SoftwareDriver::draw2DRectangle(dim::rect2di Position, color Color, EDrawingModeTypes Type)
{
    
}

void SoftwareDriver::draw2DRectangle(
    dim::rect2di Position, color lefttopColor, color righttopColor,
    color rightbottomColor, color leftbottomColor, EDrawingModeTypes Type)
{
    
}

void SoftwareDriver::draw2DRoundRectangle(dim::rect2di Position, dim::size2di CornerRadius, color Color)
{
    
}

void SoftwareDriver::draw2DCircle(dim::point2di Position, s32 Radius, color Color)
{
    
}

void SoftwareDriver::draw2DEllipse(dim::point2di Position, dim::size2di Radius, color Color)
{
    
}

void SoftwareDriver::draw2DEllipse(
    dim::point2di Position, dim::size2di Radius, f32 RotationDegree,
    color Color, bool isSolid)
{
    
}

void SoftwareDriver::draw2DPolygon(SPrimitiveVertex* VerticesList, u32 AryCount, EDrawingModeTypes Type)
{
    
}

void SoftwareDriver::draw2DPolygonImage(
    Texture* hTexture, SPrimitiveVertex* VerticesList, u32 AryCount, EDrawingModeTypes Type)
{
    
}

/*
 * ======= Extra drawing functions =======
 */


void SoftwareDriver::draw2DArc(
    dim::point2di Position, dim::size2di Radius, s32 RadialDegree1,
    s32 RadialDegree2, f32 RotationDegree, color Color)
{
    
}

void SoftwareDriver::draw2DBezier(
    dim::point2di Position1, dim::point2di Position2,
    dim::point2di Radial1, dim::point2di Radial2,
    color Color, f32 Detail)
{
    dim::point2df Pos1, Pos2;
    Pos1 = math::getBezierValue(
        0.0f, Position1.cast<f32>(), Position2.cast<f32>(), Radial1.cast<f32>(), Radial2.cast<f32>()
    );
    
    for (f32 t = 0.0f; t < 1.0f; t += Detail)
    {
        Pos2 = math::getBezierValue(
            t+Detail, Position1.cast<f32>(), Position2.cast<f32>(), Radial1.cast<f32>(), Radial2.cast<f32>()
        );
        
        draw2DLine(Pos1.cast<s32>(), Pos2.cast<s32>(), Color);
        
        Pos1 = Pos2;
    }
}

void SoftwareDriver::draw2DSpline(math::Spline2D &Spline, s32 Detail, dim::point2di Pos, color Color)
{
    
}

void SoftwareDriver::draw2DFloodFill(dim::point2di Position, color FillColor, color BorderColor)
{
    
}


/* ======= 3D drawing functions ======= */


void SoftwareDriver::draw3DPixel(dim::vector3df Position, color Color, scene::Camera* hCamera)
{
    
}

void SoftwareDriver::draw3DLine(dim::vector3df Position1, dim::vector3df Position2, color Color, scene::Camera* hCamera)
{
    
}

void SoftwareDriver::draw3DLine(
    dim::vector3df Position1, dim::vector3df Position2,
    color Color1, color Color2, scene::Camera* hCamera)
{
    
}

void SoftwareDriver::draw3DEllipse(
    dim::vector3df Position, dim::vector3df Rotation, dim::size2df Radius,
    color Color, scene::Camera* hCamera)
{
    
}

void SoftwareDriver::draw3DTriangle(
    Texture* hTexture, dim::triangle3df Triangle,
    dim::vector3df Position, dim::vector3df Rotation, dim::vector3df Scale,
    color Color, scene::Camera* hCamera)
{
}

void SoftwareDriver::draw3DBox(
    dim::vector3df Corner1, dim::vector3df Corner2,
    dim::vector3df Position, dim::vector3df Rotation, dim::vector3df Scale,
    color Color, scene::Camera* hCamera)
{
    
}

void SoftwareDriver::draw3DBone(scene::Entity* Obj, dim::matrix4f &BoneTransformation, scene::Camera* hCamera)
{
    matrixWorldViewReset();
    
    /* === Bone location === */
    
    if (hCamera)
        hCamera->updateModelviewMatrix();
    
    Obj->updateModelviewMatrix();
    
    scene::spWorldMatrix *= BoneTransformation;
    
    /* === Render the bone === */
    
    /* Bone frame */
    for (s32 i = 0; i < 24; i += 2)
    {
        draw3DLine(
            dim::vector3df(-__spBoneBodyCoords[i+0][1], __spBoneBodyCoords[i+0][2], __spBoneBodyCoords[i+0][0]),
            dim::vector3df(-__spBoneBodyCoords[i+1][1], __spBoneBodyCoords[i+1][2], __spBoneBodyCoords[i+1][0]),
            video::color(0, 0, 255)
        );
    }
    
    /* X-Axis */
    for (s32 i = 0; i < 36; ++i)
    {
        draw3DLine(
            dim::vector3df(0.0f, SIN(i*10)*0.05f, COS(i*10)*0.05f),
            dim::vector3df(0.0f, SIN((i+1)*10)*0.05f, COS((i+1)*10)*0.05f),
            video::color(255, 0, 0)
        );
    }
    
    /* Y-Axis */
    for (s32 i = 0; i < 36; ++i)
    {
        draw3DLine(
            dim::vector3df(SIN(i*10)*0.05f, 0.0f, COS(i*10)*0.05f),
            dim::vector3df(SIN((i+1)*10)*0.05f, 0.0f, COS((i+1)*10)*0.05f),
            video::color(0, 255, 0)
        );
    }
    
    /* Z-Axis */
    for (s32 i = 0; i < 36; ++i)
    {
        draw3DLine(
            dim::vector3df(SIN(i*10)*0.05f, COS(i*10)*0.05f, 0.0f),
            dim::vector3df(SIN((i+1)*10)*0.05f, COS((i+1)*10)*0.05f, 0.0f),
            video::color(255, 255, 0)
        );
    }
}


/*
 * ========== Texture loading & creating ==========
 */


/*
 * Most important function, loads all the file-formats which are integrated
 * (BMP, DIB), (JPG, JPEG, JPE, JFIF), (TGA), (PNG), (PCX)
 */

Texture* SoftwareDriver::loadTexture(io::stringc Filename)
{
    
    /* Information message */
    io::printMessage(io::stringc("Load texture: \"")+Filename+io::stringc("\""));
    
    /* Allocate some memory & temporary variables */
    
    Texture* LoadedTexture      = 0;
    textureRAWInfo* texture     = 0;
    TextureLoaderRAW* texraw    = 0;
    
    io::stringc FileFormat4     = Filename.right(4).upper();
    io::stringc FileFormat5     = Filename.right(5).upper();
    
    u32 TexID                   = 0;
    s32 Type                    = TEXTURE_RGB;
    s32 InternalFormat          = 3;
    u8* ImageData               = 0;
    dim::size2di Size;
    
    /* === Check if the file does exist === */
    
    if (!io::FileLoader().doesFileExist(Filename))
    {
        /* Error message */
        io::printError("could not found texture file");
        
        /* Create a standard empty texture */
        return createTexture(DEF_TEXTURE_SIZE);
    }
    
    /* === Select format === */
    
    if (FileFormat4 == ".BMP" || FileFormat4 == ".DIB") // Windows Bitmap
        texraw = new TextureLoaderBMP(Filename);
    else if (FileFormat4 == ".JPG" || FileFormat5 == ".JPEG" || FileFormat4 == ".JPE" || FileFormat5 == ".JFIF") // Joint Photographic Experts
        texraw = new TextureLoaderJPG(Filename);
    else if (FileFormat4 == ".TGA") // Targa Image
        texraw = new TextureLoaderTGA(Filename);
    else if (FileFormat4 == ".PNG") // Portable Network Graphics
        texraw = new TextureLoaderPNG(Filename);
    else if (FileFormat4 == ".PCX") // Picture Exchange
        texraw = new TextureLoaderPCX(Filename);
    else
    {
        /* Error message */
        io::printError(io::stringc("texture file format: \"")+Filename.getFileExtension().upper()+io::stringc("\" is not supported"));
        
        /* Allocate the texture */
        LoadedTexture = createTexture(DEF_TEXTURE_SIZE);
        
        /* Add the texture to the texture list */
        TextureList_.push_back(LoadedTexture);
        
        /* Return the texture & exit the function */
        return LoadedTexture;
    }
    
    /* === Load the "RAW" texture === */
    
    /* Get the texture raw data */
    if (texraw)
        texture = texraw->loadImageData();
    
    /* Delete the raw texture loader */
    spDeletePointer(texraw);
    
    /* === Create the "Ray" texture === */
    
    if (texture)
    {
        /* Register a new "Ray" texture */
        rayGenTexture(TexID);
        
        /* Settings */
        Type            = texture->Format;
        InternalFormat  = texture->InternalFormat;
        ImageData       = texture->ImageData;
        Size            = dim::size2di(texture->Width, texture->Height);
    }
    else
    {
        /* Error message */
        io::printError("could not load texture");
    }
    
    /* === Create "SoftPixel" texture === */
    
    /* Allocate the texture */
    if (texture)
    {
        LoadedTexture = new SoftwareTexture(
            TexID, Filename, Size, ImageData, TEXTURE_LINEAR, TexGenFilterMode_, InternalFormat, Type
        );
    }
    else
        LoadedTexture = createTexture(DEF_TEXTURE_SIZE);
    
    /* Delete the RAW texture */
    spDeletePointer(texture);
    
    /* Add the texture to the texture list */
    TextureList_.push_back(LoadedTexture);
    
    /* Return the texture & exit the function */
    return LoadedTexture;
    
}

bool SoftwareDriver::isTexture(Texture* hTexture)
{
    return hTexture && hTexture->getID();
}

Texture* SoftwareDriver::copyTexture(Texture* hTexture)
{
    /* Temporary vairables */
    u32 TexID;
    s32 InternalFormat  = hTexture->getFormat();
    s32 Type            = hTexture->getType();
    dim::size2di Size   = hTexture->getOriginalSize();
    s32 ImageDataSize   = Size.Width * Size.Height * InternalFormat;
    
    /* Generate a new "Ray" texture */
    rayGenTexture(TexID);
    
    /* Get the source image data */
    u8* pData = hTexture->getImageData();
    
    /* Allocate new image memory */
    u8* NewImageData = new u8[ImageDataSize];
    
    /* Copy the image data buffers */
    memcpy(NewImageData, pData, ImageDataSize);
    
    /* Allocate & return the new texture */
    return new SoftwareTexture(
        TexID, hTexture->getFilename(), Size, NewImageData,
        TexGenRenderMode_, TexGenFilterMode_, InternalFormat, Type, TexGenEdgeSmooth_, TexGenMipMapping_
    );
}

Texture* SoftwareDriver::createTexture(dim::size2di Dimension, u32 TextureDataType, io::stringc Filename)
{
    /* Temporary vairables */
    u32 TexID;
    s32 InternalFormat  = TextureManager::getInternalFormat(TextureDataType);
    s32 ImageDataSize   = Dimension.Width * Dimension.Height * InternalFormat;
    
    register s32 x, y, i, j = 0;
    
    /* Register a new "Ray" texture */
    rayGenTexture(TexID);
    
    /* Allocate new image memory */
    u8* ImageData = new u8[ImageDataSize];
    
    /* Fill the image buffer */
    for (y = 0; y < Dimension.Height; ++y)
    {
        for (x = 0; x < Dimension.Width; ++x)
        {
            for (i = 0; i < InternalFormat; ++i, ++j)
                ImageData[j] = StdFillColor_[i];
        }
    }
    
    /* Allocate & return the new texture */
    return new SoftwareTexture(
        TexID, Filename, Dimension, ImageData, TexGenRenderMode_, TexGenFilterMode_,
        3, TEXTURE_RGB, TexGenEdgeSmooth_, TexGenMipMapping_
    );
}

Texture* SoftwareDriver::createTexture(dim::size2di Dimension, u8* ImageData, u32 TextureDataType, io::stringc Filename)
{
    /* Temporary vairables */
    u32 TexID;
    s32 InternalFormat = TextureManager::getInternalFormat(TextureDataType);
    
    /* Register a new "Ray" texture */
    rayGenTexture(TexID);
    
    /* Allocate & return the new texture */
    return new SoftwareTexture(
        TexID, Filename, Dimension, ImageData, TexGenRenderMode_, TexGenFilterMode_,
        InternalFormat, TextureDataType, TexGenEdgeSmooth_, TexGenMipMapping_
    );
}

Texture* SoftwareDriver::createScreenShot(dim::point2di Position, dim::size2di Size)
{
    /* Use the standard dimension */
    if (Size == dim::size2di())
        Size = dim::size2di(__Screen_Width, __Screen_Height);
    
    /* Temporary vairables */
    u32 TexID;
    
    const s32 ImageDataSize = Size.Width * Size.Height * 3;
    
    register s32 x, y, i = 0, j;
    
    /* Register a new "Ray" texture */
    rayGenTexture(TexID);
    
    /* Allocate new image memory */
    u8* ImageData = new u8[ImageDataSize];
    
    /* Copy the pixel buffer into the image data */
    rayCopyTexImage2D(TexID, 0, 0, 0, Position.X, Position.Y, Size.Width, Size.Height);
    
    rayGetImage2D(TexID, 0, 3, RAY_UNSIGNED_BYTE, ImageData);
    
    /* Allocate & return the new texture */
    return new SoftwareTexture(
        TexID, "", Size, ImageData, TexGenRenderMode_, TexGenFilterMode_,
        3, TEXTURE_RGB, TexGenEdgeSmooth_, TexGenMipMapping_
    );
}

void SoftwareDriver::createScreenShot(Texture* hTexture, dim::point2di Position)
{
    /* Copy the pixel buffer into the image data */
    rayCopyTexSubImage2D(
        hTexture->getID(), 0, 0, 0, Position.X, Position.Y, hTexture->getType() == TEXTURE_DEPTH ? RAY_DEPTH : RAY_RGBA
    );
}


/*
 * ========== Font loading & drawing ==========
 */


Font* SoftwareDriver::loadFont(
    io::stringc FontName, dim::size2di FontSize, s32 TypeFlag, s32 RotationAngle, bool isSymbolUsing,
    bool is3DSpaceUsing, f32 FontDepth, f32 DetailDeviation, EDrawingModeTypes Wireframe)
{
    u32 FontID  = 0;
    HFONT hFont = 0;
    
    return new Font(FontID, FontName, FontSize, 0, (s32)hFont, is3DSpaceUsing);
}

Font* SoftwareDriver::loadFont(
    io::stringc FontName, s32 FontSize, s32 TypeFlag, s32 RotationAngle, bool isSymbolUsing,
    bool is3DSpaceUsing, f32 FontDepth, f32 DetailDeviation, EDrawingModeTypes Wireframe)
{
    return loadFont(
        FontName, dim::size2di(0, FontSize), TypeFlag, RotationAngle, isSymbolUsing,
        is3DSpaceUsing, FontDepth, DetailDeviation, Wireframe
    );
}

Font* SoftwareDriver::loadFont(Texture* hTextureFont, dim::size2di FontSize, dim::size2di ElementCount)
{
    u32 FontID  = 0;
    HFONT hFont = 0;
    
    return new Font(FontID, hTextureFont->getFilename(), FontSize, hTextureFont, (s32)hFont, false);
}

void SoftwareDriver::deleteFont(Font* hFont)
{
    /* Delete the font */
    spDeletePointer(hFont);
}

void SoftwareDriver::draw2DText(Font* hFont, dim::point2di Position, io::stringc Text, color Color)
{
    
}


/*
 * ======= Matrix controll =======
 */

void SoftwareDriver::updateModelviewMatrix()
{
    rayMatrixMode(RAY_MODELVIEW_MATRIX);
    rayLoadMatrix( (scene::spViewMatrix * scene::spWorldMatrix).getArray16() );
}

void SoftwareDriver::setProjectionMatrix(const dim::matrix4f &Matrix)
{
    scene::spProjectionMatrix = Matrix;
    
    rayMatrixMode(RAY_PROJECTION_MATRIX);
    rayLoadMatrix(Matrix.getArray16());
}

void SoftwareDriver::setTextureMatrix(const dim::matrix4f &Matrix, u8 TextureLayer)
{
    scene::spTextureMatrix[TextureLayer] = Matrix;
    
    rayMatrixMode(RAY_TEXTURE_MATRIX);
    rayLoadMatrix(Matrix.getArray16());
}


/*
 * ========== Other renderer option fuctions =======
 */


void SoftwareDriver::setViewport(s32 X, s32 Y, s32 Width, s32 Height)
{
    rayViewport(X, Y, Width, Height);
    
    dim::matrix4f ViewPortMat;
    ViewPortMat.makeViewport(dim::rect2di(X, Y, X + Width, Y + Height));
    
    rayMatrixMode(RAY_VIEWPORT_MATRIX);
    rayLoadMatrix(ViewPortMat.getArray16());
}

void SoftwareDriver::setClippingRange(f32 Near, f32 Far)
{
    RangeNear_  = Near;
    RangeFar_   = Far;
    rayPerspective(Near, Far, __Screen_Width/2);
}


/*
 * ========== Private: ==========
 */

void SoftwareDriver::build2DMipmaps(s32 InternalFormat, s32 Format, s32 Width, s32 Height, const u8* ImageData)
{
    /* Temporary memories */
    s32 i                   = 0;
    u8* pData               = 0;
    const s32 ImageDataSize = Width*Height*InternalFormat;
    
    dim::size2di Size(Width, Height);
    
    /* Allocate new image data memory */
    pData = new u8[ImageDataSize];
    
    /* Copy the image data */
    memcpy(pData, ImageData, ImageDataSize);
    
    /* Get the optimized size & scale if the input size is not correct */
    TextureManager::checkImageSize(Size);
    
    /* Clamp the image data size, must be power of to (2^n) */
    if (Size.Width != Width || Size.Height != Height)
        TextureManager::scaleImageData(pData, Width, Height, Size.Width, Size.Height, InternalFormat);
    
    /* Loop for all mipmap levels */
    while (1)
    {
        /* Build a new mipmap level */
        rayTexImage2D(rayGetCurTexture(), i++, Size.Width, Size.Height, InternalFormat, RAY_UNSIGNED_BYTE, pData);
        
        /* Check if the last mipmap level has been achieved */
        if (Size.Width == 1 && Size.Height == 1)
            break;
        
        /* Scale the image data to the half size */
        TextureManager::scaleHalfImageData(pData, Size.Width, Size.Height, InternalFormat);
        
        /* Half the size */
        if (Size.Width > 1)
            Size.Width >>= 1;
        if (Size.Height > 1)
            Size.Height >>= 1;
    }
    
    /* Delete temporary memory */
    spDeleteArray(pData);
}


} // /namespace video

} // /namespace sp






// ================================================================================
