/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * Texture loading PCX-fileformat file
 *
 */

#include "spTextureLoaderPCX.hpp"


namespace sp
{
namespace video
{


TextureLoaderPCX::TextureLoaderPCX(io::stringc &Filename) : TextureLoaderRAW(Filename)
{
}
TextureLoaderPCX::~TextureLoaderPCX()
{
}

textureRAWInfo* TextureLoaderPCX::loadImageData()
{
    /* Check if the file has opened correct */
    if (isFileError())
        return 0;
    
    /* Allocate new texture RAW data & header buffer */
    textureRAWInfo* texture = 0;
    
    headerPCXInfo HeaderInfo;
    
    // Checking for header version
    fread(&HeaderInfo, sizeof(HeaderInfo), 1, hFile_);
    if (HeaderInfo.manufacturer != 0x0A)
    {
        io::printError(io::stringc("PCX file has an incorrect version number: ")+io::stringc(HeaderInfo.manufacturer));
        return 0;
    }
    
    // Initialize texture parameters
    texture = new textureRAWInfo();
    texture->Width          = HeaderInfo.xmax - HeaderInfo.xmin + 1;
    texture->Height         = HeaderInfo.ymax - HeaderInfo.ymin + 1;
    texture->Format         = TEXTURE_RGB;
    texture->InternalFormat = 3;
    texture->ImageData      = new u8[ texture->Width * texture->Height * texture->InternalFormat ];
    
    switch (HeaderInfo.bpp)
    {
        case 1:
            readPCX1bit(&HeaderInfo, texture); break;
        case 4:
            readPCX4bit(&HeaderInfo, texture); break;
        case 8:
            readPCX8bit(&HeaderInfo, texture); break;
        case 24:
            readPCX24bit(&HeaderInfo, texture); break;
        default:
            io::printError(io::stringc("PCX file has unknown bitcount: ")+io::stringc(HeaderInfo.bpp));
            spDeleteArray(texture->ImageData);
            spDeletePointer(texture);
            return 0;
    }
    
    if (!texture->ImageData)
        return 0;
    
    // Flip Y axis
    TextureManager::flipImageDataY(texture->Width, texture->Height, texture->ImageData);
    
    return texture;
}

/* ======= Private: ======= */

void TextureLoaderPCX::readPCX1bit(headerPCXInfo* pHeader, textureRAWInfo* texture)
{
    s32 y, i, bytes;
    s32 colorIndex;
    s32 rle_count = 0, rle_value = 0;
    u8* ptr = texture->ImageData;
    
    for (y = 0; y < texture->Height; ++y)
    {
        ptr = &texture->ImageData[(texture->Height - (y + 1)) * texture->Width * 3];
        bytes = pHeader->bytesPerScanLine;
        
        // Decode line number y
        while (bytes--)
        {
            if (rle_count == 0)
            {
                if ((rle_value = fgetc(hFile_)) < 0xC0)
                {
                    rle_count = 1;
                }
                else
                {
                    rle_count = rle_value - 0xC0;
                    rle_value = fgetc(hFile_);
                }
            }
            
            --rle_count;
            
            for (i = 7; i >= 0; --i, ptr += 3)
            {
                colorIndex = ((rle_value & (1 << i)) > 0);
                
                ptr[0] = pHeader->palette[colorIndex * 3 + 0];
                ptr[1] = pHeader->palette[colorIndex * 3 + 1];
                ptr[2] = pHeader->palette[colorIndex * 3 + 2];
            }
        }
    }
}

void TextureLoaderPCX::readPCX4bit(headerPCXInfo* pHeader, textureRAWInfo* texture)
{
    u8* colorIndex, * line;
    u8* pLine, * ptr;
    s32 rle_count = 0, rle_value = 0;
    s32 x, y, c;
    s32 bytes;
    
    colorIndex = new u8[texture->Width];
    line = new u8[pHeader->bytesPerScanLine];
    
    for (y = 0; y < texture->Height; ++y)
    {
        ptr = &texture->ImageData[(texture->Height - (y + 1)) * texture->Width * 3];
        
        memset(colorIndex, 0, texture->Width * sizeof(u8));
        
        for (c = 0; c < 4; ++c)
        {
            pLine = line;
            bytes = pHeader->bytesPerScanLine;
            
            // Decode line number y
            while (bytes--)
            {
                if (rle_count == 0)
                {
                    if ((rle_value = fgetc(hFile_)) < 0xc0)
                    {
                        rle_count = 1;
                    }
                    else
                    {
                        rle_count = rle_value - 0xc0;
                        rle_value = fgetc(hFile_);
                    }
                }
                
                --rle_count;
                *(pLine++) = rle_value;
            }
            
            // Compute line's color indexes
            for (x = 0; x < texture->Width; ++x)
            {
                if (line[x / 8] & (128 >> (x % 8)))
                    colorIndex[x] += (1 << c);
            }
        }
        
        // Decode scan line. color index => rgb
        for (x = 0; x < texture->Width; ++x, ptr += 3)
        {
            ptr[0] = pHeader->palette[colorIndex[x] * 3 + 0];
            ptr[1] = pHeader->palette[colorIndex[x] * 3 + 1];
            ptr[2] = pHeader->palette[colorIndex[x] * 3 + 2];
        }
    }
    
    delete [] colorIndex;
    delete [] line;
}

void TextureLoaderPCX::readPCX8bit(headerPCXInfo* pHeader, textureRAWInfo* texture)
{
    s32 rle_count = 0, rle_value = 0;
    u8 palette[768];
    u8 magic;
    u8* ptr;
    fpos_t curpos;
    s32 y, bytes;
    
    // The palette is contained in the last 769 bytes of the file
    fgetpos(hFile_, &curpos);
    fseek(hFile_, -769, SEEK_END);
    magic = fgetc(hFile_);
    
    // First byte must be equal to 0x0c (12)
    if (magic != 0x0C)
    {
        io::printError(io::stringc("PCX file's colormap's first byte must be 0x0C"));
        spDeleteArray(texture->ImageData);
        return;
    }
    
    // Read palette
    fread(palette, sizeof(u8), 768, hFile_);
    fsetpos(hFile_, &curpos);
    
    // Read pixel data
    for (y = 0; y < texture->Height; ++y)
    {
        ptr = &texture->ImageData[(texture->Height - (y + 1)) * texture->Width * 3];
        bytes = pHeader->bytesPerScanLine;
        
        // Decode line number y
        while (bytes--)
        {
            if (rle_count == 0)
            {
                if( (rle_value = fgetc(hFile_)) < 0xC0)
                {
                    rle_count = 1;
                }
                else
                {
                    rle_count = rle_value - 0xc0;
                    rle_value = fgetc(hFile_);
                }
            }
            
            --rle_count;
            
            ptr[0] = palette[rle_value * 3 + 0];
            ptr[1] = palette[rle_value * 3 + 1];
            ptr[2] = palette[rle_value * 3 + 2];
            ptr += 3;
        }
    }
}

void TextureLoaderPCX::readPCX24bit(headerPCXInfo* pHeader, textureRAWInfo* texture)
{
    u8* ptr = texture->ImageData;
    s32 rle_count = 0, rle_value = 0;
    s32 y, c;
    s32 bytes;
    
    for (y = 0; y < texture->Height; ++y)
    {
        for (c = 0; c < 3; ++c)
        {
            ptr = &texture->ImageData[(texture->Height - (y + 1)) * texture->Width * 3];
            bytes = pHeader->bytesPerScanLine;
            
            // Decode line number y
            while (bytes--)
            {
                if (rle_count == 0)
                {
                    if((rle_value = fgetc(hFile_)) < 0xc0)
                    {
                        rle_count = 1;
                    }
                    else
                    {
                        rle_count = rle_value - 0xc0;
                        rle_value = fgetc(hFile_);
                    }
                }
                
                --rle_count;
                ptr[c] = (u8)rle_value;
                ptr += 3;
            }
        }
    }
}


} // /namespace video

} // /namespace sp






// ================================================================================
