/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * Texture loading PNG-fileformat file
 *
 */

#include "spTextureLoaderPNG.hpp"

#include <png.h> // "libpng.a"


namespace sp
{
namespace video
{


TextureLoaderPNG::TextureLoaderPNG(io::stringc &Filename) : TextureLoaderRAW(Filename)
{
}
TextureLoaderPNG::~TextureLoaderPNG()
{
}

textureRAWInfo* TextureLoaderPNG::loadImageData()
{
    /* Check if the file has opened correct */
    if (isFileError())
        return 0;
    
    /* Allocate new texture RAW data & header buffer */
    textureRAWInfo* texture = 0;
    
    png_byte    magic[8];
    png_structp png_ptr;
    png_infop   info_ptr;
    
    s32 bit_depth, color_type;
    png_bytep* row_pointers = 0;
    
    s32 i;
    
    // Read magic number
    fread(magic, 1, sizeof(magic), hFile_);
    
    // Check for valid magic number
    if (!png_check_sig(magic, sizeof(magic)))
    {
        io::printError(io::stringc("PNG file has incorrect magic number"));
        return 0;
    }
    
    // Create a png read struct
    if ( !( png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, 0, 0, 0) ) )
        return 0;
    
    // Create a png info struct
    if ( !( info_ptr = png_create_info_struct(png_ptr) ) )
    {
        png_destroy_read_struct(&png_ptr, 0, 0);
        return 0;
    }
    
    // Create our OpenGL texture object
    texture = new textureRAWInfo();
    
    // Initialize the setjmp for returning properly after a libpng error occured
    if ( setjmp(png_jmpbuf(png_ptr)) )
    {
        fclose(hFile_);
        png_destroy_read_struct (&png_ptr, &info_ptr, 0);
        
        spDeleteArray(row_pointers);
        
        if (texture)
        {
            spDeleteArray(texture->ImageData);
            spDeletePointer(texture);
        }
        
        return 0;
    }
    
    // Setup libpng for using standard C fread() function with our FILE pointer
    png_init_io(png_ptr, hFile_);
    
    // tell libpng that we have already read the magic number
    png_set_sig_bytes(png_ptr, sizeof(magic));
    
    // Read png info
    png_read_info(png_ptr, info_ptr);
    
    // Get some usefull information from header
    bit_depth = png_get_bit_depth(png_ptr, info_ptr);
    color_type = png_get_color_type(png_ptr, info_ptr);
    
    // Convert index color images to RGB images
    if (color_type == PNG_COLOR_TYPE_PALETTE)
        png_set_palette_to_rgb(png_ptr);
    
    // Convert 1-2-4 bits grayscale images to 8 bits grayscale
    if (color_type == PNG_COLOR_TYPE_GRAY && bit_depth < 8)
        png_set_gray_1_2_4_to_8(png_ptr);
    
    if (png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS))
        png_set_tRNS_to_alpha(png_ptr);
    
    if (bit_depth == 16)
        png_set_strip_16(png_ptr);
    else if (bit_depth < 8)
        png_set_packing(png_ptr);
    
    // Update info structure to apply transformations
    png_read_update_info(png_ptr, info_ptr);
    
    // Retrieve updated information
    png_get_IHDR(
        png_ptr, info_ptr,
        (png_uint_32*)(&texture->Width),
        (png_uint_32*)(&texture->Height),
		&bit_depth, &color_type,
		0, 0, 0
    );
    
    // Get image format and components per pixel
    getPNGInfo(color_type, texture);
    
    // We can now allocate memory for storing pixel data
    texture->ImageData = new u8[ texture->Width * texture->Height * texture->InternalFormat ];
    
    // Setup a pointer array. Each one points at the begening of a row
    row_pointers = new png_bytep[texture->Height];
    
    for (i = 0; i < texture->Height; ++i)
    {
        row_pointers[i] = (png_bytep)(texture->ImageData +
	                      ((texture->Height - (i + 1)) * texture->Width * texture->InternalFormat));
    }
    
    // Read pixel data using row pointers
    png_read_image(png_ptr, row_pointers);
    
    // Finish decompression and release memory
    png_read_end(png_ptr, 0);
    png_destroy_read_struct(&png_ptr, &info_ptr, 0);
    
    // We don't need row pointers anymore
    spDeleteArray(row_pointers);
    
    // Flip image data y axis
    TextureManager::flipImageDataY(texture->Width, texture->Height, texture->ImageData, texture->InternalFormat);
    
    return texture;
}

/* ======= Private: ======= */

void TextureLoaderPNG::getPNGInfo(s32 ColorType, textureRAWInfo* TextureInfo)
{
    switch (ColorType)
    {
        case PNG_COLOR_TYPE_GRAY:
            TextureInfo->Format         = TEXTURE_GRAY;
            TextureInfo->InternalFormat = 1;
            break;
        case PNG_COLOR_TYPE_GRAY_ALPHA:
            TextureInfo->Format         = TEXTURE_GRAYALPHA;
            TextureInfo->InternalFormat = 2;
            break;
        case PNG_COLOR_TYPE_RGB:
            TextureInfo->Format         = TEXTURE_RGB;
            TextureInfo->InternalFormat = 3;
            break;
        case PNG_COLOR_TYPE_RGB_ALPHA:
            TextureInfo->Format         = TEXTURE_RGBA;
            TextureInfo->InternalFormat = 4;
            break;
        default: // 'Shit happens'
            break;
    }
}


} // /namespace video

} // /namespace sp






// ================================================================================
