/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * Texture loading TGA-fileformat file
 *
 */

#include "spTextureLoaderTGA.hpp"


namespace sp
{
namespace video
{


TextureLoaderTGA::TextureLoaderTGA(io::stringc &Filename) : TextureLoaderRAW(Filename)
{
}
TextureLoaderTGA::~TextureLoaderTGA()
{
}

textureRAWInfo* TextureLoaderTGA::loadImageData()
{
    /* Check if the file has opened correct */
    if (isFileError())
        return 0;
    
    /* Allocate new texture RAW data & header buffer */
    textureRAWInfo* texture = new textureRAWInfo();
    
    /* Read the header */
    u8 Header[18];
    
    if (fread(&MainHeader_, sizeof(headerTGAInfo), 1, hFile_) == 0)
    {
        io::printError("could not read TGA header");
        spDeletePointer(texture);
        return 0;
    }
    
    if (MainHeader_.ImageType == TGA_IMGDATA_RGB)
    {
        if (!loadUncompressedTGA(texture))
            return 0;
    }
    else if (MainHeader_.ImageType == TGA_IMGDATA_RGB_COMPRESSED)
    {
        if (!loadCompressedTGA(texture))
            return 0;
    }
    else
    {
        io::printError("TGA header type must be 2 or 10");
        spDeletePointer(texture);
        return 0;
    }
    
    return texture;
}

/*
 * ======= Private: =======
 */

bool TextureLoaderTGA::loadUncompressedTGA(textureRAWInfo* texture)
{
    
    u32 BytesPerPixel;                      // Holds Number Of Bytes Per Pixel Used In The TGA File
    u32 ImageSize;                          // Used To Store The Image Size When Setting Aside Ram
    texture->Format         = TEXTURE_RGBA; // Standard format
    texture->InternalFormat = 4;            // Standard internal format
    
    texture->Width  = MainHeader_.ImageWidth;
    texture->Height = MainHeader_.ImageHeight;
    
    if ( texture->Width <= 0 || texture->Height <= 0 || ( MainHeader_.bpp != 24 && MainHeader_.bpp != 32 ) )
    {
        io::printError("invalid TGA header information (size/ bpp)");
        
        spDeletePointer(texture);
        
        return false;
    }
    
    texture->bpp    = MainHeader_.bpp;                              // Grab The TGA's Bits Per Pixel (24 or 32)
    BytesPerPixel   = texture->bpp/8;                               // Divide By 8 To Get The Bytes Per Pixel
    ImageSize       = texture->Width*texture->Height*BytesPerPixel; // Calculate The Memory Required For The TGA Data
    
    texture->ImageData = new u8[ImageSize];                         // Reserve Memory To Hold The TGA Data
    
    if (texture->ImageData == 0 || fread(texture->ImageData, 1, ImageSize, hFile_) != ImageSize)
    {
        io::printError("incorrect TGA image data size");
        
        spDeleteArray(texture->ImageData);
        spDeletePointer(texture);
        
        return false;
    }
    
    // Flip colors
    TextureManager::flipImageColors(texture->Width, texture->Height, texture->ImageData, BytesPerPixel);
    
    if (texture->bpp == 24)
    {
        texture->Format         = TEXTURE_RGB;
        texture->InternalFormat = 3;
    }
    
    return true;
    
}

bool TextureLoaderTGA::loadCompressedTGA(textureRAWInfo* texture)
{
    
    u32 bytesPerPixel;                      // Holds Number Of Bytes Per Pixel Used In The TGA File
    u32 imageSize;                          // Used To Store The Image Size When Setting Aside Ram
    texture->Format         = TEXTURE_RGBA; // Standard format
    texture->InternalFormat = 4;            // Standard internal format
    
    texture->Width  = MainHeader_.ImageWidth;
    texture->Height = MainHeader_.ImageHeight;
    
    if ( texture->Width <= 0 || texture->Height <= 0 || ( MainHeader_.bpp != 24 && MainHeader_.bpp != 32 ) )
    {
        io::printError("invalid TGA file information (size/ bpp)");
        spDeletePointer(texture);
        return false;
    }
    
    texture->bpp    = MainHeader_.bpp;                              // Grab The TGA's Bits Per Pixel (24 or 32)
    bytesPerPixel   = texture->bpp/8;                               // Divide By 8 To Get The Bytes Per Pixel
    imageSize       = texture->Width*texture->Height*bytesPerPixel; // Calculate The Memory Required For The TGA Data
    
    texture->ImageData = new u8[imageSize];                         // Reserve Memory To Hold The TGA Data
    
    // Decompressing operations
    
    u32 pixelcount = texture->Width*texture->Height;    // Number of pixels in the image
    u32 currentpixel = 0;                               // Current pixel being read
    u32 currentbyte = 0;                                // Current byte 
    u8* colorbuffer = new u8[bytesPerPixel];            // Storage for 1 pixel
    
    do
    {
        
		u8 chunkheader = 0;
        
        if (fread(&chunkheader, sizeof(u8), 1, hFile_) == 0)
        {
            io::printError("could not read TGA RLE header");
            
            spDeleteArray(texture->ImageData);
            spDeletePointer(texture);
            
            return false;
        }
        
        if (chunkheader < 128)
        {
            
            ++chunkheader;
            
            for (s16 counter = 0; counter < chunkheader; counter++)
            {
                
                if (fread(colorbuffer, 1, bytesPerPixel, hFile_) != bytesPerPixel)
                {
                    io::printError("could not read TGA file image data");
                    
                    spDeleteArray(colorbuffer);
                    spDeleteArray(texture->ImageData);
                    spDeletePointer(texture);
                    
                    return false;
                }
                
                texture->ImageData[currentbyte		] = colorbuffer[2];
                texture->ImageData[currentbyte + 1	] = colorbuffer[1];
                texture->ImageData[currentbyte + 2	] = colorbuffer[0];
                
                if (bytesPerPixel == 4)
                    texture->ImageData[currentbyte + 3] = colorbuffer[3];
                
                currentbyte += bytesPerPixel;
                ++currentpixel;
                
                if (currentpixel > pixelcount)
                {
                    io::printError("TGA file has to many pixels");
                    
                    spDeleteArray(colorbuffer);
                    spDeleteArray(texture->ImageData);
                    spDeletePointer(texture);
                    
                    return false;
                }
                
            }
            
        }
        else
        {
            
            chunkheader -= 127;
            
            if (fread(colorbuffer, 1, bytesPerPixel, hFile_) != bytesPerPixel)
            {
                io::printError("TGA file is unreadable");
                
                spDeleteArray(colorbuffer);
                spDeleteArray(texture->ImageData);
                spDeletePointer(texture);
                
                return false;
            }
            
            for (s16 counter = 0; counter < chunkheader; ++counter)
            {
                texture->ImageData[currentbyte		] = colorbuffer[2];
                texture->ImageData[currentbyte + 1	] = colorbuffer[1];
                texture->ImageData[currentbyte + 2	] = colorbuffer[0];
                
                if (bytesPerPixel == 4)
                    texture->ImageData[currentbyte + 3] = colorbuffer[3];
                
                currentbyte += bytesPerPixel;
                ++currentpixel;
                
                if (currentpixel > pixelcount)
                {
                    io::printError("TGA file has to many pixels");
                    
                    spDeleteArray(colorbuffer);
                    spDeleteArray(texture->ImageData);
                    spDeletePointer(texture);
                    
                    return false;
                }
            }
            
        }
        
    }
    while (currentpixel < pixelcount);
    
    if (texture->bpp == 24)
    {
        texture->Format         = TEXTURE_RGB;
        texture->InternalFormat = 3;
    }
    
    return true;
    
}


} // /namespace video

} // /namespace sp






// ================================================================================
