/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * Header for texture loading WAD-fileformat
 *
 */

#include "spTextureLoaderWAD.hpp"
#include "spMainDevice.hpp"


namespace sp
{

extern video::VideoDriver* __spVideoDriver;

namespace video
{


TextureLoaderWAD::TextureLoaderWAD()
    : FileLoader_(0), File_(0)
{
    init();
}
TextureLoaderWAD::~TextureLoaderWAD()
{
    clear();
}

std::vector<video::Texture*> TextureLoaderWAD::loadTextureList(io::stringc Filename)
{
    
    /* Reset the loader */
    reset();
    
    /* Information message */
    io::printMessage(io::stringc("Load WAD texture list: \"")+Filename+io::stringc("\""));
    
    /* Open a new file */
    File_ = FileLoader_->openFile(Filename, io::FILE_READ_BINARY);
    
    /* Check if the file has been opened successful */
    if (!File_)
        return TextureList_;
    
    /* Read the header */
    readHeader();
    
    /* Read the texture information */
    readTextureInfo();
    
    /* Read the textures */
    readTextures();
    
    /* Close the file */
    FileLoader_->closeFile(File_);
    
    /* Return the entity */
    return TextureList_;
    
}

void TextureLoaderWAD::init()
{
    /* Allocate a new file loader */
    FileLoader_ = new io::FileLoader();
}

void TextureLoaderWAD::clear()
{
    /* Delete the file loader */
    spDeletePointer(FileLoader_);
}

void TextureLoaderWAD::reset()
{
    /* Clear the lists */
    TextureInfoList_.clear();
    TextureList_.clear();
}

bool TextureLoaderWAD::readHeader()
{
    
    /* Read the header */
    FileLoader_->readStream(File_, &Header_, sizeof(Header_));
    
    /* Check the magic identity */
    c8 Magic[5] = { 0 };
    memcpy(Magic, Header_.Magic, 4);
    
    if (io::stringc(Magic) != "WAD2" && io::stringc(Magic) != "WAD3")
    {
        /* Print an error message */
        io::printError("WAD file has incorrect identity");
        
        /* Exit the function and return false for a failure */
        return false;
    }
    
    /* Exit the function and return true for success */
    return true;
    
}

void TextureLoaderWAD::readTextureInfo()
{
    
    /* Temporary variables */
    STextureWAD Texture;
    s32 i;
    
    /* Set the offset */
    FileLoader_->setFileSeek(File_, Header_.DirOffset);
    
    /* Read all texture information */
    for (i = 0; i < Header_.CountTextures; ++i)
    {
        
        /* Read the vertex data */
        FileLoader_->readStream(File_, &Texture, sizeof(Texture));
        
        /* Add the texture information to the list */
        TextureInfoList_.push_back(Texture);
        
    } // next texture
    
}

void TextureLoaderWAD::readTextures()
{
    
    /* Temporary variables */
    u8* ImageData       = 0;
    u8* Palette         = 0;
    u8* FinalImageData  = 0;
    
    c8 Name[17] = { 0 };
    
    SWallTextureBSP Texture;
    
    video::Texture* FinalTexture = 0;
    
    const s32 PaletteSize = 256 * 3;
    
    s32 ImageDataSize, x, y, j;
    
    /* Loop for all textures */
    for (s32 i = 0; i < Header_.CountTextures; ++i)
    {
        
        /* Set the offset */
        FileLoader_->setFileSeek(File_, TextureInfoList_[i].Offset);
        
        /* Read the BSP information */
        FileLoader_->readStream(File_, &Texture, sizeof(SWallTextureBSP));
        
        /* Get the ANSI-C string name */
        memset(Name, 0, 17);
        memcpy(Name, Texture.Name, 16);
        
        /* Information message */
        io::printMessage(io::stringc("Load WAD texture: \"")+io::stringc(Name)+io::stringc("\""));
        
        /* Set the offset */
        FileLoader_->setFileSeek(File_, TextureInfoList_[i].Offset + Texture.Offset1);
        
        /* Compute the image data size */
        ImageDataSize = Texture.Width * Texture.Height;
        
        /* Allocate new image data for the palette */
        Palette = new u8[PaletteSize];
        
        /* Allocate new image data */
        ImageData = new u8[ImageDataSize];
        
        /* Allocate new final image data */
        FinalImageData = new u8[ImageDataSize*3];
        
        /* Read the image data */
        FileLoader_->readStream(File_, ImageData, ImageDataSize);
        
        /* Set the offset */
        FileLoader_->setFileSeek(
            File_,
            TextureInfoList_[i].Offset + TextureInfoList_[i].UncompressedSize - PaletteSize - 2//4
        );
        
        //io::printMessage(FileLoader_->readValue<s16>(File_));
        
        /* Read the image data for the palette */
        FileLoader_->readStream(File_, Palette, PaletteSize);
        
        /* Loop for image data */
        for (y = 0, j = 0; y < Texture.Height; ++y)
        {
            for (x = 0; x < Texture.Width; ++x, ++j)
            {
                FinalImageData[j*3+0] = Palette[ ImageData[j]*3 + 0 ];
                FinalImageData[j*3+1] = Palette[ ImageData[j]*3 + 1 ];
                FinalImageData[j*3+2] = Palette[ ImageData[j]*3 + 2 ];
            } // next pixel
        } // next scan line
        
        /* Create a new SoftPixel texture */
        FinalTexture = __spVideoDriver->createTexture(
            dim::size2di(Texture.Width, Texture.Height), FinalImageData, video::TEXTURE_RGB, Name
        );
        
        /* Check for special attributes */
        switch (Name[0])
        {
            case '{':
            {
                video::color cr(Palette[255*3+0], Palette[255*3+1], Palette[255*3+2]);
                
                /* Determine which color-key must be used */
                if (cr != video::color(0, 0, 255))
                {
                    video::color cl(Palette[0], Palette[1], Palette[2]);
                    
                    /* Make color-key */
                    if (cl == video::color(255))
                        FinalTexture->setTransFilterAlpha(video::BLENDING_DARKNESS);
                    else
                        FinalTexture->setTransFilterAlpha(video::BLENDING_BRIGHTNESS);
                }
                else
                {
                    /* Make color-key */
                    FinalTexture->setTransFilter(video::color(0, 0, 255, 0));
                    
                    /* Replace the blue color with black */
                    for (y = 0; y < Texture.Height; ++y)
                    {
                        for (x = 0; x < Texture.Width; ++x)
                        {
                            if (FinalTexture->getPixelColor(dim::point2di(x, y)) == video::color(0, 0, 255, 0))
                                FinalTexture->setPixelColor(dim::point2di(x, y), video::color(0, 0, 0, 0));
                        }
                    }
                    
                    /* Update the image data */
                    FinalTexture->updateImageData(video::SET_IMAGEDATA);
                }
            }
            break;
        }
        
        if (io::stringc(Name).left(5) == "glass")
            FinalTexture->setAlphaIntensity(128);
        
        /* Add the final texture to the texture list */
        TextureList_.push_back(FinalTexture);
        
        spDeleteArray(ImageData);
        
        spDeleteArray(Palette);
        
    } // next texture
    
}


} // /namespace video

} // /namespace sp







// ================================================================================
