/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * Header for all video operations (drawing, coloring, textureing)
 *
 */

#ifndef __SP_VIDEODRIVER_H__
#define __SP_VIDEODRIVER_H__


#include "spStandard.hpp"
#include "spVideoDriverColor.hpp"
#include "spVideoDriverTexture.hpp"
#include "spVideoDriverTextureManipulator.hpp"
#include "spVideoDriverMovieManager.hpp"
#include "spVideoDriverGUI.hpp"
#include "spVideoDriverMaterial.hpp"
#include "spVideoDriverFont.hpp"
#include "spTextureManager.hpp"
#include "spTextureLoaderBMP.hpp"
#include "spTextureLoaderJPG.hpp"
#include "spTextureLoaderPNG.hpp"
#include "spTextureLoaderTGA.hpp"
#include "spTextureLoaderPCX.hpp"
#include "spTextureLoaderWAD.hpp"
#include "spTextureSaverBMP.hpp"
#include "spMathSpline.hpp"
#include "spVideoDriverShader.hpp"
#include "spInputOutput.hpp"
#include "spDimension.hpp"

// !!!
#if defined(WIN32)
#   include <windows.h>
#endif


namespace sp
{

namespace scene
{

class Camera;
class Entity;
class SceneManager;

namespace effect
{

class SpecialEffectsActor;

} // /namespace effect

} // /namespace scene

namespace video
{


/*
 * ========== Macros ==========
 */


static const f32 DEF_BEZIER_DETAIL          = 0.025;
static const s32 DEF_TEXTURE_SIZE           = 8;
static const s32 DEF_SCREENSHOT_SIZE        = 256;
static const s32 DEF_FONT_SIZE              = 15;
static const s32 DEF_SPLINE_DETAIL          = 100;
static const f32 DEF_NORMALMAP_AMPLITUDE    = 5.0f;


/*
 * ========== Enumerations ==========
 */


/* Driver options */

enum EDriverTypes
{
    /*
     * This is not a driver but the engine will choose
     * the most suitable video driver for the operating-system
     */
    DRIVER_AUTODETECT,
    
    /*
     * The "Null" video driver
     * Just for debugging
     */
    //DRIVER_NULL,
    
    /*
     * The "Ray - Software Renderer" video driver
     * Own software-renderer
     */
    DRIVER_SOFTWARE,
    
    /*
     * The "OpenGL" video driver
     * For 2D drawing (OpenGL 1.2); for 3D scenes (OpenGL 2.0)
     */
    DRIVER_OPENGL,
    
    /*
     * The "Direct3D9" video driver
     * Only supported under Microsoft(c) Windows(c) ;-)
     */
    DRIVER_DIRECT3D9,
};

/* Matrix options */

enum EMatrixType
{
    MATRIX_PROJECTION,
    MATRIX_MODELVIEW,
    MATRIX_TEXTURE
};

/* Render options */

enum EShadeModeTypes
{
    SHADEMODE_SMOOTH,
    SHADEMODE_FLAT
};

enum EDriverRenderModes
{
    RENDERMODE_FRONT = 0,
    RENDERMODE_BACK,
    RENDERMODE_FRONT_AND_BACK,
    
    RENDERMODE_LINE_LINEAR,
    RENDERMODE_LINE_SMOOTH,
    RENDERMODE_POINT_LINEAR,
    RENDERMODE_POINT_SMOOTH,
    
    RENDERMODE_SOLID = 0,
    RENDERMODE_LINES,
    RENDERMODE_POINTS
};

/* 2D Drawing options */

enum EMaskingTypes
{
    MASKING_SOURCE = 0,
    MASKING_TARGET
};

enum EDrawingModeTypes
{
    DRAW_SOLID = 0,
    DRAW_FRAME
};

/* Font management */

enum EFontTypes
{
    FONT_BOLD       = 0x0001,
    FONT_ITALIC     = 0x0002,
    FONT_UNDERLINED = 0x0004,
    FONT_STRIKEOUT  = 0x0008
};

/* Fog modes */

enum EFogModeTypes
{
    FOG_NONE = 0,
    FOG_STATIC,
    FOG_VOLUMETRIC
};

enum EFogTypes
{
    FOG_PALE,
    FOG_THICK,
};

/* Textrue gen flags */

enum ETextureGenFlags
{
    TEXGEN_RENDERMODE,
    TEXGEN_FILTER,
    TEXGEN_MIPMAPS,     // boolean
    TEXGEN_CLAMPEDGES,  // boolean
    TEXGEN_ANISOTROPY,
};


/*
 * ========== Structures ==========
 */

struct SPrimitiveVertex
{
    SPrimitiveVertex()
        : RHW(0.0f), Color(0xFF000000)
    {
    }
    SPrimitiveVertex(f32 X, f32 Y, f32 Z, video::color Clr, f32 U = 0.0f, f32 V = 0.0f, f32 AspectRHW = 1.0f)
    {
        Coord   .set(X, Y, Z);
        RHW     = AspectRHW;
        Color   = Clr.getSingle();
        TexCoord.set(U, V);
    }
    SPrimitiveVertex(f32 X, f32 Y, f32 Z, u32 Clr, f32 U = 0.0f, f32 V = 0.0f, f32 AspectRHW = 1.0f)
    {
        Coord   .set(X, Y, Z);
        RHW     = AspectRHW;
        Color   = Clr;
        TexCoord.set(U, V);
    }
    
    /* Members */
    dim::vector3df Coord;
    f32 RHW;
    u32 Color;
    dim::point2df TexCoord;
};

struct SFontSettings
{
    SFontSettings(
        s32 defRotationAngle = 0, bool defIsSymbolic = false, bool defIsVolumetric = false,
        f32 defVolumetricDepth = 0.1f, f32 defVolumetricDetail = 0.1f, EDrawingModeTypes defWireframe = DRAW_SOLID)
        : RotationAngle(defRotationAngle), isSymbolic(defIsSymbolic), isVolumetric(defIsVolumetric),
        VolumetricDepth(defVolumetricDepth), VolumetricDetail(defVolumetricDetail), Wireframe(defWireframe)
    {
    }
    ~SFontSettings()
    {
    }
    
    /* Members */
    s32 RotationAngle;
    bool isSymbolic;
    bool isVolumetric;
    f32 VolumetricDepth;
    f32 VolumetricDetail;
    EDrawingModeTypes Wireframe;
};


/*
 * ========== Internal members ==========
 */

#define SIZE 0.1f

const f32 __spBoneBodyCoords[24][3] =
{
    {  0.0f,  0.0f,  0.0f }, {  SIZE,  SIZE,  SIZE }, {  0.0f,  0.0f,  0.0f }, {  SIZE,  SIZE, -SIZE },
    {  0.0f,  0.0f,  0.0f }, { -SIZE,  SIZE, -SIZE }, {  0.0f,  0.0f,  0.0f }, { -SIZE,  SIZE,  SIZE },
    { -SIZE,  SIZE,  SIZE }, {  SIZE,  SIZE,  SIZE }, { -SIZE,  SIZE, -SIZE }, {  SIZE,  SIZE, -SIZE },
    { -SIZE,  SIZE, -SIZE }, { -SIZE,  SIZE,  SIZE }, {  SIZE,  SIZE, -SIZE }, {  SIZE,  SIZE,  SIZE },
    { -SIZE,  SIZE, -SIZE }, {  0.0f,  1.0f,  0.0f }, {  SIZE,  SIZE, -SIZE }, {  0.0f,  1.0f,  0.0f },
    { -SIZE,  SIZE,  SIZE }, {  0.0f,  1.0f,  0.0f }, {  SIZE,  SIZE,  SIZE }, {  0.0f,  1.0f,  0.0f }
};

#undef SIZE



/*
 * ========== VideoDriver parent class ==========
 */


class SP_EXPORT VideoDriver
{
    
    public:
        
        VideoDriver();
        
        virtual ~VideoDriver();
        
        /* Texture effect management are running with the texture manipulator */
        virtual effect::TextureManipulator* getTextureManipulator();
        
        /* Font manager for video sequence rendering */
        virtual movie::MovieManager* getMovieManager();
        
        /* GUI (GraphicsUserInterface) for simple windows-using */
        virtual gui::GUI* getGUI();
        
        /* Deletes the GUI */
        virtual void clearGUI();
        
        /*
         * ========== All video driver functions (in this case: OpenGL functions) ==========
         */
        
        /*
         * === Video driver identity ===
         */
        
        virtual inline EDriverTypes getDriverType() const
        {
            return DriverType_;
        }
        virtual inline HDC getDeviceContext() const
        {
            return hDC_;
        }
        
        virtual io::stringc getRenderer()           const = 0;
        virtual io::stringc getVersion()            const = 0;
        virtual io::stringc getVendor()             const = 0;
        
        virtual bool supportAntiAlias()             const = 0;
        virtual bool supportMultitex()              const = 0;
        virtual bool supportVBO()                   const = 0;
        virtual bool supportAnisotropicFiltering()  const = 0;
        
        virtual s32 getMultitexCount()              const = 0;
        virtual s32 getMaxAnisotropicFilter()       const = 0;
        
        virtual dim::EMatrixProjectionTypes getProjectionMatrixType() const;
        
        virtual void printWarning();
        
        /*
         * === User controll functions ===
         */
        
        /* Video buffer manipulation */
        virtual void clearBuffers();
        virtual void flipBuffers();
        
        
        /*
         * === Setting-/ getting functions ===
         */
        
        virtual void setShadeMode(EShadeModeTypes ShadeMode);
        virtual void setRenderMode(EDriverRenderModes RenderMode);
        virtual void setPolygonMode(EDriverRenderModes RenderFace, EDriverRenderModes RenderMode);
        virtual void setFrontFace(bool isFrontFace);
        
        /* Color manipulation */
        virtual void setClearColor(color Color);
        virtual void setColorMask(bool isRed, bool isGreen, bool isBlue, bool isAlpha = true);
        
        virtual void setVsync(bool isVsync);
        
        virtual void setAntiAlias(bool isAntiAlias);
        
        /*
         * === Context functions (for initializing the engine) ===
         */
        
        virtual void initRenderer() = 0;
        
        virtual void changeDisplaySettings(bool isFullScreen, dim::size2di Dimension, s32 ColorDepth) = 0;
        
        virtual bool initScreen(HWND hWnd, bool isAntiAlias, s32 MultiSamplePixelFormat, s32 ColorDepth) = 0;
        virtual void clearScreen(HWND &hWnd) = 0;
        
        virtual void initAntiAlias(HWND Wnd, bool &isAntiAlias, s32 &MultiSamplePixelFormat, s32 &MultiSampleFormat, s32 ColorDepth);
        
        
        /*
         * ========== Rendering 3D scenes ==========
         */
        
        /* Material states */
        virtual void setMaterialStatesTriangleList(SMeshMaterial &Material, dim::vector3df &Rotation);
        
        virtual void setMaterialStatesSprite(SMeshMaterial &Material);
        
        /* Render 3d objects */
        virtual void drawTriangleList(
            u32 &BufferID, u32* IndicesList, u32 &IndicesCount, std::vector<SSurfaceTexture> &TextureList
        );
        
        virtual void drawTriangleList(u32 &BufferID, u32* IndicesList, u32 &IndicesCount);
        
        virtual void drawSprite(
            Texture* &hTexture, f32 &RotationZ, video::color &Color, dim::vector3df &Scale, dim::point2df &BasePosition
        );
        
        /* Update light */
        virtual void updateLight(
            u32 &LightID, scene::ELightTypes &LightingType, bool &isVolumetric,
            dim::vector3df Direction, f32 &NearSpot, f32 &FarSpot,
            f32 &AttenuationConstant, f32 &AttenuationLinear, f32 &AttenuationQuadratic
        );
        
        /* Vertex-buffer-objects (VBO) */
        virtual void enableVertexBuffer();
        virtual void disableVertexBuffer();
        
        virtual void enableVertexArray();
        virtual void disableVertexArray();
        
        virtual void createVertexBuffer(u32 &BufferID);
        virtual void deleteVertexBuffer(u32 &BufferID);
        virtual void updateVertexBuffer(u32 &BufferID, SMeshVertex3D* pVerticesList, s32 VerticesCount);
        
        /* Disable render states */
        virtual void disableTriangleListStates();
        virtual void disable3DRenderStates();
        virtual void disableTexturing();
        
        virtual void setDefaultAlphaBlending();
        
        virtual void enableBlending();
        virtual void disableBlending();
        
        virtual void updateWireframeMode(s32 &ModeFront, s32 &ModeBack);
        
        /* Lighting */
        virtual void addDynamicLightSource(
            u32 LightID, scene::ELightTypes Type,
            video::color &Diffuse, video::color &Ambient, video::color &Specular,
            f32 AttenuationConstant_, f32 AttenuationLinear_, f32 AttenuationQuadratic_
        );
        
        virtual void setLightStatus(u32 LightID, bool isEnable);
        virtual void setLightColor(u32 LightID, video::color &Diffuse, video::color &Ambient, video::color &Specular);
        
        /* Fog effect */
        virtual void setFogMode(s32 FogMode);
        virtual void setFogColor(video::color Color);
        virtual void setFogRange(f32 Range, f32 NearFog = 1.0f, f32 FarFog = 1000.0f, s32 FogType = FOG_PALE);
        
        /* Stencil buffering */
        virtual void clearStencilBuffer();
        
        virtual void drawStencilShadowVolume(
            const dim::vector3df* pTriangleList, s32 Count, bool ZFailMethod, bool VolumetricShadow
        );
        virtual void drawStencilShadow(video::color Color);
        
        /* Clipping planes */
        virtual void setClipPlane(u32 Index, const dim::plane3df &Plane, bool Enable);
        
        /* Special effects actor */
        virtual scene::effect::SpecialEffectsActor* createSpecialEffectsActor() = 0;
        
        /* Shader effects (GLSL and HLSL) */
        virtual Shader* createShader(EStandardShaderTypes Type);
        
        virtual Shader* loadShader(
            io::stringc FilenameVertex, io::stringc FilenameFragment, bool UseHighLevel = true
        );
        virtual Shader* loadShader(
            io::stringc FilenameVertex, io::stringc FilenamePixel,
            io::stringc EntryPointVertex, io::stringc EntryPointPixel,
            EHLSLVertexShaderTypes CompilerTargetVertex = HLSL_VERTEX_1_1,
            EHLSLPixelShaderTypes CompilerTargetPixel = HLSL_PIXEL_1_1,
            bool UseHighLevel = true
        );
        
        virtual void deleteShader(Shader* &hShader);
        
        virtual void unbindShader();
        
        /* Get list - shader */
        virtual inline std::list<Shader*> getShaderList() const
        {
            return ShaderList_;
        }
        
        
        /*
         * === Stack operation functions ===
         */
        
        virtual void pushViewport();
        virtual void popViewport();
        
        virtual void pushMatrix();
        virtual void popMatrix();
        
        
        /*
         * === Texturing ===
         */
        
        virtual void bindTexture(u32 ID);
        virtual void unbindTexture();
        
        virtual void setTextureSubImage(u32 ID, s32 X, s32 Y, s32 Width, s32 Height, const u8* ImageData);
        
        
        /*
         * ========== Drawing 2D objects ==========
         */
        
        /* === General setting function === */
        
        /* Begin & end 2d drawing */
        virtual void beginDrawing2D(bool isSettingViewport = false);
        virtual void endDrawing2D();
        
        /* Begin & end 3d drawing */
        virtual void beginDrawing3D(bool isLighting = false, bool isFog = false);
        virtual void endDrawing3D();
        
        /* begin & end alpha blending */
        virtual void beginAlphaBlending(u8 Type = BLENDING_BRIGHTNESS);
        virtual void endAlphaBlending();
        
        /* Begin & end masking */
        virtual void beginMasking(u8 Type);
        virtual void endMasking();
        
        /* Begin & end texture drawing */
        virtual void beginTextureDrawing(Texture* hTexture, s32 RenderMode = TEXTURE_SMOOTH, u32 TextureDataType = TEXTURE_RGB);
        virtual void endTextureDrawing(bool isUpdateImageData = true);
        
        /* Drawing rect (scissoring) */
        virtual void beginDrawingRect(dim::rect2di Dimension);
        virtual void beginDrawingRect(dim::point2di Position, dim::size2di Size);
        virtual void endDrawingRect();
        
        /* Drawing area (viewport) */
        virtual void setDrawingArea(dim::rect2di Dimension);
        virtual void setDrawingArea(dim::point2di Position, dim::size2di Size);
        
        /* Drawing shader */
        virtual void beginDrawingShader(Shader* hShader);
        virtual void endDrawingShader();
        
        /* Binding multi textures manual (for Shaders) */
        virtual void bindTexture(Texture* hTexture, s32 Level);
        virtual void unbindTexture(s32 Level);
        
        /* Size of lines & points */
        virtual void setLineSize(s32 Size = 1);
        virtual void setPointSize(s32 Size = 1);
        
        virtual void setSolidMode(bool isSolidMode);
        
        /* === Texture drawing === */
        
        virtual void draw2DImage(
            Texture* hTexture, dim::point2di Position,
            color Color = color(255), bool isTransFilterUsing = true
        );
        
        virtual void draw2DImage(
            Texture* hTexture, dim::rect2di Position, dim::rect2df Clipping = dim::rect2df(0.0f, 0.0f, 1.0f, 1.0f),
            color Color = color(255), bool isTransFilterUsing = true
        );
        
        virtual void draw2DImage(
            Texture* hTexture, dim::point2di Position, f32 Rotation, f32 Radius,
            color Color = color(255), bool isTransFilterUsing = true
        );
        
        virtual void draw2DImage(
            Texture* hTexture,
            dim::rect2di Position,
            dim::rect2df Clipping,
            bool isTransFilterUsing,
            f32 Rotation,
            dim::point2df RotationPoint = dim::point2df(0.5f, 0.5f),
            color lefttopColor = color(255),
            color righttopColor = color(255),
            color rightbottomColor = color(255),
            color leftbottomColor = color(255)
        );
        
        virtual void draw2DImage(
            Texture* hTexture,
            dim::point2di lefttopPosition,
            dim::point2di righttopPosition,
            dim::point2di rightbottomPosition,
            dim::point2di leftbottomPosition,
            dim::point2df lefttopClipping = dim::point2df(0.0f, 0.0f),
            dim::point2df righttopClipping = dim::point2df(1.0f, 0.0f),
            dim::point2df rightbottomClipping = dim::point2df(1.0f, 1.0f),
            dim::point2df leftbottomClipping = dim::point2df(0.0f, 1.0f),
            bool isTransFilterUsing = true,
            color lefttopColor = color(255),
            color righttopColor = color(255),
            color rightbottomColor = color(255),
            color leftbottomColor = color(255)
        );
        
        /* === Primitive drawing === */
        
        virtual color getPixel(dim::point2di Position);
        
        virtual void draw2DPixel(dim::point2di Position, color Color = color(255));
        
        virtual void draw2DLine(dim::point2di Position1, dim::point2di Position2, color Color = color(255));
        virtual void draw2DLine(dim::point2di Position1, dim::point2di Position2, color Color1, color Color2);
        virtual void draw2DLine(dim::point2di Position1, dim::point2di Position2, color Color, s32 DotLength);
        
        virtual void draw2DRectangle(dim::rect2di Position, color Color = color(255), EDrawingModeTypes Type = DRAW_SOLID);
        virtual void draw2DRectangle(
            dim::rect2di Position, color lefttopColor, color righttopColor,
            color rightbottomColor, color leftbottomColor, EDrawingModeTypes Type = DRAW_SOLID
        );
        
        virtual void draw2DRoundRectangle(dim::rect2di Position, dim::size2di CornerRadius, color Color = video::color(255));
        
        virtual void draw2DCircle(dim::point2di Position, s32 Radius, color Color = color(255));
        
        virtual void draw2DEllipse(dim::point2di Position, dim::size2di Radius, color Color = color(255));
        virtual void draw2DEllipse(
            dim::point2di Position, dim::size2di Radius, f32 RotationDegree,
            color Color = color(255), bool isSolid = false
        );
        
        virtual void draw2DPolygon(SPrimitiveVertex* VerticesList, u32 AryCount, EDrawingModeTypes Type = DRAW_SOLID);
        virtual void draw2DPolygonImage(
            Texture* hTexture, SPrimitiveVertex* VerticesList, u32 AryCount, EDrawingModeTypes Type = DRAW_SOLID
        );
        
        /* === Extra drawing functions === */
        
        virtual void draw2DArc(
            dim::point2di Position, dim::size2di Radius, s32 RadialDegree1, s32 RadialDegree2,
            f32 RotationDegree = 0.0f, color Color = color(255)
        );
        
        virtual void draw2DBezier(
            dim::point2di Position1, dim::point2di Position2, dim::point2di Radial1, dim::point2di Radial2,
            color Color = color(255), f32 Detail = DEF_BEZIER_DETAIL
        );
        
        virtual void draw2DSpline(
            math::Spline2D &Spline, s32 Detail = DEF_SPLINE_DETAIL,
            dim::point2di Pos = dim::point2di(), color Color = color(255)
        );
        
        virtual void draw2DFloodFill(dim::point2di Position, color FillColor, color BorderColor = color()); // alpha(0) is someone border
        
        /* === 3D drawing functions === */
        
        virtual void draw3DPixel(dim::vector3df Position, color Color = color(255), scene::Camera* hCamera = 0);
        virtual void draw3DLine(
            dim::vector3df Position1, dim::vector3df Position2,
            color Color = color(255), scene::Camera* hCamera = 0
        );
        virtual void draw3DLine(
            dim::vector3df Position1, dim::vector3df Position2,
            color Color1, color Color2, scene::Camera* hCamera = 0
        );
        virtual void draw3DEllipse(
            dim::vector3df Position, dim::vector3df Rotation = dim::vector3df(),
            dim::size2df Radius = dim::size2df(0.5f), color Color = color(255), scene::Camera* hCamera = 0
        );
        virtual void draw3DTriangle(
            Texture* hTexture, dim::triangle3df Triangle,
            dim::vector3df Position = dim::vector3df(), dim::vector3df Rotation = dim::vector3df(),
            dim::vector3df Scale = dim::vector3df(1.0f), color Color = color(255), scene::Camera* hCamera = 0
        );
        virtual void draw3DBox(
            dim::vector3df Corner1, dim::vector3df Corner2,
            dim::vector3df Position = dim::vector3df(), dim::vector3df Rotation = dim::vector3df(),
            dim::vector3df Scale = dim::vector3df(1.0f), color Color = color(255), scene::Camera* hCamera = 0
        );
        
        virtual void draw3DBone(scene::Entity* Obj, dim::matrix4f &BoneTransformation, scene::Camera* hCamera);
        
        
        /*
         * ========== Texture loading & creating ==========
         */
        
        /* Texture generation options */
        virtual void setTextureGenFlags(
            ETextureRenderModeTypes RenderMode, ETextureFilterTypes FilterMode = TEXTURE_TRILINEAR,
            bool isMipMapping = true, bool isEdgeSmooth = false
        );
        virtual void setTextureGenFlags(ETextureGenFlags Flag, s32 Value);
        
        /* Most important function (loads the format which are integrated) */
        virtual Texture* loadTexture(io::stringc Filename) = 0;
        
        /* Loads a texture list (e.g. for WAD files) */
        virtual std::vector<Texture*> loadTextureList(io::stringc Filename);
        
        /* Saves a texture (only BMP/ DIB format supported) */
        virtual bool saveTexture(Texture* hTexture, io::stringc Filename);
        
        /* Deletes a texture */
        virtual void deleteTexture(Texture* &hTexture);
        
        /* Check the correctness of the texture */
        virtual bool isTexture(Texture* hTexture) = 0;
        
        /* Copy texture */
        virtual Texture* copyTexture(Texture* hTexture) = 0;
        
        /* Create new texture */
        virtual Texture* createTexture(
            dim::size2di Dimension, u32 TextureDataType = TEXTURE_RGB, io::stringc Filename = ""
        ) = 0;
        
        /* Create new texture (with own image data) */
        virtual Texture* createTexture(
            dim::size2di Dimension, u8* ImageData, u32 TextureDataType = TEXTURE_RGB, io::stringc Filename = ""
        ) = 0;
        
        /* Create screenshot */
        virtual Texture* createScreenShot(dim::point2di Position = 0, dim::size2di Size = 0) = 0;
        virtual void createScreenShot(Texture* hTexture, dim::point2di Position = 0);
        
        /* Cube map auto generation */
        virtual void updateCubeMap(Texture* hTexture, dim::vector3df GlobalLocation);
        
        /* Normal map generation */
        virtual void makeNormalMap(Texture* HeightMap, f32 Amplitude = DEF_NORMALMAP_AMPLITUDE);
        
        /* Standard texture creating fill color */
        virtual void setFillColor(video::color Color);
        
        /* Texture list: reloading & clearing */
        
        virtual void clearTextureList();
        virtual void reloadTextureList();
        
        virtual inline std::vector<Texture*> getTextureList() const
        {
            return TextureList_;
        }
        
        
        /*
         * ========== Font loading & drawing ==========
         */
        
        virtual Font* loadFont(
            io::stringc FontName = "",
            dim::size2di FontSize = dim::size2di(__SP_IGNORE, __SP_IGNORE),
            s32 TypeFlag = 0,
            s32 RotationAngle = 0,
            bool isSymbolUsing = false,
            bool is3DSpaceUsing = false,
            f32 FontDepth = 0.1f,
            f32 DetailDeviation = 0.1f,
            EDrawingModeTypes Wireframe = DRAW_SOLID
        ) = 0;
        
        virtual Font* loadFont(
            io::stringc FontName,
            s32 FontSize,
            s32 TypeFlag = 0,
            s32 RotationAngle = 0,
            bool isSymbolUsing = false,
            bool is3DSpaceUsing = false,
            f32 FontDepth = 0.1f,
            f32 DetailDeviation = 0.1f,
            EDrawingModeTypes Wireframe = DRAW_SOLID
        ) = 0;
        
        // !PLANED!
        //virtual Font* loadFont(const io::stringc FontName = "", s32 FontSize = 0, SFontSettings ExSettings = SFontSettings());
        //virtual Font* loadFont(const io::stringc FontName, dim::size2di FontSize, SFontSettings ExSettings = SFontSettings());
        
        virtual Font* loadFont(Texture* hTextureFont, dim::size2di FontSize, dim::size2di ElementCount = dim::size2di(16, 16)) = 0;
        
        virtual void deleteFont(Font* hFont);
        
        /* Using texture for the font (e.g. textured 3d fonts) */
        virtual void lockTexture(Texture* hTexture);
        virtual void unlockTexture();
        
        /* Drawing text */
        virtual void draw2DText(Font* hFont, dim::point2di Position, io::stringc Text, color Color = color(255));
        virtual void draw3DText(
            Font* hFont, io::stringc Text, dim::vector3df Position, dim::vector3df Rotation = dim::vector3df(),
            color Color = color(255), scene::Camera* hCamera = 0
        );
        
        
        /*
         * === Matrix controll ===
         */
        
        /* Update the render matrix (with the SoftPixel-modelview matrix) */
        virtual void updateModelviewMatrix() = 0;
        
        virtual void matrixWorldViewReset();
        
        // !CHANGE TO: setProjectionMatrix etc. ...!
        virtual void setProjectionMatrix(const dim::matrix4f &Matrix);
        virtual void setViewMatrix      (const dim::matrix4f &Matrix);
        virtual void setWorldMatrix     (const dim::matrix4f &Matrix);
        virtual void setTextureMatrix   (const dim::matrix4f &Matrix, u8 TextureLayer = 0);
        virtual void setColorMatrix     (const dim::matrix4f &Matrix);
        
        virtual dim::matrix4f getProjectionMatrix() const;
        virtual dim::matrix4f getViewMatrix      () const;
        virtual dim::matrix4f getWorldMatrix     () const;
        virtual dim::matrix4f getTextureMatrix   (u8 TextureLayer = 0) const;
        virtual dim::matrix4f getColorMatrix     () const;
        
        
        /*
         * === Other renderer option functions ===
         */
        
        virtual void setViewport(s32 X, s32 Y, s32 Width, s32 Height);
        
        virtual void setClippingRange(f32 Near, f32 Far);
        
        
    protected:
        
        friend class Font;
        
        /* === Protected functions === */
        
        VideoDriver(EDriverTypes DriverType);
        
        void updateCubeMapDirection(
            scene::Camera* Cam, dim::matrix4f CamDir, dim::vector3df Location,
            ECubeMapDirections Direction, Texture* CubeMapTexture
        );
        
        virtual void updateCubeMapDirectionTexture(ECubeMapDirections Direction, Texture* CubeMapTexture);
        
        /* === Protected members === */
        
        /* Identity member */
        EDriverTypes DriverType_;
        
        /* Context members */
        HDC hDC_;
        UINT PixelFormat_;
        
        /* Temporary members */
        f32 TempColor_[4];
        
        /* Stack members */
        dim::matrix4f StackMatrix_;
        s32 StackViewport_[4];
        dim::matrix4f Matrix2D_;
        
        /* Object lists */
        std::vector<Texture*> TextureList_;
        std::list<Shader*> ShaderList_;
        
        /* Texture generation */
        u8 StdFillColor_[4];
        
        ETextureRenderModeTypes TexGenRenderMode_;
        ETextureFilterTypes TexGenFilterMode_;
        s32 TexGenAnisotropy_;
        bool TexGenMipMapping_;
        bool TexGenEdgeSmooth_;
        
        /* Render states */
        f32 FogRange_, FogNear_, FogFar_;
        s32 FogType_, FogMode_;
        
        f32 RangeNear_, RangeFar_;
        
        u32 MaxClippingPlanes_;
        
        bool isFrontFace_;
        
};


} // /namespace video

} // /namespace sp


#endif






// ================================================================================
