/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * Header for all the material structure
 *
 */

#ifndef __SP_VIDEODRIVER_MATERIAL_H__
#define __SP_VIDEODRIVER_MATERIAL_H__


#include "spStandard.hpp"
#include "spVideoDriverColor.hpp"
#include "spInputOutputString.hpp"
#include "spVideoDriverShader.hpp"
#include "spVideoDriverTexture.hpp"
#include "spDimension.hpp"


namespace sp
{

namespace scene
{

/*
 * ========== namespace: scene ==========
 */

enum ELightTypes
{
    LIGHT_DIRECTIONAL = 0,
    LIGHT_POINT,
    LIGHT_SPOT,
};

class Entity;


} // /namespace scene

/*
 * ========== namespace: video ==========
 */

namespace video
{


/*
 * Typedefinitions & macros
 */

static const s32 MAX_COUNT_OF_TEXTURES  = 8;
static const u8 TEXTURE_IGNORE          = UCHAR_MAX;

typedef void (*PFNUSERMATERIALPROC)(scene::Entity* Obj);


/*
 * ========== Enumerations ==========
 */

enum EWireframeTypes
{
    WIREFRAME_POINTS = 0,
    WIREFRAME_LINES,
    WIREFRAME_SOLID,
};

/* Face-types (just for rendering) */
enum ERenderFaceTypes
{
    RENDER_FRONT,
    RENDER_BACK,
    RENDER_BOTH
};

/* Face-types (multi using, e.g. for rendering and collision-detection) */
enum EFaceTypes
{
    FACE_FRONT,
    FACE_BACK,
    FACE_BOTH
};

/* Cubemap directions (positive x, negative x, etc.) */
enum ECubeMapDirections
{
    CUBEMAP_POSITIVE_X = 0,
    CUBEMAP_NEGATIVE_X,
    CUBEMAP_POSITIVE_Y,
    CUBEMAP_NEGATIVE_Y,
    CUBEMAP_POSITIVE_Z,
    CUBEMAP_NEGATIVE_Z
};

/* Typical blending types */
enum EDriverBlendTypes
{
    BLEND_NORMAL,
    BLEND_BRIGHT,
    BLEND_DARK,
};

/* Size comparision types */
enum ESizeComparisionTypes
{
    CMPSIZE_NEVER = 0,
    CMPSIZE_EQUAL,
    CMPSIZE_NOTEQUAL,
    CMPSIZE_LESS,
    CMPSIZE_LESSEQUAL,
    CMPSIZE_GREATER,
    CMPSIZE_GREATEREQUAL,
    CMPSIZE_ALWAYS,
};

/* Blending types */
enum EBlendingTypes
{
    BLEND_ZERO = 0,
    BLEND_ONE,
    BLEND_SRCCOLOR,
    BLEND_INVSRCCOLOR,
    BLEND_SRCALPHA,
    BLEND_INVSRCALPHA,
    BLEND_DESTCOLOR,
    BLEND_INVDESTCOLOR,
    BLEND_DESTALPHA,
    BLEND_INVDESTALPHA,
};

/* Texture environment types */
enum ETextureEnvTypes
{
    TEXENV_MODULATE = 0,
    TEXENV_REPLACE,
    TEXENV_ADD,
    TEXENV_ADDSIGNED,
    TEXENV_SUBTRACT,
    TEXENV_INTERPOLATE,
    TEXENV_DOT3,
};

/* Texture coordinates generation types */
enum EMappingGenTypes
{
    MAPGEN_DISABLE = 0,
    MAPGEN_OBJECT_LINEAR,
    MAPGEN_EYE_LINEAR,
    MAPGEN_SPHERE_MAP,
    MAPGEN_NORMAL_MAP,
    MAPGEN_REFLECTION_MAP,
};

/* Texture mapping generation coordiantes */
enum EMappingGenCoords
{
    MAPGEN_NONE = 0x0000,
    MAPGEN_S    = 0x0001,
    MAPGEN_T    = 0x0002,
    MAPGEN_R    = 0x0004,
    MAPGEN_Q    = 0x0008,
};


/*
 * ========== Structures ==========
 */

/* === Material structure (for 3d objects) === */

struct SMeshMaterial
{
    
    SMeshMaterial()
        : hShader(0), UserMaterialBeginProc(0), UserMaterialEndProc(0)
    {
        /* General settings */
        ColorDiffuse    = video::color(255, 255, 255, 255);
        ColorAmbient    = video::color(50, 50, 50, 255);
        ColorSpecular   = video::color(255, 255, 255, 255);
        ColorEmission   = video::color(0, 0, 0, 255);
        
        Shininess       = 128.0f;
        
        isColorMaterial = true;
        isLighting      = true;
        isBlending      = true;
        isZBuffer       = true;
        isFog           = true;
        
        DepthMode       = CMPSIZE_LESSEQUAL;
        AlphaMode       = CMPSIZE_ALWAYS;
        
        BlendSource     = BLEND_SRCALPHA;
        BlendTarget     = BLEND_INVSRCALPHA;
        
        AlphaReference  = 0.0f;
        
        RenderModeFront = WIREFRAME_SOLID;
        RenderModeBack  = WIREFRAME_SOLID;
        RenderFace      = RENDER_FRONT;
    }
    ~SMeshMaterial()
    {
    }
    
    /* Material colors */
    
    video::color ColorDiffuse;
    video::color ColorAmbient;
    video::color ColorSpecular;
    video::color ColorEmission;
    
    /* Modes */
    
    f32 Shininess;
    
    s32 DepthMode;
    s32 AlphaMode;
    
    s32 BlendSource;
    s32 BlendTarget;
    
    f32 AlphaReference;
    
    s32 RenderModeFront;
    s32 RenderModeBack;
    s32 RenderFace;
    
    /* States */
    
    bool isColorMaterial;
    bool isLighting;
    bool isBlending;
    bool isZBuffer;
    bool isFog;
    
    /* Handles */
    
    Shader* hShader;
    
    PFNUSERMATERIALPROC UserMaterialBeginProc, UserMaterialEndProc;
    
};


#if 1

/* Vertex structure (for drawing triangle lists) */

struct SMeshVertex3D
{
    SMeshVertex3D() : Normal(0, 0, 1), Fog(0.0f)
    {
    }
    ~SMeshVertex3D()
    {
    }
    
    /* Coordination (X,Y,Z) */
    dim::vector3df Position;
    
    /* Color (R,G,B,A) */
    video::color Color;
    
    /* Normal (NX,NY,NZ) */
    dim::vector3df Normal;
    
    /* Texture coordinates for multitexturing (U,V) */
    dim::vector3df TexCoord[MAX_COUNT_OF_TEXTURES];
    
    /* Fog aspect */
    f32 Fog;
};


/* Vertex structure (for drawing indexed triangle lists) */

struct SMeshIndexedVertex3D
{
    SMeshIndexedVertex3D() : Normal(0, 0, 1), Color(0xFFFFFFFF)
    {
    }
    SMeshIndexedVertex3D(f32 X, f32 Y, f32 Z, u32 Clr, f32 U = 0.0f, f32 V = 0.0f) : Normal(0, 0, 1)
    {
        Position    = dim::vector3df(X, Y, Z);
        Color       = Clr;
        TexCoord[0] = dim::point2df(U, V);
    }
    
    /* Coordination (X,Y,Z) */
    dim::vector3df Position;
    
    /* Normal */
    dim::vector3df Normal;
    
    /* Color (A,R,G,B) */
    u32 Color;
    
    /* Texture coordinates for multitexturing (U,V) */
    dim::point2df TexCoord[MAX_COUNT_OF_TEXTURES];
};

#else

/* Vertex structure (for drawing triangle lists) */

struct SMeshVertex3D
{
    SMeshVertex3D() : Fog(0.0f), SingleColor(0xFFFFFFFF), Normal(0, 0, 1)
    {
    }
    SMeshVertex3D(f32 X, f32 Y, f32 Z, video::color Clr) : Fog(0.0f), Normal(0, 0, 1)
    {
        Position    = dim::vector3df(X, Y, Z);
        Color       = Clr;
        SingleColor = getColor(Clr);
    }
    SMeshVertex3D(f32 X, f32 Y, f32 Z, u32 Clr, f32 U = 0.0f, f32 V = 0.0f) : Fog(0.0f), Normal(0, 0, 1)
    {
        Position    = dim::vector3df(X, Y, Z);
        SingleColor = Clr;
        TexCoord[0] = dim::point2df(U, V);
    }
    
    /* Coordination (X,Y,Z) */
    dim::vector3df Position;
    
    /* Normal (NX,NY,NZ) */
    dim::vector3df Normal;
    
    /* Color (A,R,G,B) */
    u32 SingleColor;
    
    /* Texture coordinates for multitexturing (U,V) */
    dim::point2df TexCoord[MAX_COUNT_OF_TEXTURES];
    
    /* Color (R,G,B,A) */
    video::color Color;
    
    /* Fog aspect */
    f32 Fog;
};

#endif


/* Texture structure (for 3D surface) */

struct SSurfaceTexture
{
    SSurfaceTexture(Texture* DefTexture = 0)
        : hTexture(DefTexture)
    {
        Matrix.matrixReset();
        TextureLayer        = 0;
        TexMappingCoords    = MAPGEN_NONE;
        TexMappingGen       = MAPGEN_OBJECT_LINEAR;
        TexEnvType          = TEXENV_MODULATE;
    }
    ~SSurfaceTexture()
    {
    }
    
    /* Operators */
    
    bool operator == (const SSurfaceTexture &other) const
    {
        return
            TextureLayer        == other.TextureLayer &&
            hTexture            == other.hTexture &&
            TexMappingCoords    == other.TexMappingCoords &&
            TexMappingGen       == other.TexMappingGen &&
            TexEnvType          == other.TexEnvType;
            //Matrix              == other.Matrix;
    }
    
    /* Members */
    
    u8 TextureLayer;
    
    Texture* hTexture;
    
    s32 TexMappingCoords;
    EMappingGenTypes TexMappingGen;
    ETextureEnvTypes TexEnvType;
    
    dim::matrix4f Matrix;
    
};


} // /namespace video

} // /namespace sp


#endif






// ================================================================================
