/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * Textures
 *
 */

#include "spVideoDriverTexture.hpp"
#include "spTextureManager.hpp"


namespace sp
{
namespace video
{


Texture::Texture()
    : OriginalImageData_(0), ImageData_(0), AnimImageData_(0), DimensionType_(TEXTURE_2D)
{
    init();
}
Texture::Texture(
    u32 ID, io::stringc Filename, dim::size2di Size, u8* ImageData,
    s32 RenderMode, s32 FilterMode, s32 Format, s32 Type, bool isEdgeSmooth, bool isMipMapping)
    : OriginalImageData_(0), ImageData_(0), AnimImageData_(0), DimensionType_(TEXTURE_2D)
{
    /* Initialization */
    init();
    
    /* General settings */
    ID_                 = ID;
    Filename_           = Filename;
    
    Type_               = Type;
    
    OriginalSize_       = Size_ = Size;
    
    RenderMode_         = RenderMode;
    FilterMode_         = FilterMode;
    isEdgeSmooth_       = isEdgeSmooth;
    isMipMapping_       = isMipMapping;
    
    OriginalImageData_  = ImageData;
}
Texture::~Texture()
{
    /* Delete the original image data */
    if (!isHoldImageData_)
        spDeleteArray(OriginalImageData_);
    
    /* Delete the animation */
    deleteAnimation();
    
    /* Delete the used image data */
    spDeleteArray(ImageData_);
}

void Texture::setTransFilter(video::color Color)
{
    
    /* Check if the image is not empty */
    if (!ImageData_ || hasAnimation())
        return;
    
    /* Check if the current type is bad */
    if (InternalFormat_ < 3)
        setType(TEXTURE_RGB);
    
    /* Set the new transparency color */
    TransFilterColor_ = Color;
    
    /* Allocate new image data */
    u8* NewImageData = new u8[ Size_.Width * Size_.Height * (Color.Alpha == 255 ? 3 : 4) ];
    u8 r, g, b;
    
    /* The alpha channel to determine if the transparency shall created or removed */
    if (Color.Alpha == 255)
    {
        /* Loop for the new image data */
        for (s32 y = 0, x; y < Size_.Height; ++y)
        {
            for (x = 0; x < Size_.Width; ++x)
            {
                /* Fill the new image data */
                r = NewImageData[ (y*OriginalSize_.Width+x)*3+0 ] = ImageData_[ (y*Size_.Width+x)*InternalFormat_+0 ];
                g = NewImageData[ (y*OriginalSize_.Width+x)*3+1 ] = ImageData_[ (y*Size_.Width+x)*InternalFormat_+1 ];
                b = NewImageData[ (y*OriginalSize_.Width+x)*3+2 ] = ImageData_[ (y*Size_.Width+x)*InternalFormat_+2 ];
            }
        }
        
        /* Set the internal format & type */
        InternalFormat_ = 3;
        Type_           = TEXTURE_RGB;
    }
    else
    {
        /* Loop for the new image data */
        for (s32 y = 0, x; y < Size_.Height; ++y)
        {
            for (x = 0; x < Size_.Width; ++x)
            {
                /* Fill the new image data */
                r = NewImageData[ (y*Size_.Width+x)*4+0 ] = ImageData_[ (y*Size_.Width+x)*InternalFormat_+0 ];
                g = NewImageData[ (y*Size_.Width+x)*4+1 ] = ImageData_[ (y*Size_.Width+x)*InternalFormat_+1 ];
                b = NewImageData[ (y*Size_.Width+x)*4+2 ] = ImageData_[ (y*Size_.Width+x)*InternalFormat_+2 ];
                NewImageData[ (y*Size_.Width+x)*4+3 ] =
                    r == Color.Red && g == Color.Green && b == Color.Blue ? Color.Alpha : 255;
            }
        }
        
        /* Set the internal format & type */
        InternalFormat_ = 4;
        Type_           = TEXTURE_RGBA;
    }
    
    /* Update the new image data */
    makeNewImageData(NewImageData);
    
}

void Texture::setTransFilter(video::color Color, u8 Variation)
{
    
    /* Check if the image is not empty */
    if (!ImageData_ || hasAnimation())
        return;
    
    /* Check if the current type is bad */
    if (InternalFormat_ < 3)
        setType(TEXTURE_RGB);
    
    /* Set the new transparency color */
    TransFilterColor_ = Color;
    
    /* Allocate new image data */
    u8* NewImageData = new u8[ Size_.Width * Size_.Height * 4 ];
    u8 r, g, b;
    
    /* Loop for the new image data */
    for (s32 y = 0, x; y < Size_.Height; ++y)
    {
        for (x = 0; x < Size_.Width; ++x)
        {
            /* Fill the new image data */
            r = NewImageData[ (y*Size_.Width+x)*4+0 ] = ImageData_[ (y*Size_.Width+x)*InternalFormat_+0 ];
            g = NewImageData[ (y*Size_.Width+x)*4+1 ] = ImageData_[ (y*Size_.Width+x)*InternalFormat_+1 ];
            b = NewImageData[ (y*Size_.Width+x)*4+2 ] = ImageData_[ (y*Size_.Width+x)*InternalFormat_+2 ];
            
            /* Check the variation degree */
            if (math::Abs<s16>(s16(r - Color.Red)) <= Variation &&
                math::Abs<s16>(s16(g - Color.Green)) <= Variation &&
                math::Abs<s16>(s16(b - Color.Blue)) <= Variation)
            {
                NewImageData[ (y*Size_.Width+x)*4+3 ] = Color.Alpha;
            }
            else
            {
                NewImageData[ (y*Size_.Width+x)*4+3 ] = 255;
            }
        }
    }
    
    /* Set the internal format & type */
    InternalFormat_ = 4;
    Type_           = TEXTURE_RGBA;
    
    /* Update the new image data */
    makeNewImageData(NewImageData);
    
}

void Texture::setTransFilter(dim::point2di Pos, u8 Alpha, u8 Variation)
{
    video::color Color(getPixelColor(Pos));
    Color.Alpha = Alpha;
    
    if (Variation == 0)
        setTransFilter(Color);
    else
        setTransFilter(Color, Variation);
}

void Texture::setTransFilterAlpha(u8 Mode)
{
    
    /* Check if the image is not empty */
    if (!ImageData_ || hasAnimation())
        return;
    
    /* Check if the current type is bad */
    if (InternalFormat_ < 3)
        setType(TEXTURE_RGB);
    
    /* Set the new transparency color */
    TransFilterColor_ = video::color(Mode == BLENDING_BRIGHTNESS ? 0 : 255);
    
    /* Allocate new image data */
    u8* NewImageData = new u8[ Size_.Width * Size_.Height * 4 ];
    u8 r, g, b;
    
    if (Mode == BLENDING_BRIGHTNESS)
    {
        
        /* Loop for the new image data */
        for (s32 y = 0, x; y < Size_.Height; ++y)
        {
            for (x = 0; x < Size_.Width; ++x)
            {
                /* Fill the new image data */
                r = NewImageData[ (y*Size_.Width+x)*4+0 ] = ImageData_[ (y*Size_.Width+x)*InternalFormat_+0 ];
                g = NewImageData[ (y*Size_.Width+x)*4+1 ] = ImageData_[ (y*Size_.Width+x)*InternalFormat_+1 ];
                b = NewImageData[ (y*Size_.Width+x)*4+2 ] = ImageData_[ (y*Size_.Width+x)*InternalFormat_+2 ];
                NewImageData[ (y*Size_.Width+x)*4+3 ] = video::color(r, g, b).getBrightness<u8>();
            }
        }
        
    }
    else if (Mode == BLENDING_DARKNESS)
    {
        
        /* Loop for the new image data */
        for (s32 y = 0, x; y < Size_.Height; ++y)
        {
            for (x = 0; x < Size_.Width; ++x)
            {
                /* Fill the new image data */
                r = NewImageData[ (y*Size_.Width+x)*4+0 ] = ImageData_[ (y*Size_.Width+x)*InternalFormat_+0 ];
                g = NewImageData[ (y*Size_.Width+x)*4+1 ] = ImageData_[ (y*Size_.Width+x)*InternalFormat_+1 ];
                b = NewImageData[ (y*Size_.Width+x)*4+2 ] = ImageData_[ (y*Size_.Width+x)*InternalFormat_+2 ];
                NewImageData[ (y*Size_.Width+x)*4+3 ] = video::color(255 - r, 255 - g, 255 - b).getBrightness<u8>();
            }
        }
        
    }
    
    /* Set the internal format & type */
    InternalFormat_ = 4;
    Type_           = TEXTURE_RGBA;
    
    /* Update the new image data */
    makeNewImageData(NewImageData);
    
}

void Texture::setTransFilterMask(Texture* MaskTexture, u8 Mode)
{
    
    /* Check if the image is not empty */
    if (!ImageData_ || !MaskTexture || hasAnimation())
        return;
    
    /* Check if the current type is bad */
    if (InternalFormat_ < 3)
        setType(TEXTURE_RGB);
    
    /* Set the new transparency color */
    TransFilterColor_ = video::color(Mode == BLENDING_BRIGHTNESS ? 0 : 255);
    
    /* Get the mask image data */
    const u8* MaskImageData = MaskTexture->getImageData();
    
    /* Get the mask image size */
    dim::size2di MaskSize(MaskTexture->getSize());
    
    /* Get the internal format of the mask texture */
    s32 MaskInternalFormat = MaskTexture->getFormat();
    
    /* Check if the mask texture type is bad */
    if (MaskInternalFormat < 3)
        MaskTexture->setType(TEXTURE_RGB);
    
    /* Allocate new image data */
    u8* NewImageData = new u8[ Size_.Width * Size_.Height * 4 ];
    u8 r, g, b, a;
    
    /* Loop for the new image data */
    for (s32 y = 0, x, i; y < Size_.Height; ++y)
    {
        for (x = 0; x < Size_.Width; ++x)
        {
            /* Fill the new image data */
            NewImageData[ ( (y*Size_.Width+x) << 2 )+0 ] = ImageData_[ (y*Size_.Width+x)*InternalFormat_+0 ];
            NewImageData[ ( (y*Size_.Width+x) << 2 )+1 ] = ImageData_[ (y*Size_.Width+x)*InternalFormat_+1 ];
            NewImageData[ ( (y*Size_.Width+x) << 2 )+2 ] = ImageData_[ (y*Size_.Width+x)*InternalFormat_+2 ];
            
            /* Get the mask image data */
            if (MaskSize == Size_)
                i = (y * Size_.Width + x) * MaskInternalFormat;
            else
                i = ( (y*MaskSize.Height/Size_.Height)*MaskSize.Width + (x*MaskSize.Width/Size_.Width) ) * MaskInternalFormat;
            
            r = MaskImageData[i+0];
            g = MaskImageData[i+1];
            b = MaskImageData[i+2];
            
            /* Get the alpha channel */
            if (Mode == BLENDING_BRIGHTNESS)
                a = ( r + g + b ) / 3;
            else if (Mode == BLENDING_DARKNESS)
                a = 255 - ( r + g + b ) / 3;
            
            /* Fill the new alpha channel */
            NewImageData[ ( (y*Size_.Width+x) << 2 ) + 3 ] = a;
        }
    }
    
    /* Set the internal format & type */
    InternalFormat_ = 4;
    Type_           = TEXTURE_RGBA;
    
    /* Update the new image data */
    makeNewImageData(NewImageData);
    
}

void Texture::setSize(dim::size2di Size)
{
    /* Check if the image is not empty */
    if (!ImageData_ || Size_ == Size || hasAnimation() || Size.Width <= 0 || Size.Height <= 0)
        return;
    
    /* Delete the old memory */
    spDeleteArray(ImageData_);
    
    /* Update the image data */
    makeFlexibleIdentityImageData();
    
    /* Scale the image data */
    TextureManager::scaleImageData(
        ImageData_, OriginalSize_.Width, OriginalSize_.Height, Size.Width, Size.Height, InternalFormat_
    );
    
    /* Set the new size */
    Size_ = Size;
    
    /* Recreate the image */
    recreateImageData();
}

video::color Texture::getPixelColor(dim::point2di Pos)
{
    /* Check the correct image data range */
    if (ImageData_ && !hasAnimation() &&
        Pos.X >= 0 && Pos.Y >= 0 && Pos.X < Size_.Width && Pos.Y < Size_.Height)
    {
        s32 i = Pos.Y * Size_.Width + Pos.X;
        
        switch (InternalFormat_)
        {
            case 1:
                return video::color(ImageData_[i], ImageData_[i], ImageData_[i]);
            case 2:
                i <<= 1;
                return video::color(ImageData_[i], ImageData_[i], ImageData_[i], ImageData_[i+1]);
            case 3:
                i *= 3;
                return video::color(ImageData_[i], ImageData_[i+1], ImageData_[i+2]);
            case 4:
                i <<= 2;
                return video::color(ImageData_[i], ImageData_[i+1], ImageData_[i+2], ImageData_[i+3]);
        }
    }
    
    return video::color();
}

void Texture::setPixelColor(dim::point2di Pos, video::color Color)
{
    /* Check the correct image data range */
    if (ImageData_ && !hasAnimation() &&
        Pos.X >= 0 && Pos.Y >= 0 && Pos.X < Size_.Width && Pos.Y < Size_.Height)
    {
        s32 i = Pos.Y * Size_.Width + Pos.X;
        
        switch (InternalFormat_)
        {
            case 1:
            {
                ImageData_[i] = (Color.Red + Color.Green + Color.Blue) / 3;
            }
            break;
            
            case 2:
            {
                i <<= 1;
                ImageData_[i] = (Color.Red + Color.Green + Color.Blue) / 3;
                ImageData_[i+1] = Color.Alpha;
            }
            break;
            
            case 3:
            {
                i *= 3;
                ImageData_[i+0] = Color.Red;
                ImageData_[i+1] = Color.Green;
                ImageData_[i+2] = Color.Blue;
            }
            break;
            
            case 4:
            {
                i <<= 2;
                ImageData_[i+0] = Color.Red;
                ImageData_[i+1] = Color.Green;
                ImageData_[i+2] = Color.Blue;
                ImageData_[i+3] = Color.Alpha;
            }
            break;
        }
    }
}

// !!! MUST BE UPDATED (using the TextureManager) !!!
void Texture::mixTexture(Texture* hTexture, ETextureMixTypes Type)
{
    
    /* Check if the image is not empty */
    if (!ImageData_ || hasAnimation())
        return;
    
    /* Get the other texture information */
    u8* OtherImageData      = hTexture->getImageData();
    dim::size2di OtherSize  = hTexture->getSize();
    
    /* Temporary vairables */
    register s32 x, y, i, j, k;
    
    /* Determine the type of texture mixing */
    if (Type == TEXTUREMIX_ADD)
    {
        for (y = 0; y < Size_.Height; ++y)
        {
            for (x = 0; x < Size_.Width; ++x)
            {
                for (i = 0; i < 3; ++i)
                {
                    /* Determine the indices */
                    j = InternalFormat_ * ( y * Size_.Width + x ) + i;
                    k = InternalFormat_ * ( y * OtherSize.Height/Size_.Height*OtherSize.Width + x*OtherSize.Width/Size_.Width ) + i;
                    
                    /* Fill the image data */
                    ImageData_[j] = ( ImageData_[j] + OtherImageData[k] ) / 2;
                }
            }
        }
    }
    else if (Type == TEXTUREMIX_MULTIPLY)
    {
        for (y = 0; y < Size_.Height; ++y)
        {
            for (x = 0; x < Size_.Width; ++x)
            {
                for (i = 0; i < 3; ++i)
                {
                    /* Determine the indices */
                    j = InternalFormat_ * ( y * Size_.Width + x ) + i;
                    k = InternalFormat_ * ( y * OtherSize.Height/Size_.Height*OtherSize.Width + x*OtherSize.Width/Size_.Width ) + i;
                    
                    /* Fill the image data */
                    ImageData_[j] = ( ImageData_[j] * OtherImageData[k] ) / 255;
                }
            }
        }
    }
    
    /* Recreate the image */
    recreateImageData();
}

void Texture::setAlphaIntensity(u8 Alpha, ETextureMixTypes Type)
{
    /* Check the conditions */
    if (!ImageData_ || hasAnimation())
        return;
    
    /* Clamp the type */
    if (InternalFormat_ == 1)
        setType(TEXTURE_GRAYALPHA);
    else if (InternalFormat_ == 3)
        setType(TEXTURE_RGBA);
    
    /* Temporary variables */
    s32 x, y, i = InternalFormat_ - 1;
    
    /* Select the mixing type */
    switch (Type)
    {
        case TEXTUREMIX_ADD:
        {
            /* Modify the alpha channel */
            for (y = 0; y < Size_.Height; ++y)
                for (x = 0; x < Size_.Width; ++x, i += InternalFormat_)
                    ImageData_[i] += Alpha;
        }
        break;
        
        case TEXTUREMIX_SET:
        {
            /* Modify the alpha channel */
            for (y = 0; y < Size_.Height; ++y)
                for (x = 0; x < Size_.Width; ++x, i += InternalFormat_)
                    ImageData_[i] = Alpha;
        }
        break;
        
        case TEXTUREMIX_MULTIPLY:
        {
            /* Modify the alpha channel */
            for (y = 0; y < Size_.Height; ++y)
                for (x = 0; x < Size_.Width; ++x, i += InternalFormat_)
                    ImageData_[i] = u8( (s32)ImageData_[i] * Alpha / 255 );
        }
        break;
    }
    
    /* Recreate the image */
    recreateImageData();
}

void Texture::flipImageData(bool isHorizontal, bool isVertical)
{
    /* Check if the image is not empty */
    if (!ImageData_ || hasAnimation())
        return;
    
    /* Flip image data on the x-axis */
    if (isHorizontal)
        TextureManager::flipImageDataX(Size_.Width, Size_.Height, ImageData_, InternalFormat_);
    
    /* Flip image data on the y-axis */
    if (isVertical)
        TextureManager::flipImageDataY(Size_.Width, Size_.Height, ImageData_, InternalFormat_);
    
    /* Recreate the image */
    recreateImageData();
}

void Texture::flipImageDataX()
{
    /* Check if the image is not empty */
    if (!ImageData_ || hasAnimation())
        return;
    
    /* Flip image data on the x-axis */
    TextureManager::flipImageDataX(Size_.Width, Size_.Height, ImageData_, InternalFormat_);
    
    /* Recreate the image */
    recreateImageData();
}

void Texture::flipImageDataY()
{
    /* Check if the image is not empty */
    if (!ImageData_ || hasAnimation())
        return;
    
    /* Flip image data on the y-axis */
    TextureManager::flipImageDataY(Size_.Width, Size_.Height, ImageData_, InternalFormat_);
    
    /* Recreate the image */
    recreateImageData();
}

void Texture::setMipMapping(bool isMipMapping)
{
    isMipMapping_ = isMipMapping;
    recreateImageData();
}

void Texture::setType(s32 TextureDataType)
{
    /* Check if the image is not empty, has no animation and the type is new */
    if (!ImageData_ || hasAnimation() || Type_ == TextureDataType)
        return;
    
    /* Get the new internal format */
    s32 NewInternalFormat = TextureManager::getInternalFormat(TextureDataType);
    
    /* Get the image data size */
    const s32 ImageDataSize = Size_.Width * Size_.Height * InternalFormat_;
    
    /* Allocate new image memory */
    u8* NewImageData = new u8[ImageDataSize];
    
    /* Copy the image data */
    memcpy(NewImageData, ImageData_, ImageDataSize);
    
    /* Convert the image data */
    TextureManager::convertImageData(NewImageData, Size_.Width, Size_.Height, Type_, TextureDataType);
    
    /* Use the old image data, internal format & type */
    InternalFormat_ = NewInternalFormat;
    Type_           = TextureDataType;
    
    /* Use the new image data */
    makeNewImageData(NewImageData);
}

void Texture::makeRoot()
{
    /* Delete the animation */
    deleteAnimation();
    
    /* Make identity formats */
    Type_           = OrigType_;
    InternalFormat_ = OrigInternalFormat_;
    Size_           = OriginalSize_;
    
    /* Delete the old memory */
    spDeleteArray(ImageData_);
    
    /* Copy the root */
    makeIdentityImageData();
    
    /* Recreate the image */
    recreateImageData();
}

void Texture::addTextureFrame(Texture* NewFrame, ETextureFrameDirection FrameDirection)
{
    /* Check if the image is not empty, has no animation and the same for the new texture frame */
    if (!ImageData_ || hasAnimation() || !NewFrame->getImageData() || NewFrame->hasAnimation())
        return;
    
    /* Temporary variables */
    register s32 x, y, i, j, k;
    
    /* Get the temporary image data size */
    s32 ImageDataSize = Size_.Width * Size_.Height * InternalFormat_;
    
    /* Allocate temporary image data */
    u8* TmpImageData = new u8[ImageDataSize];
    
    /* Copy the image data */
    memcpy(TmpImageData, NewFrame->ImageData_, ImageDataSize);
    
    /* Clamp the texture frame size to this */
    if (FrameDirection == TEXTURE_FRAME_HORIZONTAL && NewFrame->Size_.Height != Size_.Height)
    {
        /* Scale the image data of the texture frame */
        TextureManager::scaleImageData(
            TmpImageData, NewFrame->Size_.Width, NewFrame->Size_.Height,
            NewFrame->Size_.Width, Size_.Height, NewFrame->InternalFormat_
        );
    }
    else if (FrameDirection == TEXTURE_FRAME_VERTICAL && NewFrame->Size_.Width != Size_.Width)
    {
        /* Scale the image data of the texture frame */
        TextureManager::scaleImageData(
            TmpImageData, NewFrame->Size_.Width, NewFrame->Size_.Height,
            Size_.Width, NewFrame->Size_.Height, NewFrame->InternalFormat_
        );
    }
    
    /* Clamp the texture frame format to this */
    if (NewFrame->InternalFormat_ != InternalFormat_)
    {
        /* Convert the image data */
        TextureManager::convertImageData(TmpImageData, Size_.Width, Size_.Height, NewFrame->Type_, Type_);
    }
    
    /* Get the new texture size */
    dim::size2di NewSize;
    dim::point2di Translation;
    
    if (FrameDirection == TEXTURE_FRAME_HORIZONTAL)
    {
        NewSize     = dim::size2di(Size_.Width + NewFrame->Size_.Width, Size_.Height);
        Translation = dim::point2di(Size_.Width, 0);
    }
    else if (FrameDirection == TEXTURE_FRAME_VERTICAL)
    {
        NewSize     = dim::size2di(Size_.Width, Size_.Height + NewFrame->Size_.Height);
        Translation = dim::point2di(0, Size_.Height);
    }
    
    /* Get the new image data size */
    ImageDataSize = NewSize.Width * NewSize.Height * InternalFormat_;
    
    /* Allocate new image data */
    u8* NewImageData = new u8[ImageDataSize];
    
    /* Copy the image data of this texture */
    for (y = 0; y < Size_.Height; ++y)
    {
        for (x = 0; x < Size_.Width; ++x)
        {
            for (i = 0; i < InternalFormat_; ++i)
            {
                j = ( y * NewSize.Width + x ) * InternalFormat_ + i;
                k = ( y * Size_.Width + x ) * InternalFormat_ + i;
                
                NewImageData[j] = ImageData_[k];
            }
        }
    }
    
    /* Copy the image data of the new texture frame */
    for (y = 0; y < NewFrame->Size_.Height; ++y)
    {
        for (x = 0; x < NewFrame->Size_.Width; ++x)
        {
            for (i = 0; i < InternalFormat_; ++i)
            {
                j = ( ( y + Translation.Y ) * NewSize.Width + Translation.X + x ) * InternalFormat_ + i;
                k = ( y * NewFrame->Size_.Width + x ) * InternalFormat_ + i;
                
                NewImageData[j] = TmpImageData[k];
            }
        }
    }
    
    /* Set the new texture size */
    Size_ = NewSize;
    
    /* Use the new image data */
    makeNewImageData(NewImageData);
    
    /* Delete the temporary image data */
    spDeleteArray(TmpImageData);
}


/* === Dimension === */

void Texture::make1Dimensional()
{
    DimensionType_  = TEXTURE_1D;
    Depth_          = 1;
    
    recreateImageData();
}

void Texture::make2Dimensional()
{
    DimensionType_  = TEXTURE_2D;
    Depth_          = 1;
    
    recreateImageData();
}

void Texture::make3Dimensional(s32 Depth)
{
    DimensionType_  = TEXTURE_3D;
    Depth_          = Depth;
    
    recreateImageData();
}

// Cube map order: +x, -x, +y, -y, +z, -z
void Texture::makeCubeMap(bool UseCubMap)
{
    if (UseCubMap)
    {
        DimensionType_  = TEXTURE_CUBEMAP;
        Depth_          = 6;
    }
    else
    {
        DimensionType_  = TEXTURE_2D;
        Depth_          = 1;
    }
    
    recreateImageData();
}


/* === Animation === */

bool Texture::createAnimation(u32 AnimCount)
{
    /* Check if the animation is able to run */
    if (AnimCount <= 0 || Size_.Width % AnimCount || !ImageData_)
        return false;
    
    /* Delete the old animation */
    if (AnimImageData_)
        deleteAnimation();
    
    /* Temporary variables */
    register s32 x, y, i, j, k, l;
    
    s32 AnimWidth = Size_.Width / AnimCount;
    s32 xStart, xEnd;
    s32 ImageDataSize;
    
    /* Set the count of the new animation */
    AnimCount_ = AnimCount;
    
    /* Allocate the animation memory */
    AnimImageData_ = new u8*[AnimCount_];
    
    /* Loop for all animation sequences */
    for (i = 0; i < AnimCount_; ++i)
    {
        /* Calculate the image data size of the current animation sequency */
        ImageDataSize = AnimWidth * Size_.Height * InternalFormat_;
        
        /* Allocate new memory */
        AnimImageData_[i] = new u8[ImageDataSize];
        
        /* Determine the start & end of the x-axis */
        xStart = i*AnimWidth;
        xEnd = (i+1)*AnimWidth;
        
        /* Fill the image data */
        for (y = 0; y < Size_.Height; ++y)
        {
            for (x = xStart; x < xEnd; ++x)
            {
                /* Calculate the indices */
                k = ( y * AnimWidth + ( x - xStart ) ) * InternalFormat_;
                l = ( y * Size_.Width + x ) * InternalFormat_;
                
                /* Loop for all pixel components */
                for (j = 0; j < InternalFormat_; ++j)
                {
                    /* Fill the image data */
                    AnimImageData_[i][k+j] = ImageData_[l+j];
                } // next pixel component
            } // next pixel
        } // next line
    } // next animation sequency
    
    /* Delete the old memory */
    spDeleteArray(ImageData_);
    
    /* Set the new image width */
    Size_.Width = AnimWidth;
    
    /* Use the current animation sequency */
    ImageData_ = AnimImageData_[0];
    
    /* Recreate the image */
    recreateImageData();
    
    /* Exit the function */
    return true;
}

void Texture::deleteAnimation()
{
    /* Check if the animation was created */
    if (!AnimImageData_)
        return;
    
    /* Temporary variables */
    register s32 x, y, i, j, k, l;
    
    s32 AnimWidth = Size_.Width;
    s32 xStart, xEnd;
    s32 ImageDataSize = Size_.Width * AnimCount_ * Size_.Height * InternalFormat_;
    
    /* Allocate the old image data */
    ImageData_ = new u8[ImageDataSize];
    
    /* Loop for all animation sequences */
    for (i = 0; i < AnimCount_; ++i)
    {
        /* Determine the start & end of the x-axis */
        xStart = i*AnimWidth;
        xEnd = (i+1)*AnimWidth;
        
        /* Fill the image data */
        for (y = 0; y < Size_.Height; ++y)
        {
            for (x = xStart; x < xEnd; ++x)
            {
                /* Calculate the indices */
                k = ( y * AnimWidth + ( x - xStart ) ) * InternalFormat_;
                l = ( y * Size_.Width + x ) * InternalFormat_;
                
                /* Loop for all pixel components */
                for (j = 0; j < InternalFormat_; ++j)
                {
                    /* Fill the image data */
                    ImageData_[l+j] = AnimImageData_[i][k+j];
                } // next pixel component
            } // next pixel
        } // next line
        
        /* Delete the old memory */
        spDeleteArray(AnimImageData_[i]);
    } // next animation sequency
    
    /* Set the old image width */
    Size_.Width *= AnimCount_;
    
    /* Delete the old memory */
    spDeleteArray(AnimImageData_);
    
    /* Clear the anim count */
    AnimCount_ = 0;
    
    /* Recreate the image */
    recreateImageData();
}

void Texture::setAnimation(u32 Frame)
{
    /* Check if the animation was created */
    if (!AnimImageData_ || Frame >= AnimCount_)
        return;
    
    /* Use the current animation sequency */
    ImageData_ = AnimImageData_[Frame];
    
    /* Recreate the image */
    recreateImageData();
}

/*
 * ======= Protected: =======
 */

void Texture::init()
{
    /* Identification */
    ID_                 = 0;
    Filename_           = "";
    Depth_              = 1;
    
    /* Internal format & type */
    Type_               = TEXTURE_RGB;
    
    /* Render mode (default smooth) */
    RenderMode_         = TEXTURE_SMOOTH;
    
    /* Filter mode (default Trilinear) */
    FilterMode_         = TEXTURE_TRILINEAR;
    AnisotropicSamples_ = 1;
    
    /* Transparency filter color */
    TransFilterColor_   = video::color(0, 0, 0, 255);
    
    /* Animation */
    AnimCount_          = 0;
    
    /* Other default values */
    isHoldImageData_    = false;
    isEdgeSmooth_       = false;
    isMipMapping_       = true;
}

void Texture::makeIdentityImageData()
{
    /* Get the image data size */
    const s32 ImageDataSize = OriginalSize_.Width * OriginalSize_.Height * InternalFormat_;
    
    /* Allocate the used image data */
    ImageData_ = new u8[ImageDataSize];
    
    /* Copy the original image data to the used image data */
    memcpy(ImageData_, OriginalImageData_, ImageDataSize);
}

void Texture::makeFlexibleIdentityImageData()
{
    /* Get the image data size */
    const s32 ImageDataSize = OriginalSize_.Width * OriginalSize_.Height * OrigInternalFormat_;
    
    /* Allocate new image memory */
    u8* NewImageData = new u8[ImageDataSize];
    
    /* Copy the image data */
    memcpy(NewImageData, OriginalImageData_, ImageDataSize);
    
    /* Convert the image data */
    TextureManager::convertImageData(NewImageData, OriginalSize_.Width, OriginalSize_.Height, OrigType_, Type_);
    
    /* Delete the old image data */
    spDeleteArray(ImageData_);
    
    /* Use the new image data */
    ImageData_ = NewImageData;
}

void Texture::recreateImageData() { }

void Texture::makeNewImageData(u8* ImageData)
{
    /* Delete the old memory */
    spDeleteArray(ImageData_);
    
    /* Use the new memory */
    ImageData_ = ImageData;
    
    /* Recreate the image data */
    recreateImageData();
}



} // /namespace video

} // /namespace sp







// ================================================================================
