/*
 * **********************************
 *                                   *
 * HLC Games, Spiele und Software     *
 *                                     *
 * (c) Copyright 2003 - Lukas Hermanns  *
 *                                       *
 * ****************************************
 *  ___    ___    ___          _____
 * /\  \  /\  \  /\  \        /  ___\
 * \ \  \_\_\  \ \ \  \      /\  \__/
 *  \ \   ___   \ \ \  \     \ \  \
 *   \ \  \__/\  \ \ \  \____ \ \  \___
 *    \ \__\ \ \__\ \ \______\ \ \_____\
 *     \/__/  \/__/  \/______/  \/_____/
 *
 * ****************************************
 *
 * ********************** HLC SoftPixelEngine - (3.1.2008) ************************
 *
 * (c) Copyright 2008 - Lukas Hermanns
 *
 * Header for textures
 *
 */

#ifndef __SP_VIDEODRIVER_TEXTURE_H__
#define __SP_VIDEODRIVER_TEXTURE_H__


#include "spStandard.hpp"
#include "spDimension.hpp"
#include "spInputOutputString.hpp"
#include "spVideoDriverColor.hpp"
#include "spTextureManager.hpp"
#include "spMath.hpp"


namespace sp
{
namespace video
{


#define __SP_NONE_TEXTURE       0

enum ETextureRenderModeTypes
{
    TEXTURE_LINEAR,
    TEXTURE_SMOOTH,
};

enum ETextureFilterTypes
{
    TEXTURE_BILINEAR,
    TEXTURE_TRILINEAR,
    TEXTURE_ANISOTROPIC,
};

enum ETextureMixTypes
{
    TEXTUREMIX_ADD = 0,
    TEXTUREMIX_SET,
    TEXTUREMIX_MULTIPLY,
};

enum EAlphaBlendingTypes
{
    BLENDING_BRIGHTNESS = 0,
    BLENDING_DARKNESS
};

enum EUpdateDataFunc
{
    // Get the image data from the renderer (renderer texture -> image data)
    GET_IMAGEDATA,
    // Recreate the texture (image data -> renderer texture)
    SET_IMAGEDATA
};

enum ETextureDimensions
{
    TEXTURE_1D = 1,
    TEXTURE_2D,
    TEXTURE_3D,
    TEXTURE_CUBEMAP,
};

enum ETextureFrameDirection
{
    TEXTURE_FRAME_HORIZONTAL,
    TEXTURE_FRAME_VERTICAL,
};


/*
 * Texture class
 * =============
 * Textures which can be 1Dimensional (for CelShading),
 * 2Dimensional (for normal texturing), 3Dimensional (for terrains or highfields)
 * and Cubic Environment Maps (for environmental reflections)
 */

class SP_EXPORT Texture
{
    
    public:
        
        Texture();
        Texture(
            u32 ID, io::stringc Filename, dim::size2di Size, u8* ImageData,
            s32 RenderMode = TEXTURE_SMOOTH, s32 FilterMode = TEXTURE_TRILINEAR,
            s32 Format = 3, s32 Type = TEXTURE_RGB, bool isEdgeSmooth = false, bool isMipMapping = true
        );
        virtual ~Texture();
        
        virtual inline io::stringc getFilename() const
        {
            return Filename_;
        }
        virtual inline dim::size2di getOriginalSize() const
        {
            return OriginalSize_;
        }
        
        /* Transfilter (color key) */
        virtual void setTransFilter(video::color Color); // Normal color key
        virtual void setTransFilter(video::color Color, u8 Variation); // Using variation
        virtual void setTransFilter(dim::point2di Pos, u8 Alpha = 0, u8 Variation = 0); // Using the color of specified coordinate
        virtual void setTransFilterAlpha(u8 Mode = BLENDING_BRIGHTNESS); // Like the alpha blending effect
        virtual void setTransFilterMask(Texture* MaskTexture, u8 Mode = BLENDING_BRIGHTNESS);
        
        virtual inline video::color getTransFilter() const
        {
            return TransFilterColor_;
        }
        
        /* Image size */
        virtual void setSize(dim::size2di Size);
        
        virtual inline dim::size2di getSize() const
        {
            return Size_;
        }
        virtual inline s32 getDepth() const
        {
            return Depth_;
        }
        
        /* Pixel color manipulation */
        
        video::color getPixelColor(dim::point2di Pos);
        void setPixelColor(dim::point2di Pos, video::color Color);
        
        /* Extra option functions */
        
        virtual void mixTexture(Texture* hTexture, ETextureMixTypes Type = TEXTUREMIX_ADD); // Mixing two textures
        
        virtual void changeAttributes(s32 RenderMode, s32 FilterMode = TEXTURE_TRILINEAR, bool isEdgeSmooth = false) = 0;
        
        virtual void setColorIntensity(f32 Red, f32 Green, f32 Blue) = 0;
        
        virtual void setAlphaIntensity(u8 Alpha, ETextureMixTypes Type = TEXTUREMIX_SET);
        
        virtual void flipColors() = 0;
        
        virtual void setType(s32 TextureDataType);
        
        virtual void makeRoot();
        
        virtual void addTextureFrame(Texture* NewFrame, ETextureFrameDirection FrameDirection = TEXTURE_FRAME_HORIZONTAL);
        
        /* Dimension */
        
        virtual void make1Dimensional();
        virtual void make2Dimensional();
        virtual void make3Dimensional(s32 Depth);
        virtual void makeCubeMap(bool UseCubMap = true);
        
        virtual inline ETextureDimensions getDimensionType() const
        {
            return DimensionType_;
        }
        
        /* Image data updating and recreating */
        
        virtual void flipImageData(bool isHorizontal, bool isVertical);
        virtual void flipImageDataX();
        virtual void flipImageDataY();
        
        virtual void setMipMapFilter(s32 MinFilter, s32 MagFilter, bool isMipMapping = true) = 0;
        
        virtual void setMipMapping(bool isMipMapping);
        virtual bool getMipMapping() const
        {
            return isMipMapping_;
        }
        
        // !PLANED!
        // Exchanges the original image data with the new image data (size, type etc. too)
        //virtual void useNewImageData(dim::size2di Size, ETextureImageDataTypes Type, const u8* ImageData);
        
        // Update the image buffer (renderer texture -> image data / image data - > renderer texture)
        virtual void updateImageData(EUpdateDataFunc UpdateDataFunc) = 0;
        
        /* Binding & unbinding */
        
        virtual void bind(s32 Level = 0) = 0;
        virtual void unbind(s32 Level = 0) = 0;
        
        /* Animation */
        
        virtual bool createAnimation(u32 AnimCount);
        
        virtual void deleteAnimation();
        
        virtual void setAnimation(u32 Frame);
        
        virtual inline bool hasAnimation() const
        {
            return AnimImageData_ != 0;
        }
        
        /* Extra inline functions */
        
        virtual inline u32 getID() const
        {
            return ID_;
        }
        
        virtual inline u8* getImageData() const
        {
            return ImageData_;
        }
        virtual inline u8* getOriginalImageData() const
        {
            return OriginalImageData_;
        }
        
        virtual inline s32 getRenderMode() const
        {
            return RenderMode_;
        }
        virtual inline s32 getFilterMode() const
        {
            return FilterMode_;
        }
        virtual inline s32 getFormat() const
        {
            return InternalFormat_;
        }
        virtual inline s32 getType() const
        {
            return Type_;
        }
        
        virtual inline void setAnisotropicSamples(s32 Samples)
        {
            AnisotropicSamples_ = Samples;
        }
        virtual inline s32 getAnisotropicSamples() const
        {
            return AnisotropicSamples_;
        }
        
        virtual inline void setHoldImageData(bool isHoldImageData)
        {
            isHoldImageData_ = isHoldImageData;
        }
        
    protected:
        
        /* === Functions === */
        
        /* Initializes the s.p. texture */
        virtual void init();
        
        /* Copies the original image data into the current image data */
        virtual void makeIdentityImageData();
        
        /*
         * Copies the original image data into the current image data
         * (in dependet of the current type)
         */
        virtual void makeFlexibleIdentityImageData();
        
        /* Recreates the renderer texture */
        virtual void recreateImageData();
        
        /* Replaces the new image memory with the old */
        virtual void makeNewImageData(u8* ImageData);
        
        /* === Members === */
        
        /* Identification */
        u32 ID_;
        s32 InternalFormat_, Type_;
        s32 OrigInternalFormat_, OrigType_;
        
        /* Name & dimension */
        io::stringc Filename_;
        dim::size2di OriginalSize_, Size_;
        ETextureDimensions DimensionType_;
        s32 Depth_;
        
        /* Render modes */
        s32 RenderMode_, FilterMode_;
        bool isEdgeSmooth_, isMipMapping_;
        video::color TransFilterColor_;
        
        /* Extra options */
        bool isHoldImageData_;
        s32 AnisotropicSamples_;
        
        /* Animation */
        s32 AnimCount_;
        
        /*
         * Used image data (pixel buffer)
         * This image data represents the current texture pixel buffer
         * (normal, resized or with manipulated surface)
         */
        u8* ImageData_;
        
        /*
         * Original image data (original size, color etc. only changed by allocating a new texture)
         * This image data will never deleted (only when deleting the texture or
         * receiving the image data from the renderer texture)
         */
        u8* OriginalImageData_;
        
        /*
         * Animated image data (default 0 (null))
         * This image data is able to animate the texture.
         * The current animation sequence will be represented in the "ImageData_" memory
         * (Width*Height*Format * AnimCount)
         */
        u8** AnimImageData_;
        
};


} // /namespace video

} // /namespace sp


#endif






// ================================================================================
